//
//  Project  : Navier-Stokes
//  File     : convDiff.cc
//  Author   : Ronald Kriemann / Sabine Gutsch Thomas Probst
//  Purpose  : test for discretization and multigrid solution
//             of a convection-diffusion problem; 
//             comparison with exact solution
//
// arch-tag: cbb14606-301a-40dc-94e1-6fff8758889d
//

#include <stdio.h>

#include "timer.hh"
#include "list.hh"
#include "cmdline.hh"

#include "vertex.hh"
#include "regular.hh"
#include "net.hh"

#include "secondNF.hh"
#include "cgSolver.hh"
#include "bicgStab.hh"
#include "kaczmarz.hh"
#include "blocksolver.hh"
#include "schur.hh"
#include "fastGaussSeidel.hh"
#include "TSOR.hh"
#include "TGMRES.hh"

#include "convDiffDoFManager.hh"
#include "convDiffFE.hh"
#include "convDiffHierarchy.hh"
#include "convDirections.hh"

#include "blockmatrix.hh"

#include "drawapp.hh"

#include "graphtool.hh"
#include "rcm.hh"
#include "fvs.hh"

//
// forward declarations
//

static void orderAndSolve ( T2DBasisEqn           * pEqn,
                            TNetHierarchy<2>      * pNetHier,
                            TBlockMatrix         ** pA,
                            TBlockVector<TFloat> ** pRHS,
                            TBlockVector<TFloat> ** pX,
                            T2DConvDiffFE         * pFE );

static void write_sparse ( TMatrixBlock & A, const char * filename );

//
// exact solution: e := x^2 + y^2
// rhs: L e = 
//      -Laplace e + b0 d/dx e + b1 d/dy e = -4 + z0 * 2 * x + z1 * 2 * y
//

class TMyConvDiffEqn : public T2DBasisEqn {

public:
    TMyConvDiffEqn ( ) 
            : T2DBasisEqn ( ) { };
    virtual TFloat f0 (const TVector<2> &v) {
        return ( -4 * epsilon + (b0(v) * 2 * v[0] + b1(v) * 2 * v[1] ) ); 
    }
    virtual TFloat dirichlet0 ( const TVector<2> &v) { 
        return ( v[0] * v[0] + v[1] * v[1] );
    }
};

int 
main( int argc, char **argv ) 
{
    uint   i = 0;
    TTimer timer;
   
    param.noOfNets    = 6;
    param.startLevel  = 1;
    param.ordering    = CONCENTRIC_MIN;
    param.xi          = 1.5;
    param.mgSort      = MGSORT_INDEP; 
    param.criterion   = 7;
    param.epsilon     = 1e-5;
    param.draw        = FALSE;
    param.steps       = 400;
    param.solver      = "bicg";
    param.prec        = "sgs";
    param.convdir     = "circle";
    param.starttri    = "grids/quad1.tri";

    param.evalCmdLine( argc, argv );

    param.endLevel = param.noOfNets-1;
    cout << "main : chosen parameter = " << param << endl;
    //param.output();

    //if (param.ordering == RCM) param.xi = 0.0;

    //
    // build and refine net
    //

    timer.start();
    
    T2DNetHierarchy * pNetHier = new T2DNetHierarchy( param.starttri );
    pNetHier->delReference();
    
    for (i = 0; i < param.noOfNets-1; i++)
        pNetHier->buildNet();

    timer.stop();
    cout << "main : time for refining = " << timer << endl;
    cout << "main : used memory = " << TMemory::used() << " kB" << endl;

    if (param.draw)
        TDrawApp::draw( * pNetHier->getList().getFirst(), "net" );
    
    //
    // build stiffness matrix sparsity structure
    //

    timer.start();
    
    TBlockMatrix         ** pA         = new TBlockMatrix * [param.noOfNets];;
    TBlockVector<TFloat> ** pRHS       = new TBlockVector<TFloat> * [param.noOfNets];
    TBlockVector<TFloat> ** pX         = new TBlockVector<TFloat> * [param.noOfNets];

    T2DConfDiffDoFManager * dofManager = new T2DConfDiffDoFManager();
    TMyConvDiffEqn        * pEqn       = new TMyConvDiffEqn();

    // set epsilon
    pEqn->setEpsilon( param.epsilon );
    
    // set reaction term
    if (param.reactionTerm)
        pEqn->setReacFunction( oneReac ); 
    else 
        pEqn->setReacFunction( noReac ); 

    // Pointer to functions for convection direction
    if (param.convdir == "none")
    {
        pEqn->setConvFunctions( noConv0, noConv1 );
    }// if
    else if (param.convdir == "xline")
    {
        pEqn->setConvFunctions( xline0, xline1 );
   
        if (param.draw)
            TDrawApp::drawFlow( pEqn, 1.0/10.0, 1.0/15.0, "flow" );
    }// if
    else if (param.convdir == "curve")
    {
        pEqn->setConvFunctions( curve0, curve1 );
   
        if (param.draw)
            TDrawApp::drawFlow( pEqn, 1.0/15.0, 1.0/15.0, "flow" );
    }// if
    else if (param.convdir == "circle")
    {
        pEqn->setConvFunctions( circle0, circle1 ); 
   
        if (param.draw)
            TDrawApp::drawFlow( pEqn, 1.0/15.0, 1.0/15.0, "flow" );
    }// if
    else if (param.convdir == "square") 
    {
        pEqn->setConvFunctions( square0, square1 );    
   
        if (param.draw)
            TDrawApp::drawFlow( pEqn, 1.0/15.0, 1.0/15.0, "flow" );
    }// if
    else if (param.convdir == "foursquares") 
    {
        pEqn->setConvFunctions( foursquares0, foursquares1 ); 
   
        if (param.draw)
            TDrawApp::drawFlow( pEqn, 1.0/15.0, 1.0/15.0, "flow" );
    }// if
    else if (param.convdir == "4circles") 
    {
        pEqn->setConvFunctions( fourcircles0, fourcircles1 );
   
        if (param.draw)
            TDrawApp::drawFlow( pEqn, 1.0/15.0, 1.0/15.0, "flow" );
    }// if
    else
    {
        cout << "main (...) : convdir unsupported" << endl;
        exit( 1 );
    }// else
   
    T2DConvDiffFE        * fe        = new T2DConvDiffFE( Tabata );
    T2DConvDiffHierarchy * hierarchy = new T2DConvDiffHierarchy();

    hierarchy->buildHierarchy( dofManager, pEqn, fe, pNetHier, pA,
                               pRHS, pX );

    timer.stop();

    cout << "main : time for disretization = " << timer << endl;
    cout << "main : used memory = " << TMemory::used() << " kB" << endl;

    cout << "main : dim of coarse net = " << (*pA[param.startLevel])[0][0].getSize() << endl;
    cout << "main : dim of fine   net = " << (*pA[param.endLevel])[0][0].getSize() << endl;

    write_sparse( (*pA[param.endLevel])[0][0], "matrix.txt" );
    return 0;
    
    /////////////////////////////////////
    //
    // reorder and solve system
    //

    if ( param.ordering == ORDER_ALL )
    {
        cout << "##################################################" << endl
             << "#"                                                  << endl
             << "#  ordering = NONE"                                 << endl
             << "#"                                                  << endl
             << "##################################################" << endl;
        param.ordering = NONE;
        orderAndSolve( pEqn, pNetHier, pA, pRHS, pX, fe );
        cout << "main : used memory = " << TMemory::used() << " kB" << endl;

        cout << "##################################################" << endl
             << "#"                                                  << endl
             << "#  ordering = RCM"                                  << endl
             << "#"                                                  << endl
             << "##################################################" << endl;
        param.ordering = RCM;
        orderAndSolve( pEqn, pNetHier, pA, pRHS, pX, fe );
        cout << "main : used memory = " << TMemory::used() << " kB" << endl;

        cout << "##################################################" << endl
             << "#"                                                  << endl
             << "#  ordering = pFVS"                                 << endl
             << "#"                                                  << endl
             << "##################################################" << endl;
        param.ordering = PLANAR_FVS;
        param.pblock   = false;
        orderAndSolve( pEqn, pNetHier, pA, pRHS, pX, fe );
        cout << "main : used memory = " << TMemory::used() << " kB" << endl;
        
        cout << "##################################################" << endl
             << "#"                                                  << endl
             << "#  ordering = pFVS (p)"                             << endl
             << "#"                                                  << endl
             << "##################################################" << endl;
        param.ordering = PLANAR_FVS;
        param.pblock   = true;
        orderAndSolve( pEqn, pNetHier, pA, pRHS, pX, fe );
        cout << "main : used memory = " << TMemory::used() << " kB" << endl;

        if ((param.convdir == "circle") || (param.convdir == "4circles"))
        {
            cout << "##################################################" << endl
                 << "#"                                                  << endl
                 << "#  ordering = CYC"                                  << endl
                 << "#"                                                  << endl
                 << "##################################################" << endl;
            param.ordering = CONCENTRIC_MIN;
            param.pblock   = false;
            orderAndSolve( pEqn, pNetHier, pA, pRHS, pX, fe );
            cout << "main : used memory = " << TMemory::used() << " kB" << endl;

            cout << "##################################################" << endl
                 << "#"                                                  << endl
                 << "#  ordering = CYC (p)"                              << endl
                 << "#"                                                  << endl
                 << "##################################################" << endl;
            param.ordering = CONCENTRIC_MIN;
            param.pblock   = true;
            orderAndSolve( pEqn, pNetHier, pA, pRHS, pX, fe );
            cout << "main : used memory = " << TMemory::used() << " kB" << endl;

            cout << "##################################################" << endl
                 << "#"                                                  << endl
                 << "#  ordering = hFVS"                                 << endl
                 << "#"                                                  << endl
                 << "##################################################" << endl;
            param.ordering = HEURISTIC_FVS;
            param.pblock   = false;
            orderAndSolve( pEqn, pNetHier, pA, pRHS, pX, fe );
            cout << "main : used memory = " << TMemory::used() << " kB" << endl;
        }// if
    }// if
    else
        orderAndSolve( pEqn, pNetHier, pA, pRHS, pX, fe );
    
    /////////////////////////////////////
    //
    // clean up
    //

    delete pNetHier;
    
    delete dofManager;
    delete pEqn;
    delete fe;
    delete hierarchy;

    for (i = 0; i < param.noOfNets; i++)
    {
        if (pX[i]) delete pX[i];
        if (pRHS[i]) delete pRHS[i];
        if (pA[i]) delete pA[i];
    }// for
    
    TMemory::statistics();
}

//
// reorder system and solve with given solver
//

static void
orderAndSolve ( T2DBasisEqn           * pEqn,
                TNetHierarchy<2>      * pNetHier,
                TBlockMatrix         ** pA,
                TBlockVector<TFloat> ** pRHS,
                TBlockVector<TFloat> ** pX,
                T2DConvDiffFE         * pFE )
{
    uint     i;
    TTimer   timer;

    timer.start();

    TBlockInfo  * blockinfo = new TBlockInfo;
    
    blockinfo->setSortMethod( SORT_OFD );

    if (param.pblock)
        blockinfo->setBlocking( TRUE );
    
    param.startLevel = param.endLevel;
    
    //
    // build data structures for each net
    //
    TVectorBlock<uint>  *vecPerm[param.noOfNets];
    TVectorBlock<uint>  *vecInvPerm[param.noOfNets];
    TVectorBlock<uint>  *cyclicBlocks[param.noOfNets];
    TVectorBlock<uint>  *arrFVSize = new TVectorBlock<uint>();
 
    if ( param.ordering != NONE )
    {
        for ( i = 0; i < param.noOfNets; i++ )
        {
            vecPerm[i] = vecInvPerm[i] = cyclicBlocks[i] = NULL;
        }// for
        
        for ( i = param.startLevel; i <= param.endLevel; i++ )
        {
            vecPerm[i]      = new TVectorBlock<uint>;
            cyclicBlocks[i] = new TVectorBlock<uint>;
        }// for
    }// if

    //
    // now build permutation vectors
    //
    if ( param.ordering & CONCENTRIC )
    {
        if (param.draw)
        {
            blockinfo->setCycleName("cycle"); 
            blockinfo->setGraphName("graph");
        }// if
        
        timer.start();
        blockinfo->buildCYCIndepPerm( vecPerm, cyclicBlocks,
                                      *pFE, pNetHier->getList(), 
                                      pA, 
                                      param.xi, 
                                      param.criterion,
                                      param.ordering,
                                      param.startLevel,
                                      param.endLevel );
        timer.stop();
        
        cout << "orderAndSolve : time for CYC = " << timer << endl;

        for (i = param.startLevel; i <= param.endLevel; i++)
            cout << "orderAndSolve : number of cycles in net " << i << " = " 
                 << cyclicBlocks[i]->getSize()/2 << endl;
    }// if
    
    if (( param.ordering == HEURISTIC_FVS ) || ( param.ordering == PLANAR_FVS ))
    {
        timer.start();
        blockinfo->buildFVSIndepPerm( vecPerm, arrFVSize,
                                      *pFE, pNetHier->getList(), 
                                      pA, 
                                      param.xi, param.criterion,
                                      SEARCH_DFS, param.ordering,
                                      param.startLevel,
                                      param.endLevel );
        timer.stop();
        
        cout << "orderAndSolve : time for FVS = " << timer << endl;

        for (i = param.startLevel; i <= param.endLevel; i++) 
            cout << "orderAndSolve : number of FVS nodes in net " << i << " = " 
                 << (*arrFVSize)[i] << endl;
    }// if

    if ( param.ordering == RCM )
    {
        timer.start();
        blockinfo->buildRCMIndepPerm( vecPerm, pA, param.startLevel, param.endLevel );
        timer.stop();
        cout << "orderAndSolve : time for RCM = " << timer << endl;
    }// if

    //
    // build the inverse permutations
    //
    
    if ( param.ordering != NONE )
    {
        uint size;
        
        cout << "orderAndSolve : building inverse permutations" << endl;

        for ( i = param.startLevel; i <= param.endLevel; i++ )
        {
            if ( vecPerm[i] != NULL)
            {
                size = vecPerm[i]->getSize();
                vecInvPerm[i] = new TVectorBlock<uint>( 0, size );
                invertPerm( (*vecInvPerm[i]), *vecPerm[i] );
            }// if
            else
                vecInvPerm[i] = NULL;
        }// for
    }// if

    //
    // reorder
    //

    if (param.ordering != NONE )
    {
        timer.start();
        for ( i = param.startLevel; i <= param.endLevel; i++ ) 
        {
            (*pRHS[i])[0].permute( vecInvPerm[i] );
            (*pA[i])[0][0].permute( vecInvPerm[i], vecInvPerm[i] );
        }
        timer.stop();
        cout << "orderAndSolve : time to permute = "<< timer << endl;
    }// if

    // draw matrix
    if (param.draw)
        TDrawApp::drawMatrix( pA[param.endLevel], param.xi, "pmatrix" );

    /////////////////////////////////////
    //
    // solver
    //

    TSolver  * pPrec;
    TSolver  * pCoarse;
    TSolver  * pPreSmoother, * pPostSmoother;
    TSolver  * pSolver;

    //
    // choose preconditioner
    //

    pPrec = pCoarse = pPreSmoother = pPostSmoother = pSolver = NULL;
    
    if      (param.prec == "none") pPrec = new TRichardson;
    else if (param.prec == "jac")  pPrec = new TJacobi;
    else if (param.prec == "sgs")  pPrec = new TSGS;
    else if (param.prec == "bgs")  pPrec = new TBGS;
    else if (param.prec == "fgs")  pPrec = new TFGS;
    else if (param.prec == "ilu")  pPrec = new TILU;
    else if (param.prec == "default")
    {
        if ((param.ordering == PLANAR_FVS) || (param.ordering == HEURISTIC_FVS))
            pPrec = new TBGS;
        else if ((param.ordering & CONCENTRIC) &&
                 (cyclicBlocks[param.endLevel]->getSize() / 2 > 0) &&
                 ( ! param.pblock ))
        {
            (*pA[param.endLevel])[0][0].decompose( BLOCK_CYC_LU, cyclicBlocks[param.noOfNets - 1] );
            pPrec = new TBlockSymmetricGS;
        }// if
        else
            pPrec = new TSGS;
    }// if
    else
    {
        cout << "orderAndSolve : unsupported preconditioner (" << param.prec << ")" << endl;
        exit( 1 );
    }// else

    if (pPrec != NULL)
    {
        pPrec->setMode( 0 );
        pPrec->setMaxit( 1 );
        pPrec->setVerbosity( 0x0 );
    }// if

    //
    // choose solver
    //
    
    if      (param.solver == "jac")    pSolver = new TJacobi;
    else if (param.solver == "ilu")    pSolver = new TILU;
    else if (param.solver == "sgs")    pSolver = new TFSGS;
    else if (param.solver == "bgs")    pSolver = new TFBGS;
    else if (param.solver == "fgs")    pSolver = new TFFGS;
    else if (param.solver == "bicg")   pSolver = new TBicgStab( pPrec );
    else if (param.solver == "gmres")  pSolver = new TGMRES( pPrec );
    else
    {
        cout << "orderAndSolve : unsupported solver (" << param.solver << ")" << endl;
        exit( 1 );
    }// else

    pSolver->setMode( MODE_REDUCEBY );
    pSolver->setAccuracy( 1e-4 );
    pSolver->setMaxit( param.steps );
    pSolver->setVerbosity( 0xff );

    //
    // solve with permuted matrix
    //

    (*pX[param.endLevel]) = 1000;
    
    timer.start();

    pSolver->solve( param.endLevel, pA, pRHS, pX );
    
    timer.stop();
    cout << "orderAndSolve : time to solve = " << timer << endl;

    //
    // compare with exact solution (in function dirichlet1)
    //
    
    /*
    TBlockVector<TFloat>      *exact      = new TBlockVector<TFloat>( *pRHS[param.endLevel] );
    TDynArray< TVector<2> * > *arrVectors = new TDynArray< TVector<2> * >;

    pFE->getCoord( pNetHier->getList().getLast(), *arrVectors );

    
    for ( i = 0; i < (*exact)[0].getSize(); i++) 
        (*exact)[0][i] = pEqn->dirichlet0( *(*arrVectors)[i] );
    
    if ( param.ordering != NONE )
        (*exact)[0].permute( vecInvPerm[param.endLevel] );

    TBlockVector<TFloat> *error = new TBlockVector<TFloat>( *exact );
    *error -= *pX[param.endLevel];
    
    cout << "orderAndSolve : error of approximation = " << error->weightedEuclidNorm() << endl;

    delete arrVectors;
    delete exact;
    delete error;

    */
    
    //
    // unpermute system
    //
    
    if (param.ordering != NONE )
    {
        for ( i = param.startLevel; i <= param.endLevel; i++ ) 
        {
            (*pRHS[i])[0].permute( vecPerm[i] );
            (*pA[i])[0][0].permute( vecPerm[i], vecPerm[i] );
        }// for
    }// if

    /////////////////////////////////////
    //
    // clean up
    //

    if (pPrec)         delete pPrec;
    if (pCoarse)       delete pCoarse;
    if (pPreSmoother)  delete pPreSmoother;
    if (pPostSmoother) delete pPostSmoother;
    if (pSolver)       delete pSolver;

    delete arrFVSize;
    delete blockinfo;

    if ( param.ordering != NONE )
    {
        for ( i = param.startLevel; i <= param.endLevel; i++ )
        {
            if (vecInvPerm[i] != NULL)
                delete (vecInvPerm[i]);
            
            if (vecPerm[i] != NULL)
                delete (vecPerm[i]);
            
            if (cyclicBlocks[i] != NULL)
                delete (cyclicBlocks[i]);
        }// for
    }// if
}

static void
write_sparse ( TMatrixBlock & A, const char * filename )
{
    FILE * file = fopen( filename, "w" );

    if ( file == NULL )
    {
        cout << "(write_sparse) could not open file " << filename << endl;
        return;
    }// if

    //
    // count no of entries
    //

    unsigned  entries = 0;
    
    for ( unsigned i = 0; i < A.getSize(); i++ )
    {
        uint  n = A[i].getNoOfEntries();

        for ( unsigned j = 0; j < n; j++ )
            if ( A[i].getValue(j) != 0.0 )
                entries++;
    }// for

    fprintf( file, " number of rows = %d\n", A.getSize() );
    fprintf( file, " number of columns = %d\n", A.getSize() );
    fprintf( file, " number of nonzero entries = %d\n", entries );

    //
    // write each row (nentries, col, coeff)
    //
    
    for ( unsigned i = 0; i < A.getSize(); i++ )
    {
        uint  n = A[i].getNoOfEntries();

        for ( unsigned j = 0; j < n; j++ )
            if ( A[i].getValue(j) != 0.0 )
            {
                fprintf( file, " %d\n", i+1 );
                fprintf( file, " %d\n", A[i].getIndex(j)+1 );
                fprintf( file, " %e\n", A[i].getValue(j) );
            }// if
    }// for
    
    fclose( file );
}
