//
//  Project   : Navier-Stokes
//  File      : stokesConv.cc
//  Author    : Ronald Kriemann
//  Purpose   : test of 3D grids
//
// arch-tag: 45424253-365c-4ded-b34c-ca11acc8849f
//

extern "C" {
#include <stdio.h>
#include <stdlib.h>
}

#include "misc/timer.hh"
#include "misc/memory.hh"
#include "misc/cmdline.hh"
#include "misc/TRandom.hh"

#include "net/nethierarchy.hh"
#include "net/T3DTetrahedron.hh"

#include "visual/drawapp.hh"

#include "graph/T3DCellEdgeGraph.hh"
#include "graph/T3DFlowSurface.hh"
#include "graph/T3DSurfaceGraph.hh"
#include "graph/graphtool.hh"

#include "discr/TCdrEqn.hh"

//
// define cdr eqn's for testing
//
class TXLineCdrEqn : public TCdrEqn<3>
{
    // evaluate convection vector
    void evalConv( TVector<3> *, TVector<3> * b )
    {
        if (b != NULL)
        {
            (*b)[X] = 1;
            (*b)[Y] = 0;
            (*b)[Z] = 0;
        }// if
        else
            cout << "TTestCdrEqn::evalConv (...) : gave NULL return vec" << endl;
    }
};

class TCurveCdrEqn : public TCdrEqn<3>
{
    // evaluate convection vector
    void evalConv( TVector<3> * x, TVector<3> * b )
    {
        if (b != NULL)
        {
            (*b)[X] = 0;
            (*b)[Y] = (*x)[Z];
            (*b)[Z] = 1 - (*x)[Y];
        }// if
        else
            cout << "TTestCdrEqn::evalConv (...) : gave NULL return vec" << endl;
    }
};

class TCircleCdrEqn : public TCdrEqn<3>
{
    // evaluate convection vector
    void evalConv( TVector<3> * x, TVector<3> * b )
    {
        if (b != NULL)
        {
            (*b)[X] = 0.5 - (*x)[Z];
            (*b)[Y] = 0;
            (*b)[Z] = (*x)[X] - 0.5;
        }// if
        else
            cout << "TTestCdrEqn::evalConv (...) : gave NULL return vec" << endl;
    }
};

void
disturbNet ( TNet<3> * pNet )
{
    uint                       i, j;
    TListIterator< TCell<3> >  iterCells( pNet->getList() );
    TCell<3>                  *pCell;
    T3DFrontier               *pFrontier;
    T3DCellEdge               *pEdge;
    TFloat                     length, delta;
    uint                       noOfFrontiers, noOfEdges;
    TVector<3>                 v0, v1;

    iterCells.first();
    while ((pCell = iterCells++) != NULL)
    {
        noOfFrontiers = pCell->getNoOfFrontiers();

        for (i = 0; i < noOfFrontiers; i++)
        {
            pFrontier = (T3DFrontier*) pCell->getFrontier( i );

            noOfEdges = pFrontier->getNoOfEdges();
            for (j = 0; j < noOfEdges; j++)
            {
                pEdge = pFrontier->getEdge( j );

                v0 = * pEdge->getNode( 0 );
                v1 = * pEdge->getNode( 2 );

                length = (v1 - v0).euclidNorm();

                if (length == 0)
                    continue;

                delta = length / 150;

                v0[X] += TRandom::random(-delta, delta);
                v0[Y] += TRandom::random(-delta, delta);
                v0[Z] += TRandom::random(-delta, delta);

                v1[X] += TRandom::random(-delta, delta);
                v1[Y] += TRandom::random(-delta, delta);
                v1[Z] += TRandom::random(-delta, delta);

                if (pEdge->getNode(0)->getBoundary() == INTERIOR)
                {
                    (*pEdge->getNode(0))[X] = v0[X];
                    (*pEdge->getNode(0))[Y] = v0[Y];
                    (*pEdge->getNode(0))[Z] = v0[Z];
                }// if

                if (pEdge->getNode(2)->getBoundary() == INTERIOR)
                {
                    (*pEdge->getNode(2))[X] = v1[X];
                    (*pEdge->getNode(2))[Y] = v1[Y];
                    (*pEdge->getNode(2))[Z] = v1[Z];
                }// if
            }// for
        }// for
    }// while
}

//
// mainfunction
//
int
main( int argc, char **argv ) 
{
    uint               i;
    TTimer             timer;
    TNetHierarchy<3>  *pHierarchy;
    TNet<3>           *pNet;

    //
    // parse command line
    //
    param.convDirName = "circle";
    param.setStartTri( "grids/cube.tri" );
    param.noOfNets = 3;
    param.evalCmdLine( argc, argv );
    cout << "main (...) : chosen parameters : " << endl << param << endl;

    //
    // build net hierarchy
    //
    pHierarchy = new TNetHierarchy<3>( param.startTri );

    timer.start();

    for (i = 0; i < param.noOfNets-1; i++)
        pHierarchy->buildNet();

    timer.stop();
    cout << "main (...) : time for refining = " << timer << endl;

    //
    // build graph
    //
    T3DCellEdgeGraph *pGraph;
    TCdrEqn<3>       *pEqn;

    if (param.convDirName == "xline")
    {
        pEqn = new TXLineCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 0.2, 0.125, 0.5, "files/flow" );
    }// if
    if (param.convDirName == "curve")
    {
        pEqn = new TCurveCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 0.5, 0.125, 0.125, "files/flow" );
    }// if
    if (param.convDirName == "circle")
    {
        pEqn = new TCircleCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 0.125, 0.5, 0.125, "files/flow" );
    }// if
    
    pNet   = pHierarchy->getList().getLast();
//    disturbNet( pNet );
    
    if ((verbosity > VERBOSE_RESULT) && (param.draw))
        TDrawApp::drawNet( pNet, "files/net", VRML );

    cout << "main (...) : no of interior nodes = "
         << pNet->getNoOfInteriorNodes() << endl;
    
    timer.start();

    pGraph = new T3DCellEdgeGraph( pNet, pEqn, param.fsKappa, param.fsGamma );
    
    timer.stop();
    cout << "main (...) : time for building graph = " << timer << endl;

    if ((verbosity > VERBOSE_RESULT) && (param.draw))
        TDrawApp::drawGraph( pGraph, "files/graph", VRML );

    cout << "main (...) : no of nodes in graph = " << pGraph->getNoOfNodes() << endl;
    
    //
    // build flow surfaces
    //
    T3DFlowSurface                    * pFlowSurface = new T3DFlowSurface;
    TList< T3DCellEdgeGraph >         * pFSList;
    TListIterator< T3DCellEdgeGraph > * pIterFS;
    T3DCellEdgeGraph                  * pFSGraph;
    
    timer.start();

    if (param.searchType == SEARCH_DFS)
        pFSList = pFlowSurface->buildFlowSurfaces( pGraph, param.fsAlpha,
                                                   param.fsBeta, SEARCH_DFS );
    else
        pFSList = pFlowSurface->buildFlowSurfaces( pGraph, param.fsAlpha,
                                                   param.fsBeta, SEARCH_BFS );
    
    pIterFS = new TListIterator< T3DCellEdgeGraph >( *pFSList );

    timer.stop();
    cout << "main (...) : time for building flow surfaces = " << timer << endl;

    //
    // output
    //
    cout << "main (...) : detected " << pFSList->getSize() << " flow surfaces :" << endl;
    i = 0;
    pIterFS->first();
    while ((pFSGraph = (*pIterFS)++) != NULL)
    {
        cout << "    fs" << i++ << " has size : " << pFSGraph->getNoOfNodes() << endl;
    }// while
    
    if (param.draw)
    {
        char  szFilename[256];
        
        timer.start();

        TDrawApp::drawFlowSurfaces( pFSList, "files/flowgraph" );
        
        i = 0;
        pIterFS->first();
        while ((pFSGraph = (*pIterFS)++) != NULL)
        {
            sprintf( szFilename, "files/flow%d", i );
            TDrawApp::drawGraph( pFSGraph, szFilename );

            // draw surface graph
            T3DSurfaceGraph  graph( pFSGraph );

            sprintf( szFilename, "files/sgraph%d", i++ );
            TDrawApp::drawGraph( & graph, szFilename );
        }// while
        
        timer.stop();
        if (verbosity >= VERBOSE_TIME)
            cout << "main (...) : time for drawing = " << timer << endl;
    }// if

    //
    // clean up
    //
    
    pIterFS->first();
    while ((pFSGraph = (*pIterFS)++) != NULL)
    {
        delete pFSGraph;
    }// while
    
    pHierarchy->free();
    delete pGraph;
    delete pEqn;
    delete pFlowSurface;
    delete pFSList;
    delete pIterFS;

    TMemory::statistics();
}
