//
//  Project   : Navier-Stokes
//  File      : convdiff.cc
//  Author    : Ronald Kriemann
//  Purpose   : test of 3D conv-diff
//
// arch-tag: cff6689e-8e88-484b-9825-6542e232c073
//

extern "C" {
#include <stdio.h>
#include <stdlib.h>
}

#include "misc/var.hh"
#include "misc/memory.hh"
#include "misc/timer.hh"
#include "misc/cmdline.hh"
#include "misc/TRandom.hh"

#include "net/T3DFrontier.hh"
#include "net/T3DCellEdge.hh"
#include "net/net.hh"
#include "net/TCdrBoundaryDesc.hh"
#include "net/TDomain.hh"
#include "net/T3DMeshGenerator.hh"

#include "discr/TConvDiffDoFManager.hh"
#include "discr/TConvDiffFE.hh"
#include "discr/TConvDiffHierarchy.hh"
#include "discr/flowdef.hh"

#include "matrix/blockvector.hh"

#include "solver/secondNF.hh"
#include "solver/fastGaussSeidel.hh"
#include "solver/bicgStab.hh"
#include "solver/cgSolver.hh"
#include "solver/TSOR.hh"
#include "solver/TGMRES.hh"

#include "graph/T3DGraph.hh"
#include "graph/T3DPermBuilder.hh"
#include "graph/criterion.hh"

#include "visual/drawapp.hh"
#include "visual/matrixview.hh"

//
// forward decl.
//
void disturbNet ( TNet<3> * pNet );

void orderAndSolve ( TCdrEqn<3>           * pEqn,
                     TNet<3>              * pNet,
                     TBlockMatrix         * pA,
                     TBlockVector<TFloat> * pRHS,
                     TBlockVector<TFloat> * pX,
                     uint                   ordering,
                     uint                   criterion = E7 ,
                     TFloat                 xi = 1.0,
                     bool                   global = FALSE );

//
// definition of equation to solve
//
class TDirichletFun : public TFunction<3>
{
public:
    TDirichletFun() {};
    ~TDirichletFun() {};

    TFloat eval ( TVector <3> *v )
    {
        TFloat a = (*v)[0];
        TFloat b = (*v)[1];
        TFloat c = (*v)[2];
        
        //  return a+b+2*c;
        return (a*a)+(b*b)+(c*c);
    }
};

//
// mainfunction
//
int
main ( int argc, char **argv ) 
{
    uint    i;
    TTimer  timer;

    //
    // parse command line
    //
    param.convDirName = "circle";
    param.noOfNets    = 4;
    param.criterion   = 3;
    param.xi          = 0.2;
    param.steps       = 100;
    param.solver      = "gmres";
    param.setStartTri( "grids/cube.tri" );

    param.evalCmdLine( argc, argv );
    cout << "main (...) : chosen parameters : " << endl << param << endl << endl;

    //
    // init discr functions
    //
    TCdrEqn<3>          * pEqn;
    TFunction<3>        * pDirichlet;

    if (param.convDirName == "none")
    {
        pEqn = new TNoConvCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 0.1, 0.1, 0.1, "flow" );
    }// if
    else if (param.convDirName == "xline")
    {
        pEqn = new TXLineCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 0.2, 1.0/2.0, 0.1, "flow" );
    }// if
    else if (param.convDirName == "curve")
    {
        pEqn = new TCurveCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 1.0/2.0, 0.1, 0.1, "flow" );
    }// if
    else if (param.convDirName == "circle")
    {
        pEqn = new TCircleCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 1.0/10.0, 1.0/10.0, 1.0/2.0, "flow" );
    }// if
    else if (param.convDirName == "4circles")
    {
        pEqn = new T4CirclesCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 1.0/15.0, 1.0/15.0, 1.0/2.0, "flow" );
    }// if
    else if (param.convDirName == "diagcircle")
    {
        pEqn = new TDiagCircleCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 0.1, 0.1, 0.1, "flow" );
    }// if
    else if (param.convDirName == "ball")
    {
        pEqn = new TBallCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 0.1, 0.1, 0.1, "flow" );
    }// if
    else if (param.convDirName == "vortex")
    {
        pEqn = new TVortexCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 0.1, 0.1, 0.1, "flow" );
    }// if
    else
    {
        cout << "main : convection (" << param.convDirName << ") unsupported" << endl;
        exit( 1 );
    }// else

    if (pEqn == NULL)
    {
        cout << "main (...) : wrong flow specified" << endl;
        exit( 1 );
    }// if
    
    pEqn->isLaplace(  TRUE );
    pEqn->Diffusion(  TRUE  );
    pEqn->Convection( TRUE  );
    pEqn->setEpsilon( param.epsilon );
  
    pDirichlet = new TDirichletFun;

    //
    // build net
    //
    TNetHierarchy<3>  * pHierarchy = new TNetHierarchy<3>( param.starttri );
    TNet<3>           * pNet;

    timer.start();
    for (i = 1; i < param.noOfNets; i++)
    {
        pNet = pHierarchy->buildNet();
    }// for

    timer.stop();
    cout << "main (...) : time for refining = " << timer << endl;

    cout << "main (...) : used memory = " << TMemory::used() << " kB" << endl;

    if (param.disturbNet)
        disturbNet( pNet );

    //
    // supply boundary Descriptons
    //
    TListIterator< TBoundary >  * pIterBound;
    TCdrBoundaryDesc<3>         * pBoundDesc;
    TBoundary                   * pBoundary;
    
    pBoundDesc = new TCdrBoundaryDesc<3>( 1 );
    
    pBoundDesc->setType( 0, DIR_TYPE );
    pBoundDesc->setFunc( 0, pDirichlet );
    pBoundDesc->setDoFInfo();

    pIterBound = new TListIterator< TBoundary >;
    pIterBound->setList( pGlobalDomain->getBoundaries() );
    
    pIterBound->first();
    while ((pBoundary = (*pIterBound)++) != NULL)
    {
        if (pBoundary->getName() == "b1" )
            pBoundary->setBoundaryDesc( pBoundDesc );
    }// while

    //
    // build stiffnessmatrix
    //
    TConvDiffDoFManager<3>  * pDoFManager;
    TBlockMatrix            * pA;
    ulong                     dim;
    TBlockVector<TFloat>    * pRHS;
    TBlockVector<TFloat>    * pX;
    TConvDiffFE<3>          * pFE;
    
    pDoFManager  = new TConvDiffDoFManager<3>();
    pFE          = new TConvDiffFE<3>( Tabata );
    pRHS         = new TBlockVector<TFloat>;
    pX           = new TBlockVector<TFloat>;

    pDoFManager->initDoFDesc( pNet );
        
    cout << "main (...) : number of DoFs = " << pDoFManager->getNoOfDoFs() << endl;

    pA  = pDoFManager->buildStiffnessMatrix( pNet, pNet );
    dim = (*pA)[0][0].getNoOfRows();
    
    pRHS->setNoOfBlocks( 1 );
    (*pRHS)[0].setSize( dim );
    
    pX->setNoOfBlocks( 1 );
    (*pX)[0].setSize( dim );
    
    (*pRHS) = 0;
    
    //
    // discretize
    //
    timer.start();
    
    pFE->discretize( pEqn, pNet, pA,  pRHS);
    
    timer.stop();
    cout << "main (...) : time for discretizing = " << timer << endl << endl;
    
    if ((verbosity > VERBOSE_RESULT) && (param.draw))
        TDrawApp::drawNet( pNet, "net", VRML );

    if (param.draw)
        TDrawApp::drawMatrix( & (*pA)[0][0], param.xi, "matrixblock" );

    cout << "main (...) : used memory = " << TMemory::used() << " kB" << endl;

    //
    // solve equation
    //

    orderAndSolve( pEqn, pNet, pA, pRHS, pX, param.ordering,
                   param.criterion, param.xi, param.global );

    /*
    cout << endl << "main (...) : solving orig. system" << endl << endl;
    orderAndSolve( pEqn, pNet, pA, pRHS, pX, NONE );
    
    cout << endl << "main (...) : solving with RCM" << endl << endl;
    orderAndSolve( pEqn, pNet, pA, pRHS, pX, RCM );
    
    cout << endl << "main (...) : solving with global fvs" << endl << endl;
    orderAndSolve( pEqn, pNet, pA, pRHS, pX, HEURISTIC_FVS,
                   param.criterion, param.xi, TRUE );
    
    cout << endl << "main (...) : solving with fs-cycles" << endl << endl;
    orderAndSolve( pEqn, pNet, pA, pRHS, pX, CONCENTRIC_MIN,
                   param.criterion, param.xi );
    
    cout << endl << "main (...) : solving with fs-fvs" << endl << endl;
    orderAndSolve( pEqn, pNet, pA, pRHS, pX, HEURISTIC_FVS,
                   param.criterion, param.xi, FALSE );
                   */
    
    //
    // clean up
    //
    
    pHierarchy->free();
    delete pEqn;
    delete pFE;
    delete pRHS;
    delete pX;
    delete pA;
    delete pDoFManager;
    delete pGlobalDomain;

    TMemory::statistics();
}

////////////////////////////////////////////////////////////////////////////
//
// misc. methods
//

//
// move inner points of net randomly
//
void
disturbNet ( TNet<3> * pNet )
{
    uint                       i, j;
    TListIterator< TCell<3> >  iterCells( pNet->getList() );
    TCell<3>                  *pCell;
    T3DFrontier               *pFrontier;
    T3DCellEdge               *pEdge;
    TFloat                     length, delta;
    uint                       noOfFrontiers, noOfEdges;
    TVector<3>                 v0, v1;

    iterCells.first();
    while ((pCell = iterCells++) != NULL)
    {
        noOfFrontiers = pCell->getNoOfFrontiers();

        for (i = 0; i < noOfFrontiers; i++)
        {
            pFrontier = (T3DFrontier*) pCell->getFrontier( i );

            noOfEdges = pFrontier->getNoOfEdges();
            for (j = 0; j < noOfEdges; j++)
            {
                pEdge = pFrontier->getEdge( j );

                v0 = * pEdge->getNode( 0 );
                v1 = * pEdge->getNode( 2 );

                length = (v1 - v0).euclidNorm();

                if (length == 0)
                    continue;

                delta = length / 115;

                v0[X] += TRandom::random(-delta, delta);
                v0[Y] += TRandom::random(-delta, delta);
                v0[Z] += TRandom::random(-delta, delta);

                v1[X] += TRandom::random(-delta, delta);
                v1[Y] += TRandom::random(-delta, delta);
                v1[Z] += TRandom::random(-delta, delta);

                if (pEdge->getNode(0)->getBoundary() == INTERIOR)
                {
                    (*pEdge->getNode(0))[X] = v0[X];
                    (*pEdge->getNode(0))[Y] = v0[Y];
                    (*pEdge->getNode(0))[Z] = v0[Z];
                }// if

                if (pEdge->getNode(2)->getBoundary() == INTERIOR)
                {
                    (*pEdge->getNode(2))[X] = v1[X];
                    (*pEdge->getNode(2))[Y] = v1[Y];
                    (*pEdge->getNode(2))[Z] = v1[Z];
                }// if
            }// for
        }// for
    }// while
}

//
// compute ordering, permute matrix, solve equation and unpermute matrix
//
void
orderAndSolve ( TCdrEqn<3>           * pEqn,
                TNet<3>              * pNet,
                TBlockMatrix         * pA,
                TBlockVector<TFloat> * pRHS,
                TBlockVector<TFloat> * pX,
                uint                   ordering,
                uint                   criterion = E7 ,
                TFloat                 xi = 1.0,
                bool                   global = FALSE )
{
    TTimer           timer;
    TSolver        * pPrec;
    TSolver        * pSolver;
    T3DPermBuilder   permBuilder;
    TPermutation   * pPerm, * pInvPerm;

    TDynArray< ulong > * pBlockStructure = NULL;
    
    timer.start();

    pPerm = NULL;
    
    if (ordering == HEURISTIC_FVS)
    {
        if (global)
            pPerm = permBuilder.buildFVSPerm( pNet, & (*pA)[0][0] );
        else
        {
            pPerm = permBuilder.buildFVSPerm( pNet, pEqn, & (*pA)[0][0], HEURISTIC_FVS );
            pBlockStructure = permBuilder.blockStructure();
        }// if
    }// if

    if (ordering == PLANAR_FVS)
    {
        pPerm = permBuilder.buildFVSPerm( pNet, pEqn, & (*pA)[0][0],
                                          PLANAR_FVS );
        pBlockStructure = permBuilder.blockStructure();
    }// if
    
    if (ordering == RCM)
        pPerm = permBuilder.buildRCMPerm( & (*pA)[0][0] );

    if (ordering == CONCENTRIC_MIN)
    {
        pPerm = permBuilder.buildCYCPerm( pNet, pEqn, & (*pA)[0][0] );
        pBlockStructure = permBuilder.blockStructure();
    }// if

    timer.stop();
    cout << "orderAndSolve (...) : time for building permutation = " << timer << endl;
    
    if (pPerm != NULL)
        pInvPerm = new TPermutation( pPerm->invert() );
    else
        pInvPerm = NULL;
    
    //
    // permute matrix
    //
    timer.start();

    if (pInvPerm != NULL)
    {
        (*pA)[0][0].permute( pInvPerm, pInvPerm );
        pRHS->permute( pInvPerm );
    }// if

    timer.stop();
    cout << "orderAndSolve (...) : time for permute = " << timer << endl;

    if (param.draw)
        TDrawApp::drawMatrix( & (*pA)[0][0], xi, pBlockStructure, "pmatrixblock" );
    
    //
    // solve with permuted matrix
    //
    (*pX) = 1;
    
    if (ordering == NONE)
        pPrec = new TSGS;

    if (ordering == HEURISTIC_FVS)
        pPrec = new TBGS;

    if (ordering == PLANAR_FVS)
        pPrec = new TBGS;

    if (ordering == RCM)
        pPrec = new TSGS;

    if (ordering == CONCENTRIC_MIN)
        pPrec = new TSGS;
    
    pPrec->setMaxit( 1 );
    pPrec->setVerbosity( 0 );

    if (param.solver == "bicg")
        pSolver = new TBicgStab( pPrec );
    else if (param.solver == "cg")
        pSolver = new TCGSolver( pPrec );
    else if (param.solver == "gmres")
        pSolver = new TGMRES( pPrec );
    else if (param.solver == "sgs")
        pSolver = new TSGS;
    else if (param.solver == "bgs")
        pSolver = new TBGS;
    else if (param.solver == "fgs")
        pSolver = new TFGS;
    else
    {
        cout << " main : no or unsupported solver (" << param.solver << ") specified" << endl;
        exit( 1 );
    }// else
        
    pSolver->setMaxit( param.steps );
    pSolver->setVerbosity( 0xff );
    
    timer.start();
    pSolver->solve( pA, pRHS, pX );
    timer.stop();
    cout << "orderAndSolve (...) : time for solving = " << timer << endl;

    //
    // unpermute
    //

    if (pPerm != NULL)
    {
        (*pA)[0][0].permute( pPerm, pPerm );
        pRHS->permute( pPerm );
    }// if
        
    //
    // clean up
    //
    
    delete pSolver;
    delete pPerm;
    delete pInvPerm;
}
