//
//  Project   : Navier-Stokes
//  File      : stokes.cc
//  Author    : Ronald Kriemann
//  Purpose   : test of 3D conv-diff
//
// arch-tag: 23191d33-7593-4918-8f3d-91596b6953e4
//

extern "C" {
#include <stdio.h>
#include <stdlib.h>
}

#include "misc/var.hh"
#include "misc/memory.hh"
#include "misc/timer.hh"
#include "misc/cmdline.hh"
#include "misc/TRandom.hh"

#include "net/T3DFrontier.hh"
#include "net/T3DCellEdge.hh"
#include "net/net.hh"
#include "net/TStokesBoundaryDesc.hh"
#include "net/TDomain.hh"

#include "discr/TStokesDoFManager.hh"
#include "discr/TStokesFE.hh"
#include "discr/flowdef.hh"

#include "matrix/blockvector.hh"

#include "solver/secondNF.hh"
#include "solver/schur.hh"
#include "solver/bicgStab.hh"

#include "graph/T3DGraph.hh"
#include "graph/T3DPermBuilder.hh"

#include "visual/drawapp.hh"

void
disturbNet ( TNet<3> * pNet )
{
    uint                       i, j;
    TListIterator< TCell<3> >  iterCells( pNet->getList() );
    TCell<3>                  *pCell;
    T3DFrontier               *pFrontier;
    T3DCellEdge               *pEdge;
    TFloat                     length, delta;
    uint                       noOfFrontiers, noOfEdges;
    TVector<3>                 v0, v1;

    iterCells.first();
    while ((pCell = iterCells++) != NULL)
    {
        noOfFrontiers = pCell->getNoOfFrontiers();

        for (i = 0; i < noOfFrontiers; i++)
        {
            pFrontier = (T3DFrontier*) pCell->getFrontier( i );

            noOfEdges = pFrontier->getNoOfEdges();
            for (j = 0; j < noOfEdges; j++)
            {
                pEdge = pFrontier->getEdge( j );

                v0 = * pEdge->getNode( 0 );
                v1 = * pEdge->getNode( 2 );

                length = (v1 - v0).euclidNorm();

                if (length == 0)
                    continue;

                delta = length / 150;

                v0[X] += TRandom::random(-delta, delta);
                v0[Y] += TRandom::random(-delta, delta);
                v0[Z] += TRandom::random(-delta, delta);

                v1[X] += TRandom::random(-delta, delta);
                v1[Y] += TRandom::random(-delta, delta);
                v1[Z] += TRandom::random(-delta, delta);

                if (pEdge->getNode(0)->getBoundary() == INTERIOR)
                {
                    (*pEdge->getNode(0))[X] = v0[X];
                    (*pEdge->getNode(0))[Y] = v0[Y];
                    (*pEdge->getNode(0))[Z] = v0[Z];
                }// if

                if (pEdge->getNode(2)->getBoundary() == INTERIOR)
                {
                    (*pEdge->getNode(2))[X] = v1[X];
                    (*pEdge->getNode(2))[Y] = v1[Y];
                    (*pEdge->getNode(2))[Z] = v1[Z];
                }// if
            }// for
        }// for
    }// while
}

class TDirichletFun0 : public TFunction<3>
{
public:
    TFloat eval(TVector<3> *v )
	{ 
        return ( sin((*v)[0]) * sin((*v)[1]) );
	}
};

class TDirichletFun1 : public TFunction<3>
{
public:
    TFloat eval(TVector <3> *v )
	{
        return  ( cos((*v)[0]) * cos((*v)[1]) );
	}
};

class TDirichletFun2 : public TFunction<3>
{
public:
    TFloat eval(TVector <3> *v )
	{
        return (0.0);
	}
};

class TPressure : public TFunction<3>
{
public:
    TFloat eval ( TVector <3> *v )
	{
        return (2.0 * cos((*v)[0]) * sin((*v)[1]) -
                2.0 * sin(1.0) * (1.0 - cos(1.0)));
	}
};

//
// mainfunction
//
int
main ( int argc, char **argv ) 
{
    uint    i;
    TTimer  timer;

    //
    // parse command line
    //
    param.convDirName = "circle";
    param.noOfNets    = 4;
    param.criterion   = 3;
    param.xi          = 0.2;
    param.sigma       = 0.1;
    param.steps       = 100;
    param.setStartTri( "grids/cube.tri" );

    param.evalCmdLine( argc, argv );
    cout << "main (...) : chosen parameters : " << endl << param << endl << endl;

    //
    // init net and discr functions
    //
    TNetHierarchy<3>  * pHierarchy = new TNetHierarchy<3>( param.starttri );
    TNet<3>           * pCNet, * pFNet;

    //
    // init discr functions
    //
    TStokesBasisEqn<3>  * pStokesEqn;
    TFunction<3>        * pDirichlet[3];
    TFunction<3>        * pPressure;

    if (param.convDirName == "none")
    {
        pStokesEqn = new TNoConvStokesEqn;

        if (param.draw)
            TDrawApp::drawFlow( pStokesEqn, 0.1, 0.1, 0.1, "flow" );
    }// if
    if (param.convDirName == "xline")
    {
        pStokesEqn = new TXLineStokesEqn;

        if (param.draw)
            TDrawApp::drawFlow( pStokesEqn, 0.2, 0.1, 1.0/3.0, "flow" );
    }// if
    if (param.convDirName == "curve")
    {
        pStokesEqn = new TCurveStokesEqn;

        if (param.draw)
            TDrawApp::drawFlow( pStokesEqn, 1.0/3.0, 0.1, 0.1, "flow" );
    }// if
    if (param.convDirName == "circle")
    {
        pStokesEqn = new TCircleStokesEqn;

        if (param.draw)
            TDrawApp::drawFlow( pStokesEqn, 0.1, 1.0/3.0, 0.1, "flow" );
    }// if
    if (param.convDirName == "4circles")
    {
        pStokesEqn = new T4CirclesStokesEqn;

        if (param.draw)
            TDrawApp::drawFlow( pStokesEqn, 0.1, 1.0/3.0, 0.1, "flow" );
    }// if
    if (param.convDirName == "diagcircle")
    {
        pStokesEqn = new TDiagCircleStokesEqn;

        if (param.draw)
            TDrawApp::drawFlow( pStokesEqn, 0.1, 0.1, 0.1, "flow" );
    }// if
    if (param.convDirName == "vortex")
    {
        pStokesEqn = new TVortexStokesEqn;

        if (param.draw)
            TDrawApp::drawFlow( pStokesEqn, 0.16, 0.25, 0.16, "flow" );
    }// if

//     pStokesEqn->isLaplace(  TRUE );
//     pStokesEqn->Diffusion(  TRUE  );
//     pStokesEqn->Convection( TRUE  );
    pStokesEqn->setEpsilon( param.epsilon );
  
    pDirichlet[0] = new TDirichletFun0();
    pDirichlet[1] = new TDirichletFun1();
    pDirichlet[2] = new TDirichletFun2();
    pPressure     = new TPressure();

    //
    // supply boundary Descriptons
    //
    TListIterator< TBoundary >  * pIterBound;
    TStokesBoundaryDesc<3>      * pBoundDesc;
    TBoundary                   * pBoundary;
    
    pBoundDesc = new TStokesBoundaryDesc<3>( 4 );

    for ( i = 0; i < 3; i++ )
    {
        pBoundDesc->setType( i, DIR_TYPE);
        pBoundDesc->setFunc( i, pDirichlet[i] );
        pBoundDesc->setDoFInfo( i, FALSE);
    }// for
    
    pBoundDesc->setType( 3, NEU_TYPE);
    pBoundDesc->setDoFInfo( 3, TRUE);

    pIterBound = new TListIterator< TBoundary >;
    pIterBound->setList( pDomain->getBoundaries() );
    
    pIterBound->first();
    while ((pBoundary = (*pIterBound)++) != NULL)
    {
        if (pBoundary->getName() == "b1" )
            pBoundary->setBoundaryDesc( pBoundDesc );
    }// while

    //
    // discretize
    //
    TStokesDoFManager<3>    * DoFManager;
    TBlockMatrix            * pA;
    ulong                     dim, noOfBlocks;
    TBlockVector<TFloat>    * pRHS;
    TBlockVector<TFloat>    * pX;
    TBlockVector<TFloat>    * pExact, * pResidual;
    TStokesFE<3>            * discretiser;
    TMatrixBlock            * pC;
    
    DoFManager  = new TStokesDoFManager<3>();
    discretiser = new TStokesFE<3>( Tabata );
    pRHS        = new TBlockVector<TFloat>;
    pX          = new TBlockVector<TFloat>;
    pExact      = new TBlockVector<TFloat>;
    pResidual   = new TBlockVector<TFloat>;

    timer.start();
    
    pCNet = pFNet = NULL;
    for (i = 1; i < param.noOfNets; i++)
    {
        pCNet = pFNet;
        pFNet = pHierarchy->buildNet();
    }// for
    
    timer.stop();
    cout << "main (...) : time for refining     = " << timer << endl;
    
    cout << "main (...) : used memory           = " << TMemory::used() << endl;

    if (param.disturbNet)
        disturbNet( pFNet );
    
    pA = DoFManager->buildStiffnessMatrix( pCNet, pFNet );
    
    dim        = 0;
    noOfBlocks = pA->getNoOfRows();

    pRHS->setNoOfBlocks( noOfBlocks );
    pX->setNoOfBlocks( noOfBlocks );
    pExact->setNoOfBlocks( noOfBlocks );
    pResidual->setNoOfBlocks( noOfBlocks );
    
    for ( i = 0; i < noOfBlocks - 1; i++ )
    {
        dim += (*pA)[i][i].getNoOfRows();
        (*pRHS)[i].setSize( (*pA)[i][i].getNoOfRows() );
        (*pX)[i].setSize( (*pA)[i][i].getNoOfRows() );
        (*pExact)[i].setSize( (*pA)[i][i].getNoOfRows() );
        (*pResidual)[i].setSize( (*pA)[i][i].getNoOfRows() );
    }// for
    (*pRHS)[i].setSize( (*pA)[i][0].getNoOfRows() );
    (*pX)[i].setSize( (*pA)[i][0].getNoOfRows() );
    (*pExact)[i].setSize( (*pA)[i][0].getNoOfRows() );
    (*pResidual)[i].setSize( (*pA)[i][0].getNoOfRows() );
    
    (*pRHS) = 0;
    
    cout << "main (...) : global dimension      = " << dim << endl;

    pC = new TMatrixBlock( (*pA)[0][0] );
    
    cout << "main (...) : used memory           = " << TMemory::used() << endl;

    //
    // Diskretisierung
    //
    timer.start();
    discretiser->discrDiff( pStokesEqn, pCNet, pA, pRHS);
    for ( i = 0; i < 3; i++)
	{
        *pC = (*pA)[i][i];
        discretiser->discrConv( pStokesEqn, pCNet, pC, pRHS, i );
        (*pA)[i][i] = *pC;
	}// for
    discretiser->discrRhs(  pStokesEqn, pCNet, pRHS);
    timer.stop();
    cout << "main (...) : time for discretizing = " << timer << endl << endl;
    
    for ( i = 0; i < 3; i++ )
        (*pA)[i][i] += *pC;

    if (param.draw)
    {
        TDrawApp::drawMatrix( pA, param.xi, "matrix" );
        TDrawApp::drawMatrix( & (*pA)[0][0], param.xi, "matrixblock" );
    }// if

    cout << endl;
    
    if ((verbosity > VERBOSE_RESULT) && (param.draw))
        TDrawApp::drawNet( pFNet, "net", VRML );

    // test discretisation:
    ulong dim_u = (*pA)[0][0].getNoOfRows();
    ulong dim_p = (*pA)[3][0].getNoOfRows();
    
    TDynArray< TVector<3> * > arrCoord_U( dim_u );
    TDynArray< TVector<3> * > arrCoord_P( dim_p );
    discretiser->getCoord ( pCNet, 0, arrCoord_U, arrCoord_P );
    
    for ( i = 0; i < dim_u; i++)
	{
        (*pExact)[0][i] = pDirichlet[0]->eval( arrCoord_U[i] );
        (*pExact)[1][i] = pDirichlet[1]->eval( arrCoord_U[i] );
        (*pExact)[2][i] = pDirichlet[2]->eval( arrCoord_U[i] );
    }// for
    
    for ( i = 0; i < dim_p; i++)
        (*pExact)[3][i] = pPressure->eval( arrCoord_P[i] );

    (*pResidual) = (*pA) * (*pExact) - (*pRHS);
    cout << "main (...) : residual of exact solution = "
         << setprecision( 5 ) << pResidual->weightedEuclidNorm() << endl;

    //
    // solve with original matrix
    //
    TSolver   * pSolver = new TSchur;

    pSolver->setAccuracy( 1e-15 );
    pSolver->setMaxit( param.steps );
    pSolver->setVerbosity( 4 );
    ((TSchur*)pSolver)->setMaxInnerSteps( 10 );
    ((TSchur*)pSolver)->setInnerReduceBy( 1.0e-1 );

    (*pX) = 1;
    
    timer.start();
    pSolver->solve( pA, pRHS, pX );
    timer.stop();
    cout << "main (...) : time for solving = " << timer << endl;

    //
    // build permutation
    //
    T3DPermBuilder * pPermBuilder = new T3DPermBuilder;
    TPermutation   * pPerm, * pInvPerm;
    
    timer.start();

    pPerm = NULL;
    
    if (param.ordering == HEURISTIC_FVS)
    {
        if (param.global)
            pPerm = pPermBuilder->buildFVSPerm( pFNet, & (*pA)[0][0] );
        else
            pPerm = pPermBuilder->buildFVSPerm( pFNet, pStokesEqn, & (*pA)[0][0] );
    }// if

    if (param.ordering == RCM)
        pPerm = pPermBuilder->buildRCMPerm( & (*pA)[0][0] );

    if (param.ordering == CONCENTRIC_MIN)
        pPerm = pPermBuilder->buildCYCPerm( pFNet, pStokesEqn, & (*pA)[0][0] );

    if (pPerm == NULL)
        exit( 1 );
    
    timer.stop();
    cout << "main (...) : time for building permutation = " << timer << endl;
    
    if (pPerm == NULL)
    {
        cout << "main (...) : have NULL permutation" << endl;
        exit( 1 );
    }// if

    pInvPerm = new TPermutation( pPerm->invert() );
    
    //
    // permute matrix
    //

    timer.start();
    
    pA->permute( pInvPerm );
    pRHS->permute( pInvPerm );

    timer.stop();
    cout << "main (...) : time for permute = " << timer << endl;
    
    if (param.draw)
    {
        TDrawApp::drawMatrix( pA, param.xi, "pmatrix" );
        TDrawApp::drawMatrix( & (*pA)[0][0], param.xi, "pmatrixblock" );
    }// if
    
    //
    // solve with permuted matrix
    //
    (*pX) = 1;
    
    pA->deleteLUdecomp();

    timer.start();
    pSolver->solve();
    timer.stop();
    cout << "main (...) : time for solving = " << timer << endl;
    
    //
    // clean up
    //
    
    delete pPerm;
    delete pInvPerm;
    
    pHierarchy->free();
    delete pStokesEqn;
    delete discretiser;
    delete pRHS;
    delete pX;
    delete pExact;
    delete pResidual;
    delete pA;
    delete pC;
    delete DoFManager;
    delete pDomain;
    delete pPermBuilder;

    delete pSolver;
  
    TMemory::statistics();
}
