//
//  Project   : Navier-Stokes
//  File      : stokesConv.cc
//  Author    : Ronald Kriemann / Sabine Gutsch
//  Purpose   : test for building a hierarchy for the Stokes
//              problem and multigrid
//  Date      : 13.06.1997
//  Changes   : 
//
// arch-tag: d7a1e8c6-f437-4d96-ad9d-299a62726443
//

#include <stdio.h>

#include "misc/timer.hh"
#include "misc/list.hh"
#include "misc/cmdline.hh"

#ifdef MATLAB
#include "misc/matlab.hh"
#endif

#include "net/vertex.hh"
#include "net/regular.hh"
#include "net/net.hh"

#include "solver/secondNF.hh"
#include "solver/cgSolver.hh"
#include "solver/bicgStab.hh"
#include "solver/kaczmarz.hh"
#include "solver/lu.hh"
#include "solver/multigrid.hh"
#include "solver/blocksolver.hh"
#include "solver/schur.hh"
#include "solver/squaredSmoother.hh"
#include "solver/optSquaredSmoother.hh"
#include "solver/optSquaredSmoother2.hh"
#include "solver/residualSmoother.hh"

#include "discr/stokesDoFManager.hh"
#include "discr/stokesFE.hh"
#include "discr/stokesHierarchy.hh"
#include "discr/convDirections.hh"

#include "matrix/blockmatrix.hh"

#include "visual/drawapp.hh"

#include "graph/blockinfo.hh"

TFloat  p ( const TVector<2> &v) {
    return  ( 2 * cos(v[0]) * sin(v[1]) - 2 * sin(1) * (1-cos(1)) );
}

class TMyStokesEqn : public TBasisEqn<2> {

public:
	TMyStokesEqn (  ) 
		: TBasisEqn<2> ( ) { };
    virtual TFloat f0 (const TVector<2> &v ) 
    { return (2.0 * ( epsilon - 1.0 ) * sin(v[0]) * sin(v[1]) + b0(v) * cos(v[0]) * sin(v[1]) + b1(v) * sin(v[0]) * cos(v[1]) ); 
    }
    virtual TFloat f1 (const TVector<2> &v ) 
	{ return ( 2.0 * ( 1.0 + epsilon) * cos(v[0]) * cos(v[1]) - b1(v) * cos(v[0]) * sin(v[1]) - b0(v) * sin(v[0]) * cos(v[1]) ); }
    virtual TFloat dirichlet0 ( const TVector<2> &v) 
	{ 
        return ( sin(v[0]) * sin(v[1]) );
	}
    virtual TFloat dirichlet1 ( const TVector<2> &v) { 
        return  ( cos(v[0]) * cos(v[1]) );
	}
};

int main( int argc, char **argv ) 

{
    uint i = 0;
	TParameter  param;
	param.steps      = 3;
	param.noOfNets   = 5;
    param.startLevel = 0;
    param.ordering   = NONE;
    param.ordering   = CONCENTRIC_MIN;
    param.xi         = 1.0/3.0;
    param.criterion  = 3;
    param.mgSort     = MGSORT_INDEP; 
    param.epsilon    = 1.0;
    param.draw       = FALSE;
    param.convDirName = "curve";
	strcpy( param.starttri, "grids/start1.tri" );

	param.evalCmdLine( argc, argv );
    param.endLevel = param.noOfNets-2;
    cout << "chosen parameters : " << endl << param;

	//
	// build and refine net
	//

    TTimer timer;
    TNetHierarchy<2> *netHier = new TNetHierarchy<2>( param.starttri );
	netHier->delReference();
    
    for (i = 0; i < param.noOfNets-1; i++)
        netHier->buildNet();
    
    ulong noOfInteriorNodesNet[param.noOfNets];
    
    TListIterator< TNet<2> > *iterNet;
	TNet<2>                  *pNet;
    
    i = 0;
    iterNet = new TListIterator< TNet<2> >( netHier->getList() );
	iterNet->first();
	while ((pNet = (*iterNet)++) != NULL)
	{
		noOfInteriorNodesNet[i++] = pNet->getNoOfInteriorNodes();
		cout << "net " << i-1 << ": " << noOfInteriorNodesNet[i-1]<<endl; 
	}// while

    timer.start();
	//
	// build stiffness matrix sparsity structure
	//

  
    TBlockMatrix **pMatrix = new (TBlockMatrix*)[param.noOfNets-1];;
    TProlongation **prolongation = new (TProlongation*)[param.noOfNets-2];
    TProlongation **restriction = new (TProlongation*)[param.noOfNets-2];
    TBlockVector<TFloat> **pRHS = new (TBlockVector<TFloat>*)[param.noOfNets-1];
    TBlockVector<TFloat> **pX = new (TBlockVector<TFloat>*)[param.noOfNets-1];
  
    T2DStokesDoFManager *dofManager = new T2DStokesDoFManager();
    TMyStokesEqn *eqn = new TMyStokesEqn();
    // set epsilon
    eqn->setEpsilon( param.epsilon );
    // set reaction term
    if (param.reactionTerm)
       eqn->setReacFunction( oneReac ); 
    else 
       eqn->setReacFunction( noReac ); 

    // Pointer to functions for convection direction
    if (param.convDirName == "none")
		eqn->setConvFunctions( noConv0, noConv1 ); 
    if (param.convDirName == "circle")
		eqn->setConvFunctions( circle0, circle1 ); 
    if (param.convDirName == "xline")
		eqn->setConvFunctions( xline0, xline1 ); 
    if (param.convDirName == "curve")
		eqn->setConvFunctions( curve0, curve1 ); 
    if (param.convDirName == "square") 
		eqn->setConvFunctions( square0, square1 );    
    if (param.convDirName == "foursquares") 
		eqn->setConvFunctions( foursquares0, foursquares1 ); 
    if (param.convDirName == "fourcircles") 
		eqn->setConvFunctions( fourcircles0, fourcircles1 ); 

    T2DStokesHierarchy  *hierarchy = new T2DStokesHierarchy();
    T2DStokesFE         *fe        = new T2DStokesFE( Tabata );  

    timer.start();
    hierarchy->buildHierarchy( dofManager, eqn, fe, netHier, pMatrix,
                               pRHS, pX, restriction, prolongation );

	/*
    for (i = 0; i < param.noOfNets - 2; i++) {
        restriction[i]->matrixDependent( &(*pMatrix[i+1])[0][0] );
        prolongation[i]->matrixDependent( &(*pMatrix[i+1])[0][0] );
	}
	*/

	// correction of RHS 
    pX[param.noOfNets-2]->set( 1.0 );
    TFloat lambda = ((*pRHS[param.noOfNets-2])[2] * (*pX[param.noOfNets-2])[2] ) / 
					((*pX[param.noOfNets-2])[2] * (*pX[param.noOfNets-2])[2] );
    cout << "lambda: " << lambda << endl;
    *pRHS[param.noOfNets-2] -= lambda * *pX[param.noOfNets-2];

    timer.stop();
    cout << endl;
	cout << "time to discretizing on " << param.noOfNets-1 
         << " levels: "  << timer << endl;
    cout << endl;
 
	//cout << (*pMatrix[param.noOfNets-2])[0][0] << endl << endl;
	//cout << (*pMatrix[param.noOfNets-3])[0][2] << endl;

	TBlockVector<TFloat> *vecSolution;
	vecSolution = new TBlockVector<TFloat>( *pX[param.noOfNets-2] ); 

	uint dimU = (*vecSolution)[0].getSize();
	uint dimP = (*vecSolution)[2].getSize();
	cout << "Dimension: U: " << dimU << ", P: " << dimP << endl;
	//cout << "Gridsize:  " << sqrt(dimU) + 1 << "x" << sqrt(dimU) + 1 
	//     << endl;

	T2DNet *net = iterNet->last();
	net = iterNet->prev();

	TDynArray< TVector<2> * > *arrVectorsU = new TDynArray< TVector<2> * >;
	TDynArray< TVector<2> * > *arrVectorsP = new TDynArray< TVector<2> * >;
	fe->getCoord(net, *arrVectorsU, *arrVectorsP); 
 
	// calculate exact solution
	for (i = 0; i < dimU; i++) {
		(*vecSolution)[0][i] = eqn->dirichlet0( *((*arrVectorsU)[i]) );
		(*vecSolution)[1][i] = eqn->dirichlet1( *((*arrVectorsU)[i]) );
	}// for
	for (i = 0; i < dimP; i++) {
		(*vecSolution)[2][i] = p( *((*arrVectorsP)[i]) );
	}// for
	// Lagrange factor
	//(*vecSolution)[2][dimP - 1] = 0.0;
	//cout << "solution\n" << *vecSolution << endl;

    //param.xi = 0.33;
	if (param.draw) 
		TDrawApp::drawMatrix( pMatrix[param.noOfNets - 2], param.xi, FALSE, "matrix");
    //TDrawApp::draw( *pMatrix[param.noOfNets - 2], 0.001, "matrix");

    // ordering
    timer.start();
    TBlockInfo  *blockinfo = new TBlockInfo;
    blockinfo->setSortMethod( SORT_OFD );
    //blockinfo->setSortMethod( SORT_EDGE_WEIGHT );
    //blockinfo->setSortMethod( SORT_SUCC_EDGE_WEIGHT );
 
    // true: parallele Anordung
    blockinfo->setBlocking( TRUE );
    blockinfo->setBlocking( FALSE );

    //
    // build data structures for each net
    //
    TVectorBlock<ulong>  *vecPerm[param.noOfNets-1];
    TVectorBlock<ulong>  *cyclicBlocks[param.noOfNets-1];
    TVectorBlock<ulong>  *arrFVSize = new TVectorBlock<ulong>();
 
    if ( param.ordering != NONE )	{
	    for ( i = 0; i < param.noOfNets-1; i++ ) { 
            vecPerm[i] = new TVectorBlock<ulong>;
            cyclicBlocks[i] = new TVectorBlock<ulong>;
		}// for
    }// if

	//
	// now build permutation vectors
	//
	if ( param.ordering & CONCENTRIC ) {
     
		if (param.draw) {
			blockinfo->setCycleName("cycle"); 
			blockinfo->setGraphName("graph");
		}
		timer.start();
	  
		switch (param.mgSort)
		{
		case MGSORT_PROJ :
			blockinfo->buildCYCProjPerm( vecPerm, 
										 cyclicBlocks,
										 *fe, 
										 netHier->getList(), 
										 pMatrix, 
										 param.xi, 
										 param.criterion,
										 param.ordering );
			break;
		case MGSORT_NESTED    :
			blockinfo->buildCYCNestedPerm( vecPerm, cyclicBlocks,
										   *fe, 
										   netHier->getList(), 
										   pMatrix, 
										   param.xi, 
										   param.criterion,
										   param.ordering  );
			break;
		case MGSORT_INDEP  :
			blockinfo->buildCYCIndepPerm( vecPerm, 
										  cyclicBlocks,
										  *fe, 
										  netHier->getList(), 
										  pMatrix, 
										  param.xi, 
										  param.criterion,
										  param.ordering  );
			break;
		default:
			break;
		}// switch

        timer.stop();
        cout << endl << "time for finding cycles : " << timer << endl;
        for (i = 0; i < param.noOfNets-1; i++)
            cout << " Number of cycles in net " << i+1 << " :  " 
                 << cyclicBlocks[i]->getSize()/2 << endl;
	}

	if ( param.ordering == PLANAR_FVS ) 
	{
        timer.start();
 		switch (param.mgSort)
		{
		case MGSORT_PROJ :
			blockinfo->buildPlanarFVSProjPerm( vecPerm, arrFVSize,
											   *fe, netHier->getList(),
											   pMatrix, 
											   param.xi, 
											   param.criterion );
			break;
		case MGSORT_NESTED    :
			blockinfo->buildPlanarFVSNestedPerm( vecPerm, arrFVSize,
												 *fe, netHier->getList(), 
												 pMatrix, 
												 param.xi, param.criterion);
			break;
		case MGSORT_INDEP  :
			blockinfo->buildPlanarFVSIndepPerm( vecPerm, arrFVSize,
												*fe, netHier->getList(), 
												pMatrix, 
												param.xi, param.criterion);
			break;
		default:
			break;
		}// switch
        timer.stop();
        cout << endl << "time for finding FVS : " << timer << endl;
        for (i = 0; i < param.noOfNets-1; i++) 
            cout << " Number of FVS nodes in net " << i+1 << " :  " 
				 << (*arrFVSize)[i] << endl;
	}
	if ( param.ordering == HEURISTIC_FVS ) 
	{
        timer.start();
 		switch (param.mgSort)
		{
		case MGSORT_PROJ :
			blockinfo->buildHeuristFVSProjPerm( vecPerm, arrFVSize,
												*fe, netHier->getList(),
												pMatrix, 
												param.xi, 
												param.criterion );
			break;
		case MGSORT_NESTED    :
			blockinfo->buildHeuristFVSNestedPerm( vecPerm, arrFVSize,
												  *fe, netHier->getList(),
												  pMatrix, 
												  param.xi, param.criterion);
			break;
		case MGSORT_INDEP  :
			blockinfo->buildHeuristFVSIndepPerm( vecPerm, arrFVSize,
												 *fe, netHier->getList(), 
												 pMatrix, 
												 param.xi, param.criterion);
			break;
		default:
			break;
		}// switch
        timer.stop();
        cout << endl << "time for finding FVS : " << timer << endl;
        for (i = 0; i < param.noOfNets-1; i++) 
            cout << " Number of FVS nodes in net " << i+1 << " :  " 
				 << (*arrFVSize)[i] << endl;
	}


    if ( param.ordering == RCM ) {
		timer.start();
		switch (param.mgSort)
		{
		case MGSORT_PROJ :
			blockinfo->buildRCMProjPerm( vecPerm, pMatrix, 
										 param.noOfNets );
			break;
		case MGSORT_INDEP  :
			blockinfo->buildRCMIndepPerm( vecPerm, pMatrix, 
										  param.noOfNets );
			break;
		default:
			break;
		}// switch
        timer.stop();
        cout << endl << "time for doing RCM : " << timer << endl;
	}

    //
	// build the inverse permutations
	//
	TVectorBlock<ulong> *vecInvPerm[param.noOfNets-1];
    
	if ( param.ordering != NONE )	{
		cout << "building inverse permutations" << endl;
	    for ( i = 0; i < param.noOfNets-1; i++ ) {
			ulong size = vecPerm[i]->getSize();
			vecInvPerm[i] = new TVectorBlock<ulong>( 0, size );

            cout << vecInvPerm[i]->getSize() << endl;
			// reverse order
			//for ( ulong j = 0; j < size; j++ ) { 
			//    (*vecInvPerm[i])[j]=(*vecPerm[i])[size-j-1];
			//}
			//*vecPerm[i] = *vecInvPerm[i];

			invertPerm( (*vecInvPerm[i]), *vecPerm[i] );
		}// for 
	}// if


	//
	// reorder
	//
    if (param.ordering != NONE ) {
	    timer.start();
        //(*vecSolution).permute( vecInvPerm[param.noOfNets-2] );
        for ( i = 0; i < param.noOfNets-1; i++ ) 
		{
            if (pMatrix[i] != NULL)
                pMatrix[i]->permute( vecInvPerm[i] );

            if (pRHS[i] != NULL)
            {
//                 cout << (*pRHS)[i] << endl;
//                 pRHS[i]->permute( vecInvPerm[i] );
            }// if
//             if ( i > 0 ) {
// 				(prolongation[i-1]->pBlockProlong)[0]->permute( vecInvPerm[i], 
// 																vecInvPerm[i-1]);
// 				(restriction[i-1]->pBlockProlong)[0]->permute( vecInvPerm[i], 
// 															   vecInvPerm[i-1]);
// 			}
		}
	    timer.stop();
	    cout << "time for permuting matrices and rhs's: "<< timer << endl;
	}	
	
    //param.xi = 0.33;
	if (param.draw) 
		TDrawApp::drawMatrix( pMatrix[param.noOfNets - 2], param.xi, FALSE, "pmatrix");
    //TDrawApp::draw( *pMatrix[param.noOfNets - 2], 0.001, "matrix");

    exit( 0 );
    
    timer.start();
    //(*pMatrix[param.startLevel]).calculateILUdecomp();
    //(*pMatrix[param.startLevel]).calculateLUdecomp();
    (*pMatrix[param.startLevel]).calcStokesLUdecomp1();
    timer.stop();
	cout << "time to decompose: " << timer << endl;

	//cout <<"matrix 0,2\n"<<(*pMatrix[param.noOfNets - 3])[0][2] << endl;
// 	matWriteCompleteMatrix("Ah.mat",(*pMatrix[param.noOfNets - 2]),"Ah");
// 	matWriteCompleteMatrix("AH.mat",(*pMatrix[param.noOfNets - 3]),"AH");
	//matWriteBlockProlongation( "prol",(*prolongation[param.noOfNets-3]),"prol");


    // solve resulting equations

    uint preSmSteps    = 1;           // number of presmoothing steps
    uint postSmSteps   = 1;           // number of postsmoothing steps
            
    uint mgMaxLevel = param.endLevel;
    uint mgMinLevel = param.startLevel;

    // V-Cycle: cycle = 1, W - Cycle: cycle = 2 
    uint cycle = 2; 	 

    // solvers
    TLU *lu = new TLU( );
    lu->setMaxit( 1 );
    TILU *ilu = new TILU( );
    TRichardson *rich = new TRichardson( );
    TKaczmarz *kaczmarz = new TKaczmarz( );
    TSquaredSmoother *squaredSmoother = new TSquaredSmoother();
    TOptSquaredSmoother *optSquaredSmoother = new TOptSquaredSmoother();
    TOptSquaredSmoother2 *optSquaredSmoother2 = new TOptSquaredSmoother2();
    TResidualSmoother *residualSmoother = new TResidualSmoother( optSquaredSmoother );
    TSchur *schur = new TSchur();
    TBicgStab *bicg = new TBicgStab( squaredSmoother );
    TCGSolver *cg = new TCGSolver( squaredSmoother );
   
    TMultigrid *multigrid = new TMultigrid( pMatrix, pRHS, pX, 
								restriction, prolongation, 
                                mgMinLevel, mgMaxLevel,
								cycle, squaredSmoother, squaredSmoother, 
								lu, param.steps );
    multigrid->setPreSmoothingSteps( preSmSteps );
    multigrid->setPostSmoothingSteps( postSmSteps );

	TBlockVector<TFloat> *start = new TBlockVector<TFloat> (*pX[mgMaxLevel]);
	start->set( 10.0 );
	//(*start)[0].set( 0.0 );
	(*start)[2].set( 0.0 );
	*pX[mgMaxLevel] = *start;

	TBlockVector<TFloat> *error=new TBlockVector<TFloat> (*pX[mgMaxLevel]);

	*error -= *vecSolution;
	cout << "\ninitial error:\n";
	cout << "norm u_ex - u_sol: " 
		 << error->weightedEuclidBlockNorm(0) << endl;     
	cout << "norm v_ex - v_sol: "
		 << error->weightedEuclidBlockNorm(1) << endl;     
	cout << "norm p_ex - p_sol: "
		 << error->weightedEuclidBlockNorm(2) << endl;     

    uint verbosity = 5;

    bicg->setAccuracy( 1e-15 );
    bicg->setMaxit( param.steps );
    bicg->setVerbosity( verbosity );
    cg->setAccuracy( 1e-15 );
    cg->setMaxit( param.steps );
    cg->setVerbosity( verbosity );
    ilu->setMaxit( param.steps );
    ilu->setVerbosity( verbosity );
    kaczmarz->setMaxit( param.steps );
    kaczmarz->setVerbosity( verbosity );
    schur->setMaxit( param.steps );
    schur->setVerbosity( verbosity );
    schur->setProblemType( UNSYMMETRIC );
    schur->setInnerReduceBy( 1.0e-1 );
    //schur->setDampFactor( 1.0 );
    schur->setMaxInnerSteps( 20 );
    squaredSmoother->setMaxit( param.steps );
    squaredSmoother->setVerbosity( verbosity );
    squaredSmoother->setEpsilon( param.epsilon );
    //squaredSmoother->setDampFactor( 1.0 );
    optSquaredSmoother->setMaxit( param.steps );
    optSquaredSmoother->setVerbosity( verbosity );
    optSquaredSmoother2->setMaxit( param.steps );
    optSquaredSmoother2->setVerbosity( verbosity );
    residualSmoother->setMaxit( param.steps );
    residualSmoother->setVerbosity( verbosity );

	/*
	cout << "\n------------ Simple solver ------------\n";
	timer.start();
	*pX[mgMaxLevel] = *start;
	squaredSmoother->setMode( 1 );
	squaredSmoother->solve(pMatrix[mgMaxLevel], pRHS[mgMaxLevel], pX[mgMaxLevel]);
	timer.stop();
	cout << "time for squared smoother: " << timer << endl;

	timer.start();
	*pX[mgMaxLevel] = *start;
	optSquaredSmoother->setMode( 1 );
	optSquaredSmoother->solve(pMatrix[mgMaxLevel], pRHS[mgMaxLevel], pX[mgMaxLevel]);
	timer.stop();
	cout << "time for optimal squared smoother: " << timer << endl;

	timer.start();
	*pX[mgMaxLevel] = *start;
	residualSmoother->setMode( 1 );
	residualSmoother->solve(pMatrix[mgMaxLevel], pRHS[mgMaxLevel], pX[mgMaxLevel]);
	timer.stop();
	cout << "time for residual smoother: " << timer << endl;

	timer.start();
	*pX[mgMaxLevel] = *start;
	optSquaredSmoother->setMode( 1 );
	optSquaredSmoother->solve(pMatrix[mgMaxLevel], pRHS[mgMaxLevel], pX[mgMaxLevel]);
	timer.stop();
	cout << "time for optimal squared smoother: " << timer << endl;

	timer.start();
	*pX[mgMaxLevel] = *start;
	optSquaredSmoother2->setMode( 1 );
	optSquaredSmoother2->solve(pMatrix[mgMaxLevel], pRHS[mgMaxLevel], pX[mgMaxLevel]);
	timer.stop();
	cout << "time for optimal squared smoother 2: " << timer << endl;

	timer.start();
	*pX[mgMaxLevel] = *start;
	kaczmarz->setMode( 1 );
	kaczmarz->solve(pMatrix[mgMaxLevel],pRHS[mgMaxLevel],pX[mgMaxLevel]);
	timer.stop();
	cout << "time for Kaczmarz: " << timer << endl;
	*/
	timer.start();
	*pX[mgMaxLevel] = *start;
	schur->setMode( 1 );
	schur->solve(pMatrix[mgMaxLevel],pRHS[mgMaxLevel],pX[mgMaxLevel]);
	timer.stop();
	cout << "time for Schur: " << timer << endl;
	/*
	timer.start();
	*pX[mgMaxLevel] = *start;
	ilu->setMode( 1 );
	ilu->solve(pMatrix[mgMaxLevel], pRHS[mgMaxLevel], pX[mgMaxLevel]);
	timer.stop();
	cout << "time for ilu: " << timer << endl;

	cout << "\n------------ Bicg solver ------------\n";

	timer.start();
	*pX[mgMaxLevel] = *start;
	bicg->setPrec( squaredSmoother );
	bicg->solve(pMatrix[mgMaxLevel], pRHS[mgMaxLevel], pX[mgMaxLevel]);
	timer.stop();
	cout << "time for bicg: " << timer << endl;

	timer.start();
	*pX[mgMaxLevel] = *start;
	cg->setPrec( squaredSmoother );
	cg->solve(pMatrix[mgMaxLevel], pRHS[mgMaxLevel], pX[mgMaxLevel]);
	timer.stop();
	cout << "time for cg: " << timer << endl;
	*/

	  timer.start();
	  *pX[mgMaxLevel] = *start;
	  bicg->setPrec( schur );
	  bicg->solve(pMatrix[mgMaxLevel], pRHS[mgMaxLevel], pX[mgMaxLevel]);
	  timer.stop();
	  cout << "time for bicg: " << timer << endl;

	/*	
	  timer.start();
	  *pX[mgMaxLevel] = *start;
	  bicg->setPrec( ilu );
	  bicg->solve(pMatrix[mgMaxLevel], pRHS[mgMaxLevel], pX[mgMaxLevel]);
	  timer.stop();
	  cout << "time for bicg: " << timer << endl;

      cout << "\n------------ Multigrid solver ------------\n";
      timer.start();
	  squaredSmoother->setVerbosity( 0 );
	  *pX[mgMaxLevel] = *start;
	  multigrid->setSmoother( squaredSmoother );
	  multigrid->setVerbosity( verbosity );
	  multigrid->solve();
	  timer.stop();
	  cout << "time for multigrid: " << timer << endl;

	  timer.start();
	  optSquaredSmoother->setVerbosity( 0 );
	  *pX[mgMaxLevel] = *start;
	  multigrid->setSmoother( optSquaredSmoother );
	  multigrid->setVerbosity( verbosity );
	  multigrid->solve();
	  timer.stop();
	  cout << "time for multigrid: " << timer << endl;

      timer.start();
	  residualSmoother->setVerbosity( 0 );
	  *pX[mgMaxLevel] = *start;
	  multigrid->setSmoother( residualSmoother );
	  multigrid->setVerbosity( verbosity );
	  multigrid->solve();
	  timer.stop();
	  cout << "time for multigrid: " << timer << endl;

	  timer.start();
	  optSquaredSmoother2->setVerbosity( 0 );
	  *pX[mgMaxLevel] = *start;
	  multigrid->setSmoother( optSquaredSmoother2 );
	  multigrid->setVerbosity( verbosity );
	  multigrid->solve();
	  timer.stop();
	  cout << "time for multigrid: " << timer << endl;

	  multigrid->setMode( 1 );
	  timer.start();
	  kaczmarz->setVerbosity( 0 );
	  *pX[mgMaxLevel] = *start;
	  multigrid->setVerbosity( verbosity );
	  multigrid->setSmoother( kaczmarz );
	  multigrid->solve();
	  timer.stop();
	  cout << "time for multigrid: " << timer << endl;

    multigrid->setMode( 1 );
    timer.start();
    schur->setVerbosity( 0 );
    *pX[mgMaxLevel] = *start;
    multigrid->setVerbosity( verbosity );
    multigrid->setSmoother( schur );
    multigrid->solve();
    timer.stop();
	cout << "time for multigrid: " << timer << endl;
    cout << "inner bicg steps: " << schur->getTotalInnerSteps() << endl;

	multigrid->setMode( 1 );
	timer.start();
	ilu->setVerbosity( 0 );
	//ilu->setDampFactor( 1.0 );
	*pX[mgMaxLevel] = *start;
	multigrid->setVerbosity( verbosity );
	multigrid->setSmoother( ilu );
	multigrid->solve();
	timer.stop();
	cout << "time for multigrid: " << timer << endl;
	*/

	TBlockVector<TFloat> *defect = 
								   new TBlockVector<TFloat>(*pRHS[mgMaxLevel]);
	*defect -= ( (*pMatrix[mgMaxLevel]) * (*pX[mgMaxLevel]) );
	//cout << "defect:\n" << (*defect)[2] << endl;   

	cout << "\ndefect of numerical solution:\n";
	cout << "weighted defect 0 : "
		 << (*defect).weightedEuclidBlockNorm(0) << endl;
	cout << "weighted defect 1 : "
		 << (*defect).weightedEuclidBlockNorm(1) << endl;
	cout << "weighted defect 2 : "
		 << (*defect).weightedEuclidBlockNorm(2) << endl;
	//(*defect)[2][(*pRHS[mgMaxLevel])[2].getSize()-1] = 0.0;
	//cout << "weighted defect 2 without lagrange: "
	//     << (*defect).weightedEuclidBlockNorm(2) << endl;
   
	// compare with exact solution
	cout << "\nerror between exact solution and numerical solution:" << endl;
	if (param.ordering != NONE ) {
	   pX[param.noOfNets-2]->permute( vecPerm[param.noOfNets-2] );
	   pRHS[param.noOfNets-2]->permute( vecPerm[param.noOfNets-2] );
       pMatrix[param.noOfNets-2]->permute( vecPerm[param.noOfNets-2] );
	}

	*error = *pX[mgMaxLevel] - *vecSolution;
	// pressure correction
	TBlockVector<TFloat> *one=new TBlockVector<TFloat> (*pX[mgMaxLevel]);
    one->set( 1.0 );
	//cout << "factor " << ((*error)[2] * (*one)[2]) << endl << (*error)[2] << endl;
    (*one)[2] *= ((*error)[2] * (*one)[2]) / ((*one)[2] * (*one)[2]);
    (*error)[2] -= (*one)[2];

	cout << "norm u_ex - u_sol: " 
		 << error->weightedEuclidBlockNorm(0) << endl;     
	cout << "norm v_ex - v_sol: "
		 << error->weightedEuclidBlockNorm(1) << endl;     
	cout << "norm p_ex - p_sol: "
		 << error->weightedEuclidBlockNorm(2) << endl;     
   
	// exact solution
	*defect = *pMatrix[mgMaxLevel] * *vecSolution - *pRHS[mgMaxLevel];
	cout << "\ndefect of exact solution:\n";
	cout << "weighted defect 0 : "
		 << (*defect).weightedEuclidBlockNorm(0) << endl;
	cout << "weighted defect 1 : "
		 << (*defect).weightedEuclidBlockNorm(1) << endl;
	cout << "weighted defect 2 : "
		 << (*defect).weightedEuclidBlockNorm(2) << endl;

    // draw exact and numerical solution

	//net = iterNet->last();
    // draw all the usual things
	if (param.draw)  
    {
        net = iterNet->prev();
	    TDrawApp::draw( *net, "coarsecoarseNet" );
        net = iterNet->next();
	    TDrawApp::draw( *net, "coarseNet" );
	    //TDrawApp::draw( *net, *vecInvPerm[param.noOfNets-3], "coarseNetPerm" );
        net = iterNet->next();
	    TDrawApp::draw( *net, "fineNet" );
	    //TDrawApp::draw( *net, *vecInvPerm[param.noOfNets-2], "fineNetPerm" );
        TDrawApp::draw( *arrVectorsU, *pX[mgMaxLevel], *net, 
                   "vecfield" );
        TDrawApp::draw( *arrVectorsU, *vecSolution, *net, 
                   "vecfield_sol" );
        //TDrawApp::drawStreakLine( *net, *pX[mgMaxLevel], *eqn, 
        //           "streaklines");
    }// if


    delete (arrVectorsU);
    delete (arrVectorsP);
   
    delete (netHier);
    delete (iterNet);
   
    delete (lu);
    delete (ilu);
    delete (rich);
    delete (kaczmarz);
    delete (bicg);
    delete (cg);
    delete (multigrid);
    delete (schur);
    delete (squaredSmoother);
    delete (optSquaredSmoother);
    delete (optSquaredSmoother2);
    delete (residualSmoother);
 
    delete (dofManager);
    delete (fe);
    delete (eqn);
    delete (hierarchy);
    
    delete (defect);
    delete (vecSolution);
    delete (error);

    delete (arrFVSize);
    delete (blockinfo);
    if ( param.ordering != NONE )
        for ( i = 1; i < param.noOfNets-1; i++ ) { 
            delete (vecInvPerm[i]);
            delete (vecPerm[i]);
            delete (cyclicBlocks[i]);
		}

    for (i = 0; i < mgMaxLevel+1; i++) {
		if (pX[i]) delete pX[i]; 
		if (pRHS[i]) delete pRHS[i];  
		if (pMatrix[i]) delete pMatrix[i];  
	}
	for (i = 0; i < mgMaxLevel; i++) {
		if (prolongation[i]) delete (prolongation[i]);
		if (restriction[i]) delete (restriction[i]);
	}
 
    TMemory::statistics();
	
}
