//
//  Project   : Navier-Stokes
//  File      : stokesConv.cc
//  Author    : sg/tpr/rok
//  Purpose   : test of 3D stokes
//
// arch-tag: 018e52dc-516f-437e-be40-3d215592659f
//

#include "misc/memory.hh"
#include "misc/TFunction.hh"

#include "net/net.hh"
#include "net/TCdrBoundaryDesc.hh"
#include "net/TStokesBoundaryDesc.hh"

#include "discr/TCdrEqn.hh"
#include "discr/TConvDiffDoFManager.hh"
#include "discr/TConvDiffFE.hh"

#include "matrix/blockvector.hh"

#include "discr/TStokesDoFManager.hh"
#include "discr/TStokesBasisEqn.hh"
#include "discr/TStokesFE.hh"

#include "solver/secondNF.hh"
#include "solver/cgSolver.hh"
#include "solver/bicgStab.hh"
#include "solver/cgSolver.hh"
#include "solver/schur.hh"

#include "visual/drawapp.hh"

class TDirichletFun0 : public TFunction<3>
{
public:
    TFloat eval(TVector<3> *v )
	{ 
        return ( sin((*v)[0]) * sin((*v)[1]) );
        //return ( 0.0 );
	}
};

class TDirichletFun1 : public TFunction<3>
{
public:
    TFloat eval(TVector <3> *v )
	{
        return  ( cos((*v)[0]) * cos((*v)[1]) );
        //return ( 0.0 );
	}
};

class TDirichletFun2 : public TFunction<3>
{
public:
    TFloat eval(TVector <3> *v )
	{
        return (0.0);
	}
};

class TPressure : public TFunction<3>
{
public:
    TPressure(){};
    ~TPressure(){};
    TFloat eval(TVector <3> *av )
	{
        return (2.0 * cos((*av)[0]) * sin((*av)[1]) -
                2.0 * sin(1.0) * (1.0 - cos(1.0)));
	}
};

class TMyStokesBasisEqn : public TStokesBasisEqn<3>
{
public:

    TMyStokesBasisEqn( TFloat aEpsilon = 1.0, 
                       TBlockVector<TFloat> *aConvVector = NULL)
            : TStokesBasisEqn<3>( aEpsilon, aConvVector) {};

    virtual ~TMyStokesBasisEqn () { };
      
    void evalConv ( TVector<3> *v, TVector<3> *b ) const
    {
        if (b != NULL)
        {
            (*b)[X] = 10.0;
            (*b)[Y] = 0.0;
            (*b)[Z] = 0;
        }// if
	}

    TFloat evalRhs(TVector<3> *v, uint i = 0 ) const
	{
        TVector<3> b;
        evalConv( v, & b );
		TFloat value = 0.0;
        
        if (i == 0)
            value = 2.0 * ( epsilon - 1.0 ) *
                sin((*v)[0]) * sin((*v)[1]) + (*b)[0] * cos((*v)[0]) *
                sin((*v)[1]) + (*b)[1] * sin((*v)[0]) * cos((*v)[1]);
        
        if (i == 1)
            value =  2.0 * ( 1.0 + epsilon) *
                cos((*v)[0]) * cos((*v)[1]) - (*b)[0] * sin((*v)[0]) *
                cos((*v)[1]) - (*b)[1] * cos((*v)[0]) * sin((*v)[1]);

        return (value);
	}  
};

const uint spaceDim = 3;

//
// mainfunction
//
int
main( int argc, char **argv ) 
{
    TNet<spaceDim>  *net = new TNet<spaceDim>( "grids/cube.tri" );
    TNet<spaceDim>  *fnet[10];      

    uint i;
    TDrawApp *app = new TDrawApp;
   
  
    TMyStokesBasisEqn* stokesEqn = new TMyStokesBasisEqn( 1.0, NULL );

    TDirichletFun0* dirichlet0 = new TDirichletFun0();
    TDirichletFun1* dirichlet1 = new TDirichletFun1();
    TDirichletFun2* dirichlet2 = new TDirichletFun2();
    TPressure* pressure = new TPressure();
   
  
    // supply boundary Descriptons

    // in "grids/cube.tri", only one TBoundary is defined
    // therefore, one Description is enough

    TStokesBoundaryDesc<spaceDim> *boundDesc = new TStokesBoundaryDesc<spaceDim>( spaceDim+1 );
    for (i = 0; i < spaceDim; i++)
    {
        boundDesc->setType( i, DIR_TYPE);
        //boundDesc->setFunc( i, dirichlet);
        boundDesc->setDoFInfo( i, FALSE);
    }
    boundDesc->setFunc( 0, dirichlet0);
    boundDesc->setFunc( 1, dirichlet1);
    boundDesc->setFunc( 2, dirichlet2);

    boundDesc->setType( spaceDim, NEU_TYPE);
    //boundDesc->setFunc( spaceDim, Neumann);
    boundDesc->setDoFInfo( spaceDim, TRUE);

    TNet<spaceDim>::TCellListIterator iter( net->getList() );
    bool ready = FALSE;
   
    // find node with boundary pointer
    TCell<spaceDim> *cell = iter.first();
    TBoundary *bound;
    while ((cell != NULL) && (ready == FALSE))
    {
        for ( i = 0; i <= spaceDim; i++ )
        {
            bound = cell->getNode(i)->getBoundary();
            if (bound != NULL) ready = TRUE; // exit loop
        }
        cell = iter.next();
    }
    
    cout << "bound found" << endl;
    bound->setBoundaryDesc( boundDesc );
    cout << "bound set" << endl;
  
    fnet[0] = net->refine(); 
  
    uint mf = 1;
    for (i = 0; i <= mf; i++) 
    {
        cout << "mf: " << i << endl;
        fnet[i+1] = fnet[i]->refine();
        cout << "Refinement done." << endl;
    }

    //app.drawNet( net, "psfiles/cnet" );
    //app.drawNet( fnet, "psfiles/fnet" );
    //app.drawNet( ffnet, "psfiles/ffnet" );

    TStokesDoFManager<spaceDim> *dofManager = new TStokesDoFManager<spaceDim>();
    TBlockMatrix *A = dofManager->buildStiffnessMatrix( fnet[mf], fnet[mf+1]);

    //cout<<"A:\n" << (*A)[0][0] << endl;

    ulong dim = 0;
    ulong u_dim = (*A)[0][0].getNoOfRows();
    ulong p_dim = (*A)[3][0].getNoOfRows();
    cout << "\nMatrixdimension" << endl;
    for (i = 0; i < spaceDim; i++)
    {
        dim += (*A)[i][i].getNoOfRows();
        cout << (*A)[i][i].getNoOfRows() << endl;
    }
    dim += (*A)[spaceDim][0].getNoOfRows();
    cout << (*A)[spaceDim][0].getNoOfRows() << endl;
 
    cout << "Altogether: " << dim << endl;

    ulong size[spaceDim+1]; 
    for ( i = 0; i < spaceDim; i++) 
        size[i] = (*A)[i][i].getNoOfRows();
    size[spaceDim] = (*A)[spaceDim][0].getNoOfRows();

    TBlockVector<TFloat> *rhs= new TBlockVector<TFloat>( spaceDim+1, size);
    TBlockVector<TFloat> *pX = new TBlockVector<TFloat>( spaceDim+1, size);
    TBlockVector<TFloat> *pOne = new TBlockVector<TFloat>( spaceDim+1, size);
    TBlockVector<TFloat> *perror = new TBlockVector<TFloat>( spaceDim+1, size);
    TBlockVector<TFloat> *pexact = new TBlockVector<TFloat>( spaceDim+1, size);
    TBlockVector<TFloat> *presidual = new TBlockVector<TFloat>( spaceDim+1, size);

    // discretization
    cout << "begin discretization\n";
    TStokesFE<spaceDim> *discretiser = new TStokesFE<spaceDim>( Tabata );
    //app->drawNet( fnet[mf], "psfiles/ffcube", VRML );
   
    // TDynArray<TVector <3> * > coords(dim);
    //discretiser->getCoord(dnet,coords);
    //for (uint i=0;i<dim;i++)
    //   cout<< "i: "<<i<<" "<<*(coords[i])<<endl;  

    TMatrixBlock *C = new TMatrixBlock( (*A)[0][0] );
    discretiser->discrDiff( stokesEqn, fnet[mf], A, rhs);
    discretiser->discrConv( stokesEqn, fnet[mf], C, rhs);
    discretiser->discrRhs( stokesEqn, fnet[mf], rhs);
    cout << "discretization done\n";

	//cout << "rhs\n" << (*rhs)[3] << endl;
	//cout << "convection\n" << C << endl;

    for ( uint j = 0; j < 3; j++)
        (*A)[j][j] += *C;

    //cout<<"A0:\n" << (*A)[0][0] << endl;
    //cout<<"A1:\n" << (*A)[1][1] << endl;
    //cout<<"A2:\n" << (*A)[3][2] << endl;
    //cout<<"C: "<<C<<endl;
    //cout << "rhs0: " << endl << (*rhs)[0] << endl;
    
    
    // test discretisation:
    TDynArray<TVector<3> * > arrCoord( u_dim );
    TDynArray<TVector<3> * > parrCoord( p_dim );
    
    discretiser->getCoord ( fnet[mf], 0, arrCoord, parrCoord );
    //cout << "size of fine grid: " << arrCoord.getSize() << endl;
    //for (i = 0; i < arrCoord.getSize(); i++)
    //    cout << "i: " << i << " " << *(arrCoord[i]) << endl;
    
      
    // solve
    TSchur *schur = new TSchur( );
    schur->setMaxInnerSteps( 10 );
    schur->setInnerReduceBy( 1.0e-1 );
    schur->setVerbosity( 2 );
    schur->setAccuracy( 1e-8 );
    schur->setMaxit( 100 );

    TSymmetricGaussSeidel *ssgs = new TSymmetricGaussSeidel( );
    TRichardson *rich = new TRichardson( );
    TBicgStab *bicgStab = new TBicgStab( schur );
    TCGSolver *cg = new TCGSolver( rich );
    bicgStab->setAccuracy( 1e-8 );
    bicgStab->setMaxit( 10 );
    bicgStab->setVerbosity( 2 );
    
    cg->setAccuracy( 1e-15 );
    cg->setMaxit( 100 );
    cg->setVerbosity( 4 );
    pX->set( 2.0 ); 
    pOne->set( 0.0 ); 
    (*pOne)[3].set( 1.0 ); 

    for (ulong j = 0; j < u_dim; j++)
	{
        (*pexact)[0][j] = dirichlet0->eval( arrCoord[j] );
        (*pexact)[1][j] = dirichlet1->eval( arrCoord[j] );
        (*pexact)[2][j] = dirichlet2->eval( arrCoord[j] );
    }
    for (ulong j = 0; j < p_dim; j++)
        (*pexact)[3][j] = pressure->eval( parrCoord[j] );

    TFloat pressureCorrection = 0.0;
    pressureCorrection -= ( (*pexact)[3] * (*pOne)[3] ) / ( (*pOne)[3] * (*pOne)[3] );
    (*pOne)[3] *= pressureCorrection;
    (*pexact)[3] += (*pOne)[3];
    (*pOne)[3] *= (1.0 / pressureCorrection);

    bicgStab->solve(A, rhs, pX);
    //cg->solve(A, rhs, pX);
    //schur->solve(A, rhs, pX);
    
    (*presidual) = (*A) * (*pexact) - (*rhs);
    cout << "Residual of exact solution: " << setprecision(5)
         << presidual->weightedEuclidNorm() << endl;;

    (*presidual) = (*A) * (*pX) - (*rhs);
    cout << "Residual of numerical solution: " << setprecision(5)
         << presidual->weightedEuclidNorm() << endl;

    pressureCorrection = ( (*pX)[3] * (*pOne)[3] ) / ( (*pOne)[3] * (*pOne)[3] );
    (*pOne)[3] *= pressureCorrection;
    (*pX)[3] -= (*pOne)[3];
    (*pOne)[3] *= (1.0 / pressureCorrection);

    (*perror) = (*pX) - (*pexact);
    cout << "Fehler Diff: " << setprecision(5)
         << perror->weightedEuclidBlockNorm(0)
         << "  " << perror->weightedEuclidBlockNorm(1)
         << "  " << perror->weightedEuclidBlockNorm(2)
         << "  " << perror->weightedEuclidBlockNorm(3)
         << "  " << endl;
    
    delete discretiser;
    delete rhs;
    delete pX;
    delete pOne;
    delete presidual;
    delete perror;
    delete A;
    delete C;
    delete dofManager;

    delete schur;
    delete bicgStab;
    delete cg;
    delete rich;
    delete ssgs;

    delete app;

    delete boundDesc;

    delete stokesEqn;
    delete dirichlet0;
    delete dirichlet1;
    delete dirichlet2;
    delete pressure;

    net->free();
    for (i = 0; i <= mf+1; i++) 
        delete fnet[i];
    delete[] fnet;

    TMemory::statistics();
}
 
