//
//  Project   : Navier-Stokes
//  File      : stokesConv.cc
//  Author    : Ronald Kriemann
//  Purpose   : test of 3D grids
//
// arch-tag: f532941a-5a0a-4968-a46b-44d286d67886
//

#include "misc/memory.hh"
#include "misc/TFunction.hh"

#include "net/net.hh"
#include "net/TCdrBoundaryDesc.hh"

#include "discr/TCdrEqn.hh"
#include "discr/TConvDiffDoFManager.hh"
#include "discr/TConvDiffFE.hh"
#include "matrix/blockvector.hh"

#include "solver/secondNF.hh"
#include "solver/cgSolver.hh"
#include "solver/bicgStab.hh"

#include "visual/drawapp.hh"
#include "visual/TPOVRay.hh"

class TDirichletFun : public TFunction<3>{
public:
   TDirichletFun(){};
   ~TDirichletFun(){};
   TFloat eval(TVector <3> *av ){
      TVector <3> *v; v=av; 
      TFloat a=(*v)[0];
      TFloat b=(*v)[1];
      TFloat c=(*v)[2];
      //  return a+b+2*c;
           return a*a+b*b+c*c;
  }
   
};
class TZeroFun : public TFunction<3>{
public:
   TZeroFun(){};
   ~TZeroFun(){};
   TFloat eval(TVector <3> *av ){
      TVector <3> *v; v=av; return 0.0;
   }
   
};
class TOneFun : public TFunction<3>{
public:
   TOneFun(){};
   ~TOneFun(){};
   TFloat eval(TVector <3> *av ){
      TVector <3> *v; v=av; return 1.0;
   }
   
};
class TTwoFun : public TFunction<3>{
public:
   TTwoFun(){};
   ~TTwoFun(){};
   TFloat eval(TVector <3> *av ){
      TVector <3> *v; v=av; return 2.0;
   }
   
};

class TMyBasisCdrEqn : public TBasisCdrEqn<3>{

public:

   TMyBasisCdrEqn( TFloat aEpsilon = 1.0, 
                   TBlockVector<TFloat> *aConvVector = NULL):TBasisCdrEqn<3>(aEpsilon,aConvVector)
   {};

   virtual ~TMyBasisCdrEqn () { }
   
   void evalDiff(TVector <3> *v, TMatrix<3>  *D){
      TVector<3> * av; av=v;
      (*D)[0][0]=1;(*D)[0][1]=0; (*D)[0][2]=0;
      (*D)[1][0]=0;(*D)[1][1]=1; (*D)[1][2]=0;
      (*D)[2][0]=0;(*D)[2][1]=0; (*D)[2][2]=1;
   }

   void evalConv(TVector <3> *v, TVector<3>  *B){
      TVector<3> * av; av=v;
      (*B)[0]=1;(*B)[1]=2; (*B)[2]=3;
   }


   TFloat evalReac(TVector <3> *v) const{
      TVector<3> * av; av=v;
      return 0.0;
   }
   
   TFloat evalRhs(TVector <3> *v ) const{
      TVector<3> * av; av=v;
     TFloat a=(*v)[0];
      TFloat b=(*v)[1];
      TFloat c=(*v)[2];
     
        return -6;
   }   
};

class TMyCdrEqn : public TCdrEqn<3> {
public:

   TMyCdrEqn( TFloat aEpsilon = 1.0, 
              TBlockVector<TFloat> *aConvVector = NULL): TCdrEqn<3>(aEpsilon,aConvVector)
   {};

   virtual ~TMyCdrEqn () { }
};


//
// mainfunction
//
int
main( int argc, char **argv ) 
{
  TNet<3>  *net3D = new TNet<3>( "grids/cube.tri" ),   *dnet;
  TNet<3>  *fnet[10];      
  
  
   TDrawApp *app = new TDrawApp;
   
  
   TMyBasisCdrEqn*  bcdreqn = new TMyBasisCdrEqn(1.0,NULL);
   TVector<3> x(0.1,2,9);
   TVector<3> b;
 
   bcdreqn->evalConv(&x,&b);
   cout<<"b: "<<b<<endl;

   
   TMyBasisCdrEqn*  cdreqn = new TMyBasisCdrEqn(1.0,NULL);
   cout<<"rhs1: "<<cdreqn->evalRhs(&x)<<endl;
  
   TDirichletFun* Dirichlet = new TDirichletFun;
   TZeroFun* Zero = new TZeroFun;
   TOneFun* One = new TOneFun;
   TTwoFun* Two = new TTwoFun;
   
 
   /*   cdreqn->setRhsFunction(One);
  cdreqn->setRhsFunction(Dirichlet);
        cdreqn->setDiffFunctions(One,0,0);
        cdreqn->setDiffFunctions(Zero,0,1);
        cdreqn->setDiffFunctions(Zero,0,2);
        cdreqn->setDiffFunctions(Two,1,1);
        cdreqn->setDiffFunctions(Zero,1,0);
        cdreqn->setDiffFunctions(Zero,1,2);

        cdreqn->setDiffFunctions(One,2,2);
        cdreqn->setDiffFunctions(Zero,2,1);
        cdreqn->setDiffFunctions(Zero,2,0);

        */


   cout<<"rhs2: "<<cdreqn->evalRhs(&x)<<endl;

   cdreqn->isLaplace(FALSE);
   cdreqn->Diffusion(TRUE);
   cdreqn->Convection(FALSE);
   cdreqn->setEpsilon(1);
 
  
   // supply boundary Descriptons

   // in "grids/cube.tri", only one TBoundary is defined
   // therefore, one Description is enough

   TCdrBoundaryDesc<3> * boundDesc = new TCdrBoundaryDesc<3>(1);
   boundDesc->setType(0,DIR_TYPE);
   boundDesc->setFunc(0,Dirichlet);
   boundDesc->setDoFInfo(0,FALSE);

   TNet<3>::TCellListIterator iter( net3D->getList() );
   bool ready=FALSE;
   
   // find node with boundary pointer
   TCell<3> *cell = iter.first();
   TBoundary *bound;
   while ((cell != NULL)&&(ready==FALSE)){
      for (uint i=0;i<4;i++){
         bound =cell->getNode(i)->getBoundary();
         if (bound!=NULL){
            ready=TRUE; // exit loop
         }
      }
      cell = iter.next();
   }
    
   cout<<"bound found"<<endl;
   bound->setBoundaryDesc(boundDesc);
   cout<<"bound set"<<endl;
  
  fnet[0] = net3D->refine(); 
  
  int mf=3;
  for (uint i=0;i<=mf;i++) {
     cout<<"mf: "<<i<<endl;
     fnet[i+1]=fnet[i]->refine();
     dnet=fnet[i+1];
     cout<<" Verfeinern fertig"<<endl;

   // app.drawNet( net3D, "psfiles/cnet" );
   //app.drawNet( fnet, "psfiles/fnet" );
   //app.drawNet( ffnet, "psfiles/ffnet" );

   TConvDiffDoFManager<3> *DoFManager= new TConvDiffDoFManager<3>();
   DoFManager->InitDoFDesc(dnet);
   TBlockMatrix *A=DoFManager->buildStiffnessMatrix(dnet,dnet);

   ulong dim=(*A)[0][0].getNoOfRows();
   cout<<"A: Dimension "<<dim<<endl;
  
   ulong size[1];    size[0]=dim;
   TBlockVector<TFloat> *rhs= new TBlockVector<TFloat>(1,size);
   TBlockVector<TFloat> *pX= new TBlockVector<TFloat>(1,size);
   TBlockVector<TFloat> *perror= new TBlockVector<TFloat>(1,size);
   TBlockVector<TFloat> *pexact= new TBlockVector<TFloat>(1,size);
   // Diskretisierung
  
   TConvDiffFE<3> *discretiser= new TConvDiffFE<3>(Tabata);
   // app->drawNet( dnet, "psfiles/ffcube", VRML );
   
   // TDynArray<TVector <3> * > coords(dim);
   //discretiser->getCoord(dnet,coords);
   //for (uint i=0;i<dim;i++)
   //   cout<< "i: "<<i<<" "<<*(coords[i])<<endl;

    TMatrixBlock C=(*A)[0][0];
  

    discretiser->discrDiff(cdreqn,dnet,A,rhs);
    //     discretiser->discrConv(cdreqn,dnet,&C,rhs);
    discretiser->discrRhs(cdreqn,dnet,rhs);
    
    //    cout<<"A: "<<(*A)[0][0]<<endl;
    // cout<<"C: "<<C<<endl;
    // cout<<"rhs: "<<endl<<(*rhs)[0]<<endl;
    
    
    // test discretisation:

    TDynArray<TVector < 3 > * > arrCoord(dim);
    
    discretiser->getCoord ( dnet,arrCoord);
    cout<<"size: "<<arrCoord.getSize()<<endl;
    
    TBlockMatrix* CC= new TBlockMatrix(C);
    
    // solve
    TSymmetricGaussSeidel *ssgs = new TSymmetricGaussSeidel( );
    TBicgStab *bicgStab = new TBicgStab(ssgs );
    bicgStab->setAccuracy( 1e-15 );
    bicgStab->setMaxit( 1000 );
    bicgStab->setVerbosity( 4 );
    pX->set( 1.0 ); 
    bicgStab->solve(A,rhs,pX);
    
    
    for (ulong j=0; j<dim;j++){
       (*pexact)[0][j]= Dirichlet->eval ((arrCoord[j]));
       //  cout<<" X: "<<(*pX)[0][j]<<" coord: "<<*(arrCoord[j])<<endl;
    }
    
    (*perror)=(*pX)-(*pexact);
    cout<<"Fehler Diff: "<<setprecision(5)<<perror->euclidNorm()/sqrt(dim)<<endl;
    
    //    cout<<"rhs: "<<(*rhs)[0]<<endl;
    //  cout<<"A*exact "<<((*A)*(*pexact))[0]<<endl;
    
    
   //  cout<<(*perror)[0]<<endl;
   /*   
    (*perror)[0]=C*(*pX)[0];
     cout<<"Fehler Conv1: "<<setprecision(5)<<perror->euclidNorm()/sqrt(dim)<<endl;
    (*perror)[0]=C*(*pX)[0]-(*rhs)[0];
     cout<<"Fehler Conv2: "<<setprecision(5)<<perror->euclidNorm()/sqrt(dim)<<endl;
     cout<<"rhs: "<<setprecision(5)<<rhs->euclidNorm()/sqrt(dim)<<endl;
     */
   //  cout<<(*perror)[0]<<endl;
   
delete discretiser;
delete rhs;
delete pX;
delete A;
delete DoFManager;
delete perror;

  }
   if (net3D->delRef()) delete net3D;
   delete[] fnet;
  
  

   


   TMemory::statistics();
}
 
