//
//  Project  : Navier-Stokes
//  File     : convDiff.cc
//  Author   : Ronald Kriemann / Sabine Gutsch Thomas Probst
//  Purpose  : test for discretization and multigrid solution
//             of a convection-diffusion problem; 
//             comparison with exact solution
//
// arch-tag: 3d87ab73-4442-4859-88a8-a4dc360d3437
//

//  Ausgabe von aufeinanderfolgenmden Iterierten sowie der Koordinaten
//  des Netes fuer Matlab und Gnuplot (Datei: /Matlab/eigenvektoren.txt)


#include <stdio.h>

#include "misc/timer.hh"
#include "misc/list.hh"
#include "misc/cmdline.hh"
#include "misc/matlab.hh"

#include "net/vertex.hh"
#include "net/regular.hh"
#include "net/net.hh"

#include "solver/secondNF.hh"
#include "solver/cgSolver.hh"
#include "solver/bicgStab.hh"
#include "solver/kaczmarz.hh"
#include "solver/multigrid.hh"
#include "solver/blocksolver.hh"
#include "solver/schur.hh"
#include "solver/fastGaussSeidel.hh"

#include "discr/convDiffDoFManager.hh"
#include "discr/convDiffFE.hh"
#include "discr/convDiffHierarchy.hh"
#include "discr/convDirections.hh"

#include "matrix/blockmatrix.hh"

#include "visual/drawapp.hh"

#include "graph/crosswind.hh"
#include "graph/nodes.hh"
#include "graph/criterion.hh"
#include "graph/graphtool.hh"
#include "graph/rcm.hh"
#include "graph/fvs.hh"

// exact solution: e := x^2 + y^2
// rhs: L e = 
//      -Laplace e + b0 d/dx e + b1 d/dy e = -4 + z0 * 2 * x + z1 * 2 * y

class TMyConvDiffEqn : public TBasisEqn<2> {

public:
    TMyConvDiffEqn ( ) 
        : TBasisEqn<2> ( ) { };
    virtual TFloat f0 (const TVector<2> &v) {
         return ( -4 * epsilon + (b0(v) * 2 * v[0] + b1(v) * 2 * v[1] ) + c(v)*dirichlet0(v) ); 
       
   }
    virtual TFloat dirichlet0 ( const TVector<2> &v) { 
           return ( v[0] * v[0] + v[1] * v[1] );
    }
};

int 
main( int argc, char **argv ) 
{
    uint i = 0;
   
    param.steps = 50;
    param.noOfNets   = 5 ;
    param.startLevel = 0;
      param.ordering   = CONCENTRIC_MIN;
    //param.ordering   = NONE;
     param.steps      = 100;
    //param.ordering   = _NONE_;
    //param.ordering   = _RCM_;
    //param.ordering   = _FVS_;
    //param.xi         = 1.0/3.0;
    param.xi         = 0.8;
    //param.mgSort     = _MGSORT_PROJECTED_; 
    //param.mgSort     = _MGSORT_NESTED_; 
    param.mgSort     = MGSORT_INDEP; 
    param.criterion  = E8;
    param.epsilon    = 1e-9;
    param.draw       = FALSE;
    param.convDirName = "circle";
    strcpy(param.starttri, "grids/start.tri" );

    param.evalCmdLine( argc, argv );

    param.endLevel = param.noOfNets-1;
    cout << "chosen parameter: " << param << endl;
    //param.output();

    //if (param.ordering == _RCM_) param.xi = 0.0;

    //
    // build and refine net
    //

    TTimer timer;
    T2DNetHierarchy *netHier = new T2DNetHierarchy( param.starttri );
    netHier->delReference();
    
    for (i = 0; i < param.noOfNets-1; i++)
        netHier->buildNet();
    
    ulong noOfInteriorNodesNet[param.noOfNets];
    
    T2DNetHierarchy::TNetListIterator *iterNet;
    T2DNet *pNet;
    i = 0;
    iterNet = new T2DNetHierarchy::TNetListIterator( netHier->getList() );
    iterNet->first();
    while ((pNet = (*iterNet)++) != NULL)
    {
        noOfInteriorNodesNet[i++] = pNet->getNoOfInteriorNodes();
        cout << "net " << i-1 << ": " << noOfInteriorNodesNet[i-1]<<endl; 
    }// while
    
    T2DNet  *fNet = iterNet->last();
    
    timer.start();
    
    //
    // build stiffness matrix sparsity structure
    //

    TBlockMatrix **pMatrix = new (TBlockMatrix*)[param.noOfNets];;
    TProlongation **prolongation = new (TProlongation*)[param.noOfNets-1];
    TBlockVector<TFloat> **pRHS = new (TBlockVector<TFloat>*)[param.noOfNets];
    TBlockVector<TFloat> **pX = new (TBlockVector<TFloat>*)[param.noOfNets];

 
    T2DConfDiffDoFManager *dofManager = new T2DConfDiffDoFManager();

    TMyConvDiffEqn *eqn = new TMyConvDiffEqn( );
    // set epsilon
    eqn->setEpsilon( param.epsilon );
    // set reaction term
    if (param.reactionTerm)
        eqn->setReacFunction( oneReac ); 
    else 
        eqn->setReacFunction( noReac ); 

    // Pointer to functions for convection direction
    if (param.convDirName == "circle")
        eqn->setConvFunctions( circle0, circle1 ); 
    if (param.convDirName == "xline")
        eqn->setConvFunctions( xline0, xline1 ); 
    if (param.convDirName == "curve")
        eqn->setConvFunctions( curve0, curve1 ); 
    if (param.convDirName == "square") 
        eqn->setConvFunctions( square0, square1 );    
    if (param.convDirName == "foursquares") 
        eqn->setConvFunctions( foursquares0, foursquares1 ); 
    if (param.convDirName == "fourcircles") 
        eqn->setConvFunctions( fourcircles0, fourcircles1 ); 
   
 
    T2DConvDiffFE *fe = new T2DConvDiffFE( Tabata );
    T2DConvDiffHierarchy *hierarchy = new T2DConvDiffHierarchy();

    hierarchy->buildHierarchy( dofManager, eqn, fe, netHier, pMatrix,
                               pRHS, pX, prolongation );

    timer.stop();
    cout << "time for disretization : " << timer << endl;
    //cout << "matrix:\n" << (*pMatrix[param.noOfNets-1])[0][0] << endl;  
 
    
    TDynArray<T2DVector *> *arrVectors = new TDynArray<T2DVector *>;
    fe->getCoord( fNet, *arrVectors );
    
    timer.start();
    
    ///////////////////////////////////////////////////
    //
    // lets draw a little bit
    //
    TDrawApp *app = new TDrawApp;
  app->draw( *fNet, "net" );
    ///////////////////////////////////////////////////
    //
    // build graph from matrix and work with it
    //
    TVectorBlock<ulong> *vecPerm[param.noOfNets];
 
    uint iNets = 0;
    iterNet->first();
   
 
    while ((pRHS[iNets]==NULL ) || (pRHS[iNets]->getSize()<10 )){
        (*iterNet)++;
        iNets++;
        vecPerm[iNets]=NULL;
    }
    uint activeNets=iNets;
    cout<<" active nets: "<<activeNets<<endl;
    uint fvss=0;
    if ( param.ordering != NONE) {
        while ((pNet = (*iterNet)++) != NULL) {
            vecPerm[iNets]=NULL;
            TDynArray<T2DVector *> *arrNetVectors     = new TDynArray<T2DVector *>;
            

            fe->getCoord( pNet, *arrNetVectors );
            eqn->initConvVector(pRHS[iNets]->getSize(), *arrNetVectors);
     
            TBlockVector<TFloat> *Convection=eqn->getConvVector();
            TVectorBlock<TFloat> c0=(*Convection)[0];
            //  cout<<" Convection: "<<c0<<endl;
        
            // cout<<" Convection: "<<endl<< *Convection<<endl;
            //cout<<" c0: "<< c0<<endl;
      
            TDynArray<T2DVector *> *arrNetConvVectors =  new TDynArray<T2DVector *>(arrNetVectors->getSize());
    
            for (uint j=0; j<arrNetVectors->getSize();j++){
                T2DVector *c=new  T2DVector((*Convection)[0][j],(*Convection)[1][j]);
                (*arrNetConvVectors)[j]=c;
            }

      

            // create graph
            TMatrixBlock *pA=&((*pMatrix[iNets])[0][0]);
            vecPerm[iNets] = new TVectorBlock<ulong>( 0, pRHS[iNets]->getSize() );
            
            //   fvss=cwOrdering(pA,*arrNetVectors,*arrNetConvVectors,*vecPerm[iNets],
            //              0.3,E8, 0.2, E4, 0.2, E3);
            cwOrdering(pA,*arrNetVectors,*arrNetConvVectors,*vecPerm[iNets],
                       0.2,E8, 0.2, E4, 0.2, E3);
        
            eqn->deleteConvVector();
            delete arrNetVectors;
            delete arrNetConvVectors;
            cout <<"ready with net  "<<iNets<<endl<<endl;
            iNets++;
        }
        timer.stop();
        cout << "time for cw ordering : " << timer << endl;
    }
   
    /*  app->draw( *fNet, "psfiles/net.ps" );
        // draw some things
            if ( arrVectors->getSize() < 1200 ) 
            {
            app->draw( *fNet, "psfiles/net.ps" );
            eqn->initConvVector(pRHS[iNets-1]->getSize(), *arrVectors);
            *(eqn->getConvVector()) *= (-1);
            app->draw( *arrVectors, *(eqn->getConvVector()), *fNet, "psfiles/vecField.ps");
            eqn->deleteConvVector();
            }
            */
    //
    // build the inverse permutations
    //
    TVectorBlock<ulong> *vecInvPerm[param.noOfNets];
   
    
    if ( param.ordering != NONE)    {
        cout << "building inverse permutations" << endl;
        for (i=activeNets; i<param.noOfNets;i++){
            vecInvPerm[i]=NULL;
            if  (vecPerm[i] != NULL){
                ulong size = vecPerm[i]->getSize();
                if (size>0){
                    vecInvPerm[i] = new TVectorBlock<ulong>( 0, size );
                    invertPerm( (*vecInvPerm[i]), *vecPerm[i] );
                }
            }
        }// for 
    }// if

  
    //    matWriteMatrixBlock("/tmp/tpr/matrixBf.mat", (*pMatrix[param.noOfNets-1])[0][0],
    //                    (*pMatrix[param.noOfNets-1])[0][0].getNoOfRows(),"Bf");
    //    matWriteMatrixBlock("/tmp/tpr/matrixBc.mat", (*pMatrix[param.noOfNets-2])[0][0],
    //                    (*pMatrix[param.noOfNets-2])[0][0].getNoOfRows(),"Bc");
  // reorder
    if (param.ordering != NONE ) {
        timer.start();
        for ( i = activeNets; i < param.noOfNets; i++ ) 
        {   cout<<" permuting level "<< i<<endl;
            (*pRHS[i])[0].permute( vecInvPerm[i] );
            (*pMatrix[i])[0][0].permute( vecInvPerm[i], vecInvPerm[i] );
            if ( i > activeNets ){
              cout<<" permuting prolong on level "<< i<<endl;
              (prolongation[i-1]->pBlockProlong)[0]->permute( vecInvPerm[i], vecInvPerm[i-1]);
       
} }
        timer.stop();
        cout << "time for permuting matrices and rhs's: "<< timer << endl;
    }
     
    matWriteMatrixBlock("/tmp/tpr/matrixf.mat", (*pMatrix[param.noOfNets-1])[0][0],
                        (*pMatrix[param.noOfNets-1])[0][0].getNoOfRows(),"Af");
   
    //  matWriteMatrixBlock("/tmp/tpr/matrixf.mat", (*pMatrix[param.noOfNets-1])[0][0],
    //                    (*pMatrix[param.noOfNets-1])[0][0].getNoOfRows(),"Af");
    //matWriteMatrixBlock("/tmp/tpr/matrixc.mat", (*pMatrix[param.noOfNets-2])[0][0],
    //                    (*pMatrix[param.noOfNets-2])[0][0].getNoOfRows(),"Ac");
    //matWriteBlockProlongation("/tmp/tpr/matrixp.mat" ,(*prolongation[param.noOfNets-2]),"p" );



    /////////////////////////////////////
    //
    // solver
    //

    TSchur *schur = new TSchur( );
    TJacobi *jacobi = new TJacobi( );
    jacobi->setDampFactor( 0.5 );
    TRichardson *rich = new TRichardson( );
    //TKaczmarz *kaczmarz = new TKaczmarz( );
    TBackwardGaussSeidel *bgs = new TBackwardGaussSeidel( );
    TSymmetricGaussSeidel *sgs = new TSymmetricGaussSeidel( );

    TFastForwardGaussSeidel *fastfgs = new TFastForwardGaussSeidel( );
    TFastBackwardGaussSeidel *fastbgs = new TFastBackwardGaussSeidel( );
    TFastSymmetricGaussSeidel *fastsgs = new TFastSymmetricGaussSeidel();

    TBicgStab *bicgStab = new TBicgStab( sgs );
    TBlockSymmetricGS *blocksgs = new TBlockSymmetricGS( );
    //sgs->setDampFactor( 0.5 );
    //blocksgs->setDampFactor( 0.5 );

    uint startLevel=activeNets;
    // increase startlevel if nets are too coarse; at least 10 unknowns
    while( ((*(pX[startLevel]))[0].getSize() < 1) && (startLevel<param.noOfNets-1)) 
        startLevel++;
    cout<<" Startlevel: "<<startLevel<<endl;
    

    schur->setDimSchurComplement(fvss);
    
    TMultigrid *multigrid = new TMultigrid( pMatrix, pRHS, pX, 
                                            prolongation, 
                                            prolongation, 
                                            startLevel,
                                            param.noOfNets-1, 
                                            param.cycle,
                                            fastbgs,
                                            NULL, 
                                            bicgStab, 
                                            param.steps);
    uint smoothingSteps    = 3;
    uint smootherVerbosity = 0;

    // mg
    multigrid->setMode( 1 );
    multigrid->setPreSmoothingSteps(smoothingSteps);
    multigrid->setPostSmoothingSteps(0);
    multigrid->setVerbosity(127);
    
    // bicg - cgridSolver
    bicgStab->setMode( 0 );
    bicgStab->setMaxit( 40);
    bicgStab->setAccuracy( 1e-8 );
    bicgStab->setPrec( bgs );
    bicgStab->setVerbosity(0);

    
    // smoother
    // modes
    bgs->setMode( 2 );
    sgs->setMode( 2 );
    fastfgs->setMode( 2 );
    fastbgs->setMode( 2 );
    fastsgs->setMode( 2 );

    // verbosity
    sgs->setVerbosity( smootherVerbosity );
    bgs->setVerbosity( smootherVerbosity );
    fastfgs->setVerbosity( smootherVerbosity );
    fastsgs->setVerbosity( smootherVerbosity );
    fastbgs->setVerbosity( smootherVerbosity );
    schur->setVerbosity(0 );

    //   bgs->setDampFactor( 0.2 );
    
    /*
      pX[param.noOfNets - 1]->set( 0.0 );
      bicgStab->setVerbosity( 5 );
      bicgStab->setMode( 1 );
      timer.start();
      bicgStab->solve( pMatrix[param.noOfNets - 1], pRHS[param.noOfNets - 1], pX[param.noOfNets - 1]);  
      timer.stop();
      cout << "time for bicgStab: " << timer << endl;
      */

  
     pX[param.noOfNets - 1]->set( 0.0 );
    timer.start();
    multigrid->solve();
    timer.stop();
    cout << "time for multigrid: " << timer << endl << endl;

   
    timer.start();
    /*
      (*pMatrix[param.noOfNets-1])[0][0].decompose( BLOCK_CYC_LU, cyclicBlocks[param.noOfNets - 1] );
      timer.stop();
      cout << "time for decomposing matrix: " << timer << endl;
      */
    /*
      timer.start();
      pX[param.noOfNets - 1]->set( 0.0 );
      bicgStab->setVerbosity( 5 );
      bicgStab->setMode( 1 );
      bicgStab->setPrec( blocksgs );
      bicgStab->solve( pMatrix[param.noOfNets - 1], pRHS[param.noOfNets - 1], pX[param.noOfNets - 1]);  
      timer.stop();
      cout << "time for bicgStab: " << timer << endl;
      */
    /*
      multigrid->setSmoother( blocksgs );
      pX[param.noOfNets - 1]->set( 0.0 );
      timer.start();
      multigrid->solve();
      timer.stop();
      cout << "time for multigrid: " << timer << endl << endl;
      */

  

    
    //
    // compare with exact solution (in function dirichlet1)
    //
    TBlockVector<TFloat> *r = new TBlockVector<TFloat>( *pRHS[param.noOfNets - 1] );
    TBlockVector<TFloat> *exact = new TBlockVector<TFloat>( *pRHS[param.noOfNets - 1] );
      TBlockVector<TFloat> *xx = new TBlockVector<TFloat>( *pRHS[param.noOfNets - 1] );
     TBlockVector<TFloat> *yy = new TBlockVector<TFloat>( *pRHS[param.noOfNets - 1] );
    
     for ( i = 0; i < (*exact)[0].getSize(); i++) {
        (*exact)[0][i] = eqn->dirichlet0( *(*arrVectors)[i] );
        (*xx)[0][i]=(*(*arrVectors)[i])[0];
        (*yy)[0][i]=(*(*arrVectors)[i])[1];
     }
        if ( param.ordering != NONE ) (*exact)[0].permute( vecInvPerm[param.noOfNets-1] );
        if ( param.ordering != NONE ) (*xx)[0].permute( vecInvPerm[param.noOfNets-1] );
        if ( param.ordering != NONE ) (*yy)[0].permute( vecInvPerm[param.noOfNets-1] );

        TBlockVector<TFloat> *x1 = new TBlockVector<TFloat>(*pX[param.noOfNets - 1] );
        multigrid->setMode(2);
        multigrid->setMaxit(1);
        multigrid->solve();
        TBlockVector<TFloat> *x2 = new TBlockVector<TFloat>(*pX[param.noOfNets - 1] );

  
        multigrid->setMode(1);
        multigrid->setMaxit(param.steps);
       
     
   
    matWriteBlockVector("/tmp/tpr/x1.mat",*x1,"x1" );
    matWriteBlockVector("/tmp/tpr/x2.mat",*x2,"x2" );
    matWriteBlockVector("/tmp/tpr/r.mat",*r,"r" );
    matWriteBlockVector("/tmp/tpr/x.mat",*xx,"xx" );
    matWriteBlockVector("/tmp/tpr/y.mat",*yy,"yy" );
    
    // cout << "(Error: " << error->weightedEuclidNorm() << ")" << endl;

    /////////////////////////////////////
    //
    // clean up
    //

    delete iterNet; 
    delete netHier;
    
    delete dofManager;
    delete eqn;
    delete fe;
    delete hierarchy;

    for (i = 0; i < param.noOfNets; i++) {
        if (pX[i]) delete pX[i];
        if (pRHS[i]) delete pRHS[i];
        if (pMatrix[i]) delete pMatrix[i];
    }
    for (i = 0; i < param.noOfNets-1; i++) {
        if (prolongation[i]) delete prolongation[i];
    }

    delete arrVectors;
    delete app;

    delete rich;
    delete jacobi;

    delete bgs;
    delete sgs;
    delete fastfgs;
    delete fastbgs;
    delete fastsgs;
    delete schur;
    delete bicgStab;
    delete blocksgs;
    delete multigrid;

    delete exact;
    //    delete error;

  

    /*  if ( param.ordering != NONE )
        for ( i = 1; i < param.noOfNets; i++ ) {
            if (vecInvPerm[i]) delete (vecInvPerm[i]);
            if (vecPerm[i]) delete (vecPerm[i]);
           
        }
        */
    TMemory::statistics();
}
