//
//  Project  : Navier-Stokes
//  File     : convDiff.cc
//  Author   : Ronald Kriemann / Sabine Gutsch Thomas Probst
//  Purpose  : test for discretization and multigrid solution
//             of a convection-diffusion problem; 
//             comparison with exact solution
//
// arch-tag: 01c24b54-09eb-4682-9dc5-39f839d93732
//

#include <stdio.h>

#include "misc/timer.hh"
#include "misc/list.hh"
#include "misc/cmdline.hh"
#include "misc/matlab.hh"

#include "net/vertex.hh"
#include "net/regular.hh"
#include "net/net.hh"

#include "solver/secondNF.hh"
#include "solver/cgSolver.hh"
#include "solver/bicgStab.hh"
#include "solver/kaczmarz.hh"
#include "solver/multigrid.hh"
#include "solver/blocksolver.hh"
#include "solver/schur.hh"
#include "solver/fastGaussSeidel.hh"

#include "discr/convDiffDoFManager.hh"
#include "discr/convDiffFE.hh"
#include "discr/convDiffHierarchy.hh"
#include "discr/convDirections.hh"

#include "matrix/blockmatrix.hh"

#include "visual/drawapp.hh"

#include "graph/graphtool.hh"
#include "graph/rcm.hh"
#include "graph/fvs.hh"

// exact solution: e := x^2 + y^2
// rhs: L e = 
//      -Laplace e + b0 d/dx e + b1 d/dy e = -4 + z0 * 2 * x + z1 * 2 * y

class TMyConvDiffEqn : public TBasisEqn<2> {

public:
    TMyConvDiffEqn ( ) 
            : TBasisEqn<2> ( ) { };
    virtual TFloat f0 (const TVector<2> &v) {
        return ( -4 * epsilon + (b0(v) * 2 * v[0] + b1(v) * 2 * v[1] ) ); 
    }
    virtual TFloat dirichlet0 ( const TVector<2> &v) { 
        return ( v[0] * v[0] + v[1] * v[1] );
    }
};

int 
main( int argc, char **argv ) 
{
    uint i = 0;
   
    param.steps      = 100;
    param.noOfNets   = 6;
    param.startLevel = 1;
    param.ordering   = CONCENTRIC_MIN;
    param.xi         = 1.0/3.0;
    param.mgSort     = MGSORT_INDEP; 
    param.criterion  = 3;
    param.epsilon    = 1e-9;
    param.draw       = FALSE;
    param.convDirName = "xline";
    strcpy(param.starttri, "grids/start1.tri" );

    param.evalCmdLine( argc, argv );

    param.endLevel = param.noOfNets-1;
    cout << "chosen parameter: " << param << endl;
    //param.output();

    //if (param.ordering == RCM) param.xi = 0.0;

    //
    // build and refine net
    //

    TTimer timer;
    T2DNetHierarchy *netHier = new T2DNetHierarchy( param.starttri );
    netHier->delReference();
    
    for (i = 0; i < param.noOfNets-1; i++)
        netHier->buildNet();
    
    ulong noOfInteriorNodesNet[param.noOfNets];
    
    T2DNetHierarchy::TNetListIterator *iterNet;
    T2DNet *pNet;
    i = 0;
    iterNet = new T2DNetHierarchy::TNetListIterator( netHier->getList() );
    iterNet->first();
    while ((pNet = (*iterNet)++) != NULL)
    {
        noOfInteriorNodesNet[i++] = pNet->getNoOfInteriorNodes();
        cout << "net " << i-1 << ": " << noOfInteriorNodesNet[i-1]<<endl; 
    }// while
    
    T2DNet  *fNet = iterNet->last();
    
    timer.start();
    
    //
    // build stiffness matrix sparsity structure
    //

    TBlockMatrix **pMatrix = new (TBlockMatrix*)[param.noOfNets];;
    TProlongation **prolongation = new (TProlongation*)[param.noOfNets-1];
    TBlockVector<TFloat> **pRHS = new (TBlockVector<TFloat>*)[param.noOfNets];
    TBlockVector<TFloat> **pX = new (TBlockVector<TFloat>*)[param.noOfNets];

    T2DConfDiffDoFManager *dofManager = new T2DConfDiffDoFManager();

    TMyConvDiffEqn *eqn = new TMyConvDiffEqn( );
    // set epsilon
    eqn->setEpsilon( param.epsilon );
    // set reaction term
    if (param.reactionTerm)
        eqn->setReacFunction( oneReac ); 
    else 
        eqn->setReacFunction( noReac ); 

    // Pointer to functions for convection direction
    if (param.convDirName == "circle")
        eqn->setConvFunctions( circle0, circle1 ); 
    if (param.convDirName == "xline")
        eqn->setConvFunctions( xline0, xline1 ); 
    if (param.convDirName == "curve")
        eqn->setConvFunctions( curve0, curve1 ); 
    if (param.convDirName == "square") 
        eqn->setConvFunctions( square0, square1 );    
    if (param.convDirName == "foursquares") 
        eqn->setConvFunctions( foursquares0, foursquares1 ); 
    if (param.convDirName == "fourcircles") 
        eqn->setConvFunctions( fourcircles0, fourcircles1 ); 
   
 
    T2DConvDiffFE *fe = new T2DConvDiffFE( Tabata );
    T2DConvDiffHierarchy *hierarchy = new T2DConvDiffHierarchy();

    hierarchy->buildHierarchy( dofManager, eqn, fe, netHier, pMatrix,
                               pRHS, pX, prolongation );

    timer.stop();
    cout << "time for disretization : " << timer << endl;
    //cout << "matrix:\n" << (*pMatrix[param.noOfNets-1])[0][0] << endl;  

    timer.start();
    
    ///////////////////////////////////////////////////
    //
    // build graph from matrix and work with it
    //

    TDynArray< TVector<2> * > *arrVectors = new TDynArray< TVector<2> * >;
    fe->getCoord( fNet, *arrVectors );

    // draw some things
    if ( arrVectors->getSize() < 1200 ) 
    {
        TDrawApp::draw( *fNet, "net" );
        eqn->initConvVector(pRHS[param.noOfNets-1]->getSize(), *arrVectors);
        *(eqn->getConvVector()) *= (-1);
        TDrawApp::draw( *arrVectors, *(eqn->getConvVector()), *fNet, "vecField");
        eqn->deleteConvVector();
    }

    timer.start();
    TBlockInfo  *blockinfo = new TBlockInfo;
    blockinfo->setSortMethod( SORT_OFD );
    //blockinfo->setSortMethod( SORT_EDGE_WEIGHT );
    //blockinfo->setSortMethod( SORT_SUCC_EDGE_WEIGHT );


    //
    // build data structures for each net
    //
    TVectorBlock<ulong>  *vecPerm[param.noOfNets];
    TVectorBlock<ulong>  *cyclicBlocks[param.noOfNets];
    TVectorBlock<ulong>  *arrFVSize = new TVectorBlock<ulong>();
 
    if ( param.ordering != NONE )    {
        i = 0;
        iterNet->first();
        while ((pNet = (*iterNet)++) != NULL)
        {
            vecPerm[i] = new TVectorBlock<ulong>;
            cyclicBlocks[i] = new TVectorBlock<ulong>;
            i++;
        }// while
    }// if

    //
    // now build permutation vectors
    //
    if ( param.ordering & CONCENTRIC ) {
     
        if (param.draw) {
            blockinfo->setCycleName("cycle"); 
            blockinfo->setGraphName("graph");
        }
        timer.start();
      
        switch (param.mgSort)
        {
            case MGSORT_PROJ :
                blockinfo->buildCYCProjPerm( vecPerm, cyclicBlocks,
                                             *fe, netHier->getList(), 
                                             pMatrix, 
                                             param.xi, 
                                             param.criterion,
                                             param.ordering );
                break;
            case MGSORT_NESTED    :
                blockinfo->buildCYCNestedPerm( vecPerm, cyclicBlocks,
                                               *fe, netHier->getList(), 
                                               pMatrix, 
                                               param.xi, 
                                               param.criterion,
                                               param.ordering  );
                break;
            case MGSORT_INDEP  :
                blockinfo->buildCYCIndepPerm( vecPerm, cyclicBlocks,
                                              *fe, netHier->getList(), 
                                              pMatrix, 
                                              param.xi, 
                                              param.criterion,
                                              param.ordering  );
                break;
            default:
                break;
        }// switch
        timer.stop();
        cout << endl << "time for finding cycles : " << timer << endl;
        for (i = 0; i < param.noOfNets; i++)
            cout << " Number of cycles in net " << i << " :  " 
                 << cyclicBlocks[i]->getSize()/2 << endl;
    }
    /*
      if ( param.mgSort == MGSORT_INDEPEND )
      blockinfo->buildIndependentPermutation( vecPerm, cyclicBlocks,
      *fe, netHier->getList(), 
      pMatrix, param.xi, 
      param.criterion,
      param.ordering);
                                             
      if ( param.mgSort == MGSORT_PROJECTED )
      blockinfo->buildProjectedPermutation( vecPerm, cyclicBlocks,
      *fe, netHier->getList(), 
      pMatrix, param.xi, 
      param.criterion,
      param.ordering);
      timer.stop();
      cout << endl << "time for finding concentric cycles : "<< timer 
      << endl << endl;
      for (i = 0; i < param.noOfNets; i++)
      cout << " Number of cycles in net " << i << " :  " 
      << cyclicBlocks[i]->getSize()/2 << endl;
      }
      */

    if (( param.ordering == HEURISTIC_FVS ) ||
        ( param.ordering == PLANAR_FVS ))
    {
        timer.start();
        switch (param.mgSort)
        {
            case MGSORT_PROJ :
                blockinfo->buildFVSProjPerm( vecPerm, arrFVSize,
                                             *fe, netHier->getList(),
                                             pMatrix, 
                                             param.xi, 
                                             param.criterion,
                                             SEARCH_DFS, param.ordering );
                break;
            case MGSORT_NESTED    :
                blockinfo->buildFVSNestedPerm( vecPerm, arrFVSize,
                                               *fe, netHier->getList(), 
                                               pMatrix, 
                                               param.xi, param.criterion,
                                               SEARCH_DFS, param.ordering );
                break;
            case MGSORT_INDEP  :
                blockinfo->buildFVSIndepPerm( vecPerm, arrFVSize,
                                              *fe, netHier->getList(), 
                                              pMatrix, 
                                              param.xi, param.criterion,
                                              SEARCH_DFS, param.ordering );
                break;
            default:
                break;
        }// switch
        timer.stop();
        cout << endl << "time for finding FVS : " << timer << endl;
        for (i = 0; i < param.noOfNets; i++) 
            cout << " Number of FVS nodes in net " << i << " :  " 
                 << (*arrFVSize)[i] << endl;
              
    }

    /*
      if ( param.mgSort == MGSORT_INDEPEND )
      blockinfo->buildIndependentPermutation( vecPerm, sizeFVS,
      *fe, netHier->getList(), 
      pMatrix, 
      param.xi, 
      param.criterion );
                                              
      if ( param.mgSort == MGSORT_PROJECTED )
      blockinfo->buildProjectedPermutation( vecPerm, sizeFVS,
      *fe, netHier->getList(), 
      pMatrix, 
      param.xi, 
      param.criterion );
      timer.stop();
      cout << endl << "time for finding FVS : " << timer << endl << endl;
      for (i = 0; i < param.noOfNets; i++) 
      cout << " Number of FVS nodes in net " << i << " :  " 
      << (*sizeFVS)[i] << endl;
              
      }
      */

    if ( param.ordering == RCM ) {
        timer.start();
        switch (param.mgSort)
        {
            case MGSORT_PROJ :
                blockinfo->buildRCMProjPerm( vecPerm, pMatrix, 
                                             param.noOfNets );
                break;
            case MGSORT_INDEP  :
                blockinfo->buildRCMIndepPerm( vecPerm, pMatrix, 
                                              param.noOfNets );
                break;
            default:
                break;
        }// switch
        timer.stop();
        cout << endl << "time for doing RCM : " << timer << endl;
    }

    //
    // build the inverse permutations
    //
    TVectorBlock<ulong> *vecInvPerm[param.noOfNets];
    
    if ( param.ordering != NONE )    {
        cout << "building inverse permutations" << endl;
        i = 0;
        iterNet->first();
        while ((pNet = (*iterNet)++) != NULL) {
            ulong size = vecPerm[i]->getSize();
            vecInvPerm[i] = new TVectorBlock<ulong>( 0, size );
            invertPerm( (*vecInvPerm[i]), *vecPerm[i] );
            i++;
        }// while 
    }// if

    // reorder
    if (param.ordering != NONE ) {
        timer.start();
        for ( i = 1; i < param.noOfNets; i++ ) 
        {
            (*pRHS[i])[0].permute( vecInvPerm[i] );
            (*pMatrix[i])[0][0].permute( vecInvPerm[i], vecInvPerm[i] );
            if ( i > 1 )
                (prolongation[i-1]->pBlockProlong)[0]->permute( vecInvPerm[i], vecInvPerm[i-1]);
        }
        timer.stop();
        cout << "time for permuting matrices and rhs's: "<< timer << endl;
    }

    // draw matrix
    param.xi = 1.0 / 3.0;
    TDrawApp::drawMatrix( pMatrix[param.noOfNets - 1], param.xi, 
                          "matrix");
    if ( arrVectors->getSize() < 1200 ) 
    {
        TDrawApp::draw( *fNet, *vecInvPerm[param.noOfNets-1], 
                   "netPerm" );
    }
    /*    
    matWriteMatrixBlock("/tmp/tpr/matrixf.mat", (*pMatrix[param.noOfNets-1])[0][0],
                        (*pMatrix[param.noOfNets-1])[0][0].getNoOfRows(),"Af");
    matWriteMatrixBlock("/tmp/tpr/matrixc.mat", (*pMatrix[param.noOfNets-2])[0][0],
                        (*pMatrix[param.noOfNets-2])[0][0].getNoOfRows(),"Ac");
    matWriteBlockProlongation("/tmp/tpr/matrixp.mat" ,(*prolongation[param.noOfNets-2]),"p" );
    */

    /////////////////////////////////////
    //
    // solver
    //

    TSchur *schur = new TSchur( );
    TJacobi *jacobi = new TJacobi( );
    jacobi->setDampFactor( 0.5 );
    TRichardson *rich = new TRichardson( );
    //TKaczmarz *kaczmarz = new TKaczmarz( );
    TForwardGaussSeidel *fgs = new TForwardGaussSeidel( );
    TBackwardGaussSeidel *bgs = new TBackwardGaussSeidel( );
    TSymmetricGaussSeidel *sgs = new TSymmetricGaussSeidel( );

    TFastForwardGaussSeidel *fastfgs = new TFastForwardGaussSeidel( );
    TFastBackwardGaussSeidel *fastbgs = new TFastBackwardGaussSeidel( );
    TFastSymmetricGaussSeidel *fastsgs = new TFastSymmetricGaussSeidel();

    TBicgStab *bicgStab = new TBicgStab( sgs );
    TBlockSymmetricGS *blocksgs = new TBlockSymmetricGS( );
    //sgs->setDampFactor( 0.5 );
    //blocksgs->setDampFactor( 0.5 );

   TMultigrid *multigrid = new TMultigrid( pMatrix, pRHS, pX, 
                                            prolongation, 
                                            prolongation, 
                                            param.startLevel, 
                                            param.noOfNets-1, 
                                            param.cycle, fastsgs, fastsgs, 
                                            bicgStab, 
                                            param.steps);
   uint smootherVerbosity = 0;

   multigrid->setMode( 1 );
   multigrid->setVerbosity( 5 );
   multigrid->setPreSmoothingSteps(param.preSmoothing);
   multigrid->setPostSmoothingSteps(param.postSmoothing);
   
   fgs->setMode( 2 );
   bgs->setMode( 2 );
   sgs->setMode( 2 );
   
   fastfgs->setMode( 2 );
   fastbgs->setMode( 2 );
   fastsgs->setMode( 2 );
   
   
    fgs->setVerbosity( smootherVerbosity );
    sgs->setVerbosity( smootherVerbosity );
    bgs->setVerbosity( smootherVerbosity );
    fastfgs->setVerbosity( smootherVerbosity );
    fastsgs->setVerbosity( smootherVerbosity );
    fastbgs->setVerbosity( smootherVerbosity );

    bicgStab->setMaxit( param.steps );
    bicgStab->setAccuracy( 1e-15 );
    bicgStab->setVerbosity( 0 );
    bicgStab->setMode( 1 );
    
    //
    // solve 
    //    
    if ( param.ordering == HEURISTIC_FVS ) {
        bicgStab->setPrec( bgs );
        multigrid->setSmoother( fastbgs );
    }
    if ( param.ordering != HEURISTIC_FVS ) {
        bicgStab->setPrec( sgs );
        multigrid->setSmoother( fastsgs );
    }

    /*
      pX[param.noOfNets - 1]->set( 0.0 );
      bicgStab->setVerbosity( 5 );
      bicgStab->setMode( 1 );
      timer.start();
      bicgStab->solve( pMatrix[param.noOfNets - 1], pRHS[param.noOfNets - 1], pX[param.noOfNets - 1]);  
      timer.stop();
      cout << "time for bicgStab: " << timer << endl;
      */

    pX[param.noOfNets - 1]->set( 0.0 );
    timer.start();
    multigrid->solve();
    timer.stop();
    cout << "time for multigrid: " << timer << endl << endl;

    if ( (param.ordering & CONCENTRIC) && 
         (cyclicBlocks[param.noOfNets - 1]->getSize() / 2 > 0) ) {
        timer.start();
        /*
          (*pMatrix[param.noOfNets-1])[0][0].decompose( BLOCK_CYC_LU, cyclicBlocks[param.noOfNets - 1] );
          timer.stop();
          cout << "time for decomposing matrix: " << timer << endl;
          */
        /*
          timer.start();
          pX[param.noOfNets - 1]->set( 0.0 );
          bicgStab->setVerbosity( 5 );
          bicgStab->setMode( 1 );
          bicgStab->setPrec( blocksgs );
          bicgStab->solve( pMatrix[param.noOfNets - 1], pRHS[param.noOfNets - 1], pX[param.noOfNets - 1]);  
          timer.stop();
          cout << "time for bicgStab: " << timer << endl;
          */
        /*
          multigrid->setSmoother( blocksgs );
          pX[param.noOfNets - 1]->set( 0.0 );
          timer.start();
          multigrid->solve();
          timer.stop();
          cout << "time for multigrid: " << timer << endl << endl;
          */

    }

    
    //
    // compare with exact solution (in function dirichlet1)
    //
    TBlockVector<TFloat> *exact = new TBlockVector<TFloat>( *pRHS[param.noOfNets - 1] );
    
    for ( i = 0; i < (*exact)[0].getSize(); i++) 
        (*exact)[0][i] = eqn->dirichlet0( *(*arrVectors)[i] );
    if ( param.ordering != NONE ) (*exact)[0].permute( vecInvPerm[param.noOfNets-1] );

    TBlockVector<TFloat> *error = new TBlockVector<TFloat>( *exact );
    *error -= *pX[param.noOfNets - 1];
    cout << "(Error: " 
         << error->weightedEuclidNorm() << ")" << endl;

    /////////////////////////////////////
    //
    // clean up
    //

    delete iterNet; 
    delete netHier;
    
    delete dofManager;
    delete eqn;
    delete fe;
    delete hierarchy;

    for (i = 0; i < param.noOfNets; i++) {
        if (pX[i]) delete pX[i];
        if (pRHS[i]) delete pRHS[i];
        if (pMatrix[i]) delete pMatrix[i];
    }
    for (i = 0; i < param.noOfNets-1; i++) {
        if (prolongation[i]) delete prolongation[i];
    }

    delete arrVectors;

    delete rich;
    delete jacobi;
    delete fgs;
    delete bgs;
    delete sgs;
    delete fastfgs;
    delete fastbgs;
    delete fastsgs;
    delete schur;
    delete bicgStab;
    delete blocksgs;
    delete multigrid;

    delete exact;
    delete error;

    delete arrFVSize;
    delete blockinfo;

    if ( param.ordering != NONE )
        for ( i = 1; i < param.noOfNets; i++ ) {
            delete (vecInvPerm[i]);
            delete (vecPerm[i]);
            delete (cyclicBlocks[i]);
        }

    TMemory::statistics();
}
