//
//	Project		: Navier-Stokes
//	File		: convDiff.cc
//	Author		: Ronald Kriemann / Sabine Gutsch Thomas Probst
//	Purpose		: test for discretization and twogrid solution
//                of a convection-diffusion problem; 
//                comparison with exact solution
//	Date		: 12.06.1997
//	Changes     : 
//
// arch-tag: 2d91a3c0-0ea4-4d25-8155-db9de90dac2b
//

#include <stdio.h>

#include "misc/timer.hh"
#include "misc/list.hh"
#include "misc/cmdline.hh"

#include "net/vertex.hh"
#include "net/regular.hh"
#include "net/net.hh"

#include "solver/secondNF.hh"
#include "solver/cgSolver.hh"
#include "solver/bicgStab.hh"
#include "solver/lu.hh"
#include "solver/kaczmarz.hh"
#include "solver/twogrid.hh"
#include "solver/blocksolver.hh"
#include "solver/schur.hh"
#include "solver/fastGaussSeidel.hh"

#include "discr/convDiffDoFManager.hh"
#include "discr/convDiffFE.hh"
#include "discr/convDiffHierarchy.hh"
#include "discr/convDirections.hh"

#include "matrix/blockmatrix.hh"

#include "visual/drawapp.hh"
#include "graph/blockinfo.hh"

//#include "graph/graphtool.hh"
//#include "graph/rcm.hh"
//#include "graph/fvs.hh"

// exact solution: e := x^2 + y^2
// rhs: L e = 
//      -Laplace e + b0 d/dx e + b1 d/dy e = -4 + z0 * 2 * x + z1 * 2 * y

class TMyConvDiffEqn : public TBasisEqn<2> {

public:
     TMyConvDiffEqn ( ) 
       : TBasisEqn<2> ( ) { };
    virtual TFloat f0 (const TVector<2> &v) {
        return ( - 4 * epsilon + (b0(v) * 2 * v[0] + b1(v) * 2 * v[1] ) ); 
    }
    virtual TFloat dirichlet0 ( const TVector<2> &v) { 
        return ( v[0] * v[0] + v[1] * v[1] );
	}
};

int 
main( int argc, char **argv ) 
{
    uint i = 0;
   
	TParameter  param;
	param.steps = 50;
    param.noOfNets   = 8;
    param.startLevel = 0;
    param.ordering   = _CONCENTRIC_MIN_;
    param.xi         =  0.333333;
    param.mgSort     = _MGSORT_INDEP_; 
    param.criterion  = 3;
    param.epsilon    = 1e+2;
    param.draw       = FALSE;
    //param.draw       = TRUE;
    //strcpy(param.convDirName, "circle");
    strcpy(param.convDirName, "xline");
    strcpy(param.starttri, "grids/init1.tri" );

	param.evalCommandLine( argc, argv );
    param.endLevel = param.noOfNets-1;
    cout << "chosen parameters : " << endl << param;

    //if (param.ordering == _RCM_) param.xi = 0.0;

	//
	// build and refine net
	//

    TTimer timer;
    T2DNetHierarchy *netHier = new T2DNetHierarchy( param.starttri );
	netHier->delReference();
    
    for (i = 0; i < param.noOfNets-1; i++)
        netHier->buildNet();
    
    ulong noOfInteriorNodesNet[param.noOfNets];
    
    T2DNetHierarchy::TNetListIterator *iterNet;
	T2DNet *pNet;
    i = 0;
    iterNet = new T2DNetHierarchy::TNetListIterator( netHier->getList() );
	iterNet->first();
	while ((pNet = (*iterNet)++) != NULL)
	{
		noOfInteriorNodesNet[i++] = pNet->getNoOfInteriorNodes();
			}// while
    
    T2DNet  *fNet = iterNet->last();
    
    timer.start();
    
	//
	// build stiffness matrix sparsity structure
	//

    TBlockMatrix **pMatrix = new (TBlockMatrix*)[param.noOfNets];;
    TProlongation **prolongation = new (TProlongation*)[param.noOfNets-1];
    TBlockVector<TFloat> **pRHS = new (TBlockVector<TFloat>*)[param.noOfNets];
    TBlockVector<TFloat> **pX = new (TBlockVector<TFloat>*)[param.noOfNets];

    T2DConfDiffDoFManager *dofManager = new T2DConfDiffDoFManager();

    TMyConvDiffEqn *eqn = new TMyConvDiffEqn( );
    // set epsilon
    eqn->setEpsilon( param.epsilon );
    // set reaction term
    if (param.reactionTerm)
       eqn->setReacFunction( oneReac ); 
    else 
       eqn->setReacFunction( noReac ); 

    // Pointer to functions for convection direction
    if (strcmp(param.convDirName,"circle")==0)
       eqn->setConvFunctions( circle0, circle1 ); 
    if (strcmp(param.convDirName,"xline")==0)
       eqn->setConvFunctions( xline0, xline1 ); 
    if (strcmp(param.convDirName,"curve")==0)
       eqn->setConvFunctions( curve0, curve1 ); 
    if (strcmp(param.convDirName,"square")==0) 
       eqn->setConvFunctions( square0, square1 );    
    if (strcmp(param.convDirName,"foursquares")==0) 
       eqn->setConvFunctions( foursquares0, foursquares1 ); 
    if (strcmp(param.convDirName,"fourcircles")==0) 
       eqn->setConvFunctions( fourcircles0, fourcircles1 ); 
    if (strcmp(param.convDirName,"triangle")==0) 
       eqn->setConvFunctions( triangle0, triangle1 ); 
   
 
    T2DConvDiffFE *fe = new T2DConvDiffFE( Tabata );
    T2DConvDiffHierarchy *hierarchy = new T2DConvDiffHierarchy();

    hierarchy->buildHierarchy( dofManager, eqn, fe, netHier, pMatrix,
                               pRHS, pX, prolongation );

    timer.stop();
  

    timer.start();
	
	///////////////////////////////////////////////////
	//
	// lets draw a little bit
	//
	TDrawApp *app = new TDrawApp;

	///////////////////////////////////////////////////
	//
	// build graph from matrix and work with it
	//

	TPArray<T2DVector> *arrVectors = new TPArray<T2DVector>;
	fe->getCoord( fNet, *arrVectors );

	// draw some things
    if ( arrVectors->getSize() < 1200 ) 
    {
        //	    app->draw( *fNet, "psfiles/net.ps" );
        eqn->initConvVector(pRHS[param.noOfNets-1]->getSize(), *arrVectors);
        *(eqn->getConvVector()) *= (-1);
        // app->draw( *arrVectors, *(eqn->getConvVector()), *fNet, "psfiles/vecField.ps");
        eqn->deleteConvVector();
	}

    timer.start();
    TBlockInfo  *blockinfo = new TBlockInfo;
    blockinfo->setSortMethod( _SORT_OFD_ );
    //blockinfo->setSortMethod( _SORT_EDGE_WEIGHT_ );
    //blockinfo->setSortMethod( _SORT_SUCC_EDGE_WEIGHT_ );

    // true: parallele Anordung
    // blockinfo->setBlocking( TRUE );
    blockinfo->setBlocking( FALSE );

    //
    // build data structures for each net
    //
    TVectorBlock<ulong>  *vecPerm[param.noOfNets];
    TVectorBlock<ulong>  *cyclicBlocks[param.noOfNets];
    TVectorBlock<ulong>  *arrFVSize = new TVectorBlock<ulong>();
 
    if ( param.ordering != _NONE_ )	{
       i = 0;
       iterNet->first();
       while ((pNet = (*iterNet)++) != NULL)
       {
          vecPerm[i] = new TVectorBlock<ulong>;
          cyclicBlocks[i] = new TVectorBlock<ulong>;
          i++;
       }// while
    }// if

   //
   // now build permutation vectors
   //
   if ( param.ordering & _CONCENTRIC_ ) {
     
	  if (param.draw) {
	     blockinfo->setCycleName("psfiles/cycle.ps"); 
	     blockinfo->setGraphName("psfiles/graph.ps");
	  }
      timer.start();
	  
		switch (param.mgSort)
		{
		case _MGSORT_PROJ_ :
			blockinfo->buildCYCProjPerm( vecPerm, cyclicBlocks,
												 *fe, netHier->getList(), 
                                                  pMatrix, 
												  param.xi, 
                                                  param.criterion,
                                                  param.ordering );
			break;
		case _MGSORT_NESTED_    :
			blockinfo->buildCYCNestedPerm( vecPerm, cyclicBlocks,
											   *fe, netHier->getList(), 
                                               pMatrix, 
											   param.xi, 
                                               param.criterion,
                                                param.ordering  );
			break;
		case _MGSORT_INDEP_  :
			blockinfo->buildCYCIndepPerm( vecPerm, cyclicBlocks,
												*fe, netHier->getList(), 
                                                pMatrix, 
												param.xi, 
                                                param.criterion,
                                                param.ordering  );
			break;
		default:
			break;
		}// switch
   }
  
   

   

   //
   // build the inverse permutations
   //
   TVectorBlock<ulong> *vecInvPerm[param.noOfNets];
    
   if ( param.ordering != _NONE_ )	{
      i = 0;
       iterNet->first();
	   while ((pNet = (*iterNet)++) != NULL) {
            ulong size = vecPerm[i]->getSize();
		    vecInvPerm[i] = new TVectorBlock<ulong>( 0, size );
		    invertPerm( (*vecInvPerm[i]), *vecPerm[i] );
            //(*vecInvPerm[i]) = invertPermutation( *vecPerm[i] );
           i++;
		}// while 
   }// if

    // reorder
    if (param.ordering != _NONE_ ) {
	    timer.start();
        for ( i = 1; i < param.noOfNets; i++ ) 
		{
	        (*pRHS[i])[0].permute( vecInvPerm[i] );
            (*pMatrix[i])[0][0].permute( vecInvPerm[i], vecInvPerm[i] );
            if ( i > 1 )
            (prolongation[i-1]->pBlockProlong)[0]->permute( vecInvPerm[i], vecInvPerm[i-1]);
		}
	    timer.stop();
	  }

    // draw matrix
    param.xi = 1.0 / 3.0;
    if ( arrVectors->getSize() < 1200 ) 
    {
        app->draw( *pMatrix[param.noOfNets - 1], param.xi, 
                   "psfiles/matrix.ps");
	    app->draw( *fNet, *vecInvPerm[param.noOfNets-1], 
                   "psfiles/netPerm.ps" );
	}

	/////////////////////////////////////
	//
	// solver
	//

    TSchur *schur = new TSchur( );
	TJacobi *jacobi = new TJacobi( );
    jacobi->setDampFactor( 0.5 );
	TRichardson *rich = new TRichardson( );
	TLU *lu = new TLU( );
    //TKaczmarz *kaczmarz = new TKaczmarz( );
    TSymmetricGaussSeidel *sgs = new TSymmetricGaussSeidel( );

    TFastForwardGaussSeidel *fastfgs = new TFastForwardGaussSeidel( );
    TFastBackwardGaussSeidel *fastbgs = new TFastBackwardGaussSeidel( );
    TFastSymmetricGaussSeidel *fastsgs = new TFastSymmetricGaussSeidel();

    TBicgStab *bicgStab = new TBicgStab( sgs );
    TBlockSymmetricGS *blocksgs = new TBlockSymmetricGS( );
    //sgs->setDampFactor( 0.5 );
    //blocksgs->setDampFactor( 0.5 );
      uint smoothingSteps    = 1;
    uint smootherVerbosity = 0;

    fastbgs->setVerbosity( smootherVerbosity );
    fastsgs->setVerbosity( smootherVerbosity );
    fastfgs->setVerbosity( smootherVerbosity );

    fastbgs->setMaxit( smoothingSteps );
    fastfgs->setMaxit(0);
    fastsgs->setMaxit(0);
	
    fastfgs->setMode( 2 );
    fastbgs->setMode( 2 );
    sgs->setMode( 2 );
    fastsgs->setMode( 2 );


    bicgStab->setVerbosity( 0 );
    bicgStab->setMode( 0 );
    bicgStab->setPrec( sgs );
    bicgStab->setMaxit( 1000 );
    bicgStab->setAccuracy( 1e-12 );
    
    uint cycle = 1;
    uint mgMaxLevel = param.noOfNets-1;
   

    for(uint aktlevel = 3; aktlevel<=mgMaxLevel;aktlevel++){

        cout<<" al: "<<aktlevel<<"  with " << (*pRHS[aktlevel])[0].getSize() << " unknowns."<<endl;
        Ttwogrid *twogrid = new Ttwogrid(pMatrix, pRHS, pX, 
                                               prolongation, prolongation, 
                                               4, 
                                               aktlevel, 
                                               cycle, fastbgs, fastbgs,
                                               lu, 
                                               param.steps);
    
        twogrid->setMode( 1 );
        twogrid->setVerbosity( 6 );
        twogrid->setSmoother( fastbgs );
        pX[aktlevel]->set( 1000.0 );
        twogrid->solve();
	
        delete twogrid;

        cout<<"TG done " <<endl;
    }

	//
	// compare with exact solution (in function dirichlet1)
	//
    TBlockVector<TFloat> *exact = new TBlockVector<TFloat>( *pRHS[param.noOfNets - 1] );
    TBlockVector<TFloat> *pOne = new TBlockVector<TFloat>( *pRHS[param.noOfNets - 1] );
    //pOne->set( 1.0 );
    //*exact = (*pMatrix[param.noOfNets - 1]) * (*pOne);
    //cout << "(Singular perturbation: " 
    //     << exact->weightedEuclidNorm() << ")" << endl;
    

    
    for ( i = 0; i < (*exact)[0].getSize(); i++) 
        (*exact)[0][i] = eqn->dirichlet0( *(*arrVectors)[i] );
    if ( param.ordering != _NONE_ ) (*exact)[0].permute( vecInvPerm[param.noOfNets-1] );

    TBlockVector<TFloat> *error = new TBlockVector<TFloat>( *exact );
    *error -= *pX[param.noOfNets - 1];
   
	//cout << "Error:\n" << *error << endl;
	//cout << "Exact solution:\n" << *exact << endl;
    //(*pOne) = (*pMatrix[param.noOfNets - 1]) * (*error);
    //cout << "(A * error: " 
    //     << pOne->weightedEuclidNorm() << ")" << endl;


    (*error) = (*pMatrix[param.noOfNets - 1]) * (*exact) - *pRHS[param.noOfNets - 1];
  //cout << "residual\n" << *error << endl;

	/////////////////////////////////////
	//
	// clean up
	//

  
    TMemory::statistics();
}
