//
//  Project   : Navier-Stokes
//  File      : cavity.cc
//  Author    : Ronald Kriemann / Sabine Gutsch
//  Purpose   : test for building a hierarchy for the Stokes
//              problem and multigrid
//  Date      : 13.09.1997
//  Changes   : 
//
// arch-tag: 825230f6-25c7-41bd-9c9f-24c86412c393
//

#include <stdio.h>

#include "misc/timer.hh"
#include "misc/list.hh"
#include "misc/cmdline.hh"

#include "net/vertex.hh"
#include "net/regular.hh"
#include "net/net.hh"

#include "solver/secondNF.hh"
#include "solver/cgSolver.hh"
#include "solver/bicgStab.hh"
#include "solver/kaczmarz.hh"
#include "solver/multigrid.hh"
#include "solver/blocksolver.hh"
#include "solver/schur.hh"
#include "solver/lu.hh"
#include "solver/fixedPointIteration.hh"
#include "solver/fixedPointDefectCorrection.hh"
#include "solver/oneStepScheme.hh"

#include "discr/stokesDoFManager.hh"
#include "discr/stokesFE.hh"
#include "discr/stokesHierarchy.hh"

#include "matrix/blockmatrix.hh"

#include "visual/drawapp.hh"

#include "graph/fvs.hh"

TFloat  p ( const TVector<2> &v) {
    return  ( 2 * cos(v[0]) * sin(v[1]) - 2 * sin(1) * (1-cos(1)) );
}

//
// driven cavity example
//
class TMyNavierStokesEqn : public TBasisEqn<2> {

public:
     TMyNavierStokesEqn ( ) 
       : TBasisEqn<2> ( ) { };
    virtual TFloat f0 (const TVector<2> &v ) { return (0.0); }
    virtual TFloat f1 (const TVector<2> &v ) { return (0.0); }
    virtual TFloat dirichlet0 ( const TVector<2> &v) 
	{ 
        if (fabs( v[1] - 1 ) < 1e-8) return 1.0; 
        else return 0.0;
	}
    virtual TFloat dirichlet1 ( const TVector<2> &v) { return (0.0); }
};

int main( int argc, char **argv ) 

{
    uint i = 0;
	TParameter  param;
	param.steps = 1;
	param.noOfNets = 4;
    param.startLevel = 0;
    param.ordering = _NONE_;
    param.xi         = 1.0/3.0;
    param.criterion  = 3;
    param.mgSort     = _MGSORT_INDEPEND_; 
    param.epsilon    = 1.0;
    //param.draw       = TRUE;
    param.draw       = FALSE;
	strcpy( param.starttri, "grids/init2.tri" );

	evalCommandLine( argc, argv, param );
    param.endLevel = param.noOfNets-2;
    param.output();

	//
	// build and refine net
	//

    TTimer timer;
    T2DNetHierarchy *netHier = new T2DNetHierarchy( param.starttri );
	netHier->delReference();
    
    for (i = 0; i < param.noOfNets-1; i++)
        netHier->buildNet();
    
    ulong noOfInteriorNodesNet[param.noOfNets];
    
    T2DNetHierarchy::TNetListIterator *iterNet;
	T2DNet *pNet;
    i = 0;
    iterNet = new T2DNetHierarchy::TNetListIterator( netHier->getList() );
	iterNet->first();
	while ((pNet = (*iterNet)++) != NULL)
	{
		noOfInteriorNodesNet[i++] = pNet->getNoOfInteriorNodes();
		cout << "net " << i-1 << ": " << noOfInteriorNodesNet[i-1]<<endl; 
	}// while

    timer.start();
    
	//
	// build stiffness matrix sparsity structure
	//

    TBlockMatrix **pMatrix = new (TBlockMatrix*)[param.noOfNets-1];;
    TProlongation **prolongation = new (TProlongation*)[param.noOfNets-2];
    TBlockVector<TFloat> **pRHS = new (TBlockVector<TFloat>*)[param.noOfNets-1];
    TBlockVector<TFloat> **pX = new (TBlockVector<TFloat>*)[param.noOfNets-1];
  
    T2DStokesDoFManager *dofManager = new T2DStokesDoFManager();
    TMyNavierStokesEqn *eqn = new TMyNavierStokesEqn( );
    // set epsilon
    eqn->setEpsilon( param.epsilon );

    T2DStokesHierarchy *hierarchy = new T2DStokesHierarchy();
    T2DStokesFE *fe = new T2DStokesFE( Tabata );  

    timer.start();
    hierarchy->buildHierarchy( dofManager, eqn, fe, netHier, pMatrix,
                               prolongation, pRHS, pX );

    timer.stop();
    cout << endl;
    timer.stat();
    cout << "(for discretizing on " << param.noOfNets << " levels)\n";
    cout << endl;

   ulong dimU = (*pX[param.noOfNets-2])[0].getSize();
   ulong dimP = (*pX[param.noOfNets-2])[2].getSize();
   cout << "Dimension: U: " << dimU << ", P: " << dimP << endl;
   cout << "Gridsize:  " << sqrt(dimU) + 1 << "x" << sqrt(dimU) + 1 
        << endl;

   T2DNet *net = iterNet->last();
   net = iterNet->prev();
 
    // solve resulting equations

    uint smSteps  = 1;                 // number of smoothing steps
    uint mgMaxLevel = param.noOfNets - 2;
    uint mgMinLevel = param.startLevel;;
    //uint mgMinLevel = param.noOfNets - 3;
    // V-Cycle: cycle = 1, W - Cycle: cycle = 2 
    uint cycle = 1; 	 


    // solvers
    TSchur *schur = new TSchur( );
    TKaczmarz *kaczmarz = new TKaczmarz( );
    TLU *lu = new TLU( );
   
    TMultigrid *multigrid = new TMultigrid( pMatrix, pRHS, pX, 
                                prolongation, mgMinLevel, mgMaxLevel,  
                                cycle, kaczmarz, kaczmarz, 
                                lu, param.steps );
    
    
	TBlockVector<TFloat> *pnX = new TBlockVector<TFloat>(*pX[mgMaxLevel]);
    pnX->set( 0.0 );
    TFixedPointIteration *fixedPointIteration = new TFixedPointIteration(
                                eqn, fe, netHier, hierarchy );
    TFixedPointDefectCorrection *fixedPointDefectCorrection 
                                = new TFixedPointDefectCorrection(
                                    eqn, fe, netHier, hierarchy );
    TOneStepScheme *oneStepScheme = new TOneStepScheme( pnX, fixedPointIteration, 1 / (sqrt(dimU) + 1) );

    multigrid->setVerbosity( 5 );

    kaczmarz->setMaxit( smSteps );
    schur->setMaxit( smSteps );
	schur->setInnerSteps( 5 );

	/*
	pX[mgMaxLevel]->set( 10.0 );
    multigrid->setSmoother( kaczmarz );
    multigrid->setCoarseSolver( kaczmarzCGC );
    timer.start();
    multigrid->solve();
    timer.stop();
	cout << "time for multigrid: " << timer << endl;

    pX[mgMaxLevel]->set( 10.0 );
    multigrid->setSmoother( schur );
    //multigrid->setCoarseSolver( schurCGC );
    timer.start();
    multigrid->solve();
    timer.stop();
	cout << "time for multigrid: " << timer << endl;
	*/

    pX[mgMaxLevel]->set( 10.0 );
    TBlockVector<TFloat> *defect = 
                        new TBlockVector<TFloat>(*pRHS[mgMaxLevel]);
    *defect -= ( (*pMatrix[mgMaxLevel]) * (*pX[mgMaxLevel]) );
   
    cout << "\nInitial defect:\n";
    cout << "weighted defect 0 : "
         << defect->weightedEuclidBlockNorm(0) << endl;
    cout << "weighted defect 1 : "
         << defect->weightedEuclidBlockNorm(1) << endl;
    cout << "weighted defect 2 : "
         << defect->weightedEuclidBlockNorm(2) << endl;
    //(*defect)[2][(*pRHS[mgMaxLevel])[2].getSize()-1] = 0.0;
    //cout << "weighted defect 2 without lagrange: "
    //     << defect->weightedBlockEuclidNorm(2) << endl;

    multigrid->setSmoother( schur );
	//pnX->set( 0.0 );
    timer.start();
    fixedPointIteration->setMaxit( 5 );
    fixedPointIteration->setOrdering( param.ordering );
    fixedPointIteration->solve( multigrid );
    //fixedPointDefectCorrection->setMaxit( 10 );
    //fixedPointDefectCorrection->solve( multigrid );
    //oneStepScheme->solve( multigrid );
    timer.stop();
	cout << "time for nonlinear solver: " << timer << endl;

   *defect = *pMatrix[mgMaxLevel] * *pX[mgMaxLevel] - *pRHS[mgMaxLevel];
   
   cout << "\nDefect of numerical solution:\n";
   cout << "weighted defect 0 : "
        << defect->weightedEuclidBlockNorm(0) << endl;
   cout << "weighted defect 1 : "
        << defect->weightedEuclidBlockNorm(1) << endl;
   cout << "weighted defect 2 : "
        << defect->weightedEuclidBlockNorm(2) << endl;
   //(*defect)[2][(*pRHS[mgMaxLevel])[2].getSize()-1] = 0.0;
   //cout << "weighted defect 2 without lagrange: "
   //    << defect->weightedEuclidBlockNorm(2) << endl;
   
    // draw exact and numerical solution

    // draw all the usual things
    TDrawApp *app = new TDrawApp;

    TPArray<T2DVector> *arrVectorsU = new TPArray<T2DVector>;
    fe->getCoord(net, *arrVectorsU ); 

   if (param.draw)  
    {
	    //app->draw( *net, "psfiles/net.ps" );
        app->draw( *arrVectorsU, *pX[mgMaxLevel], *net, "psfiles/vecfield.ps" );
        //app->draw( *arrVectorsU, *pnX, *net, "vecfield.ps" );
        app->drawStreakLine( *net, *pX[mgMaxLevel], *eqn, 
                   "psfiles/streaklines.ps");
    }// if
    delete (app);
    delete (arrVectorsU);

    delete (schur);
    delete (kaczmarz);
    delete (lu);
    delete (multigrid);

    delete (fixedPointIteration);
    delete (fixedPointDefectCorrection);
    delete (oneStepScheme);
 
    delete (netHier);
    delete (iterNet);
   
    delete (dofManager);
    delete (fe);
    delete (eqn);
    delete (hierarchy);
    
    delete (defect);
    delete (pnX);

    for (i = 0; i < mgMaxLevel+1; i++) {
      if (pX[i]) delete pX[i]; 
      if (pRHS[i]) delete pRHS[i];  
      if (pMatrix[i]) delete pMatrix[i];  
	}
      for (i = 0; i < mgMaxLevel; i++) {
          if (prolongation[i]) delete (prolongation[i]);
	}

    TMemory::statistics();
	
}
