//
//  Project   : Navier-Stokes
//  File      : stokesConv.cc
//  Author    : Ronald Kriemann / Sabine Gutsch
//  Purpose   : test for building a hierarchy for the Stokes
//              problem and multigrid
//  Date      : 13.06.1997
//  Changes   : 
//
// arch-tag: 2631c6ba-419e-4604-abb3-d6f43eb29f38
//

#include <stdio.h>

#include "misc/timer.hh"
#include "misc/list.hh"
#include "misc/cmdline.hh"
#include "misc/matlab.hh"

#include "net/vertex.hh"
#include "net/regular.hh"
#include "net/net.hh"

#include "solver/secondNF.hh"
#include "solver/cgSolver.hh"
#include "solver/bicgStab.hh"
#include "solver/kaczmarz.hh"
#include "solver/lu.hh"
#include "solver/multigrid.hh"
#include "solver/blocksolver.hh"
#include "solver/schur.hh"
#include "solver/squaredSmoother.hh"

#include "discr/stokesDoFManager.hh"
#include "discr/stokesFE.hh"
#include "discr/stokesHierarchy.hh"
#include "discr/convDirections.hh"

#include "matrix/blockmatrix.hh"

#include "visual/drawapp.hh"

#include "graph/fvs.hh"

TFloat  p ( const TVector<2> &v) {
    return  ( 2 * cos(v[0]) * sin(v[1]) - 2 * sin(1) * (1-cos(1)) );
}

class TMyStokesEqn : public TBasisEqn<2> {

public:
     TMyStokesEqn (  ) 
       : TBasisEqn<2> ( ) { };
    virtual TFloat f0 (const TVector<2> &v ) 
    { return (2.0 * ( epsilon - 1.0 ) * sin(v[0]) * sin(v[1]) + b0(v) * cos(v[0]) * sin(v[1]) + b1(v) * sin(v[0]) * cos(v[1]) ); 
    }
    virtual TFloat f1 (const TVector<2> &v ) 
	{ return ( 2.0 * ( 1.0 + epsilon) * cos(v[0]) * cos(v[1]) - b1(v) * cos(v[0]) * sin(v[1]) - b0(v) * sin(v[0]) * cos(v[1]) ); }
    virtual TFloat dirichlet0 ( const TVector<2> &v) 
	{ 
        return ( sin(v[0]) * sin(v[1]) );
	}
    virtual TFloat dirichlet1 ( const TVector<2> &v) { 
        return  ( cos(v[0]) * cos(v[1]) );
	}
};

int main( int argc, char **argv ) 

{
    uint i = 0;
	TParameter  param;
	param.steps = 3;
	param.noOfNets = 3;
    param.startLevel = 0;
    param.ordering = _NONE_;
    param.xi         = 1.0/3.0;
    param.criterion  = 3;
    param.mgSort     = _MGSORT_INDEPEND_; 
    param.epsilon    = 1.0;
    param.draw       = FALSE;
    strcpy(param.convDirName, "none");
	strcpy( param.starttri, "grids/init2.tri" );

	evalCommandLine( argc, argv, param );
    param.endLevel = param.noOfNets-2;
    param.output();

	//
	// build and refine net
	//

    TTimer timer;
    T2DNetHierarchy *netHier = new T2DNetHierarchy( param.starttri );
	netHier->delReference();
    
    for (i = 0; i < param.noOfNets-1; i++)
        netHier->buildNet();
    
    ulong noOfInteriorNodesNet[param.noOfNets];
    
    T2DNetHierarchy::TNetListIterator *iterNet;
	T2DNet *pNet;
    i = 0;
    iterNet = new T2DNetHierarchy::TNetListIterator( netHier->getList() );
	iterNet->first();
	while ((pNet = (*iterNet)++) != NULL)
	{
		noOfInteriorNodesNet[i++] = pNet->getNoOfInteriorNodes();
		cout << "net " << i-1 << ": " << noOfInteriorNodesNet[i-1]<<endl; 
	}// while

    timer.start();
	//
	// build stiffness matrix sparsity structure
	//

  
    TBlockMatrix **pMatrix = new (TBlockMatrix*)[param.noOfNets-1];;
    TProlongation **prolongation = new (TProlongation*)[param.noOfNets-2];
    TBlockVector<TFloat> **pRHS = new (TBlockVector<TFloat>*)[param.noOfNets-1];
    TBlockVector<TFloat> **pX = new (TBlockVector<TFloat>*)[param.noOfNets-1];
  
    T2DStokesDoFManager *dofManager = new T2DStokesDoFManager();
    TMyStokesEqn *eqn = new TMyStokesEqn();
    // set epsilon
    eqn->setEpsilon( param.epsilon );
    // Pointer to functions for convection direction
    if (strcmp(param.convDirName,"circle")==0)
       eqn->setConvFunctions( circle0, circle1 ); 
    if (strcmp(param.convDirName,"xline")==0)
       eqn->setConvFunctions( xline0, xline1 ); 
    if (strcmp(param.convDirName,"curve")==0)
       eqn->setConvFunctions( curve0, curve1 ); 
    if (strcmp(param.convDirName,"square")==0) 
       eqn->setConvFunctions( square0, square1 );    
    if (strcmp(param.convDirName,"foursquares")==0) 
       eqn->setConvFunctions( foursquares0, foursquares1 ); 
    if (strcmp(param.convDirName,"fourcircles")==0) 
       eqn->setConvFunctions( fourcircles0, fourcircles1 ); 

    T2DStokesHierarchy *hierarchy = new T2DStokesHierarchy();
    T2DStokesFE *fe = new T2DStokesFE( Tabata );  

    timer.start();
    hierarchy->buildHierarchy( dofManager, eqn, fe, netHier, pMatrix,
                               prolongation, pRHS,pX );

    timer.stop();
    cout << endl;
	cout << "time to discretizing on " << param.noOfNets-1 
         << " levels: "  << timer << endl;
    cout << endl;
 
	//cout << (*pMatrix[param.noOfNets-2])[0][0] << endl << endl;
	//cout << (*pMatrix[param.noOfNets-3])[0][2] << endl;

   TBlockVector<TFloat> *vecSolution = new TBlockVector<TFloat>( *pX[param.noOfNets-2] ); 

   uint dimU = (*vecSolution)[0].getSize();
   uint dimP = (*vecSolution)[2].getSize();
   cout << "Dimension: U: " << dimU << ", P: " << dimP << endl;
   //cout << "Gridsize:  " << sqrt(dimU) + 1 << "x" << sqrt(dimU) + 1 
   //     << endl;

   T2DNet *net = iterNet->last();
   net = iterNet->prev();

   TPArray<T2DVector> *arrVectorsU = new TPArray<T2DVector>;
   TPArray<T2DVector> *arrVectorsP = new TPArray<T2DVector>;
   fe->getCoord(net, *arrVectorsU, *arrVectorsP); 
 
   // calculate exact solution
   for (i = 0; i < dimU; i++) {
       (*vecSolution)[0][i] = eqn->dirichlet0( *((*arrVectorsU)[i]) );
       (*vecSolution)[1][i] = eqn->dirichlet1( *((*arrVectorsU)[i]) );
   }// for
   for (i = 0; i < dimP; i++) {
       (*vecSolution)[2][i] = p( *((*arrVectorsP)[i]) );
   }// for
   // Lagrange factor
   //(*vecSolution)[2][dimP - 1] = 0.0;
   //cout << "solution\n" << *vecSolution << endl;

    // ordering
    timer.start();
    TBlockInfo  *blockinfo = new TBlockInfo;
    blockinfo->setSortMethod( _SORT_OFD_ );
    //blockinfo->setSortMethod( _SORT_EDGE_WEIGHT_ );
    //blockinfo->setSortMethod( _SORT_SUCC_EDGE_WEIGHT_ );

    //
    // build data structures for each net
    //
    TVectorBlock<ulong>  *vecPerm[param.noOfNets-1];
    TVectorBlock<ulong>  *cyclicBlocks[param.noOfNets-1];
    TVectorBlock<ulong>  *arrFVSize = new TVectorBlock<ulong>();
 
    if ( param.ordering != _NONE_ )	{
	    for ( i = 0; i < param.noOfNets-1; i++ ) { 
            vecPerm[i] = new TVectorBlock<ulong>;
            cyclicBlocks[i] = new TVectorBlock<ulong>;
       }// for
    }// if

   //
   // now build permutation vectors
   //
   if ( param.ordering & _CONCENTRIC_ ) {
     
	  if (param.draw) {
	     blockinfo->setCycleName("psfiles/cycle.ps"); 
	     blockinfo->setGraphName("psfiles/graph.ps");
	  }
      timer.start();
	  
		switch (param.mgSort)
		{
		case _MGSORT_PROJECTED_ :
			blockinfo->buildProjectedPermutation( vecPerm, 
                                                  cyclicBlocks,
										   		  *fe, 
                                                  netHier->getList(), 
                                                  pMatrix, 
												  param.xi, 
                                                  param.criterion,
                                                  param.ordering );
			break;
		case _MGSORT_NESTED_    :
			blockinfo->buildNestedPermutation( vecPerm, cyclicBlocks,
											   *fe, 
                                               netHier->getList(), 
                                               pMatrix, 
											   param.xi, 
                                               param.criterion,
                                               param.ordering  );
			break;
		case _MGSORT_INDEPEND_  :
			blockinfo->buildIndependentPermutation( vecPerm, 
                                                cyclicBlocks,
												*fe, 
                                                netHier->getList(), 
                                                pMatrix, 
												param.xi, 
                                                param.criterion,
                                                param.ordering  );
			break;
		default:
			break;
		}// switch

        timer.stop();
        cout << endl << "time for finding cycles : " << timer << endl;
        for (i = 0; i < param.noOfNets-1; i++)
            cout << " Number of cycles in net " << i+1 << " :  " 
                 << cyclicBlocks[i]->getSize()/2 << endl;
	}
   if ( param.ordering == _FVS_ ) {
     
        timer.start();
 		switch (param.mgSort)
		{
		case _MGSORT_PROJECTED_ :
			blockinfo->buildProjectedPermutation( vecPerm, arrFVSize,
												 *fe, netHier->getList(),
                                                  pMatrix, 
												  param.xi, 
                                                  param.criterion );
			break;
		case _MGSORT_NESTED_    :
			blockinfo->buildNestedPermutation( vecPerm, arrFVSize,
											   *fe, netHier->getList(), 
                                               pMatrix, 
											   param.xi, param.criterion);
			break;
		case _MGSORT_INDEPEND_  :
			blockinfo->buildIndependentPermutation( vecPerm, arrFVSize,
											*fe, netHier->getList(), 
                                               pMatrix, 
											   param.xi, param.criterion);
			break;
		default:
			break;
		}// switch
        timer.stop();
        cout << endl << "time for finding FVS : " << timer << endl;
        for (i = 0; i < param.noOfNets-1; i++) 
            cout << " Number of FVS nodes in net " << i+1 << " :  " 
                << (*arrFVSize)[i] << endl;
   }

    if ( param.ordering == _RCM_ ) {
		timer.start();
		switch (param.mgSort)
		{
		case _MGSORT_PROJECTED_ :
			blockinfo->buildRCMProjectedPermutation( vecPerm, pMatrix, 
													 param.noOfNets );
			break;
		case _MGSORT_INDEPEND_  :
			blockinfo->buildRCMIndependentPermutation( vecPerm, pMatrix, 
													   param.noOfNets );
			break;
		default:
			break;
		}// switch
        timer.stop();
        cout << endl << "time for doing RCM : " << timer << endl;
	}

   //
   // build the inverse permutations
   //
   TVectorBlock<ulong> *vecInvPerm[param.noOfNets-1];
    
   if ( param.ordering != _NONE_ )	{
       cout << "building inverse permutations" << endl;
	    for ( i = 0; i < param.noOfNets-1; i++ ) { 
           ulong size = vecPerm[i]->getSize();
		   vecInvPerm[i] = new TVectorBlock<ulong>( 0, size );
           (*vecInvPerm[i]) = invertPermutation( *vecPerm[i] );
		}// for 
   }// if

	//
	// reorder
	//
    if (param.ordering != _NONE_ ) {
	    timer.start();
        for ( i = 1; i < param.noOfNets-1; i++ ) 
		{
	        (*pRHS[i]).permute( vecInvPerm[i] );
            (*pMatrix[i]).permute( vecInvPerm[i] );
            if ( i > 1 )
            (prolongation[i-1]->pBlockProlong)[0]->permute( vecInvPerm[i], vecInvPerm[i-1]);
            (prolongation[i-1]->pBlockProlong)[1]->permute( vecInvPerm[i], vecInvPerm[i-1]);
		}
	    timer.stop();
	    cout << "time for permuting matrices and rhs's: "<< timer << endl;
	}	
	
    // draw matrix
    TDrawApp *app = new TDrawApp;
	//if (param.draw) 
		//app->draw( *pMatrix[param.noOfNets - 2], param.xi, "psfiles/matrix.ps");


    timer.start();
    //(*pMatrix[param.startLevel]).calculateILUdecomp();
    //(*pMatrix[param.startLevel]).calculateLUdecomp();
    (*pMatrix[param.startLevel]).calcStokesLUdecomp1();
    timer.stop();
	cout << "time to decompose: " << timer << endl;

   //cout <<"matrix 0,2\n"<<(*pMatrix[param.noOfNets - 3])[0][2] << endl;
   //matWriteCompleteMatrix("Ah.mat",(*pMatrix[param.noOfNets - 2]),"Ah");
   //matWriteCompleteMatrix("AH.mat",(*pMatrix[param.noOfNets - 3]),"AH");
   //matWriteBlockProlongation( "prol",(*prolongation[param.noOfNets-3]),"prol");


    // solve resulting equations

    uint preSmSteps    = 1;           // number of presmoothing steps
    uint postSmSteps   = 1;           // number of postsmoothing steps
            
    uint mgMaxLevel = param.endLevel;
    uint mgMinLevel = param.startLevel;

    // V-Cycle: cycle = 1, W - Cycle: cycle = 2 
    uint cycle = 2; 	 

    // solvers
    TLU *lu = new TLU( );
    lu->setMaxit( 1 );
    TILU *ilu = new TILU( );
    TRichardson *rich = new TRichardson( );
    TKaczmarz *kaczmarz = new TKaczmarz( );
    TSquaredSmoother *squaredSmoother = new TSquaredSmoother();
    TSchur *schur = new TSchur();
    TBicgStab *bicg = new TBicgStab( squaredSmoother );
   
    TMultigrid *multigrid = new TMultigrid( pMatrix, pRHS, pX, 
                                prolongation, mgMinLevel, mgMaxLevel,
                                cycle, squaredSmoother, squaredSmoother, 
                                lu, param.steps );
    multigrid->setPreSmoothingSteps( preSmSteps );
    multigrid->setPostSmoothingSteps( postSmSteps );

   TBlockVector<TFloat> *start = new TBlockVector<TFloat> (*pX[mgMaxLevel]);
   start->set( 0.0 );
   (*start)[0].set( -10.0 );
   (*start)[1].set( -10.0 );
   *pX[mgMaxLevel] = *start;

   TBlockVector<TFloat> *error=new TBlockVector<TFloat> (*pX[mgMaxLevel]);

   *error -= *vecSolution;
   cout << "\ninitial error:\n";
   cout << "norm u_ex - u_sol: " 
        << error->weightedEuclidBlockNorm(0) << endl;     
   cout << "norm v_ex - v_sol: "
        << error->weightedEuclidBlockNorm(1) << endl;     
   cout << "norm p_ex - p_sol: "
        << error->weightedEuclidBlockNorm(2) << endl;     

    uint verbosity = 5;

    bicg->setAccuracy( 1e-15 );
    bicg->setMaxit( param.steps );
    bicg->setVerbosity( verbosity );
    ilu->setMaxit( param.steps * (preSmSteps + postSmSteps) );
    ilu->setVerbosity( verbosity );
    kaczmarz->setMaxit( param.steps * (preSmSteps + postSmSteps));
    kaczmarz->setVerbosity( verbosity );
    schur->setMaxit( param.steps * (preSmSteps + postSmSteps));
    schur->setVerbosity( verbosity );
    squaredSmoother->setMaxit( param.steps* (preSmSteps + postSmSteps) );
    squaredSmoother->setVerbosity( verbosity );
    squaredSmoother->setEpsilon( param.epsilon );
    //squaredSmoother->setDampFactor( 1.0 );

	/*
	cout << "\n------------ Simple solver ------------\n";
    timer.start();
    *pX[mgMaxLevel] = *start;
    squaredSmoother->setMode( 1 );
	squaredSmoother->solve(pMatrix[mgMaxLevel], pRHS[mgMaxLevel], pX[mgMaxLevel]);
    timer.stop();
	cout << "time for squared smoother: " << timer << endl;
	
    timer.start();
    *pX[mgMaxLevel] = *start;
    kaczmarz->setMode( 1 );
    kaczmarz->solve(pMatrix[mgMaxLevel],pRHS[mgMaxLevel],pX[mgMaxLevel]);
    timer.stop();
	cout << "time for Kaczmarz: " << timer << endl;
	
    timer.start();
    *pX[mgMaxLevel] = *start;
    ilu->setMode( 1 );
	ilu->solve(pMatrix[mgMaxLevel], pRHS[mgMaxLevel], pX[mgMaxLevel]);
    timer.stop();
	cout << "time for ilu: " << timer << endl;

    timer.start();
    *pX[mgMaxLevel] = *start;
    schur->setMode( 1 );
    schur->solve(pMatrix[mgMaxLevel],pRHS[mgMaxLevel],pX[mgMaxLevel]);
    timer.stop();
	cout << "time for Schur: " << timer << endl;
	*/

	/*
	cout << "\n------------ Bicg solver ------------\n";

    timer.start();
    *pX[mgMaxLevel] = *start;
    bicg->setPrec( squaredSmoother );
    bicg->solve(pMatrix[mgMaxLevel], pRHS[mgMaxLevel], pX[mgMaxLevel]);
    timer.stop();
	cout << "time for bicg: " << timer << endl;

    timer.start();
    *pX[mgMaxLevel] = *start;
    bicg->setPrec( schur );
    bicg->solve(pMatrix[mgMaxLevel], pRHS[mgMaxLevel], pX[mgMaxLevel]);
    timer.stop();
	cout << "time for bicg: " << timer << endl;

    timer.start();
    *pX[mgMaxLevel] = *start;
    bicg->setPrec( ilu );
    bicg->solve(pMatrix[mgMaxLevel], pRHS[mgMaxLevel], pX[mgMaxLevel]);
    timer.stop();
	cout << "time for bicg: " << timer << endl;
	*/

	cout << "\n------------ Multigrid solver ------------\n";

    timer.start();
    squaredSmoother->setVerbosity( 0 );
    *pX[mgMaxLevel] = *start;
    multigrid->setSmoother( squaredSmoother );
    multigrid->setVerbosity( verbosity );
    multigrid->solve();
    timer.stop();
	cout << "time for multigrid: " << timer << endl;

    multigrid->setMode( 1 );
    timer.start();
    kaczmarz->setVerbosity( 0 );
    *pX[mgMaxLevel] = *start;
    multigrid->setVerbosity( verbosity );
    multigrid->setSmoother( kaczmarz );
    multigrid->solve();
    timer.stop();
	cout << "time for multigrid: " << timer << endl;

    multigrid->setMode( 1 );
    timer.start();
    schur->setVerbosity( 0 );
    //schur->setDampFactor( 1.0 / 2.0 );
    schur->setInnerSteps( 5 );
    *pX[mgMaxLevel] = *start;
    multigrid->setVerbosity( verbosity );
    multigrid->setSmoother( schur );
    multigrid->solve();
    timer.stop();
	cout << "time for multigrid: " << timer << endl;

    multigrid->setMode( 1 );
    timer.start();
    ilu->setVerbosity( 0 );
    //ilu->setDampFactor( 1.0 );
    *pX[mgMaxLevel] = *start;
    multigrid->setVerbosity( verbosity );
    multigrid->setSmoother( ilu );
    multigrid->solve();
    timer.stop();
	cout << "time for multigrid: " << timer << endl;

   TBlockVector<TFloat> *defect = 
                        new TBlockVector<TFloat>(*pRHS[mgMaxLevel]);
   *defect -= ( (*pMatrix[mgMaxLevel]) * (*pX[mgMaxLevel]) );
   //cout << "defect:\n" << (*defect)[2] << endl;   

   cout << "\ndefect of numerical solution:\n";
   cout << "weighted defect 0 : "
        << (*defect).weightedEuclidBlockNorm(0) << endl;
   cout << "weighted defect 1 : "
        << (*defect).weightedEuclidBlockNorm(1) << endl;
   cout << "weighted defect 2 : "
        << (*defect).weightedEuclidBlockNorm(2) << endl;
   (*defect)[2][(*pRHS[mgMaxLevel])[2].getSize()-1] = 0.0;
   cout << "weighted defect 2 without lagrange: "
        << (*defect).weightedEuclidBlockNorm(2) << endl;
   
   // compare with exact solution
   cout << "\nerror between exact solution and numerical solution:" << endl;
   if (param.ordering != _NONE_ ) 
       pX[mgMaxLevel]->permute( vecPerm[param.noOfNets-2] );

   *error = *pX[mgMaxLevel] - *vecSolution;
   //cout << "error\n" << (*error)[2] << endl;
   //cout << "solution\n" << (*vecSolution)[2] << endl;

   //(*error)[2][(*pRHS[mgMaxLevel])[2].getSize()-1] = 0.0;
   cout << "norm u_ex - u_sol: " 
        << error->weightedEuclidBlockNorm(0) << endl;     
   cout << "norm v_ex - v_sol: "
        << error->weightedEuclidBlockNorm(1) << endl;     
   cout << "norm p_ex - p_sol: "
        << error->weightedEuclidBlockNorm(2) << endl;     
   
   // exact solution
   *defect = *pMatrix[mgMaxLevel] * *vecSolution - *pRHS[mgMaxLevel];
   cout << "\ndefect of exact solution:\n";
   cout << "weighted defect 0 : "
        << (*defect).weightedEuclidBlockNorm(0) << endl;
   cout << "weighted defect 1 : "
        << (*defect).weightedEuclidBlockNorm(1) << endl;
   cout << "weighted defect 2 : "
        << (*defect).weightedEuclidBlockNorm(2) << endl;
   (*defect)[2][(*pRHS[mgMaxLevel])[2].getSize()-1] = 0.0;
   cout << "weighted defect 2 without lagrange: "
        << (*defect).weightedEuclidBlockNorm(2) << endl << endl;

    // draw exact and numerical solution

	//net = iterNet->last();
    // draw all the usual things
   if (param.draw)  
    {
        net = iterNet->prev();
	    //app->draw( *net, "psfiles/coarsecoarseNet.ps" );
        net = iterNet->next();
	    //app->draw( *net, "psfiles/coarseNet.ps" );
        net = iterNet->next();
	    //app->draw( *net, "psfiles/fineNet.ps" );
	    app->draw( *net, *vecInvPerm[param.noOfNets-2], "psfiles/fineNetPerm.ps" );
        app->draw( *arrVectorsU, *pX[mgMaxLevel], *net, 
                   "psfiles/vecfield.ps" );
        //app->draw( *arrVectorsU, *vecSolution, *net, 
        //           "psfiles/vecfield_sol.ps" );
        //app->drawStreakLine( *net, *pX[mgMaxLevel], *eqn, 
        //           "psfiles/streaklines.ps");
    }// if

   //cout << (*pX[mgMaxLevel])[0][0] << endl << (*pX[mgMaxLevel])[1][0] << endl;
   //cout << (*vecSolution)[0][0] << endl << (*vecSolution)[1][0] << endl;
   //cout << (*vecSolution)[2][0] << endl << (*vecSolution)[2][1] << endl;


    delete (app);

    delete (arrVectorsU);
    delete (arrVectorsP);
   
    delete (netHier);
    delete (iterNet);
   
    delete (rich);
    delete (kaczmarz);
    delete (bicg);
    delete (multigrid);
    delete (schur);
    delete (squaredSmoother);
 
    delete (dofManager);
    delete (fe);
    delete (eqn);
    delete (hierarchy);
    
    delete (defect);
    delete (vecSolution);
    delete (error);

    delete (arrFVSize);
    delete (blockinfo);
    if ( param.ordering != _NONE_ )
        for ( i = 1; i < param.noOfNets-1; i++ ) { 
            delete (vecInvPerm[i]);
            delete (vecPerm[i]);
            delete (cyclicBlocks[i]);
		}

    for (i = 0; i < mgMaxLevel+1; i++) {
      if (pX[i]) delete pX[i]; 
      if (pRHS[i]) delete pRHS[i];  
      if (pMatrix[i]) delete pMatrix[i];  
	}
      for (i = 0; i < mgMaxLevel; i++) {
          if (prolongation[i]) delete (prolongation[i]);
	}
 
    TMemory::statistics();
	
}
