//
//  Project   : Navier-Stokes
//  File      : convdiff.cc
//  Author    : Ronald Kriemann
//  Purpose   : test of 3D conv-diff
//
// arch-tag: 075fccc7-1527-474e-880f-e92241d0755d
//

#include <stdio.h>
#include <stdlib.h>

#include "var.hh"
#include "memory.hh"
#include "timer.hh"
#include "cmdline.hh"
#include "TRandom.hh"

#include "T3DFrontier.hh"
#include "T3DCellEdge.hh"
#include "net.hh"
#include "TCdrBoundaryDesc.hh"
#include "TDomain.hh"
#include "T3DMeshGenerator.hh"

#include "TConvDiffDoFManager.hh"
#include "TConvDiffFE.hh"
#include "TConvDiffHierarchy.hh"
#include "flowdef.hh"

#include "blockvector.hh"

#include "secondNF.hh"
#include "fastGaussSeidel.hh"
#include "bicgStab.hh"
#include "cgSolver.hh"
#include "TSOR.hh"
#include "TGMRES.hh"

#include "T3DGraph.hh"
#include "T3DPermBuilder.hh"
#include "criterion.hh"

#include "drawapp.hh"
#include "matrixview.hh"

//
// forward decl.
//

static void orderAndSolve ( TCdrEqn<3>            * pEqn,
                            TNetHierarchy<3>      * pNetHier,
                            TBlockMatrix         ** pA,
                            TBlockVector<TFloat> ** pRHS,
                            TBlockVector<TFloat> ** pX );

static void write_sparse ( TMatrixBlock & A, const char * filename );
static void write_nodes  ( TNet<3> * net, const char * filename );

static TNet<3> * buildQuadNet ( T3DMeshGenerator *, TDomain *, TBoundary * );

//
// definition of equation to solve
//
class TDirichletFun : public TFunction<3>
{
public:
    TDirichletFun() {};
    ~TDirichletFun() {};

    TFloat eval ( TVector <3> *v )
    {
        TFloat a = (*v)[0];
        TFloat b = (*v)[1];
        TFloat c = (*v)[2];
        
        //  return a+b+2*c;
        return (a*a)+(b*b)+(c*c);
    }
};

//
// overload domain to set boundaries
//
class TMyDomain : public TDomain
{
public:
    TMyDomain ()
            : TDomain ()
    {}
    
    TBoundary * boundary ( TFloat x, TFloat y, TFloat z )
    {
        if ((x > TFloat(0.01)) && (x < TFloat(0.99)) &&
            (y > TFloat(0.01)) && (y < TFloat(0.99)) &&
            (z > TFloat(0.01)) && (z < TFloat(0.99)))
            return NULL;
        else
            return listBound.getFirst();
    }
};

//
// mainfunction
//
int
main ( int argc, char **argv ) 
{
    uint    i;
    TTimer  timer;

    //
    // parse command line
    //
    param.convdir     = "circle";
    param.noOfNets    = 5;
    param.criterion   = 7;
    param.xi          = 1.5;
    param.steps       = 400;
    param.solver      = "gmres";
    param.prec        = "default";
    param.startLevel  = 1;
    param.starttri    = "grids/cube.tri";

    param.evalCmdLine( argc, argv );
    param.endLevel = param.noOfNets - 1;
    cout << "main : chosen parameters = " << endl << param << endl << endl;

    //
    // init discr functions
    //
    TCdrEqn<3>          * pEqn;
    TFunction<3>        * pDirichlet;

    if (param.convdir == "none")
    {
        pEqn = new TNoConvCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 0.1, 0.1, 0.1, "flow" );
    }// if
    else if (param.convdir == "xline")
    {
        pEqn = new TXLineCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 0.2, 0.1, 1.0/2.0, "flow" );
    }// if
    else if (param.convdir == "curve")
    {
        pEqn = new TCurveCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 0.1, 0.1, 1.0/2.0, "flow" );
    }// if
    else if (param.convdir == "circle")
    {
        pEqn = new TCircleCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 1.0/10.0, 1.0/10.0, 1.0/2.0, "flow" );
    }// if
    else if (param.convdir == "4circles")
    {
        pEqn = new T4CirclesCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 1.0/15.0, 1.0/15.0, 1.0/2.0, "flow" );
    }// if
    else if (param.convdir == "diagcircle")
    {
        pEqn = new TDiagCircleCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 0.1, 0.1, 0.1, "flow" );
    }// if
    else if (param.convdir == "ball")
    {
        pEqn = new TBallCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 0.1, 0.1, 0.1, "flow" );
    }// if
    else if (param.convdir == "vortex")
    {
        pEqn = new TVortexCdrEqn;

        if (param.draw)
            TDrawApp::drawFlow( pEqn, 0.1, 0.1, 0.1, "flow" );
    }// if
    else
    {
        cout << "main : convection (" << param.convdir << ") unsupported" << endl;
        exit( 1 );
    }// else
    
    pEqn->isLaplace(  TRUE );
    pEqn->Diffusion(  TRUE  );
    pEqn->Convection( TRUE  );
    pEqn->setEpsilon( param.epsilon );

    if (param.reactionTerm)
        pEqn->Reaction( TRUE );
    
    pDirichlet = new TDirichletFun;

    //
    // build net
    //
    timer.start();

    TNetHierarchy<3>  * pNetHier;
    TNet<3>           * pNet;

    if (param.starttri == "intern")
    {
        TBoundary         * pBoundary;
        T3DMeshGenerator  * pMeshGen = new T3DMeshGenerator;
        
        pNetHier = new TNetHierarchy<3>;

        pGlobalDomain = new TMyDomain;
        pBoundary = new TBoundary( "b1" );
        
        pGlobalDomain->addBoundary( pBoundary );

        pNet = buildQuadNet( pMeshGen, pGlobalDomain, pBoundary );
        
        pNet->delRef();
        pNetHier->add( pNet );

        delete pMeshGen;
    }// if
    else
        pNetHier = new TNetHierarchy<3>( param.starttri );

    if (param.draw)
        TDrawApp::drawNet( pNetHier->getList().getFirst(), "net" );

    for (i = 1; i < param.noOfNets; i++)
    {
        pNet = pNetHier->buildNet();
    }// for

    timer.stop();
    cout << "main : time for building grids = " << timer << endl;

    cout << "main : used memory = " << TMemory::used() << " kB" << endl;

    //
    // supply boundary Descriptons
    //
    TListIterator< TBoundary >  * pIterBound;
    TCdrBoundaryDesc<3>         * pBoundDesc;
    TBoundary                   * pBoundary;
    
    pBoundDesc = new TCdrBoundaryDesc<3>( 1 );
    
    pBoundDesc->setType( 0, DIR_TYPE );
    pBoundDesc->setFunc( 0, pDirichlet );
    pBoundDesc->setDoFInfo();

    pIterBound = new TListIterator< TBoundary >;
    pIterBound->setList( pGlobalDomain->getBoundaries() );
    
    pIterBound->first();
    while ((pBoundary = (*pIterBound)++) != NULL)
    {
        if (pBoundary->getName() == "b1" )
            pBoundary->setBoundaryDesc( pBoundDesc );
    }// while

    //
    // discretize
    //
    TConvDiffDoFManager<3>  * pDoFManager;
    TConvDiffHierarchy<3>   * pDiscrHier;
    TConvDiffFE<3>          * pFE;
    TBlockMatrix           ** pA;
    TBlockVector<TFloat>   ** pRHS;
    TBlockVector<TFloat>   ** pX;
    
    pDoFManager = new TConvDiffDoFManager<3>;
    pDiscrHier  = new TConvDiffHierarchy<3>;
    pFE         = new TConvDiffFE<3>( Tabata );
    pA          = new (TBlockMatrix *)[ param.noOfNets ];
    pRHS        = new (TBlockVector<TFloat> *)[ param.noOfNets ];
    pX          = new (TBlockVector<TFloat> *)[ param.noOfNets ];

    for ( i = 0; i < param.noOfNets; i++ )
    {
        pA[i]       = NULL;
        pRHS[i]     = NULL;
        pX[i]       = NULL;
    }// for
    
    timer.start();
    
    pDiscrHier->buildHierarchy( pDoFManager, pEqn, pFE,
                                pNetHier,
                                pA, pRHS, pX );
    
    timer.stop();
    cout << "main : time for discretizing = " << timer << endl << endl;
    cout << "main : used memory = " << TMemory::used() << " kB" << endl;
    
    if (param.draw)
        TDrawApp::drawMatrix( & (*pA[param.endLevel])[0][0], param.xi, "matrix" );

    cout << "main : dim of coarse net = " << (*pA[param.startLevel])[0][0].getSize() << endl;
    cout << "main : dim of fine   net = " << (*pA[param.endLevel])[0][0].getSize() << endl;
   
    if ( param.write_matrix )
    { 
        write_sparse( (*pA[param.endLevel])[0][0], "matrix.txt" );
        write_nodes( pNet, "vertices.txt" );
        return 0;
    }// if

    //
    // solve equation
    //

    if ( param.ordering == ORDER_ALL )
    {
        cout << "##################################################" << endl
             << "#"                                                  << endl
             << "#  ordering = NONE"                                 << endl
             << "#"                                                  << endl
             << "##################################################" << endl;
        param.ordering = NONE;
        orderAndSolve( pEqn, pNetHier, pA, pRHS, pX );

        cout << "##################################################" << endl
             << "#"                                                  << endl
             << "#  ordering = RCM"                                  << endl
             << "#"                                                  << endl
             << "##################################################" << endl;
        param.ordering = RCM;
        orderAndSolve( pEqn, pNetHier, pA, pRHS, pX );

        cout << "##################################################" << endl
             << "#"                                                  << endl
             << "#  ordering = hFVS"                                 << endl
             << "#"                                                  << endl
             << "##################################################" << endl;
        param.ordering = HEURISTIC_FVS;
        param.global   = true;
        orderAndSolve( pEqn, pNetHier, pA, pRHS, pX );

        if ((param.convdir == "circle") || (param.convdir == "4circles") ||
            (param.convdir == "diagcircle"))
        {
            cout << "##################################################" << endl
                 << "#"                                                  << endl
                 << "#  ordering = CYC (FS)"                             << endl
                 << "#"                                                  << endl
                 << "##################################################" << endl;
            param.ordering = CONCENTRIC_MIN;
            param.global   = false;
            orderAndSolve( pEqn, pNetHier, pA, pRHS, pX );

            cout << "##################################################" << endl
                 << "#"                                                  << endl
                 << "#  ordering = hFVS (FS)"                            << endl
                 << "#"                                                  << endl
                 << "##################################################" << endl;
            param.ordering = HEURISTIC_FVS;
            param.global   = false;
            orderAndSolve( pEqn, pNetHier, pA, pRHS, pX );

            cout << "##################################################" << endl
                 << "#"                                                  << endl
                 << "#  ordering = pFVS (FS)"                            << endl
                 << "#"                                                  << endl
                 << "##################################################" << endl;
            param.ordering = PLANAR_FVS;
            param.global   = false;
            orderAndSolve( pEqn, pNetHier, pA, pRHS, pX );
        }// if
    }// if
    else
        orderAndSolve( pEqn, pNetHier, pA, pRHS, pX );
    
    cout << "main : used memory = " << TMemory::used() << " kB" << endl;

    //
    // clean up
    //
    
    for ( i = 0; i < param.noOfNets; i++ )
    {
        delete pA[i];
        delete pRHS[i];
        delete pX[i];
    }// for
    
    delete pA;
    delete pRHS;
    delete pX;

    pNetHier->free();
    delete pEqn;
    delete pFE;
    delete pDoFManager;
    delete pDiscrHier;
    delete pGlobalDomain;

    TMemory::statistics();
}

////////////////////////////////////////////////////////////////////////////
//
// misc. methods
//

//
// compute ordering, permute matrix, solve equation and unpermute matrix
//
static void
orderAndSolve ( TCdrEqn<3>            * pEqn,
                TNetHierarchy<3>      * pNetHier,
                TBlockMatrix         ** A,
                TBlockVector<TFloat> ** RHS,
                TBlockVector<TFloat> ** x )
{
    TTimer           timer;

    uint             i;
    TSolver        * pPrec;
    TSolver        * pCoarse;
    TSolver        * pPreSmoother, * pPostSmoother;
    TSolver        * pSolver;
    T3DPermBuilder   permBuilder;
    TPermutation  ** pPerm, ** pInvPerm;

    param.startLevel = param.endLevel;
    
    pPerm    = new (TPermutation *)[ param.noOfNets ];
    pInvPerm = new (TPermutation *)[ param.noOfNets ];
    
    for ( i = 0; i < param.noOfNets; i++ )
    {
        pPerm[i] = pInvPerm[i] = NULL;
    }// for

    //
    // build permutation
    //
    timer.start();

    if (param.ordering == HEURISTIC_FVS)
    {
        if (param.global)
            permBuilder.buildMGFVSPerm( A, pPerm,
                                        param.startLevel,
                                        param.endLevel );
        else
            permBuilder.buildMGFVSPerm( pNetHier, pEqn,
                                        A, pPerm,
                                        param.startLevel,
                                        param.endLevel,
                                        HEURISTIC_FVS );
    }// if

    if (param.ordering == PLANAR_FVS)
        permBuilder.buildMGFVSPerm( pNetHier, pEqn,
                                    A, pPerm,
                                    param.startLevel,
                                    param.endLevel,
                                    PLANAR_FVS );

    if (param.ordering & CONCENTRIC)
        permBuilder.buildMGCYCPerm( pNetHier, pEqn,
                                    A, pPerm,
                                    param.startLevel,
                                    param.endLevel );
            
    if (param.ordering == RCM)
        permBuilder.buildMGRCMPerm( A, pPerm,
                                    param.startLevel,
                                    param.endLevel );

    timer.stop();
    cout << "orderAndSolve : time for building permutation = " << timer << endl;

    //
    // invert permutations
    //
    for ( i = param.startLevel; i <= param.endLevel; i++ )
    {
        if (pPerm[i] != NULL)
            pInvPerm[i] = new TPermutation( pPerm[i]->invert() );
        else
            pInvPerm[i] = NULL;
    }// for
    
    //
    // permute matrix
    //
    timer.start();

    for ( i = param.startLevel; i <= param.endLevel; i++ )
    {
        if (pInvPerm[i] != NULL)
        {
            A[i]->permute( pInvPerm[i] );
            RHS[i]->permute( pInvPerm[i] );
        }// if
    }// for

    timer.stop();
    cout << "orderAndSolve : time for permute = " << timer << endl;

    if (param.draw)
    {
        if (permBuilder.blockStructure() != NULL)
            TDrawApp::drawMatrix( & (*A[param.endLevel])[0][0], param.xi,
                                  permBuilder.blockStructure(), "pmatrix" );
        else
            TDrawApp::drawMatrix( & (*A[param.endLevel])[0][0], param.xi, "pmatrix" );
    }// if
    
    *x[ param.endLevel ] = 1000;

    //
    // choose preconditioner
    //

    pPrec = pCoarse = pPreSmoother = pPostSmoother = pSolver = NULL;
    
    if      (param.prec == "none") pPrec = new TRichardson;
    else if (param.prec == "jac")  pPrec = new TJacobi;
    else if (param.prec == "sgs")  pPrec = new TSGS;
    else if (param.prec == "bgs")  pPrec = new TBGS;
    else if (param.prec == "fgs")  pPrec = new TFGS;
    else if (param.prec == "ilu")  pPrec = new TILU;
    else if (param.prec == "default")
    {
        if ((param.ordering == PLANAR_FVS) || (param.ordering == HEURISTIC_FVS))
            pPrec = new TBGS;
        else
            pPrec = new TSGS;
    }// if
    else
    {
        cout << "main : unsupported preconditioner (" << param.prec << ")" << endl;
        exit( 1 );
    }// else

    if (pPrec != NULL)
    {
        pPrec->setMode( 0 );
        pPrec->setMaxit( 1 );
        pPrec->setVerbosity( 0x0 );
    }// if

    //
    // choose solver
    //
    
    if      (param.solver == "jac")    pSolver = new TJacobi;
    else if (param.solver == "ilu")    pSolver = new TILU;
    else if (param.solver == "sgs")    pSolver = new TFastSymmetricGaussSeidel;
    else if (param.solver == "bgs")    pSolver = new TFastBackwardGaussSeidel;
    else if (param.solver == "fgs")    pSolver = new TFastForwardGaussSeidel;
    else if (param.solver == "bicg")   pSolver = new TBicgStab( pPrec );
    else if (param.solver == "gmres")  pSolver = new TGMRES( pPrec );
    else
    {
        cout << "orderAndSolve : unsupported solver (" << param.solver << ")" << endl;
        exit( 1 );
    }// else

    pSolver->setMode( MODE_REDUCEBY );
    pSolver->setAccuracy( 1e-4 );
    pSolver->setMaxit( param.steps );
    pSolver->setVerbosity( 0xff );

    //
    // solve with permuted matrix
    //

    timer.start();

    pSolver->solve( param.endLevel, A, RHS, x );
    
    timer.stop();
    cout << "orderAndSolve : time for solving = " << timer << endl;

    //
    // unpermute
    //

    for ( i = param.startLevel; i <= param.endLevel; i++ )
    {
        if (pPerm[i] != NULL)
        {
            A[i]->permute( pPerm[i] );
            RHS[i]->permute( pPerm[i] );
        }// if
    }// for

    if (pPerm[param.endLevel] != NULL)
        x[param.endLevel]->permute( pPerm[param.endLevel] );
    
    //
    // clean up
    //
    
    if (pPrec)         delete pPrec;
    if (pCoarse)       delete pCoarse;
    if (pPreSmoother)  delete pPreSmoother;
    if (pPostSmoother) delete pPostSmoother;
    if (pSolver)       delete pSolver;

    for ( i = 0; i < param.noOfNets; i++ )
    {
        if (pPerm[i] != NULL)
            delete pPerm[i];

        if (pInvPerm[i] != NULL)
            delete pInvPerm[i];
    }// for

    delete pPerm;
    delete pInvPerm;

}

//
// create net by hand
//
static TNet<3> *
buildQuadNet ( T3DMeshGenerator * meshGen, TDomain * domain, TBoundary * boundary )
{
    int           x, y, z;
    TVertex<3>  * aV[4][4][4];

    x = y = z = 0;
    
    aV[x++][y][z] = new TVertex<3>(    0,    0, 0, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.28,    0, 0, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.69,    0, 0, boundary );
    aV[x++][y][z] = new TVertex<3>(    1,    0, 0, boundary ); x = 0; y++;

    aV[x++][y][z] = new TVertex<3>(    0, 0.39, 0, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.30, 0.30, 0, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.69, 0.25, 0, boundary );
    aV[x++][y][z] = new TVertex<3>(    1, 0.30, 0, boundary ); x = 0; y++;

    aV[x++][y][z] = new TVertex<3>(    0, 0.63, 0, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.30, 0.73, 0, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.70, 0.69, 0, boundary );
    aV[x++][y][z] = new TVertex<3>(    1, 0.62, 0, boundary ); x = 0; y++;

    aV[x++][y][z] = new TVertex<3>(    0,    1, 0, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.36,    1, 0, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.69,    1, 0, boundary );
    aV[x++][y][z] = new TVertex<3>(    1,    1, 0, boundary ); x = 0; y = 0; z++; 



    aV[x++][y][z] = new TVertex<3>(    0,    0, 0.33, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.33,    0, 0.36, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.69,    0, 0.30, boundary );
    aV[x++][y][z] = new TVertex<3>(    1,    0, 0.33, boundary ); x = 0; y++;

    aV[x++][y][z] = new TVertex<3>(    0, 0.36, 0.33, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.30, 0.38, 0.36, INTERIOR );
    aV[x++][y][z] = new TVertex<3>( 0.72, 0.30, 0.30, INTERIOR );
    aV[x++][y][z] = new TVertex<3>(    1, 0.33, 0.33, boundary ); x = 0; y++;

    aV[x++][y][z] = new TVertex<3>(    0, 0.66, 0.33, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.37, 0.61, 0.30, INTERIOR );
    aV[x++][y][z] = new TVertex<3>( 0.68, 0.70, 0.36, INTERIOR );
    aV[x++][y][z] = new TVertex<3>(    1, 0.69, 0.33, boundary ); x = 0; y++;

    aV[x++][y][z] = new TVertex<3>(    0,    1, 0.36, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.33,    1, 0.30, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.69,    1, 0.33, boundary );
    aV[x++][y][z] = new TVertex<3>(    1,    1, 0.33, boundary ); x = 0; y = 0; z++; 



    aV[x++][y][z] = new TVertex<3>(    0,    0, 0.66, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.33,    0, 0.63, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.66,    0, 0.66, boundary );
    aV[x++][y][z] = new TVertex<3>(    1,    0, 0.69, boundary ); x = 0; y++;

    aV[x++][y][z] = new TVertex<3>(    0, 0.33, 0.66, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.37, 0.31, 0.63, INTERIOR );
    aV[x++][y][z] = new TVertex<3>( 0.64, 0.27, 0.63, INTERIOR );
    aV[x++][y][z] = new TVertex<3>(    1, 0.33, 0.66, boundary ); x = 0; y++;

    aV[x++][y][z] = new TVertex<3>(    0, 0.66, 0.66, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.37, 0.70, 0.69, INTERIOR );
    aV[x++][y][z] = new TVertex<3>( 0.68, 0.64, 0.62, INTERIOR );
    aV[x++][y][z] = new TVertex<3>(    1, 0.63, 0.66, boundary ); x = 0; y++;

    aV[x++][y][z] = new TVertex<3>(    0,    1, 0.63, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.36,    1, 0.66, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.68,    1, 0.62, boundary );
    aV[x++][y][z] = new TVertex<3>(    1,    1, 0.66, boundary ); x = 0; y = 0; z++; 



    aV[x++][y][z] = new TVertex<3>(    0,    0,    1, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.33,    0,    1, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.66,    0,    1, boundary );
    aV[x++][y][z] = new TVertex<3>(    1,    0,    1, boundary ); x = 0; y++;

    aV[x++][y][z] = new TVertex<3>(    0, 0.35,    1, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.36, 0.38,    1, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.63, 0.36,    1, boundary );
    aV[x++][y][z] = new TVertex<3>(    1, 0.30,    1, boundary ); x = 0; y++;

    aV[x++][y][z] = new TVertex<3>(    0, 0.66,    1, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.31, 0.69,    1, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.68, 0.72,    1, boundary );
    aV[x++][y][z] = new TVertex<3>(    1, 0.62,    1, boundary ); x = 0; y++;

    aV[x++][y][z] = new TVertex<3>(    0,    1,    1, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.36,    1,    1, boundary );
    aV[x++][y][z] = new TVertex<3>( 0.66,    1,    1, boundary );
    aV[x++][y][z] = new TVertex<3>(    1,    1,    1, boundary ); x = 0; y = 0; z++; 


    for ( z = 0; z < 4; z++ )
        for ( y = 0; y < 4; y++ )
            for ( x = 0; x < 4; x++ )
                aV[x][y][z]->delRef();

    for ( z = 0; z < 3; z++ )
        for ( y = 0; y < 3; y++ )
            for ( x = 0; x < 3; x++ )
                meshGen->addQuad( new T3DQuad( domain,
                                               aV[x][y][z], aV[x+1][y][z],
                                               aV[x+1][y+1][z], aV[x][y+1][z], 
                                               aV[x][y][z+1], aV[x+1][y][z+1],
                                               aV[x+1][y+1][z+1], aV[x][y+1][z+1] ) );
    
    if (param.draw)
        TDrawApp::drawMG( meshGen, "quad" );

    return meshGen->buildNet();
}

static void
write_sparse ( TMatrixBlock & A, const char * filename )
{
    FILE * file = fopen( filename, "w" );

    if ( file == NULL )
    {
        cout << "(write_sparse) could not open file " << filename << endl;
        return;
    }// if

    //
    // count no of entries
    //

    unsigned  entries = 0;
    
    for ( unsigned i = 0; i < A.getSize(); i++ )
    {
        uint  n = A[i].getNoOfEntries();

        for ( unsigned j = 0; j < n; j++ )
            if ( A[i].getValue(j) != 0.0 )
                entries++;
    }// for

    fprintf( file, " number of rows = %d\n", A.getSize() );
    fprintf( file, " number of columns = %d\n", A.getSize() );
    fprintf( file, " number of nonzero entries = %d\n", entries );

    //
    // write each row (nentries, col, coeff)
    //
    
    for ( unsigned i = 0; i < A.getSize(); i++ )
    {
        uint  n = A[i].getNoOfEntries();

        for ( unsigned j = 0; j < n; j++ )
            if ( A[i].getValue(j) != 0.0 )
            {
                fprintf( file, " %d\n", i+1 );
                fprintf( file, " %d\n", A[i].getIndex(j)+1 );
                fprintf( file, " %e\n", A[i].getValue(j) );
            }// if
    }// for
    
    fclose( file );
}

static void
write_nodes ( TNet<3> * net, const char * filename )
{
    if ( net == NULL )
        return;

    //
    // collect all nodes in grid corresponding to DoF
    //
    
    TDynArray< TVertex<3> * >  nodes( net->getNoOfInteriorNodes() );

    for ( unsigned i = 0; i < nodes.getSize(); i++ )
        nodes[i] = NULL;

    TListIterator< TCell<3> >  iterCell( net->getList() );
    TCell<3>                 * cell;
    
    iterCell.first();
    while (( cell = iterCell++ ) != NULL )
    {
        for ( uint i = 0; i < 4; i++ )
            if (( cell->getNode(i) != NULL ) &&
                ( cell->getNode(i)->getName() <= cell->getNode(i)->getMaxInteriorName() ))
                if ( nodes[ cell->getNode(i)->getName() ] == NULL )
                    nodes[ cell->getNode(i)->getName() ] = cell->getNode(i);
    }// while

    FILE * file = fopen( filename, "w" );

    if ( file == NULL )
    {
        cout << "(write_nodes) could not open file " << filename << endl;
        return;
    }// if

    fprintf( file, "dim = 3\n" );
    fprintf( file, "number of vertices = %d\n", nodes.getSize() );

    for ( unsigned i = 0; i < nodes.getSize(); i++ )
    {
        if ( nodes[i] == NULL )
        {
            cout << "node " << i << " is invalid (NULL)" << endl;
            continue;
        }// if

        fprintf( file, "%f %f %f\n", (*nodes[i])[0], (*nodes[i])[1], (*nodes[i])[2] );
    }// for

    fclose( file );
}
