//
//  Project   : libnet
//  File      : TConvDiffDoFManager.cc
//  Author    : Sabine Gutsch /tpr
//  Purpose   : class for DoF manager for piecewise linear
//              elements for the convection-diffusion equation
//
// arch-tag: 9dadf079-2df0-49f0-96c4-9e0cbd2e64ad
//

#include "TConvDiffDoFManager.hh"
#include "TDoFDesc.hh"
#include "math.hh"

// constructors 
template<int N>
TConvDiffDoFManager<N>::TConvDiffDoFManager () {
    noOfDoFs=0;
    noOfInnerDoFs=0; 
    noOfBoundaryDoFs=0;
}

// destructor
template<int N>
TConvDiffDoFManager<N>::~TConvDiffDoFManager () {}


template<int N>
void 
TConvDiffDoFManager<N>::initDoFDesc( TNet<N> *cNet, TNet<N> *fNet )
{
    typename TNet<N>::TCellListIterator iter( cNet->getList() );
        
    noOfDoFs         = 0;
    //noOfInnerDoFs    = 0; 
   
    for ( TCell<N> *cell = iter.first();   cell != NULL;   cell = iter.next() ){
        for (uint i = 0; i < N+1; i++){
            // check DoF 
            TVertex<N> *node = cell->getNode(i);
            TBoundary* boundary=node->getBoundary();
            if (boundary){
            
               
                // node is boundary node
                if (node->getDoFDesc() == NULL)
                    node->initDoFDesc(1);
              
                TBoundaryDesc* boundaryDesc=boundary->getBoundaryDesc();
                if(boundaryDesc){
                    // check whether to assign dof or not
                    if(boundaryDesc->getType(0)==NO_TYPE)
                        cerr<<"TConfDiffDoFManager<N>::InitDoFDesc: no Boundery Type for vertex "<<*node<<endl;
                    else {
                        if ( (boundaryDesc->isDoF(0)) && 
                             (node->getDoFDesc()->getName(0) == NOT_DEFINED)){
                            node->getDoFDesc()->setName(0, noOfBoundaryDoFs);
                            noOfBoundaryDoFs++;
                        }
                        else {
                            // don't init DoF
                            if (node->getDoFDesc()->getName( 0 ) == NOT_DEFINED)
                                node->getDoFDesc()->setName(0,NO_DOF);
                        }            
                    }
                }
                else cerr<<"TConfDiffDoFManager<N>::InitDoFDesc: no BounderyDesc for vertex"<<(*node)<<endl;
              
            } // if (boundary)
        } // for i
    } // for cell
    noOfDoFs=cNet->getNoOfInteriorNodes()+noOfBoundaryDoFs;
//     cout<<" inner Dofs: "<<cNet->getNoOfInteriorNodes();
//     cout<<" boundary nodes on   net : "<<cNet->getNoOfBoundaryNodes();
//     cout<<" boundary  Dofs: "<<noOfBoundaryDoFs<<endl;

} //  initDoFDesc( TNet<N> *cNet)
 
template<int N>
TBlockMatrix* 
TConvDiffDoFManager<N>::buildStiffnessMatrix( TNet<N> *cNet, TNet<N> *fNet ) 
{
    // to avoid compiler warning (fNet is not used)
    fNet = cNet;
    noOfInnerDoFs= cNet->getNoOfInteriorNodes();
    
    // create the blockmatrix
    // first the blockstructure of the matrix

        
    uint matrixSize[1][1] = { { noOfDoFs } };

    TBlockMatrix *pMat = new TBlockMatrix( 1, (uint*) matrixSize );

    uint  i, j;
    uint vertexName, neighbourName;
    uint  *noOfEntries = new uint[noOfDoFs];
        
    // 2^dim * # simplices in unit cube * constant
    const uint maxSize = uint(exp(N*log(2))*fac(N)*2);
    
    uint  newIndex[maxSize];
    TFloat newValue[maxSize];

    for ( i = 0; i < maxSize; i++ )
    {
        newValue[i] = 0;
        newIndex[i] = 0;
    }// for

    //
    // init matrixdata for every entry
    //
    for ( i = 0; i < noOfDoFs; i++ )
    {
        (*pMat)[0][0][i].set( maxSize, newIndex, newValue );
        (*pMat)[0][0][i].setIndex( 0, i );
        noOfEntries[i] = 1;
    }// for
    
    // build blockMatrix[0][0]
    TListIterator< TCell<N> >  iter( cNet->getList() );
    TCell<N>                 * cell;
    TVertex<N>               * node, * neighbourNode;

    while ((cell = iter++) != NULL)
    {
        for (i = 0; i < N+1; i++)
        {
            node = cell->getNode(i);
            if ((node->getBoundary()==NULL) ||
                (node->getBoundary()->getBoundaryDesc()->getType(0)==NEU_TYPE))
            {
                // node has DoF
                if (node->getBoundary()==NULL)
                    vertexName = node->getName();
                else
                    vertexName = node->getDoFDesc()->getName(0)+noOfInnerDoFs;

                //
                // build matrix data for vertexName 
                //
                
                // add neighbours
                for ( j = 1; j < N+1; j++ )
                {
                    // for each neighbour node  
                    neighbourNode = cell->getNode((i+j)%(N+1));
                    
                    if ((neighbourNode->getBoundary()==NULL) ||
                        (neighbourNode->getBoundary()->getBoundaryDesc()->getType(0)==NEU_TYPE))
                    {
                        // neighbourNode has DoF
                        if (neighbourNode->getBoundary()==NULL)
                            neighbourName = neighbourNode->getName();
                        else 
                            neighbourName = neighbourNode->getDoFDesc()->getName(0)+noOfInnerDoFs;
                        
                        if ((*pMat)[0][0][vertexName].hasIndex( neighbourName,
                                                                noOfEntries[vertexName] ) == FALSE)
                        {
                            (*pMat)[0][0][vertexName].setIndex( noOfEntries[vertexName],
                                                                       neighbourName );
                            noOfEntries[vertexName]++;
                        }// if
                    } // if neighbournode has DoF
                }             
            }// if node has DoF
        } // for each node in cell
    }// while

    // adjust sizes of matrixData
    for (i = 0; i < noOfDoFs; i++)
        (*pMat)[0][0][i].set( noOfEntries[i] );

    return pMat;
}


template class TConvDiffDoFManager<2>;
template class TConvDiffDoFManager<3>;
