//
//  Project    : libnet
//  File       : TStokesDoFManager.cc
//  Author     : Sabine Gutsch
//  Purpose    : DoFManager for linear conforming elements
//  Date       : 24.07.1998
//  Changes    : 
//  Questions  : 
//  ToDo       : 
//
// arch-tag: 6b4efe02-eff2-4c4a-b658-bc2909968a4b
//

#include "TStokesDoFManager.hh"
#include "TDoFDesc.hh"
#include "math.hh"
#include "T3DFrontier.hh"

// constructors	
template<int N>
TStokesDoFManager<N>::TStokesDoFManager () 
{
    for (uint i = 0; i < N+1; i++)
	{
        noOfBoundaryDoFs[i] = 0;
        noOfInnerDoFs[i] = 0;
	}
}
	
// destructor
template<int N>
TStokesDoFManager<N>::~TStokesDoFManager () {}

// initialize DoF description
template<int N>
void 
TStokesDoFManager<N>::initDoFDesc( TNet<N> *cNet, TNet<N> *fNet)
{  
  TVertex<N> *node;
  uint i, j;
  TBoundary* boundary;
  TBoundaryDesc* boundaryDesc;

  typename TNet<N>::TCellListIterator cIter( cNet->getList() );
        
  // initialize coarse grid DoFs
  for ( TCell<N> *cell = cIter.first(); cell != NULL; cell = cIter.next() )
  {
    for (i = 0; i < N+1; i++)
	{
      // check DoF 
      node = cell->getNode(i);
      boundary = node->getBoundary();
      if (boundary)
	  {
        // node is boundary node
        if (node->getDoFDesc() == NULL)
          node->initDoFDesc( N+1 );
            
        boundaryDesc = boundary->getBoundaryDesc();
        if (boundaryDesc) 
		  {
             // check whether to assign dof or not
             if(boundaryDesc->getType(N) == NO_TYPE)
                 cerr<<"TStokesDoFManager<N>::InitDoFDesc: no Boundery Type for vertex " << *node << endl;
             else 
			 {
                 if ( boundaryDesc->isDoF(N) && 
                      node->getDoFDesc()->getName( N ) == NOT_DEFINED )
				 {
                     node->getDoFDesc()->setName( N, noOfBoundaryDoFs[N]++ );
					 //cout << "set p DOF: " << node->getDoFDesc()->getName( N ) << *node << " " << noOfBoundaryDoFs[N] << endl;
				 }
                 else if (node->getDoFDesc()->getName( N ) == NOT_DEFINED)
                     // don't init DoF
                     node->getDoFDesc()->setName( N, NO_DOF );
			 }
		  }
        else cerr << "TStokesDoFManager<N>::InitDoFDesc: no BounderyDesc for vertex" << (*node) << endl;
      } // if (boundary)
    } // for i
  } // for cell

  // initialize fine grid DoFs
  typename TNet<N>::TCellListIterator fIter( fNet->getList() );
  for ( TCell<N> *cell = fIter.first(); cell != NULL; cell = fIter.next() )
  {
    for (i = 0; i < N+1; i++)
	{
      // check DoF 
      node = cell->getNode(i);
      boundary = node->getBoundary();
      if (boundary)
	  {
        // node is boundary node
        if (node->getDoFDesc() == NULL)
          node->initDoFDesc( N+1 );
            
        boundaryDesc = boundary->getBoundaryDesc();
        if (boundaryDesc) 
          for (j = 0; j < N; j++)
		  {
             // check whether to assign dof or not
             if(boundaryDesc->getType(j) == NO_TYPE)
                 cerr<<"TStokesDoFManager<N>::InitDoFDesc: no Boundery Type for vertex "<< *node << endl;
             else 
			 {
                 if( boundaryDesc->isDoF(j) && 
                      node->getDoFDesc()->getName( j ) == NOT_DEFINED)
				 {
                     node->getDoFDesc()->setName( j, noOfBoundaryDoFs[j]++ );
					 cout << "set u DOF: " << node->getDoFDesc()->getName( j ) << *node << endl;
				 }
                 else if (node->getDoFDesc()->getName( j ) == NOT_DEFINED)
                     // don't init DoF
                     node->getDoFDesc()->setName( j, NO_DOF );
			 }
		  }
        else cerr<<"TStokesDoFManager<N>::InitDoFDesc: no BounderyDesc for vertex" << (*node) << endl;
      } // if (boundary)
    } // for i
  } // for cell


  for (i = 0; i < N; i++)
  {
      noOfInnerDoFs[i] = fNet->getNoOfInteriorNodes();
      noOfDoFs[i] = noOfBoundaryDoFs[i] + noOfInnerDoFs[i];
  }
  noOfInnerDoFs[N] = cNet->getNoOfInteriorNodes();
  noOfDoFs[N] = noOfBoundaryDoFs[N] + noOfInnerDoFs[N];
}

//
// test if node has DoF
//
template<int N>
bool
TStokesDoFManager<N>::hasDoF( TVertex<N> *node, uint i )
{
   if ((node->getBoundary() == NULL) || 
       (node->getBoundary()->getBoundaryDesc()->getType(i) == NEU_TYPE))
       return( TRUE );
    return (FALSE);
}


//
// get DoF from DoF description of node
//
template<int N>
long 
TStokesDoFManager<N>::getDoFName( TVertex<N> *node, uint i )
{
    if ((node->getBoundary() == NULL) || 
        (node->getBoundary()->getBoundaryDesc()->getType(i) == NEU_TYPE))
        {  // node has DoF
            if (node->getBoundary() == NULL) return (node->getName());
            else return (node->getDoFDesc()->getName(i) + noOfInnerDoFs[i]);
		}
    return( -1 );
}

// build stiffness matrix of coarse net, using the fine net  
template<int N>
TBlockMatrix* 
TStokesDoFManager<N>::buildStiffnessMatrix( TNet<N> *cNet, TNet<N> *fNet ) 
{
    initDoFDesc( cNet, fNet );
	// create the blockmatrix
	// first the blockstructure of the matrix
	// the corresponding dof's for pressure unknowns
	uint pVertexDOF;
	uint pNeighbourDOF;
    TVertex<N> *node, *neighbourNode, *midNode;

	uint  i, j, k;
	uint matrixSize[N+1][N+1];
    for ( i = 0; i <= N; i++ ) 
        for ( j = 0; j <= N; j++ ) 
            matrixSize[i][j] = 0;

    for ( i = 0; i < N; i++ ) 
	{
        matrixSize[i][i] = noOfDoFs[i];
        matrixSize[i][N] = noOfDoFs[i];
        matrixSize[N][i] = noOfDoFs[N];
	}

	TBlockMatrix *blockMatrix = new TBlockMatrix( N+1, (uint*) matrixSize );

	uint vertexName, neighbourName, midNodeName;
    uint maxNoOfDoFs = 0;
    for ( i = 0; i < N; i++) maxNoOfDoFs = MAX( maxNoOfDoFs, noOfDoFs[i]);

    uint noOfEntries[N+1][N+1][maxNoOfDoFs];
    for (i = 0; i <= N; i++)
        for (j = 0; j <= N; j++)
            for (k = 0; k < maxNoOfDoFs; k++)
                noOfEntries[i][j][k] = 0;

    // 2^dim * # simplices in unit cube * constant;
	//const uint maxSize = uint(exp(N*log(2))*fac(N+1)*3); 
	uint maxSize = uint(exp(N*log(3))); 

	//uint  newIndex[maxSize];
	//TFloat newValue[maxSize];
	uint  *newIndex = new uint[maxSize];
	TFloat *newValue = new TFloat[maxSize];
    memset( newValue, 0, sizeof(TFloat)*maxSize );

    // first build blockMatrix[i][i]
	typename TNet<N>::TCellListIterator iter( fNet->getList() );
	for ( TCell<N> *cell = iter.first(); 
		  cell != NULL; 
		  cell = iter.next() )
		for (i = 0; i < N+1; i++) 
		{
            node = cell->getNode(i);

            for (j = 0; j < N; j++)
			{
                if ( hasDoF(node, j) )
				{
                    vertexName = getDoFName( node, j);
                    //if (vertexName == 26) cout << "26: " << *node << endl;

					// build matrix data for vertexName 
					// if it does not exist yet
				    if ( noOfEntries[j][j][vertexName] == 0 ) 
					{
					    (*blockMatrix)[j][j][vertexName].set(maxSize, newIndex, newValue);
					    (*blockMatrix)[j][N][vertexName].set(maxSize, newIndex, newValue);
						// set diagonal entry
					    (*blockMatrix)[j][j][vertexName].setIndex(0, vertexName);
					    noOfEntries[j][j][vertexName]++;
				    }// if

					// add neighbours
                    for ( k = 1; k < N+1; k++)
					{
                        // for each neighbour node  
                        neighbourNode = cell->getNode((i+k)%(N+1));
                        if ( hasDoF(neighbourNode, j) )
                        {   // neighbourNode has DoF
                            neighbourName = getDoFName( neighbourNode, j);                            //if (vertexName == 26) cout << "neighbour " << neighbourName << "  " << *neighbourNode << endl;

                            if ((*blockMatrix)[j][j][vertexName].hasIndex( neighbourName, noOfEntries[j][j][vertexName] ) == FALSE)
                                (*blockMatrix)[j][j][vertexName].setIndex( noOfEntries[j][j][vertexName]++, neighbourName );
						}// if neighbournode has DoF
					}//for           
				}// if node has DoF
			}// for each node in cell
		}//for i

	// adjust sizes of matrixData
    for ( j = 0; j < N; j++ )
	    for (i = 0; i < noOfDoFs[j]; i++)
		    (*blockMatrix)[j][j][i].set( noOfEntries[j][j][i] );

    delete[] newIndex;
    delete[] newValue;
	maxSize = uint(exp(N*log(2))*fac(N+1)*3); 
	newIndex = new uint[maxSize];
	newValue = new TFloat[maxSize];
    memset( newValue, 0, sizeof(TFloat)*maxSize );

	// build u-p coupling using the coarse net
	typename TNet<N>::TCellListIterator cIter( cNet->getList() );
	for ( TCell<N> *cell = cIter.first(); 
		  cell != NULL; 
		  cell = cIter.next() )
		for (i = 0; i < N+1; i++) 
		{
            node = cell->getNode(i);
            
			// get number for p DOF for vertex
            pVertexDOF = getDoFName( node, N);
                    
			// build matrix data for vertexName 
			// if it does not exist yet
            for (j = 0; j < N; j++)
			    if ( noOfEntries[N][j][pVertexDOF] == 0) 
				{
				    (*blockMatrix)[N][j][pVertexDOF].set(maxSize, newIndex, newValue);
                    if (hasDoF(node, j))
					{
					    // node has DoF
                        vertexName = getDoFName( node, j);
                        
				        // set diagonal entry
				        (*blockMatrix)[N][j][pVertexDOF].setIndex(0, vertexName);
				        noOfEntries[N][j][pVertexDOF]++;
				        (*blockMatrix)[j][N][vertexName].setIndex(0, pVertexDOF);
				        noOfEntries[j][N][vertexName]++;
					}//if
				}//if

			// add coarse neighbours
            for ( k = 1; k < N+1; k++)
			{
                // for each neighbour node  
                neighbourNode = cell->getNode((i+k)%(N+1));

                for ( j = 0; j < N; j++ )
				{
                    if (hasDoF(neighbourNode, j))
					{
					        // neighbourNode has DoF
                            neighbourName = getDoFName( neighbourNode, j);

				            if ( (*blockMatrix)[N][j][pVertexDOF].hasIndex( neighbourName, noOfEntries[N][j][pVertexDOF] ) == FALSE ) 
				            (*blockMatrix)[N][j][pVertexDOF].setIndex( noOfEntries[N][j][pVertexDOF]++, neighbourName );			
					}
			        // get number for p DOF of neighbour
                    pNeighbourDOF = getDoFName( neighbourNode, N);

                    if (hasDoF(node, j))
					{
					        // node has DoF
                            vertexName = getDoFName( node, j);

			                if ( (*blockMatrix)[j][N][vertexName].hasIndex( pNeighbourDOF, noOfEntries[j][N][vertexName] ) == FALSE ) 
				(*blockMatrix)[j][N][vertexName].setIndex( noOfEntries[j][N][vertexName]++, pNeighbourDOF );					
					}//if
				}//for
			}//for

			// add fine neighbours
            uint noOfMidNodes, m;
            if (N == 2) noOfMidNodes = 1; 
            else noOfMidNodes = 3;

	        for ( j = 0; j < N+1; j++ ) 
			    for ( m = 0; m < noOfMidNodes; m++) 
				{
                     if ( hasMidNode( cell, j, m ) )
					 {
                         midNode = getMidNode( cell, j, m );
                         for ( k = 0; k < N; k++ )
                             if (hasDoF( midNode, k ))
							 {
                                 midNodeName = getDoFName( midNode, k);
					             if ( (*blockMatrix)[k][N][midNodeName].hasIndex( pVertexDOF, noOfEntries[k][N][midNodeName] ) == FALSE ) 
								 {
						              (*blockMatrix)[k][N][midNodeName].setIndex( noOfEntries[k][N][midNodeName]++, pVertexDOF );					
						              (*blockMatrix)[N][k][pVertexDOF].setIndex( noOfEntries[N][k][pVertexDOF]++, midNodeName );
								 }
							 }
					 }
				}

		}//for i

	// adjust sizes of matrixData
    for ( j = 0; j < N; j++ )
	{
	    for (i = 0; i < noOfDoFs[j]; i++)
		    (*blockMatrix)[j][N][i].set( noOfEntries[j][N][i] );
	    for (i = 0; i < noOfDoFs[N]; i++)
		    (*blockMatrix)[N][j][i].set( noOfEntries[N][j][i] );
	}

    delete[] newIndex;
    delete[] newValue;

	return (blockMatrix);
}

//
// 2D: test if j'th frontier of 2d-cell has a midnode
//
bool
TStokesDoFManager<2>::hasMidNode(TCell<2> *cell, uint j, uint m)
{
    m = j; // to avoid compiler warning
    if (cell->getFrontier( j )->hasMidNode() )
        return (TRUE);
    return (FALSE);
}

//
// 2D: get midnode of j'th frontier of 2d-cell
//
TVertex<2>*
TStokesDoFManager<2>::getMidNode(TCell<2> *cell, uint j, uint m)
{
    m = j; // to avoid compiler warning
    TVertex<2>* midNode  = cell->getFrontier( j )->getNode( 1 );
    return (midNode);
}

//
// 3D: test if m'th edge of j'th frontier of 3d-cell has midnode
//
bool
TStokesDoFManager<3>::hasMidNode(TCell<3> *cell, uint j, uint m)
{
    T3DFrontier *pFrontier = (T3DFrontier*) cell->getFrontier( j );
    if (pFrontier->getEdge( m )->hasMidNode ())
        return (TRUE);
    return (FALSE);
}
        
//
// 3D: get midnode of m'th edge of j'th frontier of 3d-cell
//
TVertex<3>*
TStokesDoFManager<3>::getMidNode(TCell<3> *cell, uint j, uint m)
{
    T3DFrontier *pFrontier = (T3DFrontier*) cell->getFrontier( j );
	TVertex<3> *midNode = pFrontier->getEdge( m )->getNode( 1 );
    return (midNode);
}

template class TStokesDoFManager<2>;
template class TStokesDoFManager<3>;
