//
//  Project   : libnet
//  File      : stokesHierarchy.cc
//  Author    : Thomas Probst / Sabine Le Borne
//  Purpose   : class to build a matrix and prolongation hierarchy for
//              the Stokes equations
//
// arch-tag: 02d4e6df-6ef2-439e-a450-34cbedc58298
//

#include "TStokesHierarchy.hh"

template <int N>
TStokesHierarchy<N>::TStokesHierarchy ()
{ 
	pN       = NULL; 
    noOfNets = 0; 
}

template <int N>
TStokesHierarchy<N>::~TStokesHierarchy ()
{
	if (pN){
		for (uint i = 0; i < noOfNets-1; i++ ) {
			if (pN[i]) delete pN[i];
			if (pR[i]) delete pR[i];
		}
	}
}

template <int N>
void 
TStokesHierarchy<N>::buildHierarchy ( TBasisDoFManager<N> *dofManager,
                          TBasisEqn<N> *eqn,
                          TBasisFE<N> *fe,
                          TNetHierarchy<N> *netHier, 
                          TBlockMatrix **pMatrix, 
                          TBlockVector<TFloat> **pRHS, 
                          TBlockVector<TFloat> **pX )
{   
    TStokesBasisEqn<N> *stokesEqn = (TStokesBasisEqn<N>*) eqn;
    TDiscrType discrType = fe->getDiscrType();
    noOfNets = netHier->getNoOfNets();

    uint noOfDoFs[noOfNets][N+1];

    //uint blockSize[N+1];
    TDynArray< TVector<N> * > coordVec;
    
	T3DNetHierarchy::TNetListIterator iterNet( netHier->getList() );
   
    uint i, j;
	TNet<N> *ccNet  = iterNet.first();
	TNet<N> *cNet  = iterNet.next();
	TNet<N> *fNet  = iterNet.next();
   
    if (discrType != Null){
        pN = new TMatrixBlock * [noOfNets];
        pR = new TBlockVector<TFloat> * [noOfNets];
    }
	// allocation of a vector for
	// matrix blocks  pointers for the convective part;

    for (i = 1; i < noOfNets-1; i++) 
	{
	    pMatrix[i] = dofManager->buildStiffnessMatrix(cNet, fNet);
        for ( j = 0; j < N; j++)
			noOfDoFs[i][j] = (*(pMatrix[i]))[j][j].getSize();
        noOfDoFs[i][N] = (*(pMatrix[i]))[N][0].getSize();

		//cout << (*(pMatrix[i]))[0][0] << endl;
		/*
		cout << "level: " << i << endl;
        for ( j = 0; j <= N; j++)
			cout << noOfDoFs[i][j] << " ";
		cout << endl;
		*/

		pRHS[i] = new TBlockVector<TFloat>( N+1, noOfDoFs[i] );
        pRHS[i]->set(0.0);
		pX[i] = new TBlockVector<TFloat>( N+1, noOfDoFs[i] );
        pX[i]->set(0.0);
            
        // allocate place for the convective part:
        if (discrType != Null) {
            pN[i] = new TMatrixBlock( (*pMatrix[i])[0][0] );
            pR[i] = new TBlockVector<TFloat>( 0.0, *pRHS[i] );
		}
          
        fe->setInterior( cNet->getNoOfInteriorNodes(), 
                         fNet->getNoOfInteriorNodes() );
        fe->discrDiff( stokesEqn, cNet, pMatrix[i], pRHS[i]);
        // in the startup phase, pX[i] has zero components
        if (discrType != Null) {  
            //stokesEqn->setConvVector( pX[i] ); 
      
			// if a convection direction is defined, set pX
            if ( stokesEqn->getConvFunctions(0) != NULL ) 
			{
                //fe->getCoord( cNet, coordVec );
                //stokesEqn->initConvVector( noOfDoFs[i][0], coordVec );
			}
            fe->discrConv( stokesEqn, cNet, pN[i], pR[i]);
            for ( j = 0; j < N; j++)
			{
                (*(pMatrix[i]))[j][j] += *(pN[i]); // update matrix
                (*pRHS[i])[j] += (*pR[i])[j];
			}
		}
            // reaction term != zero
			/*
            if ( stokesEqn->getReacc() !=  noReac ) 
			{
			    TBlockVector<TFloat> *pM = new TBlockVector<TFloat>( 0.0, *pX[i]);
				*/
				// lumped mass matrix in vector pM
			/*
			    fe->discrMass(stokesEqn, cNet, &((*pM)[0])); 
			    uint j;
			    for (j = 0; j < blockSize[0]; j++) 
			    for ( k = 0; k < N; k++ )
				      (*pMatrix[i])[k][k][j][j] += (*pM)[0][j];
                delete pM;
            }
            */
            
        ccNet = cNet;
		cNet = fNet;
		fNet = iterNet.next();
	}// for

}


template <int N>
void 
TStokesHierarchy<N>::updateHierarchy ( 
                          TBasisEqn<N> *eqn,
                          TBasisFE<N> *fe,
                          TNetHierarchy<N> *netHier, 
                          TBlockMatrix **pMatrix, 
                          TBlockVector<TFloat> **pRHS, 
                          TBlockVector<TFloat> **pX )

{ 
    TStokesBasisEqn<N> *stokesEqn = (TStokesBasisEqn<N>*) eqn;
    TDiscrType discrType = fe->getDiscrType();
    noOfNets = netHier->getNoOfNets();

    uint noOfInteriorNodesNet[noOfNets-1];
    uint noOfBoundaryNodesNet[noOfNets-1];
	
    T3DNetHierarchy::TNetListIterator iterNet( netHier->getList() );
    
    uint i, j;
    TNet<N> *cNet  = iterNet.first();
    TNet<N> *fNet  = iterNet.next();
    
    noOfInteriorNodesNet[0] = cNet->getNoOfInteriorNodes();
    noOfBoundaryNodesNet[0] = cNet->getNoOfBoundaryNodes();

    for (i = 0; i < noOfNets-1; i++) 
	{
        noOfInteriorNodesNet[i+1] = fNet->getNoOfInteriorNodes();
        noOfBoundaryNodesNet[i+1] = fNet->getNoOfBoundaryNodes();
		if ( noOfInteriorNodesNet[i] != 0 ) 
		{
            // in the startup phase, pX[i] has zero components
            if (discrType != Null) {  

                //restore old matrix (so its only the diffusive part)
                for ( j = 0; j < N; j++ )
                    (*(pMatrix[i]))[j][j] -= *(pN[i]);
			    // calculate new convection
                (*pR[i]) *= 0;
                (*pN[i]) *= 0;
                stokesEqn->setConvVector( pX[i] );
                fe->discrConv(stokesEqn, cNet, pN[i], pR[i]);

                // update matrix
                for ( j = 0; j < N; j++ )
                    (*(pMatrix[i]))[j][j] += *(pN[i]); 
			    // update rhs
                for ( j = 0; j < N; j++ )
                    (*pRHS[i])[j] += (*pR[i])[j];

				// adjust LU decomposition on coarse grid
                if ( pMatrix[i]->hasLUdecomp() ) 
				{
                    pMatrix[i]->deleteLUdecomp();
                    pMatrix[i]->calcStokesLUdecomp1();
				}
            }
            

		}// else
		
		cNet = fNet;
		fNet = iterNet.next();
	}// for
}

template class TStokesHierarchy<3>;

