//
//  Project   : libmisc
//  File      : flowdef.cc
//  Author    : Ronald Kriemann
//  Purpose   : defintion of flow-examples
//
// arch-tag: d612e274-f9ec-471a-ac36-3056e9e7e0dc
//

#include "flowdef.hh"

/////////////////////////////////////////////////////
//
// forward decl. for used conv-functions
//

void evalNoConv     ( TVector<3> * pos, TVector<3> * b );
void evalXLine      ( TVector<3> * pos, TVector<3> * b );
void evalCurve      ( TVector<3> * pos, TVector<3> * b );
void evalCircle     ( TVector<3> * pos, TVector<3> * b );
void evalDiagCircle ( TVector<3> * pos, TVector<3> * b );
void eval4Circles   ( TVector<3> * pos, TVector<3> * b );
void evalBall       ( TVector<3> * pos, TVector<3> * b );
void evalVortex     ( TVector<3> * pos, TVector<3> * b );

/////////////////////////////////////////////////////
//
// base functions
//

//
// conv-diff eqn
//
TMyCdrEqn::TMyCdrEqn ( TFloat aEpsilon, 
                       TBlockVector<TFloat> *aConvVector )
        : TCdrEqn<3>( aEpsilon, aConvVector )
{}

void
TMyCdrEqn::evalDiff ( TVector <3> *, TMatrix<3> *D ) const
{
    (*D)[0][0]=epsilon; (*D)[0][1]=0;       (*D)[0][2]=0;
    (*D)[1][0]=0;       (*D)[1][1]=epsilon; (*D)[1][2]=0;
    (*D)[2][0]=0;       (*D)[2][1]=0;       (*D)[2][2]=epsilon;
}

TFloat
TMyCdrEqn::evalReac ( TVector <3> * ) const
{
    return 1.0;
}

TFloat
TMyCdrEqn::evalRhs ( TVector <3> *v, uint i ) const
{
    TVector<3>  b;

    evalConv( v, &b );
    return (-6 * epsilon + 2 * (b[0] * (*v)[0] + b[1] * (*v)[1] + b[2] * (*v)[2]) );
}

//
// stokes eqn
//
TMyStokesEqn::TMyStokesEqn ( TFloat eps,
                             TBlockVector<TFloat> * convVec)
        : TStokesBasisEqn<3>( eps, convVec )
{}

TFloat
TMyStokesEqn::evalRhs ( TVector<3> *v, uint i ) const
{
    if (v == NULL)
    {
        cout << "TMyStokesEqn::evalRhs (...) : got NULL arg" << endl;
        return 0;
    }// if
    
    TVector<3>  b;
    TFloat      value = 0.0;
    
    evalConv( v, & b );
    
    if (i == 0)
        value = 2.0 * ( epsilon - 1.0 ) * sin((*v)[0]) * sin((*v)[1]) 
                + b[0] * cos((*v)[0]) * sin((*v)[1])
                + b[1] * sin((*v)[0]) * cos((*v)[1]);
    
    if (i == 1)
        value =  2.0 * ( 1.0 + epsilon) * cos((*v)[0]) * cos((*v)[1]) 
                 - b[0] * sin((*v)[0]) * cos((*v)[1]) 
                 - b[1] * cos((*v)[0]) * sin((*v)[1]);
    
    if (i == 2)
        value =  0.0;

    return value;
}  

TFloat
TMyStokesEqn::evalReac ( TVector <3> * ) const
{
    return 0.0;
}
/////////////////////////////////////////////////////
//
// conv-diff eqn's
//

void
TNoConvCdrEqn::evalConv ( TVector<3> * pos, TVector<3> * b ) const
{
    evalNoConv( pos, b );
}

void
TXLineCdrEqn::evalConv ( TVector<3> * pos, TVector<3> * b ) const
{
    evalXLine( pos, b );
}

void
TCurveCdrEqn::evalConv ( TVector<3> * pos, TVector<3> * b ) const
{
    evalCurve( pos, b );
}

void
TCircleCdrEqn::evalConv ( TVector<3> * pos, TVector<3> * b ) const
{
    evalCircle( pos, b );
}

void
TDiagCircleCdrEqn::evalConv ( TVector<3> * pos, TVector<3> * b ) const
{
    evalDiagCircle( pos, b );
}

void
T4CirclesCdrEqn::evalConv ( TVector<3> * pos, TVector<3> * b ) const
{
    eval4Circles( pos, b );
}

void 
TBallCdrEqn::evalConv ( TVector<3> * pos, TVector<3> * b ) const
{
    evalBall( pos, b );
}

void 
TVortexCdrEqn::evalConv ( TVector<3> * pos, TVector<3> * b ) const
{
    evalVortex( pos, b );
}

/////////////////////////////////////////////////////
//
// stokes eqn's
//

void
TNoConvStokesEqn::evalConv ( TVector<3> * pos, TVector<3> * b ) const
{
    evalNoConv( pos, b );
}

void
TXLineStokesEqn::evalConv ( TVector<3> * pos, TVector<3> * b ) const
{
    evalXLine( pos, b );
}

void
TCurveStokesEqn::evalConv ( TVector<3> * pos, TVector<3> * b ) const
{
    evalCurve( pos, b );
}

void
TCircleStokesEqn::evalConv ( TVector<3> * pos, TVector<3> * b ) const
{
    evalCircle( pos, b );
}

void
TDiagCircleStokesEqn::evalConv ( TVector<3> * pos, TVector<3> * b ) const
{
    evalDiagCircle( pos, b );
}

void
T4CirclesStokesEqn::evalConv ( TVector<3> * pos, TVector<3> * b ) const
{
    eval4Circles( pos, b );
}

void 
TVortexStokesEqn::evalConv ( TVector<3> * pos, TVector<3> * b ) const
{
    evalVortex( pos, b );
}

/////////////////////////////////////////////////////
//
// definition of conv-functions
//

void
evalNoConv ( TVector<3> * pos, TVector<3> * b )
{
    if (b != NULL)
    {
        (*b)[X] = 0;
        (*b)[Y] = 0;
        (*b)[Z] = 0;
    }// if
    else
        cout << "evalNoConv (...) : got NULL arg" << endl;
}

void
evalXLine ( TVector<3> * pos, TVector<3> * b )
{
    if (b != NULL)
    {
        (*b)[X] = 1;
        (*b)[Y] = 0;
        (*b)[Z] = 0;
    }// if
    else
        cout << "evalXLine (...) : got NULL arg" << endl;
}

void
evalCurve ( TVector<3> * pos, TVector<3> * b )
{
    if ((pos != NULL) && (b != NULL))
    {
        (*b)[X] = 1.0 - (*pos)[Y];
        (*b)[Y] = (*pos)[X];
        (*b)[Z] = 0;
    }// if
    else
        cout << "evalCurve (...) : got NULL arg" << endl;
}

void
evalCircle ( TVector<3> * pos, TVector<3> * b )
{
    if ((b != NULL) && (pos != NULL))
    {
        (*b)[X] = 0.5 - (*pos)[Y];
        (*b)[Y] = (*pos)[X] - 0.5;
        (*b)[Z] = 0;
    }// if
    else
        cout << "evalCircle (...) : gave NULL arg" << endl;
}

void
evalDiagCircle ( TVector<3> * pos, TVector<3> * b )
{
    if ((pos == NULL) || (b == NULL))
    {
        cout << "evalDiagCircle (...) : got NULL args" << endl;
        return;
    }// if
    
    TFloat x = (*pos)[X];
    TFloat y = (*pos)[Y];
    TFloat z = (*pos)[Z];
    
    TFloat sqrt3 = sqrt(3);
    
    (*b)[X] = (z-y) / sqrt3;
    (*b)[Y] = (x-z) / sqrt3;
    (*b)[Z] = (y-x) / sqrt3;
}

void
eval4Circles ( TVector<3> * pos, TVector<3> * b )
{
    if ((b != NULL) && (pos != NULL))
    {
        TFloat x, y;

        x = 2 * (*pos)[X];
        y = 2 * (*pos)[Y];
            
        (*b)[Z] = 0;

        if (y < 1)
        {
            if (x < 1)
            {
                (*b)[X] = - (0.5 - y);
                (*b)[Y] = - (x - 0.5);
            }// if
            else
            {
                x -= 1;
                    
                (*b)[X] = 0.5 - y;
                (*b)[Y] = x - 0.5;
            }// else
        }// if
        else
        {
            y -= 1;
                
            if (x < 1)
            {
                (*b)[X] = 0.5 - y;
                (*b)[Y] = x - 0.5;
            }// if
            else
            {
                x -= 1;
                    
                (*b)[X] = - (0.5 - y);
                (*b)[Y] = - (x - 0.5);
            }// else
        }// else
    }// if
    else
        cout << "eval4Circles (...) : gave NULL arg" << endl;
}

void
evalBall ( TVector<3> * pos, TVector<3> * b )
{
    if ((b != NULL) && (pos != NULL))
    {
        TFloat x = (*pos)[0];
        TFloat y = (*pos)[1];
        TFloat z = (*pos)[2];
        
        (*b)[X] = x*(1-2*y)*(1-z);
        (*b)[Y] = y*(1-2*z)*(1-x);
        (*b)[Z] = z*(1-2*x)*(1-y);
    }// if
    else
        cout << "evalBall (...) : gave NULL arg" << endl;
}

void
evalVortex ( TVector<3> * pos, TVector<3> * b )
{
    if ((b != NULL) && (pos != NULL))
    {
        TFloat x = (*pos)[0];
        TFloat y = (*pos)[1];
        TFloat z = (*pos)[2];
        
        (*b)[X] = 2 * (x*(z - 0.5) + (y - 0.5));
        (*b)[Y] = 2 * ((0.5 - x) + y*(z - 0.5));
        (*b)[Z] = 2 * ((0.5 - x) + (0.5 - y) - z*z + z);
    }// if
    else
        cout << "evalVortex (...) : gave NULL arg" << endl;
}
