//
//  Project    : libnet
//  File       : stokesDoFManager.cc
//  Author     : Sabine Gutsch
//  Purpose    : DoFManager for linear conforming elements,
//               use Lagrange multipliers to make p unique
//
// arch-tag: 73551874-4df6-4f5b-b0e4-b1b7278e21b1
//

#include "stokesDoFManager.hh"

// constructors	
T2DStokesDoFManager::T2DStokesDoFManager () {}
	
// destructor
T2DStokesDoFManager::~T2DStokesDoFManager () {}

// build stiffness matrix of coarse net, using the fine net  
TBlockMatrix* 
T2DStokesDoFManager::buildStiffnessMatrix( TNet<2> *cNet, TNet<2> *fNet ) 
{
	// create the blockmatrix
	// first the blockstructure of the matrix
	uint noOfUDoFs = fNet->getNoOfInteriorNodes();
    // '+1' is for the Lagrange multipliers to enforce p in L_0	//
    // uint noOfPDoFs = cNet->getNoOfInteriorNodes() + cNet->getNoOfBoundaryNodes() + 1;  
	uint noOfPDoFs = cNet->getNoOfInteriorNodes() + cNet->getNoOfBoundaryNodes();  
	// vertexName < BOUNDARY_NAME ---> vertex is interiour
	// vertexName > BOUNDARY_NAME ---> vertex is on boundary
	uint BOUNDARY_NAME;
	BOUNDARY_NAME = UINT_MAX - fNet->getNoOfBoundaryNodes();

		// the corresponding dof's for boundary vertices
	uint pVertexDOF;
	uint pNeighbourDOF;

	uint matrixSize[3][3] = {
		{ noOfUDoFs,         0, noOfUDoFs },
		{         0, noOfUDoFs, noOfUDoFs },
		{ noOfPDoFs, noOfPDoFs, noOfPDoFs } };

	TBlockMatrix *blockMatrix = new TBlockMatrix( 3, (uint*) matrixSize );

	uint  i,j;
	uint vertexName, neighbourName, midNodeName;
	uint  *noOfUUEntries = new uint[noOfUDoFs];
	uint  *noOfUPEntries = new uint[noOfUDoFs];
	uint  *noOfPUEntries = new uint[noOfPDoFs];
	memset( noOfUUEntries, 0, sizeof(uint)*noOfUDoFs );
	memset( noOfUPEntries, 0, sizeof(uint)*noOfUDoFs );
	memset( noOfPUEntries, 0, sizeof(uint)*noOfPDoFs );
        
	const uint maxSize = 32;
	uint  newIndex[maxSize];
	TFloat newValue[maxSize];
	memset( newValue, 0, sizeof(TFloat)*maxSize );

        // first build blockMatrix[0][0]
	T2DNet::TCellListIterator iter( fNet->getList() );
	for ( TCell<2> *cell = iter.first(); 
		  cell != NULL; 
		  cell = iter.next() )
		for (i = 0; i < 3; i++)
		{
			vertexName = cell->getNode(i)->getName();
			if (vertexName < BOUNDARY_NAME)
			{
				// build matrix data for vertexName 
				// if it does not exist yet
				if ( noOfUUEntries[vertexName] == 0 ) 
				{
					(*blockMatrix)[0][0][vertexName].set(maxSize, newIndex, newValue);
					(*blockMatrix)[0][2][vertexName].set(maxSize, newIndex, newValue);
					// set diagonal entry
					(*blockMatrix)[0][0][vertexName].setIndex(0, vertexName);
					noOfUUEntries[vertexName]++;
				}// if
				// add neighbours
				neighbourName = cell->getNode((i+1)%3)->getName();
				if ( (neighbourName < BOUNDARY_NAME) &&
					 ((*blockMatrix)[0][0][vertexName].hasIndex( neighbourName, noOfUUEntries[vertexName] ) == FALSE)) 
					(*blockMatrix)[0][0][vertexName].setIndex( noOfUUEntries[vertexName]++, neighbourName );
				
				neighbourName = cell->getNode((i+2)%3)->getName();
				if ( (neighbourName < BOUNDARY_NAME) &&
					 ((*blockMatrix)[0][0][vertexName].hasIndex( neighbourName, noOfUUEntries[vertexName] ) == FALSE)) 
					(*blockMatrix)[0][0][vertexName].setIndex( noOfUUEntries[vertexName]++, neighbourName );
			}// if
		}// for

	// adjust sizes of matrixData
	for (i = 0; i < noOfUDoFs; i++)
		(*blockMatrix)[0][0][i].set( noOfUUEntries[i] );

	// copy 0x0 block to 1x1 block
	(*blockMatrix)[1][1] = (*blockMatrix)[0][0];

	// next build u-p coupling using the coarse net
	T2DNet::TCellListIterator cIter( cNet->getList() );
	for ( TCell<2> *cell = cIter.first(); 
		  cell != NULL; 
		  cell = cIter.next() )
		for (i = 0; i < 3; i++)
		{
			vertexName = cell->getNode(i)->getName();

			// get number for p DOF for vertex
			if (vertexName > BOUNDARY_NAME) 
				pVertexDOF = UINT_MAX - vertexName + cNet->getNoOfInteriorNodes();
			else
				pVertexDOF = vertexName;

			// build matrix data for vertexName 
			// if it does not exist yet
			if ( noOfPUEntries[pVertexDOF] == 0) 
			{
				(*blockMatrix)[2][0][pVertexDOF].set(maxSize, newIndex, newValue);
				if (vertexName < BOUNDARY_NAME) 
				{
					// set diagonal entry
					(*blockMatrix)[2][0][vertexName].setIndex(0, vertexName);
					noOfPUEntries[vertexName]++;
					(*blockMatrix)[0][2][vertexName].setIndex(0, vertexName);
					noOfUPEntries[vertexName]++;
				}// if
			}// if

			// add coarse neighbours
			neighbourName = cell->getNode((i+1)%3)->getName();
			// get number for p DOF of neighbour
			if (neighbourName > BOUNDARY_NAME) 
				pNeighbourDOF = UINT_MAX - neighbourName + cNet->getNoOfInteriorNodes();
			else
				pNeighbourDOF = neighbourName;

			if ( (vertexName < BOUNDARY_NAME) &&
				 ((*blockMatrix)[0][2][vertexName].hasIndex( pNeighbourDOF, noOfUPEntries[vertexName] ) == FALSE)) 
				(*blockMatrix)[0][2][vertexName].setIndex( noOfUPEntries[vertexName]++, pNeighbourDOF );

			if ( (neighbourName < BOUNDARY_NAME) &&
				 ((*blockMatrix)[2][0][pVertexDOF].hasIndex( neighbourName, noOfPUEntries[pVertexDOF] ) == FALSE)) 
				(*blockMatrix)[2][0][pVertexDOF].setIndex( noOfPUEntries[pVertexDOF]++, neighbourName );
					
			neighbourName = cell->getNode((i+2)%3)->getName();
			// get number for p DOF of neighbour
			if (neighbourName > BOUNDARY_NAME) 
				pNeighbourDOF = UINT_MAX - neighbourName + cNet->getNoOfInteriorNodes();
			else
				pNeighbourDOF = neighbourName;

			if ( (vertexName < BOUNDARY_NAME) &&
				 ((*blockMatrix)[0][2][vertexName].hasIndex( pNeighbourDOF, noOfUPEntries[vertexName] ) == FALSE)) 
				(*blockMatrix)[0][2][vertexName].setIndex( noOfUPEntries[vertexName]++, pNeighbourDOF );	
		
			if ( (neighbourName < BOUNDARY_NAME) &&
				 ((*blockMatrix)[2][0][pVertexDOF].hasIndex( neighbourName, noOfPUEntries[pVertexDOF] ) == FALSE)) 
				(*blockMatrix)[2][0][pVertexDOF].setIndex( noOfPUEntries[pVertexDOF]++, neighbourName );
					
			// add fine neighbours
			for ( j = 0; j < 3; j++ ) 
			{
				if (cell->getFrontier( j )->hasMidNode() ) 
				{
					midNodeName = cell->getFrontier( j )->getNode( 1 )->getName();
					if ( (midNodeName < BOUNDARY_NAME) &&
						 ((*blockMatrix)[0][2][midNodeName].hasIndex( pVertexDOF, noOfUPEntries[midNodeName] ) == FALSE)) 
					{
						(*blockMatrix)[0][2][midNodeName].setIndex( noOfUPEntries[midNodeName]++, pVertexDOF );					
						(*blockMatrix)[2][0][pVertexDOF].setIndex( noOfPUEntries[pVertexDOF]++, midNodeName );
					}// if
				}// if
			}// for
		}// for

	// adjust sizes of matrixData
	for (i = 0; i < noOfUDoFs; i++)
		(*blockMatrix)[0][2][i].set( noOfUPEntries[i] );
	for (i = 0; i < noOfPDoFs; i++) 
		(*blockMatrix)[2][0][i].set( noOfPUEntries[i] );

	// copy 0x2 block to 1x2 block and copy 2x0 block to 2x1 block
	(*blockMatrix)[1][2] = (*blockMatrix)[0][2];
	(*blockMatrix)[2][1] = (*blockMatrix)[2][0];

    
	// now provide for entries caused by Lagrange multipliers
	/*
    (*blockMatrix)[2][2][noOfPDoFs-1].set( noOfPDoFs - 1 );
    for (i = 0; i < noOfPDoFs-1; i++) {
        newIndex[0] = noOfPDoFs-1; newValue[0] = 0.0;
        (*blockMatrix)[2][2][i].set( 1, newIndex, newValue );
		(*blockMatrix)[2][2][noOfPDoFs-1].setIndex( i, i);
	}
	*/
	// now provide for entrie to make system non-singular
    (*blockMatrix)[2][2][noOfPDoFs-1].set( 1 );
    newIndex[0] = noOfPDoFs-1; newValue[0] = 0.0;
    (*blockMatrix)[2][2][noOfPDoFs-1].set( 1, newIndex, newValue );
		
	delete[] noOfUUEntries;
	delete[] noOfUPEntries;
	delete[] noOfPUEntries;

	return (blockMatrix);
}
