//
//  Project     : libgraph
//  File        : T2DPlanarGraph.cc
//  Author      : Ronald Kriemann
//  Purpose     : class for a planar graph in 2D
//
// arch-tag: 37129a5a-72ef-41b3-9f5f-47477b085cdb
//

#include "dynarray.hh"

#include "vector.hh"
#include "matrixblock.hh"

#include "T2DPlanarNode.hh"
#include "criterion.hh"

#include "T2DPlanarGraph.hh"

//////////////////////////////////////////////////////////////////
//
//  T2DPlanarGraph
//

//
// constructor
//
T2DPlanarGraph::T2DPlanarGraph ()
    : TGraph ()
{
    type = _DIRECTED_ | _PLANAR_;
}

//
// constructor (builds graph of a matrixblock)
//
T2DPlanarGraph::T2DPlanarGraph ( TDynArray< TVector<2> * >& arrVectors,
                                 TMatrixBlock& mat, 
                                 TFloat xi,
                                 uint criterion, bool bZeroWeight )
        : TGraph ()
{
    uint    i;
    uint  *arrIndex;
    TFloat *arrValues;
    TFloat  diagValue;
    uint    noOfEntries;

    T2DPlanarNode  *pNode;
    uint                  nodeName;
    TMarkableEdge         *pEdge;

    // matrixblock we use
    uint  noOfRows = mat.getNoOfRows();

    // Array to store pointers to nodes (for direct access)
    TDynArray< T2DPlanarNode *>  arrNodes( noOfRows, 0 );

    // first build all nodes (we assume, that index-pos equals vertexname)
    for (i = 0; i < noOfRows; i++)
    {
        pNode = new T2DPlanarNode( i, NO_MARK, arrVectors[i] );
        pNode->setWeight( fabs(mat[i][i]) );
        arrNodes[i] = pNode;
        add( *pNode );
    }// for

    // then build the successors and predecessors of every node
    TListIterator<TNode> iter( listNodes );
    while ((pNode = (T2DPlanarNode*) iter++) != NULL)
    {
        nodeName = pNode->getName();

        arrIndex    = mat[nodeName].getIndex();
        arrValues   = mat[nodeName].getValue();
        diagValue   = mat[nodeName][nodeName];

        noOfEntries = mat[nodeName].getNoOfEntries();

        if (criterion == E3)
        {
            //
            // add a succ if the value in the matrix is bigger than
            // the diagvalue with respect to xi (value > xi*diag) 
            //
            for (i = 0; i < noOfEntries; i++)
            {
                if (arrIndex[i] != nodeName)
                {
                    if (fabs(arrValues[i]) >= fabs(xi * diagValue))
                    {
                        pEdge = new TWeightedMarkableEdge( *pNode, 
                                                           *(arrNodes[arrIndex[i]]),
                                                           NO_MARK, 
                                                           fabs(arrValues[i]) );
                        pNode->addOutEdge( *pEdge );
                        arrNodes[arrIndex[i]]->addInEdge( *pEdge );
                    }// if
                    else if (bZeroWeight)
                    {
                        pEdge = new TWeightedMarkableEdge( *pNode, 
                                                           *(arrNodes[arrIndex[i]]),
                                                           NO_MARK, 0 );
                        pNode->addOutEdge( *pEdge );
                        arrNodes[arrIndex[i]]->addInEdge( *pEdge );
                    }// else if
                }// if
            }// for
        }// if
        else if (criterion == E4)
        {
            //
            // add only the strongest entry with respect to xi
            //
            TFloat max = -1.0;

            // look for this entry
            for (i = 0; i < noOfEntries; i++)
            {
                if (arrIndex[i] != nodeName)
                {
                    if (fabs(arrValues[i]) > max)
                        max = fabs(arrValues[i]);
                }// if
            }// for

            // now add all entries with maxValue
            for (i = 1; i < noOfEntries; i++)
            {
                if (arrIndex[i] != nodeName)
                {
                    if (fabs(arrValues[i]) >= fabs(xi * max))
                    {
                        pEdge = new TWeightedMarkableEdge( *pNode, 
                                                           *(arrNodes[arrIndex[i]]),
                                                           NO_MARK, 
                                                           fabs(arrValues[i]) );
                        pNode->addOutEdge( *pEdge );
                        arrNodes[arrIndex[i]]->addInEdge( *pEdge );
                    }// if
                    else if (bZeroWeight)
                    {
                        pEdge = new TWeightedMarkableEdge( *pNode, 
                                                           *(arrNodes[arrIndex[i]]),
                                                           NO_MARK, 0 );
                        pNode->addOutEdge( *pEdge );
                        arrNodes[arrIndex[i]]->addInEdge( *pEdge );
                    }// else if
                }// if
            }// for
        }// else
        else if (criterion == E5)
        {
            //
            // add only the strongest entry
            //
            TFloat max = -1.0;

            // look for this entry
            for (i = 0; i < noOfEntries; i++)
            {
                if (arrIndex[i] != nodeName)
                {
                    if (fabs(arrValues[i]) > max)
                        max = fabs(arrValues[i]);
                }// if
            }// for

            // now add all entries with maxValue
            for (i = 1; i < noOfEntries; i++)
            {
                if (arrIndex[i] != nodeName)
                {
                    if (fabs(arrValues[i]) == max)
                    {
                        pEdge = new TWeightedMarkableEdge( *pNode, 
                                                           *(arrNodes[arrIndex[i]]),
                                                           NO_MARK, 
                                                           fabs(arrValues[i]) );
                        pNode->addOutEdge( *pEdge );
                        arrNodes[arrIndex[i]]->addInEdge( *pEdge );
                    }// if
                    else if (bZeroWeight)
                    {
                        pEdge = new TWeightedMarkableEdge( *pNode, 
                                                           *(arrNodes[arrIndex[i]]),
                                                           NO_MARK, 0 );
                        pNode->addOutEdge( *pEdge );
                        arrNodes[arrIndex[i]]->addInEdge( *pEdge );
                    }// else if
                }// if
            }// for
        }// else
        else if (criterion == E7)
        {
            //
            // add a succ if a_ij is bigger than a_ji with respect to xi
            //

            // add all desired entries
            for (i = 0; i < noOfEntries; i++)
            {
                if (arrIndex[i] != nodeName)
                {
                    if (fabs(arrValues[i]) > fabs(xi * mat[arrIndex[i]][nodeName]))
                    {
                        pEdge = new TWeightedMarkableEdge( *pNode, 
                                                           *(arrNodes[arrIndex[i]]),
                                                           NO_MARK, 
                                                           fabs(arrValues[i]) );
                        pNode->addOutEdge( *pEdge );
                        arrNodes[arrIndex[i]]->addInEdge( *pEdge );
                    }// if
                    else if (bZeroWeight)
                    {
                        pEdge = new TWeightedMarkableEdge( *pNode, 
                                                           *(arrNodes[arrIndex[i]]),
                                                           NO_MARK, 0 );
                        pNode->addOutEdge( *pEdge );
                        arrNodes[arrIndex[i]]->addInEdge( *pEdge );
                    }// else if
                }// if
            }// for
        }// else
        else if (criterion == E8)
        {//
            // add  all weak edges entry
            //
            TFloat max = -1.0;
            // cout << nodeName<<"  "<<mat[nodeName]<<endl;
            // look for this entry
            for (i = 0; i < noOfEntries; i++)
            {
                if (arrIndex[i] != nodeName)
                {
                    if (fabs(arrValues[i]) > max)
                        max = fabs(arrValues[i]);
                }// if
            }// for

            // now add all entries which are smaller than  maxValue
            TFloat eps=1e-8;
      
            for (i = 1; i < noOfEntries; i++)
            {
                if (arrIndex[i] != nodeName)
                {
                    if ( (fabs(arrValues[i]) < max) && (fabs(arrValues[i]) > eps))
                    {
                        pEdge = new TWeightedMarkableEdge( *pNode, 
                                                           *(arrNodes[arrIndex[i]]),
                                                           NO_MARK, 
                                                           fabs(arrValues[i]) );
                        pNode->addOutEdge( *pEdge );
                        arrNodes[arrIndex[i]]->addInEdge( *pEdge );
                    }// if
                    else if (bZeroWeight)
                    {
                        pEdge = new TWeightedMarkableEdge( *pNode, 
                                                           *(arrNodes[arrIndex[i]]),
                                                           NO_MARK, 0 );
                        pNode->addOutEdge( *pEdge );
                        arrNodes[arrIndex[i]]->addInEdge( *pEdge );
                    }// else if
                }// if
            }// for
            
        } // else
        else
            cerr << "T2DPlanarGraph::T2DPlanarGraph (...) : "
                 << "missing second matrix or wrong criterion"
                 << endl;
    }// for
   
    type = _DIRECTED_ | _PLANAR_;
}

T2DPlanarGraph::T2DPlanarGraph ( TDynArray< TVector<2> * > & arrVectors,
                                 TDynArray< TVector<2> * > & convVectors,
                                 TMatrixBlock& mat, 
                                 TFloat xi,
                                 uint criterion,
                                 bool bZeroWeight )
        : TGraph ()
{
    uint    i;
    uint  *arrIndex;
    TFloat *arrValues;
    TFloat  diagValue;
    uint    noOfEntries;

    T2DPlanarNode  *pNode;
    uint                  nodeName;
    TMarkableEdge         *pEdge;

    // matrixblock we use
    uint  noOfRows = mat.getNoOfRows();

    // Array to store pointers to nodes (for direct access)
    TDynArray< T2DPlanarNode * >  arrNodes( noOfRows );

    // first build all nodes (we assume, that index-pos equals vertexname)
    for (i = 0; i < noOfRows; i++)
    {
        pNode = new T2DPlanarNode( i, NO_MARK, arrVectors[i] );
        pNode->setWeight( fabs(mat[i][i]) );
        arrNodes[i] = pNode;
        add( *pNode );
    }// for
    
    TListIterator<TNode> iter( listNodes );
    
    while ((pNode = (T2DPlanarNode*) iter++) != NULL)
    {
        nodeName = pNode->getName();
      
        arrIndex    = mat[nodeName].getIndex();
        arrValues   = mat[nodeName].getValue();
        diagValue   = mat[nodeName][nodeName];

        noOfEntries = mat[nodeName].getNoOfEntries();

        if (criterion == E8){
            //
            // add  all weak edges entry
            //
            TFloat max = -1.0;
            // cout << nodeName<<"  "<<mat[nodeName]<<endl;
            // look for this entry
            for (i = 0; i < noOfEntries; i++)
            {
                if (arrIndex[i] != nodeName)
                {
                    if (fabs(arrValues[i]) > max)
                        max = fabs(arrValues[i]);
                }// if
            }// for

            // now add all entries which are smaller than  maxValue
            TFloat delta=xi;
            for (i = 1; i < noOfEntries; i++)
            {
                if (arrIndex[i] != nodeName)
                {
                    TVector<2> a,b;
                    a= (*arrVectors[arrIndex[i]]) - (*arrVectors[nodeName]);
                    // b= ( *convVectors[nodeName] + *convVectors[arrIndex[i]]);
                     b= ( *convVectors[nodeName] );
                  
                    TFloat s = fabs(a*b)/sqrt((a*a)*(b*b));
                    //   if(nodeName==5){
                    //     cout<<" Node: "<<(*arrVectors[nodeName]);
                    //    cout<< "neighbour: "<<arrIndex[i]<<" "<<(*arrVectors[arrIndex[i]]);
                    //    cout<<"s: "<<s<<" b: "<<b<<endl;
                    // }
                    if (s < delta)
                    {
                        pEdge = new TWeightedMarkableEdge( *pNode, 
                                                           *(arrNodes[arrIndex[i]]),
                                                           NO_MARK, 
                                                           fabs(arrValues[i]) );
                        pNode->addOutEdge( *pEdge );
                        arrNodes[arrIndex[i]]->addInEdge( *pEdge );
                        //   if(nodeName==5)
                        //    cout<<" egde to : "<<arrIndex[i]<<endl;
                      }
                }// if
            }// for
            
        } // else
        else
            cerr << "T2DPlanarGraph::T2DPlanarGraph (...) : "
                 << "missing second matrix or wrong criterion" << endl;
    }// for
   
    type = _DIRECTED_ | _PLANAR_;
}
    
//
// constructor (builds graph out of two matrixblocks)
//
T2DPlanarGraph::T2DPlanarGraph ( TDynArray< TVector<2> * > & arrVectors,
                                 TMatrixBlock& diffMat, 
                                 TMatrixBlock& convMat,
                                 TFloat xi,
                                 uint criterion,
                                 bool bZeroWeight )
        : TGraph ()
{
    uint    i;
    uint  *arrIndex;
    TFloat *arrConvValues, *arrDiffValues;
    uint    noOfEntries;

    T2DPlanarNode  *pNode;
    uint                  nodeName;
    TMarkableEdge         *pEdge;

    // matrixblock we use
    uint  noOfRows = diffMat.getNoOfRows();

    // Array to store pointers to nodes (for direct access)
    TDynArray< T2DPlanarNode * >  arrNodes( noOfRows );

    // first build all nodes (we assume, that index-pos equals vertexname)
    for (i = 0; i < noOfRows; i++)
    {
        pNode = new T2DPlanarNode( i, NO_MARK, arrVectors[i] );
        pNode->setWeight( fabs(diffMat[i][i] + convMat[i][i]) );
        arrNodes[i] = pNode;
        add( *pNode );
    }// for

    // then build the successors and predecessors of every node
    TListIterator<TNode> iter( listNodes );
    while ((pNode = (T2DPlanarNode*) iter++) != NULL)
    {
        nodeName = pNode->getName();

        arrIndex      = diffMat[nodeName].getIndex();
        arrDiffValues = diffMat[nodeName].getValue();
        arrConvValues = convMat[nodeName].getValue();

        noOfEntries = diffMat[nodeName].getNoOfEntries();

        if (criterion == E0)
        {
            // add all entries, where the conv part is stronger than
            // the diff part (with respect to a parameter xi)
            for (i = 0; i < noOfEntries; i++)
            {
                if (arrIndex[i] != nodeName)
                {
                    if (fabs(arrConvValues[i]) > fabs(xi * arrDiffValues[i]))
                    {
                        pEdge = new TWeightedMarkableEdge( *pNode, 
                                                           *(arrNodes[arrIndex[i]]) );
                        pNode->addOutEdge( *pEdge );
                        arrNodes[arrIndex[i]]->addInEdge( *pEdge );
                    }// if
                    else if (bZeroWeight)
                    {
                        pEdge = new TWeightedMarkableEdge( *pNode, 
                                                           *(arrNodes[arrIndex[i]]),
                                                           NO_MARK, 0 );
                        pNode->addOutEdge( *pEdge );
                        arrNodes[arrIndex[i]]->addInEdge( *pEdge );
                    }// else if
                }// if
            }// for
        }// if
        else if (criterion == E1)
        {
            // add all entries with at least half of the max conv part

            // look for the max
            TFloat max = -1.0;
            for (i = 0; i < noOfEntries; i++)
            {
                if (arrIndex[i] != nodeName)
                {
                    if (fabs(arrConvValues[i]) > fabs(xi * arrDiffValues[i]))
                    {
                        max = fabs(arrConvValues[i]);
                    }// if
                }// if
            }// for

            // add all entries
            for (i = 0; i < noOfEntries; i++)
            {
                if (arrIndex[i] != nodeName)
                {
                    if (fabs(arrConvValues[i]) >= (max / 2.0))
                    {
                        pEdge = new TWeightedMarkableEdge( *pNode, 
                                                           *(arrNodes[arrIndex[i]]) );
                        pNode->addOutEdge( *pEdge );
                        arrNodes[arrIndex[i]]->addInEdge( *pEdge );
                    }// if
                    else if (bZeroWeight)
                    {
                        pEdge = new TWeightedMarkableEdge( *pNode, 
                                                           *(arrNodes[arrIndex[i]]),
                                                           NO_MARK, 0 );
                        pNode->addOutEdge( *pEdge );
                        arrNodes[arrIndex[i]]->addInEdge( *pEdge );
                    }// else if
                }// if
            }// for
        }// else
        else if (criterion == E2)
        {
            // add all entries with max conv part

            // look for the max
            TFloat max = -1.0;
            for (i = 0; i < noOfEntries; i++)
            {
                if (arrIndex[i] != nodeName)
                {
                    if (fabs(arrConvValues[i]) > fabs(xi * arrDiffValues[i]))
                    {
                        max = fabs(arrConvValues[i]);
                    }// if
                }// if
            }// for

            // add all entries
            for (i = 0; i < noOfEntries; i++)
            {
                if (arrIndex[i] != nodeName)
                {
                    if (fabs(arrConvValues[i]) == max)
                    {
                        pEdge = new TWeightedMarkableEdge( *pNode, 
                                                           *(arrNodes[arrIndex[i]]) );
                        pNode->addOutEdge( *pEdge );
                        arrNodes[arrIndex[i]]->addInEdge( *pEdge );
                    }// if
                    else if (bZeroWeight)
                    {
                        pEdge = new TWeightedMarkableEdge( *pNode, 
                                                           *(arrNodes[arrIndex[i]]),
                                                           NO_MARK, 0 );
                        pNode->addOutEdge( *pEdge );
                        arrNodes[arrIndex[i]]->addInEdge( *pEdge );
                    }// else if
                }// if
            }// for
        }// else
        else
            cerr << "T2DPlanarGraph::T2DPlanarGraph (...) : wrong criterion" << endl;
    }// for

    type = _DIRECTED_ | _PLANAR_;
}

//
// apply the one-flow-direction condition to the graph
//
void 
T2DPlanarGraph::sortOFD ()
{
    uint  i,j;
    
    TListIterator<TNode>  iterNodes( listNodes );
    T2DPlanarNode        *pNode, *pAdjNode;
    TListIterator<TEdge>  iterEdges;
    TEdge                *pEdge, *pInEdge, *pOutEdge;
    
    TVector<2>            vecEdge, vecNormEdge;
    TVector<2>            vecNode;
    TVector<2>            vecInEdge, vecOutEdge;
    
    // these arrays are for direct accessing the in- and out-edges
    // of every node and for the angles the have to the pos. x-axis
    TDynArray< TEdge *>   arrOutEdges( 5, 5 ), arrInEdges( 5, 5 );
    TDynArray< TFloat >   arrOutAngles( 5, 5 ), arrInAngles( 5, 5 );
    TDynArray< uint >    arrOrder( 5, 5 );
    uint                  noOfInEdges, noOfOutEdges;
    
    TList<TEdge>        * pInEdges, * pOutEdges;
    
    TFloat                alpha, alpha2;
    
    iterNodes.first();
    while ((pNode = (T2DPlanarNode*) iterNodes++) != NULL)
    {
        vecNode = * pNode->vector();
        
        //
        // order out-edges
        //
        if ((pNode->getInDegree() > 0) && (pNode->getOutDegree() > 1))
        {
            noOfInEdges  = pNode->getInDegree();
            noOfOutEdges = pNode->getOutDegree();
            
            arrOutEdges.setSize( noOfOutEdges );
            arrOutAngles.setSize( noOfOutEdges );
            arrOrder.setSize( noOfOutEdges );
            
            //
            // go over all out-edges and compute the angle between them
            // and the first in-edges
            //
            
            pInEdge    = pNode->getInEdges().getFirst();
            vecInEdge  = *(((T2DPlanarNode*) (pInEdge->getTail()))->vector());
            vecInEdge -= vecNode;
            
            i = 0;
            iterEdges.setList( pNode->getOutEdges() );
            iterEdges.first();
            while ((pEdge = iterEdges++) != NULL)
            {
                pAdjNode = (T2DPlanarNode*) pEdge->getHead();
                vecEdge  = *(pAdjNode->vector());
                vecEdge -= vecNode;
                
                vecNormEdge[X] =  vecEdge[Y];
                vecNormEdge[Y] = -vecEdge[X];
                
                arrOutEdges[i] = pEdge;
                
                alpha  = (vecEdge * vecInEdge) / 
                         (vecEdge.euclidNorm() * vecInEdge.euclidNorm());
                // check for unpleasant values
                if (alpha < -1.0) alpha = -1.0;
                if (alpha >  1.0) alpha =  1.0;
                alpha  = acos(alpha);
                
                alpha2 = (vecNormEdge * vecInEdge) /
                         (vecNormEdge.euclidNorm() * vecInEdge.euclidNorm());
                // check for unpleasant values
                if (alpha2 < -1.0) alpha2 = -1.0;
                if (alpha2 >  1.0) alpha2 =  1.0;
                alpha2  = acos(alpha2);
                
                // adjusting the angle to have angles in [0,PI2]
                if (alpha2 < M_PI_2)
                    arrOutAngles[i] = PI2 - alpha;
                else
                    arrOutAngles[i] = alpha;
                
                i++;
            }// while
            
            //
            // now reput all edges in the outedges-list while sorting them 
            // according to their angle
            //
            for (i = 0; i < noOfOutEdges; i++)
            {
                alpha = arrOutAngles[i];
                arrOrder[i] = 0;
                
                for (j = 0; j < noOfOutEdges; j++)
                {
                    if ((i != j) && (arrOutAngles[j] < alpha))
                        arrOrder[i]++;
                }// for
            }// for
            
            pOutEdges = &(pNode->getOutEdges());
            pOutEdges->clear();
            
            for (i = 0; i < noOfOutEdges; i++)
            {
                for (j = 0; j < noOfOutEdges; j++)
                {
                    if (arrOrder[j] == (noOfOutEdges-i-1))
                        pOutEdges->add( *(arrOutEdges[j] ) );
                }// for
            }// for
        }// if
        
        //
        // order in-edges
        //
        if ((pNode->getOutDegree() > 0) && (pNode->getInDegree() > 1))
        {
            noOfInEdges  = pNode->getInDegree();
            noOfOutEdges = pNode->getOutDegree();
            
            arrInEdges.setSize( noOfInEdges );
            arrInAngles.setSize( noOfInEdges );
            arrOrder.setSize( noOfInEdges );
            
            //
            // go over all in-edges and compute the angle between them
            // and the first out-edges
            //
            
            pOutEdge    = pNode->getOutEdges().getFirst();
            vecOutEdge  = *(((T2DPlanarNode*) (pOutEdge->getHead()))->vector());
            vecOutEdge -= vecNode;
            
            i = 0;
            iterEdges.setList( pNode->getInEdges() );
            iterEdges.first();
            while ((pEdge = iterEdges++) != NULL)
            {
                pAdjNode = (T2DPlanarNode*) pEdge->getTail();
                vecEdge  = *(pAdjNode->vector());
                vecEdge -= vecNode;
                
                vecNormEdge[X] =  vecEdge[Y];
                vecNormEdge[Y] = -vecEdge[X];
                
                arrInEdges[i] = pEdge;
                
                alpha  = (vecEdge * vecOutEdge) / 
                         (vecEdge.euclidNorm() * vecOutEdge.euclidNorm());
                // check for unpleasant values
                if (alpha < -1.0) alpha = -1.0;
                if (alpha >  1.0) alpha =  1.0;
                alpha  = acos(alpha);
                
                alpha2 = (vecNormEdge * vecOutEdge) /
                         (vecNormEdge.euclidNorm() * vecOutEdge.euclidNorm());
                // check for unpleasant values
                if (alpha2 < -1.0) alpha2 = -1.0;
                if (alpha2 >  1.0) alpha2 =  1.0;
                alpha2  = acos(alpha2);
                
                // adjusting the angle to have angles in [0,PI2]
                if (alpha2 < M_PI_2)
                    arrInAngles[i] = PI2 - alpha;
                else
                    arrInAngles[i] = alpha;
                
                i++;
            }// while
            
            //
            // now reput all edges in the outedges-list while sorting them 
            // according to their angle
            //
            for (i = 0; i < noOfInEdges; i++)
            {
                alpha = arrInAngles[i];
                arrOrder[i] = 0;
                
                for (j = 0; j < noOfInEdges; j++)
                {
                    if ((i != j) && (arrInAngles[j] < alpha))
                        arrOrder[i]++;
                }// for
            }// for
            
            pInEdges = &(pNode->getInEdges());
            pInEdges->clear();
            
            for (i = 0; i < noOfInEdges; i++)
            {
                for (j = 0; j < noOfInEdges; j++)
                {
                    if (arrOrder[j] == (noOfInEdges-i-1))
                        pInEdges->add( *(arrInEdges[j] ) );
                }// for
            }// for
        }// if
    }// while
}

//
// sort the edges of each nodes according to their weight
//
void 
T2DPlanarGraph::sortEdgeWeight ()
{
    uint  i,j;

    TListIterator<TNode>   iterNodes( listNodes );
    TNode                 *pNode;
    TListIterator<TEdge>   iterEdges;
    TWeightedMarkableEdge *pEdge;

    // these arrays are for direct accessing the out-edges
    // of every node and for array which hold the ordering
    TDynArray< TWeightedMarkableEdge * >  arrOutEdges( 5, 5 );
    TDynArray< uint >     arrOrder( 5, 5 );
    uint                   noOfOutEdges;

    TFloat  weight;

    iterNodes.first();
    while ((pNode = iterNodes++) != NULL)
    {
        if (pNode->getOutDegree() > 1)
        {
            noOfOutEdges = pNode->getOutDegree();

            arrOutEdges.setSize( noOfOutEdges );
            arrOrder.setSize( noOfOutEdges );

            //
            // put all edges in the array
            //
            iterEdges.setList( pNode->getOutEdges() );
            iterEdges.first(); i = 0;
            while ((pEdge = (TWeightedMarkableEdge*) iterEdges++) != NULL)
            {
                arrOutEdges[i] = pEdge;
                i++;
            }// while

            //
            // now reput all edges in the outedges-list while sorting them 
            // according to their weight
            //
            for (i = 0; i < noOfOutEdges; i++)
            {
                weight = arrOutEdges[i]->weight();
                arrOrder[i] = 0;

                for (j = 0; j < noOfOutEdges; j++)
                {
                    if (i != j)
                        if (arrOutEdges[j]->weight() > weight)
                            arrOrder[i]++;
                }// for
            }// for

            TList<TEdge>  *pOutEdges = &(pNode->getOutEdges());
            pOutEdges->clear();

            for (i = 0; i < noOfOutEdges; i++)
                for (j = 0; j < noOfOutEdges; j++)
                    if (arrOrder[j] == (noOfOutEdges-i-1))
                        pOutEdges->add( *(arrOutEdges[j]) );
        }// if
    }// while
}

//
// sort the edges of each nodes according to the sum of
// their and their succ. weights
//
void 
T2DPlanarGraph::sortSuccEdgeWeight ()
{
    uint  i,j;

    TListIterator<TNode>   iterNodes( listNodes );
    TNode                 *pNode;
    T2DPlanarNode         *pSucc;
    TListIterator<TEdge>   iterEdges;
    TWeightedMarkableEdge *pEdge;

    // these arrays are for direct accessing the out-edges
    // of every node and for array which hold the ordering
    TDynArray< TWeightedMarkableEdge * >     arrOutEdges( 5, 5 );
    TDynArray<uint>       arrOrder( 5, 5 );
    uint                   noOfOutEdges;

    TFloat  weight;

    iterNodes.first();
    while ((pNode = iterNodes++) != NULL)
    {
        if (pNode->getOutDegree() > 1)
        {
            noOfOutEdges = pNode->getOutDegree();

            arrOutEdges.setSize( noOfOutEdges );
            arrOrder.setSize( noOfOutEdges );

            //
            // put all edges in the array
            //
            iterEdges.setList( pNode->getOutEdges() );
            iterEdges.first(); i = 0;
            while ((pEdge = (TWeightedMarkableEdge*) iterEdges++) != NULL)
            {
                arrOutEdges[i] = pEdge;
                i++;
            }// while

            //
            // now reput all edges in the outedges-list while sorting them 
            // according to their weight
            //
            for (i = 0; i < noOfOutEdges; i++)
            {
                weight = arrOutEdges[i]->weight() + 
                         ((T2DPlanarNode*)arrOutEdges[i]->getHead())->weight();
                arrOrder[i] = 0;

                for (j = 0; j < noOfOutEdges; j++)
                {
                    if (i != j)
                    {
                        pSucc = (T2DPlanarNode*)arrOutEdges[i]->getHead();
                        if ((arrOutEdges[j]->weight() + pSucc->weight()) > weight)
                        {
                            arrOrder[i]++;
                        }// if
                    }// if
                }// for
            }// for

            TList<TEdge>  *pOutEdges = &(pNode->getOutEdges());
            pOutEdges->clear();

            for (i = 0; i < noOfOutEdges; i++)
                for (j = 0; j < noOfOutEdges; j++)
                    if (arrOrder[j] == (noOfOutEdges-i-1))
                        pOutEdges->add( *(arrOutEdges[j]) );
        }// if
    }// while
}

//
// create new graph
//
T2DPlanarGraph * 
T2DPlanarGraph::create () 
{ 
    return new T2DPlanarGraph;
}
