//
//  Project     : libgraph
//  File        : T3DCellEdgeNode.cc
//  Author      : Ronald Kriemann
//  Purpose     : class for a node holding a 3D celledge
//
// arch-tag: 75b340e8-6eb8-4acb-afa5-781482abd7b3
//

#include "T3DCellEdgeNode.hh"

// different directions of conv-diff : forward/backwards
#define FORWARD   0
#define BACKWARDS 1

// defines for face-marks
#define FACE_STRONG 0x800
#define FACE_WEAK   0x1000

// different flags used in this class
#define DELETED    0x1
#define STRONG     0x2
#define IN_SURFACE 0x4

//
// constructor
//
T3DCellEdgeNode::T3DCellEdgeNode ( uint n,
                                   T3DCellEdge * edge,
                                   T3DVertexData * v0, T3DVertexData * v1,
                                   TVector<3> * b )
        : TNode( n ), TFlagable<uchar>( 0 )
{
    if ((v0 == NULL) || (edge == NULL) || (v1 == NULL) || (b == NULL))
    {
        cout << "T3DCellEdgeNode::T3DCellEdgeNode (...) : got NULL arg" << endl;
        exit( -1 );
    }// if

    pCellEdge = edge;
    pCellEdge->addRef();
    
    pNodeData0 = v0;
    pNodeData1 = v1;

    pNodeData0->addRef();
    pNodeData1->addRef();

    pListFaces = NULL;
    
    vecConvDir = *b;

    pPhi_f = pPhi_b = NULL;

    _usedFSFunc    = NO_FUNC;
    _noOfUsedFaces = 0;

    pFSFace[0] = pFSFace[1] = NULL;
}

//
// destructor
//
T3DCellEdgeNode::~T3DCellEdgeNode ()
{
    if (pCellEdge != NULL)
        pCellEdge->free();

    if (pNodeData0 != NULL)
        pNodeData0->free();

    if (pNodeData1 != NULL)
        pNodeData1->free();

    if (pPhi_f != NULL)
        pPhi_f->free();
    
    if (pPhi_b != NULL)
        pPhi_b->free();
    
    if (pListFaces != NULL)
    {
        TListIterator< T3DFrontier >  iterFaces( *pListFaces );
        T3DFrontier                  *pFace;
        
        iterFaces.first();
        while ((pFace = iterFaces++) != NULL)
            pFace->free();

        delete pListFaces;
    }// if
}

//////////////////////////////////////////////////
//
// access member variables
//

T3DCellEdge *
T3DCellEdgeNode::cellEdge () const
{
    return pCellEdge;
}

T3DVertexData *
T3DCellEdgeNode::nodeData ( uint i ) const
{
    switch (i)
    {
        case 0  : return pNodeData0;
        case 1  : return pNodeData1;
        default : return NULL;
    }// switch
}

T3DFrontier *
T3DCellEdgeNode::phi_f () const
{
    return pPhi_f;
}

T3DFrontier *
T3DCellEdgeNode::phi_b () const
{
    return pPhi_b;
}

const int
T3DCellEdgeNode::noOfUsedFaces () const
{
    return _noOfUsedFaces;
}

void
T3DCellEdgeNode::addUsedFaces ()
{
    _noOfUsedFaces++;
}

void
T3DCellEdgeNode::setNoOfUsedFaces ( uint n )
{
    _noOfUsedFaces = n;
}
    
const bool
T3DCellEdgeNode::isDeleted () const
{
    return hasFlag( DELETED );
}

void
T3DCellEdgeNode::setDeleted ( bool b )
{
    if (b == TRUE)
        setFlag( DELETED );
    else
        clearFlag( DELETED );
}

T3DFrontier *
T3DCellEdgeNode::fsFace ( int i ) const
{
    if (i == 1)
        return pFSFace[1];
    else
        return pFSFace[0];
}

void
T3DCellEdgeNode::setFSFace ( T3DFrontier * face )
{
    if (pFSFace[0] == NULL)
        pFSFace[0] = face;
    else
        pFSFace[1] = face;
}
    
void
T3DCellEdgeNode::setFSFace ( int i, T3DFrontier * face )
{
    if (i == 1)
        pFSFace[1] = face;
    else
        pFSFace[0] = face;
}

void
T3DCellEdgeNode::clearFSFaces ()
{
    pFSFace[0] = pFSFace[1] = NULL;
}
    
const int
T3DCellEdgeNode::usedFSFunc () const
{
    return _usedFSFunc;
}

void
T3DCellEdgeNode::setUsedFSFunc ( int f )
{
    _usedFSFunc = f;
}
    
const bool
T3DCellEdgeNode::isStrong () const
{
    return hasFlag( STRONG );
}

const bool
T3DCellEdgeNode::isWeak () const
{
    return ! hasFlag( STRONG );
}

void
T3DCellEdgeNode::setStrong ( bool b )
{
    if (b == TRUE)
        setFlag( STRONG );
    else
        clearFlag( STRONG );
}
    
const bool
T3DCellEdgeNode::isInSurface () const
{
    return hasFlag( IN_SURFACE );
}

void
T3DCellEdgeNode::setInSurface ( bool b )
{
    if (b == TRUE)
        setFlag( IN_SURFACE );
    else
        clearFlag( IN_SURFACE );
}
    
//////////////////////////////////////////////////
//
// methods for facelist management
//
    
//
// add a adj face (adj. with repsect to the celledge)
//
void
T3DCellEdgeNode::addFace ( T3DFrontier * face )
{
    if (face != NULL)
    {
        if (pListFaces == NULL)
            pListFaces = new TList< T3DFrontier >;
        
        face->addRef();
        pListFaces->add( *face );
    }// if
    else
        cout << "T3DCellEdgeNode::addFace (...) : gave NULL face" << endl;
}

//
// remove a face from facelist
// (return TRUE if successfuly deleted, else FALSE)
//
bool
T3DCellEdgeNode::remFace ( T3DFrontier * face )
{
    if (face != NULL)
    {
        if (pListFaces == NULL)
            return FALSE;
        
        if (isInFaceList( face ))
        {
            pListFaces->remove( *face );
            face->free();
            
            return TRUE;
        }// if
        else
            return FALSE;
    }// if
    else
        cout << "T3DCellEdgeNode::remFace (...) : gave NULL face" << endl;

    return FALSE;
}
    
//
// check if a face is in the facelist
// return TRUE if in facelist, else FALSE
//
bool
T3DCellEdgeNode::isInFaceList ( T3DFrontier * face )
{
    if (face != NULL)
    {
        if (pListFaces == NULL)
            return FALSE;
        
        if (pListFaces->search( *face ) != NULL)
            return TRUE;
        else
            return FALSE;
    }// if
    else
        cout << "T3DCellEdgeNode::remFace (...) : gave NULL face" << endl;

    return FALSE;
}

//
// remove all adj weak faces (a face with no strong edge)
//
void
T3DCellEdgeNode::remWeakFaces ()
{
    if (pListFaces != NULL)
    {
        TListIterator< T3DFrontier >  iterFaces( *pListFaces );
        TItem< T3DFrontier >         *pActItem;
        T3DFrontier                  *pFace;

        iterFaces.first();
        while ((pFace = iterFaces()) != NULL)
        {
            pActItem = iterFaces.getItem();
            iterFaces++;
            
            if ( ! isStrongFace( pFace ) )
                pListFaces->remove( * pActItem );
        }// while
    }// if
}

//
// remove whole face list
//
void
T3DCellEdgeNode::remFaceList ()
{
    if (pListFaces != NULL)
    {
        TListIterator< T3DFrontier >  iterFaces( *pListFaces );
        T3DFrontier                  *pFace;
        
        iterFaces.first();
        while ((pFace = iterFaces++) != NULL)
            pFace->free();

        delete pListFaces;

        pListFaces = NULL;
    }// if
}
    
//////////////////////////////////////////////////
//
// methods for building flow-surfaces
//
    
//
// calc face with minimal angle with respect to
// the conv-dir (NULL if no such face exists)
// for forward/backwards direction
//

T3DFrontier *
T3DCellEdgeNode::calcPhi ( TFloat gamma )
{
    TVector<3>  *v0, *v1;
    TVector<3>   e, m, b, normal_e, normal_b;
    TVector<3>   x;
    TFloat       angle_e_b, normSqr_e, norm;

    if (pListFaces == NULL)
    {
        cout << "T3DCellEdgeNode::calcPhi () : no facelist present" << endl;
        return NULL;
    }// if

    if (pPhi_f != NULL)
        pPhi_f->free();
    
    if (pPhi_b != NULL)
        pPhi_b->free();
    
    v0 = pNodeData0->node();
    v1 = pNodeData1->node();

    if ((v0 == NULL) || (v1 == NULL))
    {
        cout << "T3DCellEdgeNode::calcPhi () : have NULL node" << endl;
        return NULL;
    }// if
    
    // calc vector of edge-dir and edge-midpoint
    e  = *v1;
    e -= *v0;

    m  = e;
    m *= 0.5;
    m += *v0;

    b = vecConvDir;

    normSqr_e = e * e;
    if (normSqr_e == 0)
        return NULL;
    
    // check if e || b (if so, no face can be found)
    normal_e  = e;
    normal_e *= 1 / sqrt( normSqr_e );

    normal_b = b;
    if ((norm = b.euclidNorm()) == 0)
        return NULL;
    else
        normal_b *= 1 / norm;
    
    angle_e_b = fabs(normal_e * normal_b);
    if (angle_e_b == 1)
        return NULL;

    angle_e_b = e * b;

    // calc vector x in plane(e,b)
    x  = e;
    x *= - angle_e_b / normSqr_e;
    x += b;

    // e || b => ||x|| = 0 => no phi_f/b
    if ((norm = x.euclidNorm()) == 0)
        return NULL;
    else
        x *= 1 / norm;

    //
    // go over all faces in facelist and determine the
    // angle between plane(e,b) and the face
    // the face with the minimal angle is our phi
    //
    uint                          i;
    TListIterator< T3DFrontier >  iterFaces( *pListFaces );
    T3DFrontier                  *pFace, *pMinFace_f, *pMinFace_b;
    TVector<3>                   *pOpposite;
    TVector<3>                    h, v;
    TVector<3>                    y, t;
    TFloat                        cosAngle, angle_f, angle_b;
    TFloat                        minAngle_f, minAngle_b;
    uint                         noOfNodes;

    minAngle_b = minAngle_f = 2 * PI2;
    pMinFace_f = pMinFace_b = NULL;

    iterFaces.first();
    while ((pFace = iterFaces++) != NULL)
    {
        // look for opposite node (with respect to e)
        
        noOfNodes = pFace->getNoOfNodes();
        pOpposite = NULL;
        
        for (i = 0; i < noOfNodes; i += 2)
        {
            pOpposite = pFace->getNode( i );

            if ((pOpposite != v0) && (pOpposite != v1))
                break;
            else
                pOpposite = NULL;
        }// for

        if (pOpposite == NULL)
        {
            cout << "T3DCellEdgeNode::calcPhi (...) : have NULL opposite node" << endl;
            continue;
        }// if

        v = *pOpposite;
        
        // calc vector y in plane(e,h), where h = v-m
        h  = v;
        h -= m;

        y  = e;
        y *= - (e * h) / normSqr_e;
        y += h;

        if ((norm = y.euclidNorm()) == 0)
            continue;
        else
            y *= 1 / norm;

        // calc angle between plane(e,b) and plane(e,h) by
        // calc the angle between x and y
        cosAngle = x * y;
        
        if (cosAngle > 1)
            cosAngle = 1;
        else if (cosAngle < -1)
            cosAngle = -1;

        angle_f = acos(   cosAngle );
//        angle_b = acos( - cosAngle );
        angle_b = PI - angle_f;
        
        // apply correction to angles
        t  = y;
        t -= x;
        norm = t.euclidNorm();
        if ((norm > SQRT2) || (norm < 0))
            angle_f = PI2 - angle_f;

        t  = y;
        t += x;
        norm = t.euclidNorm();
        if ((norm > SQRT2) || (norm < 0))
            angle_b = PI2 - angle_b;

        if (angle_f < minAngle_f)
        {
            pMinFace_f = pFace;
            minAngle_f = angle_f;
        }// if

        if (angle_b < minAngle_b)
        {
            pMinFace_b = pFace;
            minAngle_b = angle_b;
        }// if
    }// while

    // apply upper bound for angle
    if (minAngle_f > gamma)
        pMinFace_f = NULL;

    if (minAngle_b > gamma)
        pMinFace_b = NULL;

    if ((pPhi_f = pMinFace_f) != NULL)
        pPhi_f->addRef();

    if ((pPhi_b = pMinFace_b) != NULL)
        pPhi_b->addRef();

    if ((pPhi_f != NULL) && (pPhi_f == pPhi_b))
        cout << "T3DCellEdgeNode::calcPhi (...) : have equal phi_f and phi_b" << endl;
    
    return pPhi_f;
}

//
// return face with minimal angle with repect to
// given face (NULL if no such face exists)
//
T3DFrontier *
T3DCellEdgeNode::phiHat ( T3DFrontier * pFace ) const
{
    if (pListFaces == NULL)
    {
        cout << "T3DCellEdgeNode::phiHat (...) : no facelist present" << endl;
        return NULL;
    }// if
    
    if (pFace == NULL)
    {
        cout << "T3DCellEdgeNode::phiHat (...) : got NULL face" << endl;
        return NULL;
    }// if

    if (pCellEdge == NULL)
    {
        cout << "T3DCellEdgeNode::phiHat (...) : have NULL celledge" << endl;
        return NULL;
    }// if
    
    //
    // look for opposite node (opposite to local edge) in pFace
    //
    uint         i;
    uint         noOfNodes;
    TVertex<3>  *pOpposite1 = NULL;


    noOfNodes = pFace->getNoOfNodes();
    for (i = 0; i < noOfNodes; i++)
    {
        if ((i % 2) != 0)
            continue;

        pOpposite1 = pFace->getNode( i );
        if ((pOpposite1 != pCellEdge->getNode( 0 )) &&
            (pOpposite1 != pCellEdge->getNode( 2 )))
            break;
        else
            pOpposite1 = NULL;
    }// for

    if (pOpposite1 == NULL)
    {
        cout << "T3DCellEdgeNode::phiHat (...) : can't find opposite node" << endl;
        return NULL;
    }// if
    
    //
    // go over faces in facelist and look for the one with
    // minimal angle to pFace
    //

    TListIterator< T3DFrontier >  iterFaces( *pListFaces );
    T3DFrontier                  *pAdjFace, *pPhiHat;
    TFloat                        maxAngle;
    TFloat                        cosAngle, angle;
    TVector<3>                    e, m, b, h, x, y;
    TFloat                        sqrNorm_e, norm;
    TVertex<3>                   *pOpposite2 = NULL;

    
    pPhiHat  = NULL;
    maxAngle = 0;

    //
    // calc edge-dir, -midpoint and a vector which
    // represents pFace (x)
    //
    e  = * pCellEdge->getNode(2);
    e -= * pCellEdge->getNode(0);

    sqrNorm_e = e * e;
    if (sqrNorm_e == 0)
        return NULL;
    
    m  = e;
    m *= 0.5;
    m += * pCellEdge->getNode(0);
    
    b  = *pOpposite1;
    b -= m;

    x  = e;
    x *= - (b * e) / sqrNorm_e;
    x += b;

    if ((norm = x.euclidNorm()) == 0)
        return NULL;
    else
        x *= 1 / norm;

    iterFaces.first();
    while ((pAdjFace = iterFaces++) != NULL)
    {
        if (pAdjFace == pFace)
            continue;

        //
        // look for opposite node in pAdjFace
        //
        noOfNodes = pAdjFace->getNoOfNodes();
        for (i = 0; i < noOfNodes; i++)
        {
            if ((i % 2) != 0)
                continue;

            pOpposite2 = pAdjFace->getNode( i );
            if ((pOpposite2 != pCellEdge->getNode( 0 )) &&
                (pOpposite2 != pCellEdge->getNode( 2 )))
                break;
            else
                pOpposite2 = NULL;
        }// for

        if (pOpposite2 == NULL)
        {
            cout << "T3DCellEdgeNode::phiHat (...) : can't find opposite node" << endl;
            continue;
        }// if

        //
        // calc vector wich represents pAdjFace (y) and
        // determine angle between x and y
        //

        h  = *pOpposite2;
        h -= m;

        y  = e;
        y *= - (h * e) / sqrNorm_e;
        y += h;

        if ((norm = y.euclidNorm()) == 0)
            continue;
        else
            y *= 1 / norm;

        cosAngle = x * y;

        if (cosAngle > 1)
            cosAngle = 1;

        if (cosAngle < -1)
            cosAngle = -1;

        angle = acos( cosAngle );

        // handle overflow
//         y -= x;
//         norm = y.euclidNorm();
    
//         if ((norm < 0) || (norm > SQRT2))
//             angle = PI2 - angle;

        if (angle > PI)
            angle = PI2 - angle;
        
        if (angle > maxAngle)
        {
            pPhiHat  = pAdjFace;
            maxAngle = angle;
        }// if
    }// while
    
    return pPhiHat;
}

//////////////////////////////////////////////////
//
// virtual constructors
//
    
//
// create a new nodes
//
T3DCellEdgeNode *
T3DCellEdgeNode::clone ()
{
    T3DCellEdgeNode * pTemp = new T3DCellEdgeNode( name, pCellEdge,
                                                   pNodeData0, pNodeData1,
                                                   & vecConvDir );

    return pTemp;
}

//////////////////////////////////////////////////
//
// misc. help methods
//

//
// check if the given face is strong (if one edge is strong)
//
bool
T3DCellEdgeNode::isStrongFace ( T3DFrontier * pFace ) const
{
    if (pFace == NULL)
    {
        cout << "T3DCellEdgeNode::isStrongFace (...) : got NULL face" << endl;
        return FALSE;
    }// if

    uint             i, noOfEdges;
    T3DCellEdge     *pCellEdge;
    T3DCellEdgeNode *pNode;

    if (pFace->hasFlag( FACE_WEAK | FACE_STRONG) )
    {
        if (pFace->hasFlag( FACE_WEAK ))
            return FALSE;
        else
            return TRUE;
    }// if
        
    noOfEdges = pFace->getNoOfEdges();
    for (i = 0; i < noOfEdges; i++)
    {
        if ((pCellEdge = pFace->getEdge( i )) ==  NULL)
        {
            cout << "T3DCellEdgeNode::isStringFace (...) : have NULL edge" << endl;
            return FALSE;
        }// if

        // really dirty way
        if ((pNode = (T3DCellEdgeNode*) pCellEdge->data()) != NULL)
        {
            if (pNode->isStrong())
            {
                pFace->setFlag( FACE_STRONG );
                return TRUE;
            }// if
        }// if
    }// for

    pFace->setFlag( FACE_WEAK );
    
    return FALSE;
}
