//
//  Project     : libgraph
//  File        : T3DFlowSurface.cc
//  Author      : Ronald Kriemann
//  Purpose     : class for building flow surfaces in a T3DCellEdgeGraph
//
// arch-tag: 3e00fec3-51be-4a66-8f3b-703f3cf01a41
//

#include "var.hh"
#include "cmdline.hh"

#include "drawapp.hh"

#include "graphtool.hh"
#include "T3DVertexData.hh"
#include "TFSNode.hh"
#include "T3DFaceOrient.hh"

#include "T3DFlowSurface.hh"

#define NO_FACE  NULL

#define PHI_F  0
#define PHI_B  1
#define PHI_P  2

// face in flowsurface we go to or we come from
#define TO_FACE    0
#define FROM_FACE  1

// define to set/test orientation
#define SET_ORIENT
//#define CHECK_ORIENT

/////////////////////////////////////////////////
//
// constructors and destructors
//

T3DFlowSurface::T3DFlowSurface ()
{
    pFlowSurface = NULL;
}

T3DFlowSurface::~T3DFlowSurface ()
{
}

/////////////////////////////////////////////////
//
// methods for building a list of graphs, each
// representing a flow surface
//

TList< T3DCellEdgeGraph > *
T3DFlowSurface::buildFlowSurfaces ( T3DCellEdgeGraph * pGraph,
                                    TFloat alpha, TFloat beta,
                                    uint method )
{
    TList< T3DCellEdgeGraph >  *pFSList;
    TListIterator< TNode >      iterNodes;
    T3DCellEdgeNode            *pNode;
    uint                        minSize;
    
    if (pGraph == NULL)
    {
        cout << "T3DFlowSurface::buildFlowSurfaces (...) : got NULL graph" << endl;
        return NULL;
    }// if

    //
    // reset data in all nodes and vertexdata-fields
    //
    iterNodes.setList( pGraph->getNodeList() );

    iterNodes.first();
    while ((pNode = (T3DCellEdgeNode*) iterNodes++) != NULL)
    {
        pNode->setNoOfUsedFaces( 0 );
        pNode->setDeleted( FALSE );

        if (pNode->nodeData( 0 ) != NULL)
            pNode->nodeData( 0 )->setFlowSurface( NULL );

        if (pNode->nodeData( 1 ) != NULL)
            pNode->nodeData( 1 )->setFlowSurface( NULL );
    }// while

    //
    // build recursivly all flow surfaces
    //

    T3DFlowSurface::alpha = alpha;
    T3DFlowSurface::beta  = beta;

    if (verbosity >= VERBOSE_STEP)
    {
        if (method == SEARCH_DFS)
            cout << "T3DFlowSurface::buildFlowSurfaces (...) : building FSs via DFS" << endl;
        else
            cout << "T3DFlowSurface::buildFlowSurfaces (...) : building FSs via BFS" << endl;
    }// if

    // set minimal size for flow surfaces
    if ((minSize = pGraph->getNoOfNodes() / 50) > 21)
        minSize = 21;
    
    pFSList = new TList< T3DCellEdgeGraph >;
    
    iterNodes.first();
    while ((pNode = (T3DCellEdgeNode*) iterNodes++) != NULL)
    {
        if (isStartNode( pNode ) == TRUE)
        {
            pFlowSurface = new T3DCellEdgeGraph;

            startFlowSurface( pNode, method );

            if (pFlowSurface->getNoOfNodes() >= minSize)
            {
                if (verbosity >= VERBOSE_STEP)
                    cout << "T3DFlowSurface::buildFlowSurfaces (...) : flow surface found; size = "
                         << pFlowSurface->getNoOfNodes() << endl;
            
                deleteFS( pFlowSurface );
                pFSList->add( *pFlowSurface );

                //
                // check orientation of flowsurface
                //
#ifdef CHECK_ORIENT
                TListIterator< TNode >  iterFSNodes( pFlowSurface->getNodeList() );
                T3DCellEdgeNode       * pCellEdgeNode;

                iterFSNodes.first();
                while ((pCellEdgeNode = (T3DCellEdgeNode*) iterFSNodes++) != NULL)
                {
                    if ((pCellEdgeNode->fsFace( TO_FACE ) != NO_FACE) &&
                        (pCellEdgeNode->fsFace( FROM_FACE ) != NO_FACE))
                    {
                        if ( ! T3DFaceOrient::chkOrient( pCellEdgeNode->fsFace( TO_FACE ),
                                                         pCellEdgeNode->fsFace( FROM_FACE )) )
                        {
                            cout << "T3DFlowSurface::buildFlowSurfaces (...) : "
                                 << "flowsurface not oriented" << endl;
                            T3DFaceOrient::showOrient( pCellEdgeNode->fsFace( TO_FACE ) );
                            T3DFaceOrient::showOrient( pCellEdgeNode->fsFace( FROM_FACE ) );
//                            exit( 1 );
                        }// if
                    }// if
                }// while
#endif
            }// if
            else
            {
                if (verbosity >= VERBOSE_STEP)
                    cout << "T3DFlowSurface::buildFlowSurfaces (...) : flow surface too small"
                         << endl;
            
                unDeleteFS( pFlowSurface );
                delete pFlowSurface;
            }// else
        }// if
    }// while

//    orderFS( pGraph, pFSList );

    return pFSList;
}

//
// start a new flow surface
//
void
T3DFlowSurface::startFlowSurface ( T3DCellEdgeNode * pNode,
                                   uint method )
{
    if (pNode == NULL)
    {
        cout << "T3DFlowSurface::startFlowSurface (...) : got NULL arg" << endl;
        return;
    }// if

    //
    // start flow surface by expanding through phi_f and phi_b
    //
    T3DFrontier   *pPhi;

    if (method == SEARCH_DFS)
    {
        if ((pPhi = pNode->phi_f()) != NO_FACE)
        {
            if ((checkS1( pNode, pPhi, PHI_F) == FALSE) ||
                (checkS2( pNode, pPhi, PHI_F) == FALSE))
                return;

#ifdef SET_ORIENT
            T3DFaceOrient::setOrient( pPhi, 0, 2, 4 );
#endif
            
            addFSNode( pNode, pPhi, PHI_F );

            expandFS( pNode, pPhi, PHI_F );
        }// if

        if ((pPhi = pNode->phi_b()) != NO_FACE)
        {
            if ((checkS1( pNode, pPhi, PHI_B) == FALSE) ||
                (checkS2( pNode, pPhi, PHI_B) == FALSE))
                return;

            addFSNode( pNode, pPhi, PHI_B );

            expandFS( pNode, pPhi, PHI_B );
        }// if
    }// if
    else // method == SEARCH_BFS
    {
        if ((pPhi = pNode->phi_f()) != NO_FACE)
        {
            if ((checkS1( pNode, pPhi, PHI_F) == FALSE) ||
                (checkS2( pNode, pPhi, PHI_F) == FALSE))
                return;
            
            addFSNode( pNode, pPhi, PHI_F );
        }// if

        if ((pPhi = pNode->phi_b()) != NO_FACE)
        {
            if ((checkS1( pNode, pPhi, PHI_B) == FALSE) ||
                (checkS2( pNode, pPhi, PHI_B) == FALSE))
                return;

            addFSNode( pNode, pPhi, PHI_B );
        }// if

        buildFS_BFS( pNode );
    }// if
}
    
//
// recursion/iteration for building one flow surface
//
void
T3DFlowSurface::buildFS_DFS ( T3DCellEdgeNode * pNode,
                              T3DFrontier * pPhi,
                              int func )
{
    if ((pNode == NULL) || (pPhi == NULL))
    {
        cout << "T3DFlowSurface::buildFS_DFS (...) : got NULL arg" << endl;
        return;
    }// if

    //
    // check conditions S1/S2
    //
    if ((checkS1( pNode, pPhi, func ) == FALSE) ||
        (checkS2( pNode, pPhi, func ) == FALSE))
    {
        delEdges( pNode );
        return;
    }// if

    //
    // add node/face to actual flow surface
    //
    addFSNode( pNode, pPhi, func );
    
    //
    // expand flow surface by all succ/pred
    //
    expandFS( pNode, pPhi, func );
}

void
T3DFlowSurface::buildFS_BFS ( T3DCellEdgeNode * pStartNode )
{
    TList< T3DCellEdgeNode >          listS1, listS2;
    TListIterator< T3DCellEdgeNode >  iterS1, iterS2;
    T3DCellEdgeNode                  *pNode, *pAdjNode;
    TListIterator< TEdge >            listEdges;
    TEdge                            *pEdge;
    T3DFrontier                      *pFace, *pPhi;
    int                               func, usedFunc;

    if (pStartNode == NULL)
    {
        cout << "T3DFlowSurface::buildFS_BFS (...) : got NULL arg" << endl;
        return;
    }// if
    
    //
    // care about startnode : add all succ through phi_f
    //
    if (pStartNode->fsFace() != NO_FACE)
    {
        listEdges.setList( pStartNode->getOutEdges() );
        listEdges.first();

        pFace = pStartNode->fsFace();
        
        while ((pEdge = listEdges++) != NULL)
        {
            pAdjNode = (T3DCellEdgeNode*) pEdge->getHead();
            
            // check if max. number of faces was reached
            // or node was already used
            if ((pAdjNode->noOfUsedFaces() >= 2) ||
                (pAdjNode->fsFace() != NO_FACE))
                continue;

            //
            // check conditions FS/S1/S2
            //
            if (((pPhi = checkFS( pAdjNode, pFace, func )) == NO_FACE) ||
                (checkS1( pAdjNode, pPhi, func ) == FALSE) ||
                (checkS2( pAdjNode, pPhi, func ) == FALSE))
            {
                delEdges( pAdjNode );
                continue;
            }// if

            // add node/face to actual flow surface
            addFSNode( pAdjNode, pPhi, func );
            listS1.add( *pAdjNode );
        }// while
    }// if
    
    //
    // care about startnode : add all succ through phi_b
    //
    if (pStartNode->fsFace( FROM_FACE ) != NO_FACE)
    {
        listEdges.setList( pStartNode->getInEdges() );
        listEdges.first();

        pFace = pStartNode->fsFace( FROM_FACE );
        
        while ((pEdge = listEdges++) != NULL)
        {
            pAdjNode = (T3DCellEdgeNode*) pEdge->getTail();
            
            // check if max. number of faces was reached
            // or node was already used
            if ((pAdjNode->noOfUsedFaces() >= 2) ||
                (pAdjNode->fsFace() != NO_FACE))
                continue;

            //
            // check conditions FS/S1/S2
            //
            if (((pPhi = checkFS( pAdjNode, pFace, func )) == NO_FACE) ||
                (checkS1( pAdjNode, pPhi, func ) == FALSE) ||
                (checkS2( pAdjNode, pPhi, func ) == FALSE))
            {
                delEdges( pAdjNode );
                continue;
            }// if

            // add node/face to actual flow surface
            addFSNode( pAdjNode, pPhi, func );
            listS1.add( *pAdjNode );
        }// while
    }// if

    iterS1.setList( listS1 );
    iterS2.setList( listS2 );

    //
    // do BFS in graph and expand flow surface
    //
    while (listS1.getSize() > 0)
    {
        //
        // for all nodes in S1 add all adjacent nodes to S2
        //
        iterS1.first();
        while ((pNode = iterS1++) != NULL)
        {
            pFace    = pNode->fsFace();
            usedFunc = pNode->usedFSFunc();

            if (pFace == NULL)
            {
                cout << "T3DFlowSurface::buildFS_BFS (...) : have NULL face" << endl;
                continue;
            }// if

            if ((usedFunc == PHI_F) || (usedFunc == PHI_P))
                listEdges.setList( pNode->getOutEdges() );
            else // usedFunc == PHI_B
                listEdges.setList( pNode->getInEdges() );
                    
            listEdges.first();
            while ((pEdge = listEdges++) != NULL)
            {
                pAdjNode = (T3DCellEdgeNode*) pEdge->getOther( pNode );
                
                // check if max. number of faces was reached
                // or node was already used
                if ((pAdjNode->noOfUsedFaces() >= 2) ||
                    (pAdjNode->fsFace() != NO_FACE))
                    continue;
                
                //
                // check conditions FS/S1/S2
                //
                if (((pPhi = checkFS( pAdjNode, pFace, func )) == NO_FACE) ||
                    (checkS1( pAdjNode, pPhi, func ) == FALSE) ||
                    (checkS2( pAdjNode, pPhi, func ) == FALSE))
                {
                    delEdges( pAdjNode );
                    continue;
                }// if
                
                // add node/face to actual flow surface
                addFSNode( pAdjNode, pPhi, func );
                listS2.add( *pAdjNode );
            }// while
        }// while
        
        //
        // copy S2 in S1
        //
        listS1.clear();

        iterS2.first();
        while ((pNode = iterS2++) != NULL)
        {
            listS1.add( *pNode );
            listS2.remove( *pNode );
        }// while
        listS2.clear();
    }// while
}

//
// expand flow surface by adj. nodes of given node
// (this method is optimized for triangles and will not
//  work with other polygons !!!)
//
void
T3DFlowSurface::expandFS ( T3DCellEdgeNode * pNode,
                           T3DFrontier * pPhi,
                           int func )
{
    T3DCellEdgeNode       * pAdjNode1, *pAdjNode2;
    TListIterator< TEdge >  iterEdges;
    TEdge                 * pEdge;
    bool                    bUseNode1, bUseNode2;
    int                     usedEdges1, usedEdges2;
    T3DFrontier           * pPhi1 = NULL, * pPhi2 = NULL;
    int                     func1, func2;
        
    pAdjNode1  = pAdjNode2  = NULL;
    bUseNode1  = bUseNode2  = FALSE;
    usedEdges1 = usedEdges2 = -1;
    func1      = func2      = -1;

    //
    // catch successors
    //
    if ((func == PHI_F) || (func == PHI_P))
        iterEdges.setList( pNode->getOutEdges() );
    else // func == PHI_B
        iterEdges.setList( pNode->getInEdges() );

    iterEdges.first();
    if ((pEdge = iterEdges++) != NULL)
    {
        pAdjNode1 = (T3DCellEdgeNode*) pEdge->getOther( pNode );
        
        if ((pEdge = iterEdges++) != NULL)
            pAdjNode2 = (T3DCellEdgeNode*) pEdge->getOther( pNode );
    }// if

    //
    // check successors
    //
    if ((pAdjNode1 != NULL) &&
        (pAdjNode1->noOfUsedFaces() < 2) &&
        (pAdjNode1->fsFace() == NO_FACE))
    {
        if ((pPhi1 = checkFS( pAdjNode1, pPhi, func1 )) != NO_FACE)
        {
            bUseNode1  = TRUE;
            usedEdges1 = noOfUsedEdges( pPhi1 );
        }// if
    }// if

    if ((pAdjNode2 != NULL) &&
        (pAdjNode2->noOfUsedFaces() < 2) &&
        (pAdjNode2->fsFace() == NO_FACE))
    {
        if ((pPhi2 = checkFS( pAdjNode2, pPhi, func2 )) != NO_FACE)
        {
            bUseNode2 = TRUE;
            usedEdges2 = noOfUsedEdges( pPhi2 );
        }// if
    }// if

    //
    // expand flow surface
    //
    if (usedEdges1 >= usedEdges2)
    {
        if (bUseNode1)
            buildFS_DFS( pAdjNode1, pPhi1, func1 );
        if (bUseNode2)
            buildFS_DFS( pAdjNode2, pPhi2, func2 );
    }// if
    else
    {
        if (bUseNode2)
            buildFS_DFS( pAdjNode2, pPhi2, func2 );
        if (bUseNode1)
            buildFS_DFS( pAdjNode1, pPhi1, func1 );
    }// else
}
    
//
// calc angle between two faces
//
TFloat
T3DFlowSurface::calcFaceAngle ( T3DCellEdge * pEdge,
                                T3DFrontier * pFace1, T3DFrontier * pFace2 ) const
{
    uint        i;
    uint        noOfNodes;
    TVector<3> *pOpposite1 = NULL, *pOpposite2 = NULL;
    
    if ((pEdge == NULL) || (pFace1 == NULL) || (pFace2 == NULL))
    {
        cout << "T3DFlowSurface::calcFaceAngle (...) : got NULL arg" << endl;
        return 0;
    }// if

    //
    // determine nodes in pFace1 and pFace2 which are opposite
    // to pEdge
    //

    noOfNodes = pFace1->getNoOfNodes();
    for (i = 0; i < noOfNodes; i++)
    {
        // only every second node is a corner
        if ((i % 2) == 1)
            continue;
        
        pOpposite1 = pFace1->getNode( i );
        if ((pOpposite1 != pEdge->getNode(0)) &&
            (pOpposite1 != pEdge->getNode(2)))
            break;
        else
            pOpposite1 = NULL;
    }// for
    
    noOfNodes = pFace2->getNoOfNodes();
    for (i = 0; i < noOfNodes; i++)
    {
        // only every second node is a corner
        if ((i % 2) == 1)
            continue;
        
        pOpposite2 = pFace2->getNode( i );
        if ((pOpposite2 != pEdge->getNode(0)) &&
            (pOpposite2 != pEdge->getNode(2)))
            break;
        else
            pOpposite2 = NULL;
    }// for

    if ((pOpposite1 == NULL) || (pOpposite2 == NULL))
    {
        cout << "T3DFlowSurface::calcFaceAngle (...) : have NULL opposite node" << endl;
        return 0;
    }// if

    //
    // calc two normalized vectors, which represent each face
    // and we can calc angle between
    //

    TVector<3>  e, m, b, h, x, y;
    TFloat      sqrNorm_e, cosAngle, angle, norm;

    
    e  = * pEdge->getNode(2);
    e -= * pEdge->getNode(0);

    sqrNorm_e = e * e;
    if (sqrNorm_e == 0)
        return 0;
    
    m  = e;
    m *= 0.5;
    m += * pEdge->getNode(0);

    b  = *pOpposite1;
    b -= m;

    h  = *pOpposite2;
    h -= m;

    x  = e;
    x *= - (b * e) / sqrNorm_e;
    x += b;

    if ((norm = x.euclidNorm()) == 0)
        return 0;
    else
        x *= 1 / norm;

    y  = e;
    y *= - (h * e) / sqrNorm_e;
    y += h;

    if ((norm = y.euclidNorm()) == 0)
        return 0;
    else
        y *= 1 / norm;

    cosAngle = x * y;

    if (cosAngle > 1)
        cosAngle = 1;

    if (cosAngle < -1)
        cosAngle = -1;

    angle = acos( cosAngle );

    // handle overflow
//     y -= x;
//     norm = y.euclidNorm();
    
//     if ((norm < 0) || (norm > SQRT2))
//         return PI2 - angle;
//     else
//         return angle;

    if (angle > PI)
        return PI2 - angle;
    else
        return angle;
}

//
// check conditions for flow surfaces
//
bool
T3DFlowSurface::checkS1 ( T3DCellEdgeNode * pNode,
                          T3DFrontier * pFace,
                          int func ) const
{
    if ((pNode == NULL) || (pFace == NULL))
    {
        cout << "T3DFlowSurface::checkS1 (...) : got NULL arg" << endl;
        return FALSE;
    }// if

    //
    // look if a succ already has two frontiers or
    // if he belongs to another flow surface (if deleted)
    //
    
    TListIterator<TEdge>  iterEdges;
    TEdge                *pEdge;
    T3DCellEdgeNode      *pAdjNode;
    
    if ((func == PHI_F) || (func == PHI_P))
        iterEdges.setList( pNode->getOutEdges() );
    else // func == PHI_B
        iterEdges.setList( pNode->getInEdges() );
        
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pAdjNode = (T3DCellEdgeNode*) pEdge->getOther( pNode );
        
        if ((pAdjNode->noOfUsedFaces() >= 2) ||
            (pAdjNode->isDeleted() == TRUE))
            return FALSE;
    }// while

    return TRUE;
}

bool
T3DFlowSurface::checkS2 ( T3DCellEdgeNode * pNode,
                          T3DFrontier * pFace,
                          int func ) const
{
    if ((pNode == NULL) || (pFace == NULL))
    {
        cout << "T3DFlowSurface::checkS2 (...) : got NULL arg" << endl;
        return FALSE;
    }// if

    //
    // look if all succ have one face attached to them
    // or if no corner is already in a flowsurface
    //
    TListIterator<TEdge>  iterEdges;
    TEdge                *pEdge;
    T3DCellEdgeNode      *pAdjNode;
    T3DVertexData        *pVertex;
    
    if ((func == PHI_F) || (func == PHI_P))
        iterEdges.setList( pNode->getOutEdges() );
    else // func == PHI_B
        iterEdges.setList( pNode->getInEdges() );
    
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pAdjNode = (T3DCellEdgeNode*) pEdge->getOther( pNode );

        if (pAdjNode->noOfUsedFaces() == 1)
            return TRUE;
    }// while
        
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pAdjNode = (T3DCellEdgeNode*) pEdge->getOther( pNode );
            
        pVertex = pAdjNode->nodeData( 0 );
        if ((pVertex != pNode->nodeData( 0 )) &&
            (pVertex != pNode->nodeData( 1 )) &&
            (pVertex->inSurface() == TRUE))
            return FALSE;
        
        pVertex = pAdjNode->nodeData( 1 );
        if ((pVertex != pNode->nodeData( 0 )) &&
            (pVertex != pNode->nodeData( 1 )) &&
            (pVertex->inSurface() == TRUE))
            return FALSE;
    }// while

    return TRUE;
}

T3DFrontier *
T3DFlowSurface::checkFS ( T3DCellEdgeNode * pNode,
                          T3DFrontier * pFace,
                          int & func ) const
{
    if ((pNode == NULL) || (pFace == NULL))
    {
        cout << "T3DFlowSurface::checkFS (...) : got NULL arg" << endl;
        return NO_FACE;
    }// if

    T3DFrontier  *pPhi;
    TFloat        angle_f, angle_b;

    pPhi = NO_FACE;

    if ((pNode->phi_f() != NO_FACE) || (pNode->phi_b() != NO_FACE))
    {
        angle_f = 0;
        angle_b = 0;

        //
        // determine face with minimal angle to pFace
        //
        if (pNode->phi_f() != NO_FACE)
            angle_f = calcFaceAngle( pNode->cellEdge(), pFace, pNode->phi_f() );
        
        if (pNode->phi_b() != NO_FACE)
            angle_b = calcFaceAngle( pNode->cellEdge(), pFace, pNode->phi_b() );
        
        if ((angle_f >= angle_b) && (angle_f >= alpha))
        {
            pPhi = pNode->phi_f();
            func = PHI_F;
        }// if
        else if ((angle_b >= angle_f) && (angle_b >= alpha))
        {
            pPhi = pNode->phi_b();
            func = PHI_B;
        }// else if
    }// if
    else
    {
        uint                    i, noOfEdges;
        T3DCellEdge            *pCellEdge;
        T3DCellEdgeNode        *pAdjNode;
        TEdge                  *pEdge;
        
        //
        // determine face via phiHat
        //
        pPhi = pNode->phiHat( pFace );

        if (pPhi == NO_FACE)
            return NO_FACE;
        
        if (calcFaceAngle( pNode->cellEdge(), pFace, pPhi ) < alpha)
            return NO_FACE;

        func = PHI_P;

        //
        // build edges to succ with respect to found face
        //
        if (pPhi != NO_FACE)
        {
            pNode->getOutEdges().clear();
            
            noOfEdges = pPhi->getNoOfEdges();

            for (i = 0; i < noOfEdges; i++)
            {
                pCellEdge = pPhi->getEdge( i );

                if (pCellEdge == pNode->cellEdge())
                    continue;

                if ((pAdjNode = (T3DCellEdgeNode*) pCellEdge->data()) != NULL)
                {
                    pEdge = new TEdge( *pNode, *pAdjNode );
                    pNode->addOutEdge( *pEdge );
                }// if
            }// for
        }// if

        // clear facelist of node
        //        pNode->remFaceList();
    }// else

    return pPhi;
}

//
// add a node to actual flow surface
//
void
T3DFlowSurface::addFSNode ( T3DCellEdgeNode * pNode,
                            T3DFrontier * pPhi,
                            int func )
{
    T3DCellEdgeNode  *pAdjNode;
    
    if ((pNode == NULL) || (pPhi == NULL))
    {
        cout << "T3DFlowSurface::addFSNode (...) : got NULL arg" << endl;
        return;
    }// if

    //
    // add node to flow surface and set specific flags
    // to indicate, that this node is used and
    // increase facecounter for each succ/pred
    //
    
    if (verbosity > VERBOSE_RESULT)
    {
        cout << "T3DFlowSurface::addFSNode (...) : adding node "
             << pNode->getName();
        switch (func)
        {
            case PHI_F : cout << " via phi_f" << endl; break;
            case PHI_B : cout << " via phi_b" << endl; break;
            case PHI_P : cout << " via phi_p" << endl; break;
        }// switch
    }// if
    
    if ( ! pNode->isInSurface() )
    {
        pFlowSurface->add( *pNode );
        pNode->setInSurface( TRUE );
    }// if

    if (pNode->fsFace( TO_FACE ) == NO_FACE)
    {
        pNode->setFSFace( TO_FACE, pPhi );

#ifdef SET_ORIENT
        if (pNode->fsFace( FROM_FACE ) != NO_FACE)
        {
            T3DFaceOrient::setOrient( pPhi, pNode->fsFace( FROM_FACE ) );
//            T3DFaceOrient::showOrient( pPhi );
        }// if
#endif
    }// if
    else
    {
        pNode->setFSFace( FROM_FACE, pPhi );
        
#ifdef SET_ORIENT
        if (pNode->fsFace( TO_FACE ) != NO_FACE)
            T3DFaceOrient::setOrient( pPhi, pNode->fsFace( TO_FACE ) );
#endif        
    }// else
    
    pNode->setUsedFSFunc( func );
    pNode->addUsedFaces();
    pNode->nodeData( 0 )->setFlowSurface( pFlowSurface );
    pNode->nodeData( 1 )->setFlowSurface( pFlowSurface );

    if ((verbosity > VERBOSE_RESULT) && param.draw)
        TDrawApp::drawGraph( pFlowSurface, "flowsurface", VRML );
    
    TListIterator< TEdge >  iterEdges;
    TEdge                  *pEdge;

    if ((func == PHI_F) || (func == PHI_P))
        iterEdges.setList( pNode->getOutEdges() );
    else  // func == PHI_B
        iterEdges.setList( pNode->getInEdges() );
        
    while ((pEdge = iterEdges++) != NULL)
    {
        pAdjNode = (T3DCellEdgeNode*) pEdge->getOther( pNode );
        
        pAdjNode->nodeData( 0 )->setFlowSurface( pFlowSurface );
        pAdjNode->nodeData( 1 )->setFlowSurface( pFlowSurface );

        // put this node in the flowsurface because he is
        // connected via pPhi to a face in a flowsurface
        if ( ! pAdjNode->isInSurface() )
        {
            pFlowSurface->add( *pAdjNode );
            pAdjNode->setInSurface( TRUE );
        }// if

        if (pAdjNode->fsFace( FROM_FACE ) == NO_FACE)
            pAdjNode->setFSFace( FROM_FACE, pPhi );
        else if (pAdjNode->fsFace( TO_FACE ) == NO_FACE)
            pAdjNode->setFSFace( TO_FACE, pPhi );
        else
            cout << "T3DFlowSurface::addFSNode (...) : both faces set" << endl;

        pAdjNode->addUsedFaces();
    }// while
}

//
// check if given node is a suitable startnode
//
bool
T3DFlowSurface::isStartNode ( T3DCellEdgeNode * pNode ) const
{
    if (pNode == NULL)
    {
        cout << "T3DFlowSurface::isStartNode (...) : got NULL node" << endl;
        return FALSE;
    }// if
    
    if (pNode->noOfUsedFaces() >= 2)
        return FALSE;

    if (pNode->phi_f() == NO_FACE)
        return FALSE;
    
    if (pNode->getOutEdges().getSize() == 0)
        return FALSE;
    
    if (pNode->isDeleted() == TRUE)
        return FALSE;

    if (pNode->nodeData( 0 )->inSurface() == TRUE)
        return FALSE;

    if (pNode->nodeData( 1 )->inSurface() == TRUE)
        return FALSE;

    if (pNode->isWeak())
        return FALSE;
    
    return TRUE;
}

//
// virtually (un)delete given flow surface
//
void
T3DFlowSurface::deleteFS ( T3DCellEdgeGraph * pGraph )
{
    if (pGraph == NULL)
    {
        cout << "T3DFlowSurface::deleteFS () : got NULL graph" << endl;
        return;
    }// if

    TListIterator< TNode >  iterNodes;
    T3DCellEdgeNode        *pNode;

    iterNodes.setList( pGraph->getNodeList() );
    iterNodes.first();
    while ((pNode = (T3DCellEdgeNode*) iterNodes++) != NULL)
    {
        pNode->setDeleted( TRUE );
    }// while
}

void
T3DFlowSurface::unDeleteFS ( T3DCellEdgeGraph * pGraph )
{
    if (pGraph == NULL)
    {
        cout << "T3DFlowSurface::unDeleteFS () : got NULL graph" << endl;
        return;
    }// if

    TListIterator< TNode >  iterNodes;
    T3DCellEdgeNode        *pNode;

    iterNodes.setList( pGraph->getNodeList() );
    iterNodes.first();
    while ((pNode = (T3DCellEdgeNode*) iterNodes++) != NULL)
    {
        pNode->clearFSFaces();
        pNode->setUsedFSFunc( NO_FUNC );
        pNode->setNoOfUsedFaces( 0 );
        pNode->setInSurface( FALSE );
        pNode->nodeData( 0 )->setFlowSurface( NULL );
        pNode->nodeData( 1 )->setFlowSurface( NULL );
    }// while
}

//
// determine number of used (in a flowsurface) celledges
// of given face
//
uint
T3DFlowSurface::noOfUsedEdges ( T3DFrontier * pFace )
{
    if (pFace == NULL)
    {
        cout << "T3DFlowSurface::noOfUsedEdges (...) : got NULL face" << endl;
        return 0;
    }// if

    //
    // only works correctly for triangle !!!
    //
    
    uint              noOfUsed = 0;
    T3DCellEdgeNode * pNode;
    
    for (uint i = 0; i < 3; i++)
    {
        if ((pNode = (T3DCellEdgeNode*) pFace->getEdge(i)->data()) == NULL)
            continue;
        
        if (pNode->noOfUsedFaces() > 0)
            noOfUsed++;
    }// for

    return noOfUsed;
}

//
// delete all edges from/to node
//
void
T3DFlowSurface::delEdges ( T3DCellEdgeNode * pNode )
{
    return;
    
    if (pNode != NULL)
    {
        pNode->getOutEdges().clear();
        pNode->getInEdges().clear();
        pNode->remFaceList();
    }// if
    else
        cout << "T3DFlowSurface::delEdges (...) : got NULL arg" << endl;
}

//
// order given list of flowsurfaces
//
void
T3DFlowSurface::orderFS ( T3DCellEdgeGraph * pEdgeGraph,
                          TList< T3DCellEdgeGraph > * pListFS )
{
    if ((pEdgeGraph == NULL) || (pListFS == NULL))
    {
        cout << "T3DFlowSurface::orderFS (...) : got NULL arg" << endl;
        return;
    }// if

    //
    // create a graph G of flowsurfaces defined by:
    // V(G) = list of flowsurfaces
    // E(G) = { (f_1, f_2) : exist (weak?) edge in net between f_1 and f_2 }
    //
    // apply ordering to this graph and sort fs-list accord.
    //

    uint    i;
    TGraph  fsGraph;

    //
    // create nodes in fs-graph
    //

    TListIterator< T3DCellEdgeGraph >  iterFS( * pListFS );
    T3DCellEdgeGraph                 * pFS;
    TFSNode                          * pFSNode;

    i = 0;
    iterFS.first();
    while ((pFS = iterFS++) != NULL)
    {
        pFSNode = new TFSNode( i++, pFS );
        pFS->setFSNode( pFSNode );

        fsGraph.add( * pFSNode );
    }// while
    
    //
    // create edges in fs-graph
    //

    TListIterator< TNode >  iterNodes;
    T3DCellEdgeNode       * pEdgeNode;
    T3DVertexData         * pNode0, * pNode1;
    T3DCellEdgeGraph      * pFS0, * pFS1;
    TFSNode               * pFSNode0, * pFSNode1;
    TEdge                 * pEdge;

    iterNodes.setList( pEdgeGraph->getNodeList() );
    iterNodes.first();
    while ((pEdgeNode = (T3DCellEdgeNode *) iterNodes++) != NULL)
    {
        if (pEdgeNode->isInSurface() && pEdgeNode->isWeak())
            continue;

        pNode0 = pEdgeNode->nodeData( 0 );
        pNode1 = pEdgeNode->nodeData( 1 );

        pFS0 = pNode0->flowSurface();
        pFS1 = pNode1->flowSurface();
        
        if ((pFS0 == pFS1) || (pFS0 == NULL) || (pFS1 == NULL))
            continue;

        //
        // create edge between fs0 and fs1 if no such
        // edge exists
        //

        pFSNode0 = pFS0->fsNode();
        pFSNode1 = pFS1->fsNode();

        if (( ! pFSNode0->isSucc( * pFSNode1 ) ) &&
            ( ! pFSNode1->isSucc( * pFSNode0 ) ))
        {
           pEdge = new TEdge( * pFSNode0, * pFSNode1 );
           
           pFSNode0->addOutEdge( * pEdge );
           pFSNode1->addInEdge( * pEdge );
        }// if
    }// while

    cout << "T3DFlowSurface::orderFS (...) : fsGraph = " << endl
         << fsGraph << endl;
}
