//
//  Project     : libgraph
//  File        : T3DPermBuilder.cc
//  Author      : Ronald Kriemann
//  Purpose     : class to handle building of permutations in 3D
//
// arch-tag: 75b5a3aa-5c2a-49ef-b4a5-30c38f2fc177
//

extern "C" {
#include <stdio.h>
}

#include "defines.hh"
#include "var.hh"
#include "cmdline.hh"
#include "timer.hh"
#include "list.hh"

#include "net.hh"
#include "nethierarchy.hh"

#include "matrixblock.hh"

#include "TConvEqn.hh"

#include "drawapp.hh"

#include "graphtool.hh"
#include "rcm.hh"
#include "fvs.hh"
#include "heuristicFVS.hh"
#include "planarBlock.hh"

#include "T3DCellEdgeGraph.hh"
#include "T3DCellEdgeNode.hh"
#include "T3DFlowSurface.hh"
#include "T3DSurfaceGraph.hh"
#include "T3DGraph.hh"
#include "TRCMGraph.hh"
#include "TMatrixGraph.hh"

#include "T3DPermBuilder.hh"

////////////////////////////////////////////
//
// constructor and destructor
//

T3DPermBuilder::T3DPermBuilder ()
{
    orderType     = NONE;
    multiGridSort = MGSORT_INDEP;
}

T3DPermBuilder::T3DPermBuilder ( uint type, uint mgSort )
{
    orderType     = type;
    multiGridSort = mgSort;
}

T3DPermBuilder::~T3DPermBuilder ()
{
}

////////////////////////////////////////////
//
// access member variables
//

TDynArray< uint > *
T3DPermBuilder::blockStructure ()
{
    return & arrBlockStruct;
}
    
////////////////////////////////////////////
//
// basic methods for setting up permutations
//

//
// initialize var. and perm.
//
void
T3DPermBuilder::init ( uint size )
{
    lastF_CNode = 0;
    firstLNode  = 0;
    lastOther   = 0;

    permF.setSize( 0 );
    permL.setSize( 0 );
    permOther.setSize( 0 );

    permF.setGrowBy( 128 );
    permL.setGrowBy( 128 );
    permOther.setGrowBy( 128 );

    arrBlockStruct.setSize( 0 );
}

#define ORDER_FIRST

//
// add FVS to permutation
//
void
T3DPermBuilder::addFVS ( TList< TNode > & fvs )
{
    TListIterator< TNode >   iterNodes( fvs );
    TNode                  * pNode;

#ifdef ORDER_FIRST
    permF.setSize( lastF_CNode + fvs.getSize() );
    
    iterNodes.first();
    while ((pNode = iterNodes++) != NULL)
    {
        permF[ lastF_CNode++ ] = pNode->getName();
    }// while
#else
    permL.setSize( firstLNode + fvs.getSize() );
    
    iterNodes.last();
    while ((pNode = iterNodes--) != NULL)
    {
        permL[ firstLNode++ ] = pNode->getName();
    }// while
#endif
}

//
// add F,L,C or other node(s) to permutation
//
void
T3DPermBuilder::addFNode ( TNode & node )
{
    permF.setSize( lastF_CNode + 1 );

    permF[ lastF_CNode++ ] = node.getName();
}

void
T3DPermBuilder::addLNode ( TNode & node )
{
    permL.setSize( firstLNode + 1 );
    
    permL[ firstLNode++ ] = node.getName();
}

void
T3DPermBuilder::addCycle ( TGraph & cycle )
{
    TListIterator< TNode >   iterNodes( cycle.getNodeList() );
    TNode                  * pNode;

    permF.setSize( lastF_CNode + cycle.getNoOfNodes() );

    iterNodes.first();
    while ((pNode = iterNodes++) != NULL)
    {
        permF[ lastF_CNode++ ] = pNode->getName();
    }// while
}

void
T3DPermBuilder::addOther ( TNode & node )
{
    permOther.setSize( lastOther + 1 );

    permOther[lastOther++] = node.getName();
}

//
// assemble permutation
//
void
T3DPermBuilder::buildPerm ( TPermutation * pPerm )
{
    if (pPerm == NULL)
    {
        cout << "T3DPermBuilder::buildPerm (...) : got NULL arg" << endl;
        return;
    }// if

    uint  i, j;
    uint  max;
    uint  size = lastF_CNode + firstLNode + lastOther;

    pPerm->setSize( size );

    for (i = 0; i < lastF_CNode; i++)
        (*pPerm)[i] = permF[i];

    max = lastF_CNode + firstLNode;
    for (i = lastF_CNode, j = firstLNode - 1; i < max; i++, j--)
        (*pPerm)[i] = permL[j];

    max = size;
    for (i = lastF_CNode + firstLNode, j = 0; i < max; i++, j++)
        (*pPerm)[i] = permOther[j];
}

////////////////////////////////////////////
//
// various high level methods for perm building
//

//
// build permutation via concentric cycles
//
TPermutation *
T3DPermBuilder::buildCYCPerm ( TNet<3> * pNet,
                               TConvEqn<3> * pEqn,
                               TMatrixBlock * pMatrix )
{
    if ((pNet == NULL) || (pEqn == NULL) || (pMatrix == NULL))
    {
        cout << "T3DPermBuilder::buildCYCPerm (...) : got NULL args" << endl;
        return NULL;
    }// if


    TTimer  timer;
    
    
    //
    // build graph consisting of cell edges
    //
    T3DCellEdgeGraph  * pEdgeGraph;

    pEdgeGraph = buildEdgeGraph( pNet, pEqn, pMatrix );
    
    // care about special cases
    if (pEdgeGraph->getNoOfNodes() == 0)
    {
        // don't give a net, because there are problems with
        // empty edegraphs
        return buildFVSPerm( NULL, pMatrix );
    }// if
    
    //
    // build flow surfaces
    //
    TList< T3DCellEdgeGraph >         * pListFS;
    TListIterator< T3DCellEdgeGraph >   iterFS;
    T3DCellEdgeGraph                  * pFSGraph;


    pListFS = buildFS( pEdgeGraph );

    iterFS.setList( *pListFS );

    //
    // build surface graphs
    //
    TList< T3DSurfaceGraph >          * pListSG;
    TListIterator< T3DSurfaceGraph >    iterSG;
    T3DSurfaceGraph                   * pSurfaceGraph;

    pListSG = buildSurfaceGraphs( pListFS, pMatrix, TRUE );

    iterSG.setList( * pListSG );

    //
    // build permutation
    //
    uint                     i;
    TPlanarBlock             planarBlock;
    TGraphTool               graphTool;
    TList< TGraph >        * pListCycles;
    TListIterator< TGraph >  iterCycles;
    TGraph                 * pCycle;
    TPermutation           * pPermutation, localPerm;
    T3DPermBuilder           localPermBuilder;

    pPermutation = new TPermutation;

    i = 0;
    
    iterSG.first();
    while ((pSurfaceGraph = iterSG++) != NULL)
    {
        // update blockstructure
        addBlock( pSurfaceGraph->getNoOfNodes() );

        localPermBuilder.init( 0 );
        localPerm.setSize( 0 );
    
        timer.start();
        
        pListCycles = planarBlock.buildCycles( pSurfaceGraph, MIN_CYC,
                                               & localPermBuilder );

        timer.stop();
        if (verbosity >= VERBOSE_TIME)
            cout << "T3DPermBuilder::buildCYCPerm (...) : time for building cycles = "
                 << timer << endl;
        
        if (pListCycles == NULL)
        {
            cout << "T3DPermBuilder::buildCYCPerm (...) : got NULL cyclelist" << endl;
            continue;
        }// if

        if (verbosity >= VERBOSE_STEP)
            cout << "T3DPermBuilder::buildCYCPerm (...) : number of cycles = "
                 << pListCycles->getSize() << endl;
        
        if ((param.draw) && (verbosity >= VERBOSE_RESULT))
        {
            char  filename[256];
            
            sprintf( filename, "cycles%d", i++ );
            TDrawApp::drawCycles( pListCycles, filename );
        }// if

        iterCycles.setList( * pListCycles );
        iterCycles.first();
        while ((pCycle = iterCycles++) != NULL)
        {
            delete pCycle;
        }// while

        delete pListCycles;

        //
        // put local perm into global perm
        //
        localPermBuilder.buildPerm( & localPerm );
        concatPerm( pPermutation, & localPerm );
    }// while

    //
    // add all nodes, which do not belong to a FS to permutation
    //
    orderNonFSNodes( pNet, pEdgeGraph, pMatrix, & localPermBuilder );

    localPerm.setSize( 0 );
    localPermBuilder.buildPerm( & localPerm );
    concatPerm( pPermutation, & localPerm );
    
    //
    // clean up
    //
    iterSG.first();
    while ((pSurfaceGraph = iterSG++) != NULL)
    {
        delete pSurfaceGraph;
    }// while
    
    iterFS.first();
    while ((pFSGraph = iterFS++) != NULL)
    {
        delete pFSGraph;
    }// while

    delete pListSG;
    delete pListFS;
    delete pEdgeGraph;

    return pPermutation;
}

//
// build permutation via heuristic fvs
//
TPermutation *
T3DPermBuilder::buildFVSPerm ( TNet<3>      * pNet,
                               TConvEqn<3>  * pEqn,
                               TMatrixBlock * pMatrix,
                               uint           fvsMethod )
{
    if ((pNet == NULL) || (pEqn == NULL) || (pMatrix == NULL))
    {
        cout << "T3DPermBuilder::buildFVSPerm (...) : got NULL args" << endl;
        return NULL;
    }// if

    //
    // build graph consisting of cell edges
    //
    T3DCellEdgeGraph  * pEdgeGraph;

    pEdgeGraph = buildEdgeGraph( pNet, pEqn, pMatrix );

    // care about special cases
    if (pEdgeGraph->getNoOfNodes() == 0)
    {
        // don't give a net, because there are problems with
        // empty edegraphs
        return buildFVSPerm( NULL, pMatrix );
    }// if
    
    //
    // build flow surfaces
    //
    TList< T3DCellEdgeGraph >         * pListFS;
    TListIterator< T3DCellEdgeGraph >   iterFS;
    T3DCellEdgeGraph                  * pFSGraph;

    pListFS = buildFS( pEdgeGraph );

    iterFS.setList( *pListFS );

    //
    // build surface graphs
    //
    TList< T3DSurfaceGraph >          * pListSG;
    TListIterator< T3DSurfaceGraph >    iterSG;
    T3DSurfaceGraph                   * pSurfaceGraph;

    if (fvsMethod == HEURISTIC_FVS)
        pListSG = buildSurfaceGraphs( pListFS, pMatrix );
    else
        pListSG = buildSurfaceGraphs( pListFS, pMatrix, TRUE );

    iterSG.setList( * pListSG );

//     cout << "T3DPermBuilder::buildFVSPerm (...) : used memory = " 
//          << TMemory::used() << " kB" << endl;

    //
    // build permutation
    //
    TGraphTool               graphTool;
    TList< TNode >         * pListFVS;
    TListIterator< TNode >   iterFVS;
    TPermutation           * pPermutation, localPerm;
    T3DPermBuilder           localPermBuilder;
    uint                     sizeOfFVS;

    pPermutation = new TPermutation;

    sizeOfFVS = 0;
    
    iterSG.first();
    while ((pSurfaceGraph = iterSG++) != NULL)
    {
        THeuristicFVS    heuristicFVS;
        TPlanarFVS       planarFVS;
        
        // update blockstructure
        addBlock( pSurfaceGraph->getNoOfNodes() );

        // build fvs of surfacegraph
        if (fvsMethod == HEURISTIC_FVS)
            pListFVS = heuristicFVS.buildFVS( * pSurfaceGraph );
        else
            pListFVS = planarFVS.buildFVS( * pSurfaceGraph );

        if (pListFVS == NULL)
        {
            cout << "T3DPermBuilder::buildFVSPerm (...) : got NULL fvs" << endl;
            continue;
        }// if

        if (verbosity >= VERBOSE_STEP)
            cout << "T3DPermBuilder::buildFVSPerm (...) : number of fvs-nodes = "
                 << pListFVS->getSize() << endl;

        sizeOfFVS += pListFVS->getSize();
        
        localPermBuilder.init( 0 );
        localPerm.setSize( 0 );
    
        // add fvs to permutation
        localPermBuilder.addFVS( * pListFVS );

        removeFVS( pListFVS, pSurfaceGraph );

        checkCyclic( pSurfaceGraph );
        
        // add nodes in acyclic graph to permutation
        graphTool.numberNodes( * pSurfaceGraph, & localPermBuilder, FALSE );

        if (param.draw && (verbosity >= VERBOSE_RESULT))
        {
            TDrawApp::drawGraph( pSurfaceGraph, "acyclic" );
        }// if

        delete pListFVS;

        //
        // put local perm into global perm
        //
        localPermBuilder.buildPerm( & localPerm );
        concatPerm( pPermutation, & localPerm );
    }// while

    cout << "T3DPermBuilder::buildFVSPerm (...) : total number of fvs-nodes = "
         << sizeOfFVS << endl;
    
    //
    // add all nodes, which do not belong to a FS to permutation
    //
    orderNonFSNodes( pNet, pEdgeGraph, pMatrix, & localPermBuilder );

    localPerm.setSize( 0 );
    localPermBuilder.buildPerm( & localPerm );
    concatPerm( pPermutation, & localPerm );

    if (param.draw && (verbosity >= VERBOSE_RESULT))
    {
        TPermutation  invPerm( pPermutation->invert() );
        
        TDrawApp::drawGraph( pListSG, & invPerm, "surfacegraphs" );
    }// if
                             
    //
    // clean up
    //
    iterSG.first();
    while ((pSurfaceGraph = iterSG++) != NULL)
    {
        delete pSurfaceGraph;
    }// while
    
    iterFS.first();
    while ((pFSGraph = iterFS++) != NULL)
    {
        delete pFSGraph;
    }// while

    delete pListSG;
    delete pListFS;
    delete pEdgeGraph;

    return pPermutation;
}

TPermutation *
T3DPermBuilder::buildFVSPerm ( TMatrixBlock * pMatrix )
{
    return buildFVSPerm( NULL, pMatrix );
}

TPermutation *
T3DPermBuilder::buildFVSPerm ( TNet<3> * pNet,
                               TMatrixBlock * pMatrix )
{
    if (pMatrix == NULL)
    {
        cout << "T3DPermBuilder::buildFVSPerm (...) : got NULL args" << endl;
        return NULL;
    }// if

    TTimer  timer;
    
    //
    // build netgraph
    //
    TGraph      * pGraph;
    TGraphTool    graphTool;

    timer.start();
    if (pNet != NULL)
        pGraph = new T3DGraph( pNet, pMatrix, param.criterion, param.xi );
    else
        pGraph = new TMatrixGraph( pMatrix, param.criterion, param.xi );
    timer.stop();
    cout << "T3DPermBuilder::buildFVSPerm (...) : time to build graph = " << timer << endl;
    
    //
    // build permutation
    //
    THeuristicFVS            fvs;
    TList< TNode >         * pListFVS;

    init( 0 );
    
    pListFVS = fvs.buildFVS( * pGraph );

    if (param.draw && (verbosity >= VERBOSE_RESULT) && (pNet != NULL))
    {
        graphTool.numberNodes( * pGraph );
        TDrawApp::drawGraph( (T3DGraph*) pGraph, "netgraph" );
    }// if

    if (pListFVS == NULL)
    {
        cout << "T3DPermBuilder::buildFVSPerm (...) : got NULL fvs" << endl;
        return NULL;
    }// if

//    if (verbosity >= VERBOSE_STEP)
        cout << "T3DPermBuilder::buildFVSPerm (...) : number of fvs-nodes = "
             << pListFVS->getSize() << endl;
        
    // add fvs to permutation
    addFVS( * pListFVS );

    // remove fvs from graph
    removeFVS( pListFVS, pGraph );
    
    // add nodes in acyclic graph to permutation
    graphTool.numberNodes( * pGraph, this, FALSE );
    
    if (param.draw && (verbosity >= VERBOSE_RESULT) && (pNet != NULL))
        TDrawApp::drawGraph( (T3DGraph*) pGraph, "acyclic" );

    delete pListFVS;

    //
    // clean up
    //
    
    delete pGraph;

    //
    // finally build permutation and return it
    //
    TPermutation  * pPermutation;

    pPermutation = new TPermutation;
    buildPerm( pPermutation );
    
    if (param.draw && (verbosity >= VERBOSE_RESULT))
    {
        T3DGraph tmpGraph( pNet, pMatrix, param.criterion, param.xi );
        TPermutation invPerm( pPermutation->invert() );
        
        graphTool.numberNodes( tmpGraph );
        TDrawApp::drawGraph( & tmpGraph, & invPerm, "pnetgraph" );
    }// if

    return pPermutation;
}

//
// build permutation via rcm
//
TPermutation *
T3DPermBuilder::buildRCMPerm ( TMatrixBlock * pMatrix )
{
    if (pMatrix == NULL)
    {
        cout << "T3DPermBuilder::buildRCMPerm (...) : got NULL arg" << endl;
        return NULL;
    }// if

    //
    // do rcm with matrixgraph
    //
	TTimer  timer;
	TRCM    rcm;

	TRCMGraph graph( pMatrix );
	
	timer.start();
    
	rcm.rcm( graph, this );
    
	timer.stop();

	if (verbosity >= VERBOSE_TIME)
		cout << "T3DPermBuilder::buildRCMPerm (...) : time for rcm = "
             << timer << endl;

    //
    // build permutation and return it
    //
    TPermutation  * pPermutation;

    pPermutation = new TPermutation;
	buildPerm( pPermutation );

    return pPermutation;
}
    
////////////////////////////////////////////
//
// various high level methods for perm building
// - multi net
// - matrix and perm arrays must be initialized
//

void
T3DPermBuilder::buildMGCYCPerm ( TNetHierarchy<3> * pNetHier,
                                 TConvEqn<3>      * pEqn,
                                 TBlockMatrix    ** pMatrix,
                                 TPermutation    ** pPermHier,
                                 uint               startLevel,
                                 uint               endLevel )
{
    if ((pNetHier == NULL) || (pEqn == NULL))
    {
        cout << "T3DPermBuilder::buildMGCYCPerm (...) : got NULL arg" << endl;
        return;
    }// if

    TTimer                    timer;
    uint                      i;
    TListIterator< TNet<3> >  iterNet;
    TNet<3>                 * pNet;
    T3DPermBuilder            localPermBuilder;
        
    //
    // build permutation independently for every net
    //
    i = 0;

    iterNet.setList( pNetHier->getList() );
    iterNet.first();

	// in Stokes:
    if ( pMatrix[startLevel]->getNoOfRows() > 1 ) iterNet++;

    while ((pNet = iterNet++) != NULL)
    {
        if ((i < startLevel) || (i > endLevel))
        {
            i++;
            continue;
        }// if

        timer.start();
        pPermHier[i] = localPermBuilder.buildCYCPerm( pNet, pEqn,
                                                      & (*pMatrix[i])[0][0] );
        timer.stop();
        cout << "T3DPermBuilder::buildMGCYCPerm (...) : time for cyc on level "
             << i << " = " << timer << endl;

        i++;
    }// while

    if (startLevel == endLevel)
        arrBlockStruct = * (localPermBuilder.blockStructure());
}

void
T3DPermBuilder::buildMGFVSPerm ( TNetHierarchy<3> * pNetHier,
                                 TConvEqn<3>      * pEqn,
                                 TBlockMatrix    ** pMatrix,
                                 TPermutation    ** pPermHier,
                                 uint               startLevel,
                                 uint               endLevel,
                                 uint               fvsMethod )
{
    if ((pNetHier == NULL) || (pEqn == NULL))
    {
        cout << "T3DPermBuilder::buildMGFVSPerm (...) : got NULL arg" << endl;
        return;
    }// if

    TTimer                    timer;
    uint                      i;
    TListIterator< TNet<3> >  iterNet;
    TNet<3>                 * pNet;
        
    //
    // build permutation independently for every net
    //
    i = 0;

    iterNet.setList( pNetHier->getList() );
    iterNet.first();
	// in Stokes:
    if ( pMatrix[startLevel]->getNoOfRows() > 1 ) iterNet++;

    while ((pNet = iterNet++) != NULL)
    {
        if ((i < startLevel) || (i > endLevel))
        {
            i++;
            continue;
        }// if

        T3DPermBuilder  localPermBuilder;
        
        timer.start();
        pPermHier[i] = localPermBuilder.buildFVSPerm( pNet, pEqn,
                                                      & (*pMatrix[i])[0][0],
                                                      fvsMethod );
        timer.stop();
        cout << "T3DPermBuilder::buildMGFVSPerm (...) : time for fvs on level "
             << i << " = " << timer << endl;

        if (startLevel == endLevel)
            arrBlockStruct = * (localPermBuilder.blockStructure());
        
        i++;
    }// while
}
    
void
T3DPermBuilder::buildMGFVSPerm ( TBlockMatrix    ** pMatrix,
                                 TPermutation    ** pPermHier,
                                 uint               startLevel,
                                 uint               endLevel )
{
    TTimer          timer;
    uint            i;
    T3DPermBuilder  localPermBuilder;
        
    //
    // build permutation independently for every net
    //

    for ( i = startLevel; i <= endLevel; i++ )
    {
        timer.start();
        pPermHier[i] = localPermBuilder.buildFVSPerm( & (*pMatrix[i])[0][0] );
        timer.stop();
        cout << "T3DPermBuilder::buildMGFVSPerm : time for fvs on level "
             << i << " = " << timer << endl;
    }// while
}

void
T3DPermBuilder::buildMGRCMPerm ( TBlockMatrix    ** pMatrix,
                                 TPermutation    ** pPermHier,
                                 uint               startLevel,
                                 uint               endLevel )
{
    TTimer          timer;
    uint            i;
    T3DPermBuilder  localPermBuilder;
        
    //
    // build permutation independently for every net
    //

    for ( i = startLevel; i <= endLevel; i++ )
    {
        timer.start();
        pPermHier[i] = localPermBuilder.buildRCMPerm( & (*pMatrix[i])[0][0] );
        timer.stop();
        cout << "T3DPermBuilder::buildMGRCMPerm : time for rcm on level "
             << i << " = " << timer << endl;
    }// while
}

////////////////////////////////////////////
//
// encapsulate various steps
//

//
// build graph consisting of cell edges
//
T3DCellEdgeGraph *
T3DPermBuilder::buildEdgeGraph ( TNet<3> * pNet,
                                 TConvEqn<3> * pEqn,
                                 TMatrixBlock * pMatrix )
{
    TTimer              timer;
    T3DCellEdgeGraph  * pEdgeGraph;

    timer.start();
    pEdgeGraph = new T3DCellEdgeGraph( pNet, pEqn, pMatrix,
                                       param.fsKappa, param.sigma, param.fsGamma );
    timer.stop();
    if (verbosity >= VERBOSE_TIME)
        cout << "T3DPermBuilder::buildEdgeGraph (...) : time for building edge graph = "
             << timer << endl;

    return pEdgeGraph;
}

//
// build flow surfaces
//
TList< T3DCellEdgeGraph > *
T3DPermBuilder::buildFS ( T3DCellEdgeGraph * pEdgeGraph )
{
    uint                                i;
    TTimer                              timer;
    
    T3DFlowSurface                      flowSurface;
    TList< T3DCellEdgeGraph >         * pListFS;
    TListIterator< T3DCellEdgeGraph >   iterFS;
    T3DCellEdgeGraph                  * pFSGraph;


    timer.start();

    if (param.searchType == SEARCH_DFS)
        pListFS = flowSurface.buildFlowSurfaces( pEdgeGraph, param.fsAlpha,
                                                 param.fsBeta, SEARCH_DFS );
    else
        pListFS = flowSurface.buildFlowSurfaces( pEdgeGraph, param.fsAlpha,
                                                 param.fsBeta, SEARCH_BFS );

    timer.stop();
    if (verbosity >= VERBOSE_TIME)
        cout << "T3DPermBuilder::buildFS (...) : time for building flow surfaces = "
             << timer << endl;

    if (pListFS == NULL)
    {
        cout << "T3DPermBuilder::buildFS (...) : have NULL fs list" << endl;
        exit( 1 );
    }// if
    
    iterFS.setList( *pListFS );

    //
    // print results
    //
    if (verbosity >= VERBOSE_STEP)
    {
        cout << "T3DPermBuilder::buildFVSPerm (...) : detected "
             << pListFS->getSize() << " flow surfaces :" << endl;
        
        i = 0;
        iterFS.first();
        while ((pFSGraph = iterFS++) != NULL)
            cout << "T3DPermBuilder::buildFVSPerm (...) : fs" << i++
                 << " has size : " << pFSGraph->getNoOfNodes() << endl;
    }// if

    if ((param.draw) && (verbosity >= VERBOSE_RESULT))
    {
        char  szFilename[256];

        TDrawApp::drawFlowSurfaces( pListFS, "flowgraph" );
        
        i = 0;
        iterFS.first();
        while ((pFSGraph = iterFS++) != NULL)
        {
            sprintf( szFilename, "flow%d", i++ );
            TDrawApp::drawGraph( pFSGraph, szFilename );
        }// while
    }// if

    return pListFS;
}

//
// build surface graphs
//
TList< T3DSurfaceGraph > *
T3DPermBuilder::buildSurfaceGraphs ( TList< T3DCellEdgeGraph > * pListFS,
                                     TMatrixBlock * pMatrix, bool bSortOFD )
{
    uint                                i;
    TTimer                              timer;
    char                                szFilename[256];
    
    TListIterator< T3DCellEdgeGraph >   iterFS;
    T3DCellEdgeGraph                  * pFlowSurface;

    TList< T3DSurfaceGraph >          * pListGraph;
    T3DSurfaceGraph                   * pSurfaceGraph;

    TGraphTool                          graphTool;

    pListGraph = new TList< T3DSurfaceGraph >;

    timer.start();

    i = 0;
    
    iterFS.setList( * pListFS );
    iterFS.first();
    while ((pFlowSurface = iterFS++) != NULL)
    {
        pSurfaceGraph = new T3DSurfaceGraph( pFlowSurface, pMatrix,
                                             param.criterion, param.xi,
                                             bSortOFD );

        if (pSurfaceGraph != NULL)
        {
            if (verbosity >= VERBOSE_STEP)
                cout << "T3DPermBuilder::buildSurfaceGraphs (...) : "
                     << "sg" << i << " has "
                     << pSurfaceGraph->getNoOfNodes() << " nodes" << endl;

            pListGraph->add( * pSurfaceGraph );

            if ((verbosity >= VERBOSE_RESULT) && (param.draw))
            {
                graphTool.numberNodes( * pSurfaceGraph );
                
                sprintf( szFilename, "sgraph%d", i++ );
                TDrawApp::drawGraph( pSurfaceGraph, szFilename );
            }// if
            else 
                i++;
        }// if
        else
            cout << "T3DPermBuilder::buildSurfaceGraphs (...) : have NULL surface graph"
                 << endl;
    }// while

    timer.stop();
    if (verbosity >= VERBOSE_TIME)
        cout << "T3DPermBuilder::buildSurfaceGraphs (...) : time for building surface graphs = "
             << timer << endl;
    
    return pListGraph;
}

//
// care about edges not in a flowsurface
//
void
T3DPermBuilder::orderNonFSNodes ( TNet<3> * pNet,
                                  T3DCellEdgeGraph * pEdgeGraph,
                                  TMatrixBlock * pMatrix,
                                  T3DPermBuilder * pPermBuilder )
{
    if ((pNet == NULL) || (pEdgeGraph == NULL) || (pMatrix == NULL))
    {
        cout << "T3DPermBuilder::orderNonFSNodes (...) : got NULL arg" << endl;
        return;
    }// if

    //
    // build a graph of all non-fs nodes
    //

    T3DGraph                 nonFSGraph( pNet, pEdgeGraph,
                                         pMatrix, param.criterion, param.xi );
    THeuristicFVS            fvs;
    TList< TNode >         * pListFVS;
    TListIterator< TNode >   iterFVS;
    TGraphTool               graphTool;

    if (verbosity >= VERBOSE_STEP)
        cout << "T3DPermBuilder::orderNonFSNodes (...) : number of non-FS nodes = "
             << nonFSGraph.getNoOfNodes() << endl;

    if ((verbosity >= VERBOSE_RESULT) && param.draw)
    {
        graphTool.numberNodes( nonFSGraph );
        TDrawApp::drawGraph( & nonFSGraph, "nonfsgraph" );
    }// if

    //
    // update blockstructure
    //
    addBlock( nonFSGraph.getNoOfNodes() );
    
    //
    // build a fvs in non-fs graph
    //
    pListFVS = fvs.buildFVS( nonFSGraph );

    if (pListFVS == NULL)
    {
        cout << "T3DPermBuilder::orderNonFSNodes (...) : got NULL fvs" << endl;
        return;
    }// if

    if (pPermBuilder != NULL)
    {
        pPermBuilder->init( 0 );
        pPermBuilder->addFVS( * pListFVS );
    }// if
    else
        addFVS( * pListFVS );

    removeFVS( pListFVS, & nonFSGraph );

    //
    // order nodes in non-fs graph
    //
    if (pPermBuilder != NULL)
        graphTool.numberNodes( nonFSGraph, pPermBuilder, FALSE );
    else
        graphTool.numberNodes( nonFSGraph, this, FALSE );

    //
    // clean up
    //

    delete pListFVS;
}

//
// put perm2 at bottom of perm1
//
void
T3DPermBuilder::concatPerm ( TPermutation * pPerm1,
                             TPermutation * pPerm2 ) const
{
    if ((pPerm1 == NULL) || (pPerm2 == NULL))
    {
        cout << "T3DPermBuilder::concatPerm (...) : got NULL arg" << endl;
        return;
    }// if

    uint  i, j;
    uint  size1, size2;

    size1 = pPerm1->getSize();
    size2 = pPerm2->getSize();

    pPerm1->setSize( size1 + size2 );

    for (i = 0, j = size1; i < size2; i++, j++ )
    {
        (*pPerm1)[j] = (*pPerm2)[i];
    }// for
}

//
// add a block to blockstructure
//
void
T3DPermBuilder::addBlock ( uint size )
{
    if (size > 0)
    {
        arrBlockStruct.setSize( arrBlockStruct.getSize() + 1 );
        arrBlockStruct[ arrBlockStruct.getSize() - 1 ] = size;
    }// if
}

//
// check if given graph is cyclic
//
void
T3DPermBuilder::checkCyclic( TGraph * pGraph )
{
    if (pGraph == NULL)
    {
        cout << "T3DPermBuilder::checkCyclic (...) : got NULL arg" << endl;
        return;
    }// if

    TGraphTool              graphTool;
    TListIterator< TNode >  iterNodes;
    TMarkableNode         * pNode;
    
    graphTool.numberNodes( * pGraph );

    iterNodes.setList( pGraph->getNodeList() );
    iterNodes.first();
    while ((pNode = (TMarkableNode*) iterNodes++) != NULL)
    {
        if (pNode->mark() == C)
        {
            cout << "T3DPermBuilder::checkCyclic (...) : graph has cycles (node = "
                 << pNode->getName() << ")" << endl;
            return;
        }// if
    }// while
}
