//
//  Project     : libgraph
//  File        : TRedBlack.cc
//  Author      : Thomas Probst
//  Purpose     : class to build red-black permutation
//
// arch-tag: efb87467-8afc-4260-8bdb-b974e88dd18f
//

#include "TRedBlack.hh"

// used marks
#define VISITED      9

// use pInfo-Pointer in basenode for an integer Pointer
// for an esier access: 
#define GETINFO(node) (*((int*)(node)->getInfo()))  

/////////////////////////////////////////
//
// constructor and destructor
//

TRedBlack::TRedBlack ()
{
}

TRedBlack::~TRedBlack ()
{
}

/////////////////////////////////////////
//
// methods for permutation building
//

//
// determine Red-Black-Ordering
//
TPermutation *
TRedBlack::orderBFS ( TGraph & graph, uint & noOfReds )
{
    TListIterator<TNode>   iter;

    TListIterator<TEdge>   iterEdge;
    TMarkableNode        * pNode;
    TList<TMarkableNode>   listNodes;
    uint                  noOfBlacks = 0;
    int                    mark;

    noOfReds = 0;
    
    iter.setList( graph.getNodeList() );
    iter.first();

    bfs( graph );

    while ((pNode = (TMarkableNode *) iter++) != NULL)
    {
        mark = pNode->mark();
        
        if (mark == R)  
            noOfReds++;      
        else
        {
            if (mark == B)
                noOfBlacks++; 
            else
            {
                cerr << "TRedBlack::orderBFS (...) : node " << pNode->getName()
                     << " has no valid mark (Red or Black)" << endl;
                exit(-1);
            }// else
        }// else
    }// while

    //
    // create permutation vector
    //

    TPermutation  * pPerm = new TPermutation;

    uint reds   = 0;
    uint blacks = 0;

    pPerm->setSize( graph.getNoOfNodes() );
    
    iter.first();
    while ((pNode = (TMarkableNode*) iter++) != NULL)
    {
        mark = pNode->mark();
        
        if (mark == R)
        {
            (*pPerm)[reds+noOfBlacks] = pNode->getName();
            reds++;
        }// if
        else
        {                                                  
            (*pPerm)[blacks] = pNode->getName();
            blacks++;
        }// else
        
        if (pNode->getInfo())
            delete (int*) pNode->getInfo();
    }// while

    return pPerm;
}

//
// determine Black-Red-Ordering (same as RedBlack, but Black Nodes are isolated)
//
TPermutation *
TRedBlack::orderBFS_isolated ( TGraph & graph, uint & NoOfReds )
{
    return NULL;
}

//
// doing BFS
//
void                                                                                  
TRedBlack::bfs ( TGraph & graph )
{                                      
    TListIterator<TNode>    iter;
    TListIterator<TEdge>    iterEdge;
    TMarkableNode         * pNode;
    TEdge                 * pEdge;
    TList<TMarkableNode>    listNodes;

    bool                    bFinished, bAdd;
    int                   * pInfo;
                                                                 
    TList<TMarkableNode>        * pListS1;
    TList<TMarkableNode>        * pListS2;
    TList<TMarkableNode>        * pDummy;
    TListIterator<TMarkableNode>  iterAdjNodes;
    TMarkableNode               * pAdjNode;
    TMarkableNode               * pCurrentNode;


    pListS1 = new TList< TMarkableNode >;
    pListS2 = new TList< TMarkableNode >;
    pDummy  = new TList< TMarkableNode >;

    iter.setList( graph.getNodeList() );
    iter.first();
    while ((pNode = (TMarkableNode*) iter++) != NULL)
    {
        pNode->setMark( W );
        pInfo = new int( NOT_VISITED );
        pNode->setInfo( (void*)pInfo );
    }// while
  
    iter.first();
    while ((pNode = (TMarkableNode*) iter++)!= NULL)
    {
        if (GETINFO( pNode ) != VISITED)
        {
            pListS1->clear();
            pListS2->clear();
            pListS1->add( * pNode );
         
            bFinished = FALSE;
            while (!bFinished)
            {
                iterAdjNodes.setList( * pListS1 );
                iterAdjNodes.first();
                while ((pCurrentNode = iterAdjNodes()) != NULL)
                {
                    visit( pCurrentNode );
              
                    // copy adjected nodes to List S2
                    iterEdge.setList( pCurrentNode->getOutEdges() );
              
                    iterEdge.first();
                    while ((pEdge = iterEdge++) != NULL)
                    {
                        pAdjNode = (TMarkableNode*) pEdge->getHead();
                        bAdd     = TRUE;
                        
                        if (GETINFO( pAdjNode ) == NOT_VISITED)
                        {
                            pListS2->add( *pAdjNode );
                            
                            GETINFO( pAdjNode ) = ADDED;
                            bAdd                = FALSE;
                        }// if
                    }// while
                    
                    iterAdjNodes++;
                }// while currentNode=iterAdjNodes() != NULL
                
                //
                // listS1:= ListS2;
                // listS2:= empty
                //
                pDummy  = pListS1;
                pListS1 = pListS2;
                pListS2 = pDummy;
                pListS2->clear();
                
                // finished if L1 is now empty
                if (pListS1->getSize() == 0)
                    bFinished = TRUE;
            }// while
        }// visit
    }// outer loop
    
    delete pListS1;
    delete pListS2;
}

//
// visit node in BFS
//
void 
TRedBlack::visit ( TMarkableNode * pNode )
{
    TListIterator<TEdge>   iterEdge;
    TEdge                * pEdge;
    TMarkableNode        * pAdjNode;
    bool                   Cond;
 
    iterEdge.setList( pNode->getOutEdges() );
    iterEdge.first();

    GETINFO( pNode ) = VISITED; 
 
    // pNode marked as black, but not marked as visited
    if (pNode->mark() == W)
    {                   
        // pNode isolated?                    
        if (pNode->getOutEdges().getSize() == 0)
            pNode->setMark( R );
        else
        { 
            // all adjacents  of pNode W or B? (if one is R: set Cond = False)
            Cond = TRUE;
            while ((pEdge = iterEdge++) != NULL)
            {
                if (((TMarkableNode*) pEdge->getHead())->mark() == R)
                {
                    Cond = FALSE;
                    break;
                }// if
            }// while
            
            if (Cond)
            {                                                                   
                // mark node red (R)
                pNode->setMark(R);

                // mark adjacent vertices  black (B):
                iterEdge.first();
                while ((pEdge = iterEdge++) != NULL)
                {
                    pAdjNode = (TMarkableNode*) pEdge->getOther( pNode );
                    
                    if (pAdjNode->mark() == R)
                        cout << "TRedBlack::visit (...) : Houston, we have a problem !"
                             << endl;   
                    pAdjNode->setMark( B );
                }// while
            }// if
            else
                pNode->setMark( B );
        }// else
    }// if
}

