//
//  Project     : libgraph
//  File        : graph.cc
//  Author      : Ronald Kriemann
//  Purpose     : base class for all graphs
//  Date        : 30.06.1997
//
// arch-tag: 3939a62c-cc8b-4f53-b877-d74c6c2db1be
//

#include "basegraph.hh"

//////////////////////////////////////////////////////////////////
//
// TGraph
//

//
// constructor
//
TGraph::TGraph ()
{
	type = _DIRECTED_;
}

//
// destructor
//
TGraph::~TGraph ()
{
	// remove all nodes
	TListIterator<TNode>  iter( listNodes );
	TNode *pNode;
	while ((pNode = iter++) != NULL)
    {
		pNode->free();
    }// while
}

//
// return number of nodes
//
const uint 
TGraph::getNoOfNodes () const 
{ 
	return listNodes.getSize(); 
}

//
// return the node list
//
TList<TNode>& 
TGraph::getNodeList () 
{ 
	return listNodes; 
}

//
// add a node to the list
//
void 
TGraph::add ( TNode& node ) 
{ 
	node.addReference();
	listNodes.add( node );
}

//
// remove a node from the list
//
void 
TGraph::remove ( TNode& node )
{
	// search for the node in our list
	TListIterator<TNode>  iter( listNodes );
	TNode                *pNode;
	uint                 nodeName = node.getName();

	iter.first();
	while ((pNode = iter()) != NULL)
	{
		if (pNode->getName() == nodeName)
		{
			TListIterator<TEdge>  iterEdge;
			TEdge                *pEdge;

			// go through all succ of node and remove the reference
			iterEdge.setList( pNode->getOutEdges() );
			iterEdge.first();
			while ((pEdge = iterEdge++) != NULL)
			{
				pEdge->getHead()->remPred( *pNode );
				pNode->remSucc( *(pEdge->getHead()) );
				// list was changed : reset
				iterEdge.first();
			}// while
			
			// go through all pred of node and remove the reference
			iterEdge.setList( pNode->getInEdges() );
			iterEdge.first();
			while ((pEdge = iterEdge++) != NULL)
			{
				pEdge->getTail()->remSucc( *pNode );
				pNode->remPred( *(pEdge->getTail()) );
				// list was changed : reset
				iterEdge.first();
			}// while

			// now remove the reference in this graph
			// ( remove the container himself, so we do not have to
			//   search in the list again )
			listNodes.remove( *(iter.getItem()) );
			//			listNodes.remove( node );
			if (pNode->delReference()) delete pNode;

			return;
		}// if

		iter++;
	}// while

	// node is not in this graph !!!
}

//
// get and set type
//
const uint
TGraph::getType () const
{
	return (type);
}

void
TGraph::setType ( uint t )
{
	type = t;
}

//
// remove a subgraph (and only subgraphs)
//
TGraph& 
TGraph::operator-= ( TGraph& subgraph )
{
	TListIterator<TNode>  iter( subgraph.getNodeList() );
	TNode                *pNode;

	iter.first();
	while ((pNode = iter++) != NULL)
		remove( *pNode );

	return (*this);
}

// copy operator
TGraph& 
TGraph::operator= ( TGraph& graph )
{
	TListIterator<TNode>  iterNodes( getNodeList() );
	TNode                *pNode, *pNewNode;
	TNode                *pCNode1, *pCNode2;

	TListIterator<TEdge>  iterEdges;
	TEdge                *pEdge, *pNewEdge;

	// delete all nodes
	while ((pNode = iterNodes++) != NULL)
	{
		remove( *pNode );
	}// while
	
	// create the new nodes
	iterNodes.setList( graph.getNodeList() );
	iterNodes.first();
	while ((pNode = iterNodes++) != NULL)
	{
		pNewNode = pNode->clone();
		add( *pNewNode );
		pNode->setInfo( pNewNode );
		pNewNode->setInfo( pNode );
	}// while

	// create all edges
	iterNodes.first();
	while ((pNode = iterNodes++) != NULL)
	{
		pCNode1 = (TNode*) pNode->getInfo();

		iterEdges.setList( pNode->getOutEdges() );
		iterEdges.first();
		while ((pEdge = iterEdges++) != NULL)
		{
			pCNode2 = (TNode*) pEdge->getHead()->getInfo();

			pNewEdge = pEdge->clone( *pCNode1, *pCNode2 );
			pCNode1->addOutEdge( *pNewEdge );
			pCNode2->addInEdge( *pNewEdge );
		}// while
	}// while

	return *this;
}

//
// create new graph with same vertices and edges
//
TGraph* 
TGraph::clone ()
{
	TGraph  *pGraph = create();
	*pGraph = *this;

	return pGraph;
}

//
// create new graph
//
TGraph* 
TGraph::create () 
{ 
	return new TGraph; 
}

//
// output 
//
void 
TGraph::print ( ostream& os ) const
{
	TListIterator<TNode>  iter( listNodes );
	while (iter() != NULL)
		os << *(iter++) << endl;
}

ostream& 
operator<< ( ostream& os, const TGraph& graph )
{
	graph.print( os );

	return (os);
}
