//
//  Project     : libgraph
//  File        : basenode.cc
//  Author      : Ronald Kriemann
//  Purpose     : base class for all nodes
//  Date        : 04.08.1997
//
// arch-tag: b5406a85-46f3-45b0-804d-97b8418ab916
//

#include "basenode.hh"

//////////////////////////////////////////////////////////////////
//
// TNode
// 

//
// constructor
//
TNode::TNode ( uint n )
{
	name = n;
	pInfo = NULL;

    // remove one reference (so default is no reference)
	delRef();
}

TNode::TNode ( TNode& node )
{
	name = node.name;
	pInfo = node.pInfo;

    // remove one reference (so default is no reference)
	delRef();
}

//
// destructor
//
TNode::~TNode ()
{
	TListIterator<TEdge>  iter( listOutEdges );
	TEdge                *pEdge;

	// remove all out edges
	iter.first();
	while ((pEdge = iter++) != NULL)
	{
		pEdge->free();
	}// while

	// remove all in edges
	iter.setList( listInEdges );
	iter.first();
	while ((pEdge = iter++) != NULL)
	{
        pEdge->free();
	}// while
}

//
// access the name
//
const uint 
TNode::getName () const 
{ 
	return name; 
}

void
TNode::setName ( uint n )
{
    name = n;
}

//
// return number of successors and predecessors
//
const uint 
TNode::getOutDegree () const 
{ 
	return listOutEdges.getSize(); 
}

const uint 
TNode::getInDegree () const 
{ 
	return listInEdges.getSize(); 
}

const uint
TNode::getDegree () const
{
    return listOutEdges.getSize() + listInEdges.getSize();
}

//
// add a node to the succ- (pred-) list
//
void 
TNode::addOutEdge ( TEdge& edge )
{ 
	edge.addReference();
	listOutEdges.add( edge );
}

void 
TNode::addInEdge ( TEdge& edge ) 
{ 
	edge.addReference();
	listInEdges.add( edge );
}

//
// remove an in (out) edge
//
void 
TNode::remOutEdge ( TEdge& edge )
{
	TListIterator<TEdge>  iter( listOutEdges );
	TEdge                *pEdge;
	uint                 edgeName = edge.getName();

	// search for the node in the out-list
	iter.first();
	while ((pEdge = iter()) != NULL)
	{
		if (pEdge->getName() == edgeName)
		{
			listOutEdges.remove( *pEdge );
			if (pEdge->delReference()) delete pEdge;
			return;
		}// if

		iter++;
	}// while
}

void 
TNode::remInEdge ( TEdge& edge )
{
	TListIterator<TEdge>  iter( listInEdges );
	TEdge                *pEdge;
	uint                 edgeName = edge.getName();

	// search for the node in the in-list
	iter.first();
	while ((pEdge = iter()) != NULL)
	{
		if (pEdge->getName() == edgeName)
		{
			listInEdges.remove( *pEdge );
			if (pEdge->delReference()) delete pEdge;
			return;
		}// if

		iter++;
	}// while
}

//
// remove a succ (pred)
//
void 
TNode::remSucc ( TNode& node )
{
	TListIterator<TEdge>  iter( listOutEdges );
	TItem<TEdge>         *pItem;
	TEdge                *pEdge;
	uint                 nodeName = node.getName();

	// search for the node in the out-list
	iter.first();
	while ((pEdge = iter()) != NULL)
	{
		if (pEdge->getHead()->getName() == nodeName)
		{
			pItem = iter.getItem();
			if (pEdge->delReference()) delete pEdge;
			iter++;
			listOutEdges.remove( *pItem );
		}// if
		else
			iter++;
	}// while
}

void
TNode::remPred ( TNode& node )
{
	TListIterator<TEdge>  iter( listInEdges );
	TItem<TEdge>         *pItem;
	TEdge                *pEdge;
	uint                 nodeName = node.getName();

	// search for the node in the in-list
	iter.first();
	while ((pEdge = iter()) != NULL)
	{
		if (pEdge->getTail()->getName() == nodeName)
		{
			pItem = iter.getItem();
			if (pEdge->delReference()) delete pEdge;
			iter++;
			listInEdges.remove( *pItem );
		}// if
		else
			iter++;
	}// while
}

//
// check if pred (succ)
//
const bool 
TNode::isPred ( TNode &node ) const
{
	TListIterator<TEdge>  iterEdges( listInEdges );
	TEdge                *pEdge;
	uint                 nodeName = node.getName();

	iterEdges.first();
	while ((pEdge = iterEdges++) != NULL)
	{
		if (pEdge->getTail()->getName() == nodeName)
			return TRUE;
	}// while

	return FALSE;
}

const bool 
TNode::isSucc ( TNode &node ) const
{
	TListIterator<TEdge>  iterEdges( listOutEdges );
	TEdge                *pEdge;
	uint                 nodeName = node.getName();

	iterEdges.first();
	while ((pEdge = iterEdges++) != NULL)
	{
		if (pEdge->getHead()->getName() == nodeName)
			return TRUE;
	}// while

	return FALSE;
}

//
// return array of successors (predecessors)
//
TList<TEdge>& 
TNode::getOutEdges () 
{ 
	return listOutEdges; 
}

TList<TEdge>& 
TNode::getInEdges () 
{ 
	return listInEdges; 
}

const TList<TEdge>& 
TNode::getOutEdges () const 
{ 
	return listOutEdges; 
}

const TList<TEdge>& 
TNode::getInEdges () const 
{ 
	return listInEdges; 
}

//
// isolate node (remove all incident edges)
//
void 
TNode::isolate ()
{
	TListIterator<TEdge>  iterEdges;
	TEdge                *pEdge;

	iterEdges.setList( listOutEdges );
	iterEdges.first();
	while ((pEdge = iterEdges++) != NULL) 
	{
		pEdge->delReference();
		pEdge->getHead()->getInEdges().remove( *pEdge );

		pEdge->delReference();
		listOutEdges.remove( *pEdge );
		delete pEdge;

		iterEdges.first();
	}// while

	iterEdges.setList( listInEdges );
	iterEdges.first();
	while ((pEdge = iterEdges++) != NULL) 
	{
		pEdge->delReference();
		pEdge->getTail()->getOutEdges().remove( *pEdge );

		pEdge->delReference();
		listInEdges.remove( *pEdge );
		delete pEdge;

		iterEdges.first();
	}// while
}

//
// access the additional information
//
void* 
TNode::getInfo () 
{ 
	return pInfo; 
}

const void* 
TNode::getInfo () const 
{ 
	return pInfo; 
}

void
TNode::setInfo ( void *info ) 
{ 
	pInfo = info; 
}

//
// create a new nodes
//
TNode* 
TNode::clone ()
{
	return new TNode( name );
}

TNode* 
TNode::create ( uint n ) 
{ 
	return new TNode( n ); 
}

//
// output 
//
ostream& 
operator<< ( ostream& os, const TNode& node )
{
	os << "(";

	TListIterator<TEdge> iter( node.listInEdges );

	iter.first();
	while (iter() != NULL)
		os << (iter++)->getTail()->getName() << ",";

	os << ") -> " << node.name << " -> (";

	iter.setList( node.listOutEdges );
	iter.first();
	while (iter() != NULL)
		os << (iter++)->getHead()->getName() << ",";
	os << ")";

	return (os);
}
