//
//  Project     : libgraph
//  File        : blockinfo.cc
//  Author      : Ronald Kriemann
//  Purpose     : class which builds blockinfo of F, L and C blocks
//  Date        : 15.08.1997
//
// arch-tag: 59e7b566-23c5-4baa-a39c-91d47e295682
//

#include <stdio.h>

#include "timer.hh"
#include "var.hh"

#include "net.hh"

#include "basisFE.hh"

#include "drawapp.hh"

#include "TPlanarNode.hh"
#include "T2DPlanarGraph.hh"
#include "TRCMGraph.hh"
#include "TMatrixGraph.hh"
#include "graphtool.hh"

#include "fvs.hh"
#include "planarBlock.hh"
#include "rcm.hh"
#include "criterion.hh"

#include "blockinfo.hh"

// define to really delete feedback-nodes
// if not defined, only marks will be adjusted
#define DEL_FVS_NODES

//
// constructor
//
TBlockInfo::TBlockInfo ()
{
    sortMethod = SORT_OFD;

    bDrawCycles = FALSE;
    bDrawGraph  = FALSE;
    bUseBlocks  = FALSE;
}

TBlockInfo::~TBlockInfo ()
{
}

//
// init the vectors and var. with approp. values
//
void 
TBlockInfo::init ( uint size )
{
    lastF_CNode = 0;
    firstLNode  = size-1;

    vecBlocks.setSize( size );

    lastF_CBlock = lastLBlock = 0;

    vecSizeF_C.setSize( 0 );  
    vecSizeL.setSize( 0 );  

    noOfFVSNodes = 0;

    if (bUseBlocks)
    {
        uint  i;
        
        arrNodes.setSize( size );

        for ( i = 0; i < size; i++ )
            arrNodes[i] = NULL;
    }// if
}

//
// set size of fvs
//
void 
TBlockInfo::setFVSSize ( uint size )
{
    noOfFVSNodes = size;
}

//
// add a FVS to the vectors
//
void 
TBlockInfo::addFVS ( TList<TNode> &fvs )
{
    TListIterator<TNode>  iterNodes( fvs );

    if (bUseBlocks)
    {
        TListIterator<TNode>   iterBlockNodes;
        TPlanarNode           *pNode;
        TNode                 *pBlockNode;
        uint                  nodeName;

        while ((pNode = (TPlanarNode*) iterNodes++) != NULL)
        {
            nodeName = pNode->getName();
            
            if (pNode->blockList() != NULL)
            {
                iterBlockNodes.setList( *(pNode->blockList()) );
                iterBlockNodes.first();
                while ((pBlockNode = iterBlockNodes++) != NULL)
                {
                    vecBlocks[lastF_CNode++] = pBlockNode->getName();
                    noOfFVSNodes++;
                }// while
            }// if
            else if ((arrNodes[nodeName] != NULL) &&
                     (((TPlanarNode*)arrNodes[nodeName])->blockList() != NULL))
            {
                iterBlockNodes.setList( *((TPlanarNode*)arrNodes[nodeName])->blockList() );
                iterBlockNodes.first();
                while ((pBlockNode = iterBlockNodes++) != NULL)
                {
                    vecBlocks[lastF_CNode++] = pBlockNode->getName();
                    noOfFVSNodes++;
                }// while
            }// if
            else
            {
                vecBlocks[lastF_CNode++] = pNode->getName();
                noOfFVSNodes++;
            }// else
        }// while
    }// if
    else
    {
        TNode  *pNode;

        while ((pNode = iterNodes++) != NULL)
        {
            vecBlocks[lastF_CNode++] = pNode->getName();
            noOfFVSNodes++;
        }// while
    }// else
}

//
// "remove" FVS from graph
//
void
TBlockInfo::removeFVS ( TList< TNode > * pFVS, TGraph * pGraph ) const
{
    if ((pFVS == NULL) || (pGraph == NULL))
    {
        cout << "TBlockInfo::removeFVS (...) : got NULl arg" << endl;
        return;
    }// if

    if (pFVS->getSize() == 0)
        return;
    
    //
    // look for minimal and maximal name of nodes in FVS
    //

    TListIterator< TNode >  iterNodes;
    TMarkableNode         * pNode;
    uint                    nodeName;
    uint                    minName, maxName;

    minName = UINT_MAX;
    maxName = 0;

    iterNodes.setList( * pFVS );
    iterNodes.first();
    while ((pNode = (TMarkableNode*) iterNodes++) != NULL)
    {
        nodeName = pNode->getName();
        
        if (nodeName < minName)
            minName = nodeName;

        if (nodeName > maxName)
            maxName = nodeName;
    }// while

    //
    // now create array [minName,maxName] and mark
    // each name, which belongs to a feedback vertex
    //

    uint                i;
    TDynArray< bool >   arrNodeNames;
    uint                size;

    size = maxName - minName + 1;
    arrNodeNames.setSize( size );

    for ( i = 0; i < size; i++ )
        arrNodeNames[i] = FALSE;

    iterNodes.first();
    while ((pNode = (TMarkableNode*) iterNodes++) != NULL)
    {
        arrNodeNames[ pNode->getName() - minName ] = TRUE;
    }// while

    //
    // now iterate through nodes in graph and remove each node in
    // graph, if this node is marked in the name-array
    //

    TItem< TNode >  * pItem;
    
    iterNodes.setList( pGraph->getNodeList() );
    iterNodes.first();
    while ((pNode = (TMarkableNode*) iterNodes()) != NULL)
    {
        nodeName = pNode->getName();

        pItem = iterNodes.getItem();
        iterNodes++;

        if ((nodeName - minName < size) && (arrNodeNames[nodeName-minName]))
        {
#ifdef DEL_FVS_NODES
            pNode->isolate();
            pGraph->getNodeList().remove( * pItem );
#else
            pNode->setMark( FVS_NODE );
#endif
        }// if
    }// while
}

//
// add a F,L or C node(s) to the vector
//
void 
TBlockInfo::addFNode( TNode &node )
{
    if (bUseBlocks)
    {
        TListIterator<TNode>  iterNodes;
        TPlanarNode          *pNode;
        TNode                *pBlockNode;
        uint                 nodeName;
        
        pNode = (TPlanarNode*) &node;
        nodeName = pNode->getName();
        
        if (pNode->blockList() != NULL)
        {
            iterNodes.setList( *(pNode->blockList()) );
            iterNodes.first();
            while ((pBlockNode = iterNodes++) != NULL)
            {
                vecBlocks[lastF_CNode++] = pBlockNode->getName();
            }// while
        }// if
        else if ((arrNodes[nodeName] != NULL) &&
                 (((TPlanarNode*)arrNodes[nodeName])->blockList() != NULL))
        {
            iterNodes.setList( *((TPlanarNode*)arrNodes[nodeName])->blockList() );
            iterNodes.first();
            while ((pBlockNode = iterNodes++) != NULL)
            {
                vecBlocks[lastF_CNode++] = pBlockNode->getName();
            }// while
        }// if
        else
        {
            vecBlocks[lastF_CNode++] = node.getName();
        }// else
    }// if
    else
        vecBlocks[lastF_CNode++] = node.getName();
}

void 
TBlockInfo::addLNode( TNode &node )
{
    if (bUseBlocks)
    {
        TListIterator<TNode>  iterNodes;
        TPlanarNode          *pNode;
        TNode                *pBlockNode;
        uint                 nodeName;
        
        pNode = (TPlanarNode*) &node;
        nodeName = pNode->getName();

        if (pNode->blockList() != NULL)
        {
            iterNodes.setList( *(pNode->blockList()) );
            iterNodes.last();
            while ((pBlockNode = iterNodes--) != NULL)
            {
                vecBlocks[firstLNode--] = pBlockNode->getName();
            }// while
        }// if
        else if ((arrNodes[nodeName] != NULL) &&
                 (((TPlanarNode*)arrNodes[nodeName])->blockList() != NULL))
        {
            iterNodes.setList( *((TPlanarNode*)arrNodes[nodeName])->blockList() );
            iterNodes.first();
            while ((pBlockNode = iterNodes++) != NULL)
            {
                vecBlocks[firstLNode--] = pBlockNode->getName();
            }// while
        }// if
        else
        {
            vecBlocks[firstLNode--] = node.getName();
        }// else
    }// if
    else
        vecBlocks[firstLNode--] = node.getName();
}

void 
TBlockInfo::addCycle ( TGraph &cycle )
{
    TListIterator<TNode>  iterNodes( cycle.getNodeList() );
    uint                 noOfNodes = 0;

    if (bUseBlocks)
    {
        TListIterator<TNode>   iterBlock;
        T2DPlanarNode         *pNode;
        TNode                 *pBlockNode;
        uint                  nodeName;

        // go over all nodes and look for F-marked nodes
        iterNodes.last();
        while ((pNode = (T2DPlanarNode*) iterNodes--) != NULL)
        {
            nodeName = pNode->getName();
            
            if (pNode->blockList() != NULL)
            {
                iterBlock.setList( *(pNode->blockList()) );
                iterBlock.first();
                while ((pBlockNode = iterBlock++) != NULL)
                {
                    noOfNodes++;
                    vecBlocks[lastF_CNode++] = pBlockNode->getName();
                }// while
            }// if
            else if ((arrNodes[nodeName] != NULL) &&
                     (((TPlanarNode*)arrNodes[nodeName])->blockList() != NULL))
            {
                iterNodes.setList( *((TPlanarNode*)arrNodes[nodeName])->blockList() );
                iterNodes.first();
                while ((pBlockNode = iterNodes++) != NULL)
                {
                    noOfNodes++;
                    vecBlocks[lastF_CNode++] = pBlockNode->getName();
                }// while
            }// if
            else
            {
                noOfNodes++;
                vecBlocks[lastF_CNode++] = pNode->getName();
            }// else
        }// while
    }// if
    else
    {
        TNode   *pNode;

        // go over all nodes and look for F-marked nodes
        iterNodes.last();
        while ((pNode = iterNodes--) != NULL)
        {
            noOfNodes++;
            vecBlocks[lastF_CNode++] = pNode->getName();
        }// while
    }// if
        
    if (noOfNodes > 0)
    {
        vecSizeF_C.setSize( lastF_CBlock + 2 );
        vecSizeF_C[lastF_CBlock++] = lastF_CNode - noOfNodes;
        vecSizeF_C[lastF_CBlock++] = lastF_CNode - 1;
    }// if
}

//
// access sort method
//
void 
TBlockInfo::setSortMethod ( int sort )
{
    sortMethod = sort;
}

//
// access filenames
//
void 
TBlockInfo::setCycleName ( const char *name )
{
    if (name != NULL) {
        strcpy( szCycleName, name );
        bDrawCycles = TRUE;
    }// if
    else {
        bDrawCycles = FALSE;
    }// else
}

void 
TBlockInfo::setGraphName ( const char *name )
{
    if (name != NULL) {
        strcpy( szGraphName, name );
        bDrawGraph = TRUE;
    }// if
    else {
        bDrawGraph = FALSE;
    }// else
}

//
// set block-using
//
void 
TBlockInfo::setBlocking ( bool useBlocks )
{
    bUseBlocks = useBlocks;
}


//
// simply put permutation into vector
//
void 
TBlockInfo::buildPerm ( TVectorBlock<uint> &permute )
{
    permute = vecBlocks;
}

//
// build permutation vector
//

/////////////////////////////////////////////////////////////
//
// for concentric cycles
//
void 
TBlockInfo::buildCYCPerm ( TVectorBlock<uint> &permute, 
                           TVectorBlock<uint> &vecSize )
{
    uint i;
    permute = vecBlocks;

    uint noOfBlocks = lastF_CBlock + lastLBlock;
    vecSize.setSize( noOfBlocks );

    for ( i = 0; i < lastF_CBlock; i++)
        vecSize[i] = vecSizeF_C[i];

    for ( i = 0; i < lastLBlock; i++)
        vecSize[noOfBlocks-i-1] = vecSizeL[i];
}

void 
TBlockInfo::buildCYCPerm ( TVectorBlock<uint> &permute, 
                           TVectorBlock<uint> &vecSize,
                           TDynArray< TVector<2> * > &arrVectors, TMatrixBlock &mat, 
                           TFloat xi, uint criterion, 
                           uint ordering )
{
    TTimer         timer;
    TPlanarBlock   block;
    TList<TGraph> *plistCycles = NULL;

    // create graph
    T2DPlanarGraph  *pGraph;

    if (bUseBlocks) 
        pGraph = new T2DPlanarGraph( arrVectors, mat, xi, criterion, TRUE );
    else
        pGraph = new T2DPlanarGraph( arrVectors, mat, xi, criterion );

    // init. size-values for this graph
    init( pGraph->getNoOfNodes() );

    // if a weighted graph is considered, we can only use minimal cycles
    if ((ordering != CONCENTRIC_MIN) && (sortMethod != SORT_OFD))
        ordering = CONCENTRIC_MIN;

    if (bUseBlocks)
    {
        timer.start();
        block.reduce( *pGraph );
        timer.stop();
        cout << "TBlockInfo::buildCYCPerm : time to reduce graph = " << timer << endl;

        // cache nodes
        TListIterator< TNode >  iter( pGraph->getNodeList() );
        TNode                 * pNode;

        iter.first();
        while ((pNode = iter++) != NULL)
            arrNodes[ pNode->getName() ] = pNode;
    }// if

    // sort the edges
    switch (sortMethod) 
    {
        case SORT_OFD              : pGraph->sortOFD(); break;
        case SORT_EDGE_WEIGHT      : pGraph->sortEdgeWeight(); break;
        case SORT_SUCC_EDGE_WEIGHT : pGraph->sortSuccEdgeWeight(); break;
        default                    : break;
    }// switch

    timer.start();
    if (ordering == CONCENTRIC_MIN) 
        plistCycles = block.buildCycles( pGraph, MIN_CYC, this );

    if (ordering == CONCENTRIC_MAX) 
        plistCycles = block.buildCycles( pGraph, MAX_CYC, this );
    timer.stop();

    if (verbosity >= VERBOSE_TIME)
        cout << "time for building concentric cycles : " << timer << endl;

    if (verbosity >= VERBOSE_STEP)
        cout << "found " << plistCycles->getSize() << " cycles" << endl;

    buildCYCPerm( permute, vecSize );

    if ((bDrawGraph == TRUE) || (bDrawCycles == TRUE))
    {
        TVectorBlock<uint> invPerm(permute);

        invertPerm( invPerm, permute );
        if (bDrawGraph == TRUE)  TDrawApp::draw( *pGraph, invPerm, szGraphName );
        if (bDrawCycles == TRUE) TDrawApp::draw( *plistCycles, invPerm, szCycleName );
    }// if

    // clean up
    TListIterator<TGraph>  iterCycles( *plistCycles );
    TGraph                *pCycle;

    iterCycles.first();
    while ((pCycle = iterCycles++) != NULL)
        delete pCycle;

    delete plistCycles;
    delete pGraph;
}

void 
TBlockInfo::buildCYCPerm ( TVectorBlock<uint> &permute, 
                           TVectorBlock<uint> &vecSize,
                           TDynArray< TVector<2> * > &arrVectors,
                           TMatrixBlock &diff, TMatrixBlock &conv,
                           TFloat xi, uint criterion, 
                           uint ordering )
{
    TTimer         timer;
    TPlanarBlock   block;
    TList<TGraph> *plistCycles = NULL;

    // create graph
    T2DPlanarGraph  graph( arrVectors, diff, conv, xi, criterion );

    // init. size-values for this graph
    init( graph.getNoOfNodes() );

    // if a weighted graph is considered, we can only use minimal cycles
    if ((ordering != CONCENTRIC_MIN) && (sortMethod != SORT_OFD))
        ordering = CONCENTRIC_MIN;

    if (bUseBlocks)
    {
        timer.start();
        block.reduce( graph );
        timer.stop();
        if (verbosity >= VERBOSE_TIME)
            cout << "time for reducing : " << timer << endl;

        // cache nodes
        TListIterator< TNode >  iter( graph.getNodeList() );
        TNode                 * pNode;

        iter.first();
        while ((pNode = iter++) != NULL)
            arrNodes[ pNode->getName() ] = pNode;
    }// if

    // sort the edges
    switch (sortMethod) 
    {
        case SORT_OFD              : graph.sortOFD(); break;
        case SORT_EDGE_WEIGHT      : graph.sortEdgeWeight(); break;
        case SORT_SUCC_EDGE_WEIGHT : graph.sortSuccEdgeWeight(); break;
        default                      : break;
    }// switch

    timer.start();
    if (ordering == CONCENTRIC_MIN) 
        plistCycles = block.buildCycles( & graph, MIN_CYC, this );

    if (ordering == CONCENTRIC_MAX) 
        plistCycles = block.buildCycles( & graph, MAX_CYC, this );
    timer.stop();

    if (verbosity >= VERBOSE_TIME)
        cout << "time for building concentric cycles : " << timer << endl;

    if (verbosity >= VERBOSE_STEP)
        cout << "found " << plistCycles->getSize() << " cycles" << endl;

    buildCYCPerm( permute, vecSize );

    if ((bDrawGraph == TRUE) || (bDrawCycles == TRUE))
    {
        TVectorBlock<uint> invPerm(permute);

        invertPerm( invPerm, permute );
        if (bDrawGraph == TRUE)  TDrawApp::draw( graph, invPerm, szGraphName );
        if (bDrawCycles == TRUE) TDrawApp::draw( *plistCycles, invPerm, szCycleName );
    }// if

    // clean up
    TListIterator<TGraph>  iterCycles( *plistCycles );
    TGraph                *pCycle;

    iterCycles.first();
    while ((pCycle = iterCycles++) != NULL)
        delete pCycle;

    delete plistCycles;
}

/////////////////////////////////////////////////////////////
//
// for FVS
//
void 
TBlockInfo::buildFVSPerm ( TVectorBlock<uint> &permute, uint &fvsSize )
{
    permute = vecBlocks;
    fvsSize = noOfFVSNodes;
}

void 
TBlockInfo::buildFVSPerm ( TVectorBlock<uint> &permute, uint &fvsSize,
                           TDynArray< TVector<2> * > &arrVectors, TMatrixBlock &mat, 
                           TFloat xi, uint criterion, uint searchType,
                           uint fvsType )
{
    TTimer         timer;
    TFVS           fvs;
    TPlanarBlock   block;
    TList<TNode> * plistFVS;

    // create graph
    TGraph       * pGraph;

    if (bUseBlocks) 
        pGraph = new T2DPlanarGraph( arrVectors, mat, xi, criterion, TRUE );
    else
    {
        if (fvsType == HEURISTIC_FVS)
            pGraph = new TMatrixGraph( & mat, criterion, xi );
        else
            pGraph = new T2DPlanarGraph( arrVectors, mat, xi, criterion );
    }// else

    // init. size-values for this graph
    init( pGraph->getNoOfNodes() );

    if (bUseBlocks)
    {
        timer.start();
        block.reduce( * ((T2DPlanarGraph*) pGraph) );
        timer.stop();
        if (verbosity >= VERBOSE_TIME)
            cout << "time for reducing : " << timer << endl;

        // cache nodes
        TListIterator< TNode >  iter( pGraph->getNodeList() );
        TNode                 * pNode;

        iter.first();
        while ((pNode = iter++) != NULL)
            arrNodes[ pNode->getName() ] = pNode;
    }// if

    // sort the edges (weight-sort and fvs makes no sense)
    if ( bUseBlocks || (fvsType == PLANAR_FVS))
        ((T2DPlanarGraph*) pGraph)->sortOFD();

    timer.start();
    plistFVS = fvs.buildFVS( *pGraph, fvsType );
    timer.stop();
    if (verbosity >= VERBOSE_TIME)
        cout << "time for building FVS : " << timer << endl;

    //////////////////////////////////
    //
    // remove fvs from graph
    //
    TNode                *pNode;
    TGraphTool            graphtool;
    TListIterator<TNode>  iterFVS( *plistFVS );
    TListIterator<TNode>  iterNodes;

    // now build blockinfo for the FVS
    setFVSSize( 0 );
    addFVS( *plistFVS );

    removeFVS( plistFVS, pGraph );

    if (searchType == SEARCH_BFS)
    {
        TBlockInfo           blockinfo;
        TVectorBlock<uint>  rcmOrder;

        blockinfo.buildRCMPerm( rcmOrder, mat );
        graphtool.numberBFS( *pGraph, this, &rcmOrder, FALSE );
    }// if
    else
        graphtool.numberNodes( *pGraph, this, FALSE );

    buildFVSPerm( permute, fvsSize );

    if ((bDrawGraph == TRUE) && (dynamic_cast< T2DPlanarGraph * >( pGraph ) != NULL))
    {    
        TVectorBlock<uint> invPerm(permute);

        invertPerm( invPerm, permute );
        TDrawApp::draw( *pGraph, invPerm, szGraphName );
    }// if

    // clean up
    iterFVS.setList( * plistFVS );
    iterFVS.first();
    while ((pNode = iterFVS++) != NULL)
        pNode->free();
    
    delete  plistFVS;
    delete  pGraph;
}

void 
TBlockInfo::buildFVSPerm ( TVectorBlock<uint> &permute, uint &fvsSize,
                           TDynArray< TVector<2> * > &arrVectors,
                           TMatrixBlock &diff, TMatrixBlock &conv,
                           TFloat xi, uint criterion, uint searchType,
                           uint fvsType )
{
    TTimer         timer;
    TFVS           fvs;
    TList<TNode>  *plistFVS;

    // create graph
    T2DPlanarGraph  graph( arrVectors, diff, conv, xi, criterion );

    // sort the edges (weight-sort and fvs makes no sense)
    graph.sortOFD();

    // init. size-values for this graph
    init( graph.getNoOfNodes() );

    timer.start();
    plistFVS = fvs.buildFVS( graph, fvsType );
    timer.stop();
    if (verbosity >= VERBOSE_TIME)
        cout << "time for building FVS : " << timer << endl;
        
    //////////////////////////////////
    //
    // remove fvs from graph
    //
    TNode      *pNode;
    TGraphTool  graphtool;

    TListIterator<TNode>  iterFVS( *plistFVS );

    removeFVS( plistFVS, & graph );
    
    // now build blockinfo for the FVS
    setFVSSize( 0 );
    addFVS( *plistFVS );
    if (searchType == SEARCH_BFS)
    {
        TBlockInfo           blockinfo;
        TVectorBlock<uint>  rcmOrder;
        TMatrixBlock         mat;

        mat  = diff;
        mat += conv;

        blockinfo.buildRCMPerm( rcmOrder, mat );
        graphtool.numberBFS( graph, this, &rcmOrder, FALSE );
    }// if
    else
        graphtool.numberNodes( graph, this, FALSE );

    buildFVSPerm( permute, fvsSize );

    if (bDrawGraph == TRUE)
    {    
        TVectorBlock<uint> invPerm(permute);

        invertPerm( invPerm, permute );
        TDrawApp::draw( graph, invPerm, szGraphName );
    }// if

    // clean up
    iterFVS.first();
    while ((pNode = iterFVS++) != NULL)
        pNode->free();
    
    delete  plistFVS;
}

//
// for planar fvs
//
void 
TBlockInfo::buildPlanarFVSPerm ( TVectorBlock<uint> &permute, 
                                 uint &fvsSize )
{
    buildFVSPerm( permute, fvsSize );
}

void 
TBlockInfo::buildPlanarFVSPerm ( TVectorBlock<uint> &permute, uint &fvsSize,
                                 TDynArray< TVector<2> * > &arrVectors, TMatrixBlock &mat, 
                                 TFloat xi, uint criterion, uint searchType )
{
    buildFVSPerm( permute, fvsSize, arrVectors, mat, xi, criterion, 
                  searchType, PLANAR_FVS );
}

void 
TBlockInfo::buildPlanarFVSPerm ( TVectorBlock<uint> &permute, uint &fvsSize,
                                 TDynArray< TVector<2> * > &arrVectors,
                                 TMatrixBlock &diff, TMatrixBlock &conv,
                                 TFloat xi, uint criterion, uint searchType )
{
    buildFVSPerm( permute, fvsSize, arrVectors, diff, conv, xi, criterion,
                  searchType, PLANAR_FVS );
}

// for heuristic fvs
void 
TBlockInfo::buildHeuristFVSPerm ( TVectorBlock<uint> &permute, uint &fvsSize )
{
    buildFVSPerm( permute, fvsSize );
}

void 
TBlockInfo::buildHeuristFVSPerm ( TVectorBlock<uint> &permute, uint &fvsSize,
                                  TDynArray< TVector<2> * > &arrVectors, TMatrixBlock &mat, 
                                  TFloat xi, uint criterion, uint searchType )
{
    buildFVSPerm( permute, fvsSize, arrVectors, mat, xi, criterion, 
                  searchType, HEURISTIC_FVS );
}

void 
TBlockInfo::buildHeuristFVSPerm ( TVectorBlock<uint> &permute, uint &fvsSize,
                                  TDynArray< TVector<2> * > &arrVectors,
                                  TMatrixBlock &diff, TMatrixBlock &conv,
                                  TFloat xi, uint criterion, uint searchType )
{
    buildFVSPerm( permute, fvsSize, arrVectors, diff, conv, xi, criterion,
                  searchType, HEURISTIC_FVS );
}

//
// for rcm
//
void 
TBlockInfo::buildRCMPerm ( TVectorBlock<uint> &permute, TMatrixBlock &mat )
{
    TTimer  timer;
    TRCM    rcm;

    // build graph from matrix and work with it
    TRCMGraph graph( & mat );
    
    timer.start();
    rcm.rcm( graph, this );
    timer.stop();

    if (verbosity >= VERBOSE_TIME)
        cout << "time for rcm : " << timer << endl;

    buildPerm( permute );
}

/////////////////////////////////////////////////////////////
//
// for multigrid
//

//
// concentric cycles
//
void 
TBlockInfo::buildCYCProjPerm ( TVectorBlock<uint> **arrPermVec,
                               TVectorBlock<uint> **arrSizeVec,
                               T2DBasisFE &fe, const TList< TNet<2> > &listNets,
                               TBlockMatrix **arrMatrix,
                               TFloat xi, uint criterion, 
                               uint ordering )
{
    TTimer timer;

    //
    // do the ordering on finest net and project permutation to
    // coarser grids
    //
    TDynArray< TVector<2> * >        arrVectors(0,50);
    TListIterator< TNet<2> >  iterNets( listNets );
    TNet<2>                  *pNet;

    char  szString[256];
    uint  i = 0, lastPerm;
    uint  noOfNets = listNets.getSize();
    uint permSize;

    // if stokes-problem : begin ordering on 2nd level
    if ((arrMatrix[noOfNets-2] != NULL) && (arrMatrix[noOfNets-2]->getNoOfRows() > 1))
        lastPerm = noOfNets-2;
    else
        lastPerm = noOfNets-1;

    // order on finest level
    pNet = iterNets.last();
    if ((pNet != NULL) && (arrMatrix[noOfNets-1] != NULL) )
    {
        fe.getCoord( pNet, arrVectors );
        
        TBlockInfo  blockinfo;
        blockinfo.setSortMethod( sortMethod );
        
        if (bDrawCycles == TRUE) {
            sprintf( szString, "%s%d", szCycleName, i );
            blockinfo.setCycleName( szString );
        }// if
        if (bDrawGraph == TRUE) {
            sprintf( szString, "%s%d", szGraphName, i );
            blockinfo.setGraphName( szString );
        }// if
        
        timer.start();
        blockinfo.buildCYCPerm( *arrPermVec[lastPerm], *arrSizeVec[noOfNets-1],
                                arrVectors, (*arrMatrix[noOfNets-1])[0][0], 
                                xi, criterion, ordering );
        timer.stop();
        
        if (verbosity >= VERBOSE_TIME)
            cout << "time for finding cycles on finest net : " << timer << endl;
    }// if

    permSize = arrPermVec[lastPerm]->getSize();

    // project permutation on coarser nets
    iterNets.first(); i = 0;

    // if stokes-problem : begin ordering on 2nd level
    if ((arrMatrix[noOfNets-2] != NULL) && (arrMatrix[noOfNets-2]->getNoOfRows() > 1))
        iterNets++;

    while ((pNet = iterNets++) != NULL)
    {
        uint noOfInteriorNodes = pNet->getNoOfInteriorNodes();
        uint index = 0;

        arrPermVec[i]->setSize( noOfInteriorNodes );

        for (uint j = 0; j < permSize; j++)
        {
            if ((*arrPermVec[noOfNets-1])[j] < noOfInteriorNodes)
                (*arrPermVec[i])[index++] = (*arrPermVec[noOfNets-1])[j];
        }// for

        i++;
    }// while
}

void
TBlockInfo::buildCYCIndepPerm ( TVectorBlock<uint> **arrPermVec,
                                TVectorBlock<uint> **arrSizeVec,
                                T2DBasisFE &fe, 
                                const TList< TNet<2> > &listNets,
                                TBlockMatrix **arrMatrix,
                                TFloat xi, uint criterion,
                                uint ordering,
                                uint startLevel, uint endLevel )
{
    TTimer timer;

    //
    // do the ordering for each net
    //
    TDynArray< TVector<2> * > arrVectors(0,50);
    TListIterator< TNet<2> >  iterNets( listNets );
    TNet<2>                  *pNet;
    uint                      noOfNets = listNets.getSize();

    char  szString[256];
    uint  i = 0;

    iterNets.first();

    // if stokes-problem : begin ordering on 2nd level
    if ((arrMatrix[noOfNets-2] != NULL) && (arrMatrix[noOfNets-2]->getNoOfRows() > 1))
        iterNets++;

    while ((pNet = iterNets++) != NULL)
    {
        if (verbosity >= VERBOSE_STEP)
            cout << "building cycles for net (" << pNet->getList().getSize() 
                 << ")" << endl;

        if ((arrMatrix[i] != NULL) && (i >= startLevel) && (i <= endLevel))
        {
            fe.getCoord( pNet, arrVectors );

            TBlockInfo  blockinfo;
            blockinfo.setSortMethod( sortMethod );
            blockinfo.setBlocking( bUseBlocks );

            if (bDrawCycles == TRUE) {
                sprintf( szString, "%s%d", szCycleName, i );
                blockinfo.setCycleName( szString );
            }// if
            if (bDrawGraph == TRUE) {
                sprintf( szString, "%s%d", szGraphName, i );
                blockinfo.setGraphName( szString );
            }// if

            timer.start();
            blockinfo.buildCYCPerm( *arrPermVec[i], *arrSizeVec[i],
                                    arrVectors, (*arrMatrix[i])[0][0], 
                                    xi, criterion, ordering );
            timer.stop();

            cout << "TBlockInfo::buildCYCIndepPerm : time for cyc on level "
                 << i << " = "<< timer << endl;
        }// if
        
        i++;
    }// while
}

void
TBlockInfo::buildCYCNestedPerm ( TVectorBlock<uint> **arrPermVec,
                                 TVectorBlock<uint> **arrSizeVec,
                                 T2DBasisFE &fe, const TList< TNet<2> > &listNets,
                                 TBlockMatrix **arrMatrix,
                                 TFloat xi, uint criterion,
                                 uint ordering )
{
    TTimer   timer;

    //
    // do the ordering for each net
    //
    TDynArray< TVector<2> * >        arrVectors(0,50);
    TListIterator< TNet<2> >  iterNets( listNets );
    TNet<2>                  *pNet;
    TList<TGraph>             listCycles;
    TListIterator<TGraph>     iterCycles( listCycles );
    TGraph                   *pCycle;
    uint                      noOfNets = listNets.getSize();
    TPlanarBlock              block;
    TList<TGraph>            *plistCycles = NULL;


    char  szString[256];
    uint  i = 0;

    iterNets.first();

    // if stokes-problem : begin ordering on 2nd level
    if ((arrMatrix[noOfNets-2] != NULL) && (arrMatrix[noOfNets-2]->getNoOfRows() > 1))
        iterNets++;

    while ((pNet = iterNets++) != NULL)
    {
        if (verbosity >= VERBOSE_STEP)
            cout << "building cycles for net (" << pNet->getList().getSize() 
                 << ")" << endl;

        if (arrMatrix[i] != NULL) {
            fe.getCoord( pNet, arrVectors );

            // create graph
            T2DPlanarGraph  graph( arrVectors, (*arrMatrix[i])[0][0], 
                                   xi, criterion );

            if (bDrawGraph == TRUE) {
                sprintf( szString, "%s%d", szGraphName, i );
                TDrawApp::draw( graph, szString );
            }// if

            // init. size-values for this graph
            init( graph.getNoOfNodes() );

            // remove previous cycles from graph
            TListIterator<TGraph>   iterCycles( listCycles );
            TGraph                 *pCycle;
            
            while ((pCycle = iterCycles++) != NULL)
            {
                graph -= *pCycle;
                addCycle( *pCycle );
            }// while

            // sort the edges
            switch (sortMethod) {
                case SORT_OFD              : graph.sortOFD(); break;
                case SORT_EDGE_WEIGHT      : graph.sortEdgeWeight(); break;
                case SORT_SUCC_EDGE_WEIGHT : graph.sortSuccEdgeWeight(); break;
                default                      : break;
            }// switch

            // if a weighted graph is considered, we can only use minimal cycles
            if ((ordering != CONCENTRIC_MIN) && (sortMethod != SORT_OFD))
                ordering = CONCENTRIC_MIN;

            timer.start();
            if (ordering == CONCENTRIC_MIN)
                plistCycles = block.buildCycles( & graph, MIN_CYC, this );
            if (ordering == CONCENTRIC_MAX)
                plistCycles = block.buildCycles( & graph, MAX_CYC, this );
            timer.stop();

            if (verbosity >= VERBOSE_TIME)
                cout << "time for finding cycles : " << timer << endl;

            if (bDrawCycles == TRUE) 
            {
                sprintf( szString, "%s%d", szCycleName, i );
                TDrawApp::draw( *plistCycles, szString );
            }// if

            // add new cycles to cyclelist
            iterCycles.setList( *plistCycles );
            iterCycles.first();
            
            while ((pCycle = iterCycles++) != NULL)
                listCycles.add( *pCycle );

            if (verbosity >= VERBOSE_RESULT)
                cout << "total cycles : " << listCycles.getSize() << endl;

            // clean up
            delete  plistCycles;

            buildCYCPerm( *arrPermVec[i], *arrSizeVec[i] );
        }// if
        i++;
    }// while

    // clean up
    iterCycles.first();
    while ((pCycle = iterCycles++) != NULL)
        delete pCycle;
}

//
// fvs
//
void 
TBlockInfo::buildFVSProjPerm ( TVectorBlock<uint> **arrPermVec,
                               TVectorBlock<uint> *arrFVSSize,
                               T2DBasisFE &fe, const TList< TNet<2> > &listNets,
                               TBlockMatrix **arrMatrix,
                               TFloat xi, uint criterion,
                               uint searchType, uint fvsType )
{
    TTimer timer;

    //
    // do the ordering on finest net and project permutation to
    // coarser grids
    //
    TDynArray< TVector<2> * >        arrVectors(0,50);
    TListIterator< TNet<2> >  iterNets( listNets );
    TNet<2>                  *pNet, *fNet;

    char  szString[256];
    uint  i = 0, lastPerm;
    uint  noOfNets = listNets.getSize();
    uint permSize;

    arrFVSSize->setSize( listNets.getSize() );

    for ( i = 0; i < (uint) listNets.getSize(); i++ )
        arrFVSSize[ i ] = 0;

    i = 0;
    
    // if stokes-problem : begin ordering on 2nd level
    if ((arrMatrix[noOfNets-2] != NULL) && (arrMatrix[noOfNets-2]->getNoOfRows() > 1))
        lastPerm = noOfNets-2;
    else
        lastPerm = noOfNets-1;

    // order on finest level
    fNet = iterNets.last();
    if ((fNet != NULL) && (arrMatrix[noOfNets-1] != NULL) )
    {
        fe.getCoord( fNet, arrVectors );
        
        TBlockInfo  blockinfo;
        blockinfo.setSortMethod( sortMethod );
        
        if (bDrawGraph == TRUE) {
            sprintf( szString, "%s%d", szGraphName, i );
            blockinfo.setGraphName( szString );
        }// if
        
        uint noOfFVSNodes;
        timer.start();
        blockinfo.buildFVSPerm( *arrPermVec[lastPerm], noOfFVSNodes,
                                arrVectors, (*arrMatrix[noOfNets-1])[0][0], 
                                xi, criterion, searchType, fvsType );
        timer.stop();
        
        // we have no FVS on coarser levels
        for ( i = 0; i < noOfNets; i++ ) (*arrFVSSize)[i] = 0;

        (*arrFVSSize)[noOfNets-1] = noOfFVSNodes;

        if (verbosity >= VERBOSE_TIME)
            cout << "time for finding fvs on finest net : " << timer << endl;
    }// if

    permSize = arrPermVec[lastPerm]->getSize();

    // project permutation on coarser nets
    iterNets.first(); i = 0;

    // if stokes-problem : begin ordering on 2nd level
    if ((arrMatrix[noOfNets-2] != NULL) && (arrMatrix[noOfNets-2]->getNoOfRows() > 1))
        iterNets++;

    while ((pNet = iterNets++) != NULL)
    {
        if (pNet != fNet)
        {
            uint noOfInteriorNodes = pNet->getNoOfInteriorNodes();
            uint index = 0;
            
            arrPermVec[i]->setSize( noOfInteriorNodes );
            
            for (uint j = 0; j < permSize; j++)
            {
                if ((*arrPermVec[noOfNets-1])[j] < noOfInteriorNodes)
                    (*arrPermVec[i])[index++] = (*arrPermVec[noOfNets-1])[j];
            }// for

            i++;
        }// if
    }// while
}

void
TBlockInfo::buildFVSIndepPerm ( TVectorBlock<uint> **arrPermVec,
                                TVectorBlock<uint> *arrFVSSize,
                                T2DBasisFE &fe, 
                                const TList< TNet<2> > &listNets,
                                TBlockMatrix **arrMatrix,
                                TFloat xi, uint criterion,
                                uint searchType, uint fvsType,
                                uint startLevel, uint endLevel )
{
    TTimer  timer;

    //
    // do the ordering for each net
    //
    TDynArray< TVector<2> * > arrVectors(0,50);
    TListIterator< TNet<2> >  iterNets( listNets );
    TNet<2>                  *pNet;
    uint                      noOfNets = listNets.getSize();

    char  szString[256];
    uint  i = 0;
    int   j;
    uint fvsSize;

    arrFVSSize->setSize( listNets.getSize() );

    for ( j = 0; j < listNets.getSize(); j++ )
        (*arrFVSSize)[ j ] = 0;

    iterNets.first();

    // if stokes-problem : begin ordering on 2nd level
    if ((arrMatrix[noOfNets-2] != NULL) && (arrMatrix[noOfNets-2]->getNoOfRows() > 1))
        iterNets++;

    while ((pNet = iterNets++) != NULL)
    {
        if ((arrMatrix[i] != NULL) && (i >= startLevel) && (i <= endLevel))
        {
            fe.getCoord( pNet, arrVectors );

            TBlockInfo  blockinfo;
            blockinfo.setSortMethod( sortMethod );
            blockinfo.setBlocking( bUseBlocks );

            if (bDrawCycles == TRUE) {
                sprintf( szString, "%s%d", szCycleName, i );
                blockinfo.setCycleName( szString );
            }// if
            if (bDrawGraph == TRUE) {
                sprintf( szString, "%s%d", szGraphName, i );
                blockinfo.setGraphName( szString );
            }// if

            timer.start();
            blockinfo.buildFVSPerm( *arrPermVec[i], fvsSize,
                                    arrVectors, (*arrMatrix[i])[0][0], 
                                    xi, criterion, searchType, fvsType );
            timer.stop();

            (*arrFVSSize)[i] = fvsSize;

            cout << "TBlockInfo::buildFVSIndepPerm : time for fvs on level "
                 << i << " = " << timer << endl;

            if (verbosity >= VERBOSE_STEP)
                cout << fvsSize << " nodes in FVS" << endl;
        }// if
        
        i++;
    }// while
}

void 
TBlockInfo::buildFVSNestedPerm ( TVectorBlock<uint> **arrPermVec,
                                 TVectorBlock<uint> *arrFVSSize,
                                 T2DBasisFE &fe, const TList< TNet<2> > &listNets,
                                 TBlockMatrix **arrMatrix,
                                 TFloat xi, uint criterion,
                                 uint searchType, uint fvsType )
{
    TTimer   timer;

    //
    // do the ordering for each net
    //
    TDynArray< TVector<2> * >        arrVectors(0,50);
    TListIterator< TNet<2> >  iterNets( listNets );
    TNet<2>                  *pNet;
    TList<TNode>              listNodes;
    TListIterator<TNode>      iterNodes( listNodes );
    TNode                    *pNode;
    uint                     fvsSize, fvsSum;
    uint                      noOfNets = listNets.getSize();

    char  szString[256];
    uint  i = 0;
    int   j;

    fvsSum = 0;

    arrFVSSize->setSize( listNets.getSize() );

    for ( j = 0; j < listNets.getSize(); j++ )
        (*arrFVSSize)[ j ] = 0;

    iterNets.first();

    // if stokes-problem : begin ordering on 2nd level
    if ((arrMatrix[noOfNets-2] != NULL) && (arrMatrix[noOfNets-2]->getNoOfRows() > 1))
        iterNets++;

    while ((pNet = iterNets++) != NULL)
    {
        if (verbosity >= VERBOSE_STEP)
            cout << "building fvs for net (" << pNet->getList().getSize() << ")" << endl;

        if (arrMatrix[i] != NULL) 
        {
            fe.getCoord( pNet, arrVectors );

            // create graph
            T2DPlanarGraph  graph( arrVectors, (*arrMatrix[i])[0][0], 
                                         xi, criterion );

            if (bDrawGraph == TRUE) {
                sprintf( szString, "%s%d", szGraphName, i );
                TDrawApp::draw( graph, szString );
            }// if

            // init. size-values for this graph
            init( graph.getNoOfNodes() );
            setFVSSize( 0 );
            addFVS( listNodes );

            // remove previous fvs-nodes from graph
            removeFVS( & listNodes, & graph );

            // sort the edges (weight-sort and fvs makes no sense)
            graph.sortOFD();

            TFVS  fvs;
            timer.start();
            TList<TNode>  *plistNodes = fvs.buildFVS( graph, fvsType );
            timer.stop();
            if (verbosity >= VERBOSE_TIME)
                cout << "time for finding FVS : " << timer << endl;

            //////////////////////////////////
            //
            // remove fvs from graph
            //
            TNode      *pNode;
            TGraphTool  graphtool;

            removeFVS( plistNodes, & graph );
            
            // now build blockinfo for the new FVS
            addFVS( *plistNodes );
            if (searchType == SEARCH_BFS)
            {
                TBlockInfo           blockinfo;
                TVectorBlock<uint>  rcmOrder;

                blockinfo.buildRCMPerm( rcmOrder, (*arrMatrix[i])[0][0] );
                graphtool.numberBFS( graph, this, &rcmOrder, FALSE );
            }// if
            else
                graphtool.numberNodes( graph, this, FALSE );
            buildFVSPerm( *arrPermVec[i], fvsSize );
            
            // add new fvs-nodes to nodelist
            iterNodes.setList( *plistNodes );
            iterNodes.first();
            while ((pNode = iterNodes++) != NULL)
                listNodes.add( *pNode );

            // clean up
            delete  plistNodes;

            fvsSum += fvsSize;
            (*arrFVSSize)[i] = fvsSum;

            if (verbosity >= VERBOSE_STEP)
                cout << fvsSize << " nodes in local FVS" << endl;
            if (verbosity >= VERBOSE_STEP)
                cout << fvsSum << " nodes in total FVS" << endl;
        }// if
        i++;
    }// while

    // clean up
    iterNodes.setList( listNodes );
    iterNodes.first();
    while ((pNode = iterNodes++) != NULL)
        pNode->free();
}

////////////////////////////////////////////////////
//
// for multigrid (rcm)
//

void 
TBlockInfo::buildRCMProjPerm ( TVectorBlock<uint> **arrPermVec,
                               TBlockMatrix **arrMatrix,
                               uint noOfNets )
{
    TTimer timer;

    //
    // do the ordering on finest net and project permutation to
    // coarser grids
    //

    uint  i = 0, j = 0, lastPerm;
    uint permSize;

    // if stokes-problem : begin ordering on 2nd level
    if ((arrMatrix[noOfNets-2] != NULL) && (arrMatrix[noOfNets-2]->getNoOfRows() > 1))
    {
        lastPerm = noOfNets-2;
        j++;
    }
    else
        lastPerm = noOfNets-1;

    // order on finest level
    if (arrMatrix[noOfNets-1] != NULL)
    {
        TBlockInfo  blockinfo;
        
        timer.start();
        blockinfo.buildRCMPerm( *arrPermVec[lastPerm], 
                                (*arrMatrix[noOfNets-1])[0][0] );
        timer.stop();
        
        if (verbosity >= VERBOSE_TIME)
            cout << "time for rcm on finest net : " << timer << endl;
    }// if

    permSize = arrPermVec[lastPerm]->getSize();

    // project permutation on coarser levels
    for (i = 0; i < noOfNets-1; i++, j++)
    {
        uint noOfInteriorNodes = (*arrMatrix[j])[0][0].getNoOfRows();
        uint index = 0;

        arrPermVec[i]->setSize( noOfInteriorNodes );

        for (uint k = 0; k < permSize; k++)
        {
            if ((*arrPermVec[noOfNets-1])[k] < noOfInteriorNodes)
                (*arrPermVec[i])[index++] = (*arrPermVec[noOfNets-1])[k];
        }// for
    }// while
}

void 
TBlockInfo::buildRCMIndepPerm ( TVectorBlock<uint> **arrPermVec,
                                TBlockMatrix **arrMatrix,
                                uint startLevel, uint endLevel )
{
    TTimer  timer;

    //
    // do the ordering for each net
    //
    uint  i = 0, j = 0;

    // if stokes-problem : begin ordering on 2nd level
    if ((arrMatrix[endLevel-1] != NULL) && (arrMatrix[endLevel-1]->getNoOfRows() > 1))
        j++;

    for (i = startLevel; i <= endLevel - j; i++)
    {
        if (arrMatrix[i] != NULL) 
        {
            TBlockInfo  blockinfo;

            timer.start();
            blockinfo.buildRCMPerm( *arrPermVec[i], (*arrMatrix[i])[0][0] );
            timer.stop();

            cout << "TBlockInfo::buildRCMIndepPerm : time for rcm on level "
                 << i << " = " << timer << endl;
        }// if
    }// for

}

///////////////////////////////////////////////////////
//
// additional functions for permutation handling
//

void
invertPerm( TVectorBlock<uint> &invPerm, TVectorBlock<uint> &perm )
{
    uint i;
    uint n = perm.getSize();

    invPerm.setSize( n );

    for (i = 0; i < n; i++)
        invPerm[perm[i]] = i;
}

void 
reverseOrder ( TVectorBlock<uint> &perm )
{
    uint                 i, size;
    TVectorBlock<uint>  temp;

    size = perm.getSize();
    temp.setSize( size );
    
    for (i = 0; i < size; i++)
        temp[size-1-i] = perm[i];

    perm = temp;
}

bool 
orderBlocks( TVectorBlock<uint> *PermVec,
             TVectorBlock<uint> *cycle, TBlockMatrix *pMatrix )
{
    uint i,j,k;
    // dim
    uint N = (*pMatrix)[0][0].getNoOfRows();
    bool  changed = FALSE;
   
    if (cycle->getSize() > 2 )
    {
        TVectorBlock<uint>  invPerm(*PermVec);
        TVectorBlock<uint>  ba(0,2*(cycle->getSize())+1);
        TVectorBlock<uint>  be(0,2*(cycle->getSize())+1);
        TVectorBlock<uint>  cf(0,2*(cycle->getSize())+1);
        TVectorBlock<uint>  bl(0,2*(cycle->getSize())+1);
        TVectorBlock<uint>  blockindex(0,N);
     
        invertPerm( invPerm, *PermVec );
     
        // new ordering
        (*pMatrix)[0][0].permute(&invPerm,&invPerm );
     
        // determine number of blocks
        uint bend=1;
        uint cya,cyb,banz;
     
        // first block
        j   = 0;
        cya = (*cycle)[0];
        cyb = (*cycle)[1];
     
        if (cya == 0)
        {
            ba[0] = cya;
            be[0] = cyb;
            bend  = be[0];
            cf[0] = 1;
            j     = 1;
        }// if
        else {
            ba[j] = 0;
            be[j] = cya-1;
            cf[j] = 0;
            j++;
            ba[j] = cya;
            be[j] = cyb;
            cf[j] = 1;
            bend  = be[j];
            j++;
        }// else
        
        //if the first indicees in *cycle point to a block, start with the secend 
        // block in *cycle (i.e., j=1)
        // else: j =0; start with the second block.
        for (i=1;i<cycle->getSize()/2;i++)
        {
            cya = (*cycle)[2*i];
            cyb = (*cycle)[2*i+1];
            if (cya == bend+1)
            {  
                 ba[j] = cya;
                be[j] = cyb;
                bend  = be[j];
                cf[j] = 1;
                j++;
            }// if
            else
            {
                ba[j] = bend+1;
                be[j] = cya-1;
                cf[j] = 0;
                j++;
                ba[j] = cya;
                be[j] = cyb;
                cf[j] = 1;
                bend  = be[j];
                j++;
            }// else
        } // for
     
        // last block
        if ( (*pMatrix)[0][0].getNoOfRows()-1>bend) 
        {
            ba[j] = bend+1;
            be[j] = (*pMatrix)[0][0].getNoOfRows()-1;
            cf[j] = 0;
        }// if
        else 
            j--;
     
        // number of blocks  
        banz = j+1;
     
        //   cout << "banz: " << banz << endl;

        // block length
        for (k = 0; k < banz; k++)
            bl[k] = be[k]-ba[k]+1;

        ba.setSize(banz);
        be.setSize(banz);
        cf.setSize(banz);
        bl.setSize(banz);

        // cout << "find blockindex" << endl;
        // blockindex vector ( zu welchen Block gehoert Index i?)
        for (k = 0; k < banz; k++)
            for (i = ba[k]; i <= be[k]; i++)
                blockindex[i] = k;
     
        // build blockdependency-matrix
        TVectorBlock<TFloat> *pX = new TVectorBlock<TFloat> (0.0,N);
        TVectorBlock<TFloat> *pY = new TVectorBlock<TFloat> (0.0,N);
  
        // cout << "fordere B1" << endl;

        TMatrixBlock B1(N);
     
        // cout << "B1 gefordert" << endl;

        TFloat Values[banz];
        uint  Index[banz];
        bool   used[banz];
        uint  noOfEntries=0;
        TFloat dia;
        TMatrixData tmd;
        uint  nOE,l,NO;

        for (k = 0; k < banz; k++) {
            Index[k]  = k;
            Values[k] = 0.0;
            used[k]   = 0;
        }// for

        for (i = 0; i < N; i++) 
        {
            tmd = TMatrixData((*pMatrix)[0][0][i]);
            dia = tmd[i];
            nOE = tmd.getNoOfEntries();
            NO  = 0;
            for (l = 0; l < nOE; l++)
            {
                k = blockindex[tmd.getIndex(l)];
                if (used[k])
                {
                     // find Index
                    j = 0;
                    while ((j < NO) && (Index[j] != k))
                        j++;
                    // assign Value
                    Values[j] -= tmd.getValue(l)/dia;
                }// if
                else
                {
                    Values[NO] -= tmd.getValue(l)/dia;
                    Index[NO]   = k;
                    used[k]     = TRUE;
                    NO++;
                }// else
            }// for
            for (l = 0; l < NO; l++)
                if (Index[l] == blockindex[i]) 
                    Values[l]++;
            // set matrix line in B1
            B1[i].set(NO, Index, Values);
            // clear used vector
            for  (l = 0; l < nOE; l++) 
            {
                k = blockindex[tmd.getIndex(l)];
                used[k] = FALSE;
            }// for
            for( l = 0; l < NO; l++) 
            { 
                Values[l] = 0.0; 
                Index[l]  = 0;
            }// for
        } // for i

        /*
          // cout<<"Matrix: "<<(*pMatrix)[0][0]<<endl;
          for (k=0;k<banz;k++){
          for (i=ba[k];i<= be[k];i++)
          (*pX)[i]=1.0;
          *pY=((*pMatrix)[0][0])*(*pX);
          for (i=0;i<N;i++){
          B1[i][k]= (-1.0)*(*pY)[i]/(*pMatrix)[0][0][i][i];
          if( k == blockindex[i])
          B1[i][k]++;
          }
          for (i=ba[k];i<= be[k];i++)
          (*pX)[i]=0.0;
          
          }
          //   for (i=0;i<N;i++){ for (k=0;k<banz;k++){cout<<  B1[i][k]<<" ";}cout<<endl;}
          */

        // cout << "Matrix B1 calculated: " << endl;
   
        TMatrixBlock BB(banz);
     
        TFloat s;
        // TFloat Values[banz];
        // uint Index[banz];
        // uint 
     
        for (i = 0; i < banz; i++)
        {
            noOfEntries = 0;
            for (k = 0; k < banz; k++)
            {
                s = 0;
                for (j = ba[i]; j <= be[i]; j++)
                    if (B1[j].hasIndex(k,banz))
                        s += B1[j][k];
                
                if ((fabs(s) > 1e-5) || (i == k) )
                {
                     Values[noOfEntries] = s/(be[i]-ba[i]+1.0);
                    Index[noOfEntries]  = k;
                    noOfEntries++;
                }// if
            }// for
            BB[i].set(noOfEntries,Index,Values);
        }// for
     
     
        for (i = 0; i < banz; i++)
            if (fabs(BB[i][i]) < 1e-5) 
                BB[i][i]=1e-5;

        // delete minicycles
        for (i = 0; i < banz; i++)
            for (k = 0; k < i; k++)
                if((BB[i].hasIndex(k,banz)) && (BB[k].hasIndex(i,banz)))
                {
                    if (fabs(BB[i][k]) > fabs(BB[k][i]))
                        BB[k][i] = 0.0;
                    else 
                        BB[i][k] = 0.0;
                }// if
           
        // cout << "BB " << BB << endl;      
        //   
        // cout << " Matrix BB calculated" << endl;
   
        TDynArray< TVector<2> * > *arrVectors = new TDynArray< TVector<2> * >;
        TVertex<2> *zv;

        // zero initialize
        for (i = 0; i < banz; i++) 
        {
            zv = new TVertex<2>( 0.5+0.4*cos(6.27/banz*i),
                                 0.5+0.4*sin(6.27/banz*i),
                                 0L );
            (*arrVectors)[i] = zv;
            // cout << ((*arrVectors)[i])[0] << (*arrVectors)[i][1] << endl;
            // cout << (*arrVectors)[i] << endl;
        }// for
   
        /*   TDrawApp *app = new TDrawApp;
  
             T2DPlanarGraph *graph =NULL;
             graph = new T2DPlanarGraph( *arrVectors,BB,0.1,_E4_);
             cout<<*graph<<endl;
             TGraphTool  tool;
             tool.numberNodes( *graph );
             app->draw( *graph,  "graph" );
             */


        TVectorBlock<uint> BPerm;
        TVectorBlock<uint> BCyc;
     
        TBlockInfo *bi = new TBlockInfo();
        bi->setSortMethod( SORT_EDGE_WEIGHT);
        bi->buildCYCPerm(BPerm,BCyc, *arrVectors,BB,0.1,E4, CONCENTRIC_MIN);
  
        // cout << "Matrix BB permutiert" << endl;
   
        for (i = 0; i < banz; i++)
            delete (*arrVectors)[i];
     
        delete arrVectors;
        delete bi;
        delete pX;
        delete pY;
   
        // restore old ordering;
        (*pMatrix)[0][0].permute(PermVec,PermVec );

        // calculate new permutation vector
        TVectorBlock<uint> InvBPerm(BPerm);
        invertPerm( InvBPerm, BPerm );
   
        TVectorBlock<uint> NewPerm(*PermVec);
        uint m;
        // cout << "BPerm" << BPerm;
   
        m = 0;
        for (k = 0; k < banz; k++)
        {
            for (l = ba[BPerm[k]]; l <= be[BPerm[k]]; l++)
            {
                NewPerm[m] = (*PermVec)[l];
                m++;
            }// for
        }// for
  

        // has PermVector changed?
        s = 0;
        for (k = 0; k < banz; k++)
            s += fabs(BPerm[k]-k);
        if (s > 0) 
            changed = TRUE;
        else 
            changed = FALSE;

        *PermVec = NewPerm;

        // determine cyclic blocks
        cf.permute(&InvBPerm);
        ba.permute(&InvBPerm);
        be.permute(&InvBPerm);
        bl.permute(&InvBPerm);

        m = 0; l = 0;
        for(k = 0; k < banz; k++)
        {
             ba[k]  = m;
            m     += (bl[k]-1);
            be[k]  = m;
            m++;
        }// for
  
        l = 0;
        for(k = 0; k < banz; k++)
        {
            if (cf[k] == 1)
            {
                (*cycle)[l] = ba[k];
                l++;
                (*cycle)[l] = be[k];
                l++;
            }// if
        }// for
        //  cout << "Bloecke permutiert" << endl;
    }

    return changed;
}

