//
//  Project     : libgraph
//  File        : graphtool.cc
//  Author      : Ronald Kriemann
//  Purpose     : class for some common graph tools
//
// arch-tag: f7833b71-97d5-45af-90ab-b27c0b0db978
//

#include "timer.hh"
#include "parray.hh"
#include "var.hh"
#include "stack.hh"

#include "blockvector.hh"

#include "fvs.hh"
#include "blockinfo.hh"

#include "graphtool.hh"

//////////////////////////////////////////////

#define NOT_NUMBERED  -1
#define UNDEF         -1

//
// numbering algorithm
//
void 
TGraphTool::numberNodes ( TGraph& graph, TBlockInfo *pBlockInfo,
                          bool bInit )
{
    TListIterator<TNode> iter( graph.getNodeList() );
    TMarkableNode       *pNode;
    uint                 graphSize = 0;

    // set marks of all nodes to C
    iter.first();
    while ((pNode = (TMarkableNode*) iter++) != NULL)
    {
        if (pNode->mark() != FVS_NODE)
        {
            pNode->setMark( C );
            graphSize++;
        }// if
    }// while

    if ((pBlockInfo != NULL) && (bInit == TRUE)) 
        pBlockInfo->init( graphSize );

    // now adjust marks
    iter.first();
    while ((pNode = (TMarkableNode*) iter++) != NULL)
    {
        if (pNode->mark() == C)
            setF( *pNode, pBlockInfo );

        if (pNode->mark() == C)
            setL( *pNode, pBlockInfo );
    }// while
}

//
// set attr F to a node
//
void 
TGraphTool::setF ( TMarkableNode& node, TBlockInfo *pBlockInfo )
{
    TListIterator<TEdge>  iter( node.getInEdges() );
    TEdge                *pEdge;
    TMarkableNode        *pNode;

    bool  bPredHaveF = TRUE;

    iter.first();
    while ((pEdge = iter++) != NULL)
    {
        pNode = (TMarkableNode*) pEdge->getTail();

        if ((pNode->mark() != F) && (pNode->mark() != FVS_NODE))
        {
            bPredHaveF = FALSE;
            iter.last();
            iter++;
        }// if
    }// while

    if (bPredHaveF == TRUE)
    {
        node.setMark( F );
        if (pBlockInfo != NULL)
            pBlockInfo->addFNode( node );

        // do the same with all successors
        iter.setList( node.getOutEdges() );
        iter.first();
        while ((pEdge = iter++) != NULL)
        {
            pNode = (TMarkableNode*) pEdge->getHead();

            if (pNode->mark() == C)
                setF( *pNode, pBlockInfo );
        }// while
    }// if
}

//
// set attr L to a node
//
void 
TGraphTool::setL ( TMarkableNode& node, TBlockInfo *pBlockInfo )
{
    TListIterator<TEdge>  iter( node.getOutEdges() );
    TEdge                *pEdge;
    TMarkableNode        *pNode;

    bool  bSuccHaveL = TRUE;

    iter.first();
    while ((pEdge = iter++) != NULL)
    {
        pNode = (TMarkableNode*) pEdge->getHead();
        if ((pNode->mark() != L) && (pNode->mark() != FVS_NODE))
        {
            bSuccHaveL = FALSE;
            iter.last();
            iter++;
        }// if
    }// while

    if (bSuccHaveL == TRUE)
    {
        node.setMark( L );
        if (pBlockInfo != NULL)
            pBlockInfo->addLNode( node );

        // do the same with all predecessors
        iter.setList( node.getInEdges() );
        iter.first();
        while ((pEdge = iter++) != NULL)
        {
            pNode = (TMarkableNode*) pEdge->getTail();

            if (pNode->mark() == C)
                setL( *pNode, pBlockInfo );
        }// while
    }// if
}

//
// similar algorithm but with BFS-type of search
//
void 
TGraphTool::numberBFS( TGraph &graph, TBlockInfo *pBlockInfo,
                       TVectorBlock<uint>  *pvecPerm,
                       bool bInit )
{
    uint  i;

    TListIterator<TNode>   iter( graph.getNodeList() );
    TMarkableNode         *pNode;
    TList<TMarkableNode>   listNodes;
    uint                   noOfNodes = graph.getNoOfNodes();
    TPArray<TMarkableNode> arrNodes( noOfNodes, 0 );

    // set marks of all nodes to C
    iter.first(); i = 0;
    while ((pNode = (TMarkableNode*) iter++) != NULL)
    {
        pNode->setMark( C );
        arrNodes.set( i++, pNode );
    }// while

    if ((pBlockInfo != NULL) && (bInit == TRUE)) 
        pBlockInfo->init( graph.getNoOfNodes() );

    // now adjust marks
    for (i = 0; i < noOfNodes; i++)
    {
        if (pvecPerm != NULL)
            pNode = arrNodes[(*pvecPerm)[i]];
        else
            pNode = arrNodes[i];

        if (pNode != NULL)
        {
            if (pNode->mark() == C)
                setF_BFS( *pNode, listNodes, pBlockInfo );

            if (pNode->mark() == C)
                setL_BFS( *pNode, listNodes, pBlockInfo );
        }// if
    }// while
}
    
void 
TGraphTool::setF_BFS ( TMarkableNode& node, TList<TMarkableNode> &list,
                       TBlockInfo *pBlockInfo )
{
    TListIterator<TEdge>  iter( node.getInEdges() );
    TEdge                *pEdge;
    TMarkableNode        *pNode;

    bool  bPredHaveF = TRUE;

    iter.first();
    while ((pEdge = iter++) != NULL)
    {
        pNode = (TMarkableNode*) pEdge->getTail();

        if (pNode->mark() != F)
        {
            bPredHaveF = FALSE;
            iter.last();
            iter++;
        }// if
    }// while

    if (bPredHaveF == TRUE)
    {
        node.setMark( F );
        if (pBlockInfo != NULL) pBlockInfo->addFNode( node );

        // add all succ with mark = C to list to treat them later
        iter.setList( node.getOutEdges() );
        iter.first();
        while ((pEdge = iter++) != NULL)
        {
            pNode = (TMarkableNode*) pEdge->getHead();

            if (pNode->mark() == C)
                list.add( *pNode );
        }// while
    }// if

    while (list.getSize() > 0)
    {
        pNode = list.getFirst();
        if (pNode != NULL)
        {
            list.remove( *pNode );
            if (pNode->mark() == C)
                setF_BFS( *pNode, list, pBlockInfo );
        }// if
    }// while
}

void 
TGraphTool::setL_BFS ( TMarkableNode& node, TList<TMarkableNode> &list,
                       TBlockInfo *pBlockInfo )
{
    TListIterator<TEdge>  iter( node.getOutEdges() );
    TEdge                *pEdge;
    TMarkableNode        *pNode;

    bool  bSuccHaveL = TRUE;

    iter.first();
    while ((pEdge = iter++) != NULL)
    {
        pNode = (TMarkableNode*) pEdge->getHead();
        if (pNode->mark() != L)
        {
            bSuccHaveL = FALSE;
            iter.last();
            iter++;
        }// if
    }// while

    if (bSuccHaveL == TRUE)
    {
        node.setMark( L );
        if (pBlockInfo != NULL) pBlockInfo->addLNode( node );

        // add all pred with mark = C to list to treat them later
        iter.setList( node.getInEdges() );
        iter.first();
        while ((pEdge = iter++) != NULL)
        {
            pNode = (TMarkableNode*) pEdge->getTail();

            if (pNode->mark() == C)
                list.add( *pNode );
        }// while
    }// if

    while (list.getSize() > 0)
    {
        pNode = list.getFirst();
        if (pNode != NULL)
        {
            list.remove( *pNode );
            if (pNode->mark() == C)
                setL_BFS( *pNode, list, pBlockInfo );
        }// if
    }// while
}

//////////////////////////////////////////////

// data which is used in the Tarjan algorithm
struct TNodeData 
{
    long     number, lowlink;
    bool     bInStack;
    TNode   *pCreatedNode;
    TGraph  *pComponent;
};

#define NODEDATA(x) ((TNodeData*)((x)->getInfo()))

//
// decompose a given graph into the strong connected
// components (based on Tarjan [72], SIAM, Vol 1 No 2)
//
TList<TGraph>* 
TGraphTool::tarjan ( TGraph& graph )
{
//    cout << "TGraphTool::tarjan : WARNING: using recursive version !!!" << endl;
    
    TTimer  timer;

    TListIterator<TNode>  iter( graph.getNodeList() );
    TNode                *pNode;

    // list with the components
    TList<TGraph>  *plistGraph = new TList<TGraph>;

    // list for the nodes on our way
    TList<TNode>  stack;
    
    uint  noOfNodes = graph.getNoOfNodes();
    uint  index = 0;

    // create a nodedata for every node
    TNodeData  nodeData[noOfNodes];

    // reset nodedata of all nodes
    long  i = 0;
    iter.first();
    while ((pNode = iter++) != NULL)
    {
        nodeData[i].number = nodeData[i].lowlink = NOT_NUMBERED;
        nodeData[i].bInStack     = FALSE;
        nodeData[i].pCreatedNode = NULL;
        nodeData[i].pComponent   = NULL;
        pNode->setInfo( &(nodeData[i]) );
        i++;
    }// for

    ////////////////////////////////////////////////
    //
    // now go through all nodes and build components
    //
    timer.start();

    iter.first();
    while ((pNode = iter++) != NULL)
    {
        if (NODEDATA(pNode)->number == NOT_NUMBERED)
            strongConnect( *plistGraph, stack, graph, *pNode, index );
    }// while

    timer.stop();
    if (verbosity >= VERBOSE_TIME)
        cout << "tarjan : time for building components : " << timer << endl;


    ////////////////////////////////////////////////
    //
    // add all edges to the components 
    //
    timer.start();

    TListIterator<TGraph>  iterComp( *plistGraph );
    TGraph                *pComponent;

    TListIterator<TEdge>   iterEdge;
    TEdge                 *pEdge, *pNewEdge;

    TGraph                *pActComponent;

    TNode                 *pSuccNode, *pOrig;
    TNodeData             *pSuccData;

    iterComp.first();
    while ((pComponent = iterComp++) != NULL)
    {
        iter.setList( pComponent->getNodeList() );

        iter.first();
        while ((pNode = iter++) != NULL)
        {
            // get succ of node in our orig graph
            pOrig = (TNode*)(pNode->getInfo());
            iterEdge.setList( pOrig->getOutEdges() );

            pActComponent = NODEDATA(pOrig)->pComponent;

            // go through all this succ and add them in the succ array
            // of this node, if they are also in this component
            // (we can see this, because the lowlink-number is equal in
            //  all nodes of the same component)
            iterEdge.first();
            while ((pEdge = iterEdge++) != NULL)
            {
                pSuccNode = pEdge->getHead();
                pSuccData = NODEDATA(pSuccNode);

                if (pSuccData->pComponent == pActComponent)
                {
                    pNewEdge = pEdge->create( *pNode, *(pSuccData->pCreatedNode) );
                    pNode->addOutEdge( *pNewEdge );
                    pSuccData->pCreatedNode->addInEdge( *pNewEdge );
                }// if
            }// while
        }// while
    }// while

    //
    // order the edges in each comp. to fit ordering in orig. graph
    //
    iterComp.first();
    while ((pComponent = iterComp++) != NULL)
    {
        iter.setList( pComponent->getNodeList() );

        iter.first();
        while ((pNode = iter++) != NULL)
        {
            sortInEdges( pNode );
            sortOutEdges( pNode );
        }// while
    }// while
    
    timer.stop();
    if (verbosity >= VERBOSE_TIME)
        cout << "tarjan : time for adding edges : " << timer << endl;

    return (plistGraph);
}

//
// used by tarjan to detect the components
//
void 
TGraphTool::strongConnect ( TList<TGraph>& listGraph,
                            TList<TNode>& stack,
                            TGraph& graph, TNode& node, 
                            uint& index )
{
    // get list of successors
    TListIterator<TEdge>  iter( node.getOutEdges() );
    TEdge                *pEdge;
    TNode                *pNode;
    TNodeData            *pNodeData, *pSuccData;

    pNodeData = (TNodeData*) node.getInfo();

    // set new number and lowlink
    pNodeData->number = pNodeData->lowlink = index++;

    // add node to stack
    stack.insert( node );
    pNodeData->bInStack = TRUE;

    // go over all successors and look for components
    iter.first();
    while ((pEdge = iter++) != NULL)
    {
        pNode = pEdge->getHead();
        pSuccData = NODEDATA(pNode);

        // no component
        if (pSuccData->number == NOT_NUMBERED)
        {
            strongConnect( listGraph, stack,
                           graph, *pNode, index );
            pNodeData->lowlink = MIN( pNodeData->lowlink, pSuccData->lowlink );
        }// if
        // component detected
        else
        {
            if ((pSuccData->number < pNodeData->number) && 
                (pSuccData->bInStack == TRUE))
                pNodeData->lowlink = MIN( pNodeData->lowlink, pSuccData->number );
        }// else
    }// for

    // begin of component -> collect all members
    if (pNodeData->lowlink == pNodeData->number)
    {
        TGraph    *pComponent = graph.create();
        TNode     *pTopNode, *pCompNode;
        TNodeData *pTopNodeData;

        pNodeData->pComponent = pComponent;
        
        while (((pTopNode = stack.getFirst()) != NULL) &&
               (NODEDATA( pTopNode )->number >= pNodeData->number))
        {
            pTopNodeData = NODEDATA(pTopNode);
            
            pCompNode = pTopNode->clone();
            pComponent->add( *pCompNode );

            // add a link to the original node on the created node
            pCompNode->setInfo( pTopNode );
            
            // and a link to the created node on the original
            pTopNodeData->pCreatedNode = pCompNode;

            // reset lowlink to indicate component
            pTopNodeData->lowlink = pNodeData->lowlink;
            pTopNodeData->pComponent = pComponent;

            // remove top from stack
            stack.remove( *pTopNode );
            pTopNodeData->bInStack = FALSE;
        }// while

        listGraph.add( * pComponent );
    }// if
}

void 
TGraphTool::strongConnect_nr ( TList<TGraph>& listGraph,
                               TList<TNode>& stack,
                               TGraph& graph, TNode& node, 
                               uint& index )
{
    TList< TNode >          listDFS;
    TNode                 * pTop, * pSucc;
    TNodeData             * pTopData, * pSuccData;
    TListIterator< TEdge >  iterEdges;
    TEdge                 * pEdge;

    listDFS.insert( node );
    
    while (listDFS.getSize() > 0)
    {
        pTop     = listDFS.getFirst();
        pTopData = NODEDATA( pTop );

        if (pTopData->number != NOT_NUMBERED)
        {
            listDFS.remove( *pTop );

            iterEdges.setList( pTop->getOutEdges() );
            iterEdges.first();
            while ((pEdge = iterEdges++) != NULL)
            {
                pSucc     = pEdge->getHead();
                pSuccData = NODEDATA( pSucc );

                if (pSuccData->bInStack == TRUE)
                    pTopData->lowlink = MIN( pTopData->lowlink, pSuccData->lowlink );
            }// while
            
            if (pTopData->number == pTopData->lowlink)
            {
                TGraph    *pComponent = graph.create();
                TNode     *pTopNode, *pCompNode;
                TNodeData *pTopNodeData;
                
                pTopData->pComponent = pComponent;
                
                while (((pTopNode = stack.getFirst()) != NULL) &&
                       (NODEDATA( pTopNode )->number >= pTopData->number))
                {
                    pTopNodeData = NODEDATA(pTopNode);
                    
                    pCompNode = pTopNode->clone();
                    pComponent->add( *pCompNode );
                    
                    // add a link to the original node on the created node
                    pCompNode->setInfo( pTopNode );
                    
                    // and a link to the created node on the original
                    pTopNodeData->pCreatedNode = pCompNode;
                    
                    // reset lowlink to indicate component
//                    pTopNodeData->lowlink = pTopData->lowlink;
                    pTopNodeData->pComponent = pComponent;
                    
                    // remove top from stack
                    stack.remove( *pTopNode );
                    pTopNodeData->bInStack = FALSE;
                }// while

                listGraph.add( * pComponent );
            }// if
        }// if
        else
        {
            pTopData->number = pTopData->lowlink = index++;
            stack.insert( * pTop );
            pTopData->bInStack = TRUE;

            iterEdges.setList( pTop->getOutEdges() );
            iterEdges.first();
            while ((pEdge = iterEdges++) != NULL)
            {
                pSucc     = pEdge->getHead();
                pSuccData = NODEDATA( pSucc );

                if (pSuccData->number == NOT_NUMBERED)
                {
                    listDFS.insert( * pSucc );
                }// if
                else
                {
                    if ((pSuccData->number < pTopData->number) &&
                        (pSuccData->bInStack == TRUE))
                        pTopData->lowlink = MIN( pTopData->lowlink, pSuccData->number );
                }// else
            }// while
        }// else
    }// while
}

//
// order edges after tarjan
//
void
TGraphTool::sortInEdges ( TNode * pNode )
{
    if (pNode == NULL)
    {
        cout << "TGraphTool::sortInEdges (...) : got NULL arg" << endl;
        return;
    }// if

    TNode                 * pOrig;
    TList< TEdge >          listEdges, * pList;
    TListIterator< TEdge >  iterOrigEdges, iterEdges;
    TEdge                 * pOrigEdge, * pEdge;
    uint                   adjNodeName;

    pOrig = (TNode*)(pNode->getInfo());
    
    iterOrigEdges.setList( pOrig->getInEdges() );
    iterEdges.setList( pNode->getInEdges() );

    //
    // catch edges in original order
    //
    iterOrigEdges.first();
    while ((pOrigEdge = iterOrigEdges++) != NULL)
    {
        adjNodeName = pOrigEdge->getTail()->getName();

        iterEdges.first();
        while ((pEdge = iterEdges++) != NULL)
        {
            if (pEdge->getTail()->getName() == adjNodeName)
                listEdges.add( * pEdge );
        }// while
    }// while

    //
    // put ordered edges back to list
    //

    pList = & pNode->getInEdges();
    pList->clear();
    
    iterEdges.setList( listEdges );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pList->add( * pEdge );
    }// while
}

void
TGraphTool::sortOutEdges ( TNode * pNode )
{
    if (pNode == NULL)
    {
        cout << "TGraphTool::sortOutEdges (...) : got NULL arg" << endl;
        return;
    }// if

    TNode                 * pOrig;
    TList< TEdge >          listEdges, * pList;
    TListIterator< TEdge >  iterOrigEdges, iterEdges;
    TEdge                 * pOrigEdge, * pEdge;
    uint                   adjNodeName;

    pOrig = (TNode*)(pNode->getInfo());
    
    iterOrigEdges.setList( pOrig->getOutEdges() );
    iterEdges.setList( pNode->getOutEdges() );

    //
    // catch edges in original order
    //
    iterOrigEdges.first();
    while ((pOrigEdge = iterOrigEdges++) != NULL)
    {
        adjNodeName = pOrigEdge->getHead()->getName();

        iterEdges.first();
        while ((pEdge = iterEdges++) != NULL)
        {
            if (pEdge->getHead()->getName() == adjNodeName)
                listEdges.add( * pEdge );
        }// while
    }// while

    //
    // put ordered edges back to list
    //

    pList = & pNode->getOutEdges();
    pList->clear();
    
    iterEdges.setList( listEdges );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pList->add( * pEdge );
    }// while
}

//////////////////////////////////////////////

//
// do a DFS with the graph, mark each vertex and build a palm
// Assumption : there ex. a path from root to every node in the graph
//
TGraph* 
TGraphTool::dfs ( TGraph& graph, TMarkableNode& root )
{
    TListIterator<TNode>  iter( graph.getNodeList() );
    TMarkableNode        *pNode;
    TGraph               *pPalm = graph.create();

    int   dfsNumber = 0;

    // set marks of all nodes to NO_MARK
    iter.first();
    while ((pNode = (TMarkableNode*) iter++) != NULL)
        pNode->setMark( NO_MARK );

    // root is the first node in the palm
    pNode = root.clone();
    pNode->setMark( dfsNumber );
    root.setInfo( pNode );
    pPalm->add( *pNode );

    // do the dfs
    visit( *pPalm, root, root, dfsNumber );

    return pPalm;
}

//
// recursivly visit a node and all successors
//
void 
TGraphTool::visit ( TGraph& palm, 
                    TMarkableNode& node, TMarkableNode& pred, 
                    int& dfsNumber )
{
    TListIterator<TEdge>  iter( node.getOutEdges() );
    TMarkableEdge        *pEdge, *pNewEdge;
    TMarkableNode        *pNode, *pNewNode, *pOldNode;

    // set new mark
    node.setMark( dfsNumber++ );

    // now check if there is an unvisited successor
    iter.first();
    while ((pEdge = (TMarkableEdge*) iter++) != NULL)
    {
        pNode = (TMarkableNode*) pEdge->getHead();

        if ((pNode->mark() != DELETED) &&
            (pNode->mark() == NO_MARK))
        {
            // create an dag edge in the palm
            pNewNode = pNode->clone();
            pNewNode->setMark( dfsNumber+1 );
            pNode->setInfo( pNewNode );
            palm.add( *pNewNode );

            pOldNode = (TMarkableNode*) node.getInfo();
            pNewEdge = pEdge->create( *pOldNode, *pNewNode );
            pNewEdge->setMark( DAG_EDGE );
            pOldNode->addOutEdge( *pNewEdge );
            pNewNode->addInEdge( *pNewEdge );

            // mark edge in orig graph too
            pEdge->setMark( DAG_EDGE );

            // now go one and do the same with the succ
            visit( palm, *pNode, node, dfsNumber );
        }// if
        else 
            if ((pNode->mark() != DELETED) &&
                ((pNode->mark() < node.mark()) &&
                 (pNode->getName() != pred.getName())))
        {
            // create a backward edge in the palm
            pNewNode = (TMarkableNode*) pNode->getInfo();

            pOldNode = (TMarkableNode*) node.getInfo();
            pNewEdge = pEdge->create( *pOldNode, *pNewNode );
            pNewEdge->setMark( BACKWARD_EDGE );
            pOldNode->addOutEdge( *pNewEdge );
            pNewNode->addInEdge( *pNewEdge );

            // mark edge in orig graph too
            pEdge->setMark( BACKWARD_EDGE );
        }// else
    }// while
}

//////////////////////////////////////////////////////////////////////
//
// strongly connected components and depth first search from LEDA
//

struct TNData
{
    long     dfsnum, compnum;
    bool     reached;
    TNode  * pNewNode;
};

#define NDATA(x) ((TNData*)((x)->pInfo))

void
TGraphTool::dfs ( TNode            & node,
                  int              & count1,
                  int              & count2 )
{
    TStack< TNode >         stack;
    TListIterator< TEdge >  iter;
    TEdge                 * pEdge;
    TNode                 * pNode, * pAdj;

    stack.push( & node );
    while ( ! stack.empty() )
    {
        pNode = stack.head();

        if ( NDATA( pNode )->dfsnum == UNDEF )
        {
            //
            // node was not visited => add all succ to stack
            //

            NDATA( pNode )->dfsnum = count1++;

            iter.setList( pNode->getOutEdges() );
            iter.last();
            while ((pEdge = iter--) != NULL)
            {
                pAdj = pEdge->getHead();
                
                if ( NDATA( pAdj )->dfsnum == UNDEF )
                    stack.push( pEdge->getHead() );
            }// while
        }// if
        else
        {
            //
            // node was visited before => remove it from stack
            //

            if ( NDATA( pNode )->compnum == UNDEF )
                NDATA( pNode )->compnum = count2++;

            stack.pop();
        }// else
    }// while
}

void
TGraphTool::dfs_num ( TGraph  & graph )
{
    int                     count1, count2;
    TListIterator< TNode >  iter;
    TNode                 * pNode;

    count1 = count2 = 0;
    
//     iter.first();
//     while (( pNode = iter++ ) != NULL)
//     {
//         NDATA( pNode )->dfsnum  = UNDEF;
//         NDATA( pNode )->compnum = UNDEF;
//     }// while

    iter.setList( graph.getNodeList() );
    iter.first();
    while (( pNode = iter++ ) != NULL)
    {
        if (NDATA( pNode )->dfsnum == UNDEF) 
            dfs( * pNode, count1, count2 );
    }// while
}

void
TGraphTool::dfs_reach_rev ( TNode           & node,
                            TList< TNode >  & l )
{
    TStack< TNode >         stack;
    
    TListIterator< TEdge >  iter;
    TEdge                 * pEdge;
    TNode                 * pNode, * pAdj;
    
    stack.push( & node );
    while ( ! stack.empty() )
    {
        pNode = stack.head();
        
        if ( ! NDATA( pNode )->reached )
        {
            l.add( * pNode );
            NDATA( pNode )->reached = true;

            iter.setList( pNode->getInEdges() );
            iter.last();
            while ((pEdge = iter--) != NULL)
            {
                pAdj = pEdge->getTail();
                
                if ( ! NDATA( pAdj )->reached )
                    stack.push( pAdj );
            }// while
        }// if
        else
            stack.pop();
    }// while
    
}
                  
TList< TGraph > *
TGraphTool::scc ( TGraph & graph )
{
    int                     i, n;
    int                     count;
    TDynArray< TNode * >    V;

    TListIterator< TNode >  iter;
    TNode                 * pNode;

    TList< TNode >          S;

    
    n = int( graph.getNoOfNodes() );

    //
    // create datafields for each node
    //

    TNData  nodeData[ n ];
    
    iter.setList( graph.getNodeList() );
    iter.first();
    i = 0;
    while ((pNode = iter++) != NULL)
    {
        nodeData[i].dfsnum   = UNDEF;
        nodeData[i].compnum  = UNDEF;
        nodeData[i].reached  = false;
        nodeData[i].pNewNode = NULL;
        
        pNode->setInfo( & (nodeData[i]) );

        i++;
    }// while

    V.setSize( n );

    //
    // set numbers in normal dfs
    //
    
    dfs_num( graph );

    iter.first();
    while ((pNode = iter++) != NULL)
    {
        V[ NDATA( pNode )->compnum ] = pNode;
    }// while

    //
    // now do dfs in transposed graph and adjust component numbers
    //
    
    count = 0;
    iter.setList( S );
    
    for ( i = n-1; i >= 0; i-- )
    {
        if ( ! NDATA( V[i] )->reached ) 
        {
            S.clear();
            dfs_reach_rev( * V[i], S );

            iter.first();
            while ((pNode = iter++) != NULL)
                NDATA( pNode )->compnum = count;
            
            count++;
        }// if
    }// for

    //
    // build each component
    //

    // create empty components

    TGraph                * comp[ count ];
    TList< TGraph >       * pCompList;

    pCompList = new TList< TGraph >;
    
    for ( i = 0; i < count; i++ )
    {
        comp[i] = graph.create();
        pCompList->add( * comp[i] );
    }// for
    
    // create nodes

    TNode  * pNewNode;
    
    iter.setList( graph.getNodeList() );
    iter.first();
    while ((pNode = iter++) != NULL)
    {
        pNewNode = pNode->clone();
        pNewNode->setInfo( pNode );
        NDATA( pNode )->pNewNode = pNewNode;

        comp[ NDATA( pNode )->compnum ]->add( * pNewNode );
    }// while

    // create edges
    TListIterator< TEdge >  iterEdges;
    TEdge                 * pEdge, * pNewEdge;
    TNode                 * pAdj, * pNewAdj;
    long                    compnum;
    
    iter.setList( graph.getNodeList() );
    iter.first();
    while ((pNode = iter++) != NULL)
    {
        pNewNode = NDATA( pNode )->pNewNode;
        compnum  = NDATA( pNode )->compnum;

        iterEdges.setList( pNode->getOutEdges() );
        iterEdges.first();
        while ((pEdge = iterEdges++) != NULL)
        {
            pAdj = pEdge->getHead();
            
            // only add an edge, if same component
            if (NDATA( pAdj )->compnum != compnum)
                continue;
            
            pNewAdj = NDATA( pAdj )->pNewNode;

            pNewEdge = pEdge->create( * pNewNode, * pNewAdj );
            pNewNode->addOutEdge( * pNewEdge );
            pNewAdj->addInEdge( * pNewEdge );
        }// while
    }// while

    //
    // order the edges in each comp. to fit ordering in orig. graph
    //

    TListIterator< TGraph >  iterComp;
    TGraph                 * pComponent;

    iterComp.setList( * pCompList );
    iterComp.first();
    while ((pComponent = iterComp++) != NULL)
    {
        iter.setList( pComponent->getNodeList() );

        iter.first();
        while ((pNode = iter++) != NULL)
        {
            sortInEdges( pNode );
            sortOutEdges( pNode );
        }// while
    }// while

    return pCompList;
}
