//
//  Project     : libgraph
//  File        : heuristicFVS.cc
//  Author      : Ronald Kriemann
//  Purpose     : class for a heuristic FVS algorithm
//
// arch-tag: c00d937a-edff-46a1-9451-8f39432d5d71
//

#include "timer.hh"
#include "var.hh"
#include "TBTree.hh"
#include "TSList.hh"

#include "graphtool.hh"

#include "heuristicFVS.hh"

#ifdef L
#  undef L
#endif

#ifdef D
#  undef D
#endif

// used attributes
#define L  1
#define D  3

// static member variables
bool THeuristicFVS::bLookGlobally = FALSE;
bool THeuristicFVS::bUseTarjan    = FALSE;

//
// constructor and destructor
//
THeuristicFVS::THeuristicFVS ()
{
}

THeuristicFVS::~THeuristicFVS ()
{
}

//
// build fvs of given graph
//
TList< TNode > *
THeuristicFVS::buildFVS ( TGraph & graph )
{
    if ( ! bUseTarjan)
    {
        TGraph         * pGraph = graph.clone();
        TList< TNode > * pFVS;

        THeuristicFVS    localFVS;

        pFVS = localFVS.buildFVS_SCC( pGraph );

        delete pGraph;
        return pFVS;
    }// if
    else
    {
        TGraphTool               graphTool;
        TList< TGraph >        * pListComp;
        TListIterator< TGraph >  iterComp;
        TGraph                 * pComp;
        TList< TNode >         * pLocalFVS, * pFVS;
        TListIterator< TNode >   iterFVS;
        TNode                  * pNode;
        
        pFVS = new TList< TNode >;
        
        //
        // decompose graph into SCC's
        //

        pListComp = graphTool.scc( graph );
        
        if (pListComp == NULL)
        {
            cout << "THeuristicFVS::buildFVS (...) : got NULL SCC-list" << endl;
            return NULL;
        }// if
        
        iterComp.setList( * pListComp );
        
        //
        // build fvs in each SCC
        //
        
        iterComp.first();
        while ((pComp = iterComp++) != NULL)
        {
            if (pComp->getNoOfNodes() > 1)
            {
                THeuristicFVS  localFVS;
                
                pLocalFVS = localFVS.buildFVS_SCC( pComp );
                
                if (pLocalFVS == NULL)
                {
                    cout << "THeuristicFVS::buildFVS (...) : got NULL fvs-list" << endl;
                    continue;
                }// if
                
                iterFVS.setList( * pLocalFVS );
                iterFVS.first();
                while ((pNode = iterFVS++) != NULL)
                {
                    pFVS->add( * pNode );
                }// while
                
                delete pLocalFVS;
            }// if
        }// while
        
        //
        // clean up
        //
        iterComp.first();
        while ((pComp = iterComp++) != NULL)
        {
            delete pComp;
        }// while
        
        delete pListComp;
        
        //
        // finish
        //
        return pFVS;
    }// else
}

TList<TNode>* 
THeuristicFVS::buildFVS_SCC ( TGraph * pGraph )
{
    if (pGraph == NULL)
    {
        cout << "THeuristicFVS::buildFVS_SCC (...) : got NULL arg" << endl;
        return NULL;
    }// if
    
    TListIterator<TNode>          iterGraph;
    TListIterator<TMarkableNode>  iterNodes;
    TList< TMarkableNode >        listNodes;
    TMarkableNode               * pNode;
    TList< TNode >              * plistFVS;
    TItem< TMarkableNode >      * pActItem;

    // initialize pointer to NULL
    pMaxNode  = NULL;
    maxDegree = 0;

    // create list for feedback vertices
    plistFVS = new TList<TNode>;

    //
    // copy nodes from graph into local list and reset attr.
    //
    iterGraph.setList( pGraph->getNodeList() );
    iterGraph.first();
    while ((pNode = (TMarkableNode*) iterGraph++) != NULL)
    {
        listNodes.add( * pNode );
        
        pNode->setMark( L );
    }// while

    iterNodes.setList( listNodes );
    noOfNotDeleted = listNodes.getSize();

    //
    // reduce graph by all non-cycle-nodes
    //

    iterNodes.first();
    while ((pNode = iterNodes++) != NULL)
    {
        if (pNode->mark() == L) transformLoop( pNode, *plistFVS );
        if (pNode->mark() == L) transformZeroPred( *pNode, *plistFVS );
        if (pNode->mark() == L) transformZeroSucc( *pNode, *plistFVS );
        if (pNode->mark() == L) transformOnePred(  *pNode, *plistFVS );
        if (pNode->mark() == L) transformOneSucc(  *pNode, *plistFVS );
    }// while
    
    // 
    // look for fvs nodes
    //
    while (noOfNotDeleted > 0)
    {
        if ((pMaxNode != NULL) && (pMaxNode->mark() == L) && ( ! bLookGlobally )) 
        {
            if (verbosity >= VERBOSE_STEP)
                cout << "THeuristicFVS::buildFVS_SCC (...) : local look" << endl;
            
            addFVS( *pMaxNode, *plistFVS  );
        }// if
        else
        {
            if (verbosity >= VERBOSE_STEP)
                cout << "THeuristicFVS::buildFVS_SCC (...) : global look" << endl;
            
            //
            // look for global-maximal-degree node
            //
            pMaxNode  = NULL;
            maxDegree = 0;

            iterNodes.first();
            while ((pNode = iterNodes()) != NULL)
            {
                pActItem = iterNodes.getItem();
                iterNodes++;

                if (pNode->mark() == D)
                    listNodes.remove( * pActItem );
                else
                {
                    if (maxDegree < pNode->getDegree())
                    {
                        maxDegree = pNode->getDegree();
                        pMaxNode  = pNode;
                    }// if
                }// else
            }// while

            // if no maximal-degree node was found, graph is empty
            if (pMaxNode != NULL)
                addFVS( *pMaxNode, *plistFVS  );
            else
                break;
        }// else
    }// while

    //
    // finish
    //
    
    return plistFVS;
}

//
// set global-flags
//
void
THeuristicFVS::lookGlobally ( const bool & b )
{
    bLookGlobally = b;
}
    
void
THeuristicFVS::useTarjan ( const bool b )
{
    bUseTarjan = b;
}
    
//
// misc. methods for transforming the graph
//

void
THeuristicFVS::transformLoop ( TMarkableNode * pNode, TList<TNode> &listFVS )
{
    if (pNode == NULL)
    {
        cout << "THeuristicFVS::transformLoop (...) : got NULL arg" << endl;
        return;
    }// if

    TListIterator< TEdge >  iterEdges;
    TEdge                 * pEdge;
    uint                   nodeName;

    nodeName = pNode->getName();
    
    iterEdges.setList( pNode->getOutEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        // look for loop
        if (pEdge->getHead()->getName() == nodeName)
        {
            // add node to FVS
            addFVS( * pNode, listFVS );
            break;
        }// if
    }// if
}

void 
THeuristicFVS::transformZeroPred ( TMarkableNode & node, TList<TNode> &listFVS )
{
    if (node.getInDegree() != 0)
        return;
    
    TListIterator<TEdge>  iterEdges( node.getOutEdges() );
    TEdge                *pEdge;
    TList<TNode>          listNodes;
    TListIterator<TNode>  iterNodes( listNodes );
    TMarkableNode        *pNode;
    
    // add all succ to list
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        listNodes.add( *(pEdge->getHead()) );
    }// while
    
    // ``remove'' node from graph
    node.isolate();
    node.setMark( D );
    noOfNotDeleted--;
    
    // look if we can do the same with a succ
    iterNodes.first();
    while ((pNode = (TMarkableNode*) iterNodes++) != NULL)
    {
        if (pNode->mark() != D)
            transformZeroPred( *pNode, listFVS  );
        if (pNode->mark() != D) 
            transformOnePred( *pNode, listFVS );
    }// if
}

void 
THeuristicFVS::transformZeroSucc ( TMarkableNode & node, TList<TNode> &listFVS )
{
    if (node.getOutDegree() != 0)
        return;
    
    TListIterator<TEdge>  iterEdges( node.getInEdges() );
    TEdge                *pEdge;
    TList<TNode>          listNodes;
    TListIterator<TNode>  iterNodes( listNodes );
    TMarkableNode        *pNode;

    // add all pred to list
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        listNodes.add( *(pEdge->getHead()) );
    }// while
    
    // ``remove'' node from graph
    node.isolate();
    node.setMark( D );
    noOfNotDeleted--;
    
    // look if we can do the same with a succ
    iterNodes.first();
    while ((pNode = (TMarkableNode*) iterNodes++) != NULL)
    {
        if (pNode->mark() != D)
            transformZeroSucc( *pNode, listFVS  );
        if (pNode->mark() != D) 
            transformOneSucc( *pNode, listFVS );
    }// if
}

void 
THeuristicFVS::transformOnePred ( TMarkableNode & node, TList<TNode> &listFVS )
{
    TMarkableNode        *pSucc, *pPred;
    TListIterator<TEdge>  iterEdges( node.getOutEdges() );
    TEdge                *pEdge, *pNewEdge;

    pPred = NULL;

    // update maximal-degree node
    if ((node.getInDegree() > 1) && (node.getDegree() >= maxDegree) && (node.mark() == L))
    {
        maxDegree = node.getDegree();
        pMaxNode  = &node;
    }// if

    if ((node.getInDegree() != 1) || (node.mark() != L))
        return;
    
    pPred = (TMarkableNode*) node.getInEdges().getFirst()->getTail();

    // add all pred from node to succ
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pSucc = (TMarkableNode*) pEdge->getHead();

        // add edge only, if not already present
        if (pPred->isSucc( * pSucc ))
            continue;
        
        // create new edge, because old will be deleted
        pNewEdge = pEdge->clone( *pPred, *pSucc );
        
        // remove edge to node
        pSucc->remPred( node );
        node.remSucc( *pSucc );
        // list was changed -> reset
        iterEdges.first();
        
        // and create edge to succ of node
        pPred->addOutEdge( *pNewEdge );
        pSucc->addInEdge( *pNewEdge );
    }// while

    // ``remove'' node from graph
    node.isolate();
    node.setMark( D );
    noOfNotDeleted--;
    
    // check if pred has self-loop -> add to fvs
    transformLoop( pPred, listFVS );

    if ( pPred->mark() == L )
        transformOnePred( *pPred, listFVS );
}

void 
THeuristicFVS::transformOneSucc ( TMarkableNode & node, TList<TNode> &listFVS )
{
    TMarkableNode        *pSucc, *pPred;
    TListIterator<TEdge>  iterEdges( node.getInEdges() );
    TEdge                *pEdge, *pNewEdge;

    pSucc = NULL;

    if ((node.getOutDegree() > 1) && (node.getDegree() >= maxDegree) && (node.mark() == L))
    {
        maxDegree = node.getDegree();
        pMaxNode  = &node;
    }// if

    if ((node.getOutDegree() != 1) || (node.mark() != L))
        return;
    
    pSucc = (TMarkableNode*) node.getOutEdges().getFirst()->getHead();

    // add all pred from node to succ
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pPred = (TMarkableNode*) pEdge->getTail();
        
        // add edge only, if not already present
        if (pSucc->isPred( * pPred ))
            continue;
        
        // create new edge, because old will be deleted
        pNewEdge = pEdge->clone( *pPred, *pSucc );
        
        // remove edge to node
        pPred->remSucc( node );
        node.remPred( *pPred );
        // list was changed -> reset
        iterEdges.first();
        
        // and create edge to succ of node
        pPred->addOutEdge( *pNewEdge );
        pSucc->addInEdge( *pNewEdge );
    }// while
    
    // ``remove'' node from graph
    node.isolate();
    node.setMark( D );
    noOfNotDeleted--;
    
    // check if succ has self-loop -> add to fvs
    transformLoop( pSucc, listFVS );

    if ( pSucc->mark() == L )
        transformOneSucc( *pSucc, listFVS );
}

void 
THeuristicFVS::addFVS ( TMarkableNode & node, TList<TNode> & listFVS )
{
    TListIterator<TEdge>  iterEdges;
    TEdge                *pEdge;
    TList<TNode>          listSucc, listPred;
    TListIterator<TNode>  iterNodes;
    TMarkableNode        *pNode;
    
    //
    // collect pred and succ of node
    //
    iterEdges.setList( node.getInEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        listPred.add( *(pEdge->getTail()) );
    }// while

    iterEdges.setList( node.getOutEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        listSucc.add( *(pEdge->getHead()) );
    }// while

    // ``remove'' node from graph
    node.isolate();
    node.setMark( D );
    noOfNotDeleted--;

    node.addReference();
    listFVS.add( node );

    if (pMaxNode == &node)
    {
        pMaxNode  = NULL;
        maxDegree = 0;
    }// if

    //
    // now check for ``useless'' succ (pred)
    //
    iterNodes.setList( listSucc );
    iterNodes.first();
    while ((pNode = (TMarkableNode*) iterNodes++) != NULL)
    {
        if (pNode->mark() == L) transformZeroPred( *pNode, listFVS );
        if (pNode->mark() == L) transformZeroSucc( *pNode, listFVS );
        if (pNode->mark() == L) transformOnePred(  *pNode, listFVS );
        if (pNode->mark() == L) transformOneSucc(  *pNode, listFVS );
    }// while

    iterNodes.setList( listPred );
    iterNodes.first();
    while ((pNode = (TMarkableNode*) iterNodes++) != NULL)
    {
        if (pNode->mark() == L) transformZeroPred( *pNode, listFVS );
        if (pNode->mark() == L) transformZeroSucc( *pNode, listFVS );
        if (pNode->mark() == L) transformOnePred(  *pNode, listFVS );
        if (pNode->mark() == L) transformOneSucc(  *pNode, listFVS );
    }// while
}
