//
//  Project     : libgraph
//  File        : planarBlock.cc
//  Author      : Ronald Kriemann
//  Purpose     : class building cycles (blocks) in a planar graph
//
// arch-tag: 6f1f8338-a5d5-47a8-bc73-c258cfc140dc
//

#include "var.hh"
#include "drawapp.hh"

#include "graphtool.hh"

#include "planarBlock.hh"

////////////////////////////////////////////////
//
// constructor and destructor
//

TPlanarBlock::TPlanarBlock ()
{}

TPlanarBlock::~TPlanarBlock ()
{}

////////////////////////////////////////////////
//
// methods for planar block algorithm
//
    
//
// reduce graph by combining all nodes with weak edges
//
void 
TPlanarBlock::reduce ( T2DPlanarGraph &graph )
{
    TListIterator<TNode>    iterNodes( graph.getNodeList() );
    T2DPlanarNode          *pNode;
    TList<TNode>            listAdjNodes;
    TListIterator<TNode>    iterAdjNodes( listAdjNodes );

    if (verbosity >= VERBOSE_STEP)
        cout << "reducing graph" << endl;

    //
    // remove all weak edges, if an opposite strong edge exists
    //
    iterNodes.first();
    while ((pNode = (T2DPlanarNode*) iterNodes++) != NULL)
    {
        removeNonRealWeak( *pNode );
    }// while
    
    if (verbosity >= VERBOSE_RESULT)
        TDrawApp::draw( graph, "graph" );

    iterNodes.first();
    while ((pNode = (T2DPlanarNode*) iterNodes++) != NULL)
    {
        sortWeakOFD( *pNode );
    }// while

    //
    // go through all nodes and combine those, which have
    // weak connections (edges)
    //
    iterNodes.first();
    while ((pNode = (T2DPlanarNode*) iterNodes()) != NULL)
    {
        if ((pNode->isDeleted( MIN_CYC ) == FALSE) &&
            hasStrongInEdge(*pNode) && hasStrongOutEdge(*pNode))
        {
            reduceNode( graph, *pNode );
        }// if

        iterNodes++;
    }// while

    //
    // check if weak edges survived and melt nodes
    //
    iterNodes.first();
    while ((pNode = (T2DPlanarNode*) iterNodes()) != NULL)
    {
        if ((pNode->isDeleted( MIN_CYC ) == FALSE) && hasWeakEdge( *pNode ))
        {
            //            reduceWeakNode( graph, *pNode );
            removeWeak( *pNode );
        }// if
        
        iterNodes++;
    }// while

    //
    // remove all "virtually" removed nodes
    //
    TItem< TNode > * pItem;
    
    iterNodes.first();
    while ((pNode = (T2DPlanarNode*) iterNodes()) != NULL)
    {
        pItem = iterNodes.getItem();
        iterNodes++;
        
        if ((pNode->isDeleted( MIN_CYC ) == TRUE) && 
            ((pNode->getInDegree() + pNode->getOutDegree()) == 0))
        {
            graph.getNodeList().remove( * pItem );
            if (pNode->delReference())
                delete pNode;
        }// if
        else
            pNode->setDeleted( FALSE, MIN_CYC );
    }// while

    if (verbosity >= VERBOSE_RESULT)
        TDrawApp::draw( graph, "graph" );
}

//
// reduce single node
//
void 
TPlanarBlock::reduceNode ( T2DPlanarGraph &graph,
                           T2DPlanarNode &node )
{
    T2DPlanarNode          *pSucc, *pAdjNode;
    TWeightedMarkableEdge  *pEdge, *pWeakEdge, *pSuccEdge;
    TListIterator<TEdge>    iterEdges, iterSuccEdges;
    bool                    bHasWeakRight, bHasWeakLeft;
    TList<TNode>            listAdjNodes;
    TListIterator<TNode>    iterAdjNodes( listAdjNodes );

    // add all adj nodes to list
    iterEdges.setList( node.getInEdges() );
    iterEdges.first();
    while ((pEdge = (TWeightedMarkableEdge*) iterEdges++) != NULL)
    {
        pSucc = (T2DPlanarNode*) pEdge->getTail();
        
        if ((pEdge->weight() != 0) &&
            (pSucc->isDeleted( MIN_CYC ) == FALSE))
            listAdjNodes.add( *pSucc );
    }// while
    iterEdges.setList( node.getOutEdges() );
    iterEdges.first();
    while ((pEdge = (TWeightedMarkableEdge*) iterEdges++) != NULL)
    {
        pSucc = (T2DPlanarNode*) pEdge->getHead();
        
        if ((pEdge->weight() != 0) &&
            (pSucc->isDeleted( MIN_CYC ) == FALSE))
            listAdjNodes.add( *pSucc );
    }// while

    if (verbosity >= VERBOSE_RESULT)
        cout << "reducing node : " << node.getName() << endl;

    //
    // try to melt all nodes on right side
    //
    bHasWeakRight = TRUE;
    while (bHasWeakRight)
    {
        //
        // find a suitable edge on the right side
        //
        pWeakEdge = NULL;
        iterEdges.setList( node.getOutEdges() );
        iterEdges.first();
        while ((pEdge = (TWeightedMarkableEdge*) iterEdges++) != NULL)
        {
            pSucc = (T2DPlanarNode*) pEdge->getHead();

            if (pEdge->weight() != 0)
            {
                iterEdges.last();
                iterEdges++;
            }// if
            else if (pSucc->isDeleted( MIN_CYC ) == FALSE)
            {
                // check if edge is on left side at succ
                iterSuccEdges.setList( pSucc->getInEdges() );
                iterSuccEdges.first();
                while ((pSuccEdge = (TWeightedMarkableEdge*) iterSuccEdges++) != NULL)
                {
                    if (pSuccEdge->weight() != 0)
                    {
                        iterSuccEdges.last();
                        iterSuccEdges++;
                    }// if
                    else if (pSuccEdge->getTail()->getName() == node.getName())
                    {
                        pWeakEdge = pEdge;
                        iterSuccEdges.last();
                        iterSuccEdges++;
                        iterEdges.last();
                        iterEdges++;
                    }// if
                }// while
            }// else
        }// while

        if (pWeakEdge == NULL)
            bHasWeakRight = FALSE;
        else
        {
            pSucc = (T2DPlanarNode*) pWeakEdge->getHead();

            //
            // remove all weak edges on right side of node
            // and on left side of succ
            //
            iterEdges.first();
            while ((pEdge = (TWeightedMarkableEdge*) iterEdges++) != NULL)
            {
                if (pEdge->weight() != 0)
                {
                    iterEdges.last();
                    iterEdges++;
                }// if
                else
                {
                    node.remSucc( *(pEdge->getHead()) );
                    pEdge->getHead()->remPred( node );
                    node.remPred( *(pEdge->getHead()) );
                    pEdge->getHead()->remSucc( node );
                }// if
            }// if

            iterSuccEdges.first();
            while ((pEdge = (TWeightedMarkableEdge*) iterSuccEdges++) != NULL)
            {
                if (pEdge->weight() != 0)
                {
                    iterSuccEdges.last();
                    iterSuccEdges++;
                }// if
                else
                {
                    pSucc->remPred( *(pEdge->getTail()) );
                    pEdge->getTail()->remSucc( *pSucc );
                    pSucc->remSucc( *(pEdge->getTail()) );
                    pEdge->getTail()->remPred( *pSucc );
                }// if
            }// if

            // melt nodes
            if (verbosity >= VERBOSE_RESULT)
                cout << "melting nodes " << node.getName() 
                     << " and " << pSucc->getName() << endl;
            meltNodesRight( node, *pSucc );

            // add succ to blocklist
            node.addRightToBlockList( pSucc );

            // remove melted node from graph
            pSucc->isolate();
            pSucc->setDeleted( TRUE, MIN_CYC );
            //            graph.remove( *pSucc );
        }// else
    }// while

    //
    // try to melt all nodes on left side
    //
    bHasWeakLeft = TRUE;
    while (bHasWeakLeft)
    {
        //
        // find a suitable edge on the left side
        //
        pWeakEdge = NULL;
        iterEdges.setList( node.getOutEdges() );
        iterEdges.last();
        while ((pEdge = (TWeightedMarkableEdge*) iterEdges--) != NULL)
        {
            pSucc = (T2DPlanarNode*) pEdge->getHead();

            if (pEdge->weight() != 0)
            {
                iterEdges.first();
                iterEdges--;
            }// if
            else if (pSucc->isDeleted( MIN_CYC ) == FALSE)
            {
                // check if edge is on left side at succ
                iterSuccEdges.setList( pSucc->getInEdges() );
                iterSuccEdges.last();
                while ((pSuccEdge = (TWeightedMarkableEdge*) iterSuccEdges--) != NULL)
                {
                    if (pSuccEdge->weight() != 0)
                    {
                        iterSuccEdges.first();
                        iterSuccEdges--;
                    }// if
                    else if (pSuccEdge->getTail()->getName() == node.getName())
                    {
                        pWeakEdge = pEdge;
                        iterSuccEdges.first();
                        iterSuccEdges--;
                        iterEdges.first();
                        iterEdges--;
                    }// if
                }// while
            }// else
        }// while

        if (pWeakEdge == NULL)
            bHasWeakLeft = FALSE;
        else
        {
            pSucc = (T2DPlanarNode*) pWeakEdge->getHead();

            //
            // remove all weak edges on left side of node
            // and on right side of succ
            //
            iterEdges.last();
            while ((pEdge = (TWeightedMarkableEdge*) iterEdges--) != NULL)
            {
                if (pEdge->weight() != 0)
                {
                    iterEdges.first();
                    iterEdges--;
                }// if
                else
                {
                    node.remSucc( *(pEdge->getHead()) );
                    pEdge->getHead()->remPred( node );
                    node.remPred( *(pEdge->getHead()) );
                    pEdge->getHead()->remSucc( node );
                }// if
            }// if

            iterSuccEdges.last();
            while ((pEdge = (TWeightedMarkableEdge*) iterSuccEdges--) != NULL)
            {
                if (pEdge->weight() != 0)
                {
                    iterSuccEdges.first();
                    iterSuccEdges--;
                }// if
                else
                {
                    pSucc->remPred( *(pEdge->getTail()) );
                    pEdge->getTail()->remSucc( *pSucc );
                    pSucc->remSucc( *(pEdge->getTail()) );
                    pEdge->getTail()->remPred( *pSucc );
                }// if
            }// if

            // melt nodes
            if (verbosity >= VERBOSE_RESULT)
                cout << "melting nodes " << node.getName() 
                     << " and " << pSucc->getName() << endl;
            meltNodesLeft( node, *pSucc );

            // add succ to blocklist
            node.addLeftToBlockList( pSucc );

            // remove melted node from graph
            pSucc->isolate();
            pSucc->setDeleted( TRUE, MIN_CYC );
            //            graph.remove( *pSucc );
        }// else
    }// while

    //
    // node should not be melted anymore
    //
    node.setDeleted( TRUE, MIN_CYC );

    if (verbosity >= VERBOSE_RESULT)
        TDrawApp::draw( graph, "graph" );

    // now care about adj nodes
    iterAdjNodes.first();
    while ((pAdjNode = (T2DPlanarNode*) iterAdjNodes++) != NULL)
    {
        if ((pAdjNode->isDeleted( MIN_CYC ) == FALSE) &&
            hasStrongInEdge(*pAdjNode) && hasStrongOutEdge(*pAdjNode))
        {
            reduceNode( graph, *pAdjNode );
        }// if
    }// while
}

//
// melt nodes connected by a weak edge (node has not one in- or one out-edge)
//
void 
TPlanarBlock::reduceWeakNode ( T2DPlanarGraph &graph, 
                               T2DPlanarNode &node )
{
    //
    // if a weak edge has survived, it can only connect two nodes
    // which do not have one strong in- or one strong out-edge
    // so we can safely melt nodes until no weak edges left
    //

    TListIterator<TEdge>    iterEdges, iterSuccEdges;
    TWeightedMarkableEdge  *pEdge;
    T2DPlanarNode          *pSucc;

    if (verbosity >= VERBOSE_RESULT)
        cout << "TPlanarBlock::reduceWeakNode (...) : "
             << "reducing (weak) node " << node.getName()
             << endl;

    iterEdges.setList( node.getInEdges() );
    iterEdges.first();
    while ((pEdge = (TWeightedMarkableEdge*) iterEdges++) != NULL)
    {
        if (pEdge->weight() == 0)
        {
            pSucc = (T2DPlanarNode*) pEdge->getTail();

            node.remPred( *pSucc );
            pSucc->remSucc( node );
            node.remSucc( *pSucc );
            pSucc->remPred( node );

            // melt nodes
            meltNodesLeft( node, *pSucc );

            // add succ to blocklist
            node.addLeftToBlockList( pSucc );

            // remove melted node from graph
            pSucc->isolate();
            pSucc->setDeleted( TRUE, MIN_CYC );
            //            graph.remove( *pSucc );

            iterEdges.last();
            iterEdges++;
        }// if
    }// while

    //
    // node should not be melted anymore
    //
    node.setDeleted( TRUE, MIN_CYC );

    if (verbosity >= VERBOSE_RESULT)
        TDrawApp::draw( graph, "graph" );
}

//
// remove all weak edges if an opposite strong edge exists
//
void 
TPlanarBlock::removeNonRealWeak ( T2DPlanarNode &node )
{
    TListIterator<TEdge>     iterEdges, iterAdjEdges;
    TWeightedMarkableEdge   *pEdge, *pAdjEdge;

    iterEdges.setList( node.getInEdges() );
    iterEdges.first();
    while ((pEdge = (TWeightedMarkableEdge*) iterEdges++) != NULL)
    {
        if (pEdge->weight() == 0)
        {
            if (pEdge->getTail()->isPred( node ) == TRUE)
            {
                iterAdjEdges.setList( pEdge->getTail()->getInEdges() );
                iterAdjEdges.first();
                while ((pAdjEdge = (TWeightedMarkableEdge*) iterAdjEdges++) != NULL)
                {
                    if ((pAdjEdge->getTail()->getName() == node.getName()) &&
                        (pAdjEdge->weight() != 0))
                    {
                        node.remPred( *(pEdge->getTail()) );
                        pEdge->getTail()->remSucc( node );
                        iterAdjEdges.last();
                        iterAdjEdges++;
                    }// if
                }// while
            }// if
            else
            {
                node.remPred( *(pEdge->getTail()) );
                pEdge->getTail()->remSucc( node );
            }// else
        }// if
    }// while

    iterEdges.setList( node.getOutEdges() );
    iterEdges.first();
    while ((pEdge = (TWeightedMarkableEdge*) iterEdges++) != NULL)
    {
        if (pEdge->weight() == 0)
        {
            if (pEdge->getHead()->isSucc( node ) == TRUE)
            {
                iterAdjEdges.setList( pEdge->getHead()->getOutEdges() );
                iterAdjEdges.first();
                while ((pAdjEdge = (TWeightedMarkableEdge*) iterAdjEdges++) != NULL)
                {
                    if ((pAdjEdge->getHead()->getName() == node.getName()) &&
                        (pAdjEdge->weight() != 0))
                    {
                        node.remSucc( *(pEdge->getHead()) );
                        pEdge->getHead()->remPred( node );
                        iterAdjEdges.last();
                        iterAdjEdges++;
                    }// if
                }// while
            }// if
            else
            {
                node.remSucc( *(pEdge->getHead()) );
                pEdge->getHead()->remPred( node );
            }// else
        }// if
    }// while
}

//
// sort edges according to ofd-cond
//
void 
TPlanarBlock::sortWeakOFD ( T2DPlanarNode &node )
{
    uint   i,j;

    T2DPlanarNode        *pSucc, *pPred;
    TListIterator<TEdge>  iterEdges;
    TEdge                *pEdge, *pInEdge;
    T2DVector             vecEdge, vecNormEdge, vecInEdge, vecOutEdge;
    uint                 nodeName = node.getName();

    TList<TWeightedMarkableEdge>          listOrderedEdges;
    TListIterator<TWeightedMarkableEdge>  iterOrderedEdges( listOrderedEdges );
    TWeightedMarkableEdge                *pWeightedEdge;
    TWeightedMarkableEdge                *pLastOutEdge, *pLastInEdge, *pFirstOutEdge;
    bool                                  bFoundLast, bFoundFirst;

    // these arrays are for direct accessing the in- and out-edges
    // of every node and for the angles the have to the pos. x-axis
    TDynArray<TEdge*>  arrEdges( 5, 5 );
    TDynArray<TFloat>  arrAngles( 5, 5 );
    TDynArray<uint>   arrOrder( 5, 5 );
    uint               noOfInEdges, noOfOutEdges;

    TFloat             alpha, alpha2;

    //
    // first order out-edges
    //
    if ((node.getInDegree() > 0) && (node.getOutDegree() > 0))
    {
        noOfInEdges  = node.getInDegree();
        noOfOutEdges = node.getOutDegree();
        
        arrEdges.setSize( noOfOutEdges + noOfInEdges );
        arrAngles.setSize( noOfOutEdges + noOfInEdges );
        arrOrder.setSize( noOfOutEdges + noOfInEdges );

        //
        // go over all edges and compute the angle between them
        // and the first in-edge
        //

        pInEdge = NULL; //node.getInEdges().getFirst();;
        iterEdges.setList( node.getInEdges() );
        iterEdges.first();
        while ((pEdge = iterEdges++) != NULL)
        {
            if (((TWeightedMarkableEdge*) pEdge)->weight() != 0)
            {
                pInEdge = pEdge;
                iterEdges.last();
                iterEdges++;
            }// if
            else
                pInEdge = pEdge;
        }// while

        if (pInEdge != NULL)
        {
            vecInEdge  = *(((T2DPlanarNode*) (pInEdge->getTail()))->vector());
            vecInEdge -= *(node.vector());

            i = 0;
            iterEdges.setList( node.getOutEdges() );
            iterEdges.first();
            while ((pEdge = iterEdges++) != NULL)
            {
                pSucc    = (T2DPlanarNode*) pEdge->getHead();
                vecEdge  = *(pSucc->vector());
                vecEdge -= *(node.vector());

                vecNormEdge[X] =  vecEdge[Y];
                vecNormEdge[Y] = -vecEdge[X];

                arrEdges[i] = pEdge;

                alpha  = (vecInEdge * vecEdge) / 
                         (vecEdge.euclidNorm() * vecInEdge.euclidNorm());
                // check for unpleasant values
                if (alpha < -1.0) alpha = -1.0;
                if (alpha >  1.0) alpha =  1.0;
                alpha  = acos(alpha);

                alpha2 = (vecNormEdge * vecInEdge) /
                         (vecNormEdge.euclidNorm() * vecInEdge.euclidNorm());
                // check for unpleasant values
                if (alpha2 < -1.0) alpha2 = -1.0;
                if (alpha2 >  1.0) alpha2 =  1.0;
                alpha2  = acos(alpha2);

                // adjusting the angle to have angles in [0,PI2]
                if (alpha2 > M_PI_2)
                    arrAngles[i] = PI2 - alpha;
                else
                    arrAngles[i] = alpha;

                i++;
            }// while
            
            iterEdges.setList( node.getInEdges() );
            iterEdges.first();
            while ((pEdge = iterEdges++) != NULL)
            {
                pPred    = (T2DPlanarNode*) pEdge->getTail();
                vecEdge  = *(pPred->vector());
                vecEdge -= *(node.vector());

                vecNormEdge[X] =  vecEdge[Y];
                vecNormEdge[Y] = -vecEdge[X];

                arrEdges[i] = pEdge;

                alpha  = (vecEdge * vecInEdge) / 
                         (vecEdge.euclidNorm() * vecInEdge.euclidNorm());
                // check for unpleasant values
                if (alpha < -1.0) alpha = -1.0;
                if (alpha >  1.0) alpha =  1.0;
                alpha  = acos(alpha);

                alpha2 = (vecInEdge * vecNormEdge) /
                         (vecNormEdge.euclidNorm() * vecInEdge.euclidNorm());
                // check for unpleasant values
                if (alpha2 < -1.0) alpha2 = -1.0;
                if (alpha2 >  1.0) alpha2 =  1.0;
                alpha2  = acos(alpha2);

                // adjusting the angle to have angles in [0,PI2]
                if (alpha2 > M_PI_2)
                    arrAngles[i] = PI2 - alpha;
                else
                    arrAngles[i] = alpha;

                i++;
            }// while

            //
            // sort all edges according to angle into single list
            //
            for (i = 0; i < noOfOutEdges + noOfInEdges; i++)
            {
                alpha = arrAngles[i];
                arrOrder[i] = 0;

                for (j = 0; j < noOfOutEdges + noOfInEdges; j++)
                {
                    if ((i != j) && (arrAngles[j] < alpha))
                        arrOrder[i]++;
                }// for
            }// for
            
            i = 0;
            while (i <= (noOfOutEdges + noOfInEdges))
            {
                for (j = 0; j < noOfOutEdges + noOfInEdges; j++)
                {
                    if (arrOrder[j] == i)
                    {
                        pWeightedEdge = (TWeightedMarkableEdge*) arrEdges[j];
                        listOrderedEdges.add( *pWeightedEdge );
                    }// if
                }// for

                i++;
            }// while

            //
            // be sure, that the in-edges come first and out-edges come last
            //

            pLastOutEdge = NULL;
            iterOrderedEdges.last();
            while ((pWeightedEdge = iterOrderedEdges--) != NULL)
            {
                if ((pWeightedEdge->weight() != 0) &&
                    (pWeightedEdge->getTail()->getName() == nodeName))
                {
                    pLastOutEdge = pWeightedEdge;
                    iterOrderedEdges.first();
                    iterOrderedEdges--;
                }// if
            }// while

            if (pLastOutEdge != NULL)
            {
                iterOrderedEdges.last();
                while ((pWeightedEdge = iterOrderedEdges--) != NULL)
                {
                    if (pWeightedEdge == pLastOutEdge)
                    {
                        iterOrderedEdges.first();
                        iterOrderedEdges--;
                    }// if
                    else if (pWeightedEdge->getHead()->getName() == nodeName)
                    {
                        listOrderedEdges.remove( *pWeightedEdge );
                        listOrderedEdges.insert( *pWeightedEdge );
                    }// else if
                }// while
            }// if

            //
            // check ofd condition and (if wanted) remove all weak edges
            // if they do not fullfill it
            //

            pLastInEdge = NULL;
            iterOrderedEdges.last();
            while ((pWeightedEdge = iterOrderedEdges--) != NULL)
            {
                if ((pWeightedEdge->weight() != 0) &&
                    (pWeightedEdge->getHead()->getName() == nodeName))
                {
                    pLastInEdge = pWeightedEdge;
                    iterOrderedEdges.first();
                    iterOrderedEdges--;
                }// if
            }// while
            
            if (pLastInEdge != NULL)
            {
                bFoundFirst = FALSE;
                iterOrderedEdges.first();
                while ((pWeightedEdge = iterOrderedEdges++) != NULL)
                {
                    if (pWeightedEdge == pLastInEdge)
                    {
                        iterOrderedEdges.last();
                        iterOrderedEdges++;
                    }// if
                    else if (bFoundFirst == FALSE)
                    {
                        if ((pWeightedEdge->weight() != 0) &&
                            (pWeightedEdge->getHead()->getName() == nodeName))
                        {
                            bFoundFirst = TRUE;
                        }// if
                    }// else if
                    else if (pWeightedEdge->getTail()->getName() == nodeName)
                    {
                        // we have an out-edge in the in-edge-sector
                        if (pWeightedEdge->weight() == 0)
                        {
                            listOrderedEdges.remove( *pWeightedEdge );
                            pWeightedEdge->getHead()->remPred( node );
                            if (pWeightedEdge->delReference()) delete pWeightedEdge;
                        }// if
                        else
                        {
                            cerr << "T2DPlanarNode::sortOFD () : ofd-cond violated in node "
                                 << nodeName << endl;
                        }// else
                    }// else if
                }// while
            }// if

            pFirstOutEdge = NULL;
            iterOrderedEdges.first();
            while ((pWeightedEdge = iterOrderedEdges++) != NULL)
            {
                if ((pWeightedEdge->weight() != 0) &&
                    (pWeightedEdge->getTail()->getName() == nodeName))
                {
                    pFirstOutEdge = pWeightedEdge;
                    iterOrderedEdges.last();
                    iterOrderedEdges++;
                }// if
            }// while

            if (pFirstOutEdge != NULL)
            {
                bFoundLast = FALSE;
                iterOrderedEdges.last();
                while ((pWeightedEdge = iterOrderedEdges--) != NULL)
                {
                    if (pWeightedEdge == pFirstOutEdge)
                    {
                        iterOrderedEdges.first();
                        iterOrderedEdges--;
                    }// if
                    else if (bFoundLast == FALSE)
                    {
                        if ((pWeightedEdge->weight() != 0) &&
                            (pWeightedEdge->getTail()->getName() == nodeName))
                        {
                            bFoundLast = TRUE;
                        }// if
                    }// else if
                    else if (pWeightedEdge->getHead()->getName() == nodeName)
                    {
                        // we have an in-edge in the out-edge-sector
                        if (pWeightedEdge->weight() == 0)
                        {
                            listOrderedEdges.remove( *pWeightedEdge );
                            pWeightedEdge->getHead()->remSucc( node );
                            if (pWeightedEdge->delReference()) delete pWeightedEdge;
                        }// if
                        else
                        {
                            cerr << "T2DPlanarNode::sortOFD () : ofd-cond violated in node "
                                 << nodeName << endl;
                        }// else
                    }// else if
                }// while
            }// if

            //
            // now sort edges back into in- and out-lists
            //
            node.getOutEdges().clear();
            node.getInEdges().clear();

            iterOrderedEdges.first();
            while ((pWeightedEdge = iterOrderedEdges++) != NULL)
            {
                if (pWeightedEdge->getHead()->getName() == nodeName)
                    node.getInEdges().add( *pWeightedEdge );
                else if (pWeightedEdge->getTail()->getName() == nodeName)
                    node.getOutEdges().add( *pWeightedEdge );
            }// while
        }// if
    }// if
}

//
// melt nodes on right side
//
void 
TPlanarBlock::meltNodesRight ( T2DPlanarNode &node,
                               T2DPlanarNode &succ )
{
    TListIterator<TEdge>   iterEdges;
    TEdge                 *pEdge, *pNewEdge;

    //
    // add out-edges of succ to node
    //
    iterEdges.setList( succ.getOutEdges() );
    iterEdges.last();
    while ((pEdge = iterEdges--) != NULL)
    {
        if (node.isSucc( *(pEdge->getHead()) ) == FALSE)
        {
            pNewEdge = pEdge->clone( node, *(pEdge->getHead()) );

            pNewEdge->addReference();
            node.getOutEdges().insert( *pNewEdge );

            pNewEdge->addReference();
            pEdge->getHead()->getInEdges().insertAfter( *pEdge, *pNewEdge );
        }// if
    }// while

    //
    // add in-edges of succ to node
    //
    iterEdges.setList( succ.getInEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        if (node.isPred( *(pEdge->getTail()) ) == FALSE)
        {
            pNewEdge = pEdge->clone( *(pEdge->getTail()), node );

            pNewEdge->addReference();
            node.getInEdges().add( *pNewEdge );
                        
            pNewEdge->addReference();
            pEdge->getTail()->getOutEdges().insertAfter( *pEdge, *pNewEdge );
        }// if
    }// while
}

void 
TPlanarBlock::meltNodesLeft ( T2DPlanarNode &node,
                              T2DPlanarNode &succ )
{
    TListIterator<TEdge>   iterEdges;
    TEdge                 *pEdge, *pNewEdge;

    //
    // add out-edges of succ to node
    //
    iterEdges.setList( succ.getOutEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        if (node.isSucc( *(pEdge->getHead()) ) == FALSE)
        {
            pNewEdge = pEdge->clone( node, *(pEdge->getHead()) );

            pNewEdge->addReference();
            node.getOutEdges().add( *pNewEdge );

            pNewEdge->addReference();
            pEdge->getHead()->getInEdges().insertAfter( *pEdge, *pNewEdge );
        }// if
    }// while

    //
    // add in-edges of succ to node
    //
    iterEdges.setList( succ.getInEdges() );
    iterEdges.last();
    while ((pEdge = iterEdges--) != NULL)
    {
        if (node.isPred( *(pEdge->getTail()) ) == FALSE)
        {
            pNewEdge = pEdge->clone( *(pEdge->getTail()), node );

            pNewEdge->addReference();
            node.getInEdges().insert( *pNewEdge );
                        
            pNewEdge->addReference();
            pEdge->getTail()->getOutEdges().insertAfter( *pEdge, *pNewEdge );
        }// if
    }// while
}

//
// check if node has at least one strong in (out) edge
//
bool 
TPlanarBlock::hasStrongInEdge ( T2DPlanarNode &node )
{
    TListIterator<TEdge>    iterEdges;
    TWeightedMarkableEdge  *pEdge;

    iterEdges.setList( node.getInEdges() );
    iterEdges.first();
    while ((pEdge = (TWeightedMarkableEdge*) iterEdges++) != NULL)
    {
        if (pEdge->weight() != 0)
            return TRUE;
    }// while

    return FALSE;
}

bool 
TPlanarBlock::hasStrongOutEdge ( T2DPlanarNode &node )
{
    TListIterator<TEdge>    iterEdges;
    TWeightedMarkableEdge  *pEdge;

    iterEdges.setList( node.getOutEdges() );
    iterEdges.first();
    while ((pEdge = (TWeightedMarkableEdge*) iterEdges++) != NULL)
    {
        if (pEdge->weight() != 0)
            return TRUE;
    }// while

    return FALSE;
}

//
// check if node has a weak edge
//
bool 
TPlanarBlock::hasWeakEdge ( T2DPlanarNode &node )
{
    TListIterator<TEdge>    iterEdges;
    TWeightedMarkableEdge  *pEdge;

    iterEdges.setList( node.getOutEdges() );
    iterEdges.first();
    while ((pEdge = (TWeightedMarkableEdge*) iterEdges++) != NULL)
    {
        if (pEdge->weight() == 0)
            return TRUE;
    }// while

    return FALSE;
}

//
// remove all weak edges
//
void 
TPlanarBlock::removeWeak ( T2DPlanarNode &node )
{
    TListIterator<TEdge>    iterEdges;
    TWeightedMarkableEdge  *pEdge;
    T2DPlanarNode          *pSucc;

    iterEdges.setList( node.getOutEdges() );
    iterEdges.first();
    while ((pEdge = (TWeightedMarkableEdge*) iterEdges++) != NULL)
    {
        pSucc = (T2DPlanarNode*) pEdge->getHead();

        if (pEdge->weight() == 0)
        {
            node.remPred( *pSucc );
            node.remSucc( *pSucc );
            pSucc->remPred( node );
            pSucc->remSucc( node );
        }// if
    }// while
}

////////////////////////////////////////////////
//
// methods for building minimal/maximal cycles
//
    
//
// build cycles (blocks)
//
TList<TGraph> *
TPlanarBlock::buildCycles ( TGraph * pGraph, uint type,
                            TBlockInfo * pBlockInfo )
{
    if (pGraph == NULL)
    {
        cout << "TPlanarBlock::buildCycles (...) : got NULL arg" << endl;
        return NULL;
    }// if
    
    TList<TGraph> *plistCycles;
    TPlanarCycle  *pCycle;
    uint          i;

    //
    // restrict graph to C-Nodes
    //
    localNumber( pGraph, type, pBlockInfo );

    //
    // build cycles
    // 

    plistCycles = new TList<TGraph>;

    i = 0;
    while ((pCycle = getCycle( pGraph, type, i++ )) != NULL)
    {
        // build add. info and update graph
        if (pBlockInfo != NULL)
            pBlockInfo->addCycle( *pCycle );

        // locally update nodemarks
        localNumberCycle( pCycle, type, pBlockInfo );
        
        plistCycles->add( *pCycle );
    }// while

    return plistCycles;
}

//
// create single cycle
//
TPlanarCycle* 
TPlanarBlock::getCycle ( TGraph * pGraph, uint type, long cycleName )
{
    if (pGraph == NULL)
    {
        cout << "TPlanarBlock::getCycle (...) : got NULL arg" << endl;
        return NULL;
    }// if
    
    // stack for the path we go
    TList<TNode>           stack;
    TListIterator<TNode>   iterStack( stack );

    TPlanarNode          * pNode, * pStartNode, * pStackNode;
    bool                   bFinish = FALSE;
    uint                  beginOfCycle;

    //
    // look for a suitable node
    //
    if ((pStartNode = getStartNode( pGraph, type )) == NULL)
    {
        return NULL;
    }// if

    //
    // create cycle
    //
    // Because of the one-flow-dir-cond the last (first) edge 
    // is the left- (right-) most  edge. 
    // We follow this edge and go to the next node. Then we collect 
    // all these nodes until we come back to a node we visted before.
    //

    TPlanarCycle * pCycle;

    pCycle = new TPlanarCycle;
    pNode  = pStartNode;
    
    do 
    {
        if (pNode->getOutDegree() > 0) 
        {
            // check, if we visited this node before
            if (pNode->isVisited() == TRUE) 
            {
                /////////////////////////////////
                //
                // get the cycle
                //
                beginOfCycle = pNode->getName();
            
                // set cycles to the nodes
                pNode->setCycle( pCycle, type );

                pCycle->add( *pNode );

                // mark node in orig graph as deleted
                pNode->setDeleted( TRUE, type );
                pNode->setVisited( FALSE );

                // now catch all nodes till the beginning
                // (we go backwards !!!)
                iterStack.last();
                while (((pStackNode = (TPlanarNode*) iterStack--) != NULL) &&
                       (pStackNode->getName() != beginOfCycle)) 
                {
                    // set cycles to the nodes
                    pStackNode->setCycle( pCycle, type );

                    pCycle->add( *pStackNode );
                        
                    // mark node in orig graph as deleted
                    pStackNode->setDeleted( TRUE, type );
                    pStackNode->setVisited( FALSE );
                }// while

                // unvisit all visited nodes
                while ((pStackNode =  (TPlanarNode*) iterStack--) != NULL)
                    pStackNode->setVisited( FALSE );

                pCycle->setName( cycleName );

                return pCycle;
            }// if
            else 
            {
                // mark this node as visited
                pNode->setVisited( TRUE );

                // put succ on stack
                stack.add( *pNode );
                
                //
                // go to a nondeleted succ of node
                //
                pNode = getSucc( pNode, type );
            }// else
        }// if
        else 
        {
            // dead-end: there is no min (max) cycle with the OFD cond
            cerr << "TPlanarFVS::getCycle (TNode&) : no succ to build cycle" << endl;
            bFinish = TRUE;
        }// else
    } while (bFinish != TRUE);

    return pCycle;
}

//
// look for a suiteable startnode of a cycle
//
TPlanarNode *
TPlanarBlock::getStartNode ( TGraph * pGraph, uint type )
{
    if (pGraph == NULL)
    {
        cout << "TPlanarBlock::getStartNode (...) : got NULL arg" << endl;
        return NULL;
    }// if

    TListIterator< TNode >  iterNodes;
    TPlanarNode  * pNode;

    iterNodes.setList( pGraph->getNodeList() );
    iterNodes.first();
    while ((pNode = (TPlanarNode*) iterNodes++) != NULL)
    {
        if ((pNode->isDeleted( type ) == FALSE) && 
            (pNode->mark( type ) == C) &&
            (pNode->cycle( type ) == NULL))
        {
            return pNode;
        }// if
    }// while

    return NULL;
}

//
// look for suitable succ of given node
//
TPlanarNode *
TPlanarBlock::getSucc ( TNode * pNode, uint type )
{
    if (pNode == NULL)
    {
        cout << "TPlanarBlock::getSucc (...) : got NULL arg" << endl;
        return NULL;
    }// if

    TListIterator< TEdge >   iterEdges;
    TEdge                  * pEdge;
    TPlanarNode            * pSucc;

    
    iterEdges.setList( pNode->getOutEdges() );

    //
    // map to min or to max cycle ?
    //
    if (type == MIN_CYC) 
    {
        iterEdges.last();
        while ((pEdge = iterEdges--) != NULL) 
        {
            pSucc = (TPlanarNode*) pEdge->getHead();
            
            if ((pSucc->isDeleted( type ) == FALSE) &&
                (pSucc->mark( type ) == C) &&
                (pSucc->cycle( type ) == NULL))
            {
                return pSucc;
            }// if
        }// while
    }// if
    else // type == MAX_CYC
    {
        iterEdges.first();
        while ((pEdge = iterEdges++) != NULL)
        {
            pSucc = (TPlanarNode*) pEdge->getHead();
            if ((pSucc->isDeleted( type ) == FALSE) &&
                (pSucc->mark( type ) == C) &&
                (pSucc->cycle( type ) == NULL))
            {
                return pSucc;
            }// if
        }// while
    }// else

    return NULL;
}
    
////////////////////////////////////////////////
//
// local version of numbering algorithm
//

void 
TPlanarBlock::localNumber ( TGraph * pGraph, uint type,
                            TBlockInfo * pBlockInfo )
{
    if (pGraph == NULL)
    {
        cout << "TPlanarBlock::localNumber (...) : got NULL arg" << endl;
        return;
    }// if
    
    TListIterator<TNode>    iter( pGraph->getNodeList() );
    TPlanarNode           * pNode;

    //
    // set marks of all nodes to C
    //
    iter.first();
    while ((pNode = (TPlanarNode*) iter++) != NULL)
    {
        if (pNode->isDeleted( type ) == FALSE)
            pNode->setMark( C, type );
    }// while

    //
    // now adjust marks
    //
    iter.first();
    while ((pNode = (TPlanarNode*) iter++) != NULL) 
    {
        if (pNode->isDeleted( type ) == FALSE) 
        {
            if (pNode->mark( type ) == C)
                localSetF( pNode, type, pBlockInfo, NULL );

            if (pNode->mark( type ) == C)
                localSetL( pNode, type, pBlockInfo, NULL );
        }// if
    }// while
}

void 
TPlanarBlock::localNumberCycle ( TPlanarCycle * pCycle, uint type,
                                 TBlockInfo * pBlockInfo )
{
    if (pCycle == NULL)
    {
        cout << "TPlanarBlock::localNumberCycle (...) : got NULL arg" << endl;
        return;
    }// if
    
    TListIterator<TNode>    iter( pCycle->getNodeList() );
    TPlanarNode           * pNode;

    //
    // adjust marks
    //
    iter.first();
    while ((pNode = (TPlanarNode*) iter++) != NULL)
    {
        if (pNode->mark( type ) == C)
            localSetF( pNode, type, pBlockInfo, pCycle );

        if (pNode->mark( type ) == C)
            localSetL( pNode, type, pBlockInfo, pCycle );
    }// while
}

void 
TPlanarBlock::localSetF ( TPlanarNode * pNode, uint type,
                          TBlockInfo * pBlockInfo, TPlanarCycle * pCycle )
{
    TListIterator<TEdge>     iter( pNode->getInEdges() );
    TEdge                  * pEdge;
    TPlanarNode            * pAdjNode;

    bool  bPredHaveF = TRUE;
    bool  bOnCycle = FALSE;

    // if really local check if on cycle
    if (pCycle != NULL)    
    {
        if ((pNode->cycle( type ) != NULL) &&
            (pNode->cycle( type )->name() == pCycle->name()))
            bOnCycle = TRUE;
    }// if

    if (bOnCycle == FALSE) 
    {
        iter.first();
        while ((pEdge = iter++) != NULL) 
        {
            pAdjNode = (TPlanarNode*) pEdge->getTail();
            
            if ((pAdjNode->isDeleted( type ) == FALSE) &&
                (pAdjNode->mark( type ) != F))
            {
                bPredHaveF = FALSE;
                break;
            }// if
        }// while
    }// if

    if (bPredHaveF == TRUE) 
    {
        if (bOnCycle == FALSE) 
        {
            pNode->setMark( F, type );
            pNode->setDeleted( TRUE, type );
            if (pBlockInfo != NULL)
                pBlockInfo->addFNode( * pNode );
        }// if

        // do the same with all successors
        iter.setList( pNode->getOutEdges() );
        iter.first();
        while ((pEdge = iter++) != NULL) 
        {
            pAdjNode = (TPlanarNode*) pEdge->getHead();

            if ((pAdjNode->isDeleted( type ) == FALSE) &&
                (pAdjNode->mark( type ) == C))
            {
                localSetF( pAdjNode, type, pBlockInfo, pCycle );
            }// if
        }// while
    }// if
}

void 
TPlanarBlock::localSetL ( TPlanarNode * pNode, uint type,
                          TBlockInfo * pBlockInfo, TPlanarCycle * pCycle )
{
    TListIterator<TEdge>    iter( pNode->getOutEdges() );
    TEdge                 * pEdge;
    TPlanarNode           * pAdjNode;

    bool  bSuccHaveL = TRUE;
    bool  bOnCycle = FALSE;

    // if really local check if on cycle
    if (pCycle != NULL) 
    {
        if ((pNode->cycle( type ) != NULL) &&
            (pNode->cycle( type )->name() == pCycle->name()))
            bOnCycle = TRUE;
    }// if

    if (bOnCycle == FALSE)
    {
        iter.first();
        while ((pEdge = iter++) != NULL)
        {
            pAdjNode = (TPlanarNode*) pEdge->getHead();
            
            if ((pAdjNode->isDeleted( type ) == FALSE) &&
                (pAdjNode->mark( type ) != L))
            {
                bSuccHaveL = FALSE;
                break;
            }// if
        }// while
    }// if

    if (bSuccHaveL == TRUE) 
    {
        if (bOnCycle == FALSE) 
        {
            pNode->setMark( L, type );
            pNode->setDeleted( TRUE, type );
            if (pBlockInfo != NULL)
                pBlockInfo->addLNode( * pNode );
        }// if

        // do the same with all predecessors
        iter.setList( pNode->getInEdges() );
        iter.first();
        while ((pEdge = iter++) != NULL) 
        {
            pAdjNode = (TPlanarNode*) pEdge->getTail();

            if ((pAdjNode->isDeleted( type ) == FALSE) &&
                (pAdjNode->mark( type ) == C))
            {
                localSetL( pAdjNode, type, pBlockInfo, pCycle );
            }// if
        }// while
    }// if
}
