//
//  Project     : libgraph
//  File        : TPlanarFVS.cc
//  Author      : Ronald Kriemann
//  Purpose     : class building FVS for planar graphs
//
// arch-tag: 6b2ccf43-f560-4223-a2a6-71c66a288a36
//

extern "C" {
#include <stdio.h>
}

#include "var.hh"
#include "cmdline.hh"
#include "timer.hh"

#include "drawapp.hh"

#include "nodes.hh"
#include "graphtool.hh"
#include "TPlanarNode.hh"
#include "TPlanarCycle.hh"
#include "T2DPlanarNode.hh"
#include "fvs.hh"

#include "planarFVS.hh"

#ifdef DELETED
#  undef DELETED
#endif

// marks for graph-traversal
#define UN_VISITED  1
#define VISITED     2
#define IS_PRED     3
#define DELETED     4

///////////////////////////////////////////
//
// constructor
//
TPlanarFVS::TPlanarFVS ()
{
    pGraph = NULL;

    iterStack.setList( stack );
}

///////////////////////////////////////////
//
// builds the FVS
//
TList< TNode > *
TPlanarFVS::buildFVS ( TGraph & graph )
{
    TGraphTool  graphTool;

    TList< TGraph >        * pListComp;
    TListIterator< TGraph >  iterComp;
    TGraph                 * pComp;
    TList< TNode >         * pLocalFVS, * pFVS;
    TListIterator< TNode >   iterFVS;
    TNode                  * pNode;

    //
    // decompose graph into SCC's
    //
    if (param.global)                                 // workaround for BUG
        pListComp = graphTool.scc( graph );
    else
        pListComp = graphTool.tarjan( graph );

    if (pListComp == NULL)
    {
        cout << "TPlanarFVS::buildFVS (...) : got NULL SCC-list" << endl;
        return NULL;
    }// if

    iterComp.setList( * pListComp );

    //
    // build fvs in each SCC
    //

    pFVS = new TList< TNode >;
    
    iterComp.first();
    while ((pComp = iterComp++) != NULL)
    {
        if (pComp->getNoOfNodes() > 1)
        {
            TPlanarFVS  localFVS;
        
            pLocalFVS = localFVS.buildFVS_SCC( pComp );

            if (pLocalFVS == NULL)
            {
                cout << "TPlanarFVS::buildFVS (...) : got NULL fvs-list" << endl;
                continue;
            }// if
            
            iterFVS.setList( * pLocalFVS );
            iterFVS.first();
            while ((pNode = iterFVS++) != NULL)
            {
                pFVS->add( * pNode );
            }// while

            delete pLocalFVS;
        }// if
    }// while

    //
    // clean up
    //
    iterComp.first();
    while ((pComp = iterComp++) != NULL)
    {
        delete pComp;
    }// while

    delete pListComp;

    
    return pFVS;
}

//
// build planar fvs provided graph is strong connected
//
TList< TNode > *
TPlanarFVS::buildFVS_SCC ( TGraph * pGraph )
{
    if (pGraph == NULL)
    {
        cout << "TPlanarFVS::buildFVS_SCC (...) : got NULL arg" << endl;
        return NULL;
    }// if

    TTimer    timer;
    TGraph  * pZPlus, * pZMinus;

    TPlanarFVS::pGraph = pGraph;
    
    //
    // build Z+
    //

    pZPlus  = new TGraph;

    timer.start();
    buildZNodes( pZPlus, MIN_CYC );
    buildZEdges( pZPlus, MIN_CYC );
    timer.stop();
    
    if (verbosity >= VERBOSE_TIME)
        cout << "TPlanarFVS::buildFVS_SCC (...) : time to build Z+ = " << timer << endl;
    if (verbosity >= VERBOSE_RESULT)
        cout << "TPlanarFVS::buildFVS_SCC (...) : Z+ = " << endl << * pZPlus << endl;

    //
    // build Z-
    //

    pZMinus = new TGraph;
    
    timer.start();
    buildZNodes( pZMinus, MAX_CYC );
    buildZEdges( pZMinus, MAX_CYC );
    timer.stop();

    if (verbosity >= VERBOSE_TIME)
        cout << "TPlanarFVS::buildFVS_SCC (...) : time to build Z- = " << timer << endl;
    if (verbosity >= VERBOSE_RESULT)
        cout << "TPlanarFVS::buildFVS_SCC (...) : Z- = " << endl << * pZMinus << endl;

    //
    // build fvs
    //

    TList< TNode >  * pListFVS;
    TPlanarNode     * pFV;

    pListFVS = new TList< TNode >;
    
    while ( cyclic( pZPlus, pZMinus ) )
    {
        //
        // search for a node in T+ or T-
        //
        pFV = getTPlusNode( pZPlus );
        
        if (pFV == NULL)
        {
            cout << "TPlanarFVS_SCC (...) : got NULL feedback vertex" << endl;
            break;
        }// if
        
        //
        // add node to FVS and add nodes in touching succ-cycles
        // update Z+ and Z-
        //

        addFV( pFV, pListFVS, pZPlus, pZMinus );

        //
        // rebuild Z+ and Z-
        //

        rebuildZ( pZPlus, pZMinus, pListFVS );
    }// while

    //
    // clean up
    //

    if (pZPlus)
        delete pZPlus;

    if (pZMinus)
        delete pZMinus;
    
    return pListFVS;
}

///////////////////////////////////////////
//
// misc. help methods 
//

//
// build graph consisting of cycles
//
void
TPlanarFVS::buildZNodes ( TGraph * pZ, uint type )
{
    if (pZ == NULL)
    {
        cout << "TPlanarFVS::buildZNodes (...) : got NULL arg" << endl;
        return;
    }// if
    
    //
    // restrict G to cyclic nodes
    //

    localNumber( type, TRUE );

    //
    // build minimal/maximal cycles and the graph Z+/Z-
    //

    uint             i;
    TCycleNode     * pZNode;
    TPlanarCycle   * pCycle;
    TPlanarNode    * pStartNode;

    pZNode = NULL;
    
    i = 0;
    while ((pStartNode = findStartNode( pZNode, pZ, type )) != NULL)
    {
        //
        // detect cycle in graph
        //
        pCycle = buildCycle( pStartNode, type );

        if (pCycle == NULL)
        {
            cout << "TPlanarFVS::buildZNodes (...) : got NULL cycle" << endl;
            break;
        }// if

        deleteCycle( pCycle, type );
        
        pCycle->setName( i++ );

        //
        // restrict graph to cyclic nodes
        //
        localNumberCycle( pCycle, type, FALSE );

        //
        // create and insert node in Z+/Z-
        // (build edges before insert into nodelist !!!)
        //
        
        pZNode = new TCycleNode( pCycle->name(), pCycle );
        pZ->add( * pZNode );
        pCycle->setZNode( pZNode );
        
        if (verbosity >= VERBOSE_RESULT)
        {
            char  filename[64];

            if (type == MIN_CYC)
                sprintf( filename, "mincyc%d", pCycle->name() );
            else
                sprintf( filename, "maxcyc%d", pCycle->name() );
            
            TDrawApp::draw( * pCycle, filename, TRUE );
//             TDrawApp::draw( * pCycle, "cycle", TRUE );
//             TDrawApp::draw( * pGraph, "graph" );
        }// if
    }// while
}

void
TPlanarFVS::buildZEdges ( TGraph * pZ, uint type )
{
    if (pZ == NULL)
    {
        cout << "TPlanarFVS::buildZEdges (...) : got NULL arg" << endl;
        return;
    }// if

    TListIterator< TNode >  iterNodes;
    TCycleNode            * pNode;

    iterNodes.setList( pZ->getNodeList() );

    //
    // test for each node in Z reachability of other nodes
    // and add edges accord. to reached side
    //

    iterNodes.first();
    while ((pNode = (TCycleNode*) iterNodes++) != NULL)
    {
        insertCycle( pNode, pZ, type );

//         if (verbosity >= VERBOSE_RESULT)
//             cout << "TPlanarFVS::buildZEdges (...) : Z = " << endl << * pZ << endl;
    }// while
}

//
// insert given cycle into Z+/Z-
//
void
TPlanarFVS::insertCycle ( TCycleNode * pZNode, TGraph * pZ, uint type )
{
    if ((pZNode == NULL) || (pZ == NULL))
    {
        cout << "TPlanarFVS::insertCycle (...) : got NULL arg" << endl;
        return;
    }// if

//     if (verbosity >= VERBOSE_RESULT)
//         cout << "TPlanarFVS::insertCycle (...) : inserting cycle " << pZNode->getName() << endl;

    TListIterator< TNode >  iterNodes;
    TPlanarNode           * pNode, * pAdjNode;
    TPlanarCycle          * pCycle;
    TListIterator< TEdge >  iterEdges;
    TEdge                 * pEdge;
    TCycleNode            * pCycleNode;

    stack.clear();

    //
    // reset marks
    //

    iterNodes.setList( pZ->getNodeList() );
    iterNodes.first();
    while ((pCycleNode = (TCycleNode*) iterNodes++) != NULL)
    {
        pCycleNode->setMark( UN_VISITED );
    }// while

    //
    // - we try to reach other cycles in Z by following all
    //   path beginning in right-succ of nodes in cycle
    // - if we reach another cycle from the left, we are in his interior
    //   and so we have to build an edge to this cycle
    //

    pCycle = pZNode->cycle();
    iterNodes.setList( pCycle->getNodeList() );

    //
    // try all succ on right side
    //
    
    iterNodes.first();
    while ((pNode = (TPlanarNode*) iterNodes++) != NULL)
    {
        iterEdges.setList( pNode->getOutEdges() );
        iterEdges.first();
        while ((pEdge = iterEdges++) != NULL)
        {
            pAdjNode = (TPlanarNode*) pEdge->getHead();

            if ( pAdjNode->isInStack() )
                continue;

            if ( pAdjNode->cycle( type ) == pCycle )
                break;
            
            if ( checkForCycle( pAdjNode, pEdge, pCycle, type ) )
                continue;
            
            reachForward( pAdjNode, pCycle, pZ, type );
        }// while
    }// while

    //
    // remove marks from all visited nodes
    //
    iterStack.first();
    while ((pNode = (TPlanarNode*) iterStack++) != NULL)
    {
        pNode->setInStack( FALSE );
    }// while

    stack.clear();

    //
    // try all pred on right side
    //

    iterNodes.first();
    while ((pNode = (TPlanarNode*) iterNodes++) != NULL)
    {
        //
        // try all succ on right side
        //

        iterEdges.setList( pNode->getInEdges() );
        iterEdges.last();
        while ((pEdge = iterEdges--) != NULL)
        {
            pAdjNode = (TPlanarNode*) pEdge->getTail();

            if ( pAdjNode->isInStack() )
                continue;

            if ( pAdjNode->cycle( type ) == pCycle )
                break;
            
            if ( checkForCycle( pAdjNode, pEdge, pCycle, type ) )
                continue;
            
            reachBackward( pAdjNode, pCycle, pZ, type );
        }// while
    }// while

    //
    // remove marks from all visited nodes
    //
    iterStack.first();
    while ((pNode = (TPlanarNode*) iterStack++) != NULL)
    {
        pNode->setInStack( FALSE );
    }// while

    stack.clear();
}

//
// try to reach other cycles via DFS
//
void
TPlanarFVS::reachForward ( TPlanarNode * pNode, TPlanarCycle * pCycle,
                           TGraph * pZ, uint type )
{
    if ((pNode == NULL) || (pCycle == NULL) || (pZ == NULL))
    {
        cout << "TPlanarFVS::reachForward (...) : got NULL arg" << endl;
        return;
    }// if

//     if (verbosity >= VERBOSE_RESULT)
//         cout << "TPlanarFVS::reachForward (...) : visiting node " << pNode->getName() << endl;
    
    //
    // initialisation of DFS-visit
    //
    
    if (pNode->isInStack())
        return;

    pNode->setInStack( TRUE );
    stack.add( * pNode );

    //
    // go over all succ and check if different cycle was reached
    //

    TListIterator< TEdge >  iterEdges;
    TEdge                 * pEdge;
    TPlanarNode           * pAdjNode;

    iterEdges.setList( pNode->getOutEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pAdjNode = (TPlanarNode*) pEdge->getHead();

        if (pAdjNode->isInStack())
            continue;

        if ( checkForCycle( pAdjNode, pEdge, pCycle, type ) )
            continue;
        
        reachForward( pAdjNode, pCycle, pZ, type );
    }// while
}
    
void
TPlanarFVS::reachBackward ( TPlanarNode * pNode, TPlanarCycle * pCycle,
                            TGraph * pZ, uint type )
{
    if ((pNode == NULL) || (pCycle == NULL) || (pZ == NULL))
    {
        cout << "TPlanarFVS::reachBackward (...) : got NULL arg" << endl;
        return;
    }// if

//     if (verbosity >= VERBOSE_RESULT)
//         cout << "TPlanarFVS::reachBackward (...) : visiting node " << pNode->getName() << endl;
    
    //
    // initialisation of DFS-visit
    //
    
    if (pNode->isInStack())
        return;

    pNode->setInStack( TRUE );
    stack.add( * pNode );

    //
    // go over all pred and check if different cycle was reached
    //

    TListIterator< TEdge >  iterEdges;
    TEdge                 * pEdge;
    TPlanarNode           * pAdjNode;

    iterEdges.setList( pNode->getInEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pAdjNode = (TPlanarNode*) pEdge->getTail();

        if (pAdjNode->isInStack())
            continue;

        if ( checkForCycle( pAdjNode, pEdge, pCycle, type ) )
            continue;
        
        reachBackward( pAdjNode, pCycle, pZ, type );
    }// while
}

//
// find suitable startnode for buildCycle
//
TPlanarNode *
TPlanarFVS::findStartNode ( TCycleNode * pLastCycle,
                            TGraph * pZ,
                            uint type )
{
    TListIterator< TNode >   iterNodes;
    TPlanarNode            * pNode, * pAdjNode;
    TListIterator< TEdge >   iterEdges;
    TEdge                  * pEdge;
    TPlanarCycle           * pCycle;

    if (pLastCycle != NULL)
    {
        //
        // try nodes in Ext/Int(cycle) depending on min or max cycles
        //

        pCycle = pLastCycle->cycle();
    
        iterNodes.setList( pCycle->getNodeList() );

        iterNodes.first();
        while ((pNode = (TPlanarNode*) iterNodes++) != NULL)
        {
            iterEdges.setList( pNode->getOutEdges() );
            iterEdges.first();
            while ((pEdge = iterEdges++) != NULL)
            {
                pAdjNode = (TPlanarNode*) pEdge->getHead();
                
                if (pAdjNode->cycle( type ) == pCycle)
                    break;
                
                if ((pAdjNode->isDeleted( type ) == FALSE) &&
                    (pAdjNode->mark( type ) == C))
                {
                    return pAdjNode;
                }// if
            }// while
        }// while

        //
        // try nodes on other side
        //

        iterNodes.first();
        while ((pNode = (TPlanarNode*) iterNodes++) != NULL)
        {
            iterEdges.setList( pNode->getOutEdges() );
            iterEdges.last();
            while ((pEdge = iterEdges--) != NULL)
            {
                pAdjNode = (TPlanarNode*) pEdge->getHead();
                
                if (pAdjNode->cycle( type ) == pCycle)
                    break;
                
                if ((pAdjNode->isDeleted( type ) == FALSE) &&
                    (pAdjNode->mark( type ) == C))
                {
                    return pAdjNode;
                }// if
            }// while
        }// while
    }// if
    
    //
    // look for startnode in whole graph
    //

    iterNodes.setList( pGraph->getNodeList() );
    iterNodes.first();
    while ((pNode = (T2DPlanarNode*) iterNodes++) != NULL)
    {
        if ((pNode->isDeleted( type ) == FALSE) && 
            (pNode->mark( type ) == C))
        {
            return pNode;
        }// if
    }// while

    return NULL;
}

//
// check if different cycle is reached and build edges if needed
//
bool
TPlanarFVS::checkForCycle ( TPlanarNode * pAdjNode, TEdge * pEdge,
                            TPlanarCycle * pCycle, uint type )
{
    if ((pAdjNode == NULL) || (pEdge == NULL) || (pCycle == NULL))
    {
        cout << "TPlanarFVS::checkForCycle (...) : got NULL arg" << endl;
        return FALSE;
    }// if

    TPlanarCycle  * pAdjCycle;
    TCycleNode    * pZNode, * pAdjZNode;

    pAdjCycle = pAdjNode->cycle( type );
    
    if ((pAdjCycle != NULL) && (pAdjCycle != pCycle))
    {
        pAdjZNode = pAdjCycle->getZNode();
        
        if (pAdjZNode->mark() != VISITED)
        {
            pAdjZNode->setMark( VISITED );
            
            if ( isLeftEdge( pEdge, pAdjCycle, type ) )
            {
                pZNode = pCycle->getZNode();
                    
                if (type == MIN_CYC)
                {
//                     if (verbosity >= VERBOSE_RESULT)
//                         cout << "TPlanarFVS::checkForCycle (...) : edge "
//                              << pCycle->name() << " -> " << pAdjCycle->name() << endl;
                    
                    pEdge = new TEdge( * pZNode, * pAdjZNode );
                    pZNode->addOutEdge( * pEdge );
                    pAdjZNode->addInEdge( * pEdge );
                }// if
                else
                {
//                     if (verbosity >= VERBOSE_RESULT)
//                         cout << "TPlanarFVS::checkForCycle (...) : edge "
//                              << pAdjCycle->name() << " -> " << pCycle->name() << endl;
                    
                    pEdge = new TEdge( * pAdjZNode, * pZNode );
                    pAdjZNode->addOutEdge( * pEdge );
                    pZNode->addInEdge( * pEdge );
                }// else
            }// if
        }// if

        //
        // we stop here further graph-traversal
        //
        pAdjNode->setInStack( TRUE );
        stack.add( * pAdjNode );
        
        return TRUE;
    }// if

    if (pAdjCycle == pCycle)
        return TRUE;
    else
        return FALSE;
}

//
// look if given edge is on left side of given cycle
//
bool
TPlanarFVS::isLeftEdge ( TEdge * pTestEdge, TPlanarCycle * pCycle, uint type )
{
    if ((pTestEdge == NULL) || (pCycle == NULL))
    {
        cout << "TPlanarFvs::isLeftEdge (...) : got NULL arg" << endl;
        return FALSE;
    }// if

    TListIterator< TEdge >   iterEdges;
    TEdge                  * pEdge;
    TPlanarNode            * pNode, * pAdjNode;

    //
    // look if we have to check in- or out-edges
    //
    
    pNode = (TPlanarNode*) pTestEdge->getHead();
    if (pNode->cycle( type ) == pCycle)
    {
        //
        // check in-edges
        //

        iterEdges.setList( pNode->getInEdges() );
        iterEdges.first();
        while ((pEdge = iterEdges++) != NULL)
        {
            if (pEdge == pTestEdge)
                return TRUE;
            
            pAdjNode = (TPlanarNode*) pEdge->getTail();

            if (pAdjNode->cycle( type ) == pCycle)
                return FALSE;
        }// while
    }// if
    else
    {
        //
        // check out-edges
        //
        
        pNode = (TPlanarNode*) pTestEdge->getTail();

        iterEdges.setList( pNode->getOutEdges() );
        iterEdges.last();
        while ((pEdge = iterEdges--) != NULL)
        {
            if (pEdge == pTestEdge)
                return TRUE;
            
            pAdjNode = (TPlanarNode*) pEdge->getHead();

            if (pAdjNode->cycle( type ) == pCycle)
                return FALSE;
        }// while
    }// else

    return FALSE;
}

//
// maps a given node to a minimal/maximal cycle
// provided the graph fulfills OFD condition
//
TPlanarCycle * 
TPlanarFVS::buildCycle ( TPlanarNode * pStartNode, uint type )
{
    TPlanarNode                  * pNode, * pStackNode;
    bool                           bFinish;
    uint                          beginOfCycle;
    TPlanarCycle                 * pCycle;

    //
    // Because of the one-flow-dir-cond the last (first) edge 
    // is the left- (right-) most  edge. 
    // We follow this edge and go to the next node. Then we collect 
    // all these nodes until we come back to a node we visted before.
    //

    stack.clear();
    pNode = pStartNode;
    bFinish = FALSE;
    
    do 
    {
        if ((pNode != NULL) && (pNode->getOutDegree() > 0))
        {
            // check, if we visited this node before
            if (pNode->isInStack()) 
            {
                pCycle = new TPlanarCycle;
                
                /////////////////////////////////
                //
                // get the cycle
                //
                beginOfCycle = pNode->getName();
            
                pCycle->add( *pNode );

                pNode->setInStack( FALSE );

                // now catch all nodes till the beginning (backwards !!!)
                iterStack.last();
                while ((pStackNode = iterStack--) != NULL)
                {
                    if (pStackNode->getName() == beginOfCycle)
                        break;
                    
                    pCycle->add( *pStackNode );

                    pStackNode->setInStack( FALSE );
                }// while

                // unvisit all visited nodes
                while ((pStackNode = iterStack--) != NULL)
                {
                    pStackNode->setInStack( FALSE );
                }// while

                return pCycle;
            }// if

            //
            // put node on stack and choose next node accord. to OFD
            //
            stack.add( *pNode );
            pNode->setInStack( TRUE );
            
            if (type == MIN_CYC)
                pNode = pNode->left();
            else
                pNode = pNode->right();
        }// if
        else 
        {
            cout << "TPlanarFVS::buildCycle (...) : no succ to build cycle" << endl;
            bFinish = TRUE;
        }// else
    } while (bFinish != TRUE);

    // unvisit all visited nodes
    iterStack.first();
    while ((pStackNode = iterStack++) != NULL)
    {
        pStackNode->setInStack( FALSE );
    }// while
    
    return NULL;
}

TPlanarCycle *
TPlanarFVS::buildUpdatedCycle ( TPlanarNode * pStartNode,
                                TPlanarCycle * pOldCycle,
                                uint type )
{
    TPlanarNode                  * pNode, * pStackNode;
    bool                           bFinish;
    uint                          beginOfCycle;
    TPlanarCycle                 * pCycle;

    //
    // Because of the one-flow-dir-cond the last (first) edge 
    // is the left- (right-) most  edge. 
    // We follow this edge and go to the next node. Then we collect 
    // all these nodes until we come back to a node we visted before.
    //

    stack.clear();
    pNode = pStartNode;
    bFinish = FALSE;
    
    do 
    {
        if ((pNode != NULL) && (pNode->getOutDegree() > 0))
        {
            // check, if we visited this node before
            if (pNode->isInStack()) 
            {
                pCycle = new TPlanarCycle;
                
                /////////////////////////////////
                //
                // get the cycle
                //
                beginOfCycle = pNode->getName();
            
                pCycle->add( *pNode );

                pNode->setInStack( FALSE );

                // now catch all nodes till the beginning (backwards !!!)
                iterStack.last();
                while ((pStackNode = iterStack--) != NULL)
                {
                    if (pStackNode->getName() == beginOfCycle)
                        break;
                    
                    pCycle->add( *pStackNode );

                    pStackNode->setInStack( FALSE );
                }// while

                // unvisit all visited nodes
                while ((pStackNode = iterStack--) != NULL)
                {
                    pStackNode->setInStack( FALSE );

                    if (pStackNode->cycle( type ) == pOldCycle)
                        pStackNode->setDeleted( TRUE, type );
                }// while

                return pCycle;
            }// if

            //
            // put node on stack and choose next node accord. to OFD
            //
            stack.add( *pNode );
            pNode->setInStack( TRUE );
            
            if (type == MIN_CYC)
                pNode = pNode->left( TRUE );
            else
                pNode = pNode->right( TRUE );
        }// if
        else 
        {
            cout << "TPlanarFVS::buildCycle (...) : no succ to build cycle" << endl;
            bFinish = TRUE;
        }// else
    } while (bFinish != TRUE);

    // unvisit all visited nodes
    iterStack.first();
    while ((pStackNode = iterStack++) != NULL)
    {
        pStackNode->setInStack( FALSE );
        
        if (pStackNode->cycle( type ) == pOldCycle)
            pStackNode->setDeleted( TRUE, type );
    }// while
    
    return NULL;
}

//
// delete nodes on given cycle and apply add. info for nodes
//
void
TPlanarFVS::deleteCycle ( TPlanarCycle * pCycle, uint type )
{
    if (pCycle == NULL)
    {
        cout << "TPlanarFVS::deleteCycle (...) : got NULL arg" << endl;
        return;
    }// if

    //
    // delete nodes in cycle and set info
    //

    TListIterator< TNode >  iterNodes;
    TPlanarNode           * pNode;

    iterNodes.setList( pCycle->getNodeList() );
    iterNodes.first();
    while ((pNode = (TPlanarNode*) iterNodes++) != NULL)
    {
        pNode->setCycle( pCycle, type );
        pNode->setDeleted( TRUE, type );
    }// while
}
    
//
// look if graph is cyclic or not (graph is represented by Z+/Z-)
//
bool
TPlanarFVS::cyclic ( TGraph * pZPlus, TGraph * pZMinus )
{
    if ((pZPlus == NULL) || (pZMinus == NULL))
    {
        cout << "TPlanarFVS::cyclic (...) : got NULL arg" << endl;
        return FALSE;
    }// if

    //
    // if both Z+ and Z- do not contain any cycle the
    // graph is acyclic
    //
    
    if ((pZPlus->getNoOfNodes() == 0) && (pZPlus->getNoOfNodes() == 0))
        return FALSE;

    // some error checks
    if (((pZPlus->getNoOfNodes() == 0) && (pZPlus->getNoOfNodes() != 0)) ||
        ((pZPlus->getNoOfNodes() != 0) && (pZPlus->getNoOfNodes() == 0)))
    {
        cout << "TPlanarFVS::cyclic (...) : inconsistent graphs Z+/Z-" << endl;
        return TRUE;
    }// if

    return TRUE;
}

//
// return node in T+/T- ( a feedback vertex )
//
TPlanarNode *
TPlanarFVS::getTPlusNode ( TGraph * pZPlus )
{
    if (pZPlus == NULL)
    {
        cout << "TPlanarFVS::getTPlusNode (...) : got NULL arg" << endl;
        return NULL;
    }// if

    TListIterator< TNode >  iterZNodes;
    TCycleNode            * pZNode;
    TListIterator< TNode >  iterNodes;
    TPlanarNode           * pNode;

    //
    // look for a cycle in Bottom(Z+)
    //

    iterZNodes.setList( pZPlus->getNodeList() );
    iterZNodes.first();
    while ((pZNode = (TCycleNode*) iterZNodes++) != NULL)
    {
        if (pZNode->getInDegree() > 0)
            continue;

        //
        // look for a node on cycle wich lies also on a
        // maximal cycle
        //
        
        iterNodes.setList( pZNode->cycle()->getNodeList() );
        iterNodes.first();
        while ((pNode = (TPlanarNode*) iterNodes++) != NULL)
        {
            if (pNode->cycle( MAX_CYC ) != NULL)
                return pNode;
        }// while
    }// while

    return NULL;
}

TPlanarNode *
TPlanarFVS::getTMinusNode ( TGraph * pZMinus )
{
    if (pZMinus == NULL)
    {
        cout << "TPlanarFVS::getTMinusNode (...) : got NULL arg" << endl;
        return NULL;
    }// if

    TListIterator< TNode >  iterZNodes;
    TCycleNode            * pZNode;
    TListIterator< TNode >  iterNodes;
    TPlanarNode           * pNode;

    //
    // look for a cycle in Bottom(Z-)
    //

    iterZNodes.setList( pZMinus->getNodeList() );
    iterZNodes.first();
    while ((pZNode = (TCycleNode*) iterZNodes++) != NULL)
    {
        if (pZNode->getInDegree() > 0)
            continue;

        //
        // look for a node on cycle wich lies also on a
        // minimal cycle
        //
        
        iterNodes.setList( pZNode->cycle()->getNodeList() );
        iterNodes.first();
        while ((pNode = (TPlanarNode*) iterNodes++) != NULL)
        {
            if (pNode->cycle( MIN_CYC ) != NULL)
                return pNode;
        }// while
    }// while

    return NULL;
}

//
// add node to fvs
//
void
TPlanarFVS::addFV ( TPlanarNode * pFVertex, TList< TNode > * pListFVS,
                    TGraph * pZPlus, TGraph * pZMinus )
{
    if ((pFVertex == NULL) || (pListFVS == NULL) ||
        (pZPlus == NULL) || (pZMinus == NULL))
    {
        cout << "TPlanarFVS::addFV (...) : got NULL arg" << endl;
        return;
    }// if

    //
    // remove node from graph und update mark in local environment
    // of feedback vertex
    //

    pFVertex->addRef();
    pListFVS->add( * pFVertex );
    
    pFVertex->setMark( FVS_NODE, MIN_CYC );
    pFVertex->setMark( FVS_NODE, MAX_CYC );

    pFVertex->setDeleted( TRUE, MIN_CYC );
    pFVertex->setDeleted( TRUE, MAX_CYC );

    //
    // undelete min and max cycle, feedback node lies on
    //

    TPlanarCycle  * pMinCycle, * pMaxCycle;

    pMinCycle = pFVertex->cycle( MIN_CYC );
    pMaxCycle = pFVertex->cycle( MAX_CYC );

    detectSuccFV( pMinCycle->getZNode(), pMaxCycle->getZNode(),
                  pListFVS );

    undelete( pMinCycle, FALSE, MIN_CYC );
    undelete( pMaxCycle, FALSE, MAX_CYC );

    return;
    
    //
    // update marks of nodes in local environment
    //

    localNumberFV( pFVertex, MIN_CYC );
    localNumberFV( pFVertex, MAX_CYC );

    if ((verbosity >= VERBOSE_RESULT) && param.draw)
        TDrawApp::draw( * pGraph, "graph", TRUE );

    //
    // mark all succ of min cycle in Z
    // and remove oldcycle from Z
    //

    TListIterator< TEdge >  iterEdges;
    TEdge                 * pEdge;
    TPlanarCycle          * pCycle;

    iterEdges.setList( pMinCycle->getZNode()->getOutEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pCycle = ((TCycleNode*) pEdge->getHead())->cycle();

        pCycle->setSucc( TRUE );
    }// while

    pZPlus->remove( * pMinCycle->getZNode() );
    
    //
    // detect all new cycles between old cycle and
    // succ/pred of old cycle
    //

    uint                     i;
    TPlanarCycle           * pNewCycle;
    TListIterator< TNode >   iterNodes;
    TPlanarNode            * pNode;
    TList< TPlanarCycle >    listNewCycles, listTouchedCycles;

    i = pZPlus->getNoOfNodes();

    iterNodes.setList( pMinCycle->getNodeList() );
    iterNodes.first();
    while ((pNode = (TPlanarNode*) iterNodes++) != NULL)
    {
        while ((pNode->mark( MIN_CYC ) == C) && ( ! pNode->isDeleted( MIN_CYC )))
        {                                                  
            pNewCycle = buildUpdatedCycle( pNode, pMinCycle, MIN_CYC );

            if (pNewCycle == NULL)
            {
                cout << "TPlanarFVS::addFV (...) : got NULL cycle" << endl;
                continue;
            }// if
            
            pNewCycle->setName( i++ );
            pNewCycle->setDeleted( TRUE );
            
            if ((verbosity >= VERBOSE_RESULT) && param.draw)
                TDrawApp::draw( * pNewCycle, "cycle", TRUE );

            //
            // find all cycles touched by new cycle
            // (except cycle with feedback-vertex on it)
            //

            findTouchedCycles( pNewCycle, pMinCycle,
                               & listTouchedCycles, MIN_CYC );

            if ( ! pNewCycle->isNew() )
            {
                //
                // cycle which was found is already in Z
                //

                listTouchedCycles.getFirst()->setDeleted( FALSE );
                delete pNewCycle;
                continue;
            }// if
            
            //
            // remove touched cycles from graph and from Z
            //

            removeCycles( & listTouchedCycles, pZPlus, MIN_CYC );
            
            //
            // add cycle to list of new cycles and
            // delete cycle in graph
            //

            listNewCycles.add( * pNewCycle );
            deleteCycle( pNewCycle, MIN_CYC );
            localNumberCycle( pNewCycle, MIN_CYC, FALSE );

            if ((verbosity >= VERBOSE_RESULT) && param.draw)
                TDrawApp::draw( * pGraph, "graph", TRUE );
        }// while
    }// while

    //
    // remove all found cycles, which are already in Z
    //

    TListIterator< TPlanarCycle >  iterCycles;

    iterCycles.setList( listNewCycles );
    iterCycles.first();
    while ((pNewCycle = iterCycles++) != NULL)
    {
        if ( ! pNewCycle->isNew() )
        {
        }// if
    }// while
}

//
// find old cycles touched by newcycle (except oldcycle)
//
void
TPlanarFVS::findTouchedCycles ( TPlanarCycle          * pNewCycle,
                                TPlanarCycle          * pOldCycle,
                                TList< TPlanarCycle > * pListTouchedCycles,
                                uint                    type )
{
    if ((pNewCycle == NULL) || (pOldCycle == NULL) || (pListTouchedCycles == NULL))
    {
        cout << "TPlanarFVS::findTouchedCycles (...) : got NULL arg" << endl;
        return;
    }// if

    TListIterator< TNode >  iterNodes;
    TPlanarNode           * pNode;
    TPlanarCycle          * pCycle;

    iterNodes.setList( pNewCycle->getNodeList() );
    iterNodes.first();
    while ((pNode = (TPlanarNode*) iterNodes++) != NULL)
    {
        if ((pCycle = pNode->cycle( type )) == NULL)
            continue;

        // if cycle is oldcycle or already detected new cycle
        // we do not care
        if (pCycle == pOldCycle)
        {
            pNewCycle->setNew( TRUE );
            continue;
        }// if

        if (pCycle->isDeleted())
            continue;

        pListTouchedCycles->add( * pCycle );
        pCycle->setDeleted( TRUE );
    }// while
}

//
// remove given list of cycles from graph and from Z
//
void
TPlanarFVS::removeCycles ( TList< TPlanarCycle > * pListCycles,
                           TGraph                * pZ,
                           uint                    type )
{
    if ((pListCycles == NULL) || (pZ == NULL))
    {
        cout << "TPlanarFVS::removeCycles (...) : got NULL arg" << endl;
        return;
    }// if

    TListIterator< TPlanarCycle >  iterCycles;
    TPlanarCycle                 * pCycle;

    iterCycles.setList( * pListCycles );
    iterCycles.first();
    while ((pCycle = iterCycles++) != NULL)
    {
        undelete( pCycle, TRUE, type );
        pZ->remove( * pCycle->getZNode() );

        delete pCycle;
    }// while
}
    
//
// update Z graph after feedback vertex was deleted
//
void
TPlanarFVS::updateZ ( TPlanarNode * pFVertex, TGraph * pZ, uint type )
{
}
    
//
// undelete all nodes in given cycle
//
void
TPlanarFVS::undelete ( TPlanarCycle * pCycle, bool bDelCycle, uint type )
{
    if (pCycle == NULL)
    {
        cout << "TPlanarFVS::undelete (...) : got NULL arg" << endl;
        return;
    }// if

    TListIterator< TNode >  iterNodes;
    TPlanarNode           * pNode;

    iterNodes.setList( pCycle->getNodeList() );
    iterNodes.first();
    while ((pNode = (TPlanarNode*) iterNodes++) != NULL)
    {
        if (pNode->mark() != FVS_NODE)
            pNode->setDeleted( FALSE, type );

        if (bDelCycle)
            pNode->setCycle( NULL, type );
    }// while
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

//
// a local version of the numbering algorithm
//
void
TPlanarFVS::localNumber ( uint type, bool bDelNonCyclic )
{
    TListIterator< TNode >  iterNodes;
    TPlanarNode           * pNode;

    iterNodes.setList( pGraph->getNodeList() );

    //
    // reset attributes
    //

    iterNodes.first();
    while ((pNode = (TPlanarNode*) iterNodes++) != NULL)
    {
        if (( ! pNode->isDeleted( type ) ) && (pNode->mark( type ) != FVS_NODE))
            pNode->setMark( C, type );
    }// while

    //
    // number all nodes
    //
    
    iterNodes.first();
    while ((pNode = (TPlanarNode*) iterNodes++) != NULL)
    {
        if ( pNode->isDeleted( type ) )
            continue;
        
        if (pNode->mark( type ) == C)
            localSetF( pNode, type, bDelNonCyclic, FALSE );

        if (pNode->mark( type ) == C)
            localSetL( pNode, type, bDelNonCyclic, FALSE );
    }// while
}

void
TPlanarFVS::localNumberCycle ( TPlanarCycle * pCycle, uint type, bool bDelNonCyclic )
{
    if (pCycle == NULL)
    {
        cout << "TPlanarFVS::localNumberCycle (...) : got NULL arg" << endl;
        return;
    }// if

    TListIterator< TNode >  iterNodes;
    TPlanarNode           * pNode, * pAdjNode;
    TListIterator< TEdge >  iterEdges;
    TEdge                 * pEdge;

    iterNodes.setList( pCycle->getNodeList() );

    //
    // number all nodes adjacent to nodes in cycle
    //
    
    iterNodes.first();
    while ((pNode = (TPlanarNode*) iterNodes++) != NULL)
    {
        iterEdges.setList( pNode->getOutEdges() );
        iterEdges.first();
        while ((pEdge = iterEdges++) != NULL)
        {
            pAdjNode = (TPlanarNode*) pEdge->getHead();
            
            if ((pAdjNode->mark( type ) == C) &&
                ( ! pAdjNode->isDeleted( type )))
                localSetF( pAdjNode, type, bDelNonCyclic, FALSE );
        }// if

        iterEdges.setList( pNode->getInEdges() );
        iterEdges.first();
        while ((pEdge = iterEdges++) != NULL)
        {
            pAdjNode = (TPlanarNode*) pEdge->getTail();
            
            if ((pAdjNode->mark( type ) == C) &&
                ( ! pAdjNode->isDeleted( type )))
                localSetL( pAdjNode, type, bDelNonCyclic, FALSE );
        }// if
    }// while
}

void
TPlanarFVS::localNumberFV ( TPlanarNode * pFVertex, uint type )
{
    if (pFVertex == NULL)
    {
        cout << "TPlanarFVS::localNumberFV (...) : got NULL arg" << endl;
        return;
    }// if

    TListIterator< TEdge >        iterEdges;
    TEdge                       * pEdge;
    TPlanarNode                 * pAdjNode;

    //
    // set marks to C
    //
    
    iterEdges.setList( pFVertex->getOutEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pAdjNode = (TPlanarNode*) pEdge->getHead();
        
        if (pAdjNode->isDeleted( type ) == FALSE)
            pAdjNode->setMark( C, type );
    }// while

    iterEdges.setList( pFVertex->getInEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pAdjNode = (TPlanarNode*) pEdge->getTail();

        if (pAdjNode->isDeleted( type ) == FALSE)
            pAdjNode->setMark( C, type );
    }// while  

    //
    // update marks to F and L
    //
    
    iterEdges.setList( pFVertex->getOutEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pAdjNode = (TPlanarNode*) pEdge->getHead();
        
        if ((pAdjNode->isDeleted( type ) == FALSE) &&
            (pAdjNode->mark( type ) == C))
            localSetF( pAdjNode, type, TRUE, TRUE );
    }// while

    iterEdges.setList( pFVertex->getInEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pAdjNode = (TPlanarNode*) pEdge->getTail();

        if ((pAdjNode->isDeleted( type ) == FALSE) &&
            (pAdjNode->mark( type ) == C))
            localSetL( pAdjNode, type, TRUE, TRUE );
    }// while  
}

void
TPlanarFVS::localSetF ( TPlanarNode * pNode, uint type,
                        bool bDelNonCyclic, bool bCheckCycle )
{
    if (pNode == NULL)
    {
        cout << "TPlanarFVS::localSetF (...) : got NULL arg" << endl;
        return;
    }// if

    TListIterator< TEdge >  iterEdges;
    TEdge                 * pEdge;
    TPlanarNode           * pAdjNode;

    //
    // look if all pred have attr F
    //

    iterEdges.setList( pNode->getInEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pAdjNode = (TPlanarNode*) pEdge->getTail();

        // check if "really" predecessor
        if (pAdjNode->mark( type ) == FVS_NODE)
            continue;
        
        if (( ! pAdjNode->isDeleted( type )) ||
            ( pAdjNode->isDeleted( type ) && bCheckCycle && (pAdjNode->cycle( type ) != NULL)))
        {  
            if (pAdjNode->mark( type ) != F)
                return;
        }// if
    }// while

    //
    // if all pred have attr F, this node gets attr F and check all succ
    //

    pNode->setMark( F, type );

    if (bDelNonCyclic)
        pNode->setDeleted( TRUE, type );
        
    iterEdges.setList( pNode->getOutEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pAdjNode = (TPlanarNode*) pEdge->getHead();
        
        if ((pAdjNode->mark( type ) == C) && ( ! pAdjNode->isDeleted( type )))
            localSetF( pAdjNode, type, bDelNonCyclic, bCheckCycle );
    }// while
}

void
TPlanarFVS::localSetL ( TPlanarNode * pNode, uint type,
                        bool bDelNonCyclic, bool bCheckCycle )
{
    if (pNode == NULL)
    {
        cout << "TPlanarFVS::localSetL (...) : got NULL arg" << endl;
        return;
    }// if

    TListIterator< TEdge >  iterEdges;
    TEdge                 * pEdge;
    TPlanarNode           * pAdjNode;

    //
    // look if all succ have attr L
    //
    
    iterEdges.setList( pNode->getOutEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pAdjNode = (TPlanarNode*) pEdge->getHead();

        // check if "really" predecessor
        if (pAdjNode->mark( type ) == FVS_NODE)
            continue;
        
        if (( ! pAdjNode->isDeleted( type )) ||
            ( pAdjNode->isDeleted( type ) && bCheckCycle && (pAdjNode->cycle( type ) != NULL)))
        {  
            if (pAdjNode->mark( type ) != L)
                return;
        }// if
    }// while

    //
    // if all succ have attr L, this node gets attr L and check all pred
    //

    pNode->setMark( L, type );
        
    if (bDelNonCyclic)
        pNode->setDeleted( TRUE, type );
        
    iterEdges.setList( pNode->getInEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pAdjNode = (TPlanarNode*) pEdge->getTail();
        
        if ((pAdjNode->mark( type ) == C) && ( ! pAdjNode->isDeleted( type )))
            localSetL( pAdjNode, type, bDelNonCyclic, bCheckCycle );
    }// while
}

//
// testversion of adding feedback vertices by succ/pred of found feedback vertex
//

void
TPlanarFVS::markPred ( TCycleNode * pNode )
{
    if (pNode == NULL)
    {
        cout << "TPlanarFVS::markPred (...) : got NULL arg" << endl;
        return;
    }// if
    
    TListIterator< TEdge >  iterEdges;
    TEdge                 * pEdge;
    TCycleNode            * pPred;

    iterEdges.setList( pNode->getInEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pPred = (TCycleNode*) pEdge->getTail();

        if (pPred->mark() != DELETED)
            pPred->setMark( IS_PRED );
    }// while
}

void
TPlanarFVS::unmarkPred ( TCycleNode * pNode )
{
    if (pNode == NULL)
    {
        cout << "TPlanarFVS::unmarkPred (...) : got NULL arg" << endl;
        return;
    }// if
    
    TListIterator< TEdge >  iterEdges;
    TEdge                 * pEdge;
    TCycleNode            * pPred;

    iterEdges.setList( pNode->getInEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pPred = (TCycleNode*) pEdge->getTail();

        if (pPred->mark() != DELETED)
            pPred->setMark( NO_MARK );
    }// while
}

void
TPlanarFVS::detectSuccFV ( TCycleNode * pMinNode, TCycleNode * pMaxNode,
                           TList< TNode > * pListFVS )
{
    if ((pMinNode == NULL) || (pMaxNode == NULL))
    {
        cout << "TPlanarFVS::detectSuccFV (...) : got NULL arg" << endl;
        return;
    }// if

    TListIterator< TEdge >  iterEdges;
    TEdge                 * pEdge;
    TCycleNode            * pSucc, * pPred;

    TListIterator< TNode >  iterNodes;
    TPlanarNode           * pNode;
    TPlanarCycle          * pCycle;

    //
    // mark given nodes as used
    //

    pMinNode->setMark( DELETED );
    pMaxNode->setMark( DELETED );
    
    //
    // check all succ of given minnode
    //
    
    iterEdges.setList( pMinNode->getOutEdges() );
    iterEdges.first();
    while ((pEdge = iterEdges++) != NULL)
    {
        pSucc = (TCycleNode*) pEdge->getHead();

        if (pSucc->mark() == DELETED)
            continue;
        
        markPred( pMaxNode );

        //
        // check all nodes on succ if this node also lies
        // on a pred of maxnode
        //
        
        iterNodes.setList( pSucc->cycle()->getNodeList() );
        iterNodes.first();
        while ((pNode = (TPlanarNode*) iterNodes++) != NULL)
        {
            pCycle = pNode->cycle( MAX_CYC );

            if (pCycle == NULL)
                continue;

            pPred = pCycle->getZNode();

            if ((pPred->mark() == DELETED) || (pPred->mark() != IS_PRED))
                continue;

//              cout << "TPlanarFVS::detectSuccFV (...) : detected fv on cycles ("
//                   << pSucc->getName() << "/" << pPred->getName() << ")" << endl;
            
            // add node to fvs
            pNode->addRef();
            pNode->setMark( FVS_NODE, MIN_CYC );
            pNode->setMark( FVS_NODE, MAX_CYC );
            pListFVS->add( * pNode );
            
            unmarkPred( pMaxNode );
            detectSuccFV( pSucc, pPred, pListFVS );

            break;
        }// while
    }// while
}

void
TPlanarFVS::rebuildZ ( TGraph * pZPlus, TGraph * pZMinus,
                       TList< TNode > * pListFVS )
{
    if ((pZPlus == NULL) || (pZMinus == NULL))
    {
        cout << "TPlanarFVS::rebuildZ (...) : got NULL arg" << endl;
        return;
    }// if

    //
    // delete all nodes in Z+ and Z-
    //

    TListIterator< TNode >  iterNodes;
    TCycleNode            * pZNode;
    TPlanarCycle          * pCycle;

    iterNodes.setList( pZPlus->getNodeList() );
    iterNodes.first();
    while ((pZNode = (TCycleNode*) iterNodes++) != NULL)
    {
        pCycle = pZNode->cycle();

        undelete( pCycle, TRUE, MIN_CYC );
        
        delete pCycle;
        pZNode->free();
    }// while

    iterNodes.setList( pZMinus->getNodeList() );
    iterNodes.first();
    while ((pZNode = (TCycleNode*) iterNodes++) != NULL)
    {
        pCycle = pZNode->cycle();

        undelete( pCycle, TRUE, MAX_CYC );

        delete pCycle;
        pZNode->free();
    }// while

    pZPlus->getNodeList().clear();
    pZMinus->getNodeList().clear();

    //
    // reset info for feedback vertices
    //

    TPlanarNode  * pNode;
    
    iterNodes.setList( * pListFVS );
    iterNodes.first();
    while ((pNode = (TPlanarNode*) iterNodes++) != NULL)
    {
        pNode->setDeleted( TRUE, MIN_CYC );
        pNode->setDeleted( TRUE, MAX_CYC );

        pNode->setMark( FVS_NODE, MIN_CYC );
        pNode->setMark( FVS_NODE, MAX_CYC );
    }// while
    
    //
    // rebuild Z+ and Z-
    //
    
    buildZNodes( pZPlus, MIN_CYC );
    buildZEdges( pZPlus, MIN_CYC );

    buildZNodes( pZMinus, MAX_CYC );
    buildZEdges( pZMinus, MAX_CYC );
}
