//
//  Project     : libgraph
//  File        : rcm.cc
//  Author      : Ronald Kriemann
//  Purpose     : class for reversed Cuthill McKee algorithm
//
// arch-tag: 39ba0738-f2c1-4e2c-81ae-d9e805670583
//

#include <limits.h>

#include "rcm.hh"

//
// constructor
//
TRCM::TRCM ()
{}

//
// destructor
//
TRCM::~TRCM ()
{}

//
// doing breadth-first-search
//
int
TRCM::bfs ( TGraph &graph, TMarkableNode &node )
{
	TListIterator<TNode>  iterNodes( graph.getNodeList() );
	TMarkableNode        *pNode;

	int  dist = 0;
	bool bFinished = FALSE;

	TList<TMarkableNode>          listS1, listS2;
	TListIterator<TMarkableNode>  iterAdjNodes;
	TListIterator<TEdge>          iterEdges;
	TEdge                        *pEdge;
	TMarkableNode                *pAdjNode;

	//
	// create visited-info for every node
	//
	iterNodes.first();
	while ((pNode = (TMarkableNode*) iterNodes++) != NULL)
	{
		if ( *((bool*) (pNode->getInfo())) == FALSE )
			pNode->setMark( -1 );
	}// while

	//
	// now do BFS
	//
	listS1.add( node );

	while (!bFinished)
	{
		bFinished = TRUE;

		// go over alle nodes from last level
		iterAdjNodes.setList( listS1 );
		iterAdjNodes.first();
		while ((pNode = iterAdjNodes++) != NULL)
		{
			if ( *((bool*) (pNode->getInfo())) == FALSE )
			{
				// and label all nodes which adj to these nodes
				iterEdges.setList( pNode->getInEdges() );
				iterEdges.first();
				while ((pEdge = iterEdges++) != NULL)
				{
					pAdjNode = (TMarkableNode*)pEdge->getOther( pNode );
					
					if (pAdjNode->mark() == -1)
					{
						pAdjNode->setMark( dist + 1 );
						listS2.add( *pAdjNode );
						bFinished = FALSE;
					}// if
				}// while
			}// if
		}// while

		if (!bFinished)
		{
			dist++;

			// clear list 1 and move all elements from list 2 to list 1
			listS1.clear();

			iterAdjNodes.setList( listS2 );
			iterAdjNodes.first();
			while ((pNode = iterAdjNodes++) != NULL)
			{
				if ( *((bool*) (pNode->getInfo())) == FALSE )
					listS1.add( *pNode );
			}// if
			
			listS2.clear();
		}// if
	}// while

	return dist;
}

//
// find starting node
//
TNode*
TRCM::findStartNode ( TGraph &graph )
{
	TListIterator<TNode>  iterNodes( graph.getNodeList() );
	TMarkableNode        *pNode, *pPeriphNode;

	TMarkableNode  *pMinDegNode;

	bool bPeriphNodeFound = FALSE;
	int  maxDist, ecc;
	uint minDegree;

	//
	// choose arbitrary unused node in graph
	//

	pPeriphNode = NULL;

	iterNodes.first();
	while ((pNode = (TMarkableNode*) iterNodes++) != NULL)
	{
		if ( *((bool*) (pNode->getInfo())) == FALSE )
		{
			pPeriphNode = pNode;
			iterNodes.last();
			iterNodes++;
		}// if
	}// while

	// check if graph is used or empty
	if (pPeriphNode == NULL)
        return NULL;

	//
	// iterate until pseudo-peripheral nodes was found
	//

	// determine all nodes in graph with longest distance from pPeriphNode
	maxDist = bfs( graph, *pPeriphNode );

	while (!bPeriphNodeFound)
	{
		// hope we'll never get this :)
		minDegree = UINT_MAX;

		pMinDegNode = NULL;

		iterNodes.first();
		while ((pNode = (TMarkableNode*) iterNodes++) != NULL)
		{
			if (( *((bool*) (pNode->getInfo())) == FALSE ) &&
				((pNode->mark() == maxDist) && (pNode->getInDegree() < minDegree)))
			{
				pMinDegNode = pNode;
				minDegree   = pNode->getInDegree();
			}// if
		}// while

		// determine eccentricity of pMinDegNode
		ecc = bfs( graph, *pMinDegNode );

		if (ecc > maxDist)
		{
			pPeriphNode = pMinDegNode;
			maxDist     = ecc;
		}// if
		else
			bPeriphNodeFound = TRUE;
	}// while

	return pPeriphNode;
}

//
// do reversed Cuthill McKee
//
void 
TRCM::rcm ( TGraph &graph, TBlockInfo *pBlockInfo )
{
	uint   i;

	TListIterator<TNode>  iterNodes( graph.getNodeList() );
	TMarkableNode        *pNode;

	TList<TMarkableNode>          listS1, listS2, listS3;
	TListIterator<TMarkableNode>  iterAdjNodes, iterS3Nodes;
	TListIterator<TEdge>          iterEdges;
	TEdge                        *pEdge;
	TMarkableNode                *pAdjNode;
	TMarkableNode                *pStartNode;

	bool bFinished = FALSE, bFoundAdj;
	uint maxDegree;

	//
	// create use-info for all nodes: if the graph is not connected, we
	// find a new startnode in alle components which are not used
	//
	bool  bUsed[graph.getNoOfNodes()];

	iterNodes.first(); i = 0;
	while ((pNode = (TMarkableNode*) iterNodes++) != NULL)
	{
		bUsed[i] = FALSE;
		pNode->setInfo( &bUsed[i] );
		i++;
	}// while

	//
	// now do rcm by first determine a startnode unless graph is totaly
	// used and then rcm'ing in the current component
	//

	if (pBlockInfo != NULL)
		pBlockInfo->init( graph.getNoOfNodes() );

	// determine starting node
	i = 0;
	while ((pStartNode = (TMarkableNode*) findStartNode( graph )) != NULL)
	{
		// clear all marks
		iterNodes.first();
		while ((pNode = (TMarkableNode*) iterNodes++) != NULL)
		{
			if ( *((bool*) (pNode->getInfo())) == FALSE ) 
				pNode->setMark( -1 );
		}// while

		listS1.clear();
		listS2.clear();

		if (pBlockInfo != NULL)
			pBlockInfo->addLNode( *pStartNode );

		pStartNode->setMark( i++ );
		listS1.add( *pStartNode );
		// mark node as used
		*((bool*) (pStartNode->getInfo())) = TRUE;

		while (!bFinished)
		{
			bFinished = TRUE;
			
			// collect new nodes which are adjacent to old nodes
			iterAdjNodes.setList( listS1 );
			iterAdjNodes.first();
			while ((pNode = iterAdjNodes++) != NULL)
			{
				bFoundAdj = FALSE;
				maxDegree = 0;
				listS3.clear();

				iterEdges.setList( pNode->getInEdges() );
				iterEdges.first();
				while ((pEdge = iterEdges++) != NULL)
				{
					pAdjNode = (TMarkableNode*) pEdge->getOther( pNode );

					if (pAdjNode->mark() == -1)
					{
						pAdjNode->setMark( i );
						listS3.add( *pAdjNode );

						bFinished = FALSE;
						bFoundAdj = TRUE;
						
						if (pAdjNode->getInDegree() > maxDegree)
							maxDegree = pAdjNode->getInDegree();
					}// if
				}// while
			
				if (bFoundAdj)
				{
					iterS3Nodes.setList( listS3 );
				
					// sort new nodes with respect to degree
					while (listS3.getSize() > 0)
					{
						iterS3Nodes.first();
						while ((pNode = iterS3Nodes++) != NULL)
						{
							if (pNode->getInDegree() == maxDegree)
							{
								listS3.remove( *pNode );
								listS2.add( *pNode );
								iterS3Nodes.first();
								if (pBlockInfo != NULL)
									pBlockInfo->addLNode( *pNode );
								// mark node as used
								*((bool*) (pNode->getInfo())) = TRUE;
							}// if
						}// while
					
						maxDegree--;
					}// while
				}// if
			}// while

			//
			// copy nodes to S1
			//
			if (!bFinished)
			{
				i++;

				listS1.clear();
				iterAdjNodes.setList( listS2 );
				iterAdjNodes.first();
				while ((pNode = iterAdjNodes++) != NULL)
				{
					listS1.add( *pNode );
				}// while

				listS2.clear();
			}// if
		}// while
	}// while
}
