//
//  Project     : libgraph
//  File        : T3DCellEdgeNode.hh
//  Author      : Ronald Kriemann
//  Purpose     : class for a node holding a 3D celledge
//
// arch-tag: 5617bbc4-55f7-4e62-a62e-390dba8c210e
//

#ifndef __T3DCELLEDGENODE_HH
#define __T3DCELLEDGENODE_HH

#include "list.hh"
#include "TFlagable.hh"

#include "vector.hh"

#include "T3DFrontier.hh"
#include "T3DCellEdge.hh"

#include "T3DVertexData.hh"

#include "basenode.hh"

// forward decl.
class T3DFrontier;

// default value for used flow surface function
#define NO_FUNC  -1

//
//
//
class T3DCellEdgeNode : public TNode, public TFlagable<uchar>
{
protected:
    // the represented celledge (and defining net-nodes)
    T3DCellEdge           *pCellEdge;
    T3DVertexData         *pNodeData0, *pNodeData1;

    // a list of adj. faces (adj. with respect to the celledge)
    TList< T3DFrontier >  *pListFaces;

    // conv-diff-direction at the midpoint of the celledge
    TVector<3>             vecConvDir;

    // faces defined by functions phi_f and phi_b
    T3DFrontier           *pPhi_f, *pPhi_b;

    // attached faces in a flow surface
    T3DFrontier           *pFSFace[2];

    // by witch function we obtained face in flow surface
    int                    _usedFSFunc;
    
    // count how many adj. faces are in a flow surfaces
    int                    _noOfUsedFaces;

public:
    //////////////////////////////////////////////////
    //
    // constructors and destructors
    //

    // constructor
    T3DCellEdgeNode ( uint n,
                      T3DCellEdge * edge,
                      T3DVertexData * v0, T3DVertexData * v1,
                      TVector<3> * b );

    // destructor
    virtual ~T3DCellEdgeNode ();

    //////////////////////////////////////////////////
    //
    // access member variables
    //

    virtual T3DCellEdge * cellEdge () const;
    
    virtual T3DVertexData * nodeData ( uint i ) const;

    virtual T3DFrontier * phi_f () const;
    virtual T3DFrontier * phi_b () const;

    virtual const int noOfUsedFaces () const;
    virtual void addUsedFaces ();
    virtual void setNoOfUsedFaces ( uint n );
    
    virtual const bool isDeleted () const;
    virtual void setDeleted ( bool b );

    virtual T3DFrontier * fsFace ( int i = 0 ) const;
    virtual void setFSFace ( T3DFrontier * face );
    virtual void setFSFace ( int i, T3DFrontier * face );
    virtual void clearFSFaces ();

    virtual const int usedFSFunc () const;
    virtual void setUsedFSFunc ( int f );
    
    virtual const bool isStrong () const;
    virtual const bool isWeak () const;
    virtual void setStrong ( bool b );

    virtual const bool isInSurface () const;
    virtual void setInSurface ( bool b );

    //////////////////////////////////////////////////
    //
    // methods for facelist management
    //
    
    // add a adj face (adj. with repsect to the celledge)
    virtual void addFace ( T3DFrontier * face );

    // remove a face from facelist
    // (return TRUE if successfuly deleted, else FALSE)
    virtual bool remFace ( T3DFrontier * face );
    
    // check if a face is in the facelist
    // return TRUE if in facelist, else FALSE
    virtual bool isInFaceList ( T3DFrontier * face );

    // remove all adj weak faces (a face with no strong edge)
    virtual void remWeakFaces ();

    // remove whole face list
    virtual void remFaceList ();
    
    //////////////////////////////////////////////////
    //
    // methods for building flow-surfaces
    //
    
    // calc faces with minimal angle with respect to
    // the conv-dir (NULL if no such face exists)
    // for forward/backward direction
    virtual T3DFrontier * calcPhi ( TFloat gamma );
    
    // return face with minimal angle with repect to
    // given face (NULL if no such face exists)
    virtual T3DFrontier * phiHat ( T3DFrontier * face ) const;

    //////////////////////////////////////////////////
    //
    // virtual constructors
    //
    
	// create a new nodes
	virtual T3DCellEdgeNode * clone ();

protected:
    //////////////////////////////////////////////////
    //
    // misc. help methods
    //

    // determine if a face is strong
    virtual bool isStrongFace ( T3DFrontier * pFace ) const;
};

#endif  // __T3DCELLEDGENODE_HH
