//
//  Project   : libnet
//  File      : T3DMeshGenerator.hh
//  Author    : Ronald Kriemann
//  Purpose   : class for an a simple meshgenerator in 3D
//
// arch-tag: 38b144e6-ddd2-4468-bf2a-c010369b23c0
//

#ifndef __T3DMESHGENERATOR_HH
#define __T3DMESHGENERATOR_HH

#include "memory.hh"
#include "dynarray.hh"
#include "list.hh"

#include "vertex.hh"
#include "T3DCellEdge.hh"
#include "T3DFrontier.hh"
#include "net.hh"
#include "TDomain.hh"

//
// class for representing a quad
// - quad will be decomposed by Kuhn's triangulation
//
class T3DQuad : public TMemory
{
protected:
    // set of vertices assoc. with this quad
    TDynArray< TVertex<3> * >  arrVertices;

    // set of edges assoc. with this quad
    TDynArray< T3DCellEdge * > arrEdges;

    // set of frontiers assoc. with this quad
    TDynArray< T3DFrontier * > arrFrontiers;

    // set of cells assoc. with this quad
    TDynArray< TCell<3> * >    arrCells;
    
    // references to neighbour-quads
    T3DQuad  * arrNeighbours[6];

    // domain to def. boundaries for net-elements
    TDomain  * pDomain;
    
public:
    ///////////////////////////////////////////
    //
    // constructors and destructor
    //

    T3DQuad ( TDomain * domain );

    T3DQuad ( TDomain * domain,
              TVertex<3> * v0, TVertex<3> * v1,
              TVertex<3> * v2, TVertex<3> * v3,
              TVertex<3> * v4, TVertex<3> * v5,
              TVertex<3> * v6, TVertex<3> * v7 );
               
    ~T3DQuad ();

    ///////////////////////////////////////////
    //
    // access member variables
    //

    void initData ();

    T3DQuad * neighbour ( uint i );

    void setNeighbour( uint i, T3DQuad * pNeighbour );
    void setNeighbours ( TList< T3DQuad > * quads );
    
    void checkNeighbours ();

    TVertex<3> * vertex ( uint i );

    T3DCellEdge * edge ( uint i );
    T3DCellEdge * edge ( uint i, uint j );

    T3DFrontier * frontier ( uint i );
    T3DFrontier * frontier ( uint i, uint j, uint k );

    TCell<3> * cell ( uint i );
    
    ///////////////////////////////////////////
    //
    // build/handle net-elements
    //

    bool buildVertex ( uint i, TFloat x, TFloat y, TFloat z );
    void buildEdge ( uint i );
    void buildFrontier ( uint i );
    void buildCell ( uint i );

    void buildAllEdges ();
    void buildAllFrontiers ();
    void buildAllCells ();

    void addCells ( TNet<3> * pNet );
};

//
// class for a simple meshgenerator in 3D
//
class T3DMeshGenerator : public TMemory
{
protected:
    // list holding all quads
    TList< T3DQuad >  listQuads;
    
public:
    ///////////////////////////////////////////
    //
    // constructors and destructor
    //

    T3DMeshGenerator ();

    ~T3DMeshGenerator ();

    ///////////////////////////////////////////
    //
    // access member variables
    //

    TList< T3DQuad > & quads ();
    
    ///////////////////////////////////////////
    //
    // build net / hierarchy
    //
    
    TNet<3> * buildNet ();
    
    TNet<3> * buildNet ( uint sizeX, TFloat arrCoordX[],
                         uint sizeY, TFloat arrCoordY[],
                         uint sizeZ, TFloat arrCoordZ[],
                         TDomain * pDomain );

    void addQuad ( T3DQuad * quad );
};

#endif  // __T3DMESHGENERATOR_HH
