//
//  Project     : libgraph
//  File        : T3DSurfaceGraph.hh
//  Author      : Ronald Kriemann
//  Purpose     : class for a graph defined by flow surfaces
//
// arch-tag: c9af0553-8c6e-48da-822f-60a21be2d801
//

#ifndef __T3DSURFACEGRAPH_HH
#define __T3DSURFACEGRAPH_HH

#include "basegraph.hh"

// forward declaration
class T3DCellEdgeGraph;
class T3DPlanarNode;
class TMatrixBlock;
class T3DVertexData;

//
//
//
class T3DSurfaceGraph : public TGraph
{
protected:
    // matrix to decide if edge is strong
    TMatrixBlock  * pMatrixBlock;

    // criterion and parameter for building edges
    uint            criterion;
    TFloat          xi;
    
public:
    /////////////////////////////////////
    //
    // constructor and destructor
    //
    T3DSurfaceGraph ( T3DCellEdgeGraph * pFlowSurface,
                      TMatrixBlock * pMatrix,
                      uint criterion, TFloat xi,
                      bool bSortOFD = FALSE );

    virtual ~T3DSurfaceGraph ();

protected:
    /////////////////////////////////////
    //
    // encapsulate graph building
    //
    
    virtual void buildGraph ( T3DCellEdgeGraph * pFlowSurface );
    virtual void buildGraph_OFD ( T3DCellEdgeGraph * pFlowSurface );

    // check if given edge is strong
    virtual const bool isStrong ( T3DPlanarNode * pNode0,
                                  T3DPlanarNode * pNode1 ) const;

    // build a new edge from node0 to node1
    virtual void buildEdge ( T3DPlanarNode * pNode0,
                             T3DPlanarNode * pNode1 ) const;

    // build node for given nodedata or simply return node
    // if already builded
    virtual T3DPlanarNode * buildNode ( T3DVertexData * pData );

    // build edges for given node sorted by angle (OFD)
    virtual void buildEdges_OFD ( T3DPlanarNode * pNode,
                                  T3DCellEdgeNode * pCellEdge );

    // return adjacent face of celledge with orientation
    // compatible to (pNode0,pNode1)
    T3DFrontier * getFaceByOrient ( T3DCellEdgeNode * pCellEdge,
                                    T3DPlanarNode * pNode0,
                                    T3DPlanarNode * pNode1 ) const;

    // return other (than e) celledge in face with common node v
    T3DCellEdgeNode * otherCellEdge ( T3DPlanarNode * pNode,
                                      T3DCellEdgeNode * pCellEdge,
                                      T3DFrontier * pFace ) const;

    // return face adjacent to given edge and not the given face
    T3DFrontier * nextFace ( T3DCellEdgeNode * pCellEdge,
                             T3DFrontier * pFace ) const;

    // add edge to list of incident edges of node
    void addEdge ( T3DPlanarNode * pNode,
                   T3DCellEdgeNode * pEdge,
                   TList< T3DPlanarNode > * pList,
                   bool bReverse );

    // build edges of node on graph w.r.t. OFD-condition
    // out of given nodelist of adjacent nodes
    void buildSortedEdges ( T3DPlanarNode * pNode,
                            TList< T3DPlanarNode > * pList );

    // create edge or return edge if in given edgelist
    TEdge * buildInEdge  ( T3DPlanarNode  * pNode,
                           T3DPlanarNode  * pAdjNode,
                           TList< TEdge > * pList );
    TEdge * buildOutEdge ( T3DPlanarNode  * pNode,
                           T3DPlanarNode  * pAdjNode,
                           TList< TEdge > * pList );
};

#endif  // __T3DSURFACEGRAPH_HH
