//
//  Project   : libnet
//  File      : TBasisFE.hh
//  Author    : Sabine Gutsch / Thomas Probst
//  Purpose   : basis class for finite elements
//
// arch-tag: 93920702-d0e8-4df3-9799-ec57983eaf75
//

#ifndef __TBASISFE_HH
#define __TBASISFE_HH

#include "memory.hh"

#include "vector.hh"
#include "TMatrix.hh"
#include "blockvector.hh"
#include "blockmatrix.hh"
#include "nethierarchy.hh"

#include "TBasisEqn.hh"

// defines for verbosity
const uint PRINT_DISCRETISATION = 1;



typedef TFloat (*PFloatFunc)( TFloat x );

TFloat simple ( TFloat x );
TFloat Samarskij ( TFloat x );

// (upwind) discretization types for convection
enum TDiscrType { Tabata, Box, Null };
//enum TUpwindType { Simple, Samarskij };

//
// basis class for a finite element 
//
template <int N>
class TBasisFE : public TMemory {
  
protected:
    // type of discretisation of the convective part (Box/Tabata)  
    TDiscrType DiscrType; 
    // function for upwinding (simple or Samarskij)
    PFloatFunc PHI;
    // verbosity of created messages
    uint                 verbosity;
    
public:
    // constructors	
    TBasisFE ( TDiscrType aDiscrType = Null, uint averb=0 ) 
    { 
        DiscrType = aDiscrType;
        verbosity = averb;
    }
    // destructor
    virtual ~TBasisFE () { }
 
    // set and get verbosity 
    void setVerbosity (const uint averb  );
    uint getVerbosity () const;
	 
    // set new DiscrType
    void setDiscrType(const TDiscrType newDiscrType  ) 
    { DiscrType = newDiscrType; }
  
    // get new DiscrType
    TDiscrType getDiscrType() { return DiscrType; }
  
    // set Phi for box method (Samarskij or simple)
    void setSamarskij() { PHI = Samarskij; }
    void setSimple() { PHI = simple; }

    PFloatFunc getPHI () { return PHI; }

    // calculate global stiffness matrix for diffusion part
    // and write it into global matrix
    virtual void discrDiff ( TBasisEqn<N> *eqn, TNet<N> *net, 
                             TBlockMatrix *A, 
                             TBlockVector<TFloat> *rhs ) = 0; 
  
  // calculate (non-symmetric) convection part
  virtual void discrConv ( TBasisEqn<N> *eqn, TNet<N> *net, 
                           TMatrixBlock *C, 
                           TBlockVector<TFloat> *rhs, uint block = 0 ) = 0; 

    // calculate (lumped) massmatrix
    virtual void discrMass ( TBasisEqn<N> *eqn, TNet<N> *net, 
                             TVectorBlock<TFloat> *M, 
                             TBlockVector<TFloat> *rhs) = 0; 


    // get coordinates of net
    virtual void getCoord ( TNet<N> *net, TDynArray<TVector <N> * >& arrCoord, uint block = 0 ) = 0;

    // set number of interior nodes of nets
    virtual void setInterior( uint cInt, uint fInt){};
};

// some usual types
typedef TBasisFE<2>  T2DBasisFE;
typedef TBasisFE<3>  T3DBasisFE;

#endif  // __TBASISFE_HH
