//
//  Project     : libnet
//  File        : TMatrixData.hh
//  Author      : Ronald Kriemann
//  Purpose     : class for a sparse row in a matrix
//
// arch-tag: 018e51fb-cd8d-4798-a41f-bff01dc73fbc
//

#ifndef __TMATRIXDATA_HH
#define __TMATRIXDATA_HH

#include <limits.h>

#include "memory.hh"
#include "defines.hh"

// forward decl.
class TPermutation;
template <class T> class TVectorBlock;

//
// class TMatrixData
//
// this class is used for a line in a matrixblock and
// for a block in the sparse blockvector
//
class TMatrixData : public TMemory 
{
protected:
    // array with indices we store and the assoc. values
    uint   * arrIndex;
    TFloat * arrValue;
    uint     noOfEntries;
    bool     writeable;

    // flag to set usage of criterion in output
    static bool bUseCriterion;
    
public:
    // constructors
    TMatrixData ();
    TMatrixData ( uint s );
    TMatrixData ( const TMatrixData & data );

    // destructor
    ~TMatrixData ();

    // set usage of criterion in output
    static void useCriterion ( bool b );
    
    // change writeable status
    void writeableStatus ( const bool flag );

    // set the indexarray and the valuearray
    void set ( uint s, uint * index, TFloat * values );
    void set ( const TMatrixData & md );
    // keep the first s entries of the index- and valuearray
    void set ( uint s );

    // get noOfEntries
    uint getNoOfEntries () const;

    // return number of the i'th index
    uint   getIndex ( uint i ) const;
    uint * getIndex () const;

    // set number of the i'th index
    void setIndex ( uint i, uint j, TFloat k = 0.0 );

    // return i'th value
    TFloat getValue ( uint i ) const;
    TFloat* getValue () const;
    // set i'th value
    void setValue ( uint i, TFloat val ) const;

    // has Index i within the first j indices
    bool hasIndex ( uint i, uint j );

    // copy operator
    TMatrixData& operator= ( const TMatrixData& data );

    // index operator 
    TFloat& operator[] ( uint i );
    const TFloat& operator[] ( uint i ) const;

    // operator * with a scalar
    TMatrixData& operator*= ( TFloat scalar );

    // operator * (between matrix data and vector block)
    friend TFloat operator* ( const TMatrixData& matrixdata,
                              const TVectorBlock<TFloat>& v );

    // inner product between TMatrixData as VectorBlocks
    // (assumption: both matrixdata objects have same index structure)
    TFloat operator* ( const TMatrixData& data );

    // operator += with another blockline (assume, that they have 
    // same indexstructure, and all indices are set correct)
    TMatrixData& operator+= ( const TMatrixData& matrixdata );

    // operator -= with another blockline (assume, that they have 
    // same indexstructure, and all indices are set correct)
    TMatrixData& operator-= ( const TMatrixData& matrixdata );

    // invLowerMult : multiplication 
    // (assumes diagonal entry in arrValue[0] )
    TFloat invLowerMult ( const TVectorBlock<TFloat> & v ) const;
    TFloat invLowerMult ( const TVectorBlock<TFloat> & v,
                          const TFloat & diagDamp ) const;
    TFloat invUnitLowerMult ( const TVectorBlock<TFloat> &v) const;

    // multiply with all lower entries < upperLimit
    TFloat lowerMult ( const TVectorBlock<TFloat> &v,
                       uint upperLimit = UINT_MAX ) const;

    // invUpperMult : multiplication 
    // (assumes diagonal entry in arrValue[0] )
    // upperLimit for block methods
    TFloat invUpperMult ( const TVectorBlock<TFloat> &v, 
                          uint upperLimit = UINT_MAX ) const;
    TFloat invUpperMult ( const TVectorBlock<TFloat> &v, 
                          const TFloat & diagDamp ) const;

    // multiply with all upper entries > lowerLimit
    TFloat upperMult ( const TVectorBlock<TFloat> &v,
                       uint lowerLimit = 0 ) const;

    // multiply with all entries between upper and lower:
    // ( lowerLimit <= entry index < upperLimit )
    TFloat mult ( const TVectorBlock<TFloat> &v,
                  uint lowerLimit, uint upperLimit ) const;
                          

    // permutation for row of matrixblock
    void permute ( const TVectorBlock<uint> * q, const TMatrixData & md );
    void permute ( const TPermutation * pPerm, const TMatrixData & md );

    // output
    friend ostream& operator<< ( ostream& os, const TMatrixData & matrixdata );
};

#endif  // __TMATRIXDATA_HH
