//
//  Project   : libmisc
//  File      : TSList.hh
//  Author    : Ronald Kriemann
//  Purpose   : Template class for a single concatenated list
//
// arch-tag: 20d91c82-ea3b-4fb8-a33b-acf2095f52a2
//

#ifndef __TSLIST_HH
#define __TSLIST_HH

#include "defines.hh"
#include "memory.hh"

////////////////////////////////////////////////////
//
// class for a single item in the list
//
template <class T> 
class TSListItem // : public TMemory
{
public:
	T               element;
	TSListItem<T> * pNext;
	
public:
	TSListItem ( const T &       e, 
                 TSListItem<T> * next = NULL )
	{
		element = e;
		pNext   = next;
	}
};

////////////////////////////////////////////////////
//
// forward declaration for the iterator
//
template<class T> class TSListIterator;

////////////////////////////////////////////////////
// 
// Template class list
//
// - stores pointers to objects in the list
//
template <class T> 
class TSList : public TMemory 
{
	// set the iterator as friend
	friend class TSListIterator<T>;

protected:
	TSListItem<T>  *pFirst;       // pointer to the first element
	TSListItem<T>  *pLast;        // and to the last one
	TSListItem<T>  *pActItem;     // act element in list access
	int             nSize;        // act. size of the list

public:
    /////////////////////////////////////////////////
    //
    // constructors and destructors
    //

	// constructors
	TSList ();

	// destructor
	virtual ~TSList ();

    /////////////////////////////////////////////////
    //
    // access member variables
    //

	const int getSize () const;

	const T & getFirst () const;
	const T & getLast () const;

    /////////////////////////////////////////////////
    //
    // methods for manipulating list
    //

	// add a new object to the list
	virtual void append ( const T & object );

	// add a new object to the begin of the list
	virtual void prepend ( const T & object );

	// insert item after given item
	virtual void insertAfter ( const T & after, const T & object );

	// insert item before given item
	virtual void insertBefore ( const T & before, const T & object );

	// replace one item by another
	virtual void replace ( const T & oldObj, const T & newObj );

	// remove element from list
	virtual void remove ( const T & object );

	// remove all elements from list
	virtual void removeAll ();

    /////////////////////////////////////////////////
    //
    // search/sort related methods
    //

	// search for an object and return the item
	virtual TSListItem<T> * search ( const T & object );

    // look if element is in list
    const bool isIn ( const T & object );
    
protected:
    /////////////////////////////////////////////////
    //
    // misc. help-methods
    //

    // encapsulate item creation
    TSListItem<T> * createItem ( const T & element,
                                 TSListItem<T> * next ) const;
};

//////////////////////////////////////////////////////////
// 
// class for an iterator for the previous list
//
template<class T>
class TSListIterator : public TMemory 
{
protected:
	TSListItem<T>   *pCurrItem;  // current selected item
	TSList<T>       *pList;      // used list for work

public:
	// constructor
	TSListIterator ();
	
	// constructor
	TSListIterator ( TSList<T> * list );

	// set the list, we iterate
	void setList ( TSList<T> * list );

	// return next object from list
	T * next ();

	// set counter to begin of list
	T * first ();

	// set counter to end of list
	T * last ();

	// search for an object and set current pointer to it
	bool search ( T * object );

	//
	// operators for easy access
	//

	// preincrement
	T * operator ++ ();

	// postincrement
	T * operator ++ (int);

	// access actual element
	T * operator () ();
};

//////////////////////////////////////////////////////////////
//
// TSList (Implementation)
//

//
// constructors
//
template <class T>
TSList<T>::TSList () 
{
	pFirst = pLast = pActItem = NULL;
	nSize = 0;
}

//
// destructor
//
template <class T>
TSList<T>::~TSList () 
{
	if (nSize > 0)
        removeAll(); 
}

/////////////////////////////////////////////////
//
// access member variables
//

template <class T>
const int 
TSList<T>::getSize () const 
{ 
	return nSize; 
}

template <class T>
const T &
TSList<T>::getFirst () const 
{ 
	if (pFirst != NULL)
		return pFirst->element;
	else
		return NULL; 
}

template <class T>
const T &
TSList<T>::getLast () const
{ 
	if (pLast != NULL)
		return pLast->element; 
	else
		return NULL;
}

/////////////////////////////////////////////////
//
// methods for manipulating list
//

//
// append a new object to the end of list
//
template <class T>
void 
TSList<T>::append ( const T & object ) 
{
	if (pFirst == NULL)
		pLast = pFirst = createItem( object, NULL );
	else
	{
		pLast->pNext = createItem( object, NULL );
		pLast = pLast->pNext;
	}// else

	nSize++;
}

//
// add a new object to the begin of the list
//
template <class T>
void 
TSList<T>::prepend ( const T & object ) 
{
	if (pFirst == NULL)
		pLast = pFirst = createItem( object, NULL );
	else
		pFirst = createItem( object, pFirst );

	nSize++;
}

//
// insert item after given item
//
template <class T>
void 
TSList<T>::insertAfter ( const T & after, const T & object ) 
{
	TSListItem<T> *pItem = search( after );

	if (pItem != NULL)
	{
		TSListItem<T> *pTmp = createItem( object, pItem->pNext );

		if (pLast == pItem) 
			pLast = pTmp;

		pItem->pNext = pTmp;

		nSize++;
	}// if
	else
		cout << "TSList::insertAfter (...) : could not insert object" << endl;
}

//
// insert item before given item
//
template <class T>
void 
TSList<T>::insertBefore ( const T & before, const T & object ) 
{
    TSListItem<T>  * pItem    = pFirst;
    TSListItem<T>  * pOldItem = NULL;
    TSListItem<T>  * pNewItem = NULL;
    
    while (pItem != NULL)
    {
        if (pItem->element == before)
        {
            pNewItem = createItem( object, pItem->pNext );

            if (pItem == pFirst)
                pFirst = pNewItem;
            else
                pOldItem->pNext = pNewItem;

            nSize++;
            return;
        }// if
    }// while
    
    cout << "TSList::insertBefore (...) : \"before\" not in list" << endl;
}

//
// replace one item by another
//
template <class T>
void 
TSList<T>::replace ( const T & oldObject, const T & newObject )
{
    TSListItem<T>  * pItem = pFirst;

    while (pItem != NULL)
    {
        if (pItem->element == oldObject)
            pItem->element = newObject;

        pItem = pItem->pNext;
    }// while
}

//
// remove item from list (access by element)
//
template <class T>
void 
TSList<T>::remove ( const T & object ) 
{
    TSListItem<T>  * pItem    = pFirst;
    TSListItem<T>  * pOldItem = NULL;

    while (pItem != NULL)
    {
        if (pItem->element == object)
        {
            if (pItem == pFirst)
                pFirst = pItem->pNext;
            else // (pItem != pFirst) => (pOldItem != NULL)
                pOldItem->pNext = pItem->pNext;

            delete pItem;
            nSize--;

            pItem = pOldItem->pNext;
        }// if
        else
        {
            pOldItem = pItem;
            pItem = pItem->pNext;
        }// else
    }// while
}

//
// remove all elements from list
//
template <class T>
void 
TSList<T>::removeAll ()
{
	if (nSize == 0)
        return;

	TSListItem<T> * pItem = pFirst;
	TSListItem<T> * pTmp;

	while (pItem != NULL)
	{
		pTmp = pItem->pNext;
        
		if (pItem != NULL)
            delete pItem;
        
		pItem = pTmp;
	}// while

	pFirst = pLast = pActItem = NULL;
	nSize  = 0;
}

/////////////////////////////////////////////////
//
// search/sort related methods
//

//
// search for an object and return the item
//
template <class T>
TSListItem<T>* 
TSList<T>::search ( const T & object ) 
{
	TSListItem<T> * pItem   = pFirst;

	if ((pActItem != NULL) && (pActItem->element == object))
	{
		return pActItem;
	}// if

	while (pItem != NULL)
	{
		if (pItem->element == object)
		{
			pActItem = pItem;
			return pItem;
		}// if

		pItem = pItem->pNext;
	}// while

	return NULL;
}

//
// look if element is in list
//
template <class T>
const bool
TSList<T>::isIn ( const T & object )
{
    if (search( object ) != NULL)
        return TRUE;
    else
        return FALSE;
}
    
/////////////////////////////////////////////////
//
// misc. help-methods
//

//
// encapsulate item creation
//
template <class T>
TSListItem<T> *
TSList<T>::createItem ( const T & element,
                        TSListItem<T> * next ) const
{
    return new TSListItem<T>( element, next );
}

//////////////////////////////////////////////////////////////
//
// TSListIterator (Implementation)
//

//
// constructor
//
template <class T>
TSListIterator<T>::TSListIterator ()
{
	pCurrItem = NULL;
	pList     = NULL;
}
	
template <class T>
TSListIterator<T>::TSListIterator ( TSList<T> * list )
{
	setList( list );
}

//
// set the list, we iterate
//
template <class T>
void 
TSListIterator<T>::setList ( TSList<T> * list )
{
	pList = list;

	if (pList != NULL)
		pCurrItem = pList->pFirst;
	else
		pCurrItem = NULL;
}

//
// return next object from list
//
template <class T>
T * 
TSListIterator<T>::next () 
{
	return ++(*this);
}

//
// set counter to begin of list
//
template <class T>
T * 
TSListIterator<T>::first () 
{
	if (pList == NULL) return NULL;

	pCurrItem = pList->pFirst;

	if (pCurrItem != NULL)
	{
		pList->pActItem = pCurrItem;
		return & pCurrItem->element;
	}// if
	else
		return NULL;
}

//
// set counter to end of list
//
template <class T>
T * 
TSListIterator<T>::last () 
{
	if (pList == NULL) return NULL;

	pCurrItem = pList->pLast;

	if (pCurrItem != NULL)
	{
		pList->pActItem = pCurrItem;
		return & pCurrItem->element;
	}// if
	else
		return NULL;
}

//
// search for an object and set current pointer to it
//
template <class T>
bool 
TSListIterator<T>::search ( T * object )
{
	if (pList != NULL)
	{
		pCurrItem = pList->search( object );

		if (pCurrItem == NULL)
			return FALSE;
		else
			return TRUE;
	}// if
	else
	{
		pCurrItem = NULL;
		return FALSE;
	}// else
}

//
// preincrement
//
template <class T>
T * 
TSListIterator<T>::operator ++ ()
{
	if ((pCurrItem != NULL) && (pCurrItem->pNext != NULL))
	{
		pCurrItem = pCurrItem->pNext;
		pList->pActItem = pCurrItem;
		return & pCurrItem->element;
	}// if

	return NULL;
}

//
// postincrement
//
template <class T>
T * 
TSListIterator<T>::operator ++ (int)
{
	T * tmp;

	if (pCurrItem != NULL)
	{
		pList->pActItem = pCurrItem;
		tmp = & pCurrItem->element;
		pCurrItem = pCurrItem->pNext;
	}// if
	else
		tmp = NULL;

	return tmp;
}

//
// access actual element
//
template <class T>
T * 
TSListIterator<T>::operator () ()
{
	if (pCurrItem != NULL)
	{
		pList->pActItem = pCurrItem;
		return & pCurrItem->element;
	}// if
	else
		return NULL;
}

#endif   // __LIST_HH
