#ifndef __TSTOKESFE_HH
#define __TSTOKESFE_HH
//
//  Project   : libnet
//  File      : TStokesFE.hh
//  Author    : Thomas Probst / Sabine Gutsch
//  Purpose   : class for linear conforming 2D or 3D finite elements
//  Date      : 21.05.1997/31.07.98
//  Changes   : 
//  Questions : do we have to initialize floats to zero?
//  Problems  :
//
// arch-tag: 94952b44-4794-458d-b98c-7682d5541899
//

#include "TBasisFE.hh"
#include "TStokesBasisEqn.hh"
#include "parray.hh"
#include "TPDEBoundaryDesc.hh"
#include "TDoFDesc.hh"

//
// class for a conforming linear finite element 
//

template <int N>
class TStokesFE : public TBasisFE<N> {

protected:
	// number of children
    uint  noOfChildren;
	// gradients of (N+1) basis functions on (2^N) children cells
    TVector<N> gradT[1<<N][N+1];
    // determinants of transformations for children cells
    TFloat     detJ[1<<N];
    // sum of values of coarse basis functions on children cells
    TFloat     c[1<<N][N+1];
	// dofs for u, p (#childeren, #unknowns/vertex, #vertices)
    long       dof_u[1<<N][N+1][N], dof_p[N+1];
	// coordinates of vertices of children cells
    TVector<N> coord[1<<N][N+1], pCoord[N+1];   
    //  size of cell and of (children) elements;
    TFloat vol_p;
    TFloat vol_u[1<<N];
    // number of interiour nodes in coarse/fine grid 
    uint cInterior;  
    uint fInterior;   
    // boundary descriptions for u 
    TPDEBoundaryDesc<N>* boundaryDesc_u[1<<N][N+1];
public:
	// constructors	
    TStokesFE ( TDiscrType aDiscrType = Null);
	
	// destructor
    ~TStokesFE ();

	// get coordinates of vertices of children cells
    void getCoord ( TCell<N> *cell );

	// get coordinates of all vertices and write them into global vector
	void getCoord ( TNet<N> *net, uint k,
                    TDynArray<TVector<N>*>& arrCoordU,
                    TDynArray<TVector<N>*>& arrCoordP  );
    // get coordinates of net
    virtual void getCoord ( TNet<N> *, TDynArray<TVector<N>*>&, uint block = 0 );
   
	//void getVertices ( TNet<N> *net, TVertex<N> **verU );
    bool isDof (uint i, uint j, uint k);

	// init dofs
    void initDofs ( TCell<N> *cell );   

	// initialize grad, det, c, dof_u and dof_p for a particular cell
    void initDiscr ( TCell<N> *cell );
   
	// calculate global stiffnes matrix
    virtual void discrDiff ( TBasisEqn<N> *aeqn, TNet<N> *net, 
                    TBlockMatrix *A, 
                    TBlockVector<TFloat> *rhs );
   
	// calculate stiffness matrix of non-symmentric part
    virtual void discrConv ( TBasisEqn<N> *aeqn, TNet<N> *net, 
                             TMatrixBlock *C, 
                             TBlockVector<TFloat> *rhs,
                             uint block = 0 );
	// calculate mass matrix
	virtual void discrMass( TBasisEqn<N> *aeqn, TNet<N> *net, 
                            TVectorBlock<TFloat> *M, 
                            TBlockVector<TFloat> *rhs);
   
    // calculate element stiffness matrix
    virtual void discrDiff ( TStokesBasisEqn<N> *eqn, TBlockMatrix *A, 
                    TBlockVector<TFloat> *rhs );
    virtual void discrUP ( TStokesBasisEqn<N> *eqn, TBlockMatrix *A,  
                  TBlockVector<TFloat> *rhs );
    virtual void discrConv ( TStokesBasisEqn<N> *eqn, TMatrixBlock *C, 
                    TBlockVector<TFloat> *rhs );
    virtual void discrConv ( TStokesBasisEqn<N> *eqn, TMatrixBlock *C, 
                    TBlockVector<TFloat> *rhs,  
                    TVectorBlock<TFloat> &suppSize, uint block = 0 );
    virtual void  discrMass ( TStokesBasisEqn<N> *eqn, 
                              TVectorBlock<TFloat> *M, 
                            TBlockVector<TFloat> *rhs);
    virtual void discrRhs ( TBasisEqn<N> *aeqn, TNet<N> *net, 
                            TBlockVector<TFloat> *rhs ); 
    virtual void discrRhs ( TStokesBasisEqn<N> *eqn, 
                            TBlockVector<TFloat> *rhs ); 
    void updateSuppSize ( TVectorBlock<TFloat> &suppSize, uint k );
	// set number of interior nodes on coarse and fine grid,
	// needed for calculation of boundary node names
    void setInterior ( uint cInt, uint fInt );
   
};

#endif  // __TSTOKESFE_HH
