//
//  Project   : libnet
//  File      : basisEqn.hh
//  Author    : Sabine Gutsch / Thomas Probst
//  Purpose   : basis class for pdes
//
// arch-tag: d6135ec8-3526-4345-8beb-571572f93666
//

#ifndef __BASISEQN_HH
#define __BASISEQN_HH

#include "memory.hh"
#include "parray.hh"

#include "vertex.hh"

#include "vector.hh"
#include "blockvector.hh"

#include "convDirections.hh"

typedef TFloat (*P2DVectorFunc)( TVector<2> );

//
// basis class for a pde
//
class T2DBasisEqn : public TMemory
{
protected:
	// viscosity in Navier-Stokes equation / factor for diffusion
	// term in convection-diffusion equation
    TFloat epsilon;

	// previous approximate/convection direction
    TBlockVector<TFloat> *convVector;

    // convection  functions
    P2DVectorFunc b0;
    P2DVectorFunc b1;

    // reaction  function
    P2DVectorFunc c;


public:
	// constructors	
    T2DBasisEqn ( TFloat aEpsilon = 1.0, 
                TBlockVector<TFloat> *aConvVector = NULL ) 
	{ 
        epsilon = aEpsilon;
        convVector = aConvVector;
        b0   = noConv0; 
        b1   = noConv1;
        c    = noReac;
    }
	// destructor
    virtual ~T2DBasisEqn () { }

	// get and set epsilon
    TFloat getEpsilon( ) { return (epsilon); }
    void setEpsilon ( TFloat aEps ) { epsilon = aEps; }

	// get and set convection direction
    void setConvVector ( TBlockVector<TFloat> *newConvVector ) {
        convVector = newConvVector; 
    }
    TBlockVector<TFloat>* getConvVector ( ) { return (convVector); }
 
    // init convVector using the convection functions
    void initConvVector ( uint size, TDynArray< TVector<2> * >& coordVec )
    {
        uint no[2];

        if ( b0 == NULL ) 
            cerr << "TBasisEqn::initConvDirection : no convection defined\n"; 
        
        no[0] = no[1] = size;

        // allocation of convVector (if necessary)
        if ( convVector == NULL )
            convVector = new TBlockVector<TFloat> ( 2, no );
        
        // evaluate b0 and b1 at the coordinates belonging to unknown 'k'
        // write result to convVector
        for ( uint k = 0; k < size; k++ ) {
            (*convVector)[0][k] = (*b0)( *(coordVec[k]) );
            (*convVector)[1][k] = (*b1)( *(coordVec[k]) );
        }
    }

	// delete convection Vector
    void deleteConvVector () {
        delete (convVector);
        convVector = NULL;
	}
  
    // get and set convection/ reaction functions
    P2DVectorFunc getConvb0() { return (b0); }
    P2DVectorFunc getConvb1() { return (b1); }
    P2DVectorFunc getReacc() {  return (c); }
  
    void setConvFunctions(const  P2DVectorFunc z0,const  P2DVectorFunc z1) {
        b0 = z0;
        b1 = z1;
    }
    void setReacFunction( const  P2DVectorFunc z) { c = z; }
  
    // rhs functions (for velocity)
    virtual TFloat f0 ( const TVector<2> &) = 0;
    virtual TFloat f1 ( const TVector<2> &) { return (0.0); }
  
    // functions for Dirichlet boundary
    virtual TFloat dirichlet0 ( const TVector<2> &) = 0;
    virtual TFloat dirichlet1 ( const TVector<2> &) { return (0.0); }
};

#endif  // __BASISEQN_HH
