#ifndef __DYNARRAY_HH
#define __DYNARRAY_HH
//
//  Project  : libmisc
//  File     : dynarray.hh
//  Author   : Ronald Kriemann
//  Purpose  : class for a dyn. array of templates
//  Date     : 19.05.1997
//
// arch-tag: 4a81077b-f6b2-4374-8544-fe249e96c6ac
//

#include "defines.hh"
#include "memory.hh"

//
// class for a template-vector
//
template <class T> 
class TDynArray : public TMemory 
{
protected:
	T*    pData;        // the array with the stored elements
	uint  nSize;        // number of the element with the max. index
	uint  nRealSize;    // real size of the array
	uint  nGrowBy;      // if index > nRealSize the array will grow to index+nGrowBy+1

public:
	// constructor
	TDynArray ();
	TDynArray ( uint size, uint growby = 0 );
	TDynArray ( const TDynArray<T>& vec );

	// default destructor
	virtual ~TDynArray ();

	// return size of the vector
	virtual const uint getSize () const;

	// set a new size but keep the old elements
	virtual void setSize ( uint n );

	// set the growby size new
	virtual void setGrowBy ( uint growby );

	// access the vector as an array
	virtual T& operator [] ( uint i );
	virtual const T& operator [] ( uint i ) const;

	// copy operator
	virtual TDynArray<T>& operator = ( const TDynArray<T>& vec );
};

/////////////////////////////////////////////////////
//
// implementation
//

//
// default constructor
//
template <class T> 
TDynArray<T>::TDynArray ()
{
	nSize = nRealSize = 0;
	nGrowBy = 0;
	pData = 0;
}

//
// constructor with default size
//
template <class T> 
TDynArray<T>::TDynArray ( uint size, uint growby )
{
	nSize     = size;
	nGrowBy   = growby;
	nRealSize = size+nGrowBy;
		
	pData = new T[nRealSize];
}

//
// copy constructor
//
template <class T> 
TDynArray<T>::TDynArray ( const TDynArray<T>& vec )
        : pData(NULL), nSize(0)
{ 
	(*this) = vec; 
}

//
// default destructor
//
template <class T> 
TDynArray<T>::~TDynArray ()
{ 
	if (pData) delete[] pData; 
}

//
// return size of the vector
//
template <class T> 
const uint 
TDynArray<T>::getSize () const 
{ 
	return (nSize); 
}

//
// set a new size but keep the old elements
//
template <class T> 
void 
TDynArray<T>::setSize ( uint n )
{
	uint  i;

	if (nSize != n)
	{
		if (n == 0) 
		{
			if (pData) delete[] pData;
			pData = NULL;
			nSize = nRealSize = 0;
		}// if
		else if (n > nRealSize)
		{
			nRealSize = n + nGrowBy;
			T *tmpData = new T[nRealSize];
				
			if (pData != NULL)
			{
				T *p1 = tmpData;
				T *p2 = pData;
				for (i = 0; i < nSize; i++)
					*p1++ = *p2++;

				delete[] pData;
			}// if

			nSize = n;
			pData = tmpData;
		} 
		else if ((n < nSize - nGrowBy) && (nGrowBy < nSize))
		{
			nSize = n;
			nRealSize = nSize+nGrowBy;
				
			T *tmpData = new T[nRealSize];

			if (pData != NULL)
			{
				T *p1 = tmpData;
				T *p2 = pData;
				for (i = 0; i < n; i++)
					*p1++ = *p2++;
					
				delete[] pData;
			}// if
				
			pData = tmpData;
		}
		else // if (n < nSize)
		{
			nSize = n;
		} 
	}// if
}

//
// set the growby size new
//
template <class T> 
void 
TDynArray<T>::setGrowBy ( uint growby ) 
{
	nGrowBy = growby;
}

//
// access the vector as an array
//
template <class T> 
T& 
TDynArray<T>::operator [] ( uint i ) 
{ 
	if (i >= nSize)
		cerr << "TDynArray::operator[] (" << i << ") : out of bound "
			 << "(max=" << nSize << ")" << endl;

	return *(pData+i); 
}

//
// access the vector as an array (const access)
//
template <class T> 
const T& 
TDynArray<T>::operator [] ( uint i ) const 
{ 
	if (i >= nSize)
		cerr << "const TDynArray::operator[] (" << i << ") const : out of bound "
			 << "(max=" << nSize << ")" << endl;

	return *(pData+i); 
}

//
// copy operator
//
template <class T> 
TDynArray<T>& 
TDynArray<T>::operator = ( const TDynArray<T>& vec ) 
{
	uint  i;

	if (vec.nSize == nSize)
	{
		nGrowBy = vec.nGrowBy;

		T *p1 = pData;
		T *p2 = vec.pData;

		for (i = 0; i < nSize; i++)
			*p1++ = *p2++;
	}// if
	else 
	{
		nGrowBy   = vec.nGrowBy;
		nSize     = vec.nSize;
		nRealSize = vec.nRealSize;

		// set new size but no need for keeping old elements
		if (pData != NULL) delete[] pData;
		pData = new T[nRealSize];

		if (nSize > 0)
		{
			T *p1 = pData;
			T *p2 = vec.pData;
			for (i = 0; i < nSize; i++)
				*p1++ = *p2++;
		}// if
	}// else
	return (*this);
}

#endif			// __DYNARRAY_HH
