//
//  Project     : libgraph
//  File        : edge.hh
//  Author      : Ronald Kriemann
//  Purpose     : classes for different edges in a graph
//  Date        : 04.08.1997
//
// arch-tag: a21ae8ed-377a-40b7-b9bc-466cace83570
//

#ifndef __EDGE_HH
#define __EDGE_HH

#include "memory.hh"
#include "multiref.hh"

// forward declaration for a node
class TNode;

///////////////////////////////////////////
//
// base class for all edges
//
class TEdge : public TMemory, public TMultiReference
{
protected:
    // pointer to the other endnode
    TNode  *pHead, *pTail;

    // name of an edge (for identification)
    uint   name;

public:
    // constructor
    TEdge ( TNode& tail, TNode& head, uint n = 0 );

    // destructor
    virtual ~TEdge ();

    // access the name
    const uint getName () const;

    // access the node pointers
    TNode* getNode ();
    TNode* getTail ();
    TNode* getHead ();
    TNode* getOther ( TNode * pNode );

    // create new edge (and set specific info)
    virtual TEdge* create ( TNode& tail, TNode& head, uint n = 0 );
    virtual TEdge* clone ( TNode& tail, TNode& head );
};

///////////////////////////////////////////
//
// class for a markable edge
//
class TMarkableEdge : public TEdge
{
protected:
    // the mark
    int  _mark;

public:
    // constructor
    TMarkableEdge ( TNode& tail, TNode& head, 
                    int m = 0, uint n = 0 );

    // access the mark
    const int mark () const;
    void setMark ( int m );

    // create new edge (and set specific info)
    virtual TMarkableEdge* create ( TNode& tail, TNode& head, uint n = 0 );
    virtual TMarkableEdge* clone ( TNode& tail, TNode& head );
};

///////////////////////////////////////////
//
// class for a weighted,markable edge
//
class TWeightedMarkableEdge : public TMarkableEdge
{
protected:
    // the weight
    TFloat  _weight;

public:
    // constructor
    TWeightedMarkableEdge ( TNode& tail, TNode& head, 
                            int m = 0, TFloat w = 0, uint n = 0 );

    // access the weight
    const TFloat weight () const;
    void setWeight ( TFloat w );

    // create new edge (and set specific info)
    virtual TWeightedMarkableEdge* create ( TNode& tail, TNode& head,
                                            uint n = 0 ); 
    virtual TWeightedMarkableEdge* clone ( TNode& tail, TNode& head );
};

#endif  // __EDGE_HH
