#ifndef __FRONTIER_HH
#define __FRONTIER_HH
//
//  Project   : libnet
//  File      : frontier.hh
//  Author    : Sabine Gutsch / Ronald Kriemann
//  Purpose   : class for a frontier (separates two cells)
//
// arch-tag: 2ce813c3-5581-4492-99b2-7fcd714a0d74
//

#include <string.h>

#include "memory.hh"
#include "multiref.hh"
#include "dynarray.hh"

#include "TBoundary.hh"
#include "vertex.hh"

// forward declaration
template <int N> class TCell;

// some defines for the type
#define BOUNDARY  1
#define NEUMANN   2

//
// This class has two pointers, one for each cell.
// A frontier can be divided into two other frontiers.
//
template <int N>
class TFrontier : public TMemory, public TMultiReference 
{
protected:
    // pointer to parentfrontier
    // TFrontier<N>    *pParent;

    // pointers to the cells 
    TCell<N>        *pCells[2];

    // pointers to childfrontiers
    TFrontier<N>   **arrChildren;
    uint            noOfChildren;

    // size of the frontier (in 2d this would be the length)
    TFloat           fSize;

    // pointer to desc of element on boundary (NULL if interior)
    TBoundary       *pBoundary;
    
public:
    // constructors
    TFrontier ( TFrontier<N>* parent );
    TFrontier ( TFrontier<N>* parent, TBoundary *boundary );
    
    // destructor
    virtual ~TFrontier ();

    // init memory for used arrays
    virtual void initChildPtr ( uint n );

    // return the parentfrontier
    // virtual TFrontier<N>* getParent () const;

    // return i'th cell
    virtual TCell<N>* getCell ( uint i ) const;

    // return the pointer to the other cell
    virtual TCell<N>* getOtherCell ( TCell<N>* cell );

    // remove cellpointer to the given cell
    virtual void clearCell ( TCell<N> *cell );

    // set the cellpointer(s) new
    virtual void setCell ( TCell<N> *cell );

    // return the size of the frontier
    virtual const TFloat size () const;

    // make the midnode of the frontier
    virtual void buildMidNode () = 0;

    // did we set the midpoint
    virtual const bool hasMidNode () const = 0;

    // return number of nodes
    virtual const uint getNoOfNodes () const = 0;

    // return i'th node
    virtual TVertex<N>* getNode ( uint i ) const = 0;

    // return i'th node with respect to the view of the given cell :
    //   for cell[0] the nodes are ccw
    //   for cell[1] the nodes are cw
    virtual TVertex<N>* getNode ( const TCell<N>* cell, uint i ) const = 0;
    
    // return number of children
    virtual const uint getNoOfChildren () const;

    // build the child(ren)
    virtual void buildChild () = 0;

    // do we habe children
    virtual const bool hasChild () const;

    // return i'th child
    virtual TFrontier<N>* getChild ( uint i ) const;

    // return i'th child with respect to the view of the given cell
    // cell[0] has the frontiers numbered from ccw
    // cell[1] has the frontiers numbered from cw
    virtual TFrontier<N>* getChild ( const TCell<N>* cell, uint i ) const;

    // get/set boundary
    TBoundary * getBoundary () const;
    void setBoundary ( TBoundary * boundary );
};

///////////////////////////////////////////////////////////////////
//
// implementation
//

//
// constructors
//
template <int N>
TFrontier<N>::TFrontier ( TFrontier<N>* parent ) 
{
    // pParent   = parent;
    pCells[0] = pCells[1] = NULL;

    arrChildren  = NULL;
    noOfChildren = 0;

    fSize = 0;
}
    
template <int N>
TFrontier<N>::TFrontier ( TFrontier<N>* parent, TBoundary *boundary )
{
    // pParent   = parent;
    pCells[0] = pCells[1] = NULL;

    arrChildren  = NULL;
    noOfChildren = 0;

    fSize = 0;

    if (parent == NULL) 
        pBoundary = boundary;
    else 
        pBoundary = parent->getBoundary();
}

//
// destructor
//
template <int N>
TFrontier<N>::~TFrontier () 
{
    uint          i;
    TFrontier<N>  *pChild;

    if (arrChildren != NULL)
    {
        for (i = 0; i < noOfChildren; i++)
        {
            pChild = arrChildren[i];
            
            if ((pChild != NULL) && (pChild != this) &&
                (pChild->delReference())) delete pChild;
        }// if

        delete arrChildren;
    }// if
}

//
// init memory for used arrays
//
template <int N>
void 
TFrontier<N>::initChildPtr ( uint n )
{
    uint  i;

    if (n == noOfChildren) 
        return;

    if (arrChildren != NULL) 
        delete arrChildren;

    if (n == 0)
    {
        arrChildren  = NULL;
        noOfChildren = 0;
    }// if
    else
    {
        arrChildren  = new TFrontier<N>* [n];
        noOfChildren = n;

        for (i = 0; i < noOfChildren; i++)
            arrChildren[i] = NULL;
    }// else
}

//
// return the parentfrontier
//
// template <int N>
// TFrontier<N>* 
// TFrontier<N>::getParent () const 
// { 
//     return pParent; 
// }

//
// return i'th cell
//
template <int N>
TCell<N>* 
TFrontier<N>::getCell ( uint i ) const 
{ 
    if (i < 2)
        return pCells[i]; 
    else
    {
        cerr << "TFrontier::getCell (" << i << ") : index too high" << endl;
        return NULL;
    }// else
}

//
// return the pointer to the other cell
//
template <int N>
TCell<N>* 
TFrontier<N>::getOtherCell ( TCell<N>* cell )
{
    if (pCells[0] == cell)
        return pCells[1];
    else if (pCells[1] == cell)
        return pCells[0];
    else
    {
        cout << "TFrontier::getOtherCell (TCell*) : no such cell" << endl;
        return NULL;
    }// else
}

//
// remove cellpointer to the given cell
//
template <int N>
void 
TFrontier<N>::clearCell ( TCell<N> *cell ) 
{ 
    if (pCells[0] == cell)
        pCells[0] = NULL;
    else if (pCells[1] == cell)
        pCells[1] = NULL;
    else
        cout << "TFrontier::clearCell (TCell*) : no such cell" << endl;
}

//
// set the cellpointer(s) new
//
template <int N>
void 
TFrontier<N>::setCell ( TCell<N> *cell ) 
{
    if (pCells[0] == NULL)
        pCells[0] = cell;
    else if (pCells[1] == NULL)
        pCells[1] = cell;
    else
    {
        cout << "TFrontier::setCell (TCell*) : all cells are defined" << endl;
        pCells[1] = cell;
    }// else
}

//
// return the size of the frontier
//
template <int N>
const TFloat 
TFrontier<N>::size () const 
{ 
    return fSize; 
}

//
// return number of children
//
template <int N>
const uint 
TFrontier<N>::getNoOfChildren () const
{
    return noOfChildren;
}

//
// do we habe children
//
template <int N>
const bool 
TFrontier<N>::hasChild () const 
{ 
    if (noOfChildren == 0)
        return FALSE;
    else
        return TRUE;
}

//
// return i'th child<N>
//
template <int N>
TFrontier<N>* 
TFrontier<N>::getChild ( uint i ) const 
{ 
    if (i < noOfChildren) 
        return arrChildren[i];
    else 
    {
        cout << "TFrontier::getChild(uint) : index out of range" << endl;
        return NULL;
    }// else
}

//
// return i'th child with respect to the view of the given cell
// cell[0] has the frontiers numbered from ccw
// cell[1] has the frontiers numbered from cw
//
template <int N>
TFrontier<N>* 
TFrontier<N>::getChild ( const TCell<N>* cell, uint i ) const 
{
    if (i < noOfChildren)
    {
        if (hasChild())
        {
            if (hasMidNode())
            {
                if (cell == pCells[0])
                    return arrChildren[i];
                else if (cell == pCells[1])
                    return arrChildren[noOfChildren-1-i];
                else
                {
                    cout << "TFrontier::getNode (TCell," << i << ") : no such cell" << endl;
                    return NULL;
                }// else
            }// if
            else
                return arrChildren[0];
        }// if
        else
        {
            cout << "TFrontier::getChild(TCell," << i <<") : no children created" << endl;
            return NULL;
        }// else
    }// if
    else
    {
        cout << "TFrontier::getChild(TCell," << i << ") : index out of range" << endl;
        return NULL;
    }// else
}

//
// return boundary
//
template <int N>
TBoundary *
TFrontier<N>::getBoundary () const
{
    return pBoundary;
}

template <int N>
void
TFrontier<N>::setBoundary ( TBoundary * boundary )
{
    pBoundary = boundary;
}

#endif  // __FRONTIER_HH
