//
//  Project     : libgraph
//  File        : graphtool.hh
//  Author      : Ronald Kriemann
//  Purpose     : class for some common graph tools
//  Date        : 20.07.1997
//
// arch-tag: 41db3914-0233-452e-92d3-b32dda7e08cf
//

#ifndef __GRAPHTOOL_HH
#define __GRAPHTOOL_HH

#include "memory.hh"
#include "list.hh"
#include "dynarray.hh"

#include "nodes.hh"
#include "basegraph.hh"

// different edge-marks in DFS
#define DAG_EDGE       1   // (_D_irected _A_cyclic _G_raph)
#define BACKWARD_EDGE  2

// different marks for a node
#define C        1
#define F        2
#define L        3

// used marks
#define W        4
#define R        5
#define B        6
#define bB       7
#define wW       8

#define NOT_VISITED  7
#define ADDED        8

// different searchtypes
#define SEARCH_DFS   1
#define SEARCH_BFS   2

// forward decl
template <class T> class TVectorBlock;
class TBlockInfo;

//
// apply misc. algorithm to graphs
//
class TGraphTool : public TMemory
{
public:
	// constructor
	TGraphTool () {}

	// numbering algorithm
	void numberNodes ( TGraph& graph, TBlockInfo *blockinfo = NULL,
					   bool bInit = TRUE );

	// similar algorithm but with BFS-type of search
	void numberBFS( TGraph &graph, TBlockInfo *blockinfo = NULL,
					TVectorBlock<uint> *pvecPerm = NULL,
					bool bInit = TRUE );
	
	// decompose a given graph into the strong connected
	// components (based on Tarjan [72], SIAM, Vol 1 No 2)
    // recursiv !!! use scc(...) if you need a non-recursiv algo
	TList<TGraph>* tarjan ( TGraph& graph );

	// used by tarjan to detect the components
	void strongConnect ( TList<TGraph>& listGraph,
						 TList<TNode>& stack,
						 TGraph& graph, TNode& node,
                         uint& index );
	void strongConnect_nr ( TList<TGraph>& listGraph,
                            TList<TNode>& stack,
                            TGraph& graph, TNode& node,
                            uint& index );

	// do a DFS with the graph, mark each vertex and build a palm
	// Assumption : there ex. a path from root to every node in the graph
	TGraph* dfs ( TGraph& graph, TMarkableNode& root );

	// used by dfs to recursivly visit every node
	void visit ( TGraph& palm,
				 TMarkableNode& node, TMarkableNode& pred, 
				 int& dfsNumber );

    ///////////////////////////////////////////////////////////////////
    //
    // strongly connected components and depth first search from LEDA
    // (non-recursiv)
    //

protected:
    void dfs ( TNode            & node,
               int              & count1,
               int              & count2 );

    void dfs_num ( TGraph  & graph );

    void dfs_reach_rev ( TNode           & node,
                         TList< TNode >  & l );

public:
    TList< TGraph > *  scc ( TGraph & graph );
    
protected:
	// for number ()
	void setF ( TMarkableNode& node, TBlockInfo *pBlockInfo );
	void setL ( TMarkableNode& node, TBlockInfo *pBlockInfo );
    
	// for numberBFS ()
	void setF_BFS ( TMarkableNode& node, TList<TMarkableNode> &list,
					TBlockInfo *pBlockInfo );
	void setL_BFS ( TMarkableNode& node, TList<TMarkableNode> &list,
					TBlockInfo *pBlockInfo );

    // order edges after tarjan
    void sortInEdges ( TNode * pNode );
    void sortOutEdges ( TNode * pNode );
};

#endif  // __GRAPHTOOL_HH
