#ifndef __NET_HH
#define __NET_HH
//
//  Project   : libnet
//  File      : net.hh
//  Author    : Sabine Gutsch / Ronald Kriemann
//  Purpose   : a net is a list of cells with some methods
//
// arch-tag: 4422e94e-db55-4cd1-b9cc-ab0840b639ce
//

#include "multiref.hh"
#include "memory.hh"
#include "list.hh"

#include "parser.hh"
#include "cell.hh"

//
// the net class
//
template <int N>
class TNet : public TMemory, public TMultiReference 
{
public:
    // some type shortcuts
    typedef TList< TCell<N> >         TCellList;
    typedef TListIterator< TCell<N> > TCellListIterator;

protected:
    // the cells
    TCellList  *plistCells;
    uint       noOfInteriorNodes;
    uint       noOfBoundaryNodes;

public:
    // constructors
    TNet ();
    // constructor which reads cells from a file
    TNet ( const char *filename );

    // destructor
    ~TNet ();

    // adds a cell to the list
    void add ( TCell<N> *cell );

    // return the list to the cells
    const TCellList& getList () const;

    // refine all cells
    TNet<N>* refine ();

    // get number of nodes
    const uint getNoOfInteriorNodes () const;
    const uint getNoOfBoundaryNodes () const;
};

////////////////////////////////////////////////////////////
//
// implementation
//

//
// constructors
//
template <int N>
TNet<N>::TNet () 
{
    plistCells = new TCellList;

    noOfInteriorNodes = 0;
    noOfBoundaryNodes = 0;
}

//
// constructor which reads cells from a file
//
template <int N>
TNet<N>::TNet ( const char *filename ) 
{
    TParser parser( filename );

    if (parser.buildList().getDim() != N)
        cerr << "TNet::TNet (char*) : data has not correct dimension" << endl;
    
    plistCells = (TCellList*) parser.getCellList( N );

    if (plistCells == NULL)
    {
        cerr << "TNet::TNet (char*) : error while reading triangulation from file "
             << filename << endl;
    }// if
    else
    {
        // add a reference to all cells
        TCellListIterator  iter(*plistCells);
        TCell<N>          *cell;

        iter.first();
        while ((cell = iter++) != NULL)
        {
            cell->addReference();
        }// while

        noOfInteriorNodes = iter.first()->getNode(0)->getMaxInteriorName();
        noOfBoundaryNodes = iter.first()->getNode(0)->getMinBoundaryName();
    }// else
}

//
// destructor
//
template <int N>
TNet<N>::~TNet () 
{
    TCellListIterator  iter(*plistCells);
    TCell<N>          *cell;

    iter.first();
    while ((cell = iter++) != NULL)
    {
        if (cell->delReference()) delete cell;
    }// while

    delete plistCells;
}

//
// adds a cell to the list
//
template <int N>
void 
TNet<N>::add ( TCell<N> *cell )
{ 
    if (cell != NULL)
    {
        cell->addReference();
        plistCells->add( *cell ); 

        noOfInteriorNodes = cell->getNode(0)->getMaxInteriorName();
        noOfBoundaryNodes = cell->getNode(0)->getMinBoundaryName();
    }// if
    else
        cout << "TNet::add(TCell*) : no cell was given" << endl;
}

//
// return the list to the cells
//
template <int N>
const TList<TCell<N> > &
TNet<N>::getList () const
{ 
    return *plistCells;
}

//
// refine all cells
//
template <int N>
TNet<N>* 
TNet<N>::refine () 
{
    int                 i;
    TNet<N>            *pnetNew;
    TCellListIterator   iter(*plistCells);
    TCell<N>           *cell;
    int                 noOfChildren;
        
    // first prepare cell for refinement
    iter.first();
    while ((cell = iter++) != NULL)
    {
        cell->prepareRefine();
    }// while

    // and then refine
    iter.first();
    while ((cell = iter++) != NULL)
    {
        cell->refine();
    }// while

    // make new net
    pnetNew = new TNet<N>;
    pnetNew->delReference();

    // and insert all children from the old net into the new net
    iter.first();
    while ((cell = iter++) != NULL)
    {
        noOfChildren = cell->getNoOfChildren();

        for (i = 0; i < noOfChildren; i++)
        {
            if (cell->getChild(i) != NULL)
                pnetNew->add(cell->getChild(i));
        }// for
    }// while

    return pnetNew;
}

//
// get number of nodes
//
template <int N>
const uint
TNet<N>::getNoOfInteriorNodes () const
{ 
    return noOfInteriorNodes;
}

template <int N>
const uint 
TNet<N>::getNoOfBoundaryNodes () const
{ 
    return noOfBoundaryNodes;
}

//
// some typedefs
//
typedef TNet<2>  T2DNet;
typedef TNet<3>  T3DNet;

#endif  // __NET_HH
