#ifndef __NETHIERARCHY_HH
#define __NETHIERARCHY_HH
//
//  Project   : libnet
//  File      : nethierarchy.hh
//  Author    : Sabine Gutsch / Ronald Kriemann
//  Purpose   : class for a set of nets
//
// arch-tag: cab4b14f-df6a-4985-883f-ef1323799039
//

#include "memory.hh"
#include "multiref.hh"
#include "list.hh"

#include "net.hh"

//
// this class consists of a list of nets, where we can build a new
// net through refining the last net in the list, or simply add an 
// extern net to the list
//
template <int N>
class TNetHierarchy : public TMemory, public TMultiReference {
public:
    // some type shortcuts
    typedef TList<TNet<N> >          TNetList;
    typedef TListIterator<TNet<N> >  TNetListIterator;

protected:
    // list for the nets
    TNetList  *plistNets;

public:
    // constructor
    TNetHierarchy ();

    // constructor which reads a init. net from a file
    TNetHierarchy ( const TString & filename );

    // destructor
    ~TNetHierarchy ();

    // return the list to the nets
    TList<TNet<N> >& getList () const;

    // add a new net the list
    void add ( TNet<N> *net );

    // build a new net from the last in the list
    TNet<N> * buildNet ();

    // return number of nets
    const int getNoOfNets () const;
};

////////////////////////////////////////////////////////////
//
// implementation
//

//
// constructor
//
template <int N>
TNetHierarchy<N>::TNetHierarchy () 
{
    plistNets = new TNetList;
}

//
// constructor which reads a init. net from a file
//
template <int N>
TNetHierarchy<N>::TNetHierarchy ( const TString & filename ) 
{
    char  buffer[256];

    filename.copyTo( buffer, sizeof( buffer ) );
    
    TNet<N> *pNet = new TNet<N>( buffer );
    plistNets     = new TNetList;

    plistNets->add( *pNet );
}

//
// destructor
//
template <int N>
TNetHierarchy<N>::~TNetHierarchy () 
{
    TNetListIterator  iter(*plistNets);
    TNet<N>          *net;

    iter.first();
    while ((net = iter++) != NULL)
    {
        if (net->delReference()) delete net;
    }// while

    delete plistNets;
}

//
// return the list to the nets
//
template <int N>
TList<TNet<N> > & 
TNetHierarchy<N>::getList () const 
{ 
    return *plistNets; 
}

//
// add a new net the list
//
template <int N>
void 
TNetHierarchy<N>::add( TNet<N> *net ) 
{ 
    if (net != NULL)
    {
        net->addReference();
        plistNets->add( *net ); 
    }// if
    else
        cout << "TNetHierarchy::add (TNet*) : no net was given" << endl;
}

//
// build a new net from the last in the list
//
template <int N>
TNet<N> * 
TNetHierarchy<N>::buildNet () 
{
    TNetListIterator  iter(*plistNets);
    TNet<N>          *plastNet = iter.last();
    TNet<N>          *pNew = NULL;

    if (plastNet != NULL) 
    {
        // now refine the last one and insert the created net in the list
        pNew = plastNet->refine();
        add( pNew );
    }// if
    else
        cout << "TNetHierarchy::buildNet () : no net in list" << endl;

    return pNew;
}

//
// return number of nets
//
template <int N>
const int 
TNetHierarchy<N>::getNoOfNets() const 
{ 
    return plistNets->getSize();
}

//
// some typdefs
//
typedef TNetHierarchy<2> T2DNetHierarchy;
typedef TNetHierarchy<3> T3DNetHierarchy;

#endif  // __NETHIERARCHY_HH
