//
//  Project   : libnet
//  File      : oneStepScheme.hh
//  Author    : Sabine Gutsch
//  Purpose   : one step theta time discretization scheme
//
// arch-tag: 1f7fd5c7-4272-4e32-a0c6-e2ba9a83af35
//

#ifndef __ONESTEPSCHEME_HH
#define __ONESTEPSCHEME_HH

#include "nonlinearSolver.hh"

class TOneStepScheme : public TMemory {

protected:
    TFloat               theta;         // one step theta scheme
    TFloat               h;             // grid parameter
    TFloat               steplength;    // time step length
    uint                noOfSteps;     // number of time steps
    TBlockVector<TFloat> *pX;           // approximate solution
    TNonLinearSolver     *nonlinearSolver;     
    

public:
	// constructor
    TOneStepScheme ( TBlockVector<TFloat> *paX,
                     TNonLinearSolver *aNonlinearSolver,
                     TFloat aH,
                     TFloat aTheta = 1.0 ) 
    {
        pX              = paX;
        h               = aH;
        nonlinearSolver = aNonlinearSolver;
        theta           = aTheta;
        steplength      = 0.25;
        noOfSteps       = 40;
	}

	// destructor
    ~TOneStepScheme ( ) {};

    void step ( TMultigrid *mgSolver ) 
    {
        if ( theta < 1e-8 ) 
		{  // explicit Euler step
			// Poisson problem for pressure!!! (pressure implicit)
            //cout << "Explicit Euler\n";
			//cout << "steplength: " << MIN( 50.0/512.0, 0.0625 / MAX(0.0001, pX->maxNorm(0)) ) << "   " << pX->maxNorm(0) << endl;
            //steplength = MIN( 500.0/512.0, 0.0625 / MAX(0.0001, pX->maxNorm(0)) );
			// adaptive steplength: see Griebel p. 40
            *(mgSolver->getX()) = *pX;
            nonlinearSolver->updateHierarchy( mgSolver );
		    *pX += steplength * ( *(mgSolver->getRHS()) - *(mgSolver->getA()) * *pX);
		}
        if ( theta > 0.999 )
		{ // implicit Euler step
            cout << "Implicit Euler\n";
            *(mgSolver->getX()) = *pX;
            (*pX)[2].set( 0.0 );
            *(mgSolver->getRHS()) +=  ((h*h/steplength) * *pX);
            mgSolver->addIdentity( h*h/steplength );
           
            nonlinearSolver->updateHierarchy( mgSolver );
            mgSolver->solve();
            //nonlinearSolver->solve( mgSolver );

            *(mgSolver->getRHS()) -= ((h*h/steplength) * *pX);
            mgSolver->subtractIdentity( h*h/steplength );           

            *pX = *(mgSolver->getX());         
		}

    }

    void solve ( TMultigrid *mgSolver ) 
	{
        uint i = 0; 
        //nonlinearSolver->solve( mgSolver );
        //*pX = *(mgSolver->getX());
        while ( (i++ < noOfSteps)  )  step ( mgSolver );
	}
};



#endif  // __ONESTEPSCHEME_HH
