//
//  Project   : libnet
//  File      : solver.hh
//  Author    : Sabine Gutsch / Thomas Probst
//  Purpose   : virtual linear solver class
//
// arch-tag: 23088cb3-a1d0-4a94-916e-52be84cbde78
//

#ifndef __SOLVER_HH
#define __SOLVER_HH

#include "memory.hh"
#include "TString.hh"

#include "blockmatrix.hh"

#define DEFAULT_MAX_ITERATION 30
#define DEFAULT_ACCURACY 1e-8
#define DEFAULT_REDUCEBY 1e-4

//verbosity of created messages: 
// The verbosity is bitfield-encoded
#define PRINT_NOMESSAGE           0  // no messages  
#define PRINT_SOLVERNAME          1  // print name of solver
#define PRINT_CONVERGENCE_HISTORY 2  // print rate of defect reduction for each step
#define PRINT_CONVERGENCE_RATE    4  // print convergence rate after last step
#define PRINT_BREAKDOWNS          8  // print breakdown messages

// return values for iterations :
#define RETURN_CONTINUE   0  // solving step was succesfull, continue with next step 
#define RETURN_BREAKDOWN  1  // a breakdown has occured; stop solver

// mode for iteration
#define MODE_ACC_MAXIT 0  // stop if maxit or accuracy is reached
#define MODE_REDUCEBY  1  // stop if maxit is reached or accuracy is reduced by DEFAULT_REDUCEBY
#define MODE_MAXIT     2  // stop if maxit is reached

//
// basis class for a solver for a (linear or nonlinear) 
// system of equations
//
class TSolver : public TMemory
{

protected:
    TString                name;
   
    TBlockMatrix         * pA;          // hierarchy of matrices
    TBlockVector<TFloat> * pRHS;        // hierarchy of right hand sides
    TBlockVector<TFloat> * pX;          // hierarchy of vectors for approximate solution
    uint                   maxit;       // maximal number of iteration steps
    TFloat                 accuracy;    // desired accuracy or desired reduction (depends on mode)
    uint                   verbosity;   // verbosity of created messages
    uint                   mode;        // mode of solver
  
	// array with all convergence rates
	TDynArray<TFloat>    * parrConvRates;

public:
    //
	// constructor and destructor
    //
    TSolver ( TBlockMatrix         * apA    = NULL, 
              TBlockVector<TFloat> * apRHS  = NULL, 
              TBlockVector<TFloat> * apX    = NULL, 
              uint                   amaxit = DEFAULT_MAX_ITERATION, 
              TFloat                 aaccur = DEFAULT_ACCURACY,
              uint                   averb  = 0,
              uint                   amode  = 0 );

    virtual ~TSolver ();

	// get (set) name
    virtual void setName ( const TString & name );
    virtual const TString & getName ();

	// set and get verbosity 
    virtual void setVerbosity (const uint averb  );
    virtual uint getVerbosity () const;
	
	// set and get accuracy 
    virtual void setAccuracy (const TFloat aaccur );
    virtual TFloat getAccuracy () const;
	
	// set and get maxit
    virtual void setMaxit ( const uint aMaxit );
    virtual uint getMaxit () const;

	// set and get mode
    virtual void setMode ( const uint aMode ) { mode = aMode; }
    virtual uint getMode () { return (mode); }
 
	// a single iteration step
    virtual int step () = 0;
    virtual int step ( int                       level,
                       TBlockMatrix           ** A,
                       TBlockVector< TFloat >  * RHS,
                       TBlockVector< TFloat >  * x );

	// solving method
    virtual void solve ();

	// first set rhs and then solve with zero initial vector
    virtual void solve ( TBlockMatrix *newA, TBlockVector<TFloat> *newRHS, 
                         TBlockVector<TFloat> *newX );

    // multigrid solve
    virtual void solve ( int                     level,
                         TBlockMatrix         ** A,
                         TBlockVector<TFloat> ** RHS,
                         TBlockVector<TFloat> ** x );

	// consistency check for dimensions of pA and pRHS
    virtual void checkpApRHS ();

	// initialize solver
    virtual void init ();

	// set and get pA
    virtual void setA ( TBlockMatrix *newA );
    virtual TBlockMatrix * getA () const;

	// set and get RHS
    virtual void setRHS ( TBlockVector<TFloat>  *newRHS );
    virtual TBlockVector<TFloat> * getRHS () const;

	// set and get x
    virtual void setX (  TBlockVector<TFloat> *newx );
    virtual TBlockVector<TFloat> * getX () const;

	// return array with conv. rates
	virtual TDynArray<TFloat> * getConvRates ();

    // print statistics
    void printStep ( uint step, TFloat norm_old, TFloat norm );
    void printResult ( uint steps, TFloat norm_0, TFloat norm );

    // test stop-criteria
    bool testStop ( uint it, TFloat norm_0, TFloat norm );
};


#endif  // __SOLVER_HH
