//
//  Project  : libnet
//  File     : blockvector.hh
//  Author   : Sabine Gutsch / Ronald Kriemann / Thomas Probst
//  Purpose  : class for a blockvector
//
// arch-tag: 1952bfd5-c336-43ce-958c-2b800b9c9ae8
//

#include <math.h>

#include "matrixblock.hh"
#include "blockmatrix.hh"
#include "TPermutation.hh"
#include "sparsevector.hh"

#include "blockvector.hh"

///////////////////////////////////////////////////////////
//
// class TVectorBlock (template in storagetype)
// - a single block in a BlockVector
//

//
// constructor
//
template <class T>
TVectorBlock<T>::TVectorBlock ()
{
    precision = 4;
}

template <class T>
TVectorBlock<T>::TVectorBlock ( const TVectorBlock<T>& v ) 
{
    (*this)   = v;
    precision = 4;
}

template <class T>
TVectorBlock<T>::TVectorBlock ( const T & f, uint n ) 
{ 
    TDynArray<T>::setSize( n );
    for ( uint i = 0; i < n; i++ ) 
        (*this)[i] = f;
    precision = 4;
}

//
// destructor
//
template <class T>
TVectorBlock<T>::~TVectorBlock ()
{}

//
// set vectorBlock
//
template <class T>
void
TVectorBlock<T>::setVectorBlock ( const TVectorBlock<T>& v )
{
    (*this) = v;
}
	
//
// set all vector components to scalar
//
template <class T>
void
TVectorBlock<T>::set ( const T& scalar ) 
{
    uint  i;
    uint  size = TDynArray<T>::getSize();
    
    for ( i = 0; i < size; i++ )
        (*this)[i] = scalar;
}

template <class T>
void
TVectorBlock<T>::setPrecision (const uint prec )
{
    precision = prec;
}

template <class T>
uint
TVectorBlock<T>::getPrecision () const
{
    return precision;
}

//
// copy operator
//
template <class T>
TVectorBlock<T>&
TVectorBlock<T>::operator= ( const T & scalar )
{
    set( scalar );

    return *this;
}

//
// init vec with same structure, but no data
//
template <class T>
void
TVectorBlock< T >::clone ( TVectorBlock< T > & vec ) const
{
    vec.setSize( TDynArray<T>::getSize() );
}

//
// vectorblock addition ( we assume, that the vectorblocks
// have same structure )
//
template <class T>
TVectorBlock<T>&
TVectorBlock<T>::operator+= ( const TVectorBlock<T>& vec ) 
{
    uint j, blockSize;
		
    blockSize = TDynArray<T>::getSize();
    
    for (j = 0; j < blockSize; j++)
        (*this)[j] += vec[j];
		
    return *this;
}

template <class T>
TVectorBlock<T>& 
TVectorBlock<T>::operator+= ( const TMatrixData& data )
{
    uint   noOfEntries = data.getNoOfEntries();
    uint  index;
   
    for (uint i = 0; i < noOfEntries; i++)
    {
        index = data.getIndex(i);
        (*this)[index] += T(data.getValue( i ));
    }// for
   
    return (*this);
}

template <class T>
TVectorBlock<T>
TVectorBlock<T>::operator+ ( const TVectorBlock<T>& vec ) const 
{
    TVectorBlock<T> vecTmp( *this );
    
    return vecTmp += vec;
}
	
//
// vectorblock subtraction ( we assume, that the vectorblocks
// have same structure )
//
template <class T>
TVectorBlock<T>&
TVectorBlock<T>::operator-= ( const TVectorBlock<T>& vec ) 
{
    uint  j;
    uint  blockSize = TDynArray<T>::getSize();
    
    for ( j = 0; j < blockSize; j++ )
        (*this)[j] -= vec[j];
		
    return *this;
}

template <class T>
TVectorBlock<T>& 
TVectorBlock<T>::operator-= ( const TMatrixData& data )
{
    uint   noOfEntries = data.getNoOfEntries();
    uint  index;
   
    for (uint i = 0; i < noOfEntries; i++)
    {
        index = data.getIndex(i);
        (*this)[index] -= T(data.getValue(i));
    }// for
   
    return (*this);
}

template <class T>
TVectorBlock<T>
TVectorBlock<T>::operator- ( const TVectorBlock<T>& vec ) const 
{
    TVectorBlock<T> vecTmp( *this );
    
    return vecTmp -= vec;
}
	
//
// scalar mult.
//
template <class T>
TVectorBlock<T>&
TVectorBlock<T>::operator*= ( const T& scalar ) 
{
    uint  j;
    uint  blockSize = TDynArray<T>::getSize();
    
    for ( j = 0; j < blockSize; j++ )
        (*this)[j] *= scalar;
		
    return *this;
}

//
// inner product (we assume that the vectorblocks have same structure)
//
template <class T>
T
TVectorBlock<T>::operator* ( const TVectorBlock<T>& vec ) const
{
    uint  j, blockSize;
    T      product = T(0);

    blockSize = TDynArray<T>::getSize();
    
    for (j = 0; j < blockSize; j++)
        product += (*this)[j] * vec[j];

    return (product);
}

//
// some blas routines
//

// x = x + alpha y
template <class T>
const TVectorBlock< T > &
TVectorBlock< T >::daxpy ( const TVectorBlock< T > & y, const T & alpha )
{
    int  i, size;

    size = TDynArray<T>::getSize();

    for ( i = 0; i < size; i++ )
        (*this)[i] += alpha * y[i];

    return *this;
}
    
// x = x + alpha A y
const TVectorBlock< TFloat > &
TVectorBlock< TFloat >::matmulplus ( const TMatrixBlock           & A,
                                     const TVectorBlock< TFloat > & y,
                                     const TFloat                 & alpha )
{
    TVectorBlock< TFloat >  tmp;

    tmp = A * y;
    this->daxpy( tmp, alpha );

    return *this;
}
    
//
// permutation
//
template <class T>
void
TVectorBlock<T>::permute ( const TVectorBlock<uint> *p ) 
{
    uint           j;
    uint           blockSize = TDynArray<T>::getSize();
    TVectorBlock<T> vecTmp( *this );
    
    for ( j = 0; j < blockSize; j++ )
        (*this)[(*p)[j]] = vecTmp[j];
}

template <class T>
void
TVectorBlock<T>::permute ( const TPermutation * pPerm ) 
{
    if (pPerm == NULL)
    {
        cout << "TVectorBlock::permute (...) : got NULL arg" << endl;
        return;
    }// if

    uint            j;
    uint            blockSize = TDynArray<T>::getSize();
    TVectorBlock<T>  vecTmp( *this );
    
    for ( j = 0; j < blockSize; j++ )
        (*this)[ (*pPerm)[j] ] = vecTmp[j];
}

///////////////////////////////////////////////////////////
//
// class TBlockVector (template in storagetype)
//

//
// standard constructors
//
template <class T>
TBlockVector<T>::TBlockVector () 
{ 
}

//
// create only the blockstructure
//
template <class T>
TBlockVector<T>::TBlockVector ( uint number, uint size ) 
{
    setSize( number );

    if ( size > 0 )
    {
        // and then resize block to the given size
        (*this)[0].setSize( size );
    
        // reset all elements in this block to zero
        for (uint j = 0; j < size; j++)
            (*this)[0][j] = T(0);
    }// if
}

//
// set all n = number blocks to size size[]
//
template <class T>
TBlockVector<T>::TBlockVector ( uint number, uint * size ) 
{
    uint   i, j;
    
    setSize( number );

    // resize all blocks to the given size
    for ( i = 0; i < number; i++ )
    {
        (*this)[i].setSize( size[i] );
        
        // reset all elements in this block to zero
        for ( j = 0; j < size[i]; j++ )
            (*this)[i][j] = T(0);
    }// for
}

//
// copy constructor
//
template <class T>
TBlockVector<T>::TBlockVector ( const TBlockVector<T> &v ) 
{ 
    (*this) = v; 
}

//
// constructor with vector block
//
template <class T>
TBlockVector<T>::TBlockVector ( const TVectorBlock<T> &v ) 
{  
    setSize( 1 );
    (*this)[0] = v;
}
   
//
// take the blockstructure from v but init. the vector with f
//
template <class T>
TBlockVector<T>::TBlockVector ( const T& scalar, const TBlockVector<T> &v ) 
{
    uint   i, j;
    uint   size;
    uint  n = v.getNoOfBlocks();
    
    // create v.n blocks
    setSize( n );
    
    // resize the blocks, and fill with scalar
    for ( i = 0; i < n; i++ ) 
    {
        size = v[i].getSize();
        (*this)[i].setSize( size );
        
        for ( j = 0; j < size; j++ )
            (*this)[i][j] = scalar;
    }// for
}

//
// destructor
//
template <class T>
TBlockVector<T>::~TBlockVector () 
{
} 

//
// set all vector components to scalar
//
template <class T>
void
TBlockVector<T>::set ( const T& scalar ) 
{
    uint   i;
    uint  n = getNoOfBlocks();
    
    for ( i = 0; i < n; i++ )
    {
        (*this)[i].set( scalar );
    }// for
}

//
// get/set number of blocks
//
template <class T>
uint
TBlockVector<T>::getNoOfBlocks () const
{
    return TDynArray< TVectorBlock<T> >::getSize();
};

template <class T>
void
TBlockVector<T>::setNoOfBlocks ( uint no )
{
    setSize( no );
}

//
// return size (sum of sizes of all blocks)
//
template <class T>
const uint
TBlockVector<T>::getSize () const
{
    uint   i;
    uint  size = 0;
    uint  n    = getNoOfBlocks();
    
    for ( i = 0; i < n; i++ )
        size += (*this)[i].getSize();
    
    return size;
}

//
// vector addition ( we assume, that the vectors have same structure )
//
template <class T>
TBlockVector<T>&
TBlockVector<T>::operator+= ( const TBlockVector<T>& vec ) 
{
    uint   i;
    uint  j, blockSize;
    uint  n = getNoOfBlocks();

    for (i = 0; i < n; i++)
    {
        blockSize = (*this)[i].getSize();
        
        for (j = 0; j < blockSize; j++)
            (*this)[i][j] += vec[i][j];
    }// for

    return (*this);
}

template <class T>
TBlockVector<T>& 
TBlockVector<T>::operator+= ( const TSparseBlockVector& vec )
{
    uint   i;
    uint  n = getNoOfBlocks();
    
    for ( i = 0; i < n; i++ )
        (*this)[i] += vec[i];
   
    return (*this);
}

template <class T>
TBlockVector<T>
TBlockVector<T>::operator+ ( const TBlockVector<T>& vec ) 
{
    TBlockVector<T> vecTmp( *this );
    return (vecTmp += vec);
}

//
// vector subtraction ( we assume, that the vectors have same structure )
//
template <class T>
TBlockVector<T>&
TBlockVector<T>::operator-= ( const TBlockVector<T>& vec ) 
{
    uint   i;
    uint  j, blockSize;
    uint  n = getNoOfBlocks();

    for (i = 0; i < n; i++)
    {
        blockSize = (*this)[i].getSize();
        for (j = 0; j < blockSize; j++)
            (*this)[i][j] -= vec[i][j];
    }// for

    return (*this);
}

template <class T>
TBlockVector<T>& 
TBlockVector<T>::operator-= ( const TSparseBlockVector& vec )
{
    uint   i;
    uint  n = getNoOfBlocks();

    for ( i = 0; i < n; i++ )
        (*this)[i] -= vec[i];
   
    return (*this);
}

template <class T>
TBlockVector<T>
TBlockVector<T>::operator- ( const TBlockVector<T>& vec ) 
{
    TBlockVector<T> vecTmp( *this );
    return (vecTmp -= vec);
}

//
// scalar mult.
//
template <class T>
TBlockVector<T>&
TBlockVector<T>::operator*= ( const T& scalar ) 
{
    uint   i;
    uint  j, blockSize;
    uint  n = getNoOfBlocks();

    for (i = 0; i < n; i++)
    {
        blockSize = (*this)[i].getSize();
        for (j = 0; j < blockSize; j++)
            (*this)[i][j] *= scalar;
    }// for

    return (*this);
}

template <class T>
TBlockVector<T>
TBlockVector<T>::operator* ( const T& scalar ) 
{
    TBlockVector<T> vecTmp( *this );
    return (vecTmp *= scalar);
}

//
// inner product ( we assume, that the vectors have same structure )
//
template <class T>
T
TBlockVector<T>::operator* ( const TBlockVector<T>& vec ) const
{
    uint   i;
    uint  j, blockSize;
    T      product = T(0);
    uint  n = getNoOfBlocks();
    
    for (i = 0; i < n; i++)
    {
        blockSize = (*this)[i].getSize();
        for (j = 0; j < blockSize; j++)
            product += (*this)[i][j] * vec[i][j];
    }// for

    return (product);
}

//
// copy operator
//
template <class T>
TBlockVector<T> &
TBlockVector<T>::operator= ( const T & scalar )
{
    set( scalar );

    return *this;
}

//
// init vec with same structure, but no data
//
template <class T>
void
TBlockVector< T >::clone ( TBlockVector< T > & vec ) const
{
    int  i, size;

    size = getNoOfBlocks();

    vec.setNoOfBlocks( size );

    for ( i = 0; i < size; i++ )
        (*this)[i].clone( vec[i] );
}

//
// some blas routines
//

// x = x + alpha y
template <class T>
const TBlockVector< T > &
TBlockVector< T >::daxpy ( const TBlockVector< T > & y, const T & alpha )
{
    int  i, size;

    size = getNoOfBlocks();

    for ( i = 0; i < size; i++ )
        (*this)[i].daxpy( y[i], alpha );

    return *this;
}

// x = x + alpha A y
const TBlockVector< TFloat > &
TBlockVector< TFloat >::matmulplus ( const TBlockMatrix           & A,
                                     const TBlockVector< TFloat > & y,
                                     const TFloat                 & alpha )
{
    TBlockVector< TFloat >  tmp;

    // tmp = Ay
    A.matmul( tmp, y );

    // x = x + alpha tmp;
    daxpy( tmp, alpha );

    return *this;
}

//
// euclid norm
//
template <class T>
const TFloat
TBlockVector<T>::euclidNorm ( TFloat *scal ) const
{ 
    TFloat f = 0.0;
    uint  n = getNoOfBlocks();
    
    if (scal == NULL)
        return TFloat(sqrt((*this) * (*this))); 
    else
    {
        for (uint i = 0; i < n ; i++)
            f += ( scal[i] * TFloat((*this)[i] * (*this)[i]) );
        
        f = sqrt(f);
        
        return (f);
    }// else
}
	
//
// weighted euclid norm
//
template <class T>
const TFloat
TBlockVector<T>::weightedEuclidNorm ( TFloat *scal ) const
{
    uint    i;
    TFloat  f = 0.0;
    uint   n = getNoOfBlocks();
    
    if (scal == NULL)
        f = TFloat(sqrt((*this) * (*this)));
    else
    {
        for ( i = 0; i < n ; i++)
            f += ( scal[i] * TFloat((*this)[i] * (*this)[i]) );
        f = sqrt(f);
    }// else
    
    uint length = 0;
    
    // sum up the size of each block
    for ( i = 0; i < n ; i++) 
        length += (*this)[i].getSize();

    if (length != 0)
         return TFloat(f / sqrt(length));
    else
    {
        cout << "TBlockVector::weightedEuclidNorm (...) : length is zero" << endl;
        return 0;
    }// else
}
		
//
// euclid norm of a single block
//
template <class T>
const TFloat
TBlockVector<T>::euclidBlockNorm ( uint i ) const
{ 
    return TFloat(sqrt((*this)[i] * (*this)[i])); 
}

//
// weighted euclid norm of a single block
//
template <class T>
const TFloat
TBlockVector<T>::weightedEuclidBlockNorm ( uint i ) const
{
    TFloat f = TFloat(sqrt((*this)[i] * (*this)[i]));

    return TFloat( f / sqrt( (*this)[i].getSize() ) );
}
	
//
// maximum norm
//
template <class T>
const TFloat
TBlockVector<T>::maxNorm () const
{	
    uint    i;
    uint   j, blockSize;
    TFloat  maxValue = 0.0;
    uint   n = getNoOfBlocks();

    for (i = 0; i < n; i++)
    {
        blockSize = (*this)[i].getSize();
        
        for (j = 0; j < blockSize; j++)
            maxValue = MAX( maxValue, fabs((*this)[i][j]) );
    }// for

    return (maxValue);
}

//
// maximum norm of a single block
//
template <class T>
const TFloat
TBlockVector<T>::maxNorm ( uint i) const
{	
    uint   j, blockSize;
    TFloat  maxValue = 0.0;

    blockSize = (*this)[i].getSize();
    
    for (j = 0; j < blockSize; j++)
        maxValue = MAX( maxValue, fabs((*this)[i][j]) );

    return (maxValue);
}

//
// permute first n-1 blocks
//
template <class T>
void
TBlockVector<T>::permute ( const TVectorBlock<uint> *p ) 
{
    uint  n = getNoOfBlocks();

    if (n == 1)
        (*this)[0].permute( p );
    else
    {
        for (uint i = 0; i < ( n-1 ); i++) 
            (*this)[i].permute( p );
    }// else
}

template <class T>
void
TBlockVector<T>::permute ( const TPermutation * pPerm ) 
{
    uint  n = getNoOfBlocks();

    if (n == 1)
        (*this)[0].permute( pPerm );
    else
    {
        for (uint i = 0; i < ( n-1 ); i++) 
            (*this)[i].permute( pPerm );
    }// else
}

//
// some dummy objects for code creation
//
template class TVectorBlock<TFloat>;
template class TVectorBlock<uint>;
template class TBlockVector<TFloat>;
template class TBlockVector<uint>;

