//
//  Project : libnet
//  File    : prolongation.cc
//  Author  : Sabine Gutsch / Thomas Probst
//  Purpose : prolongation between refinement levels
//
// arch-tag: 2475d1b5-d69e-4e5e-8d67-087afac428d2
//

#include "T2DFrontier.hh"

#include "prolongation.hh"

//
// constructor
// fBoundary is the number of boundary nodes for which we need
// a prolongation/restriction
//
TBlockProlongation::TBlockProlongation ( TNet<2> *cNet, uint afDim, 
                                         uint acDim, uint fBoundary=0 ) 
	: TMatrixBlock ( afDim ) 
{
	cDim = acDim;
	fDim = afDim;
	
	// simple error checking
    if ( (fBoundary == 0) && (cDim != cNet->getNoOfInteriorNodes()) ) 
        cerr << "TBlockProlongation::TBlockProlongation: dimension mismatch 1" << endl;
    if ( (fBoundary != 0) && (cDim != cNet->getNoOfInteriorNodes() + cNet->getNoOfBoundaryNodes()) ) 
        cerr << "TBlockProlongation::TBlockProlongation: dimension mismatch 2" << endl;
    
	uint i,j;
    //  name of fine/coarse node
	uint fName, cName, cNewName; 
		
	typedef TCell<2> TNetCell;
	TListIterator<TNetCell>  iter(cNet->getList());
	for ( TCell<2> *cell = iter.first(); cell != NULL; cell = iter.next()) {
		for (i = 0; i < 3; i++) {
            if ( (((T2DFrontier*)cell->getFrontier(i))->getFrontierType() != BOUNDARY) ||
                  (fBoundary != 0) ) { 
                // Frontier is not a boundary frontier or prolong all
                if ( (cell->getFrontier( i )->hasMidNode()) ) {
                    fName = cell->getFrontier(i)->getNode(1)->getName(); 
                    if ( fName >= fDim ) 
                        fName = ULONG_MAX - fName + fDim - fBoundary;
					// '-1' with Lagrange
					//fName = ULONG_MAX - fName + fDim - fBoundary - 1;

                    // name of midnode (in fNet)
                    // midnode ( not a boundary node, since frontier is no boundary frontier
                    if ( (*this)[fName].getNoOfEntries() == 0 ) { 
                        // not done yet
                        uint  newIndex[2];
                        TFloat newValue[2];
                        j = 0;
                        cName = cell->getFrontier(i)->getNode(0)->getName();
                        if ( cName < cDim ) {
                            newIndex[j] = cName; 
                            newValue[j++] = 0.5;
                        }
                        if ( (cName >= cDim) && (fBoundary != 0) ) {
                            newIndex[j] = ULONG_MAX - cName + cNet->getNoOfInteriorNodes(); 
                            newValue[j++] = 0.5;
						}
                        cName = cell->getFrontier(i)->getNode(2)->getName();
                        if( cName < cDim) {
                            newIndex[j] = cName; 
                            newValue[j++] = 0.5; 
						}
                        if ( (cName >= cDim) && (fBoundary != 0) ) {
                            newIndex[j] = ULONG_MAX - cName + cNet->getNoOfInteriorNodes(); 
                            newValue[j++] = 0.5;
						}
                        // if ( j == 0 ) cerr << "TBlockProlongation::TBlockProlongation:  Frontier has only boundary nodes, but no boundary flag " << endl; 
                        if ( fName > fDim ) cerr << " f is boundary " << endl;
                        if ( j > 0 ) (*this)[fName].set( j, newIndex, newValue );

                    }// if
                }
                
				// and now the identity part of the prolongation/
				// restriction matrix

                // first (coarse) vertex
                cName = cell->getFrontier(i)->getNode(0)->getName();
                cNewName = cName;
                if ( (fBoundary != 0) && (cName >= cDim) ) {
                    cName = ULONG_MAX - cName + cNet->getNoOfInteriorNodes();
                    // '-1' with lagrange
                    //cNewName = cName - cNet->getNoOfInteriorNodes() + fDim - fBoundary - 1;
                    cNewName = cName - cNet->getNoOfInteriorNodes() + fDim - fBoundary;
				}
                if ( cName < cDim ) { 
                    // not a Boundary node or prolong all
                    if ( (*this)[cNewName].getNoOfEntries() == 0 ) {  
                        // not done yet
                        uint  newIndex[1];
                        TFloat newValue[1];
                        newIndex[0] = cName;
                        newValue[0] = 1.0;
                        (*this)[cNewName].set( 1, newIndex, newValue );
                    }	
                }
                // second (coarse) vertex
                cName = cell->getFrontier(i)->getNode(2)->getName();
                cNewName = cName;
                 if ( (fBoundary != 0) && (cName >= cDim) ) {
                    cName = ULONG_MAX - cName + cNet->getNoOfInteriorNodes();
                    // '-1' with lagrange
                    //cNewName = cName - cNet->getNoOfInteriorNodes() + fDim - fBoundary - 1;
                    cNewName = cName - cNet->getNoOfInteriorNodes() + fDim - fBoundary;
				}
                if (cName < cDim) { 
                    // not a Boundary node or prolong all
                    if ((*this)[cNewName].getNoOfEntries() == 0) {  
                        // not done yet
                        uint  newIndex[1];
                        TFloat newValue[1];
                        newIndex[0] = cName;
                        newValue[0] = 1.0;
                        (*this)[cNewName].set( 1, newIndex, newValue );
                    }
                }
            }
        }// for
	}// for
	// and finally the identity prolongation/restriction for the
	// Lagrange factor
	/*
    if ( fBoundary != 0 ) {
        uint  newIndex[1];
        TFloat newValue[1];
        newIndex[0] = cDim-1;
        newValue[0] = 1.0;
        (*this)[fDim-1].set( 1, newIndex, newValue );    
	}
	*/
}

//
//  matrix dependent prolongation/restriction
//
void
TBlockProlongation::matrixDependent ( TMatrixBlock *mb )
{
    uint i,j;
    TFloat newvalue;
    for (i = 0; i < getNoOfRows(); i++) 
        if ( (*this)[i].getNoOfEntries() > 0 && (*this)[i].getValue( 0 ) < 0.999 )
            for (j = 0; j < (*this)[i].getNoOfEntries(); j++)
		    {
                //cout << i << "  " << (*this)[i] << endl;
                //cout << j << "  " << (*this)[i].getIndex( j ) << endl;
                //cout << (*mb)[i] << endl;
                newvalue = (*mb)[i][(*this)[i].getIndex( j )] / (*mb)[i].getValue( 0 );
                (*this)[i].setValue( j, newvalue );
		    }
	
}

//
// prolongation
//
TVectorBlock<TFloat> 
TBlockProlongation::prolong( TVectorBlock<TFloat> &cVec ) 
{
    TVectorBlock<TFloat> fVec( 0.0, fDim );
    fVec = (*this) * cVec;
	return (fVec);
}

//
// restriction
//
TVectorBlock<TFloat> 
TBlockProlongation::restrict( TVectorBlock<TFloat> &fVec ) 
{
	TVectorBlock<TFloat> cVec( 0.0, cDim );
    (*this).transposeMult( fVec, cVec);
	return (cVec);
}

//////////////////////////////////////////////////////////////

//
// prolongation
//
TBlockVector<TFloat> 
TProlongation::prolong ( TBlockVector<TFloat> &cVec ) 
{
	uint i;
	uint size[n];
	
	for (i = 0; i < n; i++) 
		size[i] = pBlockProlong[i]->getfDim();
	
	TBlockVector<TFloat> fVec( n, size );

	for (i = 0; i < n; i++) 
		fVec[i] = pBlockProlong[i]->prolong( cVec[i] );
	
	return (fVec);
}

//
// restriction
//
TBlockVector<TFloat> 
TProlongation::restrict( TBlockVector<TFloat> &fVec ) 
{
	uint i;
	uint size[n];
	
	for (i = 0; i < n; i++) 
		size[i] = pBlockProlong[i]->getcDim();
	
	TBlockVector<TFloat> cVec( n, size );

	for (i = 0; i < n; i++) 
		cVec[i] = pBlockProlong[i]->restrict( fVec[i] );

	return (cVec);
}
