//
//  Project  : libnet
//  File     : sort.cc
//  Author   : Thomas Probst / Sabine Gutsch
//  Purpose  : class for linear conforming 2D finite elements
//
// arch-tag: f256e148-6905-4825-be6a-a52a67336587
//

#include "sort.hh"
#include "functions.hh"

void 
TSort::sort( TSortType aSortType ) {
  uint i;
  uint  BOUNDARY_NAME = ULONG_MAX - pNet->getNoOfBoundaryNodes();
  uint dof;

  TVector<3> *sortVec = new TVector<3> [dim];
  
  typedef TCell<2> TNetCell;
  TListIterator<TNetCell>  iter(pNet->getList());
  for (TCell<2> *cell = iter.first(); cell != NULL; cell = iter.next()) 
    for (  i = 0; i < 3; i++) {
      dof = cell->getNode(i)->getName();
      if ( dof < BOUNDARY_NAME ) {
        sortVec[dof][0] = dof;
        sortVec[dof][1] = ((*cell->getNode(i))[0]);
        sortVec[dof][2] = ((*cell->getNode(i))[1]);
      }
    }


 size_t size = sizeof( sortVec[0] );
 if ( aSortType == lexicographical )
     qsort( sortVec, dim, size, lexComp);
 if ( aSortType == concCycle )
     qsort( sortVec, dim, size, cycleComp);

 TFloat  x;
 for ( i = 0; i < dim; i++) {
     x = sortVec[i][0];
     (*invperm)[i] = (uint)x;
     (*perm)[(*invperm)[i]] = i;
     //cout<< "x: "<<sortVec[i][1]<< " y: "<<sortVec[i][2]<<endl;
 }
 
 // permute A, rhs and X:

 if ( pA->getNoOfRows() > 1 ) {
     // real block matrix
     pA->permute( perm ); 
     pRHS->permute( perm );
     pX->permute( perm );
 } 
 else { 

     // block matrix has only one  block
     (*pA)[0][0].permute( perm, perm); 
     (*pRHS)[0].permute( perm );
     (*pX)[0].permute( perm );
 } 
 delete[] sortVec;
}

void 
TSort::redBlack() {
  uint i;
  uint  BOUNDARY_NAME = ULONG_MAX - pNet->getNoOfBoundaryNodes();
  uint dof;

  TVector<3> *sortVec = new TVector<3> [dim];
  
  typedef TCell<2> TNetCell;
  TListIterator<TNetCell>  iter(pNet->getList());
  for (TCell<2> *cell = iter.first(); cell != NULL; cell = iter.next()) 
    for (  i = 0; i < 3; i++) {
      dof = cell->getNode(i)->getName();
      if ( dof < BOUNDARY_NAME ) {
          sortVec[dof][0] = dof;
          sortVec[dof][1] = ((*cell->getNode(i))[0]);
          sortVec[dof][2] = ((*cell->getNode(i))[1]);
      }
    }
  
 size_t size = sizeof(sortVec[0]);
 qsort(sortVec, dim, size, lexComp);

 //  cout<< "x: "<<sortVec[i][1]<< " y: "<<sortVec[i][2]<<endl;

 uint offset;
 if (dim %2 == 0) {
     offset=dim/2; // dim even
     cout << "even offset: " << offset << endl;
       for ( i = 0; i < offset; i++) {
         (*invperm)[i] = (uint)sortVec[2*i][0];
         (*invperm)[i+offset] = (uint)sortVec[2*i+1][0];
     }
 }
 else { 
     offset = uint(ceil(dim/2.0));
     cout<< "odd offset: " << offset << endl; //  dim  odd
     for ( i = 0; i < offset-1; i++){
         (*invperm)[i] = (uint)sortVec[2*i][0];
         (*invperm)[i+offset] = (uint)sortVec[2*i+1][0];
         //   cout<<"i: "<<i <<"  x: "<<sortVec[2*i][1]<< " y: "<<sortVec[2*i][2]<<endl;
         //   cout<<"i+off: "<<i+offset<<" x: "<<sortVec[2*i+1][1]<< " y: "<<sortVec[2*i+1][2]<<endl;
     }
     (*invperm)[offset-1] = (uint)sortVec[2*(offset-1)][0];
     // cout<<"i: "<<offset-1 <<"  x: "<<sortVec[2*(offset-1)][1]<< " y: "<<sortVec[2*i][2]<<endl;
 }

 for ( i =0; i < dim; i++) {
     (*perm)[(*invperm)[i]] = i;
 } 

 // permute A, rhs and X:

 if ( pA->getNoOfRows()>1) {
    // real block matrix
    pA->permute(perm); 
    pRHS->permute(perm);
    pX->permute(perm);} 
 else { 
    // block matrix has only one  block
    (*pA)[0][0].permute(perm,perm); 
    (*pRHS)[0].permute( perm );
    (*pX)[0].permute( perm );
 } 
 delete[] sortVec;
}

void  
TSort::order(TBlockMatrix *pM) {
    // permute M
    if ( pM->getNoOfRows() > 1)
        // stokes block matrix
        pM->permute(perm); 
    else  // block matrix has only one  block
        (*pM)[0][0].permute(perm,perm) ;
}

void  
TSort::order(TMatrixBlock *pM) { pM->permute( perm, perm); }

void  
TSort::order(TBlockVector<TFloat> *pV) {
    // permute V:
    
    if ( pV->getNoOfBlocks() > 1)
        pV->permute(perm); // stokes block vector
    else  // block vector has only one  block
        (*pV)[0].permute( perm );
}
 
void  
TSort::restore() {
 // permute A, rhs and X:

if ( pA->getNoOfRows() > 1) {
    // stokes block matrix
    pA->permute(invperm); 
   pRHS->permute(invperm);
   pX->permute(invperm);} 
   else { 
       // block matrix has only one  block
       (*pA)[0][0].permute(invperm,invperm) ;
       (*pRHS)[0].permute( invperm );
       (*pX)[0].permute( invperm );
 }
}

void  
TSort::restore(TBlockMatrix *pM) {
    // permute M
    if ( pM->getNoOfRows()>1)
        pM->permute(invperm); // stokes block matrix
    else  // block matrix has only one  block
        (*pM)[0][0].permute(invperm,invperm) ;
}

void  
TSort::restore(TBlockVector<TFloat> *pV){
    // permute V:
    
    if ( pV->getNoOfBlocks()>1)
        pV->permute(invperm); // stokes block vector
    else  // block vector has only one  block
        (*pV)[0].permute( invperm );
}

