//
//  Project   : libmisc
//  File      : TString.cc
//  Author    : Ronald Kriemann
//  Purpose   : class for a string
//  Date      : 17.01.1998
//
// arch-tag: b7d511b3-cb16-4b27-b44b-99c73fc2ce81
//

extern "C" {
#include <ctype.h>
}

#include <stdio.h>
#include <string.h>

#include "TString.hh"

//
// constructor
//
TString::TString ( size_t size )
{
    size_t  i;

    setGrowBy( 16 );
    setSize( size );

    // reset data
    for (i = 0; i < size; i++)
        (*this)[i] = 0;
}

TString::TString ( const TString &string )
{
    ::TString();

    *this = string;
}

TString::TString ( const char string[], const size_t size )
{
    ::TString();

    copyFrom( string, size );
}

TString::TString ( const uchar string[], const size_t size )
{
    ::TString();

    copyFrom( string, size );
}

//
// destructor
//
TString::~TString ()
{}

//
// access string status
//
const size_t 
TString::length () const
{
    return getSize();
}

const size_t 
TString::charSize ()
{
    return (sizeof(uchar) * 8);
}

const bool 
TString::isEmpty () const
{
    size_t  i;
    size_t  size = length();

    if (size == 0)
        return TRUE;
    else
    {
        for (i = 0; i < size; i++)
        {
            if ((*this)[i] != 0)
                return FALSE;
        }// for
    }// else

    return TRUE;
}

//
// set string length
//
void
TString::setLength ( size_t n )
{
    size_t  i;
    size_t  oldSize = length();

    if (n < 0)
    {
#if defined(EXCEPTIONS)
        throw TSize(n);
#else
        cerr << "TString::setLength (" << n << ") : gave negative size" << endl;
#endif
        return;
    }// if
    
    setSize( n );

    // reset data
    for (i = oldSize; i < n; i++)
        (*this)[i] = 0;
}

//
// import (export) from (to) normal strings
//
void
TString::copyTo ( uchar string[], size_t size ) const
{
    size_t  i;
    bool    bWithGivenSize;
    size_t  stringSize = size < length() ? size : length();

    bWithGivenSize = !(size == ULONG_MAX);
    
    for (i = 0; i < stringSize; i++)
        string[i] = (*this)[i];

    // if there is some space, set an ending 0
    if (bWithGivenSize && (size > length()))
        string[length()] = 0;
}

void
TString::copyTo ( char string[], size_t size ) const
{
    size_t  i;
    bool    bWithGivenSize;
    size_t  stringSize = size < length() ? size : length();

    bWithGivenSize = !(size == ULONG_MAX);
    
    for (i = 0; i < stringSize; i++)
        string[i] = (char) (*this)[i];

    // if there is some space, set an ending 0
    if (bWithGivenSize && (size > length()))
        string[length()] = 0;
}

void
TString::copyFrom ( const uchar string[], size_t size )
{
    size_t  i;
    size_t  stringSize = size == ULONG_MAX ? strlen( (char*) string ) : size;

    setSize( stringSize );

    for (i = 0; i < stringSize; i++)
        (*this)[i] = string[i];
}

void
TString::copyFrom ( const char string[], size_t size )
{
    size_t  i;
    size_t  stringSize = size == ULONG_MAX ? strlen( (char*) string ) : size;

    setSize( stringSize );

    for (i = 0; i < stringSize; i++)
        (*this)[i] = (uchar) string[i];
}

//
// access substrings
//
TString
TString::substr ( size_t pos, size_t len ) const
{
    size_t    j, k;
    TString   temp;
    size_t    max = (pos + len) < length() ? pos + len : length();

    temp.setLength( len );
    for ( j = pos, k = 0; j < max; j++, k++ )
        temp[k] = (*this)[j];

    return temp;
}

TString 
TString::index ( size_t pos ) const
{
    return substr( pos, length()-pos );
}

//
// locate substrings or char.
//
TString 
TString::strchr ( char c ) const
{
    return strchr( uchar(c) );
}

TString 
TString::strchr ( uchar c ) const
{
    size_t    i;
    size_t    len = length();
    TString  temp;

    for (i = 0; i < len; i++)
    {
        if ((*this)[i] == c)
            return index( i );
    }// for

    return temp;
}

TString
TString::strstr ( const TString & needle ) const
{
    size_t   i;
    bool    bFound;
    TString string( *this );

    if (needle.length() == 0)
        return string;
    
    while ((string = string.strchr( needle[0])).length() > 0)
    {
        bFound = TRUE;
        
        for (i = 1; i < needle.length(); i++)
        {
            if (string[i] != needle[i])
            {
                bFound = FALSE;
                break;
            }// if
        }// for

        if (bFound)
            return string;
        else
            string <<= 1;
    }// while

    return string;
}

//
// return substrings, seperated by delim
// (remove all char before delim in string)
//
TString
TString::strtok ( char delim )
{
    return strtok( uchar(delim) );
}

TString
TString::strtok ( uchar delim )
{
    size_t    i,j,k;
    size_t    len = length();
    TString   temp( "" );

    for (i = 0; i < len; i++)
    {
        if ((*this)[i] == delim)
        {
            //
            // remove all char before and incl c
            //
            for (j = i+1, k = 0; j < len; j++, k++)
                (*this)[k] = (*this)[j];

            setLength( len-i-1 );
            
            return temp;
        }// if
        else
            temp += (*this)[i];
    }// for

    // we did not find the char => return empty string
    temp = "";
    return temp;
}
    
//
// various operators for conversion and manipulation
//

TString & 
TString::operator += ( const TString &string )
{
    size_t i;
    size_t size1, size2, size3;
    uchar  a;

    size1 = length();
    size2 = string.length();
    size3 = size1 + size2;

    setSize( size3 );

    for (i = size1; i < size3; i++)
    {
        a = string[i-size1];
        (*this)[i] = a;
    }// for

    return *this;
}

TString & 
TString::operator += ( const uchar c )
{
    setLength( length() + 1 );
    (*this)[length()-1] = c;

    return *this;
}

const bool
TString::operator == ( const TString &string ) const
{
    long   i;
    long   size1, size2;

    uchar *pData1;
    uchar *pData2;

    size1 = length();
    size2 = string.length();

    pData1 = pData;
    pData2 = string.pData;

    if (size2 < size1)
    {
        for ( i = size1-1; i >= size2; i-- )
            if (pData1[i] != 0)
                return FALSE;

        for ( i = size2-1; i >= 0; i-- )
        {
            if (pData1[i] != pData2[i])
                return FALSE;
        }// for
    }// if
    else
    {
        for ( i = size2-1; i >= size1; i-- )
            if (pData2[i] != 0)
                return FALSE;

        for ( i = size1-1; i >= 0; i-- )
        {
            if (pData1[i] != pData2[i])
                return FALSE;
        }// if
    }// else

    return TRUE;
}

const bool
TString::operator != ( const TString &string ) const
{
    return !(*this == string);
}

TString & 
TString::operator << ( const TString &string )
{
    return *this += string;
}

TString &
TString::operator >> ( char & c )
{
    if (length() > 0)
    {
        c = (*this)[0];
        *this <<= 1;
    }// if

    return *this;
}

TString &
TString::operator >> ( uchar & c )
{
    if (length() > 0)
    {
        c = (*this)[0];
        *this <<= 1;
    }// if

    return *this;
}

TString &
TString::operator >> ( int & n )
{
    long  d = n;

    *this >> d;
    n = d;
    
    return *this;
}

TString &
TString::operator >> ( long & n )
{
    size_t  len = length();
    size_t  pos = 0;
    uchar   c   = 0;
    bool    bNeg = FALSE;

    while ((pos < len) && (isspace(c = (*this)[pos])))
        pos++;

    if (c == '-')
    {
        bNeg = TRUE;
        pos++;
    }// if
        
    while ((pos < len) && (isspace(c = (*this)[pos])))
        pos++;

    if ((pos >= len) || (!isdigit( c )))
        return *this;

    n = 0;
    while ((pos < len) && (isdigit(c = (*this)[pos])))
    {
        n *= 10;
        n += int(c) - 48;

        pos++;
    }// while

    if (bNeg)
        n = -n;
    
    *this <<= pos;
    
    return *this;
}

TString &
TString::operator >> ( float & f )
{
    double  d = f;

    *this >> d;
    f = d;
    
    return *this;
}

TString &
TString::operator >> ( double & f )
{
    size_t   len = length();
    size_t   pos = 0;
    uchar    c   = 0;
    bool     bNeg = FALSE;
    double   d;

    while ((pos < len) && (isspace(c = (*this)[pos])))
        pos++;

    if (c == '-')
    {
        bNeg = TRUE;
        pos++;
    }// if
        
    while ((pos < len) && (isspace(c = (*this)[pos])))
        pos++;

    if ((pos >= len) || (!isdigit( c )))
        return *this;

    f = 0;
    while ((pos < len) && (isdigit(c = (*this)[pos])))
    {
        f *= 10;
        f += int(c) - 48;

        pos++;
    }// while

    if ((*this)[pos] == '.')
    {
        pos++;

        d = 1;
        while ((pos < len) && (isdigit(c = (*this)[pos])))
        {
            d /= 10;
            f += double(int(c) - 48) * d;
            
            pos++;
        }// while
    }// while
    
    if (bNeg)
        f = -f;

    *this <<= pos;
    
    return *this;
}

TString &
TString::operator <<= ( size_t n )
{
    size_t  i, j;
    size_t  len = length();

    if (n > 0)
    {
        for (i = n, j = 0; i < len; i++, j++)
            (*this)[j] = (*this)[i];
        
        setLength( (len - n) >= 0 ? len - n : 0 );
    }// if
    
    return *this;
}

TString &
TString::operator >>= ( size_t n )
{
    long    i, j;
    size_t  len = length();

    if (n > 0)
    {
        if (len == 0)
            setLength( n );
        else
        {
            setLength( len + n );
    
            for (i = len-1, j = len+n-1; i >= 0; i--, j--)
                (*this)[j] = (*this)[i];
        }// else
    }// if
    
    return *this;
}
    
//
// misc. methods
//

//
// remove 0's at end
//
void
TString::shorten ()
{
    size_t  size;
    uchar  *pData1 = pData;

    //
    // check if string is almost empty and set begin of search
    //
    size = length();
    if (size > 1)
        size--;
    else
        return;
    
    //
    // look for first non-zero
    //
    while ((size > 0) && (pData1[size] == 0))
        size--;

    setSize( size+1 );
}

//
// concatenation
//
TString
operator + ( const TString & string1, const TString & string2 )
{
    TString temp = string1;

    return temp += string2;
}

//
// output
//
ostream & 
operator << ( ostream &os, const TString &string )
{
    size_t   i;
    size_t   size = string.length();

    i = 0;
    while ((i < size) && (string[i] != 0))
    {
        os << string[i++];
    }// while

    return os;
}

