/*
 * LKMALLOC by Larson and Krishnan
 * -------------------------------
 *
 * LKmalloc as you see it here is a modified version of
 * Rmalloc. But it tries to mimic LKmalloc as best as is
 * possible.
 *
 * Changelog
 * ---------
 *
 * v0.2 - changed array sizes in stat and blocks to NO_OF_CLASSES+1
 *        to hold blocks of last size class
 *      - added checking of wilderness in check_lkmalloc
 *      - removed potential bug in thread -> heap conversion (% TSD_MAX)
 *      - added flag USE_SPINLOCKS to turn on/off spinlock support
 *        (default: off, use PTHREAD mutices)
 * v0.1 - initial release
 *
 */

/*
 * the version info
 */

#define LKMALLOC_VERSION  "0.2"

/*****************************************************************
 *****************************************************************
 **
 ** some options
 **
 *****************************************************************
 *****************************************************************/

/*
 * use lkmalloc as a malloc-replacment
 */
#if defined(__cplusplus) && !defined(OVERLOAD_MALLOC)
#  define OVERLOAD_MALLOC  0
#else
#  ifndef OVERLOAD_MALLOC
#  define OVERLOAD_MALLOC  0
#  endif
#endif

/*
 * use lkmalloc as a new/delete replacment
 */
#if defined(__cplusplus) && !defined(OVERLOAD_NEW)
#define OVERLOAD_NEW     1
#endif

/*
 * use pthreads
 */
#ifndef USE_THREAD
#define USE_THREAD       1
#endif

/*
 * use spinlocks instead of pthread locks
 */
#ifndef USE_SPINLOCKS
#define USE_SPINLOCKS    0
#endif

/*
 * define system-allocation method: MALLOC, MMAP, SBRK
 */
#if OVERLOAD_MALLOC == 1
#  ifndef USE_MMAP
#  define USE_MMAP         1
#  endif
#  ifndef USE_SBRK
#  define USE_SBRK         0
#  endif
#else
#  ifndef USE_MALLOC
#  define USE_MALLOC       0
#  endif
#  ifndef USE_MMAP
#  define USE_MMAP         1
#  endif
#  ifndef USE_SBRK
#  define USE_SBRK         0
#  endif
#endif

/*
 * check heap extracted from memory-chunk
 */
#ifndef CHECK_HEAP
#define CHECK_HEAP       1
#endif


#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>

#if USE_THREAD == 1
#include <pthread.h>
#endif

#include "rmalloc.h"

#ifdef __cplusplus
extern "C" {
#endif

/*****************************************************************
 *****************************************************************
 **
 ** defines and option setting
 **
 *****************************************************************
 *****************************************************************/

/* overload malloc functions */
#if OVERLOAD_MALLOC == 1
#  define  lk_malloc         malloc
#  define  lk_calloc         calloc
#  define  lk_free           free
#  define  lk_realloc        realloc
#endif

/*
 * total number of size classes,
 * number of small classes and small size
 * and first "middle" size
 */
#define NO_OF_CLASSES   128
#define SMALL_SIZE      512
#define MIN_BLOCK_SIZE  32

/* alignment of all memory */
#if 0
#define ALIGNMENT          (sizeof(double)-1)
#define IS_ALIGNED( s )    (((s) & ALIGNMENT) == 0)
#define IS_MISALIGNED( s ) (((s) & ALIGNMENT) != 0)
#define ALIGN( s )         ((s) = (((s)+ALIGNMENT) & ~ALIGNMENT))
#else
#define ALIGNMENT          (sizeof(double))
#define IS_ALIGNED( s )    (((s) % ALIGNMENT) == 0)
#define IS_MISALIGNED( s ) (((s) % ALIGNMENT) != 0)
#define ALIGN( s )         ((s) += (ALIGNMENT - ((s) % ALIGNMENT)))
#endif
    
/*
 * set to 1 to enable statistics
 */
#define STAT_FRAGMENTATION  1

/*
 * set to 1 to enable debugging
 */
#define DEBUG_BLOCK     0
#define DEBUG_LKMALLOC  0

/* magic numbers for blocks and containers */
#define BLOCK_MAGIC  0xbabecafe
    
/* default trace file */
#define DEFAULT_TRACE_FILE  "lkmalloc.trc"

/* trace-types: 1 : trace by steps, 2: trace by allocation */
#define TRACE_STEPS       1
#define TRACE_ALLOCATION  2
    
/* default value for top-pad */
#define DEFAULT_STRIP_SIZE  4*1024*1024

/* exit/abort command */
#define ABORT  kill( getpid(), SIGKILL )
#define EXIT   exit( 1 )

/* error value in sbrk */
#define SBRK_FAIL -1

/* definitions for mmap-usage */
#if USE_MMAP == 1

#if !defined(MAP_ANONYMOUS) && defined(MAP_ANON)
#  define MAP_ANONYMOUS MAP_ANON
#endif

#if !defined(MAP_FAILED)
#  define MAP_FAILED ((char*)-1)
#endif

#if !defined(MAP_ANONYMOUS)

static int dev_zero_fd = -1;

#define MMAP(addr, size, prot, flags) ((dev_zero_fd < 0) ? \
 (dev_zero_fd = open("/dev/zero", O_RDWR), \
  mmap((addr), (size), (prot), (flags), dev_zero_fd, 0)) : \
   mmap((addr), (size), (prot), (flags), dev_zero_fd, 0))

#else

#define MMAP(addr, size, prot, flags) \
 (mmap((addr), (size), (prot), (flags)|MAP_ANONYMOUS, -1, 0))

#endif  /* !defined(MAP_ANONYMOUS) */

#endif  /* USE_MMAP == 1 */

/*****************************************************************
 *****************************************************************
 **
 ** data-types for lkmalloc
 **
 *****************************************************************
 *****************************************************************/


/* internal size and pointer types */
typedef  unsigned long  internal_size_t;

/* bool looks better than int */
#ifndef __cplusplus
typedef enum { false, true } bool;
#endif

/*
 * mutex and thread-specific-data type
 */
#if USE_SPINLOCKS == 1
typedef int              mutex_t;
#else
typedef pthread_mutex_t  mutex_t;
#endif

typedef int              tsd_key_t;

/*
 * forward declaration of types
 */

struct sysconf_s;

struct block_s;
struct block_storage_s;
struct heap_s;
    
struct heaplist_s;
struct trace_s;

typedef struct sysconf_s   sysconf_t;
typedef struct stat_s      stat_t;
    
typedef struct block_s         block_t;
typedef struct block_storage_s block_storage_t;
typedef struct heap_s          heap_t;

typedef struct heap_list_s heap_list_t;
typedef struct trace_s     trace_t;

/*
 * for holding info about system
 */
struct sysconf_s
{
    internal_size_t   pagesize;    /* number of bytes per page */
    internal_size_t   strip_size;  /* size of memory strips allocated from operating system */
};

/*
 * holding statistical information
 */
struct stat_s
{
    internal_size_t  used[ NO_OF_CLASSES+1 ];  /* number of used chunks */
    internal_size_t  free[ NO_OF_CLASSES+1 ];  /* number of free chunks */
    
    internal_size_t  used_mem;     /* total number of bytes used in heap (alloc from system) */
    internal_size_t  mem_in_use;   /* current number of bytes, used by application */
    internal_size_t  max_in_use;   /* maximal number of bytes, used by application */
    internal_size_t  allocated;    /* sum of all allocated memory in heap */
};

/*
 * datatype for memory chunks
 */
struct block_s
{
    internal_size_t  size;         /* original, requested size of block */
    void           * hc_addr;      /* heap we belong to */

#if DEBUG_BLOCK == 1
    unsigned         magic;        /* magic number for the block */
    unsigned         __tmp__;      /* for 8-byte-alignment */
#endif
    
    block_t        * next, * prev; /* next, prev pointer for list */
};

/* size of extra-data to be stored in each chunk (pointers + size-field) */
#if DEBUG_BLOCK == 1
#define PREFIX_SIZE     (sizeof(internal_size_t) + sizeof(void*) + 2*sizeof(unsigned))
#else
#define PREFIX_SIZE     (sizeof(internal_size_t) + (sizeof(void*)))
#endif
#define POSTFIX_SIZE    (sizeof(internal_size_t))
#define EXTRA_DATA_SIZE (PREFIX_SIZE)

/*
 * combines different container-types
 */
struct block_storage_s
{
    block_t  * blocks[ NO_OF_CLASSES+1 ]; /* size-classes for usual blocks */
    block_t  * wilderness;                /* last block of each data-segment */
};

/*
 * type for a heap
 */
struct heap_s
{
    int              id;                    /* unique id of heap */
    heap_t         * next;                  /* link for the heap-list */
    block_storage_t  blocks;                /* holds all stored blocks */
    stat_t           stat;                  /* status information */
    mutex_t          mutex;                 /* mutex for locking heap */
    bool             used;                  /* if false, heap is not used */
};

/*
 * list of heaps for the threads
 */
struct heap_list_s
{
    heap_t   * heaps;                       /* linked list of heaps */
    unsigned   nheaps;                      /* number of heaps in list */
    mutex_t    mutex;                       /* mutex for safety */
};

/*
 * type for tracing memory-allocation
 */
struct trace_s
{
    int               fd;                   /* file descriptor of trace-file */
    internal_size_t   old_used;             /* old number of used bytes */
    internal_size_t   old_free;             /* old number of free bytes */
    internal_size_t   old_alloc;            /* old total allocation size (time in MB) */
    unsigned          step;                 /* allocation step */
    internal_size_t   size;                 /* size of chunk to trace */
    unsigned          type;                 /* type of trace */
};



/*****************************************************************
 *****************************************************************
 **
 ** thread definition
 **
 *****************************************************************
 *****************************************************************/

#if USE_THREAD == 1

/* maximal number of threads */
#  define TSD_MAX   256

/*
 * handling of mutices
 */
#if USE_SPINLOCKS == 1
#  define LOCKED                    1
#  define MUTEX_INITIALIZER         (! LOCKED)
#  define DEFINE_MUTEX( m )         static mutex_t m = MUTEX_INITIALIZER

#  define MUTEX_INIT(mutex,status)  mutex = ! LOCKED

#  define LOCK( m )                 mutex_lock(    (m) )
#  define UNLOCK( m )               mutex_unlock(  (m) )
#  define TRYLOCK( m )              mutex_trylock( (m) )
#  define MUTEX_BUSY                1
#  define MUTEX_LOCKED( m )         (TRYLOCK( m ) == MUTEX_BUSY)
#else
#  define MUTEX_INITIALIZER  PTHREAD_MUTEX_INITIALIZER
#  define DEFINE_MUTEX( m )  static mutex_t m = PTHREAD_MUTEX_INITIALIZER

#  define MUTEX_INIT(mutex,status)  if (( status = pthread_mutex_init( & mutex, NULL )) != 0) \
                                    { fprintf( stderr, "(rmalloc) init_heap : error in mutex_init (%s)\n", \
                                               strerror( status ) ); \
                                      EXIT; }

#  define LOCK( m )         pthread_mutex_lock(    & (m) )
#  define UNLOCK( m )       pthread_mutex_unlock(  & (m) )
#  define TRYLOCK( m )      pthread_mutex_trylock( & (m) )
#  define MUTEX_BUSY        EBUSY
#  define MUTEX_LOCKED( m ) (TRYLOCK( m ) == MUTEX_BUSY)
#endif

/*
 * handling thread-specific-data
 */
#  define TSD_KEY_INIT         { int i; for ( i = 0; i < TSD_MAX; i++ ) tsd[i] = NULL; }
#  define TSD_GET_DATA         (tsd[ ((unsigned) pthread_self()) % TSD_MAX ])
#  define TSD_SET_DATA( data ) tsd[ ((unsigned) pthread_self()) % TSD_MAX ] = ((heap_t*) data);

/*
 * conversion thread <-> heap
 */

#  define GET_THR_HEAP          ((heap_t*) TSD_GET_DATA)
#  define SET_THR_HEAP( heap )  TSD_SET_DATA( (void*) heap )

#else

/* maximal number of threads */
#  define TSD_MAX   256

/*
 * handling of mutices
 */
#  define MUTEX_INITIALIZER  0
#  define DEFINE_MUTEX( m )
#  define MUTEX_INIT(mutex,status)

#  define LOCK( m )     
#  define UNLOCK( m )   
#  define TRYLOCK( m )       1
#  define MUTEX_BUSY         0
#  define MUTEX_LOCKED( m )  false

/*
 * handling thread-specific-data
 */
#  define TSD_KEY_INIT( key )  

/*
 * conversion thread <-> heap
 */
#  define THREAD_ID             0
#  define GET_THR_HEAP          & global_heap
#  define SET_THR_HEAP( heap )  

#endif  /* USE_THREAD == 1 */


/*****************************************************************
 *****************************************************************
 **
 ** heap specific data
 **
 *****************************************************************
 *****************************************************************/

/* lookup table for size classes */
static internal_size_t size_classes[ NO_OF_CLASSES ] = { 
    0, 8, 16, 24, 32, 40, 48, 56, 64, 72, 80, 88, 96, 104, 112, 120, 128, 
    136, 144, 152, 160, 168, 176, 184, 192, 200, 208, 216, 224, 232, 240, 
    248, 256, 264, 272, 280, 288, 296, 304, 312, 320, 328, 336, 344, 352, 
    360, 368, 376, 384, 392, 400, 408, 416, 424, 432, 440, 448, 456, 464, 
    472, 480, 488, 496, 504, 512, 576, 640, 704, 768, 832, 896, 960, 1024, 
    1088, 1152, 1216, 1280, 1344, 1408, 1472, 1536, 1600, 1664, 1728, 
    1792, 1856, 1920, 1984, 2048, 2112, 2176, 2240, 2304, 2368, 2432, 
    2496, 2560, 3072, 3584, 4096, 4608, 5120, 5632, 6144, 6656, 7168, 
    7680, 8192, 8704, 9216, 9728, 10240, 10752, 14848, 18944, 23040, 27136, 
    31232, 35328, 39424, 43520, 76288, 109056, 141824, 174592, 436736, 
    698880, 768776
};

/* holds info about system */
static sysconf_t  sys_conf = { 0, DEFAULT_STRIP_SIZE };

/* global heap */
static heap_t global_heap = {
    0,
    NULL,
    { { NULL }, NULL },
    { { 0 }, { 0 }, 0, 0, 0, 0 },
    MUTEX_INITIALIZER,
    true
};

/* counter for the allocated heaps */
static unsigned     heap_id = 0;

/* linked list of heaps for threads (global heap is not a part of it) */
static heap_list_t  heap_list = { NULL, 0, MUTEX_INITIALIZER };

/* thread specific data */
#if USE_THREAD == 1
static heap_t     * tsd[ TSD_MAX ];
#endif

/* is the heap initialised */
static bool         is_initialised = false;

/* how much statistics should be printed */
static unsigned     heap_stat_lvl = 0;

/* for memory trace */
static bool         trace_mem = false;
static trace_t      trace     = { -1, 0, 0, 0, 0 };

/*****************************************************************
 *****************************************************************
 **
 ** defines for easier access
 **
 *****************************************************************
 *****************************************************************/


#define BLOCK_SIZE( b )     ((b)->size & ~0x7)
#define PRED_SIZE( b )      (*(((internal_size_t*) (b)) - 1))
#define SET_SIZE( b, n )    ((b)->size = (n) + ((b)->size & 0x7))
#define SET_EOB_SIZE( b )   { char * p = ((char*) b) + BLOCK_SIZE( b ) - sizeof(internal_size_t); \
                              *((internal_size_t*) p) = BLOCK_SIZE( b ); }

#define IS_PRED_USED( b )   ((b)->size &  0x1)
#define SET_PRED_USED( b )  ((b)->size |= 0x1)
#define SET_PRED_FREE( b )  ((b)->size &= ~0x1)

#define IS_USED( b )        ((b)->size &  0x2)
#define SET_USED( b )       ((b)->size |= 0x2)
#define SET_FREE( b )       ((b)->size &= ~0x2)

#define IS_WILD( b )        ((b)->size &  0x4)
#define SET_WILD( b )       ((b)->size |= 0x4)
#define SET_TAME( b )       ((b)->size &= ~0x4)

#define BLOCK_TO_PTR( n )   ((void *)   (((char*) n) + PREFIX_SIZE))
#define PTR_TO_BLOCK( p )   ((block_t*) (((char*) p) - PREFIX_SIZE))

#ifndef MAX   
#define MAX(x,y) ((x) > (y) ? (x) : (y))
#endif

#if DEBUG_BLOCK == 1
#define SET_BLOCK_MAGIC( b )       { (b)->magic = BLOCK_MAGIC; (b)->__tmp__ = 0; }
#define CHECK_BLOCK_MAGIC( b, m, r ) if ( (b)->magic != BLOCK_MAGIC ) \
                                   { fprintf( stderr, "(lkmalloc) %s : block corrupted\n", m ); r; }
#else
#define SET_BLOCK_MAGIC( b )      
#define CHECK_BLOCK_MAGIC( b, m, r )
#endif

#if DEBUG_LKMALLOC == 1
#define CHECK_LKMALLOC      check_lkmalloc()
#else
#define CHECK_LKMALLOC      
#endif

/*****************************************************************
 *****************************************************************
 **
 ** forward declarations and inline functions
 **
 *****************************************************************
 *****************************************************************/


/*
 * initialise/finish malloc
 */
static void lkmalloc_init     ();
static void lkmalloc_finish   ();

/*
 * heap management
 */
static void     init_heap     ( heap_t * heap );
static void     insert_heap   ( heap_t * heap );
static heap_t * get_free_heap ();

/*
 * allocation/deallocation
 */
void * lk_malloc  ( size_t size );
void * lk_calloc  ( size_t nmemb, size_t size);
void   lk_free    ( void * ptr  );
void * lk_realloc ( void * ptr, size_t size );

/********************************************
 * methods for middle (usual) sizes
 ********************************************/

/*
 * allocation and deallocation
 */

static block_t * middle_alloc ( heap_t * heap, internal_size_t size, unsigned sclass );
static void      middle_free  ( heap_t * heap, block_t * block );


/********************************************
 * methods for large (unmanaged) sizes
 ********************************************/

static block_t * vm_alloc   ( internal_size_t size );
static void      vm_dealloc ( block_t * block );

/*
 * trace memory allocation
 */
static void lkmalloc_init_trace   ( const char * name );
static void lkmalloc_trace        ();
static void lkmalloc_finish_trace ();

/*
 * print statistics
 */
static void lkmalloc_stat ();

/*
 * translate size into size-class
 */
static unsigned size_to_class  ( internal_size_t  size );

/*
 * check status of lkmalloc
 */
#if DEBUG_LKMALLOC == 1
static void check_lkmalloc ();
#endif

/*****************************************************************
 *****************************************************************
 **
 ** mutex functions
 **
 *****************************************************************
 *****************************************************************/

#if USE_SPINLOCKS == 1

/*
 * atomic exchange of data in a mutex
 * (copied from Hoard)
 */
inline int
interlocked_exch ( mutex_t m, int data )
{
#if defined(sparc)
    asm volatile ("swap %1,%0"
                  :"=r" (data)
                  :"r" (m), "0" (data)
                  : "memory");
    
#endif
#if defined(i386)
    asm volatile ("xchgl %0, %1"
                  : "=r" (data)
                  : "m" (m), "0" (data)
                  : "memory");
#endif
#if defined(ppc)
    int ret;
    asm volatile ("sync;"
                  "0:    lwarx %0,0,%1 ;"
                  "      xor. %0,%3,%0;"
                  "      bne 1f;"
                  "      stwcx. %2,0,%1;"
                  "      bne- 0b;"
                  "1:    sync"
                  : "=&r"(ret)
                  : "r"(m), "r"(data), "r"(m)
                  : "cr0", "memory");
#endif
#if !(defined(sparc) || defined(i386) || defined(ppc))
#error "LKmalloc does not include support for user-level locks for this platform."
#endif
    return data;
}

/*
 * lock a mutex and block until mutex is released
 */

#define MAXSPIN  4000

inline void
mutex_lock ( mutex_t m )
{
    int  cnt;

    do
    {
        cnt = MAXSPIN;

        while ( m == LOCKED )
        {
            cnt--;

            if ( cnt < 0 )
            {
                sleep(0);
                cnt = MAXSPIN;
            }
        }
    } while ( interlocked_exch( m, LOCKED ) == LOCKED );
}

inline void
mutex_unlock ( mutex_t m )
{
    m = ! LOCKED;
}

inline int
mutex_trylock ( mutex_t m )
{
    if ( m == LOCKED )
        return MUTEX_BUSY;
    else
    {
        mutex_lock( m );
        return ! MUTEX_BUSY;
    }
}

#endif

/*****************************************************************
 *****************************************************************
 **
 ** allocation of memory from system
 **
 *****************************************************************
 *****************************************************************/

/*
 * allocation from system via sbrk / mmap etc.
 */
static void *
system_alloc ( heap_t * heap, internal_size_t size )
{
    char            * p;
    block_t         * block, * remainder;
    internal_size_t   n;
    
#if USE_MALLOC == 1 || USE_SBRK == 1
    DEFINE_MUTEX( mutex );
#endif
    
    if ( size == 0 )
        return NULL;

    assert( heap != NULL );

    /* adjust n to be multiple of 8 */
    if ( IS_MISALIGNED( size ) )
        ALIGN( size );

    /*
     * use wilderness of heap to serve malloc request
     */

    block = heap->blocks.wilderness;

    while ( block != NULL )
    {
        if ( BLOCK_SIZE( block ) >= size )
        {
            /* split block to serve request
             * and prepend block for future requests (if large enough) */

            n = BLOCK_SIZE( block ) - size;

            if ( n >= MIN_BLOCK_SIZE )
            {
                /* split block */
                block_t * new_block = block;

                SET_SIZE( new_block, size );
                SET_TAME( new_block );
                
                block = (block_t*) (((char*) block) + size);
                SET_SIZE( block, n );
                block->hc_addr = new_block->hc_addr;
                SET_WILD( block );
                SET_FREE( block );
                
                if ( new_block != heap->blocks.wilderness )
                {
                    /* remove block and prepend to wilderness */
                    if ( new_block->next != NULL )
                        new_block->next->prev = new_block->prev;
                    new_block->prev->next = new_block->next;
                    
                    block->next = heap->blocks.wilderness;
                    block->prev = NULL;

                    heap->blocks.wilderness->prev = block;
                    heap->blocks.wilderness = block;
                }
                else
                {
                    /* copy links from old (new) block */
                    block->next = new_block->next;
                    block->prev = new_block->prev;

                    /* reset wilderness */
                    heap->blocks.wilderness = block;
                    if ( block->next != NULL )
                        block->next->prev = block;
                    /* prev is NULL !!! */
                }

                return new_block;
            }
            else
            {
                /* block is too small to serve for another request,
                 * treat it as internal fragmentation */
                
                if ( block->next != NULL )
                    block->next->prev = block->prev;
                if ( block->prev != NULL )
                    block->prev->next = block->next;
                else
                    heap->blocks.wilderness = block->next;

                return block;
            }
            
            break;
        }
        else
            block = block->next;
    }

    /*
     * no suitable block was found in the wilderness,
     * create a new block + padding, serve the request
     * and put remainder into wilderness
     */
    
#if USE_MMAP == 1 || USE_MALLOC == 1
    /*
     * use system-malloc to request a large memory-chunk
     */

    if ( size > sys_conf.strip_size )
    {
        fprintf( stderr, "(lkmalloc) memory request larger than strip size: %ld <-> %ld\n",
                 size, sys_conf.strip_size );
        ABORT;
    }
                 
    n = sys_conf.strip_size;
    
#if USE_MALLOC == 1
    /* call malloc in a thread-safe way */
    LOCK( mutex );
    p = (char*) malloc( n );
    UNLOCK( mutex );
    
    if ( p == (char*) NULL )
    {
        fprintf( stderr, "(lkmalloc) system_alloc : malloc failed (%s)\n",
                 strerror( errno ) );
        ABORT;
    }
#else /* USE_MMAP == 1 */
    /* map new heap with mmap */
    p = (char*) MMAP( 0, n, PROT_READ | PROT_WRITE, MAP_PRIVATE );
    
    if ( p == (char*) MAP_FAILED )
    {
        fprintf( stderr, "(lkmalloc) system_alloc : mmap failed (%s)\n",
                 strerror( errno ) );
        ABORT;
    }
#endif

#elif USE_SBRK == 1
    /* to be done */
#else
    fprintf( stderr, "(lkmalloc) system_alloc : no heap-allocation method defined (MALLOC,MMAP,SBRK)\n" );
    EXIT;
#endif

    /*
     * update statistics
     */

    heap->stat.used_mem += n;

    /*
     * form block and insert remainder to wilderness
     */

    block = (block_t *) p;
    SET_SIZE( block, size );
    block->hc_addr = heap;

    /* this is the first block in the data-segment and therefore has no
     * predecessor. by setting pred_used, we avoid coalescing with it */
    SET_PRED_USED( block );

    remainder = (block_t*) (p + size);
    SET_SIZE( remainder, n - size );
    remainder->hc_addr = heap;
    remainder->prev    = NULL;
    remainder->next    = heap->blocks.wilderness;
    SET_WILD( remainder );

    if ( remainder->next != NULL )
        remainder->next->prev = remainder;

    heap->blocks.wilderness = remainder;
    
    return block;
}



/********************************************************************
 ********************************************************************
 **
 ** malloc/free interface
 **
 ********************************************************************
 ********************************************************************/

void *
lk_malloc ( size_t size )
{
    block_t  * block = NULL;
    heap_t   * heap = NULL;
    unsigned   sclass = 0;
    
    if ( size == 0 )
        return NULL;

    if ( ! is_initialised )
        lkmalloc_init();

    CHECK_LKMALLOC;

    /* add management data for all sizes */
    size = size + PREFIX_SIZE;
        
    /* first check if size exceeds maximal managed size */
    if ( size > size_classes[ NO_OF_CLASSES - 1 ] )
    {
        block_t  * lblock;
        
        /* request is managed by the operating system */
        lblock = vm_alloc( size );

        /* block has no heap/container, but remember size */
        SET_SIZE( lblock, size );
        lblock->hc_addr = NULL;
        
        return BLOCK_TO_PTR( lblock );
    }
    else
    {
        /* adjust to hold extra data */
        size = MAX( size, sizeof(block_t) + POSTFIX_SIZE );
        
        /* trim size to be multiple of 8 */
        if ( IS_MISALIGNED( size ) )
            ALIGN( size );

        sclass = size_to_class( size );
    }
    
    /*
     * get heap for this thread
     */
    
    /* get private heap of thread or set new heap */
    heap = GET_THR_HEAP;
    if ( heap == NULL )
    {
        /* look for free heap or create new one */
        if ((heap = get_free_heap()) == NULL)
        {
            LOCK( global_heap.mutex );
            heap = (heap_t *) system_alloc( (heap_t*) & global_heap, sizeof(heap_t) );
            UNLOCK( global_heap.mutex );
            
            init_heap( heap );

            /* lock heap BEFORE inserting so we can be sure, no other grabs it */
            LOCK( heap->mutex );
            insert_heap( heap );
        }
        else
            LOCK( heap->mutex );

        /* set this heap to be the thread-private heap */
        SET_THR_HEAP( heap );
    }
    else
        LOCK( heap->mutex );

    assert( heap != NULL );

    /*
     * get free block
     */

    block = middle_alloc( heap, size, sclass );
    
    assert( block != NULL );

    SET_BLOCK_MAGIC( block );
    
    /* update statistics */
    heap->stat.mem_in_use += BLOCK_SIZE( block );
    heap->stat.max_in_use  = MAX( heap->stat.max_in_use, heap->stat.mem_in_use );
    heap->stat.allocated  += BLOCK_SIZE( block );
    
    if ( trace_mem )
        lkmalloc_trace();

    UNLOCK( heap->mutex );

    CHECK_LKMALLOC;

    return BLOCK_TO_PTR( block );
}

void *
lk_calloc ( size_t nmemb, size_t size )
{
    void * p;
    
    if ((nmemb == 0) || (size == 0))
        return NULL;
    
    p = lk_malloc( nmemb * size );

    /* memset to zero */
    if ( p != NULL )
        memset( p, 0, nmemb * size );

    return p;
}

void
lk_free ( void * ptr )
{
    block_t  * block;

    if ( ptr == NULL )
        return;

    /* should at least have called malloc for this ptr, if not : ERROR */
    if ( ! is_initialised )
    {
        fprintf( stderr, "(lkmalloc) free : lkmalloc not initialised\n" );
        return;
    }

    CHECK_LKMALLOC;

    block = PTR_TO_BLOCK( ptr );

    CHECK_BLOCK_MAGIC( block, "free", return );
    
    /* test whether block is managed by operating system */
    if ( block->hc_addr == NULL )
        vm_dealloc( block );
    else
    {
        /* now free block */
        heap_t * heap = (heap_t*) block->hc_addr;
        
        assert( heap != NULL );
        
        /* lock mutex here, cause middle_free is also called
         * from small_free and mutex is already locked there
         */
        
        LOCK( heap->mutex );
        
        /* update statistics */
        heap->stat.mem_in_use -= BLOCK_SIZE( block );
        
        middle_free( heap, block );
        
        UNLOCK( heap->mutex );
    }

    CHECK_LKMALLOC;
}

void *
lk_realloc ( void * ptr, size_t size )
{
    if ( ptr == NULL )
        return lk_malloc( size );
    else
    {
        block_t        * block    = PTR_TO_BLOCK( ptr );
        void           * newptr   = NULL;
        internal_size_t  new_size = size + PREFIX_SIZE;
        internal_size_t  old_size = BLOCK_SIZE( block );

        CHECK_BLOCK_MAGIC( block, "realloc", return NULL );

        /*
         * adjust new size to get correct size of block
         * corresponding to this new size
         */

        if ( new_size <= size_classes[ NO_OF_CLASSES - 1 ] )
            new_size  = MAX( new_size, sizeof(block_t) + POSTFIX_SIZE );

        if ( IS_MISALIGNED( size ) )
            ALIGN( size );

        /* if block has same size, just return */
        if ( new_size == old_size )
            return ptr;

        /* create new block and copy data */
        newptr = lk_malloc( size );

        old_size -= PREFIX_SIZE;

        if ( size > old_size )
            memcpy( newptr, ptr, old_size );
        else
            memcpy( newptr, ptr, size );

        lk_free( ptr );

        return newptr;
    }
}

/********************************************************************
 ********************************************************************
 **
 ** misc. functions for heap-management
 **
 ********************************************************************
 ********************************************************************/

/*
 * init/finish heap-manager
 */

static void
lkmalloc_init ()
{
    DEFINE_MUTEX( mutex );

    if ( MUTEX_LOCKED( mutex ) )
    {
        fprintf( stderr, "(lkmalloc) lkmalloc_init : mutex locked\n" );
        LOCK( mutex );
    }

    if ( ! is_initialised )
    {
        char * value;

        sys_conf.pagesize = getpagesize();

        /* adjust strip size to be multiple of pagesize */
        if ( sys_conf.strip_size % sys_conf.pagesize != 0 )
            sys_conf.strip_size += sys_conf.pagesize - (sys_conf.strip_size % sys_conf.pagesize);

        /* init tsd-key */
        TSD_KEY_INIT;
        
        /* re-initialise global heap */
        init_heap( & global_heap );

        /*
         * setup memory trace
         */

        value = getenv( "LKMALLOC_TRACE" );
        
        if (( value != NULL ) && ((value[0] == '1') || (value[0] == '2')))
        {
            /* set trace-type */
            if ( value[0] == '1' )
                trace.type = TRACE_STEPS;
            else
                trace.type = TRACE_ALLOCATION;
            
            /* get size of trace */

            if ((value = getenv( "LKMALLOC_TRACE_SIZE" )) != NULL )
            {
                if ( strcmp( value, "all" ) == 0 )
                    trace.size = 0;
                else
                {
                    trace.size = atoi( value );

                    /* round to next multiple of 8 */
                    trace.size += 8 - (trace.size % 8);
                }
            }

            /* get name of tracefile and initialise */

            if ((value = getenv( "LKMALLOC_TRACE_FILE" )) != NULL )
                lkmalloc_init_trace( value );
            else
                lkmalloc_init_trace( DEFAULT_TRACE_FILE );
        }
        
        /*
         * register cleanup-functions
         */

        if ( atexit( lkmalloc_finish ) != 0 )
            fprintf( stderr, "(lkmalloc) init : error in atexit (%s)\n", strerror( errno ) );
        
        if ( atexit( lkmalloc_finish_trace ) != 0 )
            fprintf( stderr, "(lkmalloc) init : error in atexit (%s)\n", strerror( errno ) );

        if ((value = getenv( "LKMALLOC_STAT" )) != NULL )
        {
            heap_stat_lvl = atoi( value );

            if ( heap_stat_lvl > 0 )
            {
                if ( atexit( lkmalloc_stat ) != 0 )
                    fprintf( stderr, "(lkmalloc) init : error in atexit (%s)\n", strerror( errno ) );
            }
        }

        is_initialised = true;
    }
    
    UNLOCK( mutex );
}

/*
 * end it all
 */
static void
lkmalloc_finish ()
{
    /*
     * clean will be called at the end and then
     * the memory will be deallocated by the system
     * so why bother ???
     */
}

/*
 * initialise a heap
 */
static void
init_heap ( heap_t * heap )
{
    unsigned  i;
#if USE_SPINLOCKS == 0
    int       status;
#endif
    
    heap->next    = NULL;
    heap->id      = heap_id++;

    for ( i = 0; i <= NO_OF_CLASSES; i++ )
    {
        heap->blocks.blocks[i] = NULL;
        
        heap->stat.free[i]  = 0;
        heap->stat.used[i]  = 0;
    }

    heap->blocks.wilderness = NULL;

    heap->stat.used_mem   = 0;
    heap->stat.mem_in_use = 0;
    heap->stat.max_in_use = 0;
    heap->stat.allocated  = 0;

    heap->used = true;

    MUTEX_INIT( heap->mutex, status );
}

/*
 * inserts new heap into heap-list
 */
static void
insert_heap ( heap_t * heap )
{
    LOCK( heap_list.mutex );
    
    /* prepend into list */
    heap->next      = heap_list.heaps;
    heap_list.heaps = heap;

    heap_list.nheaps++;
    
    UNLOCK( heap_list.mutex );
}

/*
 * return unused/free heap from heap-list
 */
static heap_t *
get_free_heap ()
{
    heap_t  * heap = heap_list.heaps;

    while ( heap != NULL )
    {
        if ( ! heap->used )
        {
            heap->used = true;
            return heap;
        }
        else
            heap = heap->next;
    }

    /* no free heap found */
    return NULL;
}

/*
 * translate size into size-class
 */
static unsigned
size_to_class ( internal_size_t  size )
{
    if ( size <= SMALL_SIZE )
        return ((unsigned) ((size >> 3) - 1));
    else
    {
#if 1
        /*
         * use binary search for size-class
         */
        unsigned  lb = (SMALL_SIZE >> 3)-1;
        unsigned  ub = NO_OF_CLASSES;
        unsigned  split;
        
        while ( ub - lb > 1 )
        {
            split = (ub + lb) / 2;
            
            if (( size_classes[ split-1 ] < size ) && (size <= size_classes[ split ]))
                return split;
            
            if ( size_classes[ split ] < size )
                lb = split;
            else
                ub = split;
        }
        
        return ub;
#else
        /*
         * use linear search for size-class
         */
        unsigned  sclass = SMALL_SIZE >> 3;

        while ((sclass < NO_OF_CLASSES) && (size_classes[ sclass ] < size))
            sclass++;

        return sclass;
#endif
    }
}

/********************************************************************
 ********************************************************************
 **
 ** allocation and deallocation of middle-sized blocks
 **
 ********************************************************************
 ********************************************************************/

static block_t *
middle_alloc ( heap_t * heap, internal_size_t size, unsigned sclass )
{
    block_t  * block = NULL;
    unsigned   i;
    
    assert( heap != NULL );

    /*
     * look for an old block, and begin search with
     * given size-class
     */

    for ( i = sclass; i <= NO_OF_CLASSES; i++ )
    {
        block = heap->blocks.blocks[ i ];
        
        while ( block != NULL )
        {
            if ( BLOCK_SIZE( block ) >= size )
            {
                /* remove block from list */
                if ( block->next != NULL )
                    block->next->prev = block->prev;

                if ( block->prev != NULL )
                    block->prev->next = block->next;
                else
                    heap->blocks.blocks[ i ] = block->next;
            
                block->next = NULL;
                block->prev = NULL;

                /* update statistics */
                heap->stat.free[ i ]--;
            
                /*
                 * check if block is from a different size-class
                 * and if the remainder of the block can be used
                 * by another sclass
                 */
                
                if (( i != sclass ) && ( BLOCK_SIZE( block ) - size > MIN_BLOCK_SIZE))
                {
                    /*
                     * split block and put remainder into another size-class
                     */

                    block_t  * new_block = (block_t*) (((char*) block) + size);
                    unsigned   s;
                    
                    SET_SIZE( new_block, BLOCK_SIZE( block ) - size );
                    SET_FREE( new_block );
                    SET_TAME( new_block );
                    SET_BLOCK_MAGIC( new_block );

                    new_block->hc_addr = heap;
                    
                    /* copy size-information to end of new block */
                    SET_EOB_SIZE( new_block );

                    /* put block into proper size-class */
                    s = size_to_class( BLOCK_SIZE( new_block ) );

                    new_block->prev = NULL;
                    new_block->next = heap->blocks.blocks[ s ];
                    if ( new_block->next != NULL )
                        new_block->next->prev = new_block;
                    heap->blocks.blocks[ s ] = new_block;
                    heap->stat.free[ s ]++;

                    /* adjust blocksize of old block */
                    SET_SIZE( block, size );
                }
                else
                    sclass = i;

                /* finish search */
                i = NO_OF_CLASSES+1;
                
                break;
            }
            else
                block = block->next;
        }
    }

    /*
     * allocate block from system
     */

    if ( block == NULL )
    {
        block = (block_t*) system_alloc( heap, size );
        assert( block != NULL );

        SET_SIZE( block, size );
        block->hc_addr = heap;
    }

    /* update statistics */
    heap->stat.used[ sclass ]++;

    SET_USED( block );

    /* set boundary tag in successor */
    if ( ! IS_WILD( block ) )
    {
        block_t * succ = (block_t*) (((char*) block) + BLOCK_SIZE( block ));

        SET_PRED_USED( succ );
    }

    SET_BLOCK_MAGIC( block );

    return block;
}
               
static void
middle_free ( heap_t * heap, block_t * block )
{
    unsigned  sclass;

    assert((heap != NULL) && ( block != NULL ));

    sclass = size_to_class( BLOCK_SIZE( block ) );
    heap->stat.used[ sclass ]--;

    /*
     * try to coalesce block with succ
     */

    if ( ! IS_WILD( block ) )
    {
        block_t * succ = (block_t*) (((char*) block) + BLOCK_SIZE( block ));

        if ( ! IS_USED( succ ) )
        {
            if ( ! IS_WILD( succ ) )
            {
                /* remove successor from his list */
                sclass = size_to_class( BLOCK_SIZE( succ ) );

                if ( succ->next != NULL ) succ->next->prev = succ->prev;
                if ( succ->prev != NULL ) succ->prev->next = succ->next;

                if ( heap->blocks.blocks[ sclass ] == succ )
                    heap->blocks.blocks[ sclass ] = succ->next;

                heap->stat.free[ sclass ]--;
            }
            else
            {
                /* coalesced block is now in wilderness:
                 * set status and copy links */
                
                SET_WILD( block );
                block->next = succ->next;
                block->prev = succ->prev;

                if ( heap->blocks.wilderness == succ )
                    heap->blocks.wilderness = block;
            }
            
            /* coalesce */
            SET_SIZE( block, BLOCK_SIZE( block ) + BLOCK_SIZE( succ ) );
        }
    }
    else
    {
        /*
         * reinsert block into wilderness
         */

        block->next = heap->blocks.wilderness;
        block->prev = NULL;
        
        if ( block->next != NULL )
            block->next->prev = block;
        
        heap->blocks.wilderness = block;
    }

    /*
     * try to coalesce with predecessor
     */

    if ( ! IS_PRED_USED( block ) )
    {
        internal_size_t  psize = PRED_SIZE( block );
        block_t        * pred  = (block_t*) (((char*) block) - psize);
        
        /* remove predecessor from his list */
        sclass = size_to_class( psize );
        
        if ( pred->next != NULL ) pred->next->prev = pred->prev;
        if ( pred->prev != NULL ) pred->prev->next = pred->next;
        
        if ( heap->blocks.blocks[ sclass ] == pred )
            heap->blocks.blocks[ sclass ] = pred->next;

        heap->stat.free[ sclass ]--;
        
        /* coalesce */
        SET_SIZE( pred, BLOCK_SIZE( block ) + psize );

        /* adjust status and links if block is inside wilderness */
        if ( IS_WILD( block ) )
        {
            SET_WILD( pred );
            pred->next = block->next;
            pred->prev = block->prev;

            if ( heap->blocks.wilderness == block )
                heap->blocks.wilderness = pred;
        }

        /* replace block */
        block = pred;
    }
    
    /*
     * put block back into wilderness or size-classes
     */

    if ( IS_WILD( block ) )
    {
        /* adjust links in wilderness-list */
        if ( block->prev != NULL ) block->prev->next = block;
        if ( block->next != NULL ) block->next->prev = block;
    }
    else
    {
        block_t * succ;
        
        sclass = size_to_class( BLOCK_SIZE( block ) );

        /* put block at begining of list */
        block->next = heap->blocks.blocks[ sclass ];

        if ( block->next != NULL )
            block->next->prev = block;

        block->prev = NULL;
        heap->blocks.blocks[ sclass ] = block;
        
        heap->stat.free[ sclass ]++;

        /* copy size-information to end of block */
        SET_EOB_SIZE( block );

        /* adjust boundary tag */
        succ = (block_t*) (((char*) block) + BLOCK_SIZE( block ));
        SET_PRED_FREE( succ );
    }
    
    SET_FREE( block );
}
               
/********************************************************************
 ********************************************************************
 **
 ** handling of unmanaged chunks
 **
 ********************************************************************
 ********************************************************************/

/*
 * allocate block directly
 */
static block_t *
vm_alloc ( internal_size_t size )
{
#if USE_MMAP == 1
    char  * p;
#endif
    
    /* round size up to next multiple of pagesize */
    if ( size % sys_conf.pagesize != 0 )
        size += sys_conf.pagesize - (size % sys_conf.pagesize);
    
#if USE_MALLOC == 1
    return (block_t*) malloc( size );
#elif USE_MMAP == 1
    p = (char*) MMAP( 0, size, PROT_READ | PROT_WRITE, MAP_PRIVATE );

    if ( p == (char*) MAP_FAILED )
    {
        fprintf( stderr, "(lkmalloc) vm_alloc : mmap failed (%s)\n",
                 strerror( errno ) );
        ABORT;
    }

    return (block_t*) p;
#else
    fprintf( stderr, "(lkmalloc) vm_alloc : no supported allocation method specified\n" );
    return NULL;
#endif
}

static void
vm_dealloc ( block_t * block )
{
#if USE_MALLOC == 1
    free( block );
#else
#if USE_MMAP == 1
    if ( munmap( (char*) block, BLOCK_SIZE( block ) ) != 0 )
    {
        fprintf( stderr, "(lkmalloc) vm_dealloc : munmap failed (%s)\n",
                 strerror( errno ) );
        ABORT;
    }
#endif
#endif
}

/********************************************************************
 ********************************************************************
 **
 ** get info and set options from/for malloc
 **
 ********************************************************************
 ********************************************************************/

/*
 * report memory usage
 */

struct ul_mallinfo
lkmallinfo ( void )
{
    struct ul_mallinfo  mi;
    internal_size_t     total_size;
    internal_size_t     used_size;
    internal_size_t     free_size;
    heap_t            * heap;

    LOCK( global_heap.mutex );
    
    total_size = global_heap.stat.used_mem;
    used_size  = global_heap.stat.mem_in_use;
    free_size  = global_heap.stat.used_mem - global_heap.stat.mem_in_use;

    UNLOCK( global_heap.mutex );

    LOCK( heap_list.mutex );

    heap = heap_list.heaps;
    while ( heap )
    {
        LOCK( heap->mutex );
        
        total_size += heap->stat.used_mem;
        used_size  += heap->stat.mem_in_use;
        free_size  += heap->stat.used_mem - heap->stat.mem_in_use;
            
        UNLOCK( heap->mutex );
            
        heap = heap->next;
    }
    
    UNLOCK( heap_list.mutex );
    
    mi.arena    = total_size;      /* total space allocated from system */
    mi.ordblks  = 0;               /* number of non-inuse chunks */
    mi.smblks   = 0;               /* unused -- always zero */
    mi.hblks    = 0;               /* number of mmapped regions */
    mi.hblkhd   = 0;               /* total space in mmapped regions */
    mi.usmblks  = 0;               /* unused -- always zero */
    mi.fsmblks  = 0;               /* unused -- always zero */
    mi.uordblks = used_size;       /* total allocated space */
    mi.fordblks = free_size;       /* total non-inuse space */
    mi.keepcost = 0;               /* top-most, releasable (via malloc_trim) space */

    return mi;
}

/*
 * wrapper for mallinfo
 */

#if OVERLOAD_MALLOC == 1
struct mallinfo
mallinfo ( void )
{
    struct mallinfo     mi;
    struct ul_mallinfo  ul_mi = rmallinfo();

    mi.arena    = ul_mi.arena;
    mi.ordblks  = ul_mi.ordblks;
    mi.smblks   = ul_mi.smblks; 
    mi.hblks    = ul_mi.hblks;  
    mi.hblkhd   = ul_mi.hblkhd; 
    mi.usmblks  = ul_mi.usmblks;
    mi.fsmblks  = ul_mi.fsmblks;
    mi.uordblks = ul_mi.uordblks;
    mi.fordblks = ul_mi.fordblks;
    mi.keepcost = ul_mi.keepcost;

    return mi;
}
#endif


/********************************************************************
 ********************************************************************
 **
 ** tracing
 **
 ********************************************************************
 ********************************************************************/

/*
 * allocation trace
 */

static void
lkmalloc_init_trace ( const char * name )
{
    assert( name != NULL );

    if ((trace.fd  = open( name, O_WRONLY | O_CREAT | O_TRUNC,
                           S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH )) == -1)
    {
        fprintf( stderr, "(lkmalloc) init_trace : could not open \"%s\" (%s)\n",
                 name, strerror( errno ) );
        return;
    }
    
    trace.old_used  = 0;
    trace.old_free  = 0;
    trace.old_alloc = 0;
    trace.step      = 0;

    trace_mem = true;
}

/*
 * get status and write trace-entry
 */

static void
lkmalloc_trace ()
{
    internal_size_t  used, free, alloc;
    
    if ( ! trace_mem )
        return;

    if ( trace.size == 0 )
    {
        heap_t  * heap = heap_list.heaps;
        
        used  = global_heap.stat.mem_in_use;
        free  = global_heap.stat.used_mem - global_heap.stat.mem_in_use;
        alloc = global_heap.stat.allocated;

        while ( heap != NULL )
        {
            used  += heap->stat.mem_in_use;
            free  += (heap->stat.used_mem - heap->stat.mem_in_use);
            alloc += heap->stat.allocated;
            heap = heap->next;
        }
    }
    else
    {
        unsigned   sclass = size_to_class( trace.size );
        heap_t   * heap   = heap_list.heaps;
        
        used  = global_heap.stat.used[ sclass ] * size_classes[ sclass ];
        free  = global_heap.stat.free[ sclass ] * size_classes[ sclass ];
        alloc = global_heap.stat.allocated;

        while ( heap != NULL )
        {
            used  += heap->stat.used[ sclass ] * size_classes[ sclass ];
            free  += heap->stat.free[ sclass ] * size_classes[ sclass ];
            alloc += heap->stat.allocated;
            heap = heap->next;
        }
    } 

    used /= 1024;
    free /= 1024;

    if ( trace.type == TRACE_STEPS )
    {
        if (((used > 0) || (free > 0)) &&
            ((used != trace.old_used) || (free != trace.old_free)))
        {
            char  buffer[256];
            
            sprintf( buffer, "%ld    %ld    %ld\n", (long) trace.step, (long) used, (long) free );
            write( trace.fd, buffer, strlen(buffer) );
            trace.old_used  = used;
            trace.old_free  = free;
        }

        trace.step++;
    }
    else if ( trace.type == TRACE_ALLOCATION )
    {
        alloc /= 1024;
        
        if (( alloc != trace.old_alloc ) &&
            ( ((used > 0) || (free > 0)) &&
              ((used != trace.old_used) || (free != trace.old_free))))
        {
            char  buffer[256];
            
            sprintf( buffer, "%.3f    %ld    %ld\n", ((float)alloc)/1024.0, used, free );
            write( trace.fd, buffer, strlen(buffer) );
            trace.old_used  = used;
            trace.old_free  = free;
            trace.old_alloc = alloc;
        }
    }
}

/*
 * finish tracing
 */

static void
lkmalloc_finish_trace ()
{
    if ( trace_mem )
    {
        close( trace.fd );
        trace_mem = false;
    }
}


/********************************************************************
 ********************************************************************
 **
 ** statistical methods
 **
 ********************************************************************
 ********************************************************************/

#define BYTE_TO_MB( n ) (((double)(n)) / (1000.0 * 1000.0))

/*
 * print statistics
 */

static void
print_heap_stat ( heap_t  * heap )
{
    unsigned         i;
    unsigned         nwild;
    internal_size_t  swild;
    block_t        * block;
    
    if ( heap_stat_lvl <= 1 )
        return;
    
    /*
     * statistic for small-usage
     */

    if ( heap_stat_lvl >= 3 )
    {
        fprintf( stderr, " sizeclass |  # used    |   # free   |     kB     |    comment    \n" );
        fprintf( stderr, "-----------+------------+------------+------------+---------------\n" );
        
        for ( i = 0; i <= NO_OF_CLASSES; i++ )
        {
            block = heap->blocks.blocks[i];
            
            if ( block != NULL )
            {
                internal_size_t  mem = 0;

                while ( block != NULL )
                {
                    mem += BLOCK_SIZE( block );
                    block = block->next;
                }

                fprintf( stderr, "%10ld | %10ld | %10ld | %10ld |",
                         (long) size_classes[ i ],
                         (long) heap->stat.used[i],
                         (long) heap->stat.free[i],
                         (long) mem / 1024 );

                if ( heap->stat.used[i] != 0 )
                    fprintf( stderr, " %ld chunk(s) alive", heap->stat.used[i] );
                
                fprintf( stderr, "\n" );
            }
        }
    }

    /*
     * sum up wilderness
     */

    nwild = 0;
    swild = 0;
    block = heap->blocks.wilderness;
    
    while ( block != NULL )
    {
        nwild++;
        swild += BLOCK_SIZE( block );
        block = block->next;
    }
    
    /*
     * output
     */

    if ((heap->stat.mem_in_use != 0) || (heap->stat.used_mem != 0))
        fprintf( stderr, "    mem in use / allocated   = %.2f MB / %.2f MB\n",
                 BYTE_TO_MB(heap->stat.mem_in_use), BYTE_TO_MB(heap->stat.used_mem) );
    
    if ( nwild != 0 )
        fprintf( stderr, "    wilderness blocks/size   = %d / %.2f MB\n",
                 nwild, BYTE_TO_MB(swild) );
}

static void
lkmalloc_stat ()
{
    heap_t * heap;

    if ( heap_stat_lvl == 0 )
        return;
    
    if ( ! is_initialised )
    {
        fprintf( stderr, "(lkmalloc) stat : heap not initialised\n" );
        return;
    }

    fprintf( stderr, "(lkmalloc) lkmalloc version %s", LKMALLOC_VERSION );

#if USE_MMAP == 1
    fprintf( stderr, " (mmap)\n" );
#elif USE_MALLOC == 1
    fprintf( stderr, " (malloc)\n" );
#elif USE_SBRK == 1
    fprintf( stderr, " (sbr)\n" );
#else
    fprintf( stderr, "\n" );
#endif
    
    /*
     * print statstics about each heap
     */

    if ( heap_stat_lvl >= 2 )
    {
        print_heap_stat( & global_heap );
        
        heap = heap_list.heaps;
        while ( heap != NULL )
        {
            fprintf( stderr, "(lkmalloc) heap %ld :\n", (long) heap->id );
            print_heap_stat( heap );
            heap = heap->next;
        }
    }
    
    if ( heap_stat_lvl >= 1 )
    {
        internal_size_t  total_size = global_heap.stat.used_mem;
        internal_size_t  max_use    = global_heap.stat.max_in_use;
        internal_size_t  allocated  = global_heap.stat.allocated;

        heap = heap_list.heaps;
        while ( heap != NULL )
        {
            total_size += heap->stat.used_mem;
            max_use    += heap->stat.max_in_use;
            allocated  += heap->stat.allocated;
            heap = heap->next;
        }
        
        fprintf( stderr, "(lkmalloc) global stat:\n" );

        if ( total_size > 0 )
            fprintf( stderr, "    mem used by lkmalloc    = %.2f MB\n", BYTE_TO_MB(total_size) );
        
        if ( max_use > 0 )
            fprintf( stderr, "    mem used by app (frag) = %.2f MB (%.2f)\n",
                     BYTE_TO_MB(max_use), ((double) total_size) / ((double) max_use) );

        if ( max_use > 0 )
            fprintf( stderr, "    total allocation       = %.2f MB\n", BYTE_TO_MB(allocated) );
    }
}

#if DEBUG_LKMALLOC == 1
static void
check_lkmalloc ()
{
    /*
     * check all heaps, all containers and all blocks
     */

    uint      i;
    heap_t  * heap = heap_list.heaps;

    while ( heap != NULL )
    {
        /*
         * check wilderness in heap for errors
         */

        block_t * block = heap->blocks.wilderness;

        while ( block != NULL )
        {
            if ( ! IS_WILD( block ) )
                fprintf( stderr, "(check_lkmalloc) wilderness block is marked as tame\n" );

            block = block->next;
        }

        /*
         * check tame blocks for errors
         */
        
        for ( i = 0; i <= NO_OF_CLASSES; i++ )
        {
            block_t * block = heap->blocks.blocks[i];
            
            while ( block != NULL )
            {
#if DEBUG_BLOCK == 1
                if ( block->magic != BLOCK_MAGIC )
                {
                    fprintf( stderr, "(check_lkmalloc) middle block is corrupted\n" );
                    continue;
                }

                if ( block->__tmp__ != 0 )
                {
                    fprintf( stderr, "(check_lkmalloc) middle block is corrupted\n" );
                    continue;
                }
#endif
                block = block->next;
            }
        }
        
        heap = heap->next;
    }
}
#endif

#ifdef __cplusplus
}
#endif

/********************************************************************
 ********************************************************************
 **
 ** C++ memory management
 **
 ********************************************************************
 ********************************************************************/

/*
 * overload global new/delete
 */
  
#if defined(__cplusplus) && OVERLOAD_NEW == 1

#include <new>

void * operator new      ( size_t   n ) throw (std::bad_alloc) { return lk_malloc( n ); }
void * operator new[]    ( size_t   n ) throw (std::bad_alloc) { return lk_malloc( n ); }
void   operator delete   ( void   * p ) throw ()               {        lk_free( p ); }
void   operator delete[] ( void   * p ) throw ()               {        lk_free( p ); }

#endif  /* defined(__cplusplus) && OVERLOAD_NEW == 1 */
