//
//  Project   : libnet
//  File      : T2DFrontier.cc
//  Author    : Sabine Gutsch / Ronald Kriemann
//  Purpose   : class for a frontier (separates two cells) in 2d
//
// arch-tag: fb4f7b47-f7d0-4ed4-968a-4e513cf2ff8b
//

#include "T2DFrontier.hh"

//
// constructors
//
T2DFrontier::T2DFrontier ( T2DFrontier *parent, uint ftype ) 
    : TFrontier<2>( parent )
{
    arrNodes  = NULL;
    noOfNodes = 0;

    initNodePtr( 3 );

    arrNodes[0] = NULL;
    arrNodes[2] = NULL;
    arrNodes[1] = NULL;

    if (parent == NULL) 
        frontierType = ftype;
    else 
        frontierType = parent->getFrontierType();
}

T2DFrontier::T2DFrontier ( T2DFrontier *parent,
                           T2DVertex *vertex0, T2DVertex* vertex2, 
                           uint ftype ) 
        : TFrontier<2>( parent ) 
{
    arrNodes  = NULL;
    noOfNodes = 0;

    initNodePtr( 3 );

    arrNodes[0] = vertex0;
    arrNodes[2] = vertex2;
    arrNodes[1] = NULL;

    vertex0->addReference();
    vertex2->addReference();

    fSize = (*vertex2 - *vertex0).euclidNorm();

    if (parent == NULL) 
        frontierType = ftype;
    else 
        frontierType = parent->getFrontierType();
}

//
// destructor
//
T2DFrontier::~T2DFrontier ()
{
    uint        i;
    TVertex<2>  *pVertex;

    if (arrNodes != NULL)
    {
        for (i = 0; i < noOfNodes; i++)
        {
            pVertex = arrNodes[i];
            
            if ((pVertex != NULL) && (pVertex->delReference())) delete pVertex;
        }// for

        delete arrNodes;
    }// if
}

//
// init memory for used arrays
//
void 
T2DFrontier::initNodePtr ( uint n )
{
    uint  i;

    if (n == noOfNodes) 
        return;

    if (arrNodes != NULL) 
        delete arrNodes;

    if (n == 0)
    {
        arrNodes  = NULL;
        noOfNodes = 0;
    }// if
    else
    {
        arrNodes  = new TVertex<2> * [n];
        noOfNodes = n;

        for (i = 0; i < noOfNodes; i++)
            arrNodes[i] = NULL;
    }// else
}

//
// make the midnode of the frontier
//
void 
T2DFrontier::buildMidNode () 
{
    if (hasMidNode())
    {
        cerr << "T2DFrontier::buildMidNode () : already have midnode" << endl;
        return;
    }// if

    arrNodes[1] = new T2DVertex( (*arrNodes[0]) + (((*arrNodes[2]) - (*arrNodes[0]))*0.5), 
                                 frontierType );
}

//
// did we set the midnode
//
const bool 
T2DFrontier::hasMidNode () const 
{ 
    if (arrNodes[1] != NULL)
        return TRUE;
    else
        return FALSE;
}

//
// return frontierType
//
const uint 
T2DFrontier::getFrontierType () const 
{ 
    return frontierType; 
}

//
// return number of nodes
//
const uint 
T2DFrontier::getNoOfNodes () const
{
    return noOfNodes;
}

//
// return i'th node
//
TVertex<2>* 
T2DFrontier::getNode ( uint i ) const
{
    if (i < noOfNodes)
        return arrNodes[i];
    else
    {
        cerr << "T2DTFrontier::getNode (" << i << ") : no such node" << endl;
        return NULL;
    }// else
}

//
// return i'th node with respect to the view of the given cell
//
TVertex<2>* 
T2DFrontier::getNode ( const TCell<2> *cell, uint i ) const
{
    if (i < noOfNodes)
    {
        if (cell == pCells[0])
            return arrNodes[i];
        else if (cell == pCells[1])
            return arrNodes[noOfNodes-1-i];
        else
        {
            cerr << "T2DFrontier::getNode (TCell, " << i << ") : no such cell" << endl;
            return NULL;
        }// else
    }// if
    else
    {
        cerr << "T2DFrontier::getNode (TCell, " << i << ") : no such node" << endl;
        return NULL;
    }// else
}

//
// build the child(ren)
//
void 
T2DFrontier::buildChild () 
{
    if (noOfChildren != 0)
    {
        cout << "T2DFrontier::buildChild () : children exist" << endl;
        return;
    }// if
    else if (hasMidNode())
    {
        initChildPtr( 2 );
        arrChildren[0] = new T2DFrontier( this, arrNodes[0], arrNodes[1] );
        arrChildren[1] = new T2DFrontier( this, arrNodes[1], arrNodes[2] );
    }// else if
    else
    {
        initChildPtr( 1 );
        //        arrChildren[0] = new T2DFrontier( this, arrNodes[0], arrNodes[2] );
        // experimental
        arrChildren[0] = this;
        addReference();
    }// else
}
