//
//  Project  : libnet
//  File     : T3DTetrahedron.cc
//  Author   : Ronald Kriemann
//  Purpose  : class of cell which implements a regular triangulation in 3D
//
// arch-tag: 8b188775-28ed-4598-b76c-45516836490f
//

#include <string.h>
#include "T3DTetrahedron.hh"

// defines for nodeposition in ordered array 
// (double-indices for midnodes)
#define  _0_   0
#define  _1_   1
#define  _2_   2
#define  _3_   3
#define  _01_  4
#define  _12_  5
#define  _20_  6
#define  _03_  7
#define  _13_  8
#define  _23_  9

// defines for frontiers (which nodes)
#define _20_0_01_   0
#define _01_1_12_   1
#define _12_2_20_   2
#define _20_01_12_  3
#define _12_1_13_   4
#define _13_3_23_   5
#define _23_2_12_   6
#define _12_13_23_  7
#define _03_0_20_   8
#define _20_2_23_   9
#define _23_3_03_  10
#define _03_20_23_ 11
#define _01_0_03_  12
#define _03_3_13_  13
#define _13_1_01_  14
#define _01_03_13_ 15
#define _0_1_2_    16
#define _0_2_3_    17
#define _1_3_2_    18
#define _0_3_1_    19
#define _20_01_03_ 20
#define _23_12_20_ 21
#define _03_13_23_ 22
#define _01_12_13_ 23
#define _01_03_12_ 24
#define _20_03_12_ 25
#define _23_03_12_ 26
#define _13_03_12_ 27
#define _01_20_13_ 28
#define _12_20_13_ 29
#define _23_20_13_ 30
#define _03_20_13_ 31
#define _01_20_23_ 32
#define _01_23_12_ 33
#define _01_23_03_ 34
#define _01_23_13_ 35

#define MAX_FRONTIER  36

// defines for used edges
#define _20_01_   0
#define _01_12_   1
#define _12_20_   2
#define _12_13_   3
#define _13_23_   4
#define _23_12_   5
#define _03_20_   6
#define _20_23_   7
#define _23_03_   8
#define _01_03_   9
#define _03_13_  10
#define _13_01_  11

// direction of nodes in cell
#define CW   1
#define CCW  2


ostream & operator << ( ostream &os, const TFrontier<3> &f )
{
    if (f.getNode(0) != NULL)
        os << *(f.getNode(0)) << " --> ";
    else
        cerr << " error(0) --> ";
    
    if (f.getNode(2) != NULL)
        os << *(f.getNode(2)) << " --> ";
    else
        cerr << " error(2) --> ";

    if (f.getNode(4) != NULL)
        os << *(f.getNode(4));
    else
        cerr << "error(4)";

    return os;
}

//
// constructor
//
T3DTetrahedron::T3DTetrahedron ( T3DCell *parent,
                                 T3DTriangle *frontier0,
                                 T3DTriangle *frontier1,
                                 T3DTriangle *frontier2,
                                 T3DTriangle *frontier3 )
        : T3DCell( parent )
{
    //
    // init frontier handling
    //
    initFrontierPtr( 4 );
    
    arrFrontiers[0] = frontier0;
    arrFrontiers[1] = frontier1;
    arrFrontiers[2] = frontier2;
    arrFrontiers[3] = frontier3;

    frontier0->addReference();
    frontier1->addReference();
    frontier2->addReference();
    frontier3->addReference();

    frontier0->setCell( this );
    frontier1->setCell( this );
    frontier2->setCell( this );
    frontier3->setCell( this );
    
    _idxOfLastCorner = idxOfLastCorner();

    fSize = calcSize();
}

//
// divide all marked frontiers before real refinement
//
void 
T3DTetrahedron::prepareRefine ()
{
    uint          i;
    TFrontier<3>  *pFrontier;

    for (i = 0; i < noOfFrontiers; i++) 
    {
        pFrontier = arrFrontiers[i];

        if ((pFrontier != NULL) && (!pFrontier->hasMidNode())) 
            pFrontier->buildMidNode();
    }// for
}

//
// refine yourself (build only the children)
//
void 
T3DTetrahedron::refine ()
{
    TVertex<3>           *arrOrdNodes[10];
    TFrontier<3>         *arrOrdFrontiers[MAX_FRONTIER];
    T3DCellEdge          *arrOrdEdges[12];
    struct TFrontierInfo  arrFrontierInfo[4];
    

    // create children of all frontiers
    if (!arrFrontiers[0]->hasChild()) arrFrontiers[0]->buildChild();
    if (!arrFrontiers[1]->hasChild()) arrFrontiers[1]->buildChild();
    if (!arrFrontiers[2]->hasChild()) arrFrontiers[2]->buildChild();
    if (!arrFrontiers[3]->hasChild()) arrFrontiers[3]->buildChild();
    
    initChildPtr( 8 );
    
    //
    // determine all nodes in ordered form
    //

    collectNodes( arrOrdNodes, arrFrontierInfo );

    collectFrontierChildren( arrOrdFrontiers, arrOrdEdges, arrFrontierInfo );

    //
    // create inner frontiers
    //
    
    arrOrdFrontiers[_20_01_03_] = new T3DTriangle( NULL, 
                                                   arrOrdEdges[_01_03_],
                                                   arrOrdEdges[_03_20_],
                                                   arrOrdEdges[_20_01_] );
    arrOrdFrontiers[_20_01_03_]->delReference();
    arrOrdFrontiers[_23_12_20_] = new T3DTriangle( NULL, 
                                                   arrOrdEdges[_23_12_],
                                                   arrOrdEdges[_12_20_],
                                                   arrOrdEdges[_20_23_] );
    arrOrdFrontiers[_23_12_20_]->delReference();
    arrOrdFrontiers[_03_13_23_] = new T3DTriangle( NULL, 
                                                   arrOrdEdges[_03_13_],
                                                   arrOrdEdges[_13_23_],
                                                   arrOrdEdges[_23_03_] );
    arrOrdFrontiers[_03_13_23_]->delReference();
    arrOrdFrontiers[_01_12_13_] = new T3DTriangle( NULL, 
                                                   arrOrdEdges[_01_12_],
                                                   arrOrdEdges[_12_13_],
                                                   arrOrdEdges[_13_01_] );
    arrOrdFrontiers[_01_12_13_]->delReference();

    //
    // create children
    //
    arrChildren[0] = new T3DTetrahedron( this,
                                         (T3DTriangle*) arrOrdFrontiers[_20_0_01_],
                                         (T3DTriangle*) arrOrdFrontiers[_20_01_03_],
                                         (T3DTriangle*) arrOrdFrontiers[_03_0_20_],
                                         (T3DTriangle*) arrOrdFrontiers[_01_0_03_] );
    arrChildren[1] = new T3DTetrahedron( this,
                                         (T3DTriangle*) arrOrdFrontiers[_12_2_20_],
                                         (T3DTriangle*) arrOrdFrontiers[_23_2_12_],
                                         (T3DTriangle*) arrOrdFrontiers[_23_12_20_],
                                         (T3DTriangle*) arrOrdFrontiers[_20_2_23_] );
    arrChildren[2] = new T3DTetrahedron( this,
                                         (T3DTriangle*) arrOrdFrontiers[_03_13_23_],
                                         (T3DTriangle*) arrOrdFrontiers[_13_3_23_],
                                         (T3DTriangle*) arrOrdFrontiers[_23_3_03_],
                                         (T3DTriangle*) arrOrdFrontiers[_03_3_13_] );
    arrChildren[3] = new T3DTetrahedron( this,
                                         (T3DTriangle*) arrOrdFrontiers[_01_1_12_],
                                         (T3DTriangle*) arrOrdFrontiers[_12_1_13_],
                                         (T3DTriangle*) arrOrdFrontiers[_01_12_13_],
                                         (T3DTriangle*) arrOrdFrontiers[_13_1_01_] );

    // create inner children
    buildInnerChildren( arrOrdNodes, arrOrdFrontiers, arrOrdEdges );
}

//
// coarse (for future use)
//
void 
T3DTetrahedron::coarsen ()
{
}


//
// get i'th vertex of this cell
//
TVertex<3> *
T3DTetrahedron::getNode ( uint i ) const
{
    TFrontier<3>  *pFrontier = arrFrontiers[0];

    if (pFrontier == NULL)
    {
        cout << "T3DTetrahedron::getNode () : have NULL frontier" << endl;
        exit( -1 );
    }// if
    
    switch (i)
    {
        case 0 :
            return pFrontier->getNode( 0 );

        case 1 :
            return pFrontier->getNode( 2 );

        case 2 :
            return pFrontier->getNode( 4 );

        case 3 :
            pFrontier = arrFrontiers[1];
            if (pFrontier == NULL)
            {
                cout << "T3DTetrahedron::getNode () : have NULL frontier" << endl;
                exit( -1 );
            }// if

            return pFrontier->getNode( _idxOfLastCorner );

    default:
            cout << "T3DTetrahedron::getNode () : range exception "
                 << "(pos = " << i << ", max = 3)" << endl;
            return NULL;
    }// switch
}

// calculate size of cell 


TFloat 
T3DTetrahedron::calcSize ()
{
    TVector<3> x,y,z;
    x = (*getNode(1)) - (*getNode(0));
    y = (*getNode(2)) - (*getNode(0));
    z = (*getNode(3)) - (*getNode(0));
    
    // det [ x y z]   
    TFloat det;
    
    det =x[0]*(y[1]*z[2]-y[2]*z[1]);
    det-=x[1]*(y[0]*z[2]-y[2]*z[0]);
    det+=x[2]*(y[0]*z[1]-y[1]*z[0]);
    
    //   det =   vol(cell)/vol(T_ref)=vol(cell)/( 1/fac(N))
    
    return fabs(det/6.0);
}

//
// catch all corners
//
uint
T3DTetrahedron::idxOfLastCorner ()
{
    TVertex<3>   *v0, *v1, *v2;
    TFrontier<3> *pFrontier;

    // set first frontier as base with nodes 0-2
    pFrontier = arrFrontiers[0];
    
    if (pFrontier == NULL)
    {
        cout << "T3DTetrahedron::idxOfLastCorner () : have NULL frontier" << endl;
        exit( -1 );
    }// if
    
    v0 = pFrontier->getNode( 0 );
    v1 = pFrontier->getNode( 2 );
    v2 = pFrontier->getNode( 4 );

    // on next frontier must be a different node
    pFrontier = arrFrontiers[1];

    if (pFrontier == NULL)
    {
        cout << "T3DTetrahedron::idxOfLastCorner () : have NULL frontier" << endl;
        exit( -1 );
    }// if

    if (pFrontier->getNode( 0 ) == v0)
    {
        if ((pFrontier->getNode( 2 ) == v1) ||
            (pFrontier->getNode( 2 ) == v2))
            return 4;
        else if ((pFrontier->getNode( 4 ) == v1) ||
                 (pFrontier->getNode( 4 ) == v2))
            return 2;
    }// if
    else if (pFrontier->getNode( 0 ) == v1)
    {
        if ((pFrontier->getNode( 2 ) == v2) ||
            (pFrontier->getNode( 2 ) == v0))
            return 4;
        else if ((pFrontier->getNode( 4 ) == v2) ||
                 (pFrontier->getNode( 4 ) == v0))
            return 2;
    }// else if
    else if (pFrontier->getNode( 0 ) == v2)
    {
        if ((pFrontier->getNode( 2 ) == v0) ||
            (pFrontier->getNode( 2 ) == v1))
            return 4;
        else if ((pFrontier->getNode( 4 ) == v0) ||
                 (pFrontier->getNode( 4 ) == v1))
            return 2;
    }// else
    else
        return 0;

    // node do not have correct ordering
    cout << "T3DTetrahedron::idxOfLastCorner () : did not found matching nodes" << endl;
    exit( -1 );
}
    
//
// catch all nodes from frontiers
//
void 
T3DTetrahedron::collectNodes ( TVertex<3> * arrOrdNodes[], 
                               struct TFrontierInfo arrFrontierInfo[] )
{
    uint          i;
    TFrontier<3>  *pFrontier;

    pFrontier = arrFrontiers[0];

    if (pFrontier == NULL)
    {
        cout << "T3DTetrahedron::collectNodes (...) : have NULL frontier" << endl;
        exit( -1 );
    }// if
    
    // set first frontier as base with nodes 0-2
    arrOrdNodes[_0_]  = pFrontier->getNode( 0 );
    arrOrdNodes[_01_] = pFrontier->getNode( 1 );
    arrOrdNodes[_1_]  = pFrontier->getNode( 2 );
    arrOrdNodes[_12_] = pFrontier->getNode( 3 );
    arrOrdNodes[_2_]  = pFrontier->getNode( 4 );
    arrOrdNodes[_20_] = pFrontier->getNode( 5 );

    arrFrontierInfo[0].type      = _0_1_2_;
    arrFrontierInfo[0].firstNode = 0;

    //
    // check next frontier for common nodes and set some of the missing nodes
    //
    for (i = 1; i < 4; i++)
    {
        pFrontier = arrFrontiers[i];

        if (pFrontier == NULL)
        {
            cerr << "T3DTetrahedron::collectNodes : has NULL frontier" << endl;
            exit( -1 );
        }// if

        if (pFrontier->getNode( 0 ) == arrOrdNodes[_0_])
        {
            arrFrontierInfo[i].firstNode = 0;

            if (pFrontier->getNode( 2 ) == arrOrdNodes[_2_])
            {
                arrOrdNodes[_23_] = pFrontier->getNode( 3 );
                arrOrdNodes[_03_] = pFrontier->getNode( 5 );

                arrFrontierInfo[i].type = _0_2_3_;
                arrFrontierInfo[i].dir  = CCW;
            }// if
            else if (pFrontier->getNode( 2 ) == arrOrdNodes[_1_])
            {
                arrOrdNodes[_03_] = pFrontier->getNode( 5 );
                arrOrdNodes[_13_] = pFrontier->getNode( 3 );

                arrFrontierInfo[i].type = _0_3_1_;
                arrFrontierInfo[i].dir  = CW;
            }// else if
            else if (pFrontier->getNode( 4 ) == arrOrdNodes[_2_])
            {
                arrOrdNodes[_23_] = pFrontier->getNode( 3 );
                arrOrdNodes[_03_] = pFrontier->getNode( 1 );

                arrFrontierInfo[i].type = _0_2_3_;
                arrFrontierInfo[i].dir  = CW;
            }// else if
            else if (pFrontier->getNode( 4 ) == arrOrdNodes[_1_])
            {
                arrOrdNodes[_03_] = pFrontier->getNode( 1 );
                arrOrdNodes[_13_] = pFrontier->getNode( 3 );

                arrFrontierInfo[i].type = _0_3_1_;
                arrFrontierInfo[i].dir  = CCW;
            }// else if
            else
            {
                cerr << "T3DTetrahedron::collectNodes () : wrong nodes in frontier" << endl;
                exit( -1 );
            }// else
        }// if
        else if (pFrontier->getNode( 0 ) == arrOrdNodes[_1_])
        {
            arrFrontierInfo[i].firstNode = 1;

            if (pFrontier->getNode( 2 ) == arrOrdNodes[_0_])
            {
                arrOrdNodes[_03_] = pFrontier->getNode( 3 );
                arrOrdNodes[_13_] = pFrontier->getNode( 5 );

                arrFrontierInfo[i].type = _0_3_1_;
                arrFrontierInfo[i].dir  = CCW;
            }// if
            else if (pFrontier->getNode( 2 ) == arrOrdNodes[_2_])
            {
                arrOrdNodes[_13_] = pFrontier->getNode( 5 );
                arrOrdNodes[_23_] = pFrontier->getNode( 3 );

                arrFrontierInfo[i].type = _1_3_2_;
                arrFrontierInfo[i].dir  = CW;
            }// else if
            else if (pFrontier->getNode( 4 ) == arrOrdNodes[_0_])
            {
                arrOrdNodes[_03_] = pFrontier->getNode( 3 );
                arrOrdNodes[_13_] = pFrontier->getNode( 1 );

                arrFrontierInfo[i].type = _0_3_1_;
                arrFrontierInfo[i].dir  = CW;
            }// else if
            else if (pFrontier->getNode( 4 ) == arrOrdNodes[_2_])
            {
                arrOrdNodes[_13_] = pFrontier->getNode( 1 );
                arrOrdNodes[_23_] = pFrontier->getNode( 3 );

                arrFrontierInfo[i].type = _1_3_2_;
                arrFrontierInfo[i].dir  = CCW;
            }// else if
            else
            {
                cerr << "T3DTetrahedron::collectNodes () : wrong nodes in frontier" << endl;
                exit( -1 );
            }// else
        }// else if
        else if (pFrontier->getNode( 0 ) == arrOrdNodes[_2_])
        {
            arrFrontierInfo[i].firstNode = 2;

            if (pFrontier->getNode( 2 ) == arrOrdNodes[_1_])
            {
                arrOrdNodes[_13_] = pFrontier->getNode( 3 );
                arrOrdNodes[_23_] = pFrontier->getNode( 5 );

                arrFrontierInfo[i].type = _1_3_2_;
                arrFrontierInfo[i].dir  = CCW;
            }// if
            else if (pFrontier->getNode( 4 ) == arrOrdNodes[_1_])
            {
                arrOrdNodes[_13_] = pFrontier->getNode( 3 );
                arrOrdNodes[_23_] = pFrontier->getNode( 1 );

                arrFrontierInfo[i].type = _1_3_2_;
                arrFrontierInfo[i].dir  = CW;
            }// else if
            else if (pFrontier->getNode( 4 ) == arrOrdNodes[_0_])
            {
                arrOrdNodes[_23_] = pFrontier->getNode( 1 );
                arrOrdNodes[_03_] = pFrontier->getNode( 3 );

                arrFrontierInfo[i].type = _0_2_3_;
                arrFrontierInfo[i].dir  = CCW;
            }// else if
            else if (pFrontier->getNode( 2 ) == arrOrdNodes[_0_])
            {
                arrOrdNodes[_23_] = pFrontier->getNode( 5 );
                arrOrdNodes[_03_] = pFrontier->getNode( 3 );

                arrFrontierInfo[i].type = _0_2_3_;
                arrFrontierInfo[i].dir  = CW;
            }// else if
            else
            {
                cerr << "T3DTetrahedron::collectNodes () : wrong nodes in frontier" << endl;
                exit( -1 );
            }// else
        }// else if
        else if (pFrontier->getNode( 0 ) != NULL)
        {
            arrFrontierInfo[i].firstNode = 3;

            if (pFrontier->getNode( 2 ) == arrOrdNodes[_0_])
            {
                if (pFrontier->getNode( 4 ) == arrOrdNodes[_1_])
                {
                    arrOrdNodes[_03_] = pFrontier->getNode( 1 );
                    arrOrdNodes[_13_] = pFrontier->getNode( 5 );
     
                    arrFrontierInfo[i].type = _0_3_1_;
                    arrFrontierInfo[i].dir  = CW;
                }// if
                else if (pFrontier->getNode( 4 ) == arrOrdNodes[_2_])
                {
                    arrOrdNodes[_03_] = pFrontier->getNode( 1 );
                    arrOrdNodes[_23_] = pFrontier->getNode( 5 );
     
                    arrFrontierInfo[i].type = _0_2_3_;
                    arrFrontierInfo[i].dir  = CCW;
                }// if
                else
                {
                    cerr << "T3DTetrahedron::collectNodes () : wrong nodes in frontier" << endl;
                    exit( -1 );
                }// else
            }// if
            else if (pFrontier->getNode( 2 ) == arrOrdNodes[_1_])
            {
                if (pFrontier->getNode( 4 ) == arrOrdNodes[_0_])
                {
                    arrOrdNodes[_03_] = pFrontier->getNode( 5 );
                    arrOrdNodes[_13_] = pFrontier->getNode( 1 );
     
                    arrFrontierInfo[i].type = _0_3_1_;
                    arrFrontierInfo[i].dir  = CCW;
                }// if
                else if (pFrontier->getNode( 4 ) == arrOrdNodes[_2_])
                {
                    arrOrdNodes[_13_] = pFrontier->getNode( 1 );
                    arrOrdNodes[_23_] = pFrontier->getNode( 5 );
     
                    arrFrontierInfo[i].type = _1_3_2_;
                    arrFrontierInfo[i].dir  = CW;
                }// if
                else
                {
                    cerr << "T3DTetrahedron::collectNodes () : wrong nodes in frontier" << endl;
                    exit( -1 );
                }// else
            }// if
            else if (pFrontier->getNode( 2 ) == arrOrdNodes[_2_])
            {
                if (pFrontier->getNode( 4 ) == arrOrdNodes[_0_])
                {
                    arrOrdNodes[_03_] = pFrontier->getNode( 5 );
                    arrOrdNodes[_23_] = pFrontier->getNode( 1 );
     
                    arrFrontierInfo[i].type = _0_2_3_;
                    arrFrontierInfo[i].dir  = CW;
                }// if
                else if (pFrontier->getNode( 4 ) == arrOrdNodes[_1_])
                {
                    arrOrdNodes[_13_] = pFrontier->getNode( 5 );
                    arrOrdNodes[_23_] = pFrontier->getNode( 1 );
     
                    arrFrontierInfo[i].type = _1_3_2_;
                    arrFrontierInfo[i].dir  = CCW;
                }// if
                else
                {
                    cerr << "T3DTetrahedron::collectNodes () : wrong nodes in frontier" << endl;
                    exit( -1 );
                }// else
            }// if
            else
            {
                cerr << "T3DTetrahedron::collectNodes () : wrong nodes in frontier" << endl;
                exit( -1 );
            }// else
        }// else if
        else
        {
            cerr << "T3DTetrahedron::collectNodes () : wrong nodes in frontier" << endl;
            exit( -1 );
        }// else
    }// for
}

//
// collect all children of all frontiers
//
void
T3DTetrahedron::collectFrontierChildren ( TFrontier<3> * arrOrdFrontiers[],
                                          T3DCellEdge  * arrOrdEdges[],
                                          struct TFrontierInfo arrFrontierInfo[] )
{
    uint          i;
    TFrontier<3>  *pFrontier;

    for (i = 0; i < 4; i++)
    {
        pFrontier = arrFrontiers[i];

        if ((pFrontier == NULL) || (pFrontier->getNoOfChildren() != 4))
        {
            cerr << "collectFrontierChildren (...) : not correct noOfChildren" << endl;
            exit( -1 );
        }// if

        if (arrFrontierInfo[i].type == _0_1_2_)
        {
            // first node is (per def.) 0
            arrOrdFrontiers[_20_0_01_]  = pFrontier->getChild( 0 );
            arrOrdFrontiers[_01_1_12_]  = pFrontier->getChild( 1 );
            arrOrdFrontiers[_12_2_20_]  = pFrontier->getChild( 2 );
            arrOrdFrontiers[_20_01_12_] = pFrontier->getChild( 3 );

            arrOrdEdges[_20_01_] = ((T3DFrontier*)arrOrdFrontiers[_20_0_01_])->getEdge( 2 );
            arrOrdEdges[_01_12_] = ((T3DFrontier*)arrOrdFrontiers[_01_1_12_])->getEdge( 2 );
            arrOrdEdges[_12_20_] = ((T3DFrontier*)arrOrdFrontiers[_12_2_20_])->getEdge( 2 );
        }// if
        else if (arrFrontierInfo[i].type == _0_2_3_)
        {
            if (arrFrontierInfo[i].dir == CCW)
            {
                switch (arrFrontierInfo[i].firstNode)
                {
                    case 0:
                        arrOrdFrontiers[_03_0_20_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_20_2_23_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_23_3_03_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_03_20_23_] = pFrontier->getChild( 3 );
                        break;
                    case 2:
                        arrOrdFrontiers[_03_0_20_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_20_2_23_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_23_3_03_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_03_20_23_] = pFrontier->getChild( 3 );
                        break;
                    case 3:
                        arrOrdFrontiers[_03_0_20_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_20_2_23_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_23_3_03_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_03_20_23_] = pFrontier->getChild( 3 );
                        break;
                }// switch  
            }// if
            else
            {
                switch (arrFrontierInfo[i].firstNode)
                {
                    case 0:
                        arrOrdFrontiers[_03_0_20_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_23_3_03_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_20_2_23_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_03_20_23_] = pFrontier->getChild( 3 );
                        break;
                    case 2:
                        arrOrdFrontiers[_03_0_20_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_20_2_23_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_23_3_03_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_03_20_23_] = pFrontier->getChild( 3 );
                        break;
                    case 3:
                        arrOrdFrontiers[_03_0_20_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_20_2_23_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_23_3_03_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_03_20_23_] = pFrontier->getChild( 3 );
                        break;
                }// switch  
            }// else

            arrOrdEdges[_03_20_] = ((T3DFrontier*)arrOrdFrontiers[_03_0_20_])->getEdge( 2 );
            arrOrdEdges[_20_23_] = ((T3DFrontier*)arrOrdFrontiers[_20_2_23_])->getEdge( 2 );
            arrOrdEdges[_23_03_] = ((T3DFrontier*)arrOrdFrontiers[_23_3_03_])->getEdge( 2 );
        }// else if
        else if (arrFrontierInfo[i].type == _0_3_1_)
        {
            if (arrFrontierInfo[i].dir == CCW)
            {
                switch (arrFrontierInfo[i].firstNode)
                {
                    case 0:
                        arrOrdFrontiers[_01_0_03_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_03_3_13_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_13_1_01_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_01_03_13_] = pFrontier->getChild( 3 );
                        break;
                    case 3:
                        arrOrdFrontiers[_01_0_03_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_03_3_13_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_13_1_01_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_01_03_13_] = pFrontier->getChild( 3 );
                        break;
                    case 1:
                        arrOrdFrontiers[_01_0_03_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_03_3_13_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_13_1_01_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_01_03_13_] = pFrontier->getChild( 3 );
                        break;
                }// switch
            }// if
            else
            {
                switch (arrFrontierInfo[i].firstNode)
                {
                    case 0:
                        arrOrdFrontiers[_01_0_03_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_03_3_13_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_13_1_01_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_01_03_13_] = pFrontier->getChild( 3 );
                        break;
                    case 3:
                        arrOrdFrontiers[_01_0_03_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_03_3_13_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_13_1_01_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_01_03_13_] = pFrontier->getChild( 3 );
                        break;
                    case 1:
                        arrOrdFrontiers[_01_0_03_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_03_3_13_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_13_1_01_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_01_03_13_] = pFrontier->getChild( 3 );
                        break;
                }// switch
            }// else

            arrOrdEdges[_01_03_] = ((T3DFrontier*)arrOrdFrontiers[_01_0_03_])->getEdge( 2 );
            arrOrdEdges[_03_13_] = ((T3DFrontier*)arrOrdFrontiers[_03_3_13_])->getEdge( 2 );
            arrOrdEdges[_13_01_] = ((T3DFrontier*)arrOrdFrontiers[_13_1_01_])->getEdge( 2 );
        }// else if
        else if (arrFrontierInfo[i].type == _1_3_2_)
        {
            if (arrFrontierInfo[i].dir == CCW)
            {
                switch (arrFrontierInfo[i].firstNode)
                {
                    case 1:
                        arrOrdFrontiers[_12_1_13_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_13_3_23_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_23_2_12_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_12_13_23_] = pFrontier->getChild( 3 );
                        break;
                    case 3:
                        arrOrdFrontiers[_12_1_13_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_13_3_23_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_23_2_12_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_12_13_23_] = pFrontier->getChild( 3 );
                        break;
                    case 2:
                        arrOrdFrontiers[_12_1_13_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_13_3_23_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_23_2_12_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_12_13_23_] = pFrontier->getChild( 3 );
                        break;
                }// switch
            }// if
            else
            {
                switch (arrFrontierInfo[i].firstNode)
                {
                    case 1:
                        arrOrdFrontiers[_12_1_13_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_13_3_23_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_23_2_12_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_12_13_23_] = pFrontier->getChild( 3 );
                        break;
                    case 3:
                        arrOrdFrontiers[_12_1_13_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_13_3_23_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_23_2_12_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_12_13_23_] = pFrontier->getChild( 3 );
                        break;
                    case 2:
                        arrOrdFrontiers[_12_1_13_]  = pFrontier->getChild( 2 );
                        arrOrdFrontiers[_13_3_23_]  = pFrontier->getChild( 1 );
                        arrOrdFrontiers[_23_2_12_]  = pFrontier->getChild( 0 );
                        arrOrdFrontiers[_12_13_23_] = pFrontier->getChild( 3 );
                        break;
                }// switch
            }// else

            arrOrdEdges[_12_13_] = ((T3DFrontier*)arrOrdFrontiers[_12_1_13_])->getEdge( 2 );
            arrOrdEdges[_13_23_] = ((T3DFrontier*)arrOrdFrontiers[_13_3_23_])->getEdge( 2 );
            arrOrdEdges[_23_12_] = ((T3DFrontier*)arrOrdFrontiers[_23_2_12_])->getEdge( 2 );
        }// else if
        else
        {
            cerr << "collectFrontierChildren (...) : wrong frontiertype" << endl;
            exit( -1 );
        }// else
    }// for
}

//
// decide which refine-strategy to use and refine inner octahedron
//
void
T3DTetrahedron::buildInnerChildren ( TVertex<3>   * arrOrdNodes[],
                                     TFrontier<3> * arrOrdFrontiers[],
                                     T3DCellEdge  * arrOrdEdges[] )
{
    T3DCellEdge   *pInnerEdge;
    TFloat         length_03_12, length_01_23, length_20_13, minDiag = 0.0;
    TVector<3>     vecTemp;

    //
    // look for shortest inner diagonal to choose refinement strategy
    //
    vecTemp  = *arrOrdNodes[_03_];
    vecTemp -= *arrOrdNodes[_12_];
    length_03_12 = vecTemp.euclidNorm();
 
    vecTemp  = *arrOrdNodes[_01_];
    vecTemp -= *arrOrdNodes[_23_];
    length_01_23 = vecTemp.euclidNorm();
 
    vecTemp  = *arrOrdNodes[_20_];
    vecTemp -= *arrOrdNodes[_13_];
    length_20_13 = vecTemp.euclidNorm();

    if      ((length_20_13 <= length_01_23) && (length_20_13 <= length_03_12))
        minDiag = length_20_13;
    else if ((length_01_23 <= length_20_13) && (length_01_23 <= length_03_12))
        minDiag = length_01_23;
    else if ((length_03_12 <= length_20_13) && (length_03_12 <= length_01_23))
        minDiag = length_03_12;
    else
        cout << "######################### error #############################" << endl;

    //
    // build children according to strategy
    //
    if (length_03_12 == minDiag)
    {
        pInnerEdge = new T3DCellEdge( arrOrdNodes[_03_], arrOrdNodes[_12_] );
        pInnerEdge->delReference();

        arrOrdFrontiers[_01_03_12_] = new T3DTriangle( NULL,
                                                       arrOrdEdges[_01_03_],
                                                       pInnerEdge,
                                                       arrOrdEdges[_01_12_] );
        arrOrdFrontiers[_01_03_12_]->delReference();
        arrOrdFrontiers[_20_03_12_] = new T3DTriangle( NULL,
                                                       arrOrdEdges[_03_20_],
                                                       pInnerEdge,
                                                       arrOrdEdges[_12_20_] );
        arrOrdFrontiers[_20_03_12_]->delReference();
        arrOrdFrontiers[_23_03_12_] = new T3DTriangle( NULL,
                                                       arrOrdEdges[_23_03_],
                                                       pInnerEdge,
                                                       arrOrdEdges[_23_12_] );
        arrOrdFrontiers[_23_03_12_]->delReference();
        arrOrdFrontiers[_13_03_12_] = new T3DTriangle( NULL,
                                                       arrOrdEdges[_03_13_],
                                                       pInnerEdge,
                                                       arrOrdEdges[_12_13_] );
        arrOrdFrontiers[_13_03_12_]->delReference();
        
        
        arrChildren[4] = new T3DTetrahedron( this,
                                             (T3DTriangle*) arrOrdFrontiers[_23_12_20_],
                                             (T3DTriangle*) arrOrdFrontiers[_03_20_23_],
                                             (T3DTriangle*) arrOrdFrontiers[_20_03_12_],
                                             (T3DTriangle*) arrOrdFrontiers[_23_03_12_] );
        arrChildren[5] = new T3DTetrahedron( this,
                                             (T3DTriangle*) arrOrdFrontiers[_20_01_12_],
                                             (T3DTriangle*) arrOrdFrontiers[_20_01_03_],
                                             (T3DTriangle*) arrOrdFrontiers[_01_03_12_],
                                             (T3DTriangle*) arrOrdFrontiers[_20_03_12_] );
        arrChildren[6] = new T3DTetrahedron( this,
                                             (T3DTriangle*) arrOrdFrontiers[_01_12_13_],
                                             (T3DTriangle*) arrOrdFrontiers[_01_03_13_],
                                             (T3DTriangle*) arrOrdFrontiers[_13_03_12_],
                                             (T3DTriangle*) arrOrdFrontiers[_01_03_12_] );
        arrChildren[7] = new T3DTetrahedron( this,
                                             (T3DTriangle*) arrOrdFrontiers[_12_13_23_],
                                             (T3DTriangle*) arrOrdFrontiers[_03_13_23_],
                                             (T3DTriangle*) arrOrdFrontiers[_13_03_12_],
                                             (T3DTriangle*) arrOrdFrontiers[_23_03_12_] );
        return;
    }// if

    if (length_01_23 == minDiag)
    {
        pInnerEdge = new T3DCellEdge( arrOrdNodes[_01_], arrOrdNodes[_23_] );
        pInnerEdge->delReference();
        
        arrOrdFrontiers[_01_20_23_] = new T3DTriangle( NULL,
                                                       arrOrdEdges[_20_01_],
                                                       pInnerEdge,
                                                       arrOrdEdges[_20_23_] );
        arrOrdFrontiers[_01_20_23_]->delReference();
        arrOrdFrontiers[_01_23_12_] = new T3DTriangle( NULL,
                                                       pInnerEdge,
                                                       arrOrdEdges[_23_12_],
                                                       arrOrdEdges[_01_12_] );
        arrOrdFrontiers[_01_23_12_]->delReference();
        arrOrdFrontiers[_01_23_03_] = new T3DTriangle( NULL,
                                                       pInnerEdge,
                                                       arrOrdEdges[_23_03_],
                                                       arrOrdEdges[_01_03_] );
        arrOrdFrontiers[_01_23_03_]->delReference();
        arrOrdFrontiers[_01_23_13_] = new T3DTriangle( NULL,
                                                       arrOrdEdges[_13_01_],
                                                       pInnerEdge,
                                                       arrOrdEdges[_13_23_] );
        arrOrdFrontiers[_01_23_13_]->delReference();
        
        
        arrChildren[4] = new T3DTetrahedron( this,
                                             (T3DTriangle*) arrOrdFrontiers[_20_01_12_],
                                             (T3DTriangle*) arrOrdFrontiers[_23_12_20_],
                                             (T3DTriangle*) arrOrdFrontiers[_01_20_23_],
                                             (T3DTriangle*) arrOrdFrontiers[_01_23_12_] );
        arrChildren[5] = new T3DTetrahedron( this,
                                             (T3DTriangle*) arrOrdFrontiers[_20_01_03_],
                                             (T3DTriangle*) arrOrdFrontiers[_03_20_23_],
                                             (T3DTriangle*) arrOrdFrontiers[_01_20_23_],
                                             (T3DTriangle*) arrOrdFrontiers[_01_23_03_] );
        arrChildren[6] = new T3DTetrahedron( this,
                                             (T3DTriangle*) arrOrdFrontiers[_01_12_13_],
                                             (T3DTriangle*) arrOrdFrontiers[_12_13_23_],
                                             (T3DTriangle*) arrOrdFrontiers[_01_23_13_],
                                             (T3DTriangle*) arrOrdFrontiers[_01_23_12_] );
        arrChildren[7] = new T3DTetrahedron( this,
                                             (T3DTriangle*) arrOrdFrontiers[_01_03_13_],
                                             (T3DTriangle*) arrOrdFrontiers[_03_13_23_],
                                             (T3DTriangle*) arrOrdFrontiers[_01_23_13_],
                                             (T3DTriangle*) arrOrdFrontiers[_01_23_03_] );

        return;
    }// if

    if (length_20_13 == minDiag)
    {
        pInnerEdge = new T3DCellEdge( arrOrdNodes[_20_], arrOrdNodes[_13_] );
        pInnerEdge->delReference();
        
        arrOrdFrontiers[_01_20_13_] = new T3DTriangle( NULL,
                                                       pInnerEdge,
                                                       arrOrdEdges[_13_01_],
                                                       arrOrdEdges[_20_01_] );
        arrOrdFrontiers[_01_20_13_]->delReference();
        arrOrdFrontiers[_12_20_13_] = new T3DTriangle( NULL,
                                                       arrOrdEdges[_12_20_],
                                                       pInnerEdge,
                                                       arrOrdEdges[_12_13_] );
        arrOrdFrontiers[_12_20_13_]->delReference();
        arrOrdFrontiers[_23_20_13_] = new T3DTriangle( NULL,
                                                       pInnerEdge,
                                                       arrOrdEdges[_13_23_],
                                                       arrOrdEdges[_20_23_] );
        arrOrdFrontiers[_23_20_13_]->delReference();
        arrOrdFrontiers[_03_20_13_] = new T3DTriangle( NULL,
                                                       arrOrdEdges[_03_20_],
                                                       pInnerEdge,
                                                       arrOrdEdges[_03_13_] );
        arrOrdFrontiers[_03_20_13_]->delReference();
        
        
        arrChildren[4] = new T3DTetrahedron( this,
                                             (T3DTriangle*) arrOrdFrontiers[_20_01_12_],
                                             (T3DTriangle*) arrOrdFrontiers[_01_12_13_],
                                             (T3DTriangle*) arrOrdFrontiers[_01_20_13_],
                                             (T3DTriangle*) arrOrdFrontiers[_12_20_13_] );
        arrChildren[5] = new T3DTetrahedron( this,
                                             (T3DTriangle*) arrOrdFrontiers[_20_01_03_],
                                             (T3DTriangle*) arrOrdFrontiers[_01_03_13_],
                                             (T3DTriangle*) arrOrdFrontiers[_01_20_13_],
                                             (T3DTriangle*) arrOrdFrontiers[_03_20_13_] );
        arrChildren[6] = new T3DTetrahedron( this,
                                             (T3DTriangle*) arrOrdFrontiers[_23_12_20_],
                                             (T3DTriangle*) arrOrdFrontiers[_12_13_23_],
                                             (T3DTriangle*) arrOrdFrontiers[_12_20_13_],
                                             (T3DTriangle*) arrOrdFrontiers[_23_20_13_] );
        arrChildren[7] = new T3DTetrahedron( this,
                                             (T3DTriangle*) arrOrdFrontiers[_03_20_23_],
                                             (T3DTriangle*) arrOrdFrontiers[_03_13_23_],
                                             (T3DTriangle*) arrOrdFrontiers[_03_20_13_],
                                             (T3DTriangle*) arrOrdFrontiers[_23_20_13_] );

        return;
    }// if
}
 
TVertex<3>*
T3DTetrahedron::getOppositeNode ( uint i ) const{
    // gets node opposite to frontier i

    TFrontier<3>  *pFrontier = arrFrontiers[i];
    TVertex<3>* p;
    // shortcut: Node 3 is opposite to frontier 0 
    if (i == 0)
        return this->getNode(3);

    else{
        uint l;
        bool flagfield[4];
        TVertex<3>*  nodefield[4];
        for (uint j=0;j<4;j++){
            flagfield[j]=FALSE;
            nodefield[j]=this->getNode(j);
        }
       

        // mark all nodes which belong to this frontier
        // i.e., pFrontier->getNode(k); with k=0,2,4

        for (uint k=0; k<6;k+=2){
            p=pFrontier->getNode(k);
            l=0;
            while(nodefield[l]!=p)
                l++;
            if(l>4)
                cerr<<"T3DTetrahedron::getOppositeNode : Problems 1"<<endl;
            flagfield[l]=TRUE;
        }
        
       // search unmarked node - this is
        l=0;
        while (flagfield[l]==TRUE)
            l++;
        if (l>4)
            cerr<<"T3DTetrahedron::getOppositeNode : Problems 2"<<endl;
        return nodefield[l];
    }// else


}; // getOppositeNode


