//
//  Project   : libnet
//  File      : cyclicReduction.cc
//  Author    : tpr
//  Purpose   : cyclic reduction solver 
//             (cf. A. Reusken, IGPM-Berich 144, Okt. 1997, RWTH Aachen)
//
// arch-tag: 811591ee-1af1-48d6-8558-adf11a24d3fa
//

extern "C" {
#include <stdlib.h>
#include <stdio.h>
}

#include "TCycRedMatrixBlock.hh"

#include "TCyclicReduction.hh"

/////////////////////////////////////////////////////////
//
// class containing all the CyclicReduction parameters
//

//
// constructor
//
TCyclicReductionInfo::TCyclicReductionInfo ( uint alow, uint ahigh, uint aRecCall,
                                             TSolver* apInnerSolver,
                                             TSolver* apCoarseSolver) 
{ 
   low  = alow;
   high = ahigh; 
   noOfLevel = high - low + 1;
   pInnerSolver = apInnerSolver;
   pCoarseSolver = apCoarseSolver;
   recursiveCalls = aRecCall;
   CGCdamping = 1.0;
}

//
// get lowest/highest level
//
uint 
TCyclicReductionInfo::getHighLevel () { return (high); }

uint 
TCyclicReductionInfo::getLowLevel () { return (low); }

uint 
TCyclicReductionInfo::getNoOfLevel () { return (noOfLevel); }

//
// set InnerSolver		
//
void 
TCyclicReductionInfo::setInnerSolver ( TSolver* apInnerSolver ) 
{
   pInnerSolver = apInnerSolver;

	
}

// set coarse grid solver
void 
TCyclicReductionInfo::setCoarseSolver ( TSolver* apCoarseSolver ) 
{
   pCoarseSolver = apCoarseSolver;
}

/////////////////////////////////////////////////////////
//
// class for a CyclicReduction solver
// 

// 
// constructor
// 
TCyclicReduction::TCyclicReduction ( TBlockMatrix *apA, TBlockVector<TFloat> *apRHS, 
                                     TBlockVector<TFloat> *apX, 
                                     uint alow, uint ahigh, uint aRecCall,  
                                     TSolver* apInnerSolver,
                                     TSolver* apCoarseSolver, 
                                     uint amaxiter,
                                     TFloat aaccur,
                                     uint averb )
   : TSolver ( apA, apRHS, apX,  amaxiter, aaccur, averb ),
     TCyclicReductionInfo( alow, ahigh, aRecCall, apInnerSolver, 
                           apCoarseSolver )
{ 
   setName( "CycRed" ); 
  
   pA = apA;
   pX = apX;
   pRHS = apRHS;
}

//
// one CyclicReduction step on level 'level'
//
void 
TCyclicReduction::step ( uint level ) {

  
   if (level==0) { cerr<<" Level = 0" <<endl; exit(EXIT_FAILURE);}
  
   TCycRedMatrixBlock *pC=dynamic_cast< TCycRedMatrixBlock*> ( &(*pA)[0][0]);
   if (!pC){
      cerr<<"in TCyclicReduction::step : given matrix is not of type TCycRedMatrixBlock "<<endl;
      exit(EXIT_FAILURE);
   }
   else{  
      uint dim=(*pA)[0][0].getNoOfRows();
    

     
      // calculate defect
      TBlockVector<TFloat>* pd = new TBlockVector<TFloat>(1,dim);
  
      
      // ggf. rhs und L"osung permutieren
      if(level>1){
         if (pC->getDecompositionStatus()){ 
            (*pX)[0].permute(pC->getInvPerm());
            (*pRHS)[0].permute(pC->getInvPerm());
         }
      }

      (*pd)= ((*pA)*(*pX)-(*pRHS));

   
      if (!pC->getDecompositionStatus()){  // coarsest  Level
         cout << "TCyclicReduction::step (...) : coarsegrid-solver" << endl;
         pCoarseSolver->solve( pA, pRHS , pX );
      }
    
      else  { 
         uint i;
       
         uint NoOfReds= pC->getNoOfReds();
         uint NoOfBlacks= pC->getNoOfBlacks();
       
         //partition
         TVectorBlock<TFloat>* aRedX = new TVectorBlock<TFloat>(0.0,NoOfReds);
         TVectorBlock<TFloat>* aBlackX = new TVectorBlock<TFloat>(0.0,NoOfBlacks);
         TVectorBlock<TFloat>* aRedRHS = new TVectorBlock<TFloat>(0.0,NoOfReds);
         TVectorBlock<TFloat>* aBlackRHS = new TVectorBlock<TFloat>(0.0,NoOfBlacks);
                                             
         TVectorBlock<TFloat>* aCopyRedRHS = new TVectorBlock<TFloat>(0.0,NoOfBlacks+NoOfReds);
                                             
         TBlockVector<TFloat>* pRedX = new TBlockVector<TFloat>(*aRedX);
         TBlockVector<TFloat>* pBlackX = new TBlockVector<TFloat>(*aBlackX);
         TBlockVector<TFloat>* pRedRHS = new TBlockVector<TFloat>(*aRedRHS);
         TBlockVector<TFloat>* pBlackRHS = new TBlockVector<TFloat>(*aBlackRHS);
                                             
         TBlockVector<TFloat>* pCopyRedRHS = new TBlockVector<TFloat>(*aCopyRedRHS);
        

         for (i=0; i<NoOfBlacks;i++)
            (*pBlackRHS)[0][i]=(*pd)[0][i];
         for (i=0; i<NoOfReds;i++)
            (*pRedRHS)[0][i]=(*pd)[0][i+NoOfBlacks];
        
        
         TBlockMatrix* pArb = new TBlockMatrix(pC->getpArb());
         TBlockMatrix* pAbr = new TBlockMatrix(pC->getpAbr());
         TBlockMatrix* pArr = new TBlockMatrix(pC->getpArr());
         TBlockMatrix* pAbb = new TBlockMatrix(pC->getpAbb());
         
         //   TMatrixBlock *Abb = pC->getpAbb();
//         (*pAbb)[0].change(pC->getpAbb());
        
         // InnerSolve
         //cout<<"inner 1 solver level "<<level<<endl;;
         if (pInnerSolver) pInnerSolver->solve(pArr, pRedRHS , pRedX  );
         //   cout<<"inner 1 done "<<endl;
        
      
         // copy Red:
         //         cout<<"copy ...";
         for (i=0; i<NoOfReds;i++)
            (*pCopyRedRHS)[0][i+NoOfBlacks]=(*pRedX)[0][i];
         // cout<<"copy done"<<endl;
         //  cout <<(*pCopyRedRHS)[0]<<endl;

         // distribute
         //cout<<"dist 1 ...";
         (*pBlackRHS)[0] -=((*pAbr)[0][0])*(*pCopyRedRHS)[0];
         //cout<<" (*pBlackRHS)"<<endl<< (*pBlackRHS)<<endl;
         //cout<<"  dist 1 done "<<endl;

         // recursive call
        

         TBlockMatrix *tpA;
         TBlockVector<TFloat> *tpRHS;
         TBlockVector<TFloat> *tpX;


         tpA=getA();
         tpRHS=getRHS();
         tpX=getX();    
         
        
        
         pBlackX->set(0.0);
         
         setA(pAbb);
         setRHS(pBlackRHS);
         setX(pBlackX);

         //         TCycRedMatrixBlock *pD=dynamic_cast< TCycRedMatrixBlock*> ( &(*pAbb)[0][0]);
         
             
         for (uint k=0;k<recursiveCalls;k++)  
            step(level+1);
           
         
         setA(tpA);
         setRHS(tpRHS);
         setX(tpX);    

         // collect
         //cout<<"collect... "<<endl;
         (*pRedRHS)[0]-= ((*pArb)[0][0])*(*pBlackX)[0];
         // cout<<" collect done "<<endl;
  
      
         // InnerSolve
         //cout<<"inner 2....";
         //cout<<"inner 2 solver level "<<level<<endl;;
         if (pInnerSolver) pInnerSolver->solve(pArr, pRedRHS , pRedX  );
         //cout<<"inner 2 done "<<endl;

         TFloat damp=1.0;
         // collect red/black and correct pX
         for (i=0; i<NoOfBlacks;i++)
            (*pX)[0][i]-=damp*(*pBlackX)[0][i];
         for (i=0; i<NoOfReds;i++)
            (*pX)[0][i+NoOfBlacks]-=damp*(*pRedX)[0][i];
      
 
     

         delete aRedX;      
         delete aBlackX;        
         delete aRedRHS;    
         delete aBlackRHS ; 
           
           
         delete aCopyRedRHS;
           
         delete pRedX;      
         delete pBlackX;    
         delete pRedRHS;    
         delete pBlackRHS;  
           
         delete pCopyRedRHS;

         delete pArr;
         delete pAbr;
         delete pArb;
         delete pAbb;
      } //else highlevel
      (*pd)= ((*pA)*(*pX)-(*pRHS));

      if (level<1)
          cout << "TCyclicReduction::step () : defect in level "<<level
               <<" dim "<<(*pA)[0][0].getNoOfRows()
               <<" after step: "<<pd->weightedEuclidNorm()<<endl;
      delete pd;
      
      if (level >1){
         if (pC->getDecompositionStatus()){ 
            (*pX)[0].permute(pC->getPerm());
            (*pRHS)[0].permute(pC->getPerm()); 
         }
      }
    
   } // cast successful
} //step

//
// one CyclicReduction step on highest level
//
int 
TCyclicReduction::step () 
{ 
   step(1 ); 
   return (RETURN_CONTINUE);
}




