//
//  Project   : libnet
//  File      : TMultigrid.cc
//  Author    : Sabine Gutsch
//  Purpose   : multigrid method
//
// arch-tag: 25ec847b-092e-4c2b-9a61-39b43f0412b9
//

#include <stdio.h>

#include "TMultigrid.hh"

/////////////////////////////////////////////////////////
//
// class containing all the multigrid parameters
//

//
// constructor
//
TMultigridInfo::TMultigridInfo ( uint alow, uint ahigh, uint aRecCall,
                                 TSolver* apPreSmoother, TSolver* apPostSmoother,
                                 TSolver* apCoarseSolver) 
{ 
    low  = alow;
    high = ahigh; 
    noOfLevel = high - low + 1;
    pPreSmoother = apPreSmoother;
    pPostSmoother = apPostSmoother;
    pCoarseSolver = apCoarseSolver;
    preSmoothingSteps = 1;
    postSmoothingSteps = 1;
    recursiveCalls = aRecCall;
    CGCdamping = 1.0;
}

//
// get lowest/highest level
//
uint 
TMultigridInfo::getHighLevel () { return (high); }

uint 
TMultigridInfo::getLowLevel () { return (low); }

uint 
TMultigridInfo::getNoOfLevel () { return (noOfLevel); }

//
// set smoother		
//
void 
TMultigridInfo::setSmoother ( TSolver* apSmoother ) 
{
    pPreSmoother = apSmoother;
    pPostSmoother = apSmoother;
}

// set coarse grid solver
void 
TMultigridInfo::setCoarseSolver ( TSolver* apCoarseSolver ) 
{
    pCoarseSolver = apCoarseSolver;
}

// set cgc damping
void 
TMultigridInfo::setCGCdamping ( TFloat aDamp ) 
{
	CGCdamping = aDamp;
}

/////////////////////////////////////////////////////////
//
// class for a multigrid solver
// 

// 
// constructor
// 
template <int N>
TMultigrid<N>::TMultigrid ( TBlockMatrix         ** apA,
                            TBlockVector<TFloat> ** apRHS, 
                            TBlockVector<TFloat> ** apX, 
                            TProlongation<N>     ** apRestrict, 
                            TProlongation<N>     ** apProlong, 
                            uint                    alow,
                            uint                    ahigh,
                            uint                    aRecCall,  
                            TSolver               * apPreSmoother,
                            TSolver               * apPostSmoother,
                            TSolver               * apCoarseSolver, 
                            uint                    amaxiter,
                            TFloat                  aaccur,
                            uint                    averb )
        : TSolver ( apA[ahigh-1],
                    apRHS[ahigh-1],
                    apX[ahigh-1], 
                    amaxiter,
                    aaccur,
                    averb ),
          TMultigridInfo( alow,
                          ahigh,
                          aRecCall,
                          apPreSmoother, 
                          apPostSmoother,
                          apCoarseSolver )
{
    uint  i;
    
    if (pPreSmoother != NULL)
    {
        TString newName;

        newName  = "MG ";
        newName += "(";
        newName += pPreSmoother->getName();
        newName += ")";
        setName( newName ); 
    }// if
    else
        setName( "MG" ); 

    //
    // copy !!! given hierarchy
    //
    
    AH.setSize( ahigh+1 );
    RHSH.setSize( ahigh+1 );
    XH.setSize( ahigh+1 );

    for ( i = 0; i <= ahigh; i++ )
    {
        AH[i]   = apA[i];
        RHSH[i] = apRHS[i];
        XH[i]   = apX[i];
    }// for

    pProlong  = apProlong;
    pRestrict = apRestrict;

    pA   = apA[ahigh];
    pX   = apX[ahigh];
    pRHS = apRHS[ahigh];
}

template <int N>
TMultigrid<N>::~TMultigrid ()
{
}

//
// step with explicitly specified system
//
template <int N>
int
TMultigrid<N>::step ( int                       level,
                      TBlockMatrix           ** A,
                      TBlockVector< TFloat >  * RHS,
                      TBlockVector< TFloat >  * x )
{
    uint  i;

    for ( i = 0; i <= high; i++ )
        AH[i] = A[i];

    RHSH[level] = RHS;
    XH[level]   = x;
    
    step( level );
    
    return RETURN_CONTINUE;
}

//
// one multigrid step on level 'level'
//
template <int N>
void 
TMultigrid<N>::step ( uint level ) 
{
    if (level == low) 
        pCoarseSolver->solve( AH[level], RHSH[level], XH[level] );
    else 
    {
        // presmoothing
        if ((pPreSmoother) && (preSmoothingSteps > 0)) 
		{
            pPreSmoother->setMaxit( preSmoothingSteps );
            pPreSmoother->solve( AH[level], RHSH[level], XH[level] );
        }

        // do coarse grid correction only if necessary
        if (recursiveCalls > 0) 
		{
            // calculate defect on fine grid
            d = (*RHSH[level]) - (*AH[level]) * (*XH[level]);

            // restrict defect to coarse grid
            *RHSH[level-1] = pRestrict[level-1]->restrict( d );


            // project *RHSH[level-1] onto range of *AH[level-1]
            uint nob = d.getNoOfBlocks();
            if (nob > 1) 
            {
                TVectorBlock<TFloat> vecOne( (*RHSH[level-1])[nob-1] );
                vecOne.set( 1.0 );
                TFloat factor = (vecOne * (*RHSH[level-1])[nob-1]) / (vecOne * vecOne);
                vecOne.set( factor );
                (*RHSH[level-1])[nob-1] -= vecOne;
            }

            XH[level-1]->set( 0.0 );
            // coarse grid solution
            for (uint j = 0; j < recursiveCalls; j++) 
                step( level-1 );

            // project *XH[level] onto range of *AH[level]
            if (nob > 1) 
            {
                TVectorBlock<TFloat> vecOne( (*RHSH[level-1])[nob-1] );
                vecOne.set( 1.0 );
                TFloat factor = (vecOne * (*XH[level-1])[nob-1]) / (vecOne * vecOne);
                vecOne.set( factor );
                (*XH[level-1])[nob-1] -= vecOne;
            }

            // coarse grid correction
            (*XH[level]) += CGCdamping * pProlong[level-1]->prolong( (*XH[level-1]) );

            // project *XH[level] onto range of *AH[level]
            if (nob > 1) 
            {
                TVectorBlock<TFloat> vecOne( (*RHSH[level])[nob-1] );
                vecOne.set( 1.0 );
                TFloat factor = (vecOne * (*XH[level])[nob-1]) / (vecOne * vecOne);
                vecOne.set( factor );
                (*XH[level])[nob-1] -= vecOne;
            }
        }
        // postsmoothing
        if ((pPostSmoother) && (postSmoothingSteps > 0) ) 
		{
            pPostSmoother->setMaxit( postSmoothingSteps );
            pPostSmoother->solve( AH[level], RHSH[level], XH[level] );
        }               

    }// else
     
} 

//
// one multigrid step on highest level
//
template <int N>
int 
TMultigrid<N>::step () 
{ 
    step( high ); 
    return (RETURN_CONTINUE);
}

//
// multigrid solve
//
template <int N>
void
TMultigrid<N>::solve ( int                     level,
                       TBlockMatrix         ** A,
                       TBlockVector<TFloat> ** RHS,
                       TBlockVector<TFloat> ** x )
{
    int  i;
    
    AH.setSize( level+1 );
    RHSH.setSize( level+1 );
    XH.setSize( level+1 );

    for ( i = 0; i <= level; i++ )
    {
        AH[i]   = A[i];
        RHSH[i] = RHS[i];
        XH[i]   = x[i];
    }// for

    pA   = A[level];
    pX   = x[level];
    pRHS = RHS[level];

    high      = level;
    noOfLevel = high - low + 1;
    
    TSolver::solve();
}

//
// update pX on all levels by restricting fine grid pX to coarser levels
//
template <int N>
void 
TMultigrid<N>::updateXHierarchy()
{
    for (uint i = high; i > low; i--)
        *XH[i-1] = pRestrict[i-1]->restrict( *XH[i] );
}

//
// add/subtract multiple of identity to all levels
//
template <int N>
void 
TMultigrid<N>::addIdentity( const TFloat factor )
{
    for (uint i = high; i > low; i--)
        AH[i]->addIdentity( factor );
}

template <int N>
void 
TMultigrid<N>::subtractIdentity( const TFloat factor )
{
    for (uint i = high; i > low; i--)
        AH[i]->subtractIdentity( factor );
}

//
// set smoother		
//
template <int N>
void 
TMultigrid<N>::setSmoother ( TSolver* apSmoother ) 
{
    pPreSmoother = apSmoother;
    pPostSmoother = apSmoother;
    if (pPreSmoother != NULL)
    {
        TString newName;
        
        newName  = "MG ";
        newName += "(";
        newName += pPreSmoother->getName();
        newName += ")";
        setName( newName ); 
    }// if
}

template class TMultigrid<3>;
