//
//  Project   : libnet
//  File      : blocksolver.cc
//  Author    : Sabine Gutsch / Thomas Probst
//  Purpose   : block solver based on concentric cycles
//
// arch-tag: 660964c8-d343-4c18-8f65-7d79b052cd42
//

#include "blocksolver.hh"

//////////////////////////////////////////////////
//
// class for block lower Gauss-Seidel iteration (based on concentric 
// cycle decomposition)
//

// constructor
TBlockForwardGS::TBlockForwardGS ( TBlockMatrix *paA, 
                                   TBlockVector<TFloat> *paRHS, 
                                   TBlockVector<TFloat> *paX,  
                                   uint amaxiter, 
                                   TFloat aaccur, 
                                   uint averb, TFloat aDamp ) 
        : TSecondNF ( paA, paRHS, paX, amaxiter, aaccur, averb, aDamp ) 
{ 
	setName("Bl F-GS"); 
}

// calculate WInvers
TBlockVector<TFloat> 
TBlockForwardGS::WInvers ( const TBlockVector<TFloat> &v ) 
{ 
	if ( (*pA)[0][0].getDecompType() != BLOCK_CYC_LU ){
      //	cerr << "TBlockForwardGS::WInvers: no BLOCK_CYC_LU decomposition!\n";
      return (pA->invLowerMult( v ));}
	else 
		return (pA->invBlockLowerMult( v )); 
	return ( v );
}

//////////////////////////////////////////////////
//
// class for block upper Gauss-Seidel iteration (based on concentric 
// cycle decomposition)

// constructor
TBlockBackwardGS::TBlockBackwardGS ( TBlockMatrix *paA, 
									 TBlockVector<TFloat> *paRHS, 
									 TBlockVector<TFloat> *paX, 
									 uint amaxiter, 
									 TFloat aaccur, 
									 uint averb, TFloat aDamp ) 
        : TSecondNF ( paA, paRHS, paX, amaxiter, aaccur, averb, aDamp ) 
{ 
	setName("Bl B-GS"); 
}

// calculate WInvers
TBlockVector<TFloat> 
TBlockBackwardGS::WInvers ( const TBlockVector<TFloat> &v ) 
{ 
	if ( (*pA)[0][0].getDecompType() != BLOCK_CYC_LU ){
		//cerr << "TBlockBackwardGS::WInvers: no BLOCK_CYC_LU decomposition!\n";
		return (pA->invUpperMult( v ));}
    else
		return (pA->invBlockUpperMult( v ));
	return ( v );
}

//////////////////////////////////////////////////
//
// class for sym GS
//

// constructor
TBlockSymmetricGS::TBlockSymmetricGS ( TBlockMatrix *paA, 
                                       TBlockVector<TFloat> *paRHS, 
                                       TBlockVector<TFloat> *paX, 
                                       uint amaxiter, 
                                       TFloat aaccur, 
                                       uint averb, TFloat aDamp ) 
        : TSecondNF ( paA, paRHS, paX, amaxiter, aaccur, averb, aDamp ) 
{ 
  setName("Bl SGS"); 
}

// calculate WInvers
TBlockVector<TFloat> 
TBlockSymmetricGS::WInvers ( const TBlockVector<TFloat> &v ) 
{ 
	if ( (*pA)[0][0].getDecompType() != BLOCK_CYC_LU ){
		//cerr << "TBlockSymmetricGS::WInvers: no BLOCK_CYC_LU decomposition!\n";
      return (pA->invUpperMult( pA->diagMult(pA->invLowerMult( v ))));
   }
	else
		return (pA->invBlockUpperMult( pA->blockBlockDiagMult( 
			pA->invBlockLowerMult( v ))));
	return ( v );
}

TBlockBackSymmetricGS::TBlockBackSymmetricGS ( TBlockMatrix *paA, 
                                               TBlockVector<TFloat> *paRHS, 
                                               TBlockVector<TFloat> *paX, 
                                               uint amaxiter, 
                                               TFloat aaccur, 
                                               uint averb, TFloat aDamp ) 
        : TSecondNF ( paA, paRHS, paX, amaxiter, aaccur, averb, aDamp ) 
{ 
  setName("Bl SGS (back/forw)"); 
}

// calculate WInvers
TBlockVector<TFloat> 
TBlockBackSymmetricGS::WInvers ( const TBlockVector<TFloat> &v ) 
{ 
	if ( (*pA)[0][0].getDecompType() != BLOCK_CYC_LU ){
          //cerr << "TBlockBackSymmetricGS::WInvers: no BLOCK_CYC_LU decomposition!\n";
      return (pA->invLowerMult( pA->diagMult(pA->invUpperMult( v ))));
   }
	else
          return (pA->invBlockLowerMult( pA->blockBlockDiagMult( 
           pA->invBlockUpperMult( v ))));

	return ( v );
}





