//
//  Project  : libnet
//  File     : fastGaussSeidel.cc
//  Author   : Sabine Gutsch
//  Purpose  : fastGaussSeidel solver
//
// arch-tag: de736288-e894-4882-be8e-78a8e5e77624
//

#include "fastGaussSeidel.hh"

//
// constructor
//
TFastForwardGaussSeidel::TFastForwardGaussSeidel( TBlockMatrix *paA, 
                                                  TBlockVector<TFloat> *paRHS, 
                                                  TBlockVector<TFloat> *paX, 
                                                  uint amaxiter,
                                                  TFloat aaccur,
                                                  uint averb, TFloat aDamp ) 
        : TSolver ( paA, paRHS, paX, amaxiter, aaccur, averb) 
{
    setName("F-FGS"); 
    damp=aDamp;
    if(paRHS) 
        if (paRHS->getNoOfBlocks()>1)
        {cerr<<" FastForwardGaussSeidel only for single  block matrices"<<endl; exit(0);}
  
}

//
// one iteration step
//
int 
TFastForwardGaussSeidel::step () 
{ 
    uint  j;
    for ( j = 0; j < (*pRHS)[0].getSize(); j++)
        (*pX)[0][j]+=( (*pRHS)[0][j] - (*pA)[0][0][j]*((*pX)[0]))/ ((*pA)[0][0][j][j]);
    return (RETURN_CONTINUE);
}



//
// constructor
//
TFastBackwardGaussSeidel::TFastBackwardGaussSeidel( TBlockMatrix *paA, 
                                                    TBlockVector<TFloat> *paRHS, 
                                                    TBlockVector<TFloat> *paX, 
                                                    uint amaxiter,
                                                    TFloat aaccur,
                                                    uint averb, TFloat aDamp ) 
        : TSolver ( paA, paRHS, paX, amaxiter, aaccur, averb) 
{
    setName("F-BGS"); 
    damp=aDamp;
    if(paRHS) 
        if (paRHS->getNoOfBlocks()>1)
        {
            cerr<<" FastBackwardGaussSeidel only for single  block matrices"<<endl;
            exit(0);
        }
}

//
// one iteration step
//
int 
TFastBackwardGaussSeidel::step () 
{ 
    long  j;
    for ( j =(*pRHS)[0].getSize()-1; j >=0 ; j--)
        (*pX)[0][j]+=( (*pRHS)[0][j] - (*pA)[0][0][j]*((*pX)[0]))/ ((*pA)[0][0][j][j]);
    return (RETURN_CONTINUE);
}



//
// constructor symmetric
//
TFastSymmetricGaussSeidel::TFastSymmetricGaussSeidel( TBlockMatrix *paA, 
                                                      TBlockVector<TFloat> *paRHS, 
                                                      TBlockVector<TFloat> *paX, 
                                                      uint amaxiter,
                                                      TFloat aaccur,
                                                      uint averb, TFloat aDamp ) 
        : TSolver ( paA, paRHS, paX, amaxiter, aaccur, averb) 
{
    setName("F-SGS"); 
    damp=aDamp;    if(paRHS) 
        if (paRHS->getNoOfBlocks()>1)
        {
            cerr<<" FastSymmetricGaussSeidel only for single  block matrices"<<endl;
            exit(0);
        }
}

//
// one iteration step
//
int 
TFastSymmetricGaussSeidel::step () 
{ 
    long  j;
    long  N = (*pRHS)[0].getSize();
    
    for ( j = 0; j <N; j++)
        (*pX)[0][j] = ( (*pRHS)[0][j] -
                        (*pA)[0][0][j]*((*pX)[0])) /
            ((*pA)[0][0][j][j])+ (*pX)[0][j];

    for ( j =N-1; j >=0 ; j--)
        (*pX)[0][j] = ( (*pRHS)[0][j] -
                        (*pA)[0][0][j]*((*pX)[0])) /
            ((*pA)[0][0][j][j])+ (*pX)[0][j];
    
    return (RETURN_CONTINUE);
}

//
// constructor symmetric
//
TFastBackSymmetricGaussSeidel::TFastBackSymmetricGaussSeidel( TBlockMatrix *paA, 
                                                      TBlockVector<TFloat> *paRHS, 
                                                      TBlockVector<TFloat> *paX, 
                                                      uint amaxiter,
                                                      TFloat aaccur,
                                                      uint averb, TFloat aDamp ) 
        : TSolver ( paA, paRHS, paX, amaxiter, aaccur, averb) 
{
    setName("F-BF-SGS"); 
    damp=aDamp;    if(paRHS) 
        if (paRHS->getNoOfBlocks()>1)
        {
            cerr<<" FastBackSymmetricGaussSeidel only for single  block matrices"<<endl;
            exit(0);
        }
}

//
// one iteration step
//
int 
TFastBackSymmetricGaussSeidel::step () 
{ 
    long  j;
    long  N = (*pRHS)[0].getSize();
    
    // first backward step, then forward step
    for ( j =N-1; j >=0 ; j--)
        (*pX)[0][j] = ( (*pRHS)[0][j] -
                        (*pA)[0][0][j]*((*pX)[0])) /
            ((*pA)[0][0][j][j])+ (*pX)[0][j];
    
    for ( j = 0; j <N; j++)
        (*pX)[0][j] = ( (*pRHS)[0][j] -
                        (*pA)[0][0][j]*((*pX)[0])) /
            ((*pA)[0][0][j][j])+ (*pX)[0][j];

    
    return (RETURN_CONTINUE);
}


///////////////////////////////////////////////

//
// class for block lower Gauss-Seidel iteration (based on concentric 
// cycle decomposition)
//

// constructor
TFastBlockForwardGS::TFastBlockForwardGS ( TBlockMatrix *paA, 
                                           TBlockVector<TFloat> *paRHS, 
                                           TBlockVector<TFloat> *paX,  
                                           uint amaxiter, 
                                           TFloat aaccur, 
                                           uint averb, TFloat aDamp ) 
        : TSolver ( paA, paRHS, paX, amaxiter, aaccur, averb) 
{  
    setName("F-Block FGS"); 
    if(paRHS) 
        if (paRHS->getNoOfBlocks()>1)
        {
            cerr<<" FastSymmetricGaussSeidel only for single  block matrices"<<endl;
            exit(0);
        }

    damp=aDamp;
}

int 
TFastBlockForwardGS::step ( )
{
    uint j,N;
    uint i;
    uint startIndex,endIndex,blockSize,noOfDecompBlocks;
    noOfDecompBlocks= (*pA)[0][0].getNoOfDecompBlocks();

    N = (*pRHS)[0].getSize();


    if (noOfDecompBlocks == 0 ) { // normal Fast forward  GS 
        for ( j = 0; j <N ; j++)
            (*pX)[0][j]+=( (*pRHS)[0][j] - (*pA)[0][0][j]*((*pX)[0]))/ ((*pA)[0][0][j][j]);
    }
    else{
        TVectorBlock<uint>*   decompBlocks;

        decompBlocks = new TVectorBlock<uint>( *((*pA)[0][0].getDecompBlocksPointer()));
        
        TVectorBlock<TFloat>  vecBlock((*pX)[0]);

        // before the first  block
        startIndex = 0;
        endIndex   = (*decompBlocks)[0] ;
    
        for ( i = 0; i < noOfDecompBlocks; i++ )
        {
            // first pointwise Gauss-Seidel for entries before
            // current block
            // real block method: pointwise backward in the single blocks
      
            for ( j =endIndex; j > startIndex;  j-- )
                (*pX)[0][j-1] += ( (*pRHS)[0][j-1] -
                                   (*pA)[0][0][j-1]*((*pX)[0])) /
                    ((*pA)[0][0][j-1][j-1]);
      
            // and now treat the block
            // blocksize of current block
            startIndex = (*decompBlocks)[2*i];
            endIndex   = (*decompBlocks)[2*i+1];
            blockSize = endIndex - startIndex + 1;
      
            vecBlock.setSize( blockSize );
            // 1.: defect
            for ( j = startIndex; j <= endIndex; j++ )
                vecBlock[j-startIndex] = (*pRHS)[0][j] - (*pA)[0][0][j]*((*pX)[0]);
      
            // apply inverse of block diagonal
            vecBlock =  (*pA)[0][0].getDiagBlock(i)->solveDecomp( vecBlock );
      
            // update x
            for ( j = 0; j < blockSize; j++ )
                ((*pX)[0])[startIndex + j] += vecBlock[j];
      
            startIndex = endIndex + 1;
            if ( i < (noOfDecompBlocks - 1) ) 
                endIndex = (*decompBlocks)[2*(i+1)];
        }    
        
    
        // remaining entries: pointwise (backward) Gauss-Seidel
        for ( j = N; j >startIndex; j-- )
            (*pX)[0][j-1] += ( (*pRHS)[0][j-1] -
                               (*pA)[0][0][j-1]*((*pX)[0])) /
                ((*pA)[0][0][j-1][j-1]);
    
    }
    return (RETURN_CONTINUE);
} // step FastBlockForward GS


//////////////////////////////////////////////////
//
// class for block upper Gauss-Seidel iteration (based on concentric 
// cycle decomposition)

// constructor
TFastBlockBackwardGS::TFastBlockBackwardGS ( TBlockMatrix *paA, 
                                             TBlockVector<TFloat> *paRHS, 
                                             TBlockVector<TFloat> *paX, 
                                             uint amaxiter, 
                                             TFloat aaccur, 
                                             uint averb, TFloat aDamp ) 
        : TSolver ( paA, paRHS, paX, amaxiter, aaccur, averb) 
{ 
	setName("F-Block BGS"); 
    damp=aDamp;
}

// calculate step
int
TFastBlockBackwardGS::step ()
{
    uint j,N;
    uint i;
    uint startIndex,endIndex,blockSize,noOfDecompBlocks;
    noOfDecompBlocks= (*pA)[0][0].getNoOfDecompBlocks();

    N=(*pRHS)[0].getSize();



    if (noOfDecompBlocks == 0 ) { // normal Fast forward  GS 
        for ( j = N; j >0 ; j--)
            (*pX)[0][j-1] += ( (*pRHS)[0][j-1] -
                               (*pA)[0][0][j-1]*((*pX)[0])) /
                ((*pA)[0][0][j-1][j-1]);
    }
    else{
        TVectorBlock<uint>*   decompBlocks;

        decompBlocks = new TVectorBlock<uint>( *((*pA)[0][0].getDecompBlocksPointer()));
        
        TVectorBlock<TFloat>  vecBlock((*pX)[0]);
     
        // blocksize of current block
        startIndex = N-1;
        endIndex   = (*decompBlocks)[2*(noOfDecompBlocks-1)+1];

   
        for ( i = noOfDecompBlocks; i > 0; i-- )
        {
            // cout<<"    TFastBlockBackwardGS::step: treating block "<< i<<endl;
    
            // first pointwise Gauss-Seidel for entries before
            // current block
            // real block: pointwise backward in the single blocks
    
            for ( j = startIndex; j > endIndex; j-- )
                (*pX)[0][j] += ( (*pRHS)[0][j] -
                                 (*pA)[0][0][j]*((*pX)[0])) /
                    ((*pA)[0][0][j][j]);
      
            // and now treat the block
            // blocksize of current block
            startIndex = (*decompBlocks)[2*(i-1)+1];
            endIndex   = (*decompBlocks)[2*(i-1)];
            blockSize = startIndex - endIndex + 1;
    
            // 1.: defect
            vecBlock.setSize( blockSize );
            for ( j = endIndex; j <= startIndex; j++ )
                vecBlock[j-endIndex] = (*pRHS)[0][j] - (*pA)[0][0][j]*((*pX)[0]);
    
            // apply inverse of block diagonal
            vecBlock =  (*pA)[0][0].getDiagBlock(i-1)->solveDecomp( vecBlock );
    
            // update x
            for ( j = 0; j < blockSize; j++ )
                ((*pX)[0])[endIndex + j] += vecBlock[j];
    
            startIndex = endIndex - 1;
            if ( i > 1 ) endIndex = (*decompBlocks)[2*(i-2)+1];
    
        }    
  
    
        // remaining entries: pointwise (backward) Gauss-Seidel
        for ( j =  (*decompBlocks)[0]; j >0; j--)
            (*pX)[0][j-1] += ( (*pRHS)[0][j-1] -
                               (*pA)[0][0][j-1]*((*pX)[0])) /
                ((*pA)[0][0][j-1][j-1]);
    }
    return (RETURN_CONTINUE);
} // step FastBlockBackward GS

//////////////////////////////////////////////////
//
// class for sym GS
//

// constructor
TFastBlockSymmetricGS::TFastBlockSymmetricGS ( TBlockMatrix *paA, 
                                               TBlockVector<TFloat> *paRHS, 
                                               TBlockVector<TFloat> *paX, 
                                               uint amaxiter, 
                                               TFloat aaccur, 
                                               uint averb, TFloat aDamp ) 
        : TSolver ( paA, paRHS, paX, amaxiter, aaccur, averb ) 
{ 
    setName("F-Block SGS"); 
    damp=aDamp;
    For = new TFastBlockForwardGS( paA, paRHS, paX, amaxiter, aaccur, averb,aDamp);
    Back = new TFastBlockBackwardGS( paA, paRHS, paX, amaxiter, aaccur, averb,aDamp);
}

// calculate step
int 
TFastBlockSymmetricGS::step () 
{ 
    For->setA(pA );
    For->setRHS(pRHS );
    For->setX(pX);
    Back->setA(pA );
    Back->setRHS(pRHS );
    Back->setX(pX);
  
    For->step();
    Back->step();
    return (RETURN_CONTINUE);
}

// destructor
TFastBlockSymmetricGS::~TFastBlockSymmetricGS()
{
    delete For;
    delete Back;
}


TFastBlockBackSymmetricGS::TFastBlockBackSymmetricGS ( TBlockMatrix *paA, 
                                                       TBlockVector<TFloat> *paRHS, 
                                                       TBlockVector<TFloat> *paX, 
                                                       uint amaxiter, 
                                                       TFloat aaccur, 
                                                       uint averb, TFloat aDamp ) 
        : TSolver ( paA, paRHS, paX, amaxiter, aaccur, averb) 
{ 
    setName("F- Block SGS (back/forw)"); 
    damp=aDamp;
    For = new TFastBlockForwardGS( paA, paRHS, paX, amaxiter, aaccur, averb,aDamp);
    Back = new TFastBlockBackwardGS( paA, paRHS, paX, amaxiter, aaccur, averb,aDamp);


}

// destructor
TFastBlockBackSymmetricGS::~TFastBlockBackSymmetricGS()
{
    delete For;
    delete Back;
}

// calculate step
int
TFastBlockBackSymmetricGS::step() 
{
    For->setA(pA );
    For->setRHS(pRHS );
    For->setX(pX);
    Back->setA(pA );
    Back->setRHS(pRHS );
    Back->setX(pX);
    
    Back->step();
    For->step();
    return (RETURN_CONTINUE);
}
