//
//  Project  : libnet
//  File     : optSquaredSmoother.cc
//  Author   : Sabine Gutsch 
//  Purpose  : smoother for a general system
//
// arch-tag: be3e11ff-11ca-4cc2-9d33-4be05d1b82ea
//

extern "C" {
#include <math.h>
}

#include "optSquaredSmoother.hh"

//
// constructor
//
TOptSquaredSmoother::TOptSquaredSmoother ( TBlockMatrix *paA = NULL, 
                 TBlockVector<TFloat> *paRHS = NULL, 
				 TBlockVector<TFloat> *paX = NULL, 
				 uint amaxiter = DEFAULT_MAX_ITERATION,
				 TFloat aaccur = DEFAULT_ACCURACY, 
				 uint averb = 0 )
        : TSolver ( paA, paRHS, paX, amaxiter, aaccur, averb ) 
{ 
	setName("Optimal squared smoother");
}
 
//
// initialize 
//
void 
TOptSquaredSmoother::init ()
{
	r = (*pA) * (*pX); 
	r -= (*pRHS);        // r = Ax - b
}

//
// one iteration step
//
int
TOptSquaredSmoother::step ( )
{     
    uint n = (*pA)[0][0].getNoOfRows();
    a0.setSize( n );
    a1.setSize( n );

    (*pA)[0][0].transposeMult( r[0], a0 );
    (*pA)[1][1].transposeMult( r[1], a1 );
    b0 = (*pA)[0][2] * r[2];
    b1 = (*pA)[1][2] * r[2];
    c = (*pA)[2][0] * r[0];
    c += (*pA)[2][1] * r[1];
    
    d0 = (*pA)[0][0] * a0;
    d1 = (*pA)[1][1] * a1;
    e0 = (*pA)[0][0] * b0;
    e1 = (*pA)[1][1] * b1;

    f0 = (*pA)[0][2] * c;
    f1 = (*pA)[1][2] * c;
    g = (*pA)[2][0] * a0;
    g += (*pA)[2][1] * a1;
    h = (*pA)[2][0] * b0;
    h += (*pA)[2][1] * b1;
    k0 = e0 + f0;
    k1 = e1 + f1;

    alpha   = k0*k0 + k1*k1 + h*h;
    beta    = r[0]*d0 + r[1]*d1 + r[2]*g;
    gamma   = d0*k0 + d1*k1 + g*h;
    delta   = r[0]*k0 + r[1]*k1 + r[2]*h;
    epsilon = d0*d0 + d1*d1 + g*g;

	
    if (fabs(epsilon * alpha - gamma * gamma) < 1.0e-15)
	{
        lambda = 0.5;
        sigma = (beta + lambda * delta) / 
        (epsilon + 2*lambda*gamma + lambda*lambda*alpha);
lambda *= sigma;
	}
    else
	{
        TFloat zeta = (epsilon * alpha - gamma * gamma);
        //cout << "zeta: " << zeta << endl;
        sigma = (alpha / zeta) * beta - (gamma / zeta) * delta;
        lambda = delta / alpha - gamma * sigma / alpha;
	}

    if ((sigma <= 0) || lambda <= 0) {
        cout << "Error: Negative factors!   " << sigma << "  " << lambda << "   rABr  " << (1.0 + lambda) * (g*r[2])+a0*a0+a1*a1+b0*b0+b1*b1+c*c << "    " << (g*r[2]+a0*a0+a1*a1+b0*b0+b1*b1+c*c) / (g*r[2]) << endl;
        lambda = 50.0;
        sigma = (beta + lambda * delta) / 
        (epsilon + 2*lambda*gamma + lambda*lambda*alpha);
lambda *= sigma;
	}
	else { cout << "Positive factors! " << (g*r[2]+a0*a0+a1*a1+b0*b0+b1*b1+c*c) / (g*r[2]) << "  " << (1.0 + lambda) * (g*r[2])+a0*a0+a1*a1+b0*b0+b1*b1+c*c << endl;}
	/*
lambda = 49.0*49.0;
sigma = 1.0 / (12.37*12.37);
sigma = (beta + lambda * delta) / 
        (epsilon + 2*lambda*gamma + lambda*lambda*alpha);
lambda *= sigma;
*/
	//cout << "sigma, lambda: " << sigma << "  " << lambda/sigma << endl;

    a0 *= sigma;
    b0 *= lambda;
    d0 *= sigma;
    e0 *= lambda;
    f0 *= lambda;
    a1 *= sigma;
    b1 *= lambda;
    d1 *= sigma;
    e1 *= lambda;
    f1 *= lambda;
    c *= lambda;
    g *= sigma;
    h *= lambda;

    (*pX)[0] -= a0;
    (*pX)[0] -= b0;
    (*pX)[1] -= a1;
    (*pX)[1] -= b1;
    (*pX)[2] -= c;

    r[0] -= d0;
    r[0] -= e0;
    r[0] -= f0;
    r[1] -= d1;
    r[1] -= e1;
    r[1] -= f1;
    r[2] -= g;
    r[2] -= h;
 
	// pressure correction
	TVectorBlock<TFloat> vecOne( r[2] );
    vecOne.set( 1.0 );
    TFloat factor = (vecOne * (*pX)[2]) / (vecOne * vecOne);
	//cout << "factor: " << factor << endl;
    vecOne.set( factor );
    (*pX)[2] -= vecOne;

    return (RETURN_CONTINUE);
}
