//
//  Project  : libnet
//  File     : schur.cc
//  Author   : Sabine Gutsch / Thomas Probst
//  Purpose  : Schur complement solver in second normal form
//
// arch-tag: 7c38c2c9-a723-45b5-80fb-efb0e0f2358e
//

#include <math.h>

#include "schur.hh"

//
// constructor
//
TSchur::TSchur ( TBlockMatrix *paA, 
                 TBlockVector<TFloat> *paRHS, 
				 TBlockVector<TFloat> *paX, 
				 uint amaxiter,
				 TFloat aaccur, 
				 uint averb,
				 TFloat aDamp )
        : TSecondNF ( paA, paRHS, paX, amaxiter, aaccur, averb, aDamp) 
{ 
	setName("Block Schur complement");
    maxInnerSteps  = 1;            // max no of steps per outer step
    totalInnerSteps  = 0;          // counter for inner steps
    innerReduceBy = 1.0e-2;
    problemType = UNSYMMETRIC;     // 1: symmetric, 0: nonsymmetric
    dimSchurComplement = 0; 
    tol = 1.0e-20;                 // tolerance to detect breakdowns 
    innerGS = BACKWARD;            // 1: symmetric, 0: backward
    innerGS = TWOBACKWARD;         // 2: two backward steps
    //innerGS = SYMMETRIC;          
}

//
// calculate WInvers
//
TBlockVector<TFloat> 
TSchur::WInvers ( const TBlockVector<TFloat> &r )
{ 
    if ( r.getNoOfBlocks() == 1 ) return (WInversScalar( r )); 
    return (WInversStokes( r ));
}

//
// calculate WInversStokes
//
TBlockVector<TFloat> 
TSchur::WInversStokes ( const TBlockVector<TFloat> &r )
{ 
    uint j;
	uint nob = r.getNoOfBlocks();

	TBlockVector<TFloat> vecTmp( r );
	TBlockVector<TFloat> vecTmp2( r );
	TVectorBlock<TFloat> v3( vecTmp[0] );


	// vecTmp[0] = C^{-1} r[0]
    if ( innerGS == SYMMETRIC )  // symmetric Gauss-Seidel
	    for ( j = 0; j < (nob-1); j++ ) 
		{
		    vecTmp[j] = (*pA)[j][j].invUpperMult(vecTmp[j]);
	        vecTmp[j] = (*pA)[j][j].diagMult(vecTmp[j]);
		    vecTmp[j] = (*pA)[j][j].invLowerMult(vecTmp[j]);
		}
    else  // backward Gauss-Seidel
	{
	    for ( j = 0; j < (nob-1); j++ ) 
		    vecTmp[j] = (*pA)[j][j].invUpperMult(vecTmp[j]);
        if ( innerGS == TWOBACKWARD )  // a second backward step
		{
	        for ( j = 0; j < (nob-1); j++ ) 
		        vecTmp2[j] = (*pA)[j][j] * (vecTmp[j]);
	        for ( j = 0; j < (nob-1); j++ ) 
		    {
		        vecTmp2[j] = (*pA)[j][j].invUpperMult(vecTmp2[j]);
                vecTmp[j] *= 2;
                vecTmp[j] -= vecTmp2[j];
		    }
		}
	}

	// vecTmp[1] = B * vecTmp[0] - r[1]
	vecTmp[nob-1] *= (-1);
	for ( j = 0; j < (nob-1); j++ )
		vecTmp[nob-1] += (*pA)[nob-1][j] * vecTmp[j]; 
	
    // vecTmp[nob-1] *= (alpha);

	// vecTmp[1] = S^{-1} vecTmp[1]
    if ( problemType == SYMMETRIC ) 
	    vecTmp[nob-1] = innerSchurSymm( vecTmp[nob-1] );
    else 
        vecTmp[nob-1] = innerSchurNonSymm( vecTmp[nob-1] );
			
	// vecTmp[0] = vecTmp[0] - C^{-1} B^T vecTmp[1]
	for ( j = 0; j < (nob-1); j++ )
	{
		v3 = (*pA)[j][nob-1] * vecTmp[ nob - 1 ];
        v3 = (*pA)[j][j].invUpperMult( v3 );

        if ( innerGS == TWOBACKWARD )  // a second backward step
		{ 
	        vecTmp2[j] = (*pA)[j][j] * v3;
	        vecTmp2[j] = (*pA)[j][j].invUpperMult(vecTmp2[j]);
            v3 *= 2;
            v3 -= vecTmp2[j];
		}
        if ( innerGS == SYMMETRIC ) 
		{
            v3 = (*pA)[j][j].diagMult( v3 );
		    v3 = (*pA)[j][j].invLowerMult( v3 );
		}
        //v3 *= alpha;
        vecTmp[j] -= v3;
	}// for    
		
    //vecTmp[0] *= alpha;
    //vecTmp[1] *= alpha;

	return (vecTmp);
}

//
// perform inner iterations to solve for Schur complement
//
TVectorBlock<TFloat>
TSchur::innerSchurSymm( TVectorBlock<TFloat> &b )
{
	// cg method to solve Schur complement equation
    TFloat lambda, rhoNew, rhoOld;
	uint nob = pX->getNoOfBlocks();
	TSparseBlockVector sparseRowVec( nob );
    TVectorBlock<TFloat> vecDiag( 0.0, b.getSize() );
    uint i, j;
	TBlockVector<TFloat> vecBlocks( *pX );

    uint blockSize[ nob ];
    for ( i = 0; i < nob; i++ ) blockSize[i] = (*pX)[i].getSize();

	uint size = b.getSize();
	TVectorBlock<TFloat> vec_x( 0.0, size );
	TVectorBlock<TFloat> vec_a( 0.0, size );
	TVectorBlock<TFloat> vec_p( 0.0, size );
	TVectorBlock<TFloat> vec_q( 0.0, size );
	TVectorBlock<TFloat> vec_r( 0.0, size );
	
	// approximation of S^{-1}
	for ( i = 0; i < size; i++ )
	{		
		pA->getTransposedRow( sparseRowVec, nob-1, i );
		/*
        for ( j = 0; j < sparseRowVec[0].getNoOfEntries(); j++ ) {
            sparseRowVec[0].setValue( j, (*pA)[0][0][j][j] * sparseRowVec[0].getValue(j) );
            sparseRowVec[1].setValue( j, (*pA)[1][1][j][j] * sparseRowVec[1].getValue(j) );
		}

		for ( j = 0; j < (nob - 1); j++ ) 
		    vecDiag[i] +=  (sparseRowVec[j] * sparseRowVec[j] );
		*/
        vecDiag[i] = 1.0;
	}
	
	// initialization
    vec_r = b;
	for ( j = 0; j < size; j++ )
		vec_p[j] = vec_r[j] / vecDiag[j] ;
    rhoOld = vec_p * vec_r;

    TFloat initialRes = (vec_r * vec_r) / vec_r.getSize();
    TFloat currentRes = initialRes;
    i = 0;

	// cg method
    while ( (i++ < maxInnerSteps) && (currentRes > innerReduceBy * initialRes) ) 
	{
        totalInnerSteps++;
        // S vec_p
	    for ( j = 0; j < nob-1; j++ )
                vecBlocks[j] = (*pA)[j][nob-1] * vec_p;

		// set Lagrange component to zero
        //vecBlocks[ nob - 1 ][size-1] = 0.0;
			
		for ( j = 0; j < nob-1; j++ ) {
	        vecBlocks[j] = (*pA)[j][j].invUpperMult( vecBlocks[j]);
			vecBlocks[j] = (*pA)[j][j].diagMult( vecBlocks[j] );
			vecBlocks[j] = (*pA)[j][j].invLowerMult(vecBlocks[j]);
		}

        vec_a = (*pA)[nob-1][0] * vecBlocks[0];
        vec_a += (*pA)[nob-1][1] * vecBlocks[1];
				
        lambda = rhoOld / (vec_a * vec_p);
        vec_p *= lambda;
        vec_x += vec_p;
        vec_a *= lambda;
        vec_r -= vec_a;
        currentRes = (vec_r * vec_r) / vec_r.getSize();

        //if ( i == innerSteps-1 || i == 0) 
		//    cout << "schur " << i << " " << vec_r * vec_r << endl;

        // W^{-1} r
	    for ( j = 0; j < size; j++ )
		    vec_q[j] = vec_r[j] / vecDiag[j] ;

        rhoNew = vec_q * vec_r;

        vec_p *= (rhoNew / (rhoOld * lambda) );
        vec_p += vec_q;
 
        rhoOld = rhoNew;
	}
	//cout << i-1 << "   " << currentRes << endl;
    return ( vec_x );
}


//
// perform inner iterations to solve for Schur complement
//
TVectorBlock<TFloat> 
TSchur::innerSchurNonSymm( TVectorBlock<TFloat> &b )
{
	// bicgStab method to solve Schur complement equation
    TFloat rhoOld, rhoNew, alpha, omega, beta;
	uint nob = pX->getNoOfBlocks();
	TSparseBlockVector sparseRowVec( nob );
    TVectorBlock<TFloat> vecDiag( 0.0, b.getSize() );
    uint i, j;
	TBlockVector<TFloat> vecBlocks( *pX );
	TBlockVector<TFloat> vecTmp2( *pX );

    uint blockSize[ nob ];
    for ( i = 0; i < nob; i++ ) blockSize[i] = (*pX)[i].getSize();

	uint size = b.getSize();
	TVectorBlock<TFloat> vec_x( 0.0, size );
	TVectorBlock<TFloat> vec_best_x( 0.0, size );
	TVectorBlock<TFloat> vec_p( 0.0, size );
	TVectorBlock<TFloat> vec_y( 0.0, size );
	TVectorBlock<TFloat> vec_v( 0.0, size );
	TVectorBlock<TFloat> vec_s( 0.0, size );
	TVectorBlock<TFloat> vec_z( 0.0, size );
	TVectorBlock<TFloat> vec_t( 0.0, size );
	TVectorBlock<TFloat> vec_r( 0.0, size );
	TVectorBlock<TFloat> vec_rp( 0.0, size );
	
	// diagonal approximation of S
	
	/*
	for ( i = 0; i < size; i++ )
	{		
		pA->getTransposedRow( sparseRowVec, nob-1, i );
		for ( j = 0; j < (nob - 1); j++ ) 
		    vecDiag[i] +=  ( sparseRowVec[j] * sparseRowVec[j] );
	}
	*/
	
    if (pA->hasLUdecomp() != true) pA->calculateILUdecomp();
	for ( i = 0; i < size; i++ )
	    vecDiag[i] = (*(pA->getLUdecomposition()))[nob-1][nob-1][i][i];

	// initialization
    vec_r = b;
    TFloat initialRes = (vec_r * vec_r) / vec_r.getSize();
    TFloat currentRes = initialRes;
    TFloat bestRes = initialRes;
    vec_rp = b; 
    rhoOld = 1.0; 
    omega = 1.0;
    alpha = 1.0;
    i = 0; 

	// bicgStab method
    while ( (i++ < maxInnerSteps) && (currentRes > innerReduceBy * initialRes) ) 
	{
        totalInnerSteps++;
        rhoNew = vec_r * vec_rp;
        if (fabs(rhoOld) < tol || fabs(omega) < tol)
		{ 
            //innerSteps = 1;
            cout << "bicg breakdown0 after " << i << " steps " << (vec_r * vec_r) / vec_r.getSize() << " " << endl;
            return (vec_best_x);
		}
        beta = (rhoNew / rhoOld) * (alpha / omega);

        vec_v *= (-omega);
        vec_p += vec_v;
        vec_p *= beta;
        vec_p += vec_r;


        // W^{-1} p
        if (0 == 1) 
        {
			// diagonal preconditioning
	        for ( j = 0; j < size; j++ ) 
		        vec_y[j] = vec_p[j] / vecDiag[j] ;
		}
        else
		{
			// ILU-preconditioning
		    vec_y = (*pA->getLUdecomposition())[nob-1][nob-1].invUpperMult( vec_p );
		    vec_y = (*pA->getLUdecomposition())[nob-1][nob-1].invUnitLowerMult( vec_y );
		}

        //vec_v = S vec_y 
	    for ( j = 0; j < nob-1; j++ )
            vecBlocks[j] = (*pA)[j][nob-1] * vec_y;

		for ( j = 0; j < nob-1; j++ ) {
		    vecBlocks[j] = (*pA)[j][j].invUpperMult( vecBlocks[j]);
            if ( innerGS == TWOBACKWARD )  // a second backward step
			{
	            vecTmp2[j] = (*pA)[j][j] * (vecBlocks[j]);
	            vecTmp2[j] = (*pA)[j][j].invUpperMult(vecTmp2[j]);
                vecBlocks[j] *= 2;
                vecBlocks[j] -= vecTmp2[j];
		    }
		
            if ( innerGS == SYMMETRIC ) // symmetric Gauss-Seidel
			{
			    vecBlocks[j] = (*pA)[j][j].diagMult( vecBlocks[j] );
			    vecBlocks[j] = (*pA)[j][j].invLowerMult(vecBlocks[j]);
			}
	    }

        vec_v = (*pA)[nob-1][0] * vecBlocks[0];
		for ( j = 1; j < nob-1; j++ ) 
            vec_v += (*pA)[nob-1][j] * vecBlocks[j];
  	
        omega = vec_rp * vec_v;
        if (fabs(omega) < tol)
		{ 
            cout << "bicg breakdown1 after " << i << " steps " << (vec_r * vec_r) / vec_r.getSize() << endl;
            return (vec_best_x);
		}

        alpha = rhoNew / omega;
        vec_s = vec_v;
        vec_s *= (-alpha);
        vec_s += vec_r;

        // W^{-1} s
        if (0 == 1) 
        {
			// diagonal preconditioning
	        for ( j = 0; j < size; j++ ) 
		        vec_z[j] = vec_s[j] / vecDiag[j] ;
		}
        else
		{
			// ILU-preconditioning
		    vec_z = (*pA->getLUdecomposition())[nob-1][nob-1].invUpperMult( vec_s );
		    vec_z = (*pA->getLUdecomposition())[nob-1][nob-1].invUnitLowerMult( vec_z );
		}

        //vec_t = S vec_z 
	    for ( j = 0; j < nob-1; j++ )
            vecBlocks[j] = (*pA)[j][nob-1] * vec_z;
			

		for ( j = 0; j < nob-1; j++ ) {
			vecBlocks[j] = (*pA)[j][j].invUpperMult( vecBlocks[j]);
            if ( innerGS == TWOBACKWARD )  // a second backward step
			{
	            vecTmp2[j] = (*pA)[j][j] * (vecBlocks[j]);
	            vecTmp2[j] = (*pA)[j][j].invUpperMult(vecTmp2[j]);
                vecBlocks[j] *= 2;
                vecBlocks[j] -= vecTmp2[j];
		    }

            if ( innerGS == SYMMETRIC ) // symmetric Gauss-Seidel
			{
			    vecBlocks[j] = (*pA)[j][j].diagMult( vecBlocks[j] );
			    vecBlocks[j] = (*pA)[j][j].invLowerMult(vecBlocks[j]);
			}
		}


        vec_t = (*pA)[nob-1][0] * vecBlocks[0];
		for ( j = 1; j < nob-1; j++ ) 
            vec_t += (*pA)[nob-1][j] * vecBlocks[j];

        beta = (vec_t * vec_t);
        if (fabs(beta) < 1e-15)
		{ 
           vec_y *= alpha;
           vec_x += vec_y;
           
           //return (vec_x);
		}
        else
		{
            omega = (vec_t * vec_s) / beta;
            vec_y *= alpha;
            vec_z *= omega;
            vec_x += vec_y;
            vec_x += vec_z;
            vec_t *= (-omega);
		}

        vec_r = vec_s + vec_t; 
        rhoOld = rhoNew;

        currentRes = (vec_r * vec_r) / vec_r.getSize();
        if (currentRes < bestRes)
		{
            bestRes = currentRes;
            vec_best_x = vec_x;
		}
	}
	if (i-1 == maxInnerSteps)
    {
        cout << i-1 << " bicg steps" << "  " << bestRes / initialRes << "  " << bestRes <<  " (" << vec_r.getSize() << " unknowns)" << endl;
        vec_x = vec_best_x;
	}
	//cout << "innerSteps: " << i << endl;
    return ( vec_x );
}


//
// calculate WInversScalar
//
// ( F B )  =  ( S B ) ( I      0 )
// ( C U )     ( 0 U ) ( U^{-1} I ),    S = F - B U^{-1} C
//
//  r = (a, b)^T
//
TBlockVector<TFloat> 
TSchur::WInversScalar ( const TBlockVector<TFloat> &r )
{ 
	uint i;
    uint startIndex, endIndex;
	TBlockVector<TFloat> vecTmp( r );

    if ( dimSchurComplement == 0 ) {
        cout << "Warning: Schur complement size is zero!\n";
        return (vecTmp);
	}

	TVectorBlock<TFloat> vecBlock ( 0, dimSchurComplement );
    startIndex = r.getSize() - 1;
    endIndex   = dimSchurComplement;

    // yy = U^{-1} b (symmetric/backward Gauss-Seidel)
   	for ( i = startIndex; i >= endIndex; i-- )
		vecTmp[0][i] = (*pA)[0][0][i].invUpperMult( vecTmp[0] );
   	//for ( i = startIndex; i >= endIndex; i-- )
	//	vecTmp[0][i] *= (*pA)[0][0][i].getValue( 0 );
   	//for ( i = endIndex; i <= startIndex; i++ )
	//	vecTmp[0][i] = (*pA)[0][0][i].invLowerMult( vecTmp[0] );
  
	// xx = S^{-1} ( a - B yy)
    for ( i = 0; i < endIndex; i++ )
        vecBlock[i] = vecTmp[0][i] - (*pA)[0][0][i].upperMult(r[0], dimSchurComplement-1);

    vecBlock = innerSchurScalar( vecBlock );

	// x = xx
    for ( i = 0; i < endIndex; i++ )
        vecTmp[0][i] = vecBlock[i];

    // y = yy - U^{-1} C x    
    vecBlock.setSize( startIndex - endIndex + 1 );

    for ( i = endIndex; i <= startIndex; i++ )
        vecBlock[i - endIndex] = vecTmp[0][i];

    for ( i = endIndex; i <= startIndex; i++ )
        vecTmp[0][i] = (*pA)[0][0][i].lowerMult( vecTmp[0], dimSchurComplement);

   	for ( i = startIndex; i >= endIndex; i-- )
		vecTmp[0][i] = (*pA)[0][0][i].invUpperMult( vecTmp[0] );
   	//for ( i = startIndex; i >= endIndex; i-- )
	//	vecTmp[0][i] *= (*pA)[0][0][i].getValue( 0 );
   	//for ( i = endIndex; i <= startIndex; i++ )
	//	vecTmp[0][i] = (*pA)[0][0][i].invLowerMult( vecTmp[0] );

    for ( i = endIndex; i <= startIndex; i++ ) {
        vecTmp[0][i] -= vecBlock[i - endIndex];
        vecTmp[0][i] *= (-1);
	}

	return (vecTmp);
}


//
// perform inner iterations to solve for Schur complement (bicgStab)
//
TVectorBlock<TFloat> 
TSchur::innerSchurScalar( TVectorBlock<TFloat> &b )
{
    uint i;
    uint j;
    TFloat rhoNew, rhoOld, alpha, beta, omega;
    rhoOld = 1.0; alpha = 1.0; omega = 1.0;
    TVectorBlock<TFloat> vecH ( (*pX)[0] );
    TVectorBlock<TFloat> vecX ( b );
    vecX.set( 0.0 );
    TVectorBlock<TFloat> vecR ( b );
    TVectorBlock<TFloat> vecR0 ( b );
    TVectorBlock<TFloat> vecV ( b );
    vecV.set( 0.0 );
    TVectorBlock<TFloat> vecP ( b );
    vecP.set( 0.0 );
    TVectorBlock<TFloat> vecT ( b );
    TVectorBlock<TFloat> vecY ( b );
    TVectorBlock<TFloat> vecS ( b );
    TVectorBlock<TFloat> vecZ ( b );

    for (i = 0; i < maxInnerSteps; i++) {

        rhoNew = vecR0 * vecR;
        if (fabs(rhoOld) < 10e-10 || fabs(omega) < 10e-10)
		{ 
            //cout << "breakdown0 " << i << " " << vecR * vecR << endl;
            return (vecX);
		}
        beta = rhoNew / rhoOld * alpha / omega;
        vecV *= omega;
        vecP -= vecV;
        vecP *= beta;
        vecP += vecR;

        vecY = vecP;

		// multiplication with S
        for ( j = 0; j < dimSchurComplement; j++ )
            vecH[j] = vecY[j];
		// calculate C vecX
        for ( j = dimSchurComplement; j < vecH.getSize(); j++)
            vecH[j] = (*pA)[0][0][j].lowerMult(vecH, dimSchurComplement);

        // calculate U^{-1} vecH (symmetric/backward Gauss-Seidel)
   	    for ( j = vecH.getSize()-1; j >= dimSchurComplement; j-- )
		    vecH[j] = (*pA)[0][0][j].invUpperMult( vecH );
   	    //for ( j = vecH.getSize()-1; j >= dimSchurComplement; j-- )
		//    vecH[j] *= (*pA)[0][0][j].getValue( 0 );
       	//for ( j = dimSchurComplement; j <= vecH.getSize()-1; j++ )
		//    vecH[j] = (*pA)[0][0][j].invLowerMult( vecH );

		// calculate B vecH
        for ( j = 0; j < dimSchurComplement; j++ )
            vecH[j] = (*pA)[0][0][j].upperMult(vecH, dimSchurComplement-1);
        for ( j = 0; j < dimSchurComplement; j++ )
            vecV[j] = (*pA)[0][0][j].getValue( 0 ) * vecY[j] - vecH[j];

        omega = vecR0 * vecV; 
        if (fabs(omega) < 10e-10)
		{ 
            //cout << "breakdown1 " << i << " " << vecR * vecR << endl;
            return (vecX);
		}

        alpha = rhoNew / omega;
        vecS = vecV;
        vecS *= (-alpha);
        vecS += vecR;
        vecZ = vecS;

		// multiplication with S
        for ( j = 0; j < dimSchurComplement; j++ )
            vecH[j] = vecZ[j];
		// calculate C vecX
        for ( j = dimSchurComplement; j < vecH.getSize(); j++)
            vecH[j] = (*pA)[0][0][j].lowerMult(vecH, dimSchurComplement);

        // calculate U^{-1} vecH (symmetric/backward Gauss-Seidel)
   	    for ( j = vecH.getSize()-1; j >= dimSchurComplement; j-- )
		    vecH[j] = (*pA)[0][0][j].invUpperMult( vecH );
   	    //for ( j = vecH.getSize()-1; j >= dimSchurComplement; j-- )
		//    vecH[j] *= (*pA)[0][0][j].getValue( 0 );
       	//for ( j = dimSchurComplement; j <= vecH.getSize()-1; j++ )
		//    vecH[j] = (*pA)[0][0][j].invLowerMult( vecH );

		// calculate B vecH
        for ( j = 0; j < dimSchurComplement; j++ )
            vecH[j] = (*pA)[0][0][j].upperMult(vecH, dimSchurComplement-1);
        for ( j = 0; j < dimSchurComplement; j++ )
            vecT[j] = (*pA)[0][0][j].getValue( 0 ) * vecZ[j] - vecH[j];

        beta = (vecT * vecT);
        if (fabs(beta) < 10e-10)
		{ 
            //cout << "breakdown2 " << i << " " << vecR * vecR << endl;
            return (vecX);
		}
        omega = ( vecT * vecS ) / beta;

        vecY *= alpha;
        vecX += vecY;
        vecZ *= omega;
        vecX += vecZ;
        vecT *= omega;
        vecR = vecS;
        vecR -= vecT;
        
        //if ( i==0 || i == innerSteps-1 )
		//    cout << "residual " << i << " " << vecR * vecR << endl; 

        rhoOld = rhoNew;
	}

    return( vecX );
}

//
// A = ( F B )
//     ( C U ) 
//
//  Schur complement S = F - B U^{-1} C, assuming that F is diagonal
//
void
TSchur::calcSchurComplement( )
{
    uint i, j, k;
    TFloat value;

    uint size = pX->getSize();
    TVectorBlock<TFloat> vecTmp ( 0.0, size );

    TMatrixBlock matTmp( dimSchurComplement );
    for ( k = 0; k < dimSchurComplement; k++ )
        matTmp[k].set( dimSchurComplement);

	// store first dimSchurComplement rows (F B) in vecTmpSparse
 	TSparseBlockVector vecTmpSparse[dimSchurComplement];
    for (i = 0; i < dimSchurComplement; i++) {
        pA->getTransposedRow( vecTmpSparse[i], 0, i );
		// assuming that the Schurblock is diagonal
		// (not always true, e.g. example with circles on regular grid)
        vecTmpSparse[i][0].setValue( 0, 0.0 );
	}

	// calculate U^{-1} C where U^{-1} is approximated by backward GS
    for (j = 0; j < dimSchurComplement; j++) {
        vecTmp = (*pA)[0][0].getColumn( j );
        for ( k = size - 1; k >= dimSchurComplement; k-- )
            vecTmp[k] = (*pA)[0][0][k].invUpperMult( vecTmp );
        for ( k = 0; k < dimSchurComplement; k++ )
            matTmp[k].setIndex(j, j, - (vecTmpSparse[k][0] * vecTmp) );
	}

	// switch so that diagonal entry is at the beginning
	// and add diagonal entry of Schurblock F
    for ( k = 0; k < dimSchurComplement; k++ ) {
        value = matTmp[k].getValue( k ) + (*pA)[0][0][k].getValue( 0 );
        matTmp[k].setIndex(k, 0, matTmp[k].getValue(0) );
        matTmp[k].setIndex(0, k, value);
	}

    cout << "Schur \n" << matTmp << endl;
}
