//
//  Project   : libnet
//  File      : solver.cc
//  Author    : Sabine Gutsch / Thomas Probst
//  Purpose   : virtual linear solver class
//
// arch-tag: 43ffc289-b6dd-4ee8-8e28-329fd92d04f7
//

extern "C" {
#include <stdio.h>
#include <math.h>
}

#include "defines.hh"
#include "TTerminal.hh"

#include "solver.hh"

//
// constructor
//
TSolver::TSolver ( TBlockMatrix *apA, 
                   TBlockVector<TFloat> *apRHS, 
				   TBlockVector<TFloat> *apX, 
                   uint amaxit, 
                   TFloat aaccur, 
                   uint averb,
                   uint amode )
{ 
    pA        = apA;
    pRHS      = apRHS;
    pX        = apX;
    maxit     = amaxit;
    accuracy  = aaccur;
    verbosity = averb;
    mode      = amode;
  
    // check dimension of matrix and rhs
    // (we assume that A is a n x n - blockmatrix)
    if ((pA != NULL) && (pRHS != NULL))
        checkpApRHS();

	parrConvRates = NULL;

    setName( "" );
}

//
// destructor 
//
TSolver::~TSolver () 
{
	if (parrConvRates != NULL)
		delete parrConvRates;
}

//
// get (set) name
//
const TString & 
TSolver::getName ()
{
	return name;
}

void 
TSolver::setName ( const TString & n )  
{
    name = n;
}


//
// step with explicitly specified system
//
int
TSolver::step ( int                       level,
                TBlockMatrix           ** A,
                TBlockVector< TFloat >  * RHS,
                TBlockVector< TFloat >  * x )
{
    pA   = A[level];
    pRHS = RHS;
    pX   = x;
        
    return step();
}

//
// solving method
//
void 
TSolver::solve () 
{
    TFloat  convRate = 0.0;
    
    init();

    // check dimension of matrix and rhs
    // (we assume that A is a n x n - blockmatrix)
    if ((pA != NULL) && (pRHS != NULL))
        checkpApRHS();
  
    uint  it = 0;
    if (verbosity & PRINT_SOLVERNAME )
        cout << getName() << "::solve : running (" << pX->getSize() << " unknowns)" << endl;
    
    uint                  returnvalue = RETURN_CONTINUE;
    TFloat                norm_old, norm_new, norm_prev;
    TBlockVector<TFloat>  defectVector;

    norm_old = norm_new = norm_prev = 0;
    
    if ( (mode == MODE_ACC_MAXIT) || (mode == MODE_REDUCEBY) 
         || (verbosity & PRINT_CONVERGENCE_HISTORY) 
         || (verbosity & PRINT_CONVERGENCE_RATE) )
    {
        // calculate defect's norm
        defectVector  = *pRHS;
        defectVector -= (*pA) * (*pX);
        
        // ignore very last component of defect (Lagrange multiplier)
        //if ( nob > 1 ) defectVector[nob - 1][dofs - 1] = 0.0;
        norm_old  = defectVector.weightedEuclidNorm( );
        norm_new  = norm_old;
        norm_prev = norm_old;
    }// if

    if (mode == MODE_REDUCEBY)
        accuracy = DEFAULT_REDUCEBY * norm_old;
    
    // call step according to mode criteria
    while ( (((mode == MODE_MAXIT) && (it++ < maxit))
             || (((mode == MODE_ACC_MAXIT) || (mode == MODE_REDUCEBY))
                 && (it++ < maxit) && (norm_new > accuracy)))
            && (returnvalue == RETURN_CONTINUE) )
    {
        returnvalue = step();
        
        if ( (mode == MODE_ACC_MAXIT) || (mode == MODE_REDUCEBY) 
             || (verbosity & PRINT_CONVERGENCE_HISTORY) )
        {
            defectVector  = *pRHS;
            defectVector -= (*pA) * (*pX);
            
            //if ( nob > 1 ) defectVector[nob - 1][dofs - 1] = 0.0;

            norm_new = defectVector.weightedEuclidNorm();
            
            if (verbosity & PRINT_CONVERGENCE_HISTORY)
            {
                printStep( it, norm_prev, norm_new );

                norm_prev = norm_new;

                if (parrConvRates == NULL) 
                    parrConvRates = new TDynArray<TFloat>( 0, 32 );
                
                parrConvRates->setSize( parrConvRates->getSize() + 1 );
                (*parrConvRates)[parrConvRates->getSize()-1] = norm_new / norm_prev;
            }// if
        }// if
    }// while
    
    if ( verbosity & PRINT_CONVERGENCE_RATE )
    {
        defectVector  = *pRHS;
        defectVector -= (*pA) * (*pX);
        
        //if ( nob > 1 ) defectVector[nob - 1][dofs - 1] = 0.0;
        norm_new = defectVector.weightedEuclidNorm( );
      
        if (it == maxit)
            cout << getName() << "::solve : maximal number of iterations reached"
                 << endl;
        
        if ((norm_new < accuracy) && (mode == MODE_ACC_MAXIT))
            cout << getName() << "::solve : desired accuracy reached" << endl;
        
        if ((norm_new < accuracy) && (mode == MODE_REDUCEBY))
            cout << getName() << "::solve : desired reduction in accuracy reached"
                 << endl;

        if ((it > 1) && (norm_old != 0))
            convRate = pow( (norm_new / norm_old), 1.0 / (it-1) );
        
        cout << getName() << "::solve : average convergence rate after "
             << it-1 << " steps = " << convRate << endl;
        
        cout.setf( ios::scientific, ios::floatfield );
        
        cout << getName() << "::solve : start accuracy   = " << norm_old << endl;
        cout << getName() << "::solve : reached accuracy = " << norm_new << endl;
        
        // cout.setf( 0, ios::floatfield );
    }// if
    
    if ((returnvalue == RETURN_BREAKDOWN) && 
        (verbosity & PRINT_BREAKDOWNS)) 
    {
        cout << getName() << "::solve : a breakdown occured after "
             << it - 1 << " steps" << endl;
    }// if
}

//
// first set rhs and then solve with zero initial vector
//
void 
TSolver::solve ( TBlockMatrix *newA, TBlockVector<TFloat> *newRHS, 
                 TBlockVector<TFloat> *newX ) 
{ 
    setA( newA );
    setRHS( newRHS );
    setX( newX );
  
    solve();
}

//
// multigrid solve
//
void
TSolver::solve ( int                     level,
                 TBlockMatrix         ** A,
                 TBlockVector<TFloat> ** RHS,
                 TBlockVector<TFloat> ** x )
{
    // only solve on fine grid
    solve( A[level], RHS[level], x[level] );
}

void 
TSolver::checkpApRHS()
{
    uint noOfRows = pA->getNoOfRows();
  
    if (noOfRows != pRHS->getNoOfBlocks())
        cerr << "TSolver::solve : blockstructure in A and RHS is not equal" << endl;
    else
    {
        // check the first diag-blocks
        for ( uint i = 0; i < noOfRows-1; i++ )
            if ((*pA)[i][i].getNoOfRows() != (*pRHS)[i].getSize())
                cerr << "TSolver::solve : number of rows in the " << i+1 << "'th blockrow "
                     << "is not equal (" << getName() << ")" << endl;
      
        // and in the last last compare with the first block
        if ((*pA)[noOfRows-1][0].getNoOfRows() != (*pRHS)[noOfRows-1].getSize())
            cerr << "TSolver::solve : number of rows in the " << noOfRows << "'th blockrow "
                 << "is not equal (" << getName() << ")" << endl;
    }// else
  
}


//
// set and get verbosity 
//
void 
TSolver::setVerbosity (const uint averb  ) 
{
    verbosity = averb;
}

uint 
TSolver::getVerbosity () const 
{
    return verbosity;
}

//
// set and get accuracy 
//
void 
TSolver::setAccuracy (const TFloat aaccur ) 
{
    accuracy = aaccur;
}

TFloat 
TSolver::getAccuracy () const 
{
    return (accuracy);
}
	
//
// set and get maxit
//
void 
TSolver::setMaxit ( const uint aMaxit ) 
{
    maxit = aMaxit;
}

uint 
TSolver::getMaxit () const  
{
    return maxit;
}

//
// initialize solver
//
void 
TSolver::init () {}

//
// set and get pA
//
void 
TSolver::setA ( TBlockMatrix *newA ) 
{
    pA = newA;
}

TBlockMatrix*
TSolver::getA () const  
{
    return (pA);
}

//
// set and get RHS
//
void 
TSolver::setRHS ( TBlockVector<TFloat>  *newRHS ) 
{
    pRHS = newRHS;
}

TBlockVector<TFloat>*
TSolver::getRHS () const
{
    return (pRHS);
}

//
// set and get x
//
void 
TSolver::setX (  TBlockVector<TFloat> *newx ) 
{
    pX = newx;
}

//	void setx ( TFloat f ) { pX->set( f ); }

TBlockVector<TFloat>*
TSolver::getX () const  
{
    return (pX);
}

//
// return array with conv. rates
//
TDynArray<TFloat>* 
TSolver::getConvRates ()
{
	return parrConvRates;
}

//
// print statistics
//

void
TSolver::printStep ( uint step, TFloat norm_old, TFloat norm )
{
    if ((verbosity & PRINT_CONVERGENCE_HISTORY) && (norm_old != 0))
    {
        if (TTerminal::columns() > 80)
            cout << getName() << "::solve : ";
        
        printf( "step %4d  |  norm  %12.8Le  |  reduction  %.6Lf \n",
                step, (long double)( norm ), (long double)(norm / norm_old) );
    }// if
}

void
TSolver::printResult ( uint steps, TFloat norm_0, TFloat norm )
{
    if ( verbosity & PRINT_CONVERGENCE_RATE )
    {
        TFloat  convRate;
        
        if (steps == maxit)
            cout << getName() << "::solve : maximal number of iterations reached"
                 << endl;
        
        if ((norm < accuracy) && (mode == MODE_ACC_MAXIT))
            cout << getName() << "::solve : desired accuracy reached" << endl;

        if ((mode == MODE_REDUCEBY) && (norm < (accuracy * norm_0)))
            cout << getName() << "::solve : desired reduction reached" << endl;
        
        if ((steps > 0) && (norm_0 != 0))
        {
            convRate = pow( (norm / norm_0), 1.0 / TFloat( steps ) );
        
            cout << getName() << "::solve : average convergence rate after "
                 << steps << " steps = " << convRate << endl;
        }// if
        
        cout.setf( ios::scientific, ios::floatfield );
        
        cout << getName() << "::solve : start accuracy   = " << norm_0 << endl;
        cout << getName() << "::solve : reached accuracy = " << norm << endl;
        
        // cout.setf( 0, ios::floatfield );
    }// if
}

//
// test stop-criteria
//
bool
TSolver::testStop ( uint it, TFloat norm_0, TFloat norm )
{
    if (it >= maxit)
        return TRUE;
    
    if ((mode == MODE_ACC_MAXIT) && (norm < accuracy))
        return TRUE;
    
    if ((mode == MODE_REDUCEBY) && (norm < (accuracy * norm_0)))
        return TRUE;

    return FALSE;
}
