//
//  Project :  libvisual
//  File    :  T3DPrinter.hh
//  Purpose :  baseclass for a file-based 3D output interface
//  Author  :  Ronald Kriemann
//
// arch-tag: 267c2c68-7fcc-45ba-b562-885fb66cc30b
//

#include "T3DPrinter.hh"

// indentation
#define INDENT_SPACE  "    "

/////////////////////////////////////////////////////////////
//
// TEdgeData
//

//
// constructor
//
TEdgeData::TEdgeData ()
        : pNode0(NULL), pNode1(NULL)
{}

TEdgeData::TEdgeData ( TVector<3> *v0, TVector<3> *v1 )
        : pNode0(v0), pNode1(v1)
{}

//
// access nodes
//
TVector<3> *
TEdgeData::node0 () const
{
    return pNode0;
}

TVector<3> *
TEdgeData::node1 () const
{
    return pNode1;
}
    
//
// some operators
//
TEdgeData &
TEdgeData::operator = ( const TEdgeData & data )
{
    pNode0 = data.pNode0;
    pNode1 = data.pNode1;

    return *this;
}

bool
TEdgeData::operator == ( const TEdgeData & data ) const
{
    if (pNode0 == data.pNode0)
    {
        if (pNode1 == data.pNode1)
            return TRUE;
        else
            return FALSE;
    }// if
    else if (pNode0 == data.pNode1)
    {
        if (pNode1 == data.pNode0)
            return TRUE;
        else
            return FALSE;
    }// else if
    else
        return FALSE;
}

//
// compare two edgedata objects
//
int
TEdgeCmp::compare ( const TEdgeData & e1, const TEdgeData & e2 )
{
    static TVector<3> cmp1, cmp2;
    
    if (e1 == e2)
        return CMP_EQUAL;
    else
    {
        cmp1  = * e1.node0();
        cmp1 += * e1.node1();
        
        cmp2  = * e2.node0();
        cmp2 += * e2.node1();

        if (cmp1.euclidNorm() > cmp2.euclidNorm())
            return CMP_BIGGER;
        else
            return CMP_SMALLER;
    }// else
}

/////////////////////////////////////////////////////////////
//
// T3DPrinter
//

//
// constructor
//
T3DPrinter::T3DPrinter ( const char *filename )
{
	_file = NULL;
	strcpy( _fileName, filename );

	_indentLevel = 0;
	_state       = UNDEF;

    _color.r = _color.g = _color.b = 0;
    _color.amb_r = _color.amb_g = _color.amb_b = 0;
    _color.filter = _color.transmit = 0;

    
    _font.font          = FONT_SANS;
    _font.size          = 0.1;
    _font.style         = STYLE_NONE;
    _font.width         = 0.01;
    _font.spacing       = 0;
    _font.justification = JUST_LEFT;
}

//
// destructor
//
T3DPrinter::~T3DPrinter ()
{
	if (_file != NULL)
        fclose( _file );
}

//
// access data
//
const int
T3DPrinter::indentLevel () const
{
    return _indentLevel;
}

const TStateType &
T3DPrinter::state () const
{
    return _state;
}
    
const TColorType &
T3DPrinter::color () const
{
    return _color;
}

const TFontType &
T3DPrinter::font () const
{
    return _font;
}

const TPolygonData &
T3DPrinter::polygonData () const
{
    return _polygonData;
}

//
// begin/end printing
//
void
T3DPrinter::begin ()
{
    _file = fopen( _fileName, "w" );

    _state = UNDEF;
        
    if (_file == NULL)
        cerr << "T3DPrinter::begin () : can not open file (" << _fileName << ")" 
             << endl;
    else
        _state = READY;
}

void
T3DPrinter::end ()
{
    if (_file != NULL)
    {
        _state = UNDEF;

        fclose( _file );
    }// if

    _file = NULL;
}

//
// indent commands
//
void
T3DPrinter::indent ()
{
    int  i;

    if (_state == UNDEF)
        return;
    
    if (_file != NULL)
    {
        for (i = 0; i < _indentLevel; i++)
            fputs( INDENT_SPACE, _file );
    }// if
}

void
T3DPrinter::indentUp ()
{
    _indentLevel++;
}

void
T3DPrinter::indentDown ()
{
    _indentLevel--;

    if (_indentLevel < 0)
        _indentLevel = 0;
}

void
T3DPrinter::indentWrite ( const char *string )
{
    if (_state == UNDEF)
        return;
    
    if (_file != NULL)
    {
        indent();
        fputs( string, _file );
        fputc( '\n', _file );
    }// if
}

//
// misc. draw routines
//
void
T3DPrinter::beginPolygon ( bool filled )
{
    _state = BUILD_POLY;

    _polygonData.filled     = filled;
    _polygonData.pNode0     = NULL;
    _polygonData.pNode1     = NULL;
    _polygonData.pFirstNode = NULL;
}

void
T3DPrinter::addNode ( TVector<3> *v )
{
    if (_state != BUILD_POLY)
    {
        cerr << "T3DPrinter::addNode (...) : not building a polygon" << endl;
        exit( 0 );
    }// if

    if (v == NULL)
    {
        cerr << "T3DPrinter::addNode (...) : gave NULL node" << endl;
        exit( 0 );
    }// if

    if (_polygonData.filled)
    {
        _polygonData.listNodes.add( *v );
    }// if
    else
    {
        if (_polygonData.pFirstNode == NULL)
            _polygonData.pFirstNode = v;
        
        if (_polygonData.pNode0 != NULL)
        {
            TEdgeData  edgeData( _polygonData.pNode0, v );
        
            if ( ! isInEdgeList( edgeData ))
            {
                treeEdges.insert( edgeData );
                drawLine( edgeData.node0(), edgeData.node1() );
            }// if
        }// if

        _polygonData.pNode0 = v;
    }// else
}

void
T3DPrinter::endPolygon ()
{
    if (_polygonData.filled)
    {
        drawPolygon();
        _polygonData.listNodes.clear();
    }// if
    else
    {
        // draw an edge from last to first node
        if (_polygonData.pFirstNode != NULL)
        {
            TEdgeData  edgeData( _polygonData.pNode0,
                                 _polygonData.pFirstNode );
        
            if ( ! isInEdgeList( edgeData ))
            {
                treeEdges.insert( edgeData );
                drawLine( edgeData.node0(), edgeData.node1() );
            }// if
        }// if
    }// else
    
    _state = READY;
}

void
T3DPrinter::drawCoordSystem ( TFloat length )
{
    TVector<3>  v0, v1;
    TColorType  oldColor = color();
    TFontType   oldFont = font();

    setFilter( 0.0 );

    setFont( 0.1, FONT_SANS, STYLE_NONE, 0.01, 0, JUST_LEFT );
        
    v0[X] = v0[Y] = v0[Z] = 0;

    setRGBColor( COLOR_DARKRED );
    v1[X] = length;
    v1[Y] = v1[Z] = 0;
    drawArrow( &v0, &v1, LINE_THICKNESS * 5, 0.8 );
    v1[X] += _font.size;
    drawText( &v1, "x" );
    
    setRGBColor( COLOR_DARKGREEN );
    v1[Y] = length;
    v1[X] = v1[Z] = 0;
    drawArrow( &v0, &v1, LINE_THICKNESS * 5, 0.8 );
    v1[X] -= _font.size;
    v1[Y] += _font.size;
    drawText( &v1, "y" );
    
    setRGBColor( COLOR_DARKBLUE );
    v1[Z] = length;
    v1[X] = v1[Y] = 0;
    drawArrow( &v0, &v1, LINE_THICKNESS * 5, 0.8 );
    v1[Z] += _font.size;
    drawText( &v1, "z" );

    setRGBColor( COLOR_BLACK );
    v0[X] = 1; v0[Y] = 0; v0[Z] = 0;
    v1[X] = 1; v1[Y] = 1; v1[Z] = 0;
    drawLine( &v0, &v1 );
    v0[X] = 1; v0[Y] = 0; v0[Z] = 0;
    v1[X] = 1; v1[Y] = 0; v1[Z] = 1;
    drawLine( &v0, &v1 );

    v0[X] = 0; v0[Y] = 1; v0[Z] = 0;
    v1[X] = 0; v1[Y] = 1; v1[Z] = 1;
    drawLine( &v0, &v1 );
    v0[X] = 0; v0[Y] = 1; v0[Z] = 0;
    v1[X] = 1; v1[Y] = 1; v1[Z] = 0;
    drawLine( &v0, &v1 );
        
    v0[X] = 0; v0[Y] = 0; v0[Z] = 1;
    v1[X] = 0; v1[Y] = 1; v1[Z] = 1;
    drawLine( &v0, &v1 );
    v0[X] = 0; v0[Y] = 0; v0[Z] = 1;
    v1[X] = 1; v1[Y] = 0; v1[Z] = 1;
    drawLine( &v0, &v1 );

    v0[X] = 1; v0[Y] = 1; v0[Z] = 1;
    v1[X] = 0; v1[Y] = 1; v1[Z] = 1;
    drawLine( &v0, &v1 );
    v0[X] = 1; v0[Y] = 1; v0[Z] = 1;
    v1[X] = 1; v1[Y] = 0; v1[Z] = 1;
    drawLine( &v0, &v1 );
    v0[X] = 1; v0[Y] = 1; v0[Z] = 1;
    v1[X] = 1; v1[Y] = 1; v1[Z] = 0;
    drawLine( &v0, &v1 );

    setColor( oldColor );
    setFont( oldFont );
}

void
T3DPrinter::drawArrow ( TVector<3> *orig, TVector<3> *dest,
                        TFloat radius, TFloat lengthArrow )
{
    TVector<3>  dir, endCyl;

    if ((orig == NULL) || (dest == NULL))
    {
        cerr << "T3DPrinter::drawArrow (...) : gave NULL node" << endl;
        return;
    }// if

    dir  = *dest;
    dir -= *orig;
    dir *= lengthArrow;

    endCyl  = *orig;
    endCyl += dir;
    
    drawCylinder( orig, &endCyl, radius );

    drawCone( &endCyl, radius * 2.5, dest, 0 );
}
    
//
// misc. style settings
//
void
T3DPrinter::setRGBColor ( uint rgb, uint type )
{
    TFloat  r,g,b;

    r = TFloat((rgb & 0xFF0000) >> 16) / TFloat(0xFF);
    g = TFloat((rgb & 0x00FF00) >> 8)  / TFloat(0xFF);
    b = TFloat( rgb & 0x0000FF)        / TFloat(0xFF);

    setRGBColor( r, g, b, type );
}

void
T3DPrinter::setRGBColor ( TFloat r, TFloat g, TFloat b, uint type )
{
    if (type == PIGMENT)
    {
        _color.r = r;
        _color.g = g;
        _color.b = b;
    }// if
    else if (type == AMBIENT)
    {
        _color.amb_r = r;
        _color.amb_g = g;
        _color.amb_b = b;
    }// if
    else
        cout << "T3DPrinter::setRGBColor (...) : unsupported type" << endl;

    checkColor( _color );
}

void
T3DPrinter::setHSVColor ( TFloat h, TFloat s, TFloat v, uint type )
{
    int  loc_h, loc_s, loc_v;
    int  r = 0, g = 0, b = 0;

    if (h != -1)
        loc_h = (loc_h = (h > 360 ? 360 : int( h ))) < 0 ? 0 : loc_h;
    else
        loc_h = -1;
    
    loc_s = (loc_s = (s > 1 ? 255 : int( 255*s ))) < 0 ? 0 : loc_s;
    loc_v = (loc_v = (v > 1 ? 255 : int( 255*v ))) < 0 ? 0 : loc_v;

    if (( loc_s == 0 ) || ( loc_h == -1 ))
    {
        r = loc_v;
        g = loc_v;
        b = loc_v;
    }// if
    else
    {
        if ( uint( loc_h ) >= 360 )
            loc_h %= 360;
        
        uint f = loc_h % 60;
        loc_h /= 60;
        
        uint p = (uint)(2*loc_v*(255-loc_s)+255)/510;
        uint q, t;
        
        if ( loc_h & 1 )
        {
            q = (uint)(2*loc_v*(15300-loc_s*f)+15300)/30600;

            switch( loc_h )
            {
                case 1: r=(int)q;     g=(int)loc_v, b=(int)p; break;
                case 3: r=(int)p;     g=(int)q,     b=(int)loc_v; break;
                case 5: r=(int)loc_v; g=(int)p,     b=(int)q; break;
            }// switch
        }// if
        else
        {
            t = (uint)(2*loc_v*(15300-(loc_s*(60-f)))+15300)/30600;

            switch( loc_h )
            {
                case 0: r = (int)loc_v; g = (int)t,     b = (int)p;     break;
                case 2: r = (int)p;     g = (int)loc_v, b = (int)t;     break;
                case 4: r = (int)t;     g = (int)p,     b = (int)loc_v; break;
            }// switch
        }// else
    }// else

    if (type == PIGMENT)
    {
        _color.r = TFloat(r) / TFloat(255);
        _color.g = TFloat(g) / TFloat(255);
        _color.b = TFloat(b) / TFloat(255);
    }// if
    else if (type == AMBIENT)
    {
        _color.amb_r = TFloat(r) / TFloat(255);
        _color.amb_g = TFloat(g) / TFloat(255);
        _color.amb_b = TFloat(b) / TFloat(255);
    }// if
    else
        cout << "T3DPrinter::setHSVColor (...) : unsupported type" << endl;

    checkColor( _color );
}

void
T3DPrinter::setColor ( const TColorType & c )
{
    _color = c;
}

void
T3DPrinter::setFilter ( TFloat filter )
{
    _color.filter = filter;

    if (_color.filter > 1)
        _color.filter = 1;

    if (_color.filter < 0)
        _color.filter = 0;
}

void
T3DPrinter::setFont ( TFloat size, int font, int style,
                      TFloat width, TFloat spacing,
                      int justification )
{
    _font.size    = size;

    if (font != -1)
        _font.font = font;

    if (style != -1)
        _font.style = style;
    
    if (width != -1)
        _font.width = width;
    
    if (spacing != -1)
        _font.spacing = spacing;

    if (justification != -1)
        _font.justification = justification;
}

void
T3DPrinter::setFont ( const TFontType & f )
{
    _font = f;
}

//
// transform coord into local representation
//
TFloat
T3DPrinter::transX ( TVector<3> * v ) const
{
    if (v == NULL)
    {
        cout << "T3DPrinter::transX (...) : got NULL arg" << endl;
        return 0;
    }// if
    
    return (*v)[X];
}

TFloat
T3DPrinter::transY ( TVector<3> * v ) const
{
    if (v == NULL)
    {
        cout << "T3DPrinter::transY (...) : got NULL arg" << endl;
        return 0;
    }// if
    
    return (*v)[Z];
}

TFloat
T3DPrinter::transZ ( TVector<3> * v ) const
{
    if (v == NULL)
    {
        cout << "T3DPrinter::transZ (...) : got NULL arg" << endl;
        return 0;
    }// if
    
    return (*v)[Y];
}

void
T3DPrinter::trans ( TVector<3> * v ) const
{
    if (v == NULL)
    {
        cerr << "T3DPrinter::trans (...) : got NULL arg" << endl;
        return;
    }// if

    TVector<3>  tmp( *v );
    
    (*v)[X] = transX( & tmp );
    (*v)[Y] = transY( & tmp );
    (*v)[Z] = transZ( & tmp );
}

//
// look if the given edge is in the edgelist
//
bool
T3DPrinter::isInEdgeList ( TEdgeData & edge )
{
    return treeEdges.isIn( edge );
}

//
// correct given color def.
//
void
T3DPrinter::checkColor ( TColorType & c )
{
    c.r = c.r > 1.0 ? 1.0 : c.r;
    c.r = c.r < 0.0 ? 0.0 : c.r;

    c.g = c.g > 1.0 ? 1.0 : c.g;
    c.g = c.g < 0.0 ? 0.0 : c.g;

    c.b = c.b > 1.0 ? 1.0 : c.b;
    c.b = c.b < 0.0 ? 0.0 : c.b;


    c.amb_r = c.amb_r > 1.0 ? 1.0 : c.amb_r;
    c.amb_r = c.amb_r < 0.0 ? 0.0 : c.amb_r;

    c.amb_g = c.amb_g > 1.0 ? 1.0 : c.amb_g;
    c.amb_g = c.amb_g < 0.0 ? 0.0 : c.amb_g;

    c.amb_b = c.amb_b > 1.0 ? 1.0 : c.amb_b;
    c.amb_b = c.amb_b < 0.0 ? 0.0 : c.amb_b;
}
