//
//  Project :  libvisual
//  File    :  streakline.cc
//  Purpose :  class for streakline visualization
//  Author  :  Ronald Kriemann
//
// arch-tag: 0fba5152-c0ba-40eb-99a9-48fc425426f8
//

#include "streakline.hh"

/////////////////////////////////////////////////
//
// class for a particle
//

//
// constructor
//
TParticle::TParticle ( const T2DVector &pos )
{
	vecPos = pos;
}

//
// return pos.
//
const T2DVector& 
TParticle::getPos () const
{
	return vecPos;
}

//
// compute next pos. of particle
//
TParticle*
TParticle::calcNextStep ( TFloat deltaT, T2DNet &net,
						  TBlockVector<TFloat> &arrVelocity,
						  T2DBasisEqn &eqn )
{
	uint        i;
	T2DCell    *pInCell = NULL;

	//
	// find the cell this particle is in
	//
	T2DVector               vecNode0, vecNode1, vecNode2;
	uint                    noOfFrontiers;
	TListIterator<T2DCell>  iterCells( net.getList() );
	T2DCell                *pCell;

	TFloat    x;
	uint      crossings;

	iterCells.first();
	while ((pCell = iterCells++) != NULL)
	{
		//
		// count the numbers of crossed edges, while going along the x-axis
		// (the particle was set in the origin)
		//
		crossings = 0;
		noOfFrontiers = pCell->getNoOfFrontiers();
		for (i = 0; i < noOfFrontiers; i++)
		{
			vecNode0  = *(pCell->getFrontier( i )->getNode( pCell, 0 ));
			vecNode0 -= vecPos;

			vecNode2  = *(pCell->getFrontier( i )->getNode( pCell, 2 ));
			vecNode2 -= vecPos;

			if (((vecNode2[Y] > 0) && (vecNode0[Y] <= 0)) ||
				((vecNode0[Y] > 0) && (vecNode2[Y] <= 0))) {
				x = (vecNode0[X] * vecNode2[Y] - vecNode2[X] * vecNode0[Y]) /
					(vecNode2[Y] - vecNode0[Y]);
				
				if (x > 0)
					crossings++;
			}// if
		}// for

		// if particle is inside cell : crossings is odd
		if ((crossings % 2) == 1)
		{
			pInCell = pCell;
			iterCells.last();
			iterCells++;
		}// if
	}// while

	vecNode0 = *(pInCell->getFrontier( 0 )->getNode( pInCell, 0 ));
	vecNode1 = *(pInCell->getFrontier( 1 )->getNode( pInCell, 0 ));
	vecNode2 = *(pInCell->getFrontier( 2 )->getNode( pInCell, 0 ));

	vecNode1 -= vecNode0;
	vecNode2 -= vecNode0;

	//
	// project vecPos into <vecNode1,vecNode2>
	//
	TFloat  det;
	TFloat  t, length, projLength1, projLength2;

	det = (-1.0 * vecNode1[Y] * vecNode1[Y]) - (vecNode1[X] * vecNode1[X]);

	t = ((vecNode1[X] / det) * (vecNode0[X]-vecPos[X])) +
		((vecNode1[Y] / det) * (vecNode0[Y]-vecPos[Y]));

	length = vecNode1.euclidNorm();
	vecNode1 *= t;
	projLength1 = vecNode1.euclidNorm();

	//	cout << "length = " << length << endl << "projlength = " << projLength1 << endl;

	det = (-1.0 * vecNode2[Y] * vecNode2[Y]) - (vecNode2[X] * vecNode2[X]);

	t = ((vecNode2[X] / det) * (vecNode0[X]-vecPos[X])) +
		((vecNode2[Y] / det) * (vecNode0[Y]-vecPos[Y]));

	length = vecNode2.euclidNorm();
	vecNode2 *= t;
	projLength2 = vecNode2.euclidNorm();

	//	cout << "length = " << length << endl << "projlength = " << projLength2 << endl;

	//
	// interpolate velocity at vecPos (here only for triangles)
	//

	uint  noOfInteriorNodes = net.getNoOfInteriorNodes();
	TFloat interU, interV;

	// velocities in all 3 corners
	TFloat  u0, u1, u2, v0, v1, v2;

	uint nodeName;

	nodeName = pInCell->getFrontier( 0 )->getNode( pInCell, 0 )->getName();
	if (nodeName <= noOfInteriorNodes) {
		u0 = arrVelocity[0][nodeName]; 
		v0 = arrVelocity[1][nodeName];
	}// if
	else {
		u0 = eqn.dirichlet0( vecPos ); 
		v0 = eqn.dirichlet1( vecPos );
	}// else

	nodeName = pInCell->getFrontier( 1 )->getNode( pInCell, 0 )->getName();
	if (nodeName <= noOfInteriorNodes) {
		u1 = arrVelocity[0][nodeName]; 
		v1 = arrVelocity[1][nodeName];
	}// if
	else {
		u1 = eqn.dirichlet0( vecPos ); 
		v1 = eqn.dirichlet1( vecPos );
	}// else

	nodeName = pInCell->getFrontier( 2 )->getNode( pInCell, 0 )->getName();
	if (nodeName <= noOfInteriorNodes) {
		u2 = arrVelocity[0][nodeName]; 
		v2 = arrVelocity[1][nodeName];
	}// if
	else {
		u2 = eqn.dirichlet0( vecPos ); 
		v2 = eqn.dirichlet1( vecPos );
	}// else

	// interpolate velocity in the triangle
	interU = (1.0 - projLength1 - projLength2)*u0 + projLength1*u1 + projLength2*u2;
	interV = (1.0 - projLength1 - projLength2)*v0 + projLength1*v1 + projLength2*v2;

	//	cout << "(" << u0 << "," << v0 << ")" << endl;
	//	cout << "(" << u1 << "," << v1 << ")" << endl;
	//	cout << "(" << u2 << "," << v2 << ")" << endl;
	//	cout << "u = " << interU << "     v = " << interV << endl;

	TFloat  newX = vecPos[X] + (interU * deltaT);
	TFloat  newY = vecPos[Y] + (interV * deltaT);

	//	cout << "newX = " << newX << " newY = " << newY << endl;

	// check if coord exceed boundary
	if ((newX > 1.0) || (newX < 0.0) || (newY > 1.0) || (newY < 0.0))
		return NULL;
	else
		return new TParticle( T2DVector( newX, newY ));
}

/////////////////////////////////////////////////
//
//
// class for a particletrace
//

//
// constructor
//
TParticleTrace::TParticleTrace ( TParticle &particle,
								 T2DNet &net, TBlockVector<TFloat> &velocity,
								 T2DBasisEqn &eqn )
{
	pNet         = &net;
	parrVelocity = &velocity;
	pEqn         = &eqn;

	listParticles.add( particle );
}

//
// destructor
//
TParticleTrace::~TParticleTrace ()
{
	TListIterator<TParticle>  iterParticles( listParticles );
	TParticle                *pParticle;

	iterParticles.first();
	while ((pParticle = iterParticles++) != NULL)
		delete pParticle;

	listParticles.clear();
}

//
// return list
//
const TList<TParticle>& 
TParticleTrace::getList () const
{
	return listParticles;
}

//
// compute next pos. of particle
//
void 
TParticleTrace::calcNextStep ( TFloat deltaT )
{
	TParticle *pParticle = listParticles.getLast();
	TParticle *pNextParticle = NULL;

	if (pParticle != NULL)
		pNextParticle = pParticle->calcNextStep( deltaT, *pNet, *parrVelocity, *pEqn );

	if (pNextParticle != NULL)
		listParticles.add( *pNextParticle );
}

/////////////////////////////////////////////////
//
// class for streakline visualization
//

//
// constructor
//
TStreakLineView::TStreakLineView ( T2DNet &net, TBlockVector<TFloat> &velocity,
								   T2DBasisEqn &eqn )
{
	// create initial particletraces
	TParticle       *pParticle;
	TParticleTrace  *pTrace;
	TFloat           stepWidth = 0.02, s1, s2;

	for (s1 = stepWidth; s1 < 1.0; s1 += stepWidth)
	{
		for (s2 = stepWidth; s2 < 1.0; s2 += stepWidth)
		{
			pParticle = new TParticle( T2DVector( s1, s2 ) );
			pTrace    = new TParticleTrace( *pParticle, net, velocity, eqn );
			listTraces.add( *pTrace );
		}// for
	}// for
}

//
// destructor
//
TStreakLineView::~TStreakLineView ()
{
	TListIterator<TParticleTrace>  iterTraces( listTraces );
	TParticleTrace                *pTrace;

	iterTraces.first();
	while ((pTrace = iterTraces++) != NULL)
		delete pTrace;

	listTraces.clear();
}

//
// build traces
//
void 
TStreakLineView::buildTraces ( TFloat deltaT, uint intervals )
{
	TListIterator<TParticleTrace>  iterTraces( listTraces );
	TParticleTrace                *pTrace;

	for (uint i = 0; i < intervals; i++)
	{
		iterTraces.first();
		while ((pTrace = iterTraces++) != NULL)
			pTrace->calcNextStep( deltaT );
	}// for
}

//
// draw traces on screen
//
void 
TStreakLineView::draw ( TPrinter& p )
{
	TListIterator<TParticleTrace>  iterTraces( listTraces );
	TParticleTrace                *pTrace;

	TListIterator<TParticle>       iterParticles;
	TParticle                     *pLastParticle, *pParticle;

	T2DVector                      vecNode1, vecNode2;

	// draw the boundary
	p.drawLine( 0, 0, 1, 0 );
	p.drawLine( 1, 0, 1, 1 );
	p.drawLine( 1, 1, 0, 1 );
	p.drawLine( 0, 1, 0, 0 );

	// and now the traces
	iterTraces.first();
	while ((pTrace = iterTraces++) != NULL)
	{
		iterParticles.setList( pTrace->getList() );

		pLastParticle = iterParticles.first();
		iterParticles++;

		while ((pParticle = iterParticles++) != NULL)
		{
			vecNode1 = pLastParticle->getPos();
			vecNode2 = pParticle->getPos();

			p.drawLine( vecNode1[X], vecNode1[Y],
						vecNode2[X], vecNode2[Y] );

			pLastParticle = pParticle;
		}// while
	}// while
}
