//
//  Project :  libvisual
//  File    :  vectorfield.cc
//  Purpose :  class for visualization of a vectorfield
//  Author  :  Ronald Kriemann / Sabine Gutsch
//
// arch-tag: d0fcd513-673b-4d9b-bb9d-6e8f8a69c360
//

#include <stdlib.h>
#include <unistd.h>

#include <math.h>

#include "cell.hh"

#include "vectorfield.hh"

// cos and sin of the angle for the arrow (20 deg)
#define COSANGLE  0.9396936208
#define SINANGLE  0.3420201433

// a needed type shortcut
typedef TCell<2>  T2DCell;

//
// constructor
//
TVectorField::TVectorField ( TDynArray< TVector<2> * >& vertices,
							 TBlockVector<TFloat>& solution,
							 const T2DNet& net,
							 bool scale )
{
	parrVertices = &vertices;
	pvecSolution = &solution;

	bScale = scale;

	// search first two blocks (u and v) for the greatest value 
	// to set scaling factor
	maxSolution = 0;
	uint i, size;

	size = solution[0].getSize();
	for (i = 0; i < size; i++)
	{
		if (fabs(solution[0][i]) > maxSolution)
			maxSolution = fabs(solution[0][i]);
		if (fabs(solution[1][i]) > maxSolution)
			maxSolution = fabs(solution[1][i]);
	}// for

	// get the min (max) distance between two vertices
	TListIterator<T2DCell>  iterCells( net.getList() );
	T2DCell                *pCell;
	TFloat                  distV0_V1, distV1_V2, distV2_V0;

	minDistNodes = 1.0;
	maxDistNodes = 0.0;

	iterCells.first();
	while ((pCell = iterCells++) != NULL)
	{
		distV0_V1 = ((*(pCell->getNode(0))) - (*(pCell->getNode(1)))).euclidNorm();
		distV1_V2 = ((*(pCell->getNode(1))) - (*(pCell->getNode(2)))).euclidNorm();
		distV2_V0 = ((*(pCell->getNode(2))) - (*(pCell->getNode(0)))).euclidNorm();
		
		if (minDistNodes > distV0_V1) minDistNodes = distV0_V1;
		if (minDistNodes > distV1_V2) minDistNodes = distV1_V2;
		if (minDistNodes > distV2_V0) minDistNodes = distV2_V0;

		if (maxDistNodes < distV0_V1) maxDistNodes = distV0_V1;
		if (maxDistNodes < distV1_V2) maxDistNodes = distV1_V2;
		if (maxDistNodes < distV2_V0) maxDistNodes = distV2_V0;
	}// while
}

//
// draw the vectorfield
// (draw at most 1000 arrows, which are choosen randomly)
//
void 
TVectorField::draw ( TPrinter& p )
{
	T2DVector *vector;
	uint      name = 0;
	
	TFloat x, y, u, v, su, sv, ru, rv;
	
	uint size     = parrVertices->getSize();
	int  maxIndex = int((TFloat(size) / 1000.0) + 0.5);
	int  index;

	if (maxIndex < 0) maxIndex = 0;

	// init. random-number-gen
	srand( (int) getpid() );

	p.setRGBColor( COLOR_BLACK );
	p.setLineWidth( 0 );

	for (uint i = 0; i < size; i++)
	{
		// draw only if the max. index was generated
		index = int((float(rand()) / float(RAND_MAX)) * maxIndex);
		if ((size <= 1000) || (index == maxIndex - 1))
		{
			vector = (*parrVertices)[i];

			// beginning
			x = (*vector)[X];
			y = (*vector)[Y];

			// direction (we divide by 2 because we center the vector)
			if (bScale == TRUE) {
				u = (maxDistNodes * (*pvecSolution)[0][name]) / 2.0;
				v = (maxDistNodes * (*pvecSolution)[1][name]) / 2.0;
			}// if
			else {
				u = (*pvecSolution)[0][i];
				v = (*pvecSolution)[1][i];

				TFloat l = sqrt(u*u+v*v);
				if (l > 0.0) {
					u /= l;
					v /= l;
				}// if
				else {
					u = 0.0;
					v = 0.0;
				}// else

				u *= (maxDistNodes * maxSolution);
				v *= (maxDistNodes * maxSolution);
			}// else
		
			// the direction of the vector
			p.drawLine( x, y, x+u, y+v );
		
			// show direction
			su = - u / 2.75;
			sv = - v / 2.75;
		
			ru = su * COSANGLE - sv * SINANGLE;
			rv = su * SINANGLE + sv * COSANGLE;
			p.drawLine( x+u, y+v, x+u + ru, y+v + rv );
		
			ru =   su * COSANGLE + sv * SINANGLE;
			rv = - su * SINANGLE + sv * COSANGLE;
			p.drawLine( x+u, y+v, x+u + ru, y+v + rv );
		}// if
	}// for
}
